
#include "etcpip.h"
#define __KERNEL__
#include "UsbCfg.h"
#include "PoList.h"
//#include "PoTimer.h"
#include "UsMemLib.h"
#include "UsOsport.h"
#include "PoDevice.h"
#include "UsbPoPci.h"
//#include "usb.h"
//#include "UsbHcd.h"
//#include "hcdx.h"


/*
  enable this definition will enable EHCI activities
*/
#define USB_ENABLE_VT6212_EHCI

/*
  enable this definition will enable UHCI activities
*/
#define USB_ENABLE_VT6212_UHCI



extern volatile int usb_hub_init_end;
extern int hcd_init_ok;


enum pci_function_enum {
	FUNCTION_UHCI_1 = 0,
	FUNCTION_UHCI_2,
	FUNCTION_EHCI,

	FUNCTION_TOTAL
};

static unsigned char vt6212_software_interrupt_disabled;
static unsigned char vt6212_pending_irq[FUNCTION_TOTAL];

#define PENDING_IRQ_EXIST()			(vt6212_pending_irq[FUNCTION_UHCI_1] \
									| vt6212_pending_irq[FUNCTION_UHCI_2] \
									| vt6212_pending_irq[FUNCTION_EHCI])

#define TASK_SLEEP(tick_cnt)	dlytsk(cur_task, DLY_TICKS, tick_cnt)

#define	BLOCK_IRQ()				{ MASK_INTS(); }
#define UNBLOCK_IRQ()			{ UNMASK_INTS(); }

static unsigned long					vt6212_locked_task_id, vt6212_locked_task_reentry;
#define SAVE_AND_ENABLE_PREEMPTION(n)	{n = mt_busy; mt_busy = 0;}
#define SET_PREEMPTION_TO(n)			{mt_busy = n;}

static struct device vt6212_dev[FUNCTION_TOTAL];
#ifdef NO_SYS_FS
static struct device_driver localhc_dev_driver[FUNCTION_TOTAL];
#ifdef USB_DBGSTRING_ON
char localhc_dev_driver_name[FUNCTION_TOTAL][] = { "uhci-driver", "uhci-driver", "ehci-driver" };
#endif // USB_DBGSTRING_ON
#endif // NO_SYS_FS

static int vt6212_pci_device_id[FUNCTION_TOTAL], vt6212_pci_device_irq[FUNCTION_TOTAL];
static unsigned char vt6212_pci_device_irq_init[FUNCTION_TOTAL];
static unsigned long vt6212_pci_device_addr[FUNCTION_TOTAL];
static unsigned long memory_addr_in_cpu, memory_addr_in_pci;
long cpu_to_usb_addr_shift, usb_to_cpu_addr_shift;

extern const struct pci_device_id uhci_pci_ids [];
extern const struct pci_device_id ehci_pci_ids [];
static const struct pci_device_id *pci_ids[FUNCTION_TOTAL] =
									{ uhci_pci_ids, uhci_pci_ids, ehci_pci_ids };

/* -------------------------------------------------------------------- */
void usb_hc_task_lock(void) {
	unsigned char	getres_required;

	getres_required = 1;
	BLOCK_IRQ();
	if (cur_task == vt6212_locked_task_id)
		getres_required = 0;
	UNBLOCK_IRQ();

	if (getres_required) {
		unsigned long	current_preempt_status;
		SAVE_AND_ENABLE_PREEMPTION(current_preempt_status);
		while (getres(MT_RES_USB_HW_DEVICE, 0xffffffff) != SUCCESS) {
			TASK_SLEEP(1);
		}
		SET_PREEMPTION_TO(current_preempt_status);
		BLOCK_IRQ();
		vt6212_locked_task_id = cur_task;
		UNBLOCK_IRQ();
	}
	vt6212_locked_task_reentry ++;
}

void usb_hc_task_unlock(void) {
	if (vt6212_locked_task_reentry > 0) {
		vt6212_locked_task_reentry --;
		if (vt6212_locked_task_reentry == 0) {
			BLOCK_IRQ();
			vt6212_locked_task_id = 0xffffffff;
			UNBLOCK_IRQ();
			relres(MT_RES_USB_HW_DEVICE);
		}
	}
}

/* -------------------------------------------------------------------- */

#define PCI_COMMAND		0x04	/* 16 bits */
#define  PCI_COMMAND_INVALIDATE	0x10	/* Use memory write and invalidate */

int
pci_set_mwi(struct pci_dev *dev)
{
	int	function_idx;
	u16 cmd;

	function_idx = ((struct device *)dev)->dev_idx_for_pci_access;
	cmd = PciCfgRdWord( vt6212_pci_device_id[function_idx], PCI_COMMAND );
	if (! (cmd & PCI_COMMAND_INVALIDATE)) {
		cmd |= PCI_COMMAND_INVALIDATE;
		PciCfgWrWord( vt6212_pci_device_id[function_idx], PCI_COMMAND, cmd );
	}
	return 0;
}

int
//pci_set_dma_mask(struct pci_dev *dev, u64 mask)
pci_set_dma_mask(struct pci_dev *dev, u32 mask)
{
	return -EIO;
}

int pci_read_config_dword(struct pci_dev *dev, int where, u32 *val)
{
	(*val) = PciCfgRdDword( vt6212_pci_device_id[((struct device *)dev)->dev_idx_for_pci_access], where );
	return 0;
}

int pci_read_config_word(struct pci_dev *dev, int where, u16 *val)
{
	(*val) = PciCfgRdWord( vt6212_pci_device_id[((struct device *)dev)->dev_idx_for_pci_access], where );
	return 0;
}

int pci_read_config_byte(struct pci_dev *dev, int where, u8 *val)
{
	(*val) = PciCfgRdByte( vt6212_pci_device_id[((struct device *)dev)->dev_idx_for_pci_access], where );
	return 0;
}

int pci_write_config_dword(struct pci_dev *dev, int where, u32 val)
{
	PciCfgWrDword( vt6212_pci_device_id[((struct device *)dev)->dev_idx_for_pci_access], where, val );
	return 0;
}

int pci_write_config_word(struct pci_dev *dev, int where, u16 val)
{
	PciCfgWrWord( vt6212_pci_device_id[((struct device *)dev)->dev_idx_for_pci_access], where, val );
	return 0;
}

int pci_write_config_byte(struct pci_dev *dev, int where, u8 val)
{
	PciCfgWrByte( vt6212_pci_device_id[((struct device *)dev)->dev_idx_for_pci_access], where, val );
	return 0;
}

unsigned int pci_read_io_dword(void *dev, unsigned int where)
{
	return *((unsigned int *)where);
}

unsigned short pci_read_io_word(void *dev, unsigned int where)
{
#ifdef _ATHEROS
	return (unsigned short)(*((unsigned int *)(where & 0xfffffffc)) >> ((where & 3) << 3));
//	return *((unsigned short *)(where ^ 2));
#else
	return *((unsigned short *)where);
#endif
}


unsigned char pci_read_io_byte(void *dev, unsigned int where)
{
#ifdef _ATHEROS
	return	(unsigned char)(*((unsigned int *)(where & 0xfffffffc)) >> ((where & 3) << 3));
//	return *((unsigned char *)(where ^ 3));
#else
	return *((unsigned char *)where);
#endif
}

void pci_write_io_dword(void *dev, unsigned int where, u32 val)
{
	*((u32 *)where) = val;
}

void pci_write_io_word(void *dev, unsigned int where, u16 val)
{
#ifdef _ATHEROS
	*((u16 *)(where ^ 2)) = val;
#else
	*((u16 *)where) = val;
#endif
}

int usb_hc_polling_irq_idx(void *dev_base_mem)
{
	int		idx;

	idx = 0;
	while ( ((unsigned long)dev_base_mem != vt6212_pci_device_addr[idx])
			&& (idx < (int)FUNCTION_TOTAL) )
		idx ++;
	if (idx >= (int)FUNCTION_TOTAL) {
		dprintf("[USB HC] polling unknown device - x%08x\n", dev_base_mem);
		idx = -1;
	}

	return	idx;
}

void usb_hc_polling_irq(int dev_idx, void *urb)
{
	int		irq_active;
	unsigned long	flags;
	unsigned char	pending_irq;

	/* dev_idx should be obtained from usb_hc_polling_irq_idx() */
	irq_active = 1;

	BLOCK_IRQ();
	flags = vt6212_software_interrupt_disabled;
	vt6212_software_interrupt_disabled = 1;
	pending_irq = vt6212_pending_irq[dev_idx];
	vt6212_pending_irq[dev_idx] = 0;
	UNBLOCK_IRQ();

	switch (dev_idx) {
#ifdef USB_ENABLE_VT6212_EHCI
	case FUNCTION_EHCI:
		irq_active = ehci__n_irq(0, urb);
		break;
#endif	// defined(USB_ENABLE_VT6212_EHCI)
#ifdef USB_ENABLE_VT6212_UHCI
	case FUNCTION_UHCI_1:
		irq_active = uhci__n_irq(0, urb);
		break;
	case FUNCTION_UHCI_2:
		irq_active = uhci__n_irq(1, urb);
		break;
#endif	// defined(USB_ENABLE_VT6212_UHCI)
	default:
		break;
	}

	BLOCK_IRQ();
	if (irq_active == 0) {
		vt6212_pending_irq[dev_idx] |= pending_irq;
	}
	vt6212_software_interrupt_disabled = flags;
	UNBLOCK_IRQ();
}

void UsbDisableInt(void)
{
	// disable IRQ from USB device
	BLOCK_IRQ();
	vt6212_software_interrupt_disabled = 1;
	UNBLOCK_IRQ();
}

void UsbEnableInt(void)
{
	unsigned long	current_preempt_status;

	// enable IRQ from USB device
	BLOCK_IRQ();
	vt6212_software_interrupt_disabled = 0;
	if (PENDING_IRQ_EXIST()) {
		SIGNAL_IR(MT_EVT_USB2_TIMER);
	}
	UNBLOCK_IRQ();
}

/* -------------------------------------------------------------------- */

/*
  timer task for USB
*/

#define USB_TIMER_NO			3
#ifdef INT_SCHEDULING
#define USB_TIMER_PRIORITY		100
#else
#define USB_TIMER_PRIORITY		100
#endif // INT_SCHEDULING
#define USB_TIMER_STACK_SIZE	6000
#define INTERVAL_10MS			(CLOCKHZ/100)
#define USB_TIMER_CHECK_PERIOD	(0xFFFFFFFF)	// (INTERVAL_10MS)
#define GET_CUR_TIME()			get_sys_time()
#define GET_CUR_TIME_IR()		((unsigned long)sys_time)

static int usb_hc_timer_task_is_active, usb_hc_timer_status[USB_TIMER_NO];
static unsigned long usb_hc_timer_active_value[USB_TIMER_NO], usb_hc_timer_timeout_value[USB_TIMER_NO];
static void (*usb_hc_timeout_func[USB_TIMER_NO])(int);

unsigned long usb_hc_jiffies(void) {
	unsigned long	flags, jiffies;

	BLOCK_IRQ();
	flags = vt6212_software_interrupt_disabled;
	vt6212_software_interrupt_disabled = 1;

	jiffies = GET_CUR_TIME_IR();

	vt6212_software_interrupt_disabled = flags;
	UNBLOCK_IRQ();

	jiffies = jiffies * 1000/HZ;
	return	jiffies;
}

void usb_hc_timer_init(int timer_id, void (*timeout_func)(int)) {
	usb_hc_timeout_func[timer_id] = timeout_func;
}

int usb_hc_timer_is_active(int timer_id) {
	int	ret_value;
	unsigned long	flags;

	BLOCK_IRQ();
	flags = vt6212_software_interrupt_disabled;
	vt6212_software_interrupt_disabled = 1;

	ret_value = usb_hc_timer_status[timer_id];

	vt6212_software_interrupt_disabled = flags;
	UNBLOCK_IRQ();
	return	ret_value;
}

void usb_hc_timer_active(int timer_id, unsigned long tick) {
	unsigned long	flags, curr_time;

	BLOCK_IRQ();
	flags = vt6212_software_interrupt_disabled;
	vt6212_software_interrupt_disabled = 1;

	curr_time = GET_CUR_TIME_IR();
	if (usb_hc_timer_status[timer_id] == 0) {
		usb_hc_timer_active_value[timer_id] = curr_time;
		usb_hc_timer_timeout_value[timer_id] = tick;
		usb_hc_timer_status[timer_id] = 1;
		SIGNAL_IR(MT_EVT_USB2_TIMER);
	} else {
		if ( (curr_time + tick) >
			(usb_hc_timer_active_value[timer_id] + usb_hc_timer_timeout_value[timer_id]) ) {
			usb_hc_timer_active_value[timer_id] = curr_time;
			usb_hc_timer_timeout_value[timer_id] = tick;
			SIGNAL_IR(MT_EVT_USB2_TIMER);
		}
	}

	vt6212_software_interrupt_disabled = flags;
	UNBLOCK_IRQ();
}

void usb_hc_timer_stop(int timer_id) {
	unsigned long	flags;

	BLOCK_IRQ();
	flags = vt6212_software_interrupt_disabled;
	vt6212_software_interrupt_disabled = 1;

	usb_hc_timer_timeout_value[timer_id] = 0xffffffff;
	usb_hc_timer_status[timer_id] = 0;

	vt6212_software_interrupt_disabled = flags;
	UNBLOCK_IRQ();
}

static
void usb_hc_timer_task() {
	int timer_id, timer_status[USB_TIMER_NO];
	unsigned long timer_active_value[USB_TIMER_NO], timer_curr_value, timer_offset_value, timer_timeout_value[USB_TIMER_NO];
	unsigned long delay_timer_cnt;

{
extern unsigned long get_data_segment(void);
extern void set_data_segment(unsigned long new_gp);
set_data_segment(get_data_segment());
}
	for (timer_id=0; timer_id<USB_TIMER_NO; timer_id++) {
		usb_hc_timer_status[timer_id] = 0;
		usb_hc_timer_timeout_value[timer_id] = 0xffffffff;
	}

	BLOCK_IRQ();
	usb_hc_timer_task_is_active = 1;

	while (1) {
		for (timer_id=0; timer_id<USB_TIMER_NO; timer_id++) {
			timer_status[timer_id] = usb_hc_timer_status[timer_id];
			timer_active_value[timer_id] = usb_hc_timer_active_value[timer_id];
			timer_timeout_value[timer_id] = usb_hc_timer_timeout_value[timer_id];
		}
		timer_curr_value = GET_CUR_TIME_IR();
		UNBLOCK_IRQ();

		delay_timer_cnt = USB_TIMER_CHECK_PERIOD;

		for (timer_id=0; timer_id<USB_TIMER_NO; timer_id++) {
			if (timer_status[timer_id] != 0) {
				if (timer_curr_value >= timer_active_value[timer_id])
					timer_offset_value = timer_curr_value - timer_active_value[timer_id];
				else
					timer_offset_value = (0xffffffff - timer_active_value[timer_id]) + (1 + timer_active_value[timer_id]);
				if (timer_timeout_value[timer_id] <= timer_offset_value)
				{
					BLOCK_IRQ();
					usb_hc_timer_status[timer_id] = 0;
					usb_hc_timer_timeout_value[timer_id] = 0xffffffff;
					UNBLOCK_IRQ();
				}
				else
				{
					timer_offset_value = timer_timeout_value[timer_id] - timer_offset_value;
					BLOCK_IRQ();
					usb_hc_timer_active_value[timer_id] = timer_curr_value;
					usb_hc_timer_timeout_value[timer_id] = timer_offset_value;
					UNBLOCK_IRQ();
					if (timer_offset_value < delay_timer_cnt)
						delay_timer_cnt = timer_offset_value;
					timer_status[timer_id] = 0;
				}
			}
		}

		for (timer_id=0; timer_id<USB_TIMER_NO; timer_id++) {
			if (timer_status[timer_id] != 0)
				(*usb_hc_timeout_func[timer_id])(timer_id);
		}

		if (delay_timer_cnt > 0) {
			wteset(MT_EVT_USB2_TIMER, delay_timer_cnt);
			clrevt(MT_EVT_USB2_TIMER);
		}
		BLOCK_IRQ();
	}
}

/* -------------------------------------------------------------------- */

//void localhc_free_config(struct usb_hcd *hcd, struct usb_device *udev)
void localhc_free_config(void *hcd, void *udev)
{
#ifdef USB_ENABLE_VT6212_EHCI
	ehci__n_free_config(0, hcd, udev);
#endif	// defined(USB_ENABLE_VT6212_EHCI)
#ifdef USB_ENABLE_VT6212_UHCI
	uhci__n_free_config(0, hcd, udev);
	uhci__n_free_config(1, hcd, udev);
#endif	// defined(USB_ENABLE_VT6212_UHCI)
}

/* -------------------------------------------------------------------- */

void localhc_irq(void)
{
#ifdef USB_ENABLE_VT6212_EHCI
	if (vt6212_pending_irq[FUNCTION_EHCI] == 0) {
		vt6212_pending_irq[FUNCTION_EHCI] = 1;
	}
#endif	// defined(USB_ENABLE_VT6212_EHCI)
}

void localhc_uhci1_irq(void)
{
#ifdef USB_ENABLE_VT6212_UHCI
	if (vt6212_pending_irq[FUNCTION_UHCI_1] == 0) {
		vt6212_pending_irq[FUNCTION_UHCI_1] = 1;
	}
#endif	// defined(USB_ENABLE_VT6212_UHCI)
}

void localhc_uhci2_irq(void)
{
#ifdef USB_ENABLE_VT6212_UHCI
	if (vt6212_pending_irq[FUNCTION_UHCI_2] == 0) {
		vt6212_pending_irq[FUNCTION_UHCI_2] = 1;
	}
#endif	// defined(USB_ENABLE_VT6212_UHCI)
}

void usb_hc_install_interrupt(int irq)
{
	// VLYNQ irq install
	int	function_idx;

	for (function_idx=0; function_idx<(int)FUNCTION_TOTAL; function_idx++)
	{
		if (vt6212_pci_device_irq[function_idx] == irq) {
			vt6212_pci_device_irq_init[function_idx] = 1;
			switch (function_idx) {
#if defined(_AR7300)
			case FUNCTION_UHCI_1:
				usb_uhci1_isr_install();
				break;
			case FUNCTION_UHCI_2:
				usb_uhci2_isr_install();
				break;
			case FUNCTION_EHCI:
				usb_isr_install();
				break;
#elif defined(_ATHEROS)
			case FUNCTION_UHCI_1:
				break;
			case FUNCTION_UHCI_2:
				break;
			case FUNCTION_EHCI:
				break;
#else
			installing IRQ here !!
#endif
			default:
				vt6212_pci_device_irq_init[function_idx] = 0;
				break;
			}
		}
	}
}

/*
  pci_given_irq - IRQ assigned by PCI
  pci_given_memory - memory space assigned by PCI (at USBBASE in PCI configuration)
*/
int __devinit hc_found_localhc(struct device *dev, int pci_given_irq, void *pci_given_memory)
{
	int	function_idx;

	function_idx = ((struct device *)dev)->dev_idx_for_pci_access;
	device_initialize(dev);

#ifdef NO_SYS_FS
	dev->driver = &(localhc_dev_driver[function_idx]);
 #ifdef USB_DBGSTRING_ON
 	dev->driver->name = localhc_dev_driver_name[function_idx];
 #endif // USB_DBGSTRING_ON
#endif // NO_SYS_FS

	if(usb_hcd_pci_probe(dev, pci_ids[function_idx], pci_given_irq, pci_given_memory) < 0)
	{
		iprintf("[USB HC] PCI probe failed! (%d)\n", function_idx);
		return -ENODEV;
	}

	return 0;
}

int usb_hcd_init(void)
{
	extern int	nousb;
	int			function_idx;
	const int	bDevInit[FUNCTION_TOTAL] = {
#ifdef USB_ENABLE_VT6212_UHCI
		1, 1
#else	// defined(USB_ENABLE_VT6212_UHCI)
		0, 0
#endif	// defined(USB_ENABLE_VT6212_UHCI)
		,
#ifdef USB_ENABLE_VT6212_EHCI
		1
#else	// defined(USB_ENABLE_VT6212_EHCI)
		0
#endif	// defined(USB_ENABLE_VT6212_EHCI)
	};

	vt6212_pci_device_irq_init[FUNCTION_UHCI_1] = 0;
	vt6212_pci_device_irq_init[FUNCTION_UHCI_2] = 0;
	vt6212_pci_device_irq_init[FUNCTION_EHCI] = 0;

#if defined(_AR7300)
	vt6212_pci_device_id[FUNCTION_UHCI_1] = 2;
	vt6212_pci_device_id[FUNCTION_UHCI_2] = 3;
	vt6212_pci_device_id[FUNCTION_EHCI] = 4;
	vt6212_pci_device_irq[FUNCTION_UHCI_1] = 3;
	vt6212_pci_device_irq[FUNCTION_UHCI_2] = 4;
	vt6212_pci_device_irq[FUNCTION_EHCI] = 5;
	// VLINQ to PCI interface
	vt6212_pci_device_addr[FUNCTION_UHCI_1] = 0xA4021400;	// (CPU addr)0xA4021400 -> 0x60000000(PCI space)
	vt6212_pci_device_addr[FUNCTION_UHCI_2] = 0xA4021420;	// (CPU addr)0xA4021420 -> 0x60000020(PCI space)
#ifdef _Vendor_SIEMENS_SX553
	vt6212_pci_device_addr[FUNCTION_EHCI] = 0xA4001400;	// (CPU addr)0xA4001400 -> 0x40000000(PCI space)
#else
	vt6212_pci_device_addr[FUNCTION_EHCI] = 0xA4011400;	// (CPU addr)0xA4011400 -> 0x40010000(PCI space)
#endif
	memory_addr_in_cpu = 0xB4000000;		// non-cacheable address
	memory_addr_in_pci = 0xFC000000;
#elif defined(_ATHEROS)
	vt6212_pci_device_id[FUNCTION_UHCI_1] = 2;
	vt6212_pci_device_id[FUNCTION_UHCI_2] = 3;
	vt6212_pci_device_id[FUNCTION_EHCI] = 0x2<<8; //4;
	vt6212_pci_device_irq[FUNCTION_UHCI_1] = 3;
	vt6212_pci_device_irq[FUNCTION_UHCI_2] = 4;
	vt6212_pci_device_irq[FUNCTION_EHCI] = 5;
	// VLINQ to PCI interface
	vt6212_pci_device_addr[FUNCTION_UHCI_1] = 0xE0880000; 	// (CPU addr)0xA4021400 -> 0x60000000(PCI space)
	vt6212_pci_device_addr[FUNCTION_UHCI_2] = 0xE0880020; 	// (CPU addr)0xA4021420 -> 0x60000020(PCI space)
	vt6212_pci_device_addr[FUNCTION_EHCI] = 0xE0800000;	// (CPU addr)0xA4011400 -> 0x40010000(PCI space)
	memory_addr_in_cpu = 0xA0000000;		// non-cacheable address
	memory_addr_in_pci = 0x20000000;
#else
	address & io mapping mush be defined !!
#endif
	cpu_to_usb_addr_shift = (long)(-memory_addr_in_cpu + memory_addr_in_pci);
	usb_to_cpu_addr_shift = (long)(-memory_addr_in_pci + memory_addr_in_cpu);

	vt6212_software_interrupt_disabled = 0;
	for (function_idx=0; function_idx<(int)FUNCTION_TOTAL; function_idx++)
		vt6212_pending_irq[function_idx] = 0;

	vt6212_locked_task_id = 0xffffffff;
	vt6212_locked_task_reentry = 0;

	/* create timer task for USB */
	{
		int task_id, timer_is_active;

		usb_hc_timer_task_is_active = 0;
		task_id = runtsk(USB_TIMER_PRIORITY, (void (fAR *)(void))usb_hc_timer_task, USB_TIMER_STACK_SIZE);
		if (task_id < SUCCESS)
			return -ENODEV;

		do {
			TASK_SLEEP(1);
			BLOCK_PREEMPTION;
			timer_is_active = usb_hc_timer_task_is_active;
			UNBLOCK_PREEMPTION;
		} while (timer_is_active == 0);
	}

	/* initialize VIA VT6212 host controller */
	nousb = 0;
	usb_hub_init_end = 0;

	for (function_idx=0; function_idx<(int)FUNCTION_TOTAL; function_idx++) {
		if (bDevInit[function_idx] != 0) {
			vt6212_dev[function_idx].dev_idx_for_pci_access = function_idx;
			if (hc_found_localhc(&(vt6212_dev[function_idx])
								, vt6212_pci_device_irq[function_idx]
								, (void *)(vt6212_pci_device_addr[function_idx])) < 0) {
				iprintf("VT6212 function %d not initialized !!\n", function_idx);
				return	-ENODEV;
			}
		}
	}

	hcd_init_ok = 1;

	return 1;
}

