#ifndef _HCDX_H_
#define _HCDX_H_

#include "UsbCfg.h"

#ifdef _EHCI_HCDX_H_	// Linux 2.6.12.2 USB HCD

/*
	EHCI_REMOTE_WAKEUP definition :
		0 - remote wakeup function is not enabled
		1 - remote wakeup function is enabled
*/
#define EHCI_REMOTE_WAKEUP	0

/*
	CONFIG_USB_EHCI_SPLIT_ISO definition :
		enable this definition will enable "Split ISO TDs"
		- used for USB 1.1 devices going through the
			TTs in USB 2.0 hubs.  These need microframe scheduling.
*/
#define CONFIG_USB_EHCI_SPLIT_ISO

#define EHCI_TIMER()					0

#define __le32							__u32
#define __le16							__u16
#define	__iomem							volatile
#define __force

#define HC_IS_RUNNING					HCD_IS_RUNNING
#define HC_STATE_HALT					USB_STATE_HALT
#define HC_STATE_RUNNING				USB_STATE_RUNNING

#undef usb_get_dev
#define usb_get_dev(x)					x
#define usb_put_dev(x)					{;}
#define to_pci_dev(n)					container_of(n, struct pci_dev, dev)

#define dma_pool_destroy				pci_pool_destroy
#define dma_pool_create					pci_pool_create

#define pci_pool_destroy(x)				{;}
#define pci_pool_create(a,b,c,d,e)		(void *)(1)

#define wmb()							{;}
#define rmb()							{;}

#define msecs_to_jiffies(x)				(HZ*x/1000)
#define	msleep(x)						wait_ms(x)
#define jiffies							usb_hc_jiffies()
#define del_timer_sync(x)				usb_hc_timer_stop(EHCI_TIMER())

#define irqreturn_t						int
#define IRQ_NONE						0
#define IRQ_HANDLED						1

#define BUG_ON(x)						{;}

struct kref {
	atomic_t refcount;
};
#define	kref_init(x)	atomic_set(&(((struct kref *)x)->refcount),1);
#define	kref_get(x)		atomic_inc(&(((struct kref *)x)->refcount));
#define	kref_put(x,fc)	{if (atomic_dec_and_test(&(((struct kref *)x)->refcount)))	fc(x);}

static inline unsigned char test_bit(int nr, volatile void *addr) {
	return ( ((unsigned char *) addr)[nr >> 3] & (1U << (nr & 7)) );
}
static inline unsigned char test_and_set_bit(int nr, volatile void *addr) {
	unsigned char	oldval, mask = 1U << (nr & 7);
	oldval = ((unsigned char *) addr)[nr >> 3];
	((unsigned char *) addr)[nr >> 3] |= mask;
	return	(oldval & ~mask);
}
static inline unsigned long ffs(unsigned long bits_word) {
	int k=31;
	if (bits_word & 0x0000ffff) { k -= 16; bits_word <<= 16; }
	if (bits_word & 0x00ff0000) { k -= 8;  bits_word <<= 8;  }
	if (bits_word & 0x0f000000) { k -= 4;  bits_word <<= 4;  }
	if (bits_word & 0x30000000) { k -= 2;  bits_word <<= 2;  }
	if (bits_word & 0x40000000) { k -= 1; }
	return k;
}

#endif	// defined(_EHCI_HCDX_H_)


#ifdef _UHCI_HCDX_H_	// Linux 2.6.12.2 USB HCD

#define UHCI_TIMER_BASE					1
#define UHCI_TIMER(x)					(UHCI_TIMER_BASE+x)
#define UHCI_IDX_FROM_TIMER(n)			(n-UHCI_TIMER_BASE)

#define __le32							__u32
#define __le16							__u16

#undef usb_get_dev
#define usb_get_dev(x)					{;}
#define usb_put_dev(x)					{;}
#define to_pci_dev(n)					container_of(n, struct pci_dev, dev)

#define dma_pool						pci_pool
#define dma_pool_destroy				pci_pool_destroy
#define dma_pool_create					pci_pool_create

#define pci_pool_destroy(x)				{;}
#define pci_pool_create(a,b,c,d,e)		(void *)(1)

#define wmb()							{;}
#define rmb()							{;}
#define mb()							{;}

#define	inb(addr)						pci_read_io_byte(uhci, addr)
#define	inw(addr)						pci_read_io_word(uhci, addr)
#define	inl(addr)						pci_read_io_dword(uhci, addr)
#define	outw(v,addr)					pci_write_io_word(uhci, addr, v)
#define	outl(v,addr)					pci_write_io_dword(uhci, addr, v)

#define msecs_to_jiffies(x)				(HZ*x/1000)
#define	msleep(x)						wait_ms(x)
#define jiffies							usb_hc_jiffies()
#define del_timer_sync(x)				usb_hc_timer_stop(uhci->timer_id)

#define irqreturn_t						int
#define IRQ_NONE						0
#define IRQ_HANDLED						1

#define WARN_ON(x)						{;}

#define	wait_queue_head_t				int
#define init_waitqueue_head(wq)			{;}
#define wake_up_all(wq)					{;}
#define wait_event_interruptible(wq, condition)	{;}

#define barrier()						{;}

static inline unsigned char test_bit(int nr, volatile void *addr) {
	return ( ((unsigned char *) addr)[nr >> 3] & (1U << (nr & 7)) );
}
static inline unsigned char test_and_set_bit(int nr, volatile void *addr) {
	unsigned char	oldval, mask = 1U << (nr & 7);
	oldval = ((unsigned char *) addr)[nr >> 3];
	((unsigned char *) addr)[nr >> 3] |= mask;
	return	(oldval & ~mask);
}
static inline unsigned long __ffs(unsigned long bits_word) {
	int k=31;
	if (bits_word & 0x0000ffff) { k -= 16; bits_word <<= 16; }
	if (bits_word & 0x00ff0000) { k -= 8;  bits_word <<= 8;  }
	if (bits_word & 0x0f000000) { k -= 4;  bits_word <<= 4;  }
	if (bits_word & 0x30000000) { k -= 2;  bits_word <<= 2;  }
	if (bits_word & 0x40000000) { k -= 1; }
	return k;
}

#endif	// defined(_UHCI_HCDX_H_)


#define likely(x)	(x)
#define unlikely(x)	(x)

#ifdef NO_SPINLOCK
#undef	spinlock_t
#undef	spin_lock
#undef	spin_unlock
#undef	spin_lock_irqsave
#undef	spin_unlock_irqrestore
#define spinlock_t					int
#define	spin_lock_init(x)			{;}
#define spin_lock_irq(x)			{;}
#define spin_unlock_irq(x)			{;}
#define spin_lock(x)				{;}
#define spin_unlock(x)				{;}
#define	spin_lock_irqsave(a,b)		{;}
#define	spin_unlock_irqrestore(a,b)	{;}
#endif	// defined (NO_SPINLOCK)

#define time_after(a,b)				((long)(b) - (long)(a) < 0)
#define time_after_eq(a,b)			((long)(a) - (long)(b) >= 0)

extern void usb_hc_timer_init(int timer_id, void (*timeout_func)(int));
extern int usb_hc_timer_is_active(int timer_id);
extern void usb_hc_timer_active(int timer_id, unsigned long tick);
extern void usb_hc_timer_stop(int timer_id);

#endif	// ndef _HCDX_H_
