/*
 * class.c - basic device class management
 *
 * Copyright (c) 2002-3 Patrick Mochel
 * Copyright (c) 2002-3 Open Source Development Labs
 * Copyright (c) 2003 Greg Kroah-Hartman
 * Copyright (c) 2003 IBM Corp.
 * 
 * This file is released under the GPLv2
 *
 */

#undef DEBUG

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
// #include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
// #include "UsbMsDbg.h"
// #include "UsbMsInit.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP


/*
#include <linux/device.h>
*/
#include "PoDevice.h" // include/linux/device.h
#include "kdev_t.h"   // also need
// B. S. Liaw Add this
#include "base/PoDeviceSc.h" // #include/linux/device.h , 2.6.0-test7
/*
#include <linux/module.h>
#include <linux/init.h>
#include <linux/string.h>
*/
#include "PoString.h" // linux/string.h

#include "base.h"

#include "PoKernel.h"

// B. S Liaw add this !!
#include "kobject.h" // for decl_subsys()


#define to_class_attr(_attr) container_of(_attr,struct class_attribute,attr)
#define to_class(obj) container_of(obj,struct class,subsys.kset.kobj)
/*
#ifndef NO_SYS_FS
// struct attribute on sysfs.h
static ssize_t
class_attr_show(struct kobject * kobj, struct attribute * attr, char * buf)
{
	struct class_attribute * class_attr = to_class_attr(attr);
	struct class * dc = to_class(kobj);
	ssize_t ret = 0;

	if (class_attr->show)
		ret = class_attr->show(dc,buf);
	return ret;
}

// struct attribute on sysfs.h
static ssize_t
class_attr_store(struct kobject * kobj, struct attribute * attr, 
		 const char * buf, size_t count)
{
	struct class_attribute * class_attr = to_class_attr(attr);
	struct class * dc = to_class(kobj);
	ssize_t ret = 0;

	if (class_attr->store)
		ret = class_attr->store(dc,buf,count);
	return ret;
}

static struct sysfs_ops class_sysfs_ops = {
	.show	= class_attr_show,
	.store	= class_attr_store,
};
#endif // ndef NO_SYS_FS
*/
static struct kobj_type ktype_class = {
 #ifndef NO_SYS_FS
	.sysfs_ops	= &class_sysfs_ops,
 #endif // ndef NO_SYS_FS
};

/* Hotplug events for classes go to the class_obj subsys */
/*
static decl_subsys(class,&ktype_class,NULL);

static decl_subsys(class,&ktype_class);
*/

#ifndef NO_SYS_FS
int class_create_file(struct class * cls, const struct class_attribute * attr)
{
 int error;

 USB_DBG_FLOW(iprintf("[USB FLOW] class_create_file() Begin .. \n"));
 if (cls) {
	#ifndef NO_SYS_FS
	    // struct attribute on sysfs.h
		error = sysfs_create_file(&cls->subsys.kset.kobj,&attr->attr);
	 #else
		error = 0 ;
	#endif
	} else
		error = -EINVAL;
 return error;
}

void class_remove_file(struct class * cls, const struct class_attribute * attr)
{
   // struct attribute on sysfs.h
	if (cls)
		sysfs_remove_file(&cls->subsys.kset.kobj,&attr->attr);
}
#endif // ndef NO_SYS_FS

struct class * class_get(struct class * cls)
{
	if (cls)
		return container_of(subsys_get(&cls->subsys),struct class,subsys);
	return NULL;
}

void class_put(struct class * cls)
{
 // #ifndef _ETCPIP
	subsys_put(&cls->subsys);
 // #endif // ndef _ETCPIP
}

int class_register(struct class * cls)
{
 USB_DBG_FLOW(iprintf("[USB FLOW] class_register() Begin .. \n"));
 
 #ifdef USB_DBGSTRING_ON
	/*
	pr_debug("device class '%s': registering\n",cls->name);
	*/
	dprintf("[USBMS] device class [%s]: registering \n",cls->name);
 #endif // USB_DBGSTRING_ON
 
 INIT_LIST_HEAD(&cls->children);
 INIT_LIST_HEAD(&cls->interfaces);
 #ifdef USB_DBGSTRING_ON
	USB_DBG_FLOW(iprintf("[USB FLOW] class_register() call kobject_set_name() .. \n"));
	kobject_set_name(&cls->subsys.kset.kobj,cls->name);
 #endif // USB_DBGSTRING_ON
 #ifndef _ETCPIP
	USB_DBG_FLOW(iprintf("[USB FLOW] class_register() call subsys_set_kset() .. \n"));
	subsys_set_kset(cls,class_subsys);
 #endif // ndef _ETCPIP
 #ifndef _ETCPIP
	USB_DBG_FLOW(iprintf("[USB FLOW] class_register() call subsystem_register() .. \n"));
	subsystem_register(&cls->subsys);
 #endif // ndef _ETCPIP

 USB_DBG_FLOW(iprintf("[USB FLOW] class_register() End. \n"));
 return 0;
}

void class_unregister(struct class * cls)
{
 
 #ifdef USB_DBGSTRING_ON
	pr_debug("device class '%s': unregistering\n",cls->name);
 #endif // USB_DBGSTRING_ON
 
 dprintf("class_unregister() : device class unregistering\n") ;
 #ifndef _ETCPIP
	subsystem_unregister(&cls->subsys);
 #endif // ndef _ETCPIP
}

/* Class Device Stuff */

int class_device_create_file(struct class_device * class_dev,
			     const struct class_device_attribute * attr)
{
 int error = -EINVAL;

 USB_DBG_FLOW(iprintf("[USB FLOW] class_device_create_file() Begin .. \n"));
 if(class_dev)
   {
	#ifndef NO_SYS_FS
	    // struct attribute on sysfs.h
		error = sysfs_create_file(&class_dev->kobj, &attr->attr);
	 #else
		error = 0 ;
	#endif
   }
 USB_DBG_FLOW(iprintf("[USB FLOW] class_device_create_file() End. \n"));
 return error;
}

void class_device_remove_file(struct class_device * class_dev,
			      const struct class_device_attribute * attr)
{
		#ifndef NO_SYS_FS
		    // struct attribute on sysfs.h
	if (class_dev)
		sysfs_remove_file(&class_dev->kobj, &attr->attr);
		#endif // ndef NO_SYS_FS
}

static int class_device_dev_link(struct class_device * class_dev)
{
 #ifndef NO_SYS_FS
	if (class_dev->dev)
		return sysfs_create_link(&class_dev->kobj,
					 &class_dev->dev->kobj, "device");
 #endif // ndef NO_SYS_FS
	return 0;
}

static void class_device_dev_unlink(struct class_device * class_dev)
{
 #ifndef NO_SYS_FS
	if (class_dev->dev)
		sysfs_remove_link(&class_dev->kobj, "device");
 #endif // ndef NO_SYS_FS
}

static int class_device_driver_link(struct class_device * class_dev)
{
		#ifndef NO_SYS_FS
	if ((class_dev->dev) && (class_dev->dev->driver))
		return sysfs_create_link(&class_dev->kobj,
					 &class_dev->dev->driver->kobj, "driver");
		#endif // ndef NO_SYS_FS
	return 0;
}

static void class_device_driver_unlink(struct class_device * class_dev)
{
 #ifndef NO_SYS_FS
	if ((class_dev->dev) && (class_dev->dev->driver))
		sysfs_remove_link(&class_dev->kobj, "driver");
 #endif // ndef NO_SYS_FS
}


#ifndef NO_SYS_FS
// struct attribute on sysfs.h
static ssize_t
class_device_attr_show(struct kobject * kobj, struct attribute * attr,
		       char * buf)
{
	struct class_device_attribute * class_dev_attr = to_class_dev_attr(attr);
	struct class_device * cd = to_class_dev(kobj);
	ssize_t ret = 0;

	if (class_dev_attr->show)
		ret = class_dev_attr->show(cd,buf);
	return ret;
}

// struct attribute on sysfs.h
static ssize_t
class_device_attr_store(struct kobject * kobj, struct attribute * attr, 
			const char * buf, size_t count)
{
	struct class_device_attribute * class_dev_attr = to_class_dev_attr(attr);
	struct class_device * cd = to_class_dev(kobj);
	ssize_t ret = 0;

	if (class_dev_attr->store)
		ret = class_dev_attr->store(cd,buf,count);
	return ret;
}

static struct sysfs_ops class_dev_sysfs_ops = {
	.show	= class_device_attr_show,
	.store	= class_device_attr_store,
};
#endif // ndef NO_SYS_FS

static void class_dev_release(struct kobject * kobj)
{
	struct class_device *cd = to_class_dev(kobj);
	struct class * cls = cd->class;

	pr_debug("device class '%s': release.\n",cd->class_id);

 if(cls->release)
   {
	USB_DBG_FLOW(iprintf("[USB FLOW] class_dev_release() : cls->release() .. \n")) ;
	cls->release(cd);
   }
  else {
	/*
	printk(KERN_ERR "Device class '%s' does not have a release() function, "
			"it is broken and must be fixed.\n",
			cd->class_id);
	WARN_ON(1);
	*/
	dprintf("[USBMS] Error : Device class [%s] does not have a release() function, "
			"it is broken and must be fixed.\n",
			cd->class_id);
   }
}

#ifndef NO_SYS_FS
static struct kobj_type ktype_class_device = {
	.sysfs_ops	= &class_dev_sysfs_ops,
	.release	= class_dev_release,
};
#endif // ndef NO_SYS_FS

static int class_hotplug_filter(struct kset *kset, struct kobject *kobj)
{
 /*
 #ifndef NO_SYS_FS
 
 struct kobj_type *ktype = get_ktype(kobj);

	if (ktype == &ktype_class_device) {
		struct class_device *class_dev = to_class_dev(kobj);
		if (class_dev->class)
			return 1;
	}
 #endif // ndef NO_SYS_FS
 */
	return 0;
}
/*
static char *class_hotplug_name(struct kset *kset, struct kobject *kobj)
{
	struct class_device *class_dev = to_class_dev(kobj);

	return class_dev->class->name;
}
*/

static int class_hotplug(struct kset *kset, struct kobject *kobj, char **envp,
			 int num_envp, char *buffer, int buffer_size)
{
	struct class_device *class_dev = to_class_dev(kobj);
	int retval = 0;

	pr_debug("%s - name = %s\n", __FUNCTION__, class_dev->class_id);
	if (class_dev->class->hotplug) {
		/* have the bus specific function add its stuff */
		retval = class_dev->class->hotplug (class_dev, envp, num_envp,
						    buffer, buffer_size);
			if (retval) {
			pr_debug ("%s - hotplug() returned %d\n",
				  __FUNCTION__, retval);
		}
	}

	return retval;
}
/*
static struct kset_hotplug_ops class_hotplug_ops = {
	.filter =	class_hotplug_filter,
 #ifdef USB_DBGSTRING_ON
	.name =		class_hotplug_name,
 #endif // USB_DBGSTRING_ON 
	.hotplug =	class_hotplug,
};
*/
/*
static decl_subsys(class_obj, &ktype_class_device, &class_hotplug_ops);

static decl_subsys(class_obj,&ktype_class_device);
*/
void class_device_initialize(struct class_device *class_dev)
{
	kobject_init(&class_dev->kobj);
	INIT_LIST_HEAD(&class_dev->node);
}

int class_device_add(struct class_device *class_dev)
{
 struct class * parent;
 struct class_interface * class_intf;
 struct list_head * entry;
 int error;


 USB_DBG_FLOW(iprintf("[USB FLOW] class_device_add() Begin .. \n"));

 class_dev = class_device_get(class_dev);
 if (!class_dev || !strlen(class_dev->class_id))
		return -EINVAL;

 parent = class_get(class_dev->class);

 pr_debug("CLASS: registering class device: ID = '%s'\n",
		 class_dev->class_id);

 /* first, register with generic layer. */
 #ifdef USB_DBGSTRING_ON
	USB_DBG_FLOW(iprintf("[USB FLOW] class_device_add() call kobject_set_name() .. \n"));
	kobject_set_name(&class_dev->kobj, class_dev->class_id);
 #endif // USB_DBGSTRING_ON
 
 #ifndef _ETCPIP
	USB_DBG_FLOW(iprintf("[USB FLOW] class_register() call kobj_set_kset_s() .. \n"));
	kobj_set_kset_s(class_dev, class_obj_subsys);
 #endif // ndef _ETCPIP
 
 if (parent)
		class_dev->kobj.parent = &parent->subsys.kset.kobj;

 USB_DBG_INIT(iprintf("[USB INIT] class_device_add() : kobject_add() .. \n"));
 if ((error = kobject_add(&class_dev->kobj)))
		goto register_done;

 /* now take care of our own registration */
 if (parent) {
	#ifndef NO_SEMPHORE
		down_write(&parent->subsys.rwsem);
	#endif // ndef NO_SEMPHORE
	USB_DBG_MEM(iprintf("[USB MEM] class_device_add() : parent->children=%08XH \n",
			parent->children));
		#ifdef USB_BS_ASSERT
			if ( ! (IS_VALID_ADDR(&parent->children)) ) {
				iprintf("[ASSERT][USB] class_device_add() : " ) ;
    		    iprintf("Error : Invlid address &parent->children=%08XH \n",
						&parent->children) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
		list_add_tail(&class_dev->node, &parent->children);
		list_for_each(entry, &parent->interfaces) {
			class_intf = container_of(entry, struct class_interface, node);
			if (class_intf->add)
			   {
		USB_DBG_FLOW(iprintf("[USB FLOW] class_device_add() : class_intf->add() .. \n")) ;
				class_intf->add(class_dev);
			   }
		}
	#ifndef NO_SEMPHORE
		up_write(&parent->subsys.rwsem);
	#endif // ndef NO_SEMPHORE
   }

 class_device_dev_link(class_dev);
 class_device_driver_link(class_dev);

 register_done:

 if (error && parent)
		class_put(parent);
 class_device_put(class_dev);

 USB_DBG_FLOW(iprintf("[USB FLOW] class_device_add() End. \n"));
 return error;
}

int class_device_register(struct class_device *class_dev)
{
 class_device_initialize(class_dev);
 return class_device_add(class_dev);
}

void class_device_del(struct class_device *class_dev)
{
 struct class * parent = class_dev->class;
 struct class_interface * class_intf;
 struct list_head * entry;

 if(parent) {
	#ifndef NO_SEMPHORE
		down_write(&parent->subsys.rwsem);
	#endif // ndef NO_SEMPHORE
	USB_DBG_FLOW(iprintf("[USB FLOW] class_device_del() : list_del_init() .. \n"));
	list_del_init(&class_dev->node);
	list_for_each(entry, &parent->interfaces) {
		class_intf = container_of(entry, struct class_interface, node);
		if (class_intf->remove)
			class_intf->remove(class_dev);
		}
	#ifndef NO_SEMPHORE
		up_write(&parent->subsys.rwsem);
	#endif // ndef NO_SEMPHORE
   }

	class_device_dev_unlink(class_dev);
	class_device_driver_unlink(class_dev);
	
 #ifndef _ETCPIP
	kobject_del(&class_dev->kobj);
 #endif // ndef _ETCPIP

 if (parent)
		class_put(parent);
}

void class_device_unregister(struct class_device *class_dev)
{
	pr_debug("CLASS: Unregistering class device. ID = '%s'\n",
		 class_dev->class_id);
	class_device_del(class_dev);
	class_device_put(class_dev);
}
/*
int class_device_rename(struct class_device *class_dev, char *new_name)
{
	class_dev = class_device_get(class_dev);
	if (!class_dev)
		return -EINVAL;

	pr_debug("CLASS: renaming '%s' to '%s'\n", class_dev->class_id,
		 new_name);

	strlcpy(class_dev->class_id, new_name, KOBJ_NAME_LEN);

 #ifndef _ETCPIP
	kobject_rename(&class_dev->kobj, new_name);
 #endif // ndef _ETCPIP

	class_device_put(class_dev);

	return 0;
}
*/
struct class_device * class_device_get(struct class_device *class_dev)
{
	if (class_dev)
		return to_class_dev(kobject_get(&class_dev->kobj));
	return NULL;
}

void class_device_put(struct class_device *class_dev)
{
 // #ifndef _ETCPIP
	kobject_put(&class_dev->kobj);
 // #endif // ndef _ETCPIP
}


int class_interface_register(struct class_interface *class_intf)
{
 struct class * parent;
 struct class_device * class_dev;
 struct list_head * entry;

 if (!class_intf || !class_intf->class)
		return -ENODEV;

 parent = class_get(class_intf->class);
 if (!parent)
		return -EINVAL;

 #ifndef NO_SEMPHORE
	down_write(&parent->subsys.rwsem);
 #endif // ndef NO_SEMPHORE
 USB_DBG_MEM(iprintf("[USB MEM] class_interface_register() : parent->interfaces=%08XH \n",
			parent->interfaces));
 #ifdef USB_BS_ASSERT
	/*
			if (!(   (unsigned long) &parent->interfaces
				   > (unsigned long) 0x200000L
		         )
			   )
			   {
	*/
	if ( ! (IS_VALID_ADDR(&parent->interfaces)) ) {
				iprintf("[ASSERT][USB] class_interface_register() : " ) ;
    		    iprintf("Error : Invlid address &parent->interfaces=%08XH \n",
						&parent->interfaces) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
 #endif // USB_BS_ASSERT

 list_add_tail(&class_intf->node, &parent->interfaces);

 if (class_intf->add) {
		list_for_each(entry, &parent->children) {
		class_dev = container_of(entry, struct class_device, node);
		USB_DBG_FLOW(iprintf("[USB FLOW] class_interface_register() : class_intf->add() .. \n")) ;
		class_intf->add(class_dev);
	   }
   }
 #ifndef NO_SEMPHORE
	up_write(&parent->subsys.rwsem);
 #endif // ndef NO_SEMPHORE

 return 0;
}

void class_interface_unregister(struct class_interface *class_intf)
{
	struct class * parent = class_intf->class;
	struct list_head * entry;

	if (!parent)
		return;

 #ifndef NO_SEMPHORE
	down_write(&parent->subsys.rwsem);
 #endif // ndef NO_SEMPHORE
	USB_DBG_FLOW(iprintf("[USB FLOW] class_interface_unregister() : list_del_init() .. \n"));
	list_del_init(&class_intf->node);

	if (class_intf->remove) {
		list_for_each(entry, &parent->children) {
			struct class_device *class_dev = container_of(entry, struct class_device, node);
			class_intf->remove(class_dev);
		}
	}
 #ifndef NO_SEMPHORE
	up_write(&parent->subsys.rwsem);
 #endif // ndef NO_SEMPHORE

	class_put(parent);
}


/*
int __init classes_init(void)
*/
int classes_init(void)
{
 int retval;


 USB_DBG_FLOW(iprintf("[USB FLOW] classes_init() Begin .. \n"));
 /*
 #ifndef _ETCPIP
	retval = subsystem_register(&class_subsys);
	if (retval)
		return retval;
 #endif // ndef _ETCPIP
 */
 /* ick, this is ugly, the things we go through to keep from showing up
  * in sysfs... */
 /*
 #ifndef _ETCPIP
	subsystem_init(&class_obj_subsys);
 
	if (!class_obj_subsys.kset.subsys)
			class_obj_subsys.kset.subsys = &class_obj_subsys;
 #endif // ndef _ETCPIP
 */
 return 0;
}
/*
EXPORT_SYMBOL(class_create_file);
EXPORT_SYMBOL(class_remove_file);
EXPORT_SYMBOL(class_register);
EXPORT_SYMBOL(class_unregister);
EXPORT_SYMBOL(class_get);
EXPORT_SYMBOL(class_put);

EXPORT_SYMBOL(class_device_register);
EXPORT_SYMBOL(class_device_unregister);
EXPORT_SYMBOL(class_device_initialize);
EXPORT_SYMBOL(class_device_add);
EXPORT_SYMBOL(class_device_del);
EXPORT_SYMBOL(class_device_get);
EXPORT_SYMBOL(class_device_put);
EXPORT_SYMBOL(class_device_create_file);
EXPORT_SYMBOL(class_device_remove_file);

EXPORT_SYMBOL(class_interface_register);
EXPORT_SYMBOL(class_interface_unregister);
*/

// B. S. Liaw TEST for compiler worning !!

