/*
 * platform.c - platform 'psuedo' bus for legacy devices
 *
 * Copyright (c) 2002-3 Patrick Mochel
 * Copyright (c) 2002-3 Open Source Development Labs
 * 
 * This file is released under the GPLv2
 *
 * Please see Documentation/driver-model/platform.txt for more
 * information.
 */

// #ifdef _ETCPIP
#define __KERNEL__
        #ifdef _NEWLIB
          #include "string.h"
        #endif
   	#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsbMsInit.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP



/*
#include <linux/device.h>
*/
#include "PoDevice.h" // include/linux/device.h
#include "kdev_t.h"   // also need
// B. S. Liaw Add this
#include "base/PoDeviceSc.h" // #include/linux/device.h , 2.6.0-test7
/*
#include <linux/module.h>
#include <linux/init.h>
*/

#include "PoKernel.h"

struct device legacy_bus = {
	.bus_id		= "legacy",
};

/**
 *	platform_device_register - add a platform-level device
 *	@dev:	platform device we're adding
 *
 */
int platform_device_register(struct platform_device * pdev)
{
	if (!pdev)
		return -EINVAL;

	if (!pdev->dev.parent)
		pdev->dev.parent = &legacy_bus;

	pdev->dev.bus = &platform_bus_type;
 #ifdef USB_DBGSTRING_ON
	snprintf(pdev->dev.bus_id,BUS_ID_SIZE,"%s%u",pdev->name,pdev->id);
 #endif // USB_DBGSTRING_ON
	snprintf(pdev->dev.bus_id,BUS_ID_SIZE,"%u",pdev->id);

	pr_debug("Registering platform device '%s'. Parent at %s\n",
		 pdev->dev.bus_id,pdev->dev.parent->bus_id);
	return device_register(&pdev->dev);
}

void platform_device_unregister(struct platform_device * pdev)
{
	if (pdev)
		device_unregister(&pdev->dev);
}


/**
 *	platform_match - bind platform device to platform driver.
 *	@dev:	device.
 *	@drv:	driver.
 *
 *	Platform device IDs are assumed to be encoded like this: 
 *	"<name><instance>", where <name> is a short description of the 
 *	type of device, like "pci" or "floppy", and <instance> is the 
 *	enumerated instance of the device, like '0' or '42'.
 *	Driver IDs are simply "<name>". 
 *	So, extract the <name> from the device, and compare it against 
 *	the name of the driver. Return whether they match or not.
 */
/*
static int platform_match(struct device * dev, struct device_driver * drv)
{
	struct platform_device *pdev = container_of(dev, struct platform_device, dev);

	return (strncmp(pdev->name, drv->name, BUS_ID_SIZE) == 0);
}
*/

static int platform_suspend(struct device * dev, u32 state)
{
	int ret = 0;

	if (dev->driver && dev->driver->suspend) {
		ret = dev->driver->suspend(dev, state, SUSPEND_DISABLE);
		if (ret == 0)
			ret = dev->driver->suspend(dev, state, SUSPEND_SAVE_STATE);
		if (ret == 0)
			ret = dev->driver->suspend(dev, state, SUSPEND_POWER_DOWN);
	}
	return ret;
}

static int platform_resume(struct device * dev)
{
	int ret = 0;

	if (dev->driver && dev->driver->resume) {
		ret = dev->driver->resume(dev, RESUME_POWER_ON);
		if (ret == 0)
			ret = dev->driver->resume(dev, RESUME_RESTORE_STATE);
		if (ret == 0)
			ret = dev->driver->resume(dev, RESUME_ENABLE);
	}
	return ret;
}

struct bus_type platform_bus_type = {
 
 #ifdef USB_DBGSTRING_ON
	.name		= "platform",
 #endif // USB_DBGSTRING_ON 
 /*
	.match		= platform_match,
	.suspend	= platform_suspend,
	.resume		= platform_resume,
 */
};

/*
int __init platform_bus_init(void)
*/
int platform_bus_init(void)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] platform_bus_init() Begin .. \n"));
 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] platform_bus_init() call device_register() .. \n"));
 device_register(&legacy_bus);
 #ifndef NO_KSET
	return bus_register(&platform_bus_type);
 #endif // ndef NO_KSET
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] platform_bus_init() End. \n"));
}

/*
EXPORT_SYMBOL(legacy_bus);
EXPORT_SYMBOL(platform_bus_type);
EXPORT_SYMBOL(platform_device_register);
EXPORT_SYMBOL(platform_device_unregister);
*/

