/*
 *  linux/arch/arm/mach-s3c2510.c            
 *  Special pci_map/unmap_single routines for S3C2510 
 * 
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 */

#ifdef _ETCPIP
   #define __KERNEL__
#ifdef _NEWLIB
  #include "sys_type.h"
#endif
           #ifdef _NEWLIB
          #include "string.h"
        #endif
   	#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
          // for deciding whether to compile
          //  Debug Version or Release Version
   #include "etcpip.h"
   #include "UsbDbg.h"
   
   #include "PoList.h" // osport_list.h
   
   #include "UsbCfg.h" // Config and Setting
   #include "UsOsport.h" // "usb_osport.h"
#endif // _ETCPIP

#ifndef _ETCPIP
   #include <linux/config.h>
   #include <linux/module.h>
   #include <linux/init.h>
   #include <linux/slab.h>
#endif // ndef _ETCPIP
    
#ifndef _ETCPIP
	#include <linux/pci.h>
#else
    #include "PoDevice.h" // osport_device.h
	#ifdef CONFIG_PCI
	    #include "UsbPoPci.h" // usb_osport_pci.h
    #endif	// CONFIG_PCI
#endif

#ifndef USB_NO_DMA_POOL
	#include "UsDmaPol.h" // usb_pcipool.h
#endif // ndef USB_NO_DMA_POOL

/*
 * simple buffer allocator for copying of unsafe to safe buffers
 * uses __alloc/__free for actual buffers
 * keeps track of safe buffers we've allocated so we can recover the
 * unsafe buffers.
 */

#define MAX_SAFE	32
#define SIZE_SMALL	1024
#ifdef USB_NO_DMA_POOL
	#define SIZE_LARGE	(8*1024)
 #else
	#define SIZE_LARGE	(16*1024)
#endif

static long mapped_alloc_size;
static char *safe_buffers[MAX_SAFE][2];


#ifndef USB_NO_DMA_POOL // 2003/7/2 B. S. Liaw
	static struct pci_pool *small_buffer_cache, 
    					   *large_buffer_cache;
#endif // ndef USB_NO_DMA_POOL 

static int
init_safe_buffers(struct pci_dev *dev)
{
 USB_DBG_FLOW(iprintf("[USB FLOW] init_safe_buffers() Begin ..\n"));
 
 #ifndef USB_NO_DMA_POOL // 2003/7/2 B. S. Liaw
 USB_DBG_MEM(iprintf(
	"[USB MEM] init_safe_buffers() : alloc small_buffer_cache..\n"
    ));

 #ifdef USB_DBGSTRING_ON
	small_buffer_cache = 
 		usb_pci_pool_create("pci_small_buffer",
					    dev,
					    SIZE_SMALL,
					    0 /* byte alignment */,
					    0 /* no page-crossing issues */,
					    GFP_KERNEL | GFP_DMA);
  #else
	small_buffer_cache = 
 		usb_pci_pool_create(
					    dev,
					    SIZE_SMALL,
					    0 /* byte alignment */,
					    0 /* no page-crossing issues */,
					    GFP_KERNEL | GFP_DMA);
 #endif


 USB_DBG_MEM(iprintf(
		"[USB MEM] init_safe_buffers() : small_buffer_cache = %08XH \n",
		small_buffer_cache
		));
     
 if (small_buffer_cache == 0)
   {
	USB_DBG_FLOW(iprintf(
    	"[USB FLOW] init_safe_buffers() End 1 with Fail !!\n"));
	return -1;
   }

 USB_DBG_MEM(iprintf(
	"[USB MEM] init_safe_buffers() : alloc large_buffer_cache..\n"
    ));
    
 #ifdef USB_DBGSTRING_ON
	large_buffer_cache = 
		usb_pci_pool_create("pci_large_buffer",
					    dev,
					    SIZE_LARGE,
					    0 /* byte alignment */,
					    0 /* no page-crossing issues */,
					    GFP_KERNEL | GFP_DMA);
  #else
	large_buffer_cache = 
		usb_pci_pool_create(
					    dev,
					    SIZE_LARGE,
					    0 /* byte alignment */,
					    0 /* no page-crossing issues */,
					    GFP_KERNEL | GFP_DMA);
 #endif
                        
 USB_DBG_MEM(iprintf(
		"[USB MEM] init_safe_buffers() : large_buffer_cache = %08XH \n",
		large_buffer_cache
		));
        
 if (large_buffer_cache == 0)
   {
	USB_DBG_FLOW(iprintf(
    	"[USB FLOW] init_safe_buffers() End 2 with Fail !!\n"));
	return -1;
   }

 #endif // ndef USB_NO_DMA_POOL
 
 USB_DBG_FLOW(iprintf("[USB FLOW] init_safe_buffers() End.\n"));
 return 0;
}

/* allocate a 'safe' buffer and keep track of it */
static char *
alloc_safe_buffer(char *unsafe, int size, dma_addr_t *pbus)
{
 char *safe;
 dma_addr_t busptr;
 struct pci_pool *pool;
 int i;

 USB_DBG_FLOW(iprintf("[USB FLOW] alloc_safe_buffer() Begin ..\n"));

 if (0) printk("alloc_safe_buffer(size=%d)\n", size);

 #ifndef USB_NO_DMA_POOL // 2003/7/2 B. S. Liaw
	if (size <= SIZE_SMALL)
		pool = small_buffer_cache;
	else
		if (size < SIZE_LARGE)
			pool = large_buffer_cache;
				else
					return 0;
 #endif // ndef USB_NO_DMA_POOL

 #ifdef USB_NO_DMA_POOL
	// safe = usb_pci_pool_alloc(&busptr,size);
	safe = (char *) UsbAllocMemU(USB_MEM_U_S
				#ifdef USB_DBG_MEM_U_ON
                 , size
                #endif
                 );
    busptr = (dma_addr_t) safe ;
    
  #else
	safe = usb_pci_pool_alloc(pool, SLAB_ATOMIC, &busptr);
  #endif
  
 if (safe == 0)
   {
	return 0;
   }
 for (i = 0; i < MAX_SAFE; i++)
   {
	if (safe_buffers[i][0] == 0)
       {
		break;
	   }
   } // for i
   
 if (i == MAX_SAFE)
   {
	#ifndef _ETCPIP
		panic(__FILE__ ": exceeded MAX_SAFE buffers");
     #else
		dprintf("[USB] Error : alloc_safe_buffer() : exceeded "
        	"MAX_SAFE buffers \n");
    #endif
   }

 /* place the size index and the old buffer ptr in the first 8 bytes
  * and return a ptr + 12 to caller
  */
 ((int *)safe)[0] = i;
 ((char **)safe)[1] = (char *)pool;
 ((char **)safe)[2] = unsafe;

 busptr += sizeof(int) + sizeof(char *) + sizeof(char *);

 safe_buffers[i][0] = (void *)busptr;
 safe_buffers[i][1] = (void *)safe;

 safe += sizeof(int) + sizeof(char *) + sizeof(char *);

 *pbus = busptr;
 USB_DBG_FLOW(iprintf("[USB FLOW] alloc_safe_buffer() End.\n"));
 return safe;
}

/* determine if a buffer is from our "safe" pool */
static char *
find_safe_buffer(char *busptr, char **unsafe)
{
	int i;
	char *buf;
	

	for (i = 0; i < MAX_SAFE; i++) {
		if (safe_buffers[i][0] == busptr) {
			if (0) printk("find_safe_buffer(%p) found @ %d\n", busptr, i);
			buf = safe_buffers[i][1];
			*unsafe = ((char **)buf)[2];
			return buf + sizeof(int) + sizeof(char *) + sizeof(char *);
		}
	}

	return (char *)0;
}

static void
free_safe_buffer(char *buf)
{
 int index;
 struct pci_pool *pool;
 char *dma;
	

 if (0) printk("free_safe_buffer(buf=%p)\n", buf);

 /* retrieve the buffer size index */
 buf -= sizeof(int) + sizeof(char*) + sizeof(char*);
 index = ((int *)buf)[0];
 pool = (struct pci_pool *)((char **)buf)[1];

 if (0) printk("free_safe_buffer(%p) index %d\n",
		      buf, index);

 if (index < 0 || index >= MAX_SAFE) {
		printk(__FILE__ ": free_safe_buffer() corrupt buffer\n");
		return;
	}

 dma = safe_buffers[index][0];
 safe_buffers[index][0] = 0;

 #ifdef USB_NO_DMA_POOL
	// usb_pci_pool_free(buf);
	UsbFreeMemU(USB_MEM_U_M, (unsigned char *)dma
	    #if defined(USB_DBG_MEM_U_ON)
            , 444
		#endif
        ) ;
  #else
	usb_pci_pool_free(pool, buf, (u32)dma);
 #endif
}

/*
  NOTE:
  replace pci_map/unmap_single with local routines which will
  do buffer copies if buffer is above 1mb...
*/

/*
 * see if a buffer address is in an 'unsafe' range.  if it is
 * allocate a 'safe' buffer and copy the unsafe buffer into it.
 * substitute the safe buffer for the unsafe one.
 * (basically move the buffer from an unsafe area to a safe one)
 *
 * we assume calls to map_single are symmetric with calls to unmap_single...
 */
dma_addr_t
s3c2510_map_single(struct pci_dev *hwdev, void *virtptr,
	       size_t size, int direction)
{
	dma_addr_t busptr;

	mapped_alloc_size += size;

	if (0) printk("pci_map_single(hwdev=%p,ptr=%p,size=%d,dir=%x) "
		      "alloced=%ld\n",
		      hwdev, virtptr, size, direction, mapped_alloc_size);

	busptr = virt_to_bus(virtptr);

	/* we assume here that a buffer will never be >=64k */
	if ( (((unsigned long)busptr) & 0x100000) ||
	     ((((unsigned long)busptr)+size) & 0x100000) )
	{
		char *safe;

		safe = alloc_safe_buffer(virtptr, size, &busptr);
		if (safe == 0) {
			printk("unable to map unsafe buffer %p!\n", virtptr);
			return 0;
		}

		if (0) printk("unsafe buffer %p (phy=%p) mapped to %p (phy=%p)\n",
			      virtptr, (void *)virt_to_bus(virtptr),
			      safe, (void *)busptr);

		memcpy(safe, virtptr, size);
		consistent_sync(safe, size, direction);

		return busptr;
	}

	consistent_sync(virtptr, size, direction);
	return busptr;
}

/*
 * see if a mapped address was really a "safe" buffer and if so,
 * copy the data from the safe buffer back to the unsafe buffer
 * and free up the safe buffer.
 * (basically return things back to the way they should be)
 */
void
s3c2510_unmap_single(struct pci_dev *hwdev, dma_addr_t dma_addr,
		 size_t size, int direction)
{
	char *safe, *unsafe;
	void *buf;

	/* hack; usb-ohci.c never sends hwdev==NULL, all others do */
	if (hwdev == NULL) {
		return;
	}

	mapped_alloc_size -= size;

	if (0)
       {
		#ifdef CONFIG_PCI
        	printk("pci_unmap_single(hwdev=%p,ptr=%p,size=%d,dir=%x) "
		      "alloced=%ld\n",
		      hwdev, (void *)dma_addr, size, direction,
		      mapped_alloc_size);
        #endif
       }
	if ((safe = find_safe_buffer((void *)dma_addr, &unsafe))) {
		if (0) printk("copyback unsafe %p, safe %p, size %d\n",
			      unsafe, safe, size);

		consistent_sync(safe, size, PCI_DMA_FROMDEVICE);
		memcpy(unsafe, safe, size);
		free_safe_buffer(safe);
	} else {
		/* assume this is normal memory */
		buf = bus_to_virt(dma_addr);
		consistent_sync(buf, size, PCI_DMA_FROMDEVICE);
	}
}

#ifndef _ETCPIP
	EXPORT_SYMBOL(s3c2510_map_single);
	EXPORT_SYMBOL(s3c2510_unmap_single);
#endif // ndef _ETCPIP

#ifndef _ETCPIP
  static 
#endif // ndef _ETCPIP
int __init s3c2510_init_safe_buffers(void)
{
 USB_DBG_FLOW(iprintf("[USB FLOW] s3c2510_init_safe_buffers() Begin ..\n"));
 
 USB_DBG_INIT(iprintf(
	"[USB INIT] Initializing S3C2510 buffer pool "
	"for DMA workaround\n"
    )) ;
 
 init_safe_buffers(NULL);
 
 USB_DBG_FLOW(iprintf("[USB FLOW] s3c2510_init_safe_buffers() End.\n"));
 return 0;
}

static void free_safe_buffers(void)
{
 #ifndef USB_NO_DMA_POOL // 2003/7/2 B. S. Liaw
	usb_pci_pool_destroy(small_buffer_cache);
	usb_pci_pool_destroy(large_buffer_cache);
 #endif // ndef USB_NO_DMA_POOL
}

#ifndef _ETCPIP
   module_init(s3c2510_init_safe_buffers);
   module_exit(free_safe_buffers);
#endif // ndef _ETCPIP
