
// Name of File : hcd-init.c
// ====================================================================
/*
 * (C) Copyright David Brownell 2000-2002
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

// #ifdef _ETCPIP
#define __KERNEL__

#ifdef _NEWLIB
     #include "sys_type.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
          // for deciding whether to compile
          //  Debug Version or Release Version
#include "etcpip.h"
#include "ramAddr.h"
#include "UsbDbg.h"
   
    #include "PoList.h" // osport_list.h
   
    #include "UsbCfg.h" // Config and Setting
    #include "UsOsport.h" // "usb_osport.h"
	#ifndef NO_TIMER
		#include "PoTimer.h" // osport_timer.h
	#endif // ndef NO_TIMER

    #ifndef NO_WAITQUEUE
		#include  "PoWait.h" // osport_wait.h
    #endif // ndef NO_WAITQUEUE
    #include "UsMemLib.h"
    
// ohci-hcd.c need USB_HOST_INT
#if defined(_AR7300)
	#include "sysdefs.h"
	#include "if.h"
	#include "hw.h"
	#include "cpmac.h"
	#include "cpmac.h"
	#include "intc_reg.h"
	#include "irqlib.h"
	/*
	#warning: IRQ_HANDLER wrong !!
	#define IRQ_HANDLER
	#warning: IRQ_INT_MODE wrong !!
	#define IRQ_INT_MODE 1
	*/
#elif defined(_NETARM)
        #include "all.h"  // for EXTMASK
        // #include "irqs.h" // for nUSB_HOST  13 USB HOST Interrupt  0x D
        #include "isr.h" // for nUSB_HOST  13 USB HOST Interrupt  0x D
        #include "CfgIsr.h"
#elif defined(_ADM5120)
        typedef unsigned long UINT32;
        #include "adm5120.h"
        // #include <adm5120.h>
        #include "except.h"
        #include "irqlib.h"
        #include "pci_lib.h"
#elif defined(_ATHEROS)
#endif   
// #include "CfgIsr.h"
// #endif // _ETCPIP

#ifndef _ETCPIP
	#include <linux/config.h>
	#include <linux/kernel.h>
	#include <linux/module.h>
#endif // ndef _ETCPIP
    
#ifndef _ETCPIP
	#include <linux/pci.h>
#else
    #include "PoDevice.h" // osport_device.h
	#ifdef CONFIG_PCI
	    #include "UsbPoPci.h" // usb_osport_pci.h
    #endif	// CONFIG_PCI
#endif
    
#ifndef _ETCPIP
	#include <asm/io.h>
	#include <asm/irq.h>
#endif // ndef _ETCPIP


#ifdef CONFIG_USB_DEBUG
	#define DEBUG
#else
	#undef DEBUG
#endif

#ifndef _ETCPIP
	#include <linux/usb.h>
 #else
    #include "usb.h"
#endif
#include "UsbHcd.h"  // ../core/hcd.h


/* PCI-based HCs are normal, but custom bus glue should be ok */

// ====================================================================
/*
#ifdef NO_SYS_FS
	static LIST_HEAD (ohci_hcd_list);
#endif // NO_SYS_FS
*/
// ====================================================================

/*-------------------------------------------------------------------*/

/* configure so an HC device and id are always provided */
/* always called with process context; sleeping is OK */

/**
 * usb_hcd_pci_probe - initialize PCI-based HCDs
 * @dev: USB Host Controller being probed
 * @id: pci hotplug id connecting controller to HCD framework
 * Context: !in_interrupt()
 *
 * Allocates basic PCI resources for this USB host controller, and
 * then invokes the start() method for the HCD associated with it
 * through the hotplug entry's driver_data.
 *
 * Store this function in the HCD's struct pci_driver as probe().
 */
#ifndef _ETCPIP
	int usb_hcd_pci_probe (struct pci_dev *dev,
    		const struct pci_device_id *id )
#elif defined(CONFIG_PCI)
	int usb_hcd_pci_probe (struct pci_dev *dev,
			const struct pci_device_id *id, int irq, void *mem_base)
#else
	int usb_hcd_probe( struct device *dev_dev , int irq,
    				   void *mem_base, 
                       const struct hc_driver *the_hc_driver )
#endif
{
 struct hc_driver	*driver;
 unsigned long		resource, len;
 #ifndef _ETCPIP
	void			*base; 
 #endif // ndef _ETCPIP
 struct usb_hcd		*hcd;
 int			retval ;
 #ifdef CONFIG_PCI
	int region;
 #endif // CONFIG_PCI
 char		buf [8], *bufp = buf;
 #ifdef SOTA_SOFT_ISR   // 2004/1/7 B. S. Liaw
	#ifdef USB_HC_WAIT_BY_EVENT
		int status1 ;
	#endif // USB_HC_WAIT_BY_EVENT
 #endif

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_hcd_pci_probe() Begin ..\n"));
 
 #ifdef CONFIG_PCI
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : dev=%08XHH \n",
	    dev
	    ));
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : *dev=%08XH \n",
	    *dev
	   ));
	/*  
	#ifdef USB_BS_ASSERT
    
    	//if  ( !( (__u32)(*dev) > (__u32)(0x00001000L)
    	//    )  )
    
    
	    if  ( !( (*dev) == (struct device *)(0x00000000)
    	    )  )
	    
    	    {
	         iprintf("[ASSERT][USB] usb_hcd_pci_probe() : " ) ;
    	     iprintf("Error : value *dev !!\n" ) ;
        	 iprintf("[ASSERT][USB] Continue ....\n") ;
	        }
	#endif
	*/   
 #endif // CONFIG_PCI
 
 #ifdef CONFIG_PCI
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : &(dev->dev)=%08XH \n",
    	&(dev->dev)
    	));
  #else
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : &(dev_dev)=%08XH \n",
    	&(dev_dev)
    	));
 #endif
 
 #ifdef CONFIG_PCI
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : dev->dev=%08XH \n",
    	dev->dev
    	));
  #else
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : dev_dev=%08XH \n",
    	dev_dev
    	));
 #endif
 /*   
 #ifdef USB_BS_ASSERT
    
    if  ( !( (__u32)(dev->dev) > (__u32)(0x00001000L)
        )  )
    
    
    //if  ( !( (dev->dev) >= (struct device *)(0x00001000L)
    //    )  )
    
        {
         iprintf("[ASSERT][USB] usb_hcd_pci_probe() : " ) ;
         iprintf("Error : value dev->dev !!\n" ) ;
         iprintf("[ASSERT][USB] Continue ....\n") ;
        }
 #endif
 */
    
 if (usb_disabled())
		return -ENODEV;

 #ifdef CONFIG_PCI
 if (!id || !(driver = (struct hc_driver *) id->driver_data))
		return -EINVAL;
 #else
    driver = (struct hc_driver *) the_hc_driver ;
 #endif
 #ifndef _ETCPIP
	if (pci_enable_device (dev) < 0)
		return -ENODEV;
 #endif // ndef _ETCPIP

 #ifdef CONFIG_PCI
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : &(dev->irq)=%08XH \n",
    	&(dev->irq)
    	));
 #endif // CONFIG_PCI
    
 #ifdef _ETCPIP
    #ifdef CONFIG_PCI
	   dev->irq = irq ;
    #endif 
 #endif // _ETCPIP
	
 #ifdef CONFIG_PCI
	if (!dev->irq)
  #else
 	if (irq <=0 )
 #endif 
	   {
	    /*
		dprintf("[USB] Error : Found HC with no IRQ.  "
    		"Check BIOS/PCI %s setup!\n",
		  dev->slot_name);
	    */
		#ifdef USB_DBGSTRING_ON
		#ifdef CONFIG_PCI
		dprintf("[USB] Error : %s : HC with no IRQ.  !!\n",
		  	dev->slot_name);
		#endif // CONFIG_PCI
        #else
		dprintf("[USB] Error : HC with no IRQ.  !!\n"
		  	);
        #endif
	    return -ENODEV;
	   }
	
 // #ifndef _ETCPIP
 if (driver->flags & HCD_MEMORY) {	// EHCI, OHCI
	/* we read its hardware registers as memory */
 #ifdef CONFIG_PCI
	region = 0;
 #endif // CONFIG_PCI
    
	#ifndef _ETCPIP
		resource = pci_resource_start (dev, 0);
		len = pci_resource_len (dev, 0);
		if (!request_mem_region (resource, len, driver->description))
       {
			dbg ("controller already in use");
			return -EBUSY;
	   }
       
		base = (void *) ioremap_nocache (resource, len);
		if (base == NULL)
       {
			dbg ("error mapping memory");
			retval = -EFAULT;
 clean_1:
			release_mem_region (resource, len);
			err ("init %s fail, %d", dev->slot_name, retval);
			return retval;
	   }
	#endif // ndef _ETCPIP

   } else { 				// UHCI
	resource = len = 0;
	#ifndef _ETCPIP
		for (region = 0; region < PCI_ROM_RESOURCE; region++) 
       {
		if (!(pci_resource_flags (dev, region) & IORESOURCE_IO))
				continue;

		resource = pci_resource_start (dev, region);
		len = pci_resource_len (dev, region);
		if (request_region (resource, len,
					driver->description))
				break;
	   } // for
		if (region == PCI_ROM_RESOURCE) {
			dbg ("no i/o regions available");
			return -EBUSY;
		}
		base = (void *) resource;
	#endif // ndef _ETCPIP
   }
 // #endif // ndef _ETCPIP

 // driver->start(), later on, will transfer device from
 // control by SMM/BIOS to control by Linux (if needed)

 #ifndef _ETCPIP
	pci_set_master (dev);
 #endif // ndef _ETCPIP
 
 hcd = driver->hcd_alloc ();
 
 if (hcd == NULL)
   {
	dbg ("hcd alloc fail");
	retval = -ENOMEM;
        
 clean_2:
 
	#ifndef _ETCPIP
		if (driver->flags & HCD_MEMORY) {
			iounmap (base);
			goto clean_1;
		} else {
			release_region (resource, len);
			err ("init %s fail, %d", dev->slot_name, retval);
			return retval;
		}
	#endif
   }
 /*   
 #ifdef _ETCPIP
	ohci = hcd_to_ohci(hcd);
    
    // 2003/3/31 B. S. Liaw : this need before ISR Enable on hc_start()
    
    glob_ohci = ohci ;
 
    USB_DBG_MEM(iprintf(
        "[USB MEM] hc_found_ohci() : glob_ohci = %08XH \n", glob_ohci
        ));
 #endif // _ETCPIP
 */
 
 #ifdef _ETCPIP
#if _USB2 == 1	// VIA vt6212
	hcd->dev_idx_for_polling = usb_hc_polling_irq_idx((void *)mem_base);
	if (hcd->dev_idx_for_polling < 0) {
		driver->hcd_free (hcd);
		goto clean_2;
	}
#endif
	hcd->irq = -1;
	hcd->regs = mem_base;   
 #endif // _ETCPIP
 
 #ifdef CONFIG_PCI
 USB_DBG_FLOW(iprintf(
	"[USB FLOW] usb_hcd_pci_probe() call pci_set_drvdata() ..\n"));
    
 pci_set_drvdata (dev, hcd);
 #endif // CONFIG_PCI

 hcd->driver = driver;
 #ifdef USB_DBGSTRING_ON
	hcd->description = driver->description;
 #endif // USB_DBGSTRING_ON
 #ifdef CONFIG_PCI
 	hcd->pdev = dev;
 #endif // CONFIG_PCI
 #ifdef USB_DBGSTRING_ON
	#ifdef CONFIG_PCI
		hcd->self.bus_name = dev->slot_name;
	#endif // CONFIG_PCI
 #endif
 
    
 #ifdef CONFIG_PCI
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : &(dev->dev)=%08XH \n",
    	&(dev->dev)
    	));
  #else
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : &(dev_dev)=%08XH \n",
    	&(dev_dev)
    	));
 #endif
 
 #ifdef CONFIG_PCI
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : dev->dev=%08XH \n",
    	dev->dev
    	));
  #else
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : dev_dev=%08XH \n",
    	dev_dev
    	));
 #endif
 
 /*   
 #ifdef USB_BS_ASSERT
    if  ( !( (__u32)(dev->dev) > (__u32)0x00001000L
        )  )
        {
         iprintf("[ASSERT][USB] usb_hcd_pci_probe() : " ) ;
         iprintf("Error : value dev->dev !!\n" ) ;
         iprintf("[ASSERT][USB] Continue ....\n") ;
        }
 #endif
 */
 
 #ifdef USB_DBGSTRING_ON
 #ifdef CONFIG_PCI
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : &(dev->dev.name)=%08XH \n",
    	&(dev->dev.name)
    	));
  #else
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : &(dev_dev->name)=%08XH \n",
    	&(dev_dev->name)
    	));
 #endif

 #ifdef CONFIG_PCI
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : dev->dev.name=%08XH \n",
    	dev->dev.name
    	));
  #else
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : dev_dev->name=%08XH \n",
    	dev_dev->name
    	));
 #endif
 #endif // USB_DBGSTRING_ON

 #ifdef USB_DBGSTRING_ON
 #ifdef CONFIG_PCI
	hcd->product_desc = dev->dev.name;
  #else
	hcd->product_desc = dev_dev->name;
 #endif
 #endif // USB_DBGSTRING_ON

 #ifdef CONFIG_PCI
	hcd->self.controller = &dev->dev;
  #else
	hcd->self.controller = dev_dev;
 #endif
 
 USB_DBG_MEM(iprintf(
	"[USB MEM] usb_hcd_pci_probe() : "
    "hcd->self.controller=%08XH \n",
    hcd->self.controller
    ));
 USB_DBG_MEM(iprintf(
	"[USB MEM] usb_hcd_pci_probe() : "
    "hcd->self.controller->driver=%08XH \n",
    hcd->self.controller->driver
    ));
    
 #ifdef USB_BS_ASSERT
    if  ( !( (__u32)(hcd->self.controller->driver) > (__u32)0x00001000L
        )  )
        {
         iprintf("[ASSERT][USB] usb_hcd_pci_probe() : " ) ;
         iprintf("Error : value hcd->self.controller->driver !!\n" ) ;
         iprintf("[ASSERT][USB] Continue ....\n") ;
        }
 #endif
 
 hcd->controller = hcd->self.controller;

 USB_DBG_FLOW(iprintf(
	"[USB FLOW] usb_hcd_pci_probe() call hcd_buffer_create() ..\n"));
    
 if ((retval = hcd_buffer_create (hcd)) != 0)
   {
    // -----------
	clean_3:
    // -----------
	driver->hcd_free (hcd);
	goto clean_2;
   }
    
 USB_DBG_FLOW(iprintf(
	"[USB FLOW] usb_hcd_pci_probe() after call "
    "hcd_buffer_create()\n"));

 USB_DBG_MEM(iprintf(
	"[USB MEM] &(hcd)=%08XH \n",
	&(hcd)
    )) ;
 USB_DBG_MEM(iprintf(
	"[USB MEM] hcd=%08XH \n",
	hcd
    )) ;
    
 USB_DBG_MEM(iprintf(
	"[USB MEM] usb_hcd_pci_probe() : "
    "&(hcd->controller)=%08XH \n",
    &(hcd->controller)
    ));
 USB_DBG_MEM(iprintf(
	"[USB MEM] usb_hcd_pci_probe() : "
    "hcd->controller=%08XH \n",
    hcd->controller
    ));
 USB_DBG_MEM(iprintf(
	"[USB MEM] usb_hcd_pci_probe() : "
    "hcd->controller->driver=%08XH \n",
    hcd->controller->driver
    ));
    
 #ifdef USB_BS_ASSERT
    if  ( !( (__u32)(hcd->controller->driver) > (__u32)0x00001000L
        )  )
        {
         iprintf("[ASSERT][USB] usb_hcd_pci_probe() : " ) ;
         iprintf("Error : value hcd->controller->driver !!\n" ) ;
         iprintf("[ASSERT][USB] Continue ....\n") ;
        }
 #endif
 
 #ifdef USB_DBGSTRING_ON
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_pci_probe() : "
    	"&(hcd->controller->driver->name)=%08XH \n",
    	&(hcd->controller->driver->name)
    	));
 #endif

 #ifdef USB_DBGSTRING_ON
	USB_DBG_MEM(iprintf(
		"[USB MEM] &(hcd->product_desc)=%08XH \n",
		&(hcd->product_desc)
    	)) ;
	USB_DBG_MEM(iprintf(
		"[USB MEM]  hcd->product_desc=%08XH \n",
		hcd->product_desc
    	)) ;
 #endif // USB_DBGSTRING_ON
 USB_DBG_MEM(iprintf(
	"[USB MEM] &(hcd->controller->bus_id)=%08XH \n",
	&(hcd->controller->bus_id)
    )) ;
 USB_DBG_MEM(iprintf(
	"[USB MEM] hcd->controller->bus_id=%08XH \n",
	hcd->controller->bus_id
    )) ;
 
 #ifdef USB_DBGSTRING_ON
	dev_info (hcd->controller, "%s\n", hcd->product_desc);
  #else
	dev_info (hcd->controller, "HCD PCI\n");
 #endif

 #ifndef __sparc__
	#ifdef CONFIG_PCI
		sprintf (buf, "%d", dev->irq);
	 #else
		sprintf (buf, "%d", irq);
	 #endif

  #else
	bufp = __irq_itoa(dev->irq);
 #endif
 /*
 #ifndef _ETCPIP
    
	if (request_irq (dev->irq, usb_hcd_irq, SA_SHIRQ,
    				 hcd->description, hcd)
			!= 0)
       {
		dev_err (hcd->controller,
				"request interrupt %s failed\n", bufp);
		retval = -EBUSY;
		goto clean_3;
	   }
    
  #else
 */

 #ifdef SOTA_SOFT_ISR   // 2004/1/7 B. S. Liaw
	#ifdef USB_HC_WAIT_BY_EVENT
		 #ifdef INT_SCHEDULING
			status1 = runtsk(100, (void (fAR *)(void))sota_soft_isr, 6000);
		  #else
			status1 = runtsk(80, (void (fAR *)(void))sota_soft_isr, 6000);
		 #endif // INT_SCHEDULING

		 if (status1 < SUCCESS ) {
			dprintf("RUNTASK sota_soft_isr fails\n");
		    // goto exi2;
		   }
		  else {
    		 dprintf("RUNTASK id=%d sota_soft_isr ...\n", status1);
		   }
	 #else
		hc_wdh = 0 ;
	#endif // ndef USB_HC_WAIT_BY_EVENT
 #endif // SOTA_SOFT_ISR
 /*
 //interrupt disable
 USB_DBG_INIT(iprintf("[USB INIT] HwMaskAllInt ..\n"));
 HwMaskAllInt() ;
 */
 // Register USB interrupt service routine 
 USB_DBG_INIT(iprintf(
       "[USB INIT] Register USB interrupt service routine ..\n"));
 /*
 // #define nUSB_HOST       13  // USB HOST Interrupt       0x D
 USB_DBG_INIT(iprintf(
       "[USB INIT] nUSB_HOST = %08XH \n",nUSB_HOST));
 SysSetInterrupt(nUSB_HOST, hc_interrupt);
 */
 USB_DBG_INIT(iprintf(
       "[USB INIT] usb_hcd_pci_probe() : irq = %d \n",irq));
       
 #ifdef USB_BS_ASSERT
		#ifdef USB_OHCI
		#ifdef _AR7300
		if  ( !( irq == 0x04
           )  )
		#else
		if  ( !( irq == 13
           )  )
		#endif
		#elif defined(_ADM5120)
		if  ( !( irq == 0x03
           )  )
        #endif
           {
            iprintf("[ASSERT][USB] usb_hcd_pci_probe() : " ) ;
            iprintf("Error : irq value wrong : %02XH !!\n",
             		irq ) ;
            iprintf("[ASSERT][USB] Continue ....\n") ;
           }
 #endif
    
 #ifdef USB_DBG_SIMULTAE_VERIFY
		/*
    	irqConnect(INT_LVL_USBHOST, IRQ_INT_MODE, (IRQ_HANDLER) UsbIsr, 0, 0);   
    	IrqEnable(INT_LVL_USBHOST);
    	*/
 #endif // USB_DBG_SIMULTAE_VERIFY
 
 // #ifndef USB_INT_SCHEDULING
    
 // B. S. Liaw : Only Regist ISR
 // B. S. Liaw : Donot Try to Enable USB interrupt Now !!
 
 // #ifdef USB_OHCI
 #if _USB2 == 1		// VIA VT6212(PCI interface)
		usb_hc_install_interrupt(irq);
 #elif defined(_S3C2510)
	    // SysSetInterrupt(irq, hc_interrupt);
    	SysSetInterrupt(irq, localhc_irq);
 #elif defined(_AR7300)
/*		irqConnect(irq, IRQ_INT_MODE,
    			localhc_irq, 0, 0); */

 // #elif defined(USB_SOTA)   // Sota USB 1.1 Host
 #elif defined(_ADM5120)
    	/*
		irqConnect(INT_LVL_USBHOST, IRQ_INT_MODE,
    			(IRQ_HANDLER) sota_irq, 0, 0);
    	*/
		irqConnect(irq, IRQ_INT_MODE,
    			(IRQ_HANDLER) localhc_irq, 0, 0);
 #elif defined(_ATHEROS)
 #endif 
        
 // #endif // ndef USB_INT_SCHEDULING
    
 #ifdef CONFIG_PCI
    	dev->irq =  irq ;
 #endif // USB_OHCI
 /*
 #endif
 */
 #ifdef CONFIG_PCI
	hcd->irq = dev->irq;
  #else
	hcd->irq = irq;
 #endif

 #ifndef _ETCPIP
	hcd->regs = base;
 #endif // ndef _ETCPIP
 
 #ifdef CONFIG_PCI
	hcd->region = region;
 #endif // CONFIG_PCI
 
 #ifndef _ETCPIP
 	dev_info (hcd->controller, "irq %s, %s %p\n", bufp,
		(driver->flags & HCD_MEMORY) ? "pci mem" : "io base",
		base);
  #else
 	dev_info(hcd->controller, "irq %s, %s %p\n", bufp,
		(driver->flags & HCD_MEMORY) ? "reg-mem" : "io base",
		mem_base);
 #endif
        
 #ifdef USB_DBG_ABORT   
   // vvv for debug Prefetch/Data Abort Error !!
   
	#ifndef NO_TIMER
	   dprintf("[TEST] usb_hcd_pci_probe() 1 : jiffies = %08XH\n",jiffies) ;
	#endif // ndef NO_TIMER
   
   {
    unsigned long * __temp_ptr = 0 ;
    
	#if !defined(_ADM5120)
    	dprintf(
        	"[TEST] usb_hcd_pci_probe() :  *000000000H = %08XH \n",
        	* __temp_ptr
        	);
    #endif
   }
   dprintf("[TEST] usb_hcd_pci_probe() : GetTime() = %08XH\n",
           GetTime()) ;
   if(!in_interrupt())
     {
      dprintf("[TEST] usb_hcd_pci_probe() : dlytsk()/YIELD() ..\n") ;

		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] wait_ms() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
		#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
			// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
         #else
        	YIELD();
        #endif
     }
    else
     {
      dprintf("[TEST] usb_hcd_pci_probe() : in interrupt !!\n");
     }
   // ^^^ for debug Prefetch/Data Abort Error !!  
 #endif   


 USB_DBG_FLOW(iprintf(
	"[USB FLOW] usb_hcd_pci_probe() call usb_bus_init() ..\n"));
 usb_bus_init (&hcd->self);
 
 hcd->self.op = &usb_hcd_operations;
 hcd->self.hcpriv_hcd = (void *) hcd;

 USB_DBG_HC(iprintf("[USB HC] usb_hcd_probe() : hcd= %08XH \n",hcd )) ;
 USB_DBG_HC(iprintf("[USB HC] usb_hcd_probe() : hcd->self.hcpriv_hcd= %08XH \n",hcd->self.hcpriv_hcd )) ;

 INIT_LIST_HEAD (&hcd->dev_list);

 #ifdef _ETCPIP
    /*
	#ifdef NO_SYS_FS
		list_add (&hcd->dev_list, &ohci_hcd_list);
	#endif // NO_SYS_FS
    */
	/*
	INIT_LIST_HEAD (&ohci->timeout_list);
	*/
	USB_DBG_INIT(iprintf(
    	"[USB INIT] usb_hcd_pci_probe() : usb_hcd_operations."
        "deallocate = %08XH\n",
    	usb_hcd_operations.deallocate
    	));
 #endif // _ETCPIP
 
 usb_register_bus (&hcd->self);

 if ((retval = driver->start (hcd)) < 0)
   {
	#ifndef _ETCPIP
		usb_hcd_pci_remove (dev);
	#endif // ndef _ETCPIP
   }

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_hcd_pci_probe() End.\n"));
 return retval;
} 

#ifndef _ETCPIP
	EXPORT_SYMBOL (usb_hcd_pci_probe);
#endif // ndef _ETCPIP


/* may be called without controller electrically present */
/* may be called with controller, bus, and devices active */

/**
 * usb_hcd_pci_remove - shutdown processing for PCI-based HCDs
 * @dev: USB Host Controller being removed
 * Context: !in_interrupt()
 *
 * Reverses the effect of usb_hcd_pci_probe(), first invoking
 * the HCD's stop() method.  It is always called from a thread
 * context, normally "rmmod", "apmd", or something similar.
 *
 * Store this function in the HCD's struct pci_driver as remove().
 */
#ifdef CONFIG_PCI
#ifndef _ETCPIP
	void usb_hcd_pci_remove (struct pci_dev *dev)
	{
	 struct usb_hcd		*hcd;
	 struct usb_device	*hub;

	 hcd = pci_get_drvdata(dev);
	 if (!hcd)
		return;
	 dev_info (hcd->controller, "remove, state %x\n", hcd->state);

	 if (in_interrupt ()) BUG ();

	 hub = hcd->self.root_hub;
	 USB_DBG_INIT(iprintf(
		"[USB INIT] usb_hcd_pci_remove() : Set Value : hcd->state = "
	    "USB_STATE_QUIESCING\n"));
	 hcd->state = USB_STATE_QUIESCING;

	 dev_dbg (*hcd->controller, "roothub graceful disconnect\n");
	 usb_disconnect (&hub);

	 hcd->driver->stop (hcd);
	 hcd_buffer_destroy (hcd);
    
	 USB_DBG_INIT(iprintf(
		"[USB INIT] usb_hcd_pci_remove() : Set Value : hcd->state = "
	    "USB_STATE_HALT\n"));
	 hcd->state = USB_STATE_HALT;
 
	 pci_set_drvdata (dev, 0);

	 free_irq (hcd->irq, hcd);
	 if (hcd->driver->flags & HCD_MEMORY) {
		iounmap (hcd->regs);
		release_mem_region (pci_resource_start (dev, 0),
			pci_resource_len (dev, 0));
	 } else {
		release_region (pci_resource_start (dev, hcd->region),
			pci_resource_len (dev, hcd->region));
	 }

	 usb_deregister_bus (&hcd->self);
	 if (atomic_read (&hcd->self.refcnt) != 1) {
		dev_warn (hcd->controller,
			"dangling refs (%d) to bus %d!\n",
			atomic_read (&hcd->self.refcnt) - 1,
			hcd->self.busnum);
	 }
	 hcd->driver->hcd_free (hcd);
	}
	EXPORT_SYMBOL (usb_hcd_pci_remove);
#endif // ndef _ETCPIP
#endif // CONFIG_PCI


#ifdef	CONFIG_PM

/*
 * Some "sleep" power levels imply updating struct usb_driver
 * to include a callback asking hcds to do their bit by checking
 * if all the drivers can suspend.  Gets involved with remote wakeup.
 *
 * If there are pending urbs, then HCs will need to access memory,
 * causing extra power drain.  New sleep()/wakeup() PM calls might
 * be needed, beyond PCI suspend()/resume().  The root hub timer
 * still be accessing memory though ...
 *
 * FIXME:  USB should have some power budgeting support working with
 * all kinds of hubs.
 *
 * FIXME:  This assumes only D0->D3 suspend and D3->D0 resume.
 * D1 and D2 states should do something, yes?
 *
 * FIXME:  Should provide generic enable_wake(), calling pci_enable_wake()
 * for all supported states, so that USB remote wakeup can work for any
 * devices that support it (and are connected via powered hubs).
 *
 * FIXME:  resume doesn't seem to work right any more...
 */


// 2.4 kernels have issued concurrent resumes (w/APM)
// we defend against that error; PCI doesn't yet.

/**
 * usb_hcd_pci_suspend - power management suspend of a PCI-based HCD
 * @dev: USB Host Controller being suspended
 *
 * Store this function in the HCD's struct pci_driver as suspend().
 */
#ifdef CONFIG_PCI
   int usb_hcd_pci_suspend (struct pci_dev *dev, u32 state)
   {
	struct usb_hcd		*hcd;
	int			retval;

	hcd = pci_get_drvdata(dev);
	dev_info (hcd->controller, "suspend to state %d\n", state);

	pci_save_state (dev, hcd->pci_state);

	// FIXME for all connected devices, leaf-to-root:
	// driver->suspend()
	// proposed "new 2.5 driver model" will automate that

 /* driver may want to disable DMA etc */
 retval = hcd->driver->suspend (hcd, state);
    
 USB_DBG_INIT(iprintf(
	"[USB INIT] usb_hcd_pci_suspend() : Set Value : hcd->state = "
    "USB_STATE_SUSPENDED\n"));
 hcd->state = USB_STATE_SUSPENDED;

 pci_set_power_state (dev, state);
 return retval;
   }

	#ifndef _ETCPIP
		EXPORT_SYMBOL (usb_hcd_pci_suspend);
	#endif // ndef _ETCPIP
#endif // CONFIG_PCI

/**
 * usb_hcd_pci_resume - power management resume of a PCI-based HCD
 * @dev: USB Host Controller being resumed
 *
 * Store this function in the HCD's struct pci_driver as resume().
 */
#ifdef CONFIG_PCI
   int usb_hcd_pci_resume (struct pci_dev *dev)
   {
	struct usb_hcd		*hcd;
	int			retval;

	hcd = pci_get_drvdata(dev);
	dev_info (hcd->controller, "resume\n");

	/* guard against multiple resumes (APM bug?) */
	atomic_inc (&hcd->resume_count);
	if (atomic_read (&hcd->resume_count) != 1) {
		dev_err (hcd->controller, "concurrent PCI resumes\n");
		retval = 0;
		goto done;
	}

	retval = -EBUSY;
	if (hcd->state != USB_STATE_SUSPENDED) {
		dev_dbg (*hcd->controller, "can't resume, not suspended!\n");
		goto done;
	}
 USB_DBG_INIT(iprintf(
	"[USB INIT] usb_hcd_pci_resume() : Set Value : hcd->state = "
    "USB_STATE_RESUMING\n"));
    
 hcd->state = USB_STATE_RESUMING;

	pci_set_power_state (dev, 0);
	pci_restore_state (dev, hcd->pci_state);

	retval = hcd->driver->resume (hcd);
	if (!HCD_IS_RUNNING (hcd->state)) {
		dev_dbg (*hcd->controller, "resume fail, retval %d\n", retval);
		usb_hc_died (hcd);
// FIXME:  recover, reset etc.
	} else {
		// FIXME for all connected devices, root-to-leaf:
		// driver->resume ();
		// proposed "new 2.5 driver model" will automate that
	}

done:
	atomic_dec (&hcd->resume_count);
	return retval;
}

	#ifndef _ETCPIP
		EXPORT_SYMBOL (usb_hcd_pci_resume);
	#endif // ndef _ETCPIP
#endif // CONFIG_PCI

#endif	/* CONFIG_PM */


