
// Name of File : hcd.c
// ====================================================================

/*
 * (C) Copyright Linus Torvalds 1999
 * (C) Copyright Johannes Erdfelt 1999-2001
 * (C) Copyright Andreas Gal 1999
 * (C) Copyright Gregory P. Smith 1999
 * (C) Copyright Deti Fliegl 1999
 * (C) Copyright Randy Dunlap 2000
 * (C) Copyright David Brownell 2000-2002
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#ifdef _ETCPIP
   	#define __KERNEL__
   #ifdef _NEWLIB
     #include "sys_type.h"
   #endif

   	        #ifdef _NEWLIB
          #include "string.h"
        #endif
   	#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
          // for deciding whether to compile
          //  Debug Version or Release Version
   	#include "etcpip.h"

   	#include "UsbDbg.h"
    #include "UsbSysInit.h"
   	#include "PoList.h" // osport_list.h
   
   	#include "UsbCfg.h" // Config and Setting
   	#include "UsOsport.h" // "usb_osport.h"
	#ifndef NO_TIMER
		#include "PoTimer.h" // osport_timer.h
	#endif // ndef NO_TIMER
   
   	#ifndef NO_WAITQUEUE
		#include  "PoWait.h" // osport_wait.h
   	#endif // ndef NO_WAITQUEUE
    /*
   	#include "irqs.h" // for nUSB_HOST  13 USB HOST Interrupt  0x D
   	#include "isr.h" // for nUSB_HOST  13 USB HOST Interrupt  0x D
   	#include "CfgIsr.h"
    */
	#include "UsMemLib.h" // UsbMemLib.h
#endif // _ETCPIP

#ifndef _ETCPIP
	#include <linux/config.h>
#endif // ndef _ETCPIP

#ifdef CONFIG_USB_DEBUG
	#define DEBUG
#else
	#undef DEBUG
#endif

#ifndef _ETCPIP
	#include <linux/module.h>
	#include <linux/version.h>
	#include <linux/kernel.h>
	#include <linux/slab.h>
	#include <linux/completion.h>
	#include <linux/uts.h>			/* for UTS_SYSNAME */
#endif // ndef _ETCPIP

#ifndef _ETCPIP
	#include <linux/pci.h>			/* for hcd->pdev and dma addressing */
#else
    #include "PoDevice.h" // osport_device.h
	#ifdef CONFIG_PCI
	    #include "UsbPoPci.h" // usb_osport_pci.h
    #endif	// CONFIG_PCI
#endif
    
#ifndef _ETCPIP
	#include <linux/dma-mapping.h>
	#include <asm/byteorder.h>
#endif // ndef _ETCPIP


#ifndef _ETCPIP
	#include <linux/usb.h>
 #else
   #include "usb.h"
#endif
#ifndef _ETCPIP
	#include "hcd.h"
 #else
	#include "UsbHcd.h"  // ../core/hcd.h
#endif




// #define USB_BANDWIDTH_MESSAGES

/*-------------------------------------------------------------------------*/

/*
 * USB Host Controller Driver framework
 *
 * Plugs into usbcore (usb_bus) and lets HCDs share code, minimizing
 * HCD-specific behaviors/bugs.
 *
 * This does error checks, tracks devices and urbs, and delegates to a
 * "hc_driver" only for code (and data) that really needs to know about
 * hardware differences.  That includes root hub registers, i/o queues,
 * and so on ... but as little else as possible.
 *
 * Shared code includes most of the "root hub" code (these are emulated,
 * though each HC's hardware works differently) and PCI glue, plus request
 * tracking overhead.  The HCD code should only block on spinlocks or on
 * hardware handshaking; blocking on software events (such as other kernel
 * threads releasing resources, or completing actions) is all generic.
 *
 * Happens the USB 2.0 spec says this would be invisible inside the "USBD",
 * and includes mostly a "HCDI" (HCD Interface) along with some APIs used
 * only by the hub driver ... and that neither should be seen or used by
 * usb client device drivers.
 *
 * Contributors of ideas or unattributed patches include: David Brownell,
 * Roman Weissgaerber, Rory Bolt, Greg Kroah-Hartman, ...
 *
 * HISTORY:
 * 2002-02-21	Pull in most of the usb_bus support from usb.c; some
 *		associated cleanup.  "usb_hcd" still != "usb_bus".
 * 2001-12-12	Initial patch version for Linux 2.5.1 kernel.
 */

/*-------------------------------------------------------------------------*/

/* host controllers we manage */

LIST_HEAD (usb_bus_list);
/*
#ifndef _ETCPIP
	EXPORT_SYMBOL_GPL (usb_bus_list);
#endif // ndef _ETCPIP
*/
// #ifndef NO_SYS_FS
	/* used when allocating bus numbers */
	#define USB_MAXBUS		64
	struct usb_busmap {
		unsigned long busmap [USB_MAXBUS / (8*sizeof (unsigned long))];
	};
	static struct usb_busmap busmap;
// #endif // ndef NO_SYS_FS

/* used when updating list of hcds */
#ifndef NO_SEMPHORE
	DECLARE_MUTEX (usb_bus_list_lock);	/* exported only for usbfs */
	EXPORT_SYMBOL_GPL (usb_bus_list_lock);
#endif // ndef NO_SEMPHORE

/* used when updating hcd data */
#ifndef NO_SPINLOCK
	static spinlock_t hcd_data_lock = SPIN_LOCK_UNLOCKED;
#endif // ndef NO_SPINLOCK

int hcd_init_ok = 0 ;

/*-------------------------------------------------------------------------*/

/*
 * Sharable chunks of root hub code.
 */

/*-------------------------------------------------------------------------*/

#ifndef _ETCPIP
	#define KERNEL_REL	((LINUX_VERSION_CODE >> 16) & 0x0ff)
	#define KERNEL_VER	((LINUX_VERSION_CODE >> 8) & 0x0ff)
#endif // ndef _ETCPIP

/* usb 2.0 root hub device descriptor */
static const u8 usb2_rh_dev_descriptor [18] = {
	0x12,       /*  __u8  bLength; */
	0x01,       /*  __u8  bDescriptorType; Device */
	0x00, 0x02, /*  __u16 bcdUSB; v2.0 */

	0x09,	    /*  __u8  bDeviceClass; HUB_CLASSCODE */
	0x00,	    /*  __u8  bDeviceSubClass; */
	0x01,       /*  __u8  bDeviceProtocol; [ usb 2.0 single TT ]*/
	0x08,       /*  __u8  bMaxPacketSize0; 8 Bytes */

	0x00, 0x00, /*  __u16 idVendor; */
 	0x00, 0x00, /*  __u16 idProduct; */
 #ifndef _ETCPIP
	KERNEL_VER, KERNEL_REL, /*  __u16 bcdDevice */
  #else
    0x013 , 0x01 ,
 #endif

	0x03,       /*  __u8  iManufacturer; */
	0x02,       /*  __u8  iProduct; */
	0x01,       /*  __u8  iSerialNumber; */
	0x01        /*  __u8  bNumConfigurations; */
};

/* no usb 2.0 root hub "device qualifier" descriptor: one speed only */

/* usb 1.1 root hub device descriptor */
static const u8 usb11_rh_dev_descriptor [18] = {
	0x12,       /*  __u8  bLength; */
	0x01,       /*  __u8  bDescriptorType; Device */
	0x10, 0x01, /*  __u16 bcdUSB; v1.1 */

	0x09,	    /*  __u8  bDeviceClass; HUB_CLASSCODE */
	0x00,	    /*  __u8  bDeviceSubClass; */
	0x00,       /*  __u8  bDeviceProtocol; [ low/full speeds only ] */
	0x08,       /*  __u8  bMaxPacketSize0; 8 Bytes */

	0x00, 0x00, /*  __u16 idVendor; */
 	0x00, 0x00, /*  __u16 idProduct; */
 #ifndef _ETCPIP
	KERNEL_VER, KERNEL_REL, /*  __u16 bcdDevice */
  #else
    0x13 , 0x01 ,
 #endif

	0x03,       /*  __u8  iManufacturer; */
	0x02,       /*  __u8  iProduct; */
	0x01,       /*  __u8  iSerialNumber; */
	0x01        /*  __u8  bNumConfigurations; */
};


/*-------------------------------------------------------------------------*/

/* Configuration descriptors for our root hubs */

static const u8 fs_rh_config_descriptor [] = {

	/* one configuration */
	0x09,       /*  __u8  bLength; */
	0x02,       /*  __u8  bDescriptorType; Configuration */
	0x19, 0x00, /*  __u16 wTotalLength; */
	0x01,       /*  __u8  bNumInterfaces; (1) */
	0x01,       /*  __u8  bConfigurationValue; */
	0x00,       /*  __u8  iConfiguration; */
	0x40,       /*  __u8  bmAttributes; 
				 Bit 7: Bus-powered,
				     6: Self-powered,
				     5 Remote-wakwup,
				     4..0: resvd */
	0x00,       /*  __u8  MaxPower; */
      
	/* USB 1.1:
	 * USB 2.0, single TT organization (mandatory):
	 *	one interface, protocol 0
	 *
	 * USB 2.0, multiple TT organization (optional):
	 *	two interfaces, protocols 1 (like single TT)
	 *	and 2 (multiple TT mode) ... config is
	 *	sometimes settable
	 *	NOT IMPLEMENTED
	 */

	/* one interface */
	0x09,       /*  __u8  if_bLength; */
	0x04,       /*  __u8  if_bDescriptorType; Interface */
	0x00,       /*  __u8  if_bInterfaceNumber; */
	0x00,       /*  __u8  if_bAlternateSetting; */
	0x01,       /*  __u8  if_bNumEndpoints; */
	0x09,       /*  __u8  if_bInterfaceClass; HUB_CLASSCODE */
	0x00,       /*  __u8  if_bInterfaceSubClass; */
	0x00,       /*  __u8  if_bInterfaceProtocol; [usb1.1 or single tt] */
	0x00,       /*  __u8  if_iInterface; */
     
	/* one endpoint (status change endpoint) */
	0x07,       /*  __u8  ep_bLength; */
	0x05,       /*  __u8  ep_bDescriptorType; Endpoint */
	0x81,       /*  __u8  ep_bEndpointAddress; IN Endpoint 1 */
 	0x03,       /*  __u8  ep_bmAttributes; Interrupt */
 	0x02, 0x00, /*  __u16 ep_wMaxPacketSize; 1 + (MAX_ROOT_PORTS / 8) */
	0xff        /*  __u8  ep_bInterval; (255ms -- usb 2.0 spec) */
};

static const u8 hs_rh_config_descriptor [] = {

	/* one configuration */
	0x09,       /*  __u8  bLength; */
	0x02,       /*  __u8  bDescriptorType; Configuration */
	0x19, 0x00, /*  __u16 wTotalLength; */
	0x01,       /*  __u8  bNumInterfaces; (1) */
	0x01,       /*  __u8  bConfigurationValue; */
	0x00,       /*  __u8  iConfiguration; */
	0x40,       /*  __u8  bmAttributes; 
				 Bit 7: Bus-powered,
				     6: Self-powered,
				     5 Remote-wakwup,
				     4..0: resvd */
	0x00,       /*  __u8  MaxPower; */
      
	/* USB 1.1:
	 * USB 2.0, single TT organization (mandatory):
	 *	one interface, protocol 0
	 *
	 * USB 2.0, multiple TT organization (optional):
	 *	two interfaces, protocols 1 (like single TT)
	 *	and 2 (multiple TT mode) ... config is
	 *	sometimes settable
	 *	NOT IMPLEMENTED
	 */

	/* one interface */
	0x09,       /*  __u8  if_bLength; */
	0x04,       /*  __u8  if_bDescriptorType; Interface */
	0x00,       /*  __u8  if_bInterfaceNumber; */
	0x00,       /*  __u8  if_bAlternateSetting; */
	0x01,       /*  __u8  if_bNumEndpoints; */
	0x09,       /*  __u8  if_bInterfaceClass; HUB_CLASSCODE */
	0x00,       /*  __u8  if_bInterfaceSubClass; */
	0x00,       /*  __u8  if_bInterfaceProtocol; [usb1.1 or single tt] */
	0x00,       /*  __u8  if_iInterface; */
     
	/* one endpoint (status change endpoint) */
	0x07,       /*  __u8  ep_bLength; */
	0x05,       /*  __u8  ep_bDescriptorType; Endpoint */
	0x81,       /*  __u8  ep_bEndpointAddress; IN Endpoint 1 */
 	0x03,       /*  __u8  ep_bmAttributes; Interrupt */
 	0x02, 0x00, /*  __u16 ep_wMaxPacketSize; 1 + (MAX_ROOT_PORTS / 8) */
	0x0c        /*  __u8  ep_bInterval; (256ms -- usb 2.0 spec) */
};

/*-------------------------------------------------------------------------*/

/*
 * helper routine for returning string descriptors in UTF-16LE
 * input can actually be ISO-8859-1; ASCII is its 7-bit subset
 */
static int ascii2utf (char *s, u8 *utf, int utfmax)
{
	int retval;

	for (retval = 0; *s && utfmax > 1; utfmax -= 2, retval += 2) {
		*utf++ = *s++;
		*utf++ = 0;
	}
	return retval;
}

/*
 * rh_string - provides manufacturer, product and serial strings for root hub
 * @id: the string ID number (1: serial number, 2: product, 3: vendor)
 * @hcd: the host controller for this root hub
 * @type: string describing our driver 
 * @data: return packet in UTF-16 LE
 * @len: length of the return packet
 *
 * Produces either a manufacturer, product or serial number string for the
 * virtual root hub device.
 */
static int rh_string (
	int		id,
	struct usb_hcd	*hcd,
	u8		*data,
	int		len
) {
	char buf [100];

	// language ids
	if (id == 0) {
		*data++ = 4; *data++ = 3;	/* 4 bytes string data */
		*data++ = 0x09; *data++ = 0x04;	/* MSFT-speak for "en-us" */
		return 4;

	// serial number
	}
     else if (id == 1)
     {
		#ifdef USB_DBGSTRING_ON
			strcpy (buf, hcd->self.bus_name);
		#endif

	// product description
	} else if (id == 2) {
	#ifdef USB_DBGSTRING_ON
                strcpy (buf, hcd->product_desc);
    #endif

 	// id 3 == vendor description
	} else if (id == 3) {
 #ifndef _ETCPIP
        sprintf (buf, "%s %s %s", UTS_SYSNAME, UTS_RELEASE,
			hcd->description);
  #else
 #ifdef USB_DBGSTRING_ON
        sprintf (buf, "[?] [?] %s",	hcd->description);
  #else
        sprintf (buf, "[?] [?] HCD ");
 #endif // USB_DBGSTRING_ON
 #endif

	// unsupported IDs --> "protocol stall"
   }
  else
   {
	return 0;
   }
 data [0] = 2 * (strlen (buf) + 1);
 data [1] = 3;	/* type == string */
 return 2 + ascii2utf (buf, data + 2, len - 2);
}


/* Root hub control transfers execute synchronously */

static int rh_call_control(struct usb_hcd *hcd, struct urb *urb)
{
 struct usb_ctrlrequest *cmd = (struct usb_ctrlrequest *) urb->setup_packet;
 u16		typeReq, wValue, wIndex, wLength;
 const u8	*bufp = 0;
 u8		*ubuf = urb->transfer_buffer;
 int		len = 0;
 #ifdef RH_INT_BY_POLL
    int time_out ;
 #endif


 USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] rh_call_control() Begin ..\n"));
 
 USB_DBG_MEM(iprintf("[USB MEM] rh_call_control() : urb->setup_packet = %08XH \n",urb->setup_packet )) ;
        
 typeReq  = (cmd->bRequestType << 8) | cmd->bRequest;
 wValue   = le16_to_cpu (cmd->wValue);
 wIndex   = le16_to_cpu (cmd->wIndex);
 wLength  = le16_to_cpu (cmd->wLength);

 #ifdef USB_DBG_MEM_AREA_VERIFY_ON
	USB_DBG_AREA_MSG(iprintf("[USB AREA] rh_call_control(1) : Verify Memory Area ..\n")) ;
		if (! VerifyMemArea() ) {
			iprintf("[USB] rh_call_control(1) Verify Memory Area Error !\n") ;
	       }
 #endif // USB_DBG_MEM_AREA_VERIFY_ON

 #ifdef RH_INT_BY_POLL
 	/*
	USB_DBG_ROOTHUB(iprintf(
       "[USB ROOTHUB] rh_call_control() : Wait RH Free ..\n"
       ));
       
    for (time_out=RH_WAITFREE_TIME ; time_out > 0 ; time_out-- )
	   {
	    if (hub_in_submit == 0 )
           {
            break ;
           }
          else
	       {
	        // iprintf("[USB] usb_start_wait_urb() : getres Fail !!") ;
			#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
				// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
             #else
            	YIELD();
            #endif
	        continue ;
	       }
	   } // for 
       
    if (hub_in_submit != 0 )
        {
	     dprintf("[USB] Wait RH Free Time Out Fail !!") ;
        }
        
    hub_in_submit = 1 ;
    
	USB_DBG_ROOTHUB(iprintf(
    	"[USB ROOTHUB] rh_call_control() : Set Value : hub_in_submit=%d \n",
        hub_in_submit
        )) ;
    */
 #endif
 
 #ifdef RH_INT_BY_POLL
 	/*
	hub_in_submit = 0 ;
	USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] rh_call_control() : Set Value : hub_in_submit=%d \n", hub_in_submit )) ;
    */
 #endif //RH_INT_BY_POLL


 if (wLength > urb->transfer_buffer_length)
   {
	goto error;
   }

 /* set up for success */
 urb->status = 0;
 urb->actual_length = wLength;
 switch (typeReq) {

	/* DEVICE REQUESTS */

	case DeviceRequest | USB_REQ_GET_STATUS:
		// DEVICE_REMOTE_WAKEUP
		ubuf [0] = 1; // selfpowered
		ubuf [1] = 0;
			/* FALLTHROUGH */
	case DeviceOutRequest | USB_REQ_CLEAR_FEATURE:
	case DeviceOutRequest | USB_REQ_SET_FEATURE:
		dev_dbg (hcd->controller, "no device features yet yet\n");
		break;
	case DeviceRequest | USB_REQ_GET_CONFIGURATION:
		ubuf [0] = 1;
			/* FALLTHROUGH */
	case DeviceOutRequest | USB_REQ_SET_CONFIGURATION:
		break;
	case DeviceRequest | USB_REQ_GET_DESCRIPTOR:
		switch (wValue & 0xff00) {
		case USB_DT_DEVICE << 8:
			if (hcd->driver->flags & HCD_USB2)
				bufp = usb2_rh_dev_descriptor;
			else if (hcd->driver->flags & HCD_USB11)
				bufp = usb11_rh_dev_descriptor;
			else
				goto error;
			len = 18;
			break;
		case USB_DT_CONFIG << 8:
			if (hcd->driver->flags & HCD_USB2) {
				bufp = hs_rh_config_descriptor;
				len = sizeof hs_rh_config_descriptor;
			} else {
				bufp = fs_rh_config_descriptor;
				len = sizeof fs_rh_config_descriptor;
			}
			break;
		case USB_DT_STRING << 8:
			urb->actual_length = rh_string (
				wValue & 0xff, hcd,
				ubuf, wLength);
			break;
		default:
			goto error;
		}
		break;
	case DeviceRequest | USB_REQ_GET_INTERFACE:
		ubuf [0] = 0;
			/* FALLTHROUGH */
	case DeviceOutRequest | USB_REQ_SET_INTERFACE:
		break;
	/*
	case RH_SET_ADDRESS:
*/
	
	case DeviceOutRequest | USB_REQ_SET_ADDRESS:
        USB_DBG_CTLMSG(iprintf(
            "[USB CTLMSG] rh_call_control() : RH_SET_ADDRESS \n"));
		// wValue == urb->dev->devaddr
		#ifdef USB_DBG_INIT_ON
			dev_dbg (hcd->controller, "root hub device address %d\n",
				wValue);
        #endif
		break;

	/* INTERFACE REQUESTS (no defined feature/status flags) */

	/* ENDPOINT REQUESTS */

	case EndpointRequest | USB_REQ_GET_STATUS:
		// ENDPOINT_HALT flag
		ubuf [0] = 0;
		ubuf [1] = 0;
			/* FALLTHROUGH */
	case EndpointOutRequest | USB_REQ_CLEAR_FEATURE:
	case EndpointOutRequest | USB_REQ_SET_FEATURE:
		dev_dbg (hcd->controller, "no endpoint features yet\n");
		break;

	/* CLASS REQUESTS (and errors) */

	default:
		/* non-generic request */
		urb->status = hcd->driver->hub_control (hcd,
			typeReq, wValue, wIndex,
			ubuf, wLength);
		break;
error:
		/* "protocol stall" on error */
		urb->status = -EPIPE;
		dev_dbg (hcd->controller, "unsupported hub control message (maxchild %d)\n",
				urb->dev->maxchild);
	}
	if (urb->status) {
		urb->actual_length = 0;
		dev_dbg (hcd->controller, "CTRL: TypeReq=0x%x val=0x%x idx=0x%x len=%d ==> %d\n",
			typeReq, wValue, wIndex, wLength, urb->status);
	}
	if (bufp) {
		if (urb->transfer_buffer_length < len)
			len = urb->transfer_buffer_length;
		urb->actual_length = len;
		// always USB_DIR_IN, toward host
		memcpy (ubuf, bufp, len);
	}

 #ifdef USB_NO_ISR_FREE
    /*
	USB_DBG_FLOWTIMCRY(iprintf(
		"[USB FLOWTIMCRY] rh_call_control() Call hcd->driver->"
        "urb_free_priv() ...\n"
    	));
	
 	//USB_DBG_FLOW(iprintf(
    // "[USB FLOW] rh_call_control() call urb_free_priv() ... \n" )) ;
 	
 	urb_free_priv(ohci, urb->hcpriv);
	// hcd->driver->urb_free_priv(hcd, urb->hcpriv);
	hcd->driver->urb_free_priv(hcd, urb);
	urb->hcpriv = NULL;
    */
 #endif // USB_NO_ISR_FREE
 
 #ifdef USB_DBG_MEM_AREA_VERIFY_ON
	USB_DBG_AREA_MSG(iprintf("[USB AREA] rh_call_control(2) : Verify Memory Area ..\n")) ;
		if (! VerifyMemArea() ) {
			iprintf("[USB] rh_call_control(2) Verify Memory Area Error !\n") ;
	       }
 #endif // USB_DBG_MEM_AREA_VERIFY_ON

 /* any errors get returned through the urb completion */
 
 USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] rh_call_control() call usb_hcd_giveback_urb() ..\n" )) ;
 /*   
 #ifndef _ETCPIP
	usb_hcd_giveback_urb (hcd, urb, NULL);
 #else
 */
 usb_hcd_giveback_urb (hcd, urb);
 /*
 #endif
 */

 #ifdef USB_WAIT_BY_EVENT  // 2003/12/31 B. S. Liaw
	// MT_EVT_USB is on kernel/local.h
	USB_DBG_INIT(iprintf("[USB INIT] rh_call_control() : SIGNAL(MT_EVT_USB) .. \n")) ;
	// USB_DBG_INFO(iprintf("[USB INFO] rh_call_control() : SIGNAL(MT_EVT_USB) .. \n")) ;

	// kernel/mtmacro.h : #define SIGNAL_IR(signo) MTqcmd_c(SETEVT, signo)
	if (SIGNAL(MT_EVT_USB) != SUCCESS) {
	// if (MTqcmd_c(SETEVT,MT_EVT_USB) != SUCCESS) {
		USB_DBG_INFO(iprintf(
			"[USB INFO] rh_call_control() : Error : SIGNAL MT_EVT_USB=%d fail !! \n",
			MT_EVT_USB
			)) ;
		dprintf("[USB] Error : SIGNAL MT_EVT_USB %d fail !! \n",
			MT_EVT_USB ) ;
	   }
	/*
	// MTqcmd_c(SETEVT,MT_EVT_USB) ;
	SIGNAL_IR(MT_EVT_USB) ;
	*/
 #endif // USB_WAIT_BY_EVENT

 #ifdef USB_DBG_MEM_AREA_VERIFY_ON
	USB_DBG_AREA_MSG(iprintf("[USB AREA] rh_call_control(3) : Verify Memory Area ..\n")) ;
		if (! VerifyMemArea() ) {
			iprintf("[USB] rh_call_control(3) Verify Memory Area Error !\n") ;
	       }
 #endif // USB_DBG_MEM_AREA_VERIFY_ON

 USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] rh_call_control() End.\n"));
 return 0;
}

/*-------------------------------------------------------------------------*/

/*
 * Root Hub interrupt transfers are synthesized with a timer.
 * Completions are called in_interrupt() but not in_irq().
 */

#ifndef NO_TIMER
	static void rh_report_status (unsigned long ptr);
 #else
    #ifdef RH_POLL_TSK
		extern void usb_rh_task(unsigned long ptr);
    #else
		extern void rh_report_status(struct urb	*urb);
    #endif
#endif

static int rh_status_urb (struct usb_hcd *hcd, struct urb *urb) 
{
 // int	len = 1 + (urb->dev->maxchild / 8);
 int	len = 0 ;
 #ifdef NO_TIMER
	int status1 ;
 #endif // NO_TIMER

 USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] rh_status_urb() Begin ..\n"));
    
 if ( ! hcd_init_ok ) {
    USB_DBG_INFO(iprintf("[USB INFO] rh_status_urb(1) : HCD init Not ok !! \n")) ;
    return 0 ;
   }


 if (!hcd)
   {
    USB_DBG_INFO(iprintf("[USB INFO] rh_status_urb() : Error : Null urb !!\n"))
	return 0 ;
   }
 if (!urb)
   {
    USB_DBG_INFO(iprintf("[USB INFO] rh_status_urb() : Error : Null urb !!\n"))
	return 0 ;
   }
 if (!urb->dev)
   {
    USB_DBG_INFO(iprintf("[USB INFO] rh_status_urb() : Error : Null urb->dev !!\n"))
	return 0 ;
   }
 if (!urb->dev->maxchild)
   {
    USB_DBG_INFO(iprintf("[USB INFO] rh_status_urb() : Error : Null urb->dev->maxchild !!\n"))
	return 0 ;
   }
   
 len = 1 + (urb->dev->maxchild / 8);
    
 /* rh_timer protected by hcd_data_lock */
 #ifndef NO_TIMER
	if (  timer_pending (&hcd->rh_timer)
  #else
	if (  0
 #endif
		|| urb->status != -EINPROGRESS
		|| urb->transfer_buffer_length < len) 
	   {
	    /*
		dev_dbg (hcd->controller, "not queuing status urb, stat %d\n", urb->status);
        */
		dprintf("[USB] Root HUB not queuing status urb, stat %d\n", urb->status);
		return -EINVAL;
	   }

 USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] rh_status_urb() : Set Vaue : urb->hcpriv = hcd ..\n"));
    
 urb->hcpriv = hcd;	/* nonzero to indicate it's queued */
 
 #ifndef NO_TIMER
	init_timer (&hcd->rh_timer);
	hcd->rh_timer.function = rh_report_status;
	hcd->rh_timer.data = (unsigned long) urb;
	/* USB 2.0 spec says 256msec; this is close enough */
	hcd->rh_timer.expires = jiffies + HZ/4;
	add_timer (&hcd->rh_timer);
  #else
    hcd->rh_urb = urb ;
    /*
    usb_rh_task((unsigned long) urb) ;
    */
    
    #ifdef RH_POLL_TSK
	#ifdef INT_SCHEDULING
	    status1 = runtsk(100,(void (fAR *)(void))usb_rh_task, 6000,
        					(unsigned long) urb);
	#else
	    status1 = runtsk(80, (void (fAR *)(void))usb_rh_task, 6000,
        					(unsigned long) urb); // 100 too high !!
	#endif // INT_SCHEDULING

    if (status1 < SUCCESS )
       {
        dprintf("[USB] Error : RUNTASK usb_rh_task fails !!\n");
       }
      else
       {
        USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] RUNTASK id=%d usb_rh_task ...\n", status1));
       }
    #else

	if ( ! hcd_init_ok )
       {
        dprintf("[USB] rh_status_urb(2) : HCD init Not ok !! \n") ;
		// dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
	    // continue ;
        return 0 ;
	   }
	 /*
	 if(! in_interrupt())
	   {
	    BLOCK_PREEMPTION;
		USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] rh_status_urb() : mt_busy=%d\n",mt_busy)) ;
	   }
	*/

	#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] rh_status_urb(1) : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
	#endif // USB_BS_ASSERT

	USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] rh_status_urb() : rh_report_status() ..\n"));
    rh_report_status(urb);

	#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] rh_status_urb(2) : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
	#endif // USB_BS_ASSERT

	/*
	 if(! in_interrupt())
	   {
		USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] rh_status_urb() : mt_busy=%d\n",mt_busy)) ;
	    UNBLOCK_PREEMPTION;
	   }
    */
    #endif
 #endif
 USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] rh_status_urb() End. \n"));
 return 0;
}

/* timer callback */

// Lower Layer Task !!

#ifndef NO_TIMER
	static void rh_report_status (unsigned long ptr)
#else
    #ifdef RH_POLL_TSK
	void usb_rh_task(unsigned long ptr)
    #else
		extern void rh_report_status(struct urb	*urb)
    #endif
#endif
{
 #ifdef RH_POLL_TSK
 struct urb	*urb;
 #endif
 struct usb_hcd	*hcd;
 int		length;
 unsigned long	flags;


 USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] usb_rh_task() Begin ..\n"));
 
 USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] usb_rh_task() : mt_busy=%d\n",mt_busy)) ;

 #ifdef USB_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
    	     )
		   )
		   {
			iprintf("[ASSERT][USB] usb_rh_task() : " ) ;
    	    iprintf("Error : mt_busy Not Zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
 #endif // USB_BS_ASSERT

 #ifdef USB_DBG_MEM_AREA_VERIFY_ON
	USB_DBG_AREA_MSG(iprintf("[USB AREA] usb_rh_task() : Verify Memory Area ..\n")) ;
		if (! VerifyMemArea() ) {
			iprintf("[USB] usb_rh_task() Verify Memory Area Error !\n") ;
	       }
 #endif // USB_DBG_MEM_AREA_VERIFY_ON

 #ifdef RH_POLL_TSK
	if (!ptr) {
    USB_DBG_INFO(iprintf("[USB INFO] usb_rh_task() : Error : Null ptr !!\n"))
	return;
   }
 
 #endif // RH_POLL_TSK

 #ifdef RH_POLL_TSK
 #ifdef NO_TIMER
	// if ( ! hcd->rh_urb )
	while ( ! hcd_init_ok )
       {
		#ifdef RH_POLL_TSK
          dprintf("[USB] usb_rh_task(1) : RH wait for HCD init ok .. \n") ;
		#endif // def RH_POLL_TSK
		// usb_hub_events();
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] rh_report_status() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
	    continue ;
	   }
 #endif
   #else
	if ( ! hcd_init_ok )
       {
        dprintf("[USB] rh_report_status() : HCD Not init ok !! \n") ;
		// dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
	    // continue ;
        return ;
	   }
 #endif
 
 #ifdef RH_POLL_TSK
 #ifdef NO_TIMER
	#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] usb_rh_task(1) : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
	#endif // USB_BS_ASSERT
	// USB 2.0 spec says 256msec 
	// USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] usb_rh_task() : wait 256 ms ..\n"));
	USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] usb_rh_task() : wait 256 ms ..\n"));
	wait_ms(256) ; 
	// wait_ms(3000) ; 
	// USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] usb_rh_task() : wait 256 ms end \n"));
	USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] usb_rh_task() : wait 256 ms end \n"));
 #endif // NO_TIMER
    
 urb = (struct urb *) ptr;
 #endif

 if (!urb)
   {
    USB_DBG_INFO(iprintf("[USB INFO] usb_rh_task() : Error : Null urb !!\n"))
	return;
   }

 #ifndef NO_SPINLOCK
 	/*
	spin_lock_irqsave (&urb->lock, flags);
    */
 #endif // ndef NO_SPINLOCK
 if (!urb->dev)
   {
	#ifndef NO_SPINLOCK
		spin_unlock_irqrestore (&urb->lock, flags);
	#endif // ndef NO_SPINLOCK
    USB_DBG_INFO(iprintf("[USB INFO] usb_rh_task() : Error : Null urb->dev !!\n"))
	return;
   }
   
 if (!urb->dev->bus)
   {
    USB_DBG_INFO(iprintf("[USB INFO] usb_rh_task() : Error : Null urb->dev->bus !!\n"))
	return;
   }
 if (!urb->dev->bus->hcpriv_hcd)
   {
    USB_DBG_INFO(iprintf("[USB INFO] usb_rh_task() : Error : Null urb->dev->bus->hcpriv_hcd !!\n"))
	return;
   }
 if (!urb->transfer_buffer)
   {
    USB_DBG_INFO(iprintf("[USB INFO] usb_rh_task() : Error : Null urb->transfer_buffer !!\n"))
	return;
   }

 hcd = urb->dev->bus->hcpriv_hcd;
 
 /*
 USB_DBG_HC(iprintf("[USB HC] usb_rh_task() : hcd=%08XH \n",hcd )) ;
 USB_DBG_HC(iprintf("[USB HC] usb_rh_task() : hcd->state=%04XH \n",hcd->state )) ;
 */

 #ifdef USB_BS_ASSERT
	if ( !(     ( hcd->state
                  & ( ~(__ACTIVE | __SLEEPY | __SUSPEND | __TRANSIENT )
                    )
                )
             == 0x00
       )  ) {
		iprintf("[ASSERT][USB] usb_rh_task(2) : " ) ;
        iprintf("Error : hcd->state Crashed !!\n") ;
        iprintf("[ASSERT][USB] hcd=%08XH , &(hcd->state)=%08XH , hcd->state=%04XH \n",hcd,&(hcd->state) , hcd->state) ;
		iprintf("[ASSERT][USB] Continue .. \n" ) ;
	   }   
 #endif

 #ifdef USB_BS_ASSERT
	if (!(  (unsigned long) urb->transfer_buffer >= (unsigned long) 0x010000L
         )
	   )
	   {
		iprintf("[ASSERT][USB] usb_rh_task(3) : " ) ;
        iprintf("Error : Invalid urb->transfer_buffer !! %08XH\n",urb->transfer_buffer) ;
	    iprintf("[ASSERT][USB] Continue ..\n") ;
       }
 #endif // USB_BS_ASSERT

 #ifdef USB_BS_ASSERT
	if (!(  (unsigned long) hcd >= (unsigned long) 0x010000L
         )
	   )
	   {
		iprintf("[ASSERT][USB] usb_rh_task(4) : " ) ;
        iprintf("Error : Invalid hcd !! %08XH\n",hcd) ;
	    iprintf("[ASSERT][USB] Continue ..\n") ;
       }
 #endif // USB_BS_ASSERT
 
 if (urb->status == -EINPROGRESS)
   {
	if (HCD_IS_RUNNING (hcd->state))
       {
		length = hcd->driver->hub_status_data(hcd,
					urb->transfer_buffer);
                    
		#ifndef NO_SPINLOCK
        	/*
			spin_unlock_irqrestore (&urb->lock, flags);
            */
		#endif // ndef NO_SPINLOCK
            
		USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] usb_rh_task() : length=%d \n" , length)) ;
            
		if (length >= 0)
           {
			urb->actual_length = length;
			urb->status = 0;
			USB_DBG_ROOTHUB_LOOP(iprintf(
				"[USB ROOTHUB_LOOP] usb_rh_task() : Set Vaue : urb->hcpriv = 0 ..\n"
		    	));
			urb->hcpriv = 0;
            
			#ifdef USB_NO_ABC // No usb_api_blocking_completion()
				USB_DBG_CALLBACK(iprintf("[USB CALLBACK] usb_rh_task() Call usb_urb_complete() ...\n")) ;
				USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] usb_rh_task() Call usb_urb_complete() ...\n")) ;
				usb_urb_complete(urb) ; // for urb->awd_done
			#endif // USB_NO_ABC

			#ifdef RH_POLL_TSK
			 if(urb->callback)
			   {
				USB_DBG_CALLBACK(iprintf("[USB CALLBACK] usb_rh_task() : Call urb->callback() .. \n")) ;
				USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] usb_rh_task() : Call urb->callback() .. \n")) ;
				#ifndef _ETCPIP
        	    	/*
					urb->complete (urb, NULL);
                	*/
	             #else
					urb->callback(urb);
        	    #endif
			   }
			  else
			   {
				USB_DBG_CALLBACK(iprintf("[USB CALLBACK] usb_rh_task() : Null urb->callback() !! \n" ));
			   }
			#endif // RH_POLL_TSK

		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] usb_rh_task(5) : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
			   }
		#endif // USB_BS_ASSERT

			USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] usb_rh_task() 1 End.\n"));
			return;
           }
          else
           {
		    USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] usb_rh_task() : length < 0 !!\n")) ;
		   }
	   }
      else
       {
		USB_DBG_INFO(iprintf("[USB INFO] usb_rh_task() : Error : HCD Not Running !!\n"))
		USB_DBG_INFO(iprintf("[USB INFO] usb_rh_task() : hcd= %08XH \n",hcd )) ;
		USB_DBG_INFO(iprintf("[USB INFO] usb_rh_task() : hcd->state= %04XH \n",hcd->state )) ;
		#ifndef NO_SPINLOCK
        	/*
			spin_unlock_irqrestore (&urb->lock, flags);
            */
		#endif // ndef NO_SPINLOCK
 	   }

	/* retrigger timer until completion:  success or unlink */

    #ifdef RH_POLL_TSK // 2003/9/8 B. S. Liaw , use usb_rh_task
    // #error: Please Make Sure : RH_POLL_TSK

    // B. S. Liaw BugFix , RH_POLL_TSK !! Donot Comment Out !!
    // rh_status_urb() is caller of this function , donot call it !!
    //                 otherwise lead mt_busy to overflow

	#ifndef NO_SPINLOCK
    	
		// spin_lock_irqsave (&hcd_data_lock, flags);
        
     #else
        BLOCK_PREEMPTION;
	#endif // ndef NO_SPINLOCK
    
	USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] usb_rh_task() : before mt_busy=%d\n",mt_busy)) ;
	USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] usb_rh_task() : rh_status_urb ..\n"));
    
	rh_status_urb(hcd, urb);
    
	USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] usb_rh_task() : after mt_busy=%d\n",mt_busy)) ;
	#ifndef NO_SPINLOCK
        
		// spin_unlock_irqrestore (&hcd_data_lock, flags);
        
     #else
        UNBLOCK_PREEMPTION;
	#endif // ndef NO_SPINLOCK
    

	/*
	if (urb->callback) {
		USB_DBG_CALLBACK(iprintf("[USB CALLBACK] usb_rh_task() : Call urb->callback() ...\n" )) ;
		urb->callback(urb);
	   }
	  else {
		USB_DBG_CALLBACK(iprintf("[USB CALLBACK] usb_rh_task() : Null urb->callback() !! \n" )) ;
	   }
    */
    #endif // RH_POLL_TSK
   }
  else
   {
	/* this urb's been unlinked */
	USB_DBG_ROOTHUB_LOOP(iprintf(
		"[USB ROOTHUB_LOOP] usb_rh_task() : Set Vaue : urb->hcpriv = 0 ..\n"
    	));
        
	urb->hcpriv = 0;
    
	#ifndef NO_SPINLOCK
		spin_unlock_irqrestore (&urb->lock, flags);
	#endif // ndef NO_SPINLOCK

	#ifdef USB_NO_ISR_FREE
    /*
	USB_DBG_FLOWTIMCRY(iprintf(
		"[USB FLOWTIMCRY] usb_rh_task() Call hcd->driver->urb_free_priv() ...\n"
    	));
	
 	//USB_DBG_FLOW(iprintf(
    // "[USB FLOW] rh_call_control() call urb_free_priv() ... \n" )) ;
 	
    
 	// urb_free_priv(ohci, urb->hcpriv);
	// hcd->driver->urb_free_priv(hcd, urb->hcpriv);
	hcd->driver->urb_free_priv(hcd, urb);
	urb->hcpriv = NULL;
    */
	#endif // USB_NO_ISR_FREE
 
	USB_DBG_ROOTHUB_LOOP(iprintf(
		"[USB ROOTHUB_LOOP] usb_rh_task() call usb_hcd_giveback_urb() ..\n"
	    ));
	/*
	#ifndef _ETCPIP
    	
		usb_hcd_giveback_urb (hcd, urb, NULL);
        
	#else
        */
		usb_hcd_giveback_urb(hcd, urb);
	/*
	#endif
	*/
   }
   
 USB_DBG_ROOTHUB_LOOP(iprintf("[USB ROOTHUB_LOOP] usb_rh_task() 2 End.\n"));
}

/*-------------------------------------------------------------------------*/

static int rh_urb_enqueue (struct usb_hcd *hcd, struct urb *urb)
{
 int retval;
 
 USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] rh_urb_enqueue() Begin ..\n"));
 
 #ifdef USB_BS_SYSCAL_ASSERT
	if (!(  mt_busy == 0
       ) ) {
		iprintf("[ASSERT][USB] rh_urb_enqueue() : " ) ;
	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
	    iprintf("[ASSERT][USB] Continue ..\n") ;
	   }
 #endif // USB_BS_ASSERT
 
 #ifdef USB_DBG_MEM_AREA_VERIFY_ON
	USB_DBG_AREA_MSG(iprintf("[USB AREA] rh_urb_enqueue(1) : Verify Memory Area ..\n")) ;
		if (! VerifyMemArea() ) {
			iprintf("[USB] rh_urb_enqueue(1) Verify Memory Area Error !\n") ;
	       }
 #endif // USB_DBG_MEM_AREA_VERIFY_ON

 #ifdef USB_BS_ASSERT
	if (!(     (unsigned long) &(urb->pipe) 
            >= (unsigned long) 0x010000L
         )
	   )
	   {
		iprintf("[ASSERT][USB] rh_urb_enqueue() : " ) ;
        iprintf("Error : Zero &(urb->pipe) = %08XH !!\n",&(urb->pipe) ) ;
	    iprintf("[ASSERT][USB] Continue ..\n") ;
       }
 #endif // USB_BS_ASSERT

 if (usb_pipeint (urb->pipe))
   {
	// int		retval;
	unsigned long	flags;

	#ifdef RH_POLL_TSK
	#ifndef NO_SPINLOCK
		spin_lock_irqsave (&hcd_data_lock, flags);
     #else
        BLOCK_PREEMPTION;
	#endif // ndef NO_SPINLOCK
	#endif // RH_POLL_TSK
    
	#ifdef USB_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
    	     )
		   )
		   {
			iprintf("[ASSERT][USB] rh_urb_enqueue(1) : " ) ;
    	    iprintf("Error : mt_busy Not Zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif // USB_BS_ASSERT

	USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] rh_urb_enqueue() : rh_status_urb() ..\n"));
	retval = rh_status_urb(hcd, urb);
    
	#ifdef USB_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
    	     )
		   )
		   {
			iprintf("[ASSERT][USB] rh_urb_enqueue(2) : " ) ;
    	    iprintf("Error : mt_busy Not Zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif // USB_BS_ASSERT

	#ifdef RH_POLL_TSK
	#ifndef NO_SPINLOCK
		spin_unlock_irqrestore (&hcd_data_lock, flags);
     #else
        UNBLOCK_PREEMPTION;
	#endif // ndef NO_SPINLOCK
	#endif // RH_POLL_TSK
	USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] rh_urb_enqueue() End 1. \n"));
	return retval;
   }

 #ifdef USB_DBG_MEM_AREA_VERIFY_ON
	USB_DBG_AREA_MSG(iprintf("[USB AREA] rh_urb_enqueue(2) : Verify Memory Area ..\n")) ;
		if (! VerifyMemArea() ) {
			iprintf("[USB] rh_urb_enqueue(2) Verify Memory Area Error !\n") ;
	       }
 #endif // USB_DBG_MEM_AREA_VERIFY_ON

 if (usb_pipecontrol (urb->pipe))
   {
    /*
	USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] rh_urb_enqueue() End with rh_call_control() !!\n"));
	return rh_call_control (hcd, urb);
	*/
	USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] rh_urb_enqueue() : rh_call_control() ..\n"));
	retval = rh_call_control (hcd, urb);

	#ifdef USB_DBG_MEM_AREA_VERIFY_ON
		USB_DBG_AREA_MSG(iprintf("[USB AREA] rh_urb_enqueue(3) : Verify Memory Area ..\n")) ;
		if (! VerifyMemArea() ) {
			iprintf("[USB] rh_urb_enqueue(3) Verify Memory Area Error !\n") ;
	       }
	#endif // USB_DBG_MEM_AREA_VERIFY_ON

	USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] rh_urb_enqueue() End 2 . \n"));
	return retval ;
   }
  else
   {
	USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] rh_urb_enqueue() End with Fail !!\n"));
	return -EINVAL;
   }
}

/*-------------------------------------------------------------------------*/

void usb_rh_status_dequeue (struct usb_hcd *hcd, struct urb *urb)
{
 unsigned long	flags;
 

 #ifndef NO_SPINLOCK
	spin_lock_irqsave (&hcd_data_lock, flags);
 #endif // ndef NO_SPINLOCK
 
 #ifndef NO_TIMER
	del_timer_sync (&hcd->rh_timer);
	hcd->rh_timer.data = 0;
  #else
    hcd->rh_urb = 0 ;
 #endif
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore (&hcd_data_lock, flags);
 #endif // ndef NO_SPINLOCK

 #ifdef USB_NO_ISR_FREE
    /*
	USB_DBG_FLOWTIMCRY(iprintf(
		"[USB FLOWTIMCRY] usb_rh_status_dequeue() Call hcd->driver->urb_free_priv() ...\n"
    	));
	
 	// USB_DBG_FLOW(iprintf(
    //	"[USB FLOW] rh_call_control() call urb_free_priv() ... \n" )) ;
 	
 	// urb_free_priv(ohci, urb->hcpriv);
	// hcd->driver->urb_free_priv(hcd, urb->hcpriv);
	hcd->driver->urb_free_priv(hcd, urb);
	urb->hcpriv = NULL;
    */
 #endif // USB_NO_ISR_FREE
 
 /* we rely on RH callback code not unlinking its URB! */
 USB_DBG_FLOWTIMCRY(iprintf(
	"[USB FLOWTIMCRY] usb_rh_status_dequeue() call usb_hcd_giveback_urb() ..\n"
    ));
 /*
 #ifndef _ETCPIP
	usb_hcd_giveback_urb (hcd, urb, NULL);
 #else
 */
	usb_hcd_giveback_urb (hcd, urb);
 /*
 #endif
 */
}

/*-------------------------------------------------------------------------*/

/* exported only within usbcore */
void usb_bus_get (struct usb_bus *bus)
{
	atomic_inc (&bus->refcnt);
}

/* exported only within usbcore */
void usb_bus_put (struct usb_bus *bus)
{
 if (atomic_dec_and_test (&bus->refcnt))
   {
	kfree (bus);
   }
}

/*-------------------------------------------------------------------------*/

/**
 * usb_bus_init - shared initialization code
 * @bus: the bus structure being initialized
 *
 * This code is used to initialize a usb_bus structure, memory for which is
 * separately managed.
 */
void usb_bus_init (struct usb_bus *bus)
{
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_bus_init() Begin ..\n"));
	memset (&bus->devmap, 0, sizeof(struct usb_devmap));

	bus->devnum_next = 1;

	bus->root_hub = NULL;
	bus->hcpriv_hcd = NULL;
	bus->busnum = -1;
	bus->bandwidth_allocated = 0;
	bus->bandwidth_int_reqs  = 0;
	bus->bandwidth_isoc_reqs = 0;

	INIT_LIST_HEAD (&bus->bus_list);

	atomic_set (&bus->refcnt, 1);
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_bus_init() End.\n"));
}
#ifndef _ETCPIP
	EXPORT_SYMBOL (usb_bus_init);
#endif // ndef _ETCPIP

/**
 * usb_alloc_bus - creates a new USB host controller structure
 * @op: pointer to a struct usb_operations that this bus structure should use
 * Context: !in_interrupt()
 *
 * Creates a USB host controller bus structure with the specified 
 * usb_operations and initializes all the necessary internal objects.
 *
 * If no memory is available, NULL is returned.
 *
 * The caller should call usb_free_bus() when it is finished with the structure.
 */
struct usb_bus *usb_alloc_bus (struct usb_operations *op)
{
 struct usb_bus *bus;


 USB_DBG_FLOW(iprintf("[USB FLOW] usb_alloc_bus() Begin ..\n"));
 
 USB_DBG_MEM(iprintf(
	"[USB MEM] usb_alloc_bus(): alloc bus .. \n"
    ));
 #ifndef _ETCPIP
    #error : Please Make Sure : Not _ETCPIP
	bus = kmalloc (sizeof *bus, GFP_KERNEL);
  #else
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    // BLOCK_PREEMPTION;

    bus = (struct usb_bus *)
		#ifdef USB_DBG_MEM_C_ON
				UsbAllocMemC(USB_MEM_C_L, sizeof *bus ) ;
         #else
          UsbAllocMemC(USB_MEM_C_L) ;
        #endif
                 
    // UNBLOCK_PREEMPTION;
    USB_DBG_MEM(iprintf("[USB MEM] bus = %08XH \n",
       bus ));

    if ( bus == NULL )
       {
		dprintf("[USB] couldn't allocate bus memory \n");
	   }
 #endif
 
 if (!bus)
		return NULL;
 usb_bus_init (bus);
 bus->op = op;
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_alloc_bus() End.\n"));
 return bus;
}
#ifndef _ETCPIP
	EXPORT_SYMBOL (usb_alloc_bus);
#endif // ndef _ETCPIP

/**
 * usb_free_bus - frees the memory used by a bus structure
 * @bus: pointer to the bus to free
 *
 * To be invoked by a HCD, only as the last step of decoupling from
 * hardware.  It is an error to call this if the reference count is
 * anything but one.  That would indicate that some system component
 * did not correctly shut down, and thought the hardware was still
 * accessible.
 */
#ifndef _ETCPIP
void usb_free_bus (struct usb_bus *bus)
{
	if (!bus)
		return;
	if (atomic_read (&bus->refcnt) != 1)
		err ("usb_free_bus #%d, count != 1", bus->busnum);
	usb_bus_put (bus);
}
EXPORT_SYMBOL (usb_free_bus);
#endif // ndef _ETCPIP

/*-------------------------------------------------------------------------*/

/**
 * usb_register_bus - registers the USB host controller with the usb core
 * @bus: pointer to the bus to register
 * Context: !in_interrupt()
 *
 * Assigns a bus number, and links the controller into usbcore data
 * structures so that it can be seen by scanning the bus list.
 */
void usb_register_bus(struct usb_bus *bus)
{
 int busnum;

 #ifndef NO_SEMPHORE
	down (&usb_bus_list_lock);
 #endif // ndef NO_SEMPHORE
 busnum = find_next_zero_bit (busmap.busmap, USB_MAXBUS, 1);
 if (busnum < USB_MAXBUS) {
		set_bit (busnum, busmap.busmap);
		bus->busnum = busnum;
	} else
		warn ("too many buses");

	usb_bus_get (bus);

	/* Add it to the list of buses */
	list_add (&bus->bus_list, &usb_bus_list);
 #ifndef NO_SEMPHORE
	up (&usb_bus_list_lock);
 #endif // ndef NO_SEMPHORE

 usbfs_add_bus (bus);

 dev_info (bus->controller, 
			"new USB bus registered, assigned bus number %d\n",
	         bus->busnum);
}
#ifndef _ETCPIP
	EXPORT_SYMBOL (usb_register_bus);
#endif // ndef _ETCPIP

/**
 * usb_deregister_bus - deregisters the USB host controller
 * @bus: pointer to the bus to deregister
 * Context: !in_interrupt()
 *
 * Recycles the bus number, and unlinks the controller from usbcore data
 * structures so that it won't be seen by scanning the bus list.
 */
void usb_deregister_bus (struct usb_bus *bus)
{
	dev_info (bus->controller, "USB bus %d deregistered\n", bus->busnum);

	/*
	 * NOTE: make sure that all the devices are removed by the
	 * controller code, as well as having it call this when cleaning
	 * itself up
	 */
 #ifndef NO_SEMPHORE
	down (&usb_bus_list_lock);
 #endif // ndef NO_SEMPHORE
 
 list_del (&bus->bus_list);
 
 #ifndef NO_SEMPHORE
	up (&usb_bus_list_lock);
 #endif // ndef NO_SEMPHORE

 usbfs_remove_bus (bus);

 clear_bit (bus->busnum, busmap.busmap);

 usb_bus_put (bus);
}
#ifndef _ETCPIP
	EXPORT_SYMBOL (usb_deregister_bus);
#endif // ndef _ETCPIP

/**
 * usb_register_root_hub - called by HCD to register its root hub 
 * @usb_dev: the usb root hub device to be registered.
 * @parent_dev: the parent device of this root hub.
 *
 * The USB host controller calls this function to register the root hub
 * properly with the USB subsystem.  It sets up the device properly in
 * the driverfs tree, and then calls usb_new_device() to register the
 * usb device.
 */
int usb_register_root_hub (struct usb_device *usb_dev, struct device *parent_dev)
{
 int retval;

 #ifdef USB_DBGSTRING_ON
	sprintf (&usb_dev->dev.bus_id[0], "usb%d", usb_dev->bus->busnum);
  #else
	usb_dev->dev.bus_id[0] = 0x00 ;
 #endif
 retval = usb_new_device (usb_dev, parent_dev);
 if (retval)
   {
	dev_err (parent_dev, "can't register root hub for %s, %d\n",
				usb_dev->dev.bus_id, retval);
   }
 return retval;
}
#ifndef _ETCPIP
	EXPORT_SYMBOL (usb_register_root_hub);
#endif // ndef _ETCPIP


/*-------------------------------------------------------------------------*/

/**
 * usb_calc_bus_time - approximate periodic transaction time in nanoseconds
 * @speed: from dev->speed; USB_SPEED_{LOW,FULL,HIGH}
 * @is_input: true iff the transaction sends data to the host
 * @isoc: true for isochronous transactions, false for interrupt ones
 * @bytecount: how many bytes in the transaction.
 *
 * Returns approximate bus time in nanoseconds for a periodic transaction.
 * See USB 2.0 spec section 5.11.3; only periodic transfers need to be
 * scheduled in software, this function is only used for such scheduling.
 */
long usb_calc_bus_time (int speed, int is_input, int isoc, int bytecount)
{
	unsigned long	tmp;

	switch (speed) {
	case USB_SPEED_LOW: 	/* INTR only */
		if (is_input) {
			tmp = (67667L * (31L + 10L * BitTime (bytecount))) / 1000L;
			return (64060L + (2 * BW_HUB_LS_SETUP) + BW_HOST_DELAY + tmp);
		} else {
			tmp = (66700L * (31L + 10L * BitTime (bytecount))) / 1000L;
			return (64107L + (2 * BW_HUB_LS_SETUP) + BW_HOST_DELAY + tmp);
		}
	case USB_SPEED_FULL:	/* ISOC or INTR */
		if (isoc) {
			tmp = (8354L * (31L + 10L * BitTime (bytecount))) / 1000L;
			return (((is_input) ? 7268L : 6265L) + BW_HOST_DELAY + tmp);
		} else {
			tmp = (8354L * (31L + 10L * BitTime (bytecount))) / 1000L;
			return (9107L + BW_HOST_DELAY + tmp);
		}
	case USB_SPEED_HIGH:	/* ISOC or INTR */
		// FIXME adjust for input vs output
		if (isoc)
			tmp = HS_USECS (bytecount);
		else
			tmp = HS_USECS_ISO (bytecount);
		return tmp;
	default:
		dbg ("bogus device speed!");
		return -1;
	}
}
#ifndef _ETCPIP
	EXPORT_SYMBOL (usb_calc_bus_time);
#endif // ndef _ETCPIP

/*
 * usb_check_bandwidth():
 *
 * old_alloc is from host_controller->bandwidth_allocated in microseconds;
 * bustime is from calc_bus_time(), but converted to microseconds.
 *
 * returns <bustime in us> if successful,
 * or -ENOSPC if bandwidth request fails.
 *
 * FIXME:
 * This initial implementation does not use Endpoint.bInterval
 * in managing bandwidth allocation.
 * It probably needs to be expanded to use Endpoint.bInterval.
 * This can be done as a later enhancement (correction).
 *
 * This will also probably require some kind of
 * frame allocation tracking...meaning, for example,
 * that if multiple drivers request interrupts every 10 USB frames,
 * they don't all have to be allocated at
 * frame numbers N, N+10, N+20, etc.  Some of them could be at
 * N+11, N+21, N+31, etc., and others at
 * N+12, N+22, N+32, etc.
 *
 * Similarly for isochronous transfers...
 *
 * Individual HCDs can schedule more directly ... this logic
 * is not correct for high speed transfers.
 */
int usb_check_bandwidth (struct usb_device *dev, struct urb *urb)
{
	unsigned int	pipe = urb->pipe;
	long		bustime;
	int		is_in = usb_pipein (pipe);
	int		is_iso = usb_pipeisoc (pipe);
	int		old_alloc = dev->bus->bandwidth_allocated;
	int		new_alloc;


	bustime = NS_TO_US (usb_calc_bus_time (dev->speed, is_in, is_iso,
			usb_maxpacket (dev, pipe, !is_in)));
	if (is_iso)
		bustime /= urb->number_of_packets;

	new_alloc = old_alloc + (int) bustime;
	if (new_alloc > FRAME_TIME_MAX_USECS_ALLOC) {
#ifdef	DEBUG
		char	*mode = 
#ifdef CONFIG_USB_BANDWIDTH
			"";
#else
			"would have ";
#endif
		dev_dbg (&dev->dev, "usb_check_bandwidth %sFAILED: %d + %ld = %d usec\n",
			mode, old_alloc, bustime, new_alloc);
#endif
#ifdef CONFIG_USB_BANDWIDTH
		bustime = -ENOSPC;	/* report error */
#endif
	}

	return bustime;
}
#ifndef _ETCPIP
	EXPORT_SYMBOL (usb_check_bandwidth);
#endif // ndef _ETCPIP


/**
 * usb_claim_bandwidth - records bandwidth for a periodic transfer
 * @dev: source/target of request
 * @urb: request (urb->dev == dev)
 * @bustime: bandwidth consumed, in (average) microseconds per frame
 * @isoc: true iff the request is isochronous
 *
 * Bus bandwidth reservations are recorded purely for diagnostic purposes.
 * HCDs are expected not to overcommit periodic bandwidth, and to record such
 * reservations whenever endpoints are added to the periodic schedule.
 *
 * FIXME averaging per-frame is suboptimal.  Better to sum over the HCD's
 * entire periodic schedule ... 32 frames for OHCI, 1024 for UHCI, settable
 * for EHCI (256/512/1024 frames, default 1024) and have the bus expose how
 * large its periodic schedule is.
 */
void usb_claim_bandwidth (struct usb_device *dev, struct urb *urb, int bustime, int isoc)
{
	dev->bus->bandwidth_allocated += bustime;
	if (isoc)
		dev->bus->bandwidth_isoc_reqs++;
	else
		dev->bus->bandwidth_int_reqs++;
	urb->bandwidth = bustime;

#ifdef USB_BANDWIDTH_MESSAGES
	dev_dbg (&dev->dev, "bandwidth alloc increased by %d (%s) to %d for %d requesters\n",
		bustime,
		isoc ? "ISOC" : "INTR",
		dev->bus->bandwidth_allocated,
		dev->bus->bandwidth_int_reqs + dev->bus->bandwidth_isoc_reqs);
#endif
}
#ifndef _ETCPIP
	EXPORT_SYMBOL (usb_claim_bandwidth);
#endif // ndef _ETCPIP


/**
 * usb_release_bandwidth - reverses effect of usb_claim_bandwidth()
 * @dev: source/target of request
 * @urb: request (urb->dev == dev)
 * @isoc: true iff the request is isochronous
 *
 * This records that previously allocated bandwidth has been released.
 * Bandwidth is released when endpoints are removed from the host controller's
 * periodic schedule.
 */
void usb_release_bandwidth (struct usb_device *dev, struct urb *urb, int isoc)
{
	dev->bus->bandwidth_allocated -= urb->bandwidth;
	if (isoc)
		dev->bus->bandwidth_isoc_reqs--;
	else
		dev->bus->bandwidth_int_reqs--;

#ifdef USB_BANDWIDTH_MESSAGES
	dev_dbg (&dev->dev, "bandwidth alloc reduced by %d (%s) to %d for %d requesters\n",
		urb->bandwidth,
		isoc ? "ISOC" : "INTR",
		dev->bus->bandwidth_allocated,
		dev->bus->bandwidth_int_reqs + dev->bus->bandwidth_isoc_reqs);
#endif
	urb->bandwidth = 0;
}
#ifndef _ETCPIP
	EXPORT_SYMBOL (usb_release_bandwidth);
#endif // ndef _ETCPIP


/*-------------------------------------------------------------------------*/

/*
 * Generic HC operations.
 */

/*-------------------------------------------------------------------------*/

/* called from khubd, or root hub init threads for hcd-private init */

static int hcd_alloc_dev(struct usb_device *udev)
{
 struct hcd_dev		*dev;
 struct usb_hcd		*hcd;
 unsigned long		flags;

 if (!udev || udev->hcpriv)
		return -EINVAL;
 if (!udev->bus || !udev->bus->hcpriv_hcd) {
	    USB_DBG_INFO(iprintf("[USB INFO] localhc_urb_enqueue() return -ENODEV !! \n")) ;
		return -ENODEV;
    }    
 hcd = udev->bus->hcpriv_hcd;
 
 if (hcd->state == USB_STATE_QUIESCING)
		return -ENOLINK;

 USB_DBG_MEM(iprintf(
	"[USB MEM] hcd_alloc_dev(): alloc dev .. \n"
    ));
 /*
 #ifndef _ETCPIP
	dev = (struct hcd_dev *) kmalloc (sizeof *dev, GFP_KERNEL);
  #else
 */
 // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
 //              But now on ap/upnp/upnpport.c
 //              void *alloc_mem(int size) 
 //              this will fill all 0s

 // BLOCK_PREEMPTION;

 dev = (struct hcd_dev *) // 144
			UsbAllocMemC(USB_MEM_C_M
				#ifdef USB_DBG_MEM_C_ON
					, sizeof *dev 
		        #endif
                      ) ;
 // UNBLOCK_PREEMPTION;
 USB_DBG_MEM(iprintf("[USB MEM] hcd_alloc_dev() : dev = %08XH \n", dev ));

 if ( dev == NULL ) {
		dprintf("[USB] Error : couldn't allocate dev memory \n");
	   }
 /*
 #endif
 */

 if (dev == NULL)
		return -ENOMEM;
 /*
 #ifndef _ETCPIP
	memset (dev, 0, sizeof *dev);
 #endif // ndef _ETCPIP
 */
 INIT_LIST_HEAD (&dev->dev_list);
 INIT_LIST_HEAD (&dev->urb_list);
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave (&hcd_data_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 list_add (&dev->dev_list, &hcd->dev_list);
 
 // refcount is implicit
 
 udev->hcpriv = dev;
 
 USB_DBG_INIT(iprintf("[USB INIT] Set Value : udev->hcpriv = %08XH \n"
	, udev->hcpriv
	));
 
 // #ifndef USB_GLOBAL_EP // 2003/7/15 B. S. Liaw
	((struct hcd_dev *) udev->hcpriv)->ep_inited = 0 ; // 2003/7/3 B. S. Liaw
 
	USB_DBG_INIT(iprintf("[USB INIT] Set Value : udev->hcpriv->ep_inited = %d \n"
		, ((struct hcd_dev *) udev->hcpriv)->ep_inited
		));
 // #endif // ndef USB_GLOBAL_EP
 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore (&hcd_data_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 return 0;
}

/*-------------------------------------------------------------------------*/

static void urb_unlink(struct urb *urb)
{
 unsigned long		flags;
 struct usb_device	*dev;

 /* Release any periodic transfer bandwidth */
	if (urb->bandwidth)
		usb_release_bandwidth (urb->dev, urb,
			usb_pipeisoc (urb->pipe));

 /* clear all state linking urb to this dev (and hcd) */
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave (&hcd_data_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 list_del_init (&urb->urb_list);
 /*
 #ifndef _ETCPIP
	dev = urb->dev;
	usb_put_dev (dev);
 #endif // ndef _ETCPIP
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore (&hcd_data_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
}


/* may be called in any context with a valid urb->dev usecount
 * caller surrenders "ownership" of urb
 * expects usb_submit_urb() to have sanity checked and conditioned all
 * inputs in the urb
 */
#ifndef _ETCPIP
	static int hcd_submit_urb (struct urb *urb, int mem_flags)
#else
	static int hcd_submit_urb(struct urb *urb)
#endif
{
 int			status;
 struct usb_hcd		*hcd = urb->dev->bus->hcpriv_hcd;
 struct hcd_dev		*dev = urb->dev->hcpriv;
 unsigned long		flags;
 /*
 USB_DBG_ROOTHUB(iprintf(
	"[USB ROOTHUB] hcd_submit_urb() Begin ..\n"));
 */
 
 USB_DBG_CTLMSG(iprintf(
	"[USB CTLMSG] hcd_submit_urb() Begin ..\n"));
    
 USB_DBG_CTLMSG(iprintf(
    "[USB CTLMSG] hcd_submit_urb() &(urb->dev->bus->hcpriv_hcd) = %08XH "
    "urb->dev->bus->hcpriv_hcd=%08XH\n",
    &(urb->dev->bus->hcpriv_hcd) , urb->dev->bus->hcpriv_hcd
    ));
    
 #ifdef USB_DBG_CTLMSG_ON
	USB_DBG_MEM(iprintf(
		"[USB MEM] hcd_submit_urb() : hcd = %08XH \n" ,
    	hcd
    	));
 #endif //USB_DBG_CTLMSG_ON
    
 USB_DBG_CTLMSG(iprintf(
    "[USB CTLMSG] hcd_submit_urb() &(urb->dev->hcpriv) = %08XH "
    "urb->dev->hcpriv=%08XH\n",
    &(urb->dev->hcpriv) , urb->dev->hcpriv
    ));
    
 if (!hcd || !dev)
   {
	USB_DBG_INFO(iprintf("[USB INFO] hcd_submit_urb() End 1 with -ENODEV.\n"));
	return -ENODEV;
   }

 /*
  * FIXME:  make urb timeouts be generic, keeping the HCD cores
  * as simple as possible.
  */

 // NOTE:  a generic device/urb monitoring hook would go here.
 // hcd_monitor_hook(MONITOR_URB_SUBMIT, urb)
 // It would catch submission paths for all urbs.

 /*
  * Atomically queue the urb,  first to our records, then to the HCD.
  * Access to urb->status is controlled by urb->lock ... changes on
  * i/o completion (normal or fault) or unlinking.
  */

 // FIXME:  verify that quiescing hc works right (RH cleans up)

 #ifndef NO_SPINLOCK
	spin_lock_irqsave (&hcd_data_lock, flags);
 #endif // ndef NO_SPINLOCK
 
 #ifdef USB_DBG_CTLMSG_ON
	USB_DBG_MEM(iprintf(
		"[USB MEM] hcd_submit_urb() : &(hcd->state) = %08XH \n" ,
    	&(hcd->state)
    	));
 #endif //USB_DBG_CTLMSG_ON
 
 USB_DBG_CTLMSG(iprintf(
    "[USB CTLMSG] hcd_submit_urb() : hcd->state = %02XH \n",
    hcd->state 
    ));
    
 if (HCD_IS_RUNNING (hcd->state) && hcd->state != USB_STATE_QUIESCING)
   {
	usb_get_dev (urb->dev);
	list_add_tail (&urb->urb_list, &dev->urb_list);
	status = 0;
   }
  else
   {
	INIT_LIST_HEAD (&urb->urb_list);
	status = -ESHUTDOWN;
   }
   
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore (&hcd_data_lock, flags);
 #endif // ndef NO_SPINLOCK
 
 if (status)
   {
	USB_DBG_CTLMSG(iprintf(
		"[USB CTLMSG] hcd_submit_urb() End 2.\n"));
	return status;
   }
 /* increment urb's reference count as part of giving it to the HCD
  * (which now controls it).  HCD guarantees that it either returns
  * an error or calls giveback(), but not both.
  */
 urb = usb_get_urb (urb);
 if (urb->dev == hcd->self.root_hub)
   {
    /*
	USB_DBG_ROOTHUB(iprintf(
		"[USB ROOTHUB] hcd_submit_urb() : root hub\n"));
    */
	USB_DBG_CTLMSG(iprintf(
		"[USB CTLMSG] hcd_submit_urb() : root hub\n"));
	/* NOTE:  requirement on hub callers (usbfs and the hub
	 * driver, for now) that URBs' urb->transfer_buffer be
	 * valid and usb_buffer_{sync,unmap}() not be needed, since
	 * they could clobber root hub response data.
	 */
	urb->transfer_flags |= URB_NO_DMA_MAP;
	status = rh_urb_enqueue (hcd, urb);
	USB_DBG_CTLMSG(iprintf("[USB CTLMSG] after rh_urb_enqueue() \n" )) ;
	goto done;
   }

 /* lower level hcd code should use *_dma exclusively,
  * unless it uses pio or talks to another transport.
  */
 if (!(urb->transfer_flags & URB_NO_DMA_MAP)
			&& hcd->controller->dma_mask) 
   {
	USB_DBG_MEM(iprintf(
		"[USB MEM] hcd_submit_urb() : urb->setup_packet = %08XH \n",
    	urb->setup_packet
    	));
	if (usb_pipecontrol (urb->pipe))
       {
		#ifndef _ETCPIP
		/*
			urb->setup_dma = dma_map_single ( hcd->controller,
					urb->setup_packet,
					sizeof (struct usb_ctrlrequest),
					DMA_TO_DEVICE);
                    
			USB_DBG_MEM(iprintf(
				"[USB MEM] hcd_submit_urb() : urb->setup_dma = %08XH \n",
    			urb->setup_dma
    			));
    	 */
		#endif
       }             
	if (urb->transfer_buffer_length != 0)
       {
		#ifndef USB_NO_VMM // no vitrtul memory management
        
			urb->transfer_dma = dma_map_single ( hcd->controller,
					urb->transfer_buffer,
					urb->transfer_buffer_length,
					usb_pipein (urb->pipe)
					    ? DMA_FROM_DEVICE
					    : DMA_TO_DEVICE);
			USB_DBG_MEM(iprintf("[USB MEM] hcd_submit_urb() : urb->transfer_dma = %08XH \n", urb->transfer_dma )) ;
			#ifdef USB_BS_ASSERT
				if ( !(USB_IS_UN_CACHED(urb->transfer_dma)) ) {
					iprintf("[ASSERT][USB] hcd_submit_urb() : " ) ;
    			    iprintf("Error : Not Un-Cached !!\n" ) ;
			        iprintf("[ASSERT][USB] urb->transfer_dma addr = %08XH \n", urb->transfer_dma) ;
			        iprintf("[ASSERT][USB] Continue ..\n") ;
    			   }
			#endif // USB_BS_ASSERT
		#else
        	/*      
			// 2003/6/30 B. S. Liaw
			urb->transfer_dma = urb->transfer_buffer ;
            */
			#ifdef USB_BS_ASSERT
				if ( !(USB_IS_UN_CACHED(urb->transfer_buffer)) ) {
					iprintf("[ASSERT][USB] hcd_submit_urb() : " ) ;
    			    iprintf("Error : Not Un-Cached !!\n" ) ;
			        iprintf("[ASSERT][USB] urb->transfer_buffer addr = %08XH \n", urb->transfer_buffer) ;
			        iprintf("[ASSERT][USB] Continue ..\n") ;
    			   }
			#endif // USB_BS_ASSERT
		#endif       
       }
   }

	#ifndef _ETCPIP
		status = hcd->driver->urb_enqueue (hcd, urb, mem_flags);
    #else
		status = hcd->driver->urb_enqueue (hcd, urb);
    #endif
    /*
    iprintf("[USB] --------------------------------------------- \n") ;
    iprintf("[USB] after hcd->driver->urb_enqueue() ------------ \n") ;
    iprintf("[USB] --------------------------------------------- \n") ;
    */
 done:
	if (status)
       {
		USB_DBG_INFO(iprintf(
        	"[USB INFO] hcd_submit_urb() : error !! \n"
            )) ;
		usb_put_urb (urb);
		urb_unlink (urb);
	   }
    
	USB_DBG_CTLMSG(iprintf(
		"[USB CTLMSG] hcd_submit_urb() End 3.\n"));
	return status;
}

/*-------------------------------------------------------------------*/

/* called in any context */
static int hcd_get_frame_number (struct usb_device *udev)
{
 struct usb_hcd	*hcd = (struct usb_hcd *)udev->bus->hcpriv_hcd;
 return hcd->driver->get_frame_number (hcd);
}

/*-------------------------------------------------------------------------*/

struct completion_splice {		// modified urb context:
 /* did we complete? */
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
	struct completion	done;
 #endif // ndef _ETCPIP

 /* original urb data */
 usb_callback_t	callback ;
 void			*context;
};

#ifndef _ETCPIP
	/*
	static void unlink_complete (struct urb *urb, struct pt_regs *regs)
    */
#else
	static void unlink_complete (struct urb *urb)
#endif
{
 struct completion_splice	*splice;

 splice = (struct completion_splice *) urb->context;

 /* issue original completion call */
 urb->callback = splice->callback;
 urb->context = splice->context;
 
 #ifndef USB_NO_ISR_UNLINK  // Under Debug !!
 // #ifdef USB_NO_ABC // No usb_api_blocking_completion()
	USB_DBG_CALLBACK(iprintf("[USB CALLBACK] unlink_complete() Call usb_urb_complete() ...\n")) ;
	// #if defined(USB_DBG_URB_ON) || defined(USB_DBG_URBBULK_ON) || defined(USB_DBG_URBINT_ON)
    #ifdef USB_DBG_URB_INFO_ON
		iprintf("[USB URB] unlink_complete() Call usb_urb_complete() ...\n") ;
	#endif
	usb_urb_complete(urb) ; // for urb->awd_done
 // #endif // USB_NO_ABC
 #endif // ndef USB_NO_ISR_UNLINK

 #ifndef USB_NO_ISR_UNLINK // 2003/9/13 B. S. Liaw BugFix , Donot Call !!
 if(urb->callback)
   {
	USB_DBG_CALLBACK(iprintf("[USB CALLBACK] unlink_complete() : Call urb->callback() .. \n")) ;
	#ifndef _ETCPIP
    	/*
		urb->complete (urb, regs);
        */
	  #else
		urb->callback(urb);
	#endif
   }
  else
   {
	USB_DBG_CALLBACK(iprintf(
		"[USB CALLBACK] unlink_complete() : Null urb->callback() !! \n"
	    ));
   }
 #endif // ndef USB_NO_ISR_UNLINK

	/* then let the synchronous unlink call complete */
 #ifndef _ETCPIP
 	/*
	complete (&splice->done);
    */
 #endif // ndef _ETCPIP
} // function unlink_complete()

/*
 * called in any context; note ASYNC_UNLINK restrictions
 *
 * caller guarantees urb won't be recycled till both unlink()
 * and the urb's completion function return
 */
static int hcd_unlink_urb(struct urb *urb)
{
 struct hcd_dev			*dev;
 struct usb_hcd			*hcd = 0;
 struct device			*sys = 0;
 unsigned long			flags;
 #if 0  /* Charles 20040423 */
 static struct completion_splice splice;
 #endif
 int				retval;

 #ifdef USB_BS_ASSERT
	if (!(  urb
       ) ) {
		iprintf("[ASSERT][USB] hcd_unlink_urb(1) : " ) ;
	    iprintf("Error : !urb !!\n" ) ;
        iprintf("[ASSERT][USB] Continue ..\n") ;
	   }

	if (!(  urb->dev
       ) ) {
		iprintf("[ASSERT][USB] hcd_unlink_urb(2) : " ) ;
	    iprintf("Error : !urb->dev !!\n" ) ;
        iprintf("[ASSERT][USB] Continue ..\n") ;
	   }

 #endif // USB_BS_ASSERT

 #ifndef USB_NO_ISR_UNLINK
	#ifdef USB_BS_ASSERT
	if (!(  (urb->dev->state >= USB_STATE_DEFAULT) 
       ) ) {
		iprintf("[ASSERT][USB] hcd_unlink_urb(3) : " ) ;
	    iprintf("Error : urb->dev->state=%d !!\n",urb->dev->state ) ;
        iprintf("[ASSERT][USB] Continue ..\n") ;
	   }
	#endif // USB_BS_ASSERT
 #endif // ndef USB_NO_ISR_UNLINK

 #ifdef USB_BS_ASSERT

	if (!(  urb->dev->bus
       ) ) {
		iprintf("[ASSERT][USB] hcd_unlink_urb(4) : " ) ;
	    iprintf("Error : !urb->dev->bus !!\n" ) ;
        iprintf("[ASSERT][USB] Continue ..\n") ;
	   }

	if (!(  urb->dev->bus->op
       ) ) {
		iprintf("[ASSERT][USB] hcd_unlink_urb(5) : " ) ;
	    iprintf("Error : !urb->dev->bus->op !!\n" ) ;
        iprintf("[ASSERT][USB] Continue ..\n") ;
	   }
 #endif // USB_BS_ASSERT

 if(!urb) {
	USB_DBG_INFO(iprintf("[USB INFO] hcd_unlink_urb : Null urb !!\n"));
	return -EINVAL;
   }

 /*
  * we contend for urb->status with the hcd core,
  * which changes it while returning the urb.
  *
  * Caller guaranteed that the urb pointer hasn't been freed, and
  * that it was submitted.  But as a rule it can't know whether or
  * not it's already been unlinked ... so we respect the reversed
  * lock sequence needed for the usb_hcd_giveback_urb() code paths
  * (urb lock, then hcd_data_lock) in case some other CPU is now
  * unlinking it.
  */

 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave (&urb->lock, flags);
	spin_lock (&hcd_data_lock);
 #endif // ndef NO_SPINLOCK
 */

 if (!urb->dev || !urb->dev->bus) {
		USB_DBG_INFO(iprintf("[USB INFO] hcd_unlink_urb : !urb->dev || !urb->dev->bus !!\n"));
		retval = -ENODEV;
		goto done;
	}

 dev = urb->dev->hcpriv;
 sys = &urb->dev->dev;
 hcd = urb->dev->bus->hcpriv_hcd;

 if(!dev || !hcd) {
	USB_DBG_INFO(iprintf("[USB INFO] hcd_unlink_urb : !dev || !hcd !!\n"));
	retval = -ENODEV;
	goto done;
   }

 if(!urb->hcpriv) {
	USB_DBG_INFO(iprintf("[USB INFO] hcd_unlink_urb : !urb->hcpriv !!\n"));
	retval = -EINVAL;
	goto done;
   }

 /* Any status except -EINPROGRESS means something already started to
  * unlink this URB from the hardware.  So there's no more work to do.
  *
  * FIXME use better explicit urb state
  */

 /* 2003/90/27 B. S. Liaw , comment out
 #ifndef USB_NO_ISR_UNLINK   
	if (urb->status != -EINPROGRESS) {
		USB_DBG_INFO(iprintf("[USB INFO] hcd_unlink_urb : urb->status != -EINPROGRESS (%d) !!\n",urb->status)) ;
		retval = -EBUSY;
		goto done;
	}
 #endif
 */
 /* maybe set up to block until the urb's completion fires.  the
  * lower level hcd code is always async, locking on urb->status
  * updates; an intercepted completion unblocks us.
  */
 if(!(urb->transfer_flags & URB_ASYNC_UNLINK)) {
	if (in_interrupt ()) {
		/*
		dev_dbg (hcd->controller, "non-async unlink in_interrupt");
		*/
		USB_DBG_INFO(iprintf(
			"[USB INFO] hcd_unlink_urb() : non-async unlink in_interrupt !!\n"));
		retval = -EWOULDBLOCK;
		goto done;
	   }
		/* synchronous unlink: block till we see the completion */
	#ifndef _ETCPIP
    	/*
		init_completion (&splice.done);
        */
	#endif // ndef _ETCPIP
#if 0
	splice.callback = urb->callback;
	splice.context = urb->context;

	#ifdef USB_NO_ISR_UNLINK
		USB_DBG_CALLBACK(iprintf(
			"[USB CALLBACK] hcd_unlink_urb() : set call back = unlink_complete .. \n"
			)) ;
		urb->callback = unlink_complete;
	#endif // USB_NO_ISR_UNLINK

	urb->context = &splice;
#endif
	#ifdef USB_NO_ISR_UNLINK
		if (urb->status!=0)
	#endif
		   {
			urb->status = -ENOENT;
		   }
   }
  else {
		/* asynchronous unlink */
	#ifdef USB_NO_ISR_UNLINK
		if (urb->status!=0)
	#endif
		   {
			urb->status = -ECONNRESET;
    	   }
   }

 /*
 #ifndef NO_SPINLOCK
	spin_unlock (&hcd_data_lock);
	spin_unlock_irqrestore (&urb->lock, flags);
 #endif // ndef NO_SPINLOCK
 */

 #ifndef NO_TIMER
	if (urb == (struct urb *) hcd->rh_timer.data)
  #else
    if (urb == hcd->rh_urb)
 #endif
     {
		USB_DBG_INFO(iprintf("[USB INFO] hcd_unlink_urb : usb_rh_status_dequeue() ..\n"));
		usb_rh_status_dequeue (hcd, urb);
		retval = 0;
	   }
	  else {
		USB_DBG_INIT(iprintf("[USB INIT] hcd_unlink_urb : hcd->driver->urb_dequeue() ..\n"));
		// USB_DBG_INFO(iprintf("[USB INFO] hcd_unlink_urb : hcd->driver->urb_dequeue() ..\n"));
		retval = hcd->driver->urb_dequeue (hcd, urb);

		/* hcds shouldn't really fail these calls, but... */
		if (retval) {
			dev_dbg (sys, "dequeue %p --> %d\n", urb, retval);
			USB_DBG_INFO(iprintf("[USB INFO] hcd_unlink_urb : hcds shouldn't really fail these calls !!\n"));
			if (!(urb->transfer_flags & URB_ASYNC_UNLINK)) {
 #ifndef NO_SPINLOCK
				spin_lock_irqsave (&urb->lock, flags);
 #endif // ndef NO_SPINLOCK
 #if 0
				urb->callback = splice.callback;
				urb->context = splice.context;
 #endif
 #ifndef NO_SPINLOCK
				spin_unlock_irqrestore (&urb->lock, flags);
 #endif // ndef NO_SPINLOCK
			}
			goto bye;
		}
	}

 /* block till giveback, if needed */

 if (urb->transfer_flags & URB_ASYNC_UNLINK) {
		USB_DBG_INFO(iprintf("[USB INFO] hcd_unlink_urb : URB_ASYNC_UNLINK !!\n"));
		return -EINPROGRESS;
	   }
 /*
 #ifndef _ETCPIP
    // wait_for_completion() is in <linux/completion.h>
	wait_for_completion (&splice.done);
 #endif // ndef _ETCPIP
 */
 return 0;

 // ------------
 done:
 // ------------

 /*
 #ifndef NO_SPINLOCK
	spin_unlock (&hcd_data_lock);
	spin_unlock_irqrestore (&urb->lock, flags);
 #endif // ndef NO_SPINLOCK
 */

 // ------------
 bye:
 // ------------
	if (retval && sys) {
		/*
		dev_dbg (sys, "hcd_unlink_urb %p fail %d\n", urb, retval);
		*/
		USB_DBG_INFO(iprintf("[USB INFO] hcd_unlink_urb : Error : urb %08XH , Fail %d !!\n", urb, retval));
		// dprintf("[USB] Unlink URB fail !!\n") ;
       }
	return retval;
} // function hcd_unlink_urb()

/*-------------------------------------------------------------------------*/

/* called by khubd, rmmod, apmd, or other thread for hcd-private cleanup */

// FIXME:  likely best to have explicit per-setting (config+alt)
// setup primitives in the usbcore-to-hcd driver API, so nothing
// is implicit.  kernel 2.5 needs a bunch of config cleanup...

static int hcd_free_dev(struct usb_device *udev)
{
 struct hcd_dev		*dev;
 struct usb_hcd		*hcd;
 unsigned long		flags;

 USB_DBG_INIT(iprintf("[USB INIT] hcd_free_dev() .. \n"));

 /*
 usb_device_count -- ; // >= 1 : enable ISR
 
    if(usb_device_count<=0) {
		#if defined(USB_INT_SCHEDULING) || defined(USB_INT_ON_DEMAND)
			UsbDisableInt() ;
		  #else
			#if defined(_S3C2510)
		    	#error : SS2510 Need Define USB_INT_ON_DEMAND !!
		    #endif
		#endif // USB_INT_ON_DEMAND
	USB_DBG_INIT(iprintf("[USB INIT] hcd_free_dev() : reset usb_device_count to 0 \n")) ;
		usb_device_count = 0 ; // >= 1 : enable ISR
      }
 */

 if (!udev || !udev->hcpriv)
		return -EINVAL;

 if (!udev->bus || !udev->bus->hcpriv_hcd) {
	    USB_DBG_INFO(iprintf("[USB INFO] hcd_free_dev() return -ENODEV !! \n")) ;
		return -ENODEV;
   }
 // should udev->devnum == -1 ??

 USB_DBG_INIT(iprintf("[USB INIT] udev->hcpriv = %08XH \n"
	, udev->hcpriv
	));
 dev = udev->hcpriv;
 
 USB_DBG_INIT(iprintf("[USB INIT] udev->bus->hcpriv_hcd = %08XH \n"
	, udev->bus->hcpriv_hcd
	));
 hcd = udev->bus->hcpriv_hcd;

 /* device driver problem with refcounts? */
 if( !list_empty (&dev->urb_list)) {
	/*
	#ifdef USB_DBGSTRING_ON
		dev_dbg (hcd->controller, "free busy dev, %s devnum %d (bug!)\n",
			hcd->self.bus_name, udev->devnum);
	  #else
		dev_dbg (hcd->controller, "free busy dev, devnum %d (bug!)\n",
			udev->devnum);
	*/
		USB_DBG_INFO(iprintf("[USB INFO] hcd_free_dev() : Error : free busy dev, devnum %d (bug!)\n",udev->devnum)) ;
	/*
	#endif
	*/
    #ifdef USB_SOTA   // Sota USB 1.1 Host
		USB_DBG_INFO(iprintf("[USB INFO] hcd_free_dev() : Continue .. \n")) ;
		// return -EINVAL;
	 #else
		USB_DBG_INFO(iprintf("[USB INFO] hcd_free_dev() : Not Free !! \n")) ;
		// return -EINVAL;  /* Charles 20040423 */
	#endif
   }

 /*
 if (hcd->driver->free_config)
   {
	USB_DBG_INIT(iprintf("[USB INIT] hcd->driver->free_config() .. \n"));
	hcd->driver->free_config (hcd, udev);
   }
  else
   {
	USB_DBG_INIT(iprintf(
    	"[USB INIT] Error : No hcd->driver->free_config !! \n"
        ));
   }
 */
 USB_DBG_INIT(iprintf(
	"[USB INIT] hcd_free_dev() : localhc_free_config() .. \n"));
 localhc_free_config(hcd, udev);

 if(usb_device_count<=1) {
	#ifdef USB_DMA_ON_DEMAND
			UsbDisableDma() ;
	 #else
		#if defined(_ADM5120)
	    	// #error : ADMtek 5120  Need Define USB_DMA_ON_DEMAND !!
	    #endif
	#endif // USB_DMA_ON_DEMAND

	#if defined(USB_INT_SCHEDULING) || defined(USB_INT_ON_DEMAND)
		UsbDisableInt() ;
	 #else
		#if defined(_S3C2510)
	    	#error : SS2510 Need Define USB_INT_ON_DEMAND !!
	#elif defined(_AR7300)
    	#error : _AR7300 Need Define USB_INT_ON_DEMAND !!
		#elif defined(_ATHEROS)
			#error : _ATHEROS Need Define USB_INT_ON_DEMAND !!
	    #endif
	#endif // USB_INT_ON_DEMAND
	USB_DBG_INIT(iprintf("[USB INIT] hcd_free_dev() : reset usb_device_count to 0 \n")) ;
	usb_device_count = 0 ; // >= 1 : enable ISR
   }
  else {
	usb_device_count -- ; // >= 1 : enable ISR
   }

 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave (&hcd_data_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 
 list_del (&dev->dev_list);
 udev->hcpriv = NULL;
 
 /* 
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore (&hcd_data_lock, flags);
 #endif // ndef NO_SPINLOCK
 */

 USB_DBG_MEM(iprintf("[USB MEM] hcd_free_dev(): free dev .. \n"));
 /*
 #ifndef USB_MEM_LIB
 	kfree (dev);
 #elif defined USB_DBG_MEM_C_ON
 */
    UsbFreeMemC(USB_MEM_C_M,(unsigned char *)dev
			#ifdef USB_DBG_MEM_C_ON
				,sizeof(*dev)
			#endif
		   ) ;
 /*
 #else
    UsbFreeMemC(USB_MEM_C_M,(unsigned char *)dev) ;
 #endif
 */
 return 0;
} // function hcd_free_dev()

/*
 * usb_hcd_operations - adapts usb_bus framework to HCD framework (bus glue)
 *
 * When registering a USB bus through the HCD framework code, use this
 * usb_operations vector.  The PCI glue layer does so automatically; only
 * bus glue for non-PCI system busses will need to use this.
 */
struct usb_operations usb_hcd_operations = {
 .allocate =		hcd_alloc_dev,
 .get_frame_number =	hcd_get_frame_number,
 .submit_urb =		hcd_submit_urb,
 .unlink_urb =		hcd_unlink_urb,
 .deallocate =		hcd_free_dev,
 .buffer_alloc =		hcd_buffer_alloc,
 // #ifndef _ETCPIP
	.buffer_free =		hcd_buffer_free,
 // #endif // ndef _ETCPIP
};
/*
#ifndef _ETCPIP
	EXPORT_SYMBOL (usb_hcd_operations);
#endif // ndef _ETCPIP
*/
/*-------------------------------------------------------------------------*/

/**
 * usb_hcd_giveback_urb - return URB from HCD to device driver
 * @hcd: host controller returning the URB
 * @urb: urb being returned to the USB device driver.
 * @regs: pt_regs, passed down to the URB completion handler
 * Context: in_interrupt()
 *
 * This hands the URB from HCD to its USB device driver, using its
 * completion function.  The HCD has freed all per-urb resources
 * (and is done using urb->hcpriv).  It also released all HCD locks;
 * the device driver won't cause problems if it frees, modifies,
 * or resubmits this URB.
 */
/*
#ifndef _ETCPIP
	void usb_hcd_giveback_urb (struct usb_hcd *hcd, struct urb *urb, struct pt_regs *regs)
#else
*/
	void usb_hcd_giveback_urb (struct usb_hcd *hcd, struct urb *urb)
/*
#endif
*/
{
 USB_DBG_FLOWTIMCRY(iprintf("[USB FLOWTIMCRY] usb_hcd_giveback_urb() Begin ..\n"));
 /*
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_hcd_giveback_urb() Call urb_unlink() ...\n"
    ));
 */
 urb_unlink (urb);

 // NOTE:  a generic device/urb monitoring hook would go here.
 // hcd_monitor_hook(MONITOR_URB_FINISH, urb, dev)
 // It would catch exit/unlink paths for all urbs.

 /* lower level hcd code should use *_dma exclusively */
 if (!(urb->transfer_flags & URB_NO_DMA_MAP))
   {
	if (usb_pipecontrol (urb->pipe))
       {
		/*
		#ifndef _ETCPIP  // no urb->setup_dma
			pci_unmap_single (hcd->pdev, urb->setup_dma,
					sizeof (struct usb_ctrlrequest),
					PCI_DMA_TODEVICE);
		#endif
		*/
       }
	if (urb->transfer_buffer_length != 0)
       {
		#ifndef USB_NO_VMM // no vitrtul memory management
			#ifdef CONFIG_PCI
				pci_unmap_single (hcd->pdev, urb->transfer_dma,
					urb->transfer_buffer_length,
					usb_pipein (urb->pipe)
					    ? PCI_DMA_FROMDEVICE
					    : PCI_DMA_TODEVICE);
	        #endif
		#endif // ndef USB_NO_VMM
       }
   }

 /* pass ownership to the completion handler */
 
 // #ifdef USB_NO_ABC // No usb_api_blocking_completion()
 USB_DBG_CALLBACK(iprintf(
		"[USB CALLBACK] usb_hcd_giveback_urb() Call usb_urb_complete() ...\n"
		)) ;
 #if defined(USB_DBG_URB_ON) || defined(USB_DBG_URBBULK_ON) || defined(USB_DBG_URBINT_ON)
 // #ifdef USB_DBG_URB_INFO_ON
		if (urb->dev != hcd->self.root_hub) {
			iprintf("[USB URB] usb_hcd_giveback_urb() : urb->status = %d \n" ,urb->status) ;
			iprintf("[USB URB] usb_hcd_giveback_urb() Call usb_urb_complete() ...\n") ;
		   }
 #endif

 usb_urb_complete(urb) ; // for urb->awd_done
 // #endif // USB_NO_ABC

 #ifndef USB_NO_ISR_CALLBACK
 	/*
	 if(urb->callback)
	   {
		USB_DBG_CALLBACK(iprintf("[USB CALLBACK] usb_hcd_giveback_urb() Call urb->callback() ...\n" )) ;
 
		#ifndef _ETCPIP
 			
			// urb->complete (urb, regs);
		    
		  #else
			urb->callback(urb);
		#endif
	   }
	  else
	   {
		USB_DBG_CALLBACK(iprintf(
			"[USB CALLBACK] usb_hcd_giveback_urb() : Null urb->callback() !! \n"
		    ));
	   }
    */
 #endif // ndef USB_NO_ISR_CALLBACK

 #ifndef USB_NO_ISR_FREE
    /*
	USB_DBG_FLOWTIMCRY(iprintf(
		"[USB FLOWTIMCRY] usb_hcd_giveback_urb() Call usb_free_urb() ..\n"
    	));
 	usb_put_urb(urb);
    */
 #else
	USB_DBG_MEM(iprintf(
		"[USB MEM] usb_hcd_giveback_urb() Not Call usb_free_urb() "
        "!!\n"
        )) ;
    /*
	USB_DBG_INIT(iprintf(
		"[USB INIT] usb_hcd_giveback_urb() Not Call usb_free_urb() "
        "!!\n"
        )) ;
    
    dprintf(
       	"[USB] usb_hcd_giveback_urb() Not Call usb_free_urb() !! \n" ) ;
    */
 #endif 
 
 USB_DBG_FLOWTIMCRY(iprintf("[USB FLOWTIMCRY] usb_hcd_giveback_urb() End.\n" )) ;
}
/*
#ifndef _ETCPIP
	EXPORT_SYMBOL (usb_hcd_giveback_urb);
#endif // ndef _ETCPIP
*/
/*-------------------------------------------------------------------------*/

/**
 * usb_hcd_irq - hook IRQs to HCD framework (bus glue)
 * @irq: the IRQ being raised
 * @__hcd: pointer to the HCD whose IRQ is beinng signaled
 * @r: saved hardware registers
 *
 * When registering a USB bus through the HCD framework code, use this
 * to handle interrupts.  The PCI glue layer does so automatically; only
 * bus glue for non-PCI system busses will need to use this.
 */
#ifndef _ETCPIP
/*
	void usb_hcd_irq(int irq, void *__hcd, struct pt_regs * r)
#else
	void usb_hcd_irq(int irq, void *__hcd)
#endif
{
 struct usb_hcd		*hcd = __hcd;
 int			start = hcd->state;

 #ifndef _ETCPIP
	if (unlikely (hcd->state == USB_STATE_HALT))	// irq sharing? 
  #else
	if (hcd->state == USB_STATE_HALT)
 #endif
       {
		return;
       }

 #ifndef _ETCPIP
	hcd->driver->irq (hcd, r);
  #else
	hcd->driver->irq();
 #endif
    
 if (hcd->state != start && hcd->state == USB_STATE_HALT)
   {
	usb_hc_died (hcd);
   }
}

#ifndef _ETCPIP
	EXPORT_SYMBOL (usb_hcd_irq);
*/
#endif // ndef _ETCPIP

/*-------------------------------------------------------------------------*/

static void hcd_panic (void *_hcd)
{
	struct usb_hcd *hcd = _hcd;
	hcd->driver->stop (hcd);
}

/**
 * usb_hc_died - report abnormal shutdown of a host controller (bus glue)
 * @hcd: pointer to the HCD representing the controller
 *
 * This is called by bus glue to report a USB host controller that died
 * while operations may still have been pending.  It's called automatically
 * by the PCI glue, so only glue for non-PCI busses should need to call it. 
 */
void usb_hc_died (struct usb_hcd *hcd)
{
	struct list_head	*devlist, *urblist;
	struct hcd_dev		*dev;
	struct urb		*urb;
	unsigned long		flags;
	
	/* flag every pending urb as done */
 #ifndef NO_SPINLOCK
	spin_lock_irqsave (&hcd_data_lock, flags);
 #endif // ndef NO_SPINLOCK
	list_for_each (devlist, &hcd->dev_list) {
		dev = list_entry (devlist, struct hcd_dev, dev_list);
		list_for_each (urblist, &dev->urb_list) {
			urb = list_entry (urblist, struct urb, urb_list);
 #ifdef USB_DBGSTRING_ON
			dev_dbg (hcd->controller, "shutdown %s urb %p pipe %x, current status %d\n",
				hcd->self.bus_name, urb, urb->pipe, urb->status);
  #else
			dev_dbg (hcd->controller, "shutdown urb %p pipe %x, current status %d\n",
				urb, urb->pipe, urb->status);
 #endif
			if (urb->status == -EINPROGRESS)
				urb->status = -ESHUTDOWN;
		}
	}
    
 #ifndef NO_TIMER
	urb = (struct urb *) hcd->rh_timer.data;
  #else
    urb = hcd->rh_urb ;
 #endif
	if (urb)
		urb->status = -ESHUTDOWN;
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore (&hcd_data_lock, flags);
 #endif // ndef NO_SPINLOCK

 /* hcd->stop() needs a task context */
 
 #ifndef NO_WORKQUEUE
	INIT_WORK (&hcd->work, hcd_panic, hcd);
	(void) schedule_work (&hcd->work);
 #endif // NO_WORKQUEUE
}

#ifndef _ETCPIP
	EXPORT_SYMBOL (usb_hc_died);
#endif // ndef _ETCPIP

// ====================================================================
// End of File : hcd.c

