/*
 * USB hub driver.
 *
 * (C) Copyright 1999 Linus Torvalds
 * (C) Copyright 1999 Johannes Erdfelt
 * (C) Copyright 1999 Gregory P. Smith
 * (C) Copyright 2001 Brad Hards (bhards@bigpond.net.au)
 *
 */

// #ifdef _ETCPIP
#ifdef _NEWLIB
  #include "sys_type.h"
  #include "time.h"
#endif
#include "feature.h"
#define __KERNEL__
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_TIMER
	#include "PoTimer.h" // osport_timer.h
#endif // ndef NO_TIMER
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
// #endif // _ETCPIP

#ifndef _ETCPIP
   #include <linux/config.h>
   #include <linux/kernel.h>
#endif // ndef _ETCPIP
#ifndef _ETCPIP
    #include <linux/errno.h>
 #else
	#include "PoErrno.h" // osport_errno.h
#endif
#ifndef _ETCPIP
   #include <linux/module.h>
   #include <linux/completion.h>
   #include <linux/sched.h>
#endif // ndef _ETCPIP

#ifndef _ETCPIP
   #include <linux/list.h>
 #else
   #include "PoList.h" // osport_list.h <linux/list.h>
#endif

#ifndef _ETCPIP
   #include <linux/slab.h>
   #include <linux/smp_lock.h>
    #include <linux/ioctl.h>
#endif // ndef _ETCPIP

   #ifdef CONFIG_USB_DEBUG
      #define DEBUG
    #else
	  #undef DEBUG
   #endif

#ifndef _ETCPIP
   #include <linux/usb.h>
 #else
   #include "usb.h"
	#include "UsbSysInit.h"
#endif

#ifndef _ETCPIP
	#include <linux/usbdevice_fs.h>
	#include <linux/suspend.h>
#endif // ndef _ETCPIP
   
#ifndef NO_SEMPHORE
   #include <asm/semaphore.h>
#endif // ndef NO_SEMPHORE
#ifndef _ETCPIP
   #include <asm/uaccess.h>
   #include <asm/byteorder.h>
#endif // ndef _ETCPIP

#include "UsbHcd.h"  // ../core/hcd.h
#include "hub.h"

// #ifdef _ETCPIP
// 2003/3/21 B. S. Liaw
// #include "timer.h"
/*
#if defined(USB_OHCI)
	#include "ohci.h" // not usb-ohci.h
#elif defined(USB_SOTA)   // Sota USB 1.1 Host
	// 2003/10/23 B. S. Liaw
	#include "sota.h"
#endif
*/
#include "hcdx.h" // ohci.h not usb-ohci.h , sota.h
// #endif // _ETCPIP

#ifdef USB_MEM_LIB
	#include "UsMemLib.h" // UsbMemLib.h
#endif // USB_MEM_LIB
#ifdef _USB_SYSLOG
	#include "syslog.h"
	#include "logStr.h"
	#include "config.h"
	extern struct III_Config_t gSetting;
 #else
	// #error: Please Make Sure : Not _USB_SYSLOG !!
	//#warning: Not _USB_SYSLOG !!
#endif // _SYSLOG


#ifdef SOTA_HUB_BUGFIX
   /*
   void usb_hub_port_enable(struct usb_device *hub, int port) ;
   */
#endif // SOTA_HUB_BUGFIX
#ifdef _USB_SYSLOG
#include "config.h"
extern struct III_Config_t gSetting;
#endif // _USB_SYSLOG

/* Wakes up khubd */
#ifndef NO_SPINLOCK
   static spinlock_t hub_event_lock = SPIN_LOCK_UNLOCKED;
#endif // ndef NO_SPINLOCK

#ifndef NO_SEMPHORE
   static DECLARE_MUTEX(usb_address0_sem);
#endif // ndef NO_SEMPHORE

static LIST_HEAD(hub_event_list);	/* List of hubs needing servicing */

static LIST_HEAD(hub_list);		// List of all hubs (for cleanup) 

// struct list_head  global_hub_list_head;	// all hubs 
/*
static LIST_HEAD(global_hub_list_head);		// List of all hubs (for hub_task) 
*/

#ifndef NO_WAITQUEUE
	/*
    static DECLARE_WAIT_QUEUE_HEAD(khubd_wait);
    */
#endif // ndef NO_WAITQUEUE

static int khubd_pid = 0;			/* PID of khubd */

#ifndef _ETCPIP
    // in <linux/completion.h>
	static DECLARE_COMPLETION(khubd_exited);
#endif // ndef _ETCPIP

/*
#ifdef _ETCPIP
   #define MAX_USB_PORT  8
	unsigned char usb_port_status_array[MAX_USB_PORT][sizeof(struct usb_port_status)] ;
   unsigned char usb_hub_status_array[sizeof(struct usb_hub_status)] ;
   unsigned char usb_device_descriptor_array[sizeof(struct usb_device_descriptor)] ;
#endif //_ETCPIP
*/

#ifdef NO_SYS_FS
	// static struct device  dev_dev ;
	static struct device_driver hub_dev_driver ;
    char		hub_dev_driver_name[] = "hub-driver" ;
#endif // _ETCPIP

#ifdef RH_INT_BY_POLL
	/*
	struct usb_hub *global_hub = (struct usb_hub *)0 ;
    */
    /*
	struct usb_hub *(global_hub[USB_HUB_MAX]) ;
    int    global_hub_count = 0 ;
    */
    /*
	int hub_in_submit = 0 ;
    */
#endif

volatile int usb_hub_init_end = 0 ;
/*
volatile int usb_hub_plug_out = 1 ;
*/  

// #ifdef	DEBUG
static inline char *portspeed (int portstatus)
{
 if (portstatus & (1 << USB_PORT_FEAT_HIGHSPEED))
    	return "480 Mb/s";
 else if (portstatus & (1 << USB_PORT_FEAT_LOWSPEED))
		return "1.5 Mb/s";
 else
		return "12 Mb/s";
}
// #endif

/* for dev_info, dev_dbg, etc */
static inline struct device *hubdev (struct usb_device *dev)
{
 return &dev->actconfig->interface [0].dev;
}

/* USB 2.0 spec Section 11.24.4.5 */
static int usb_get_hub_descriptor(struct usb_device *dev, void *data,
	 int size)
{
    
 #ifdef USB_NO_VMM // no vitrtul memory management
	#ifdef USB_BS_ASSERT
		if ( !(USB_IS_UN_CACHED(data)) ) {
				iprintf("[ASSERT][USB] usb_get_hub_descriptor() : " ) ;
    		    iprintf("Error : Not Un-Cached !!\n" ) ;
			    iprintf("[ASSERT][USB] data addr = %08XH \n", data ) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
	#endif // USB_BS_ASSERT
 #endif // USB_NO_VMM
 
 USB_DBG_PIPE(iprintf(
    "[USB PIPE] usb_get_hub_descriptor() : usb_rcvctrlpipe(dev, 0) = %08XH \n",
    usb_rcvctrlpipe(dev, 0))) ;
 USB_DBG_PIPE(iprintf(
    "[USB PIPE] usb_get_hub_descriptor() : usb_pipedevice(usb_rcvctrlpipe(dev, 0)) = %d \n",
    usb_pipedevice(usb_rcvctrlpipe(dev, 0))
    )) ;

 USB_DBG_HUB(iprintf("[USB HUB] dev->devnum = %d \n",dev->devnum )) ;
 /*
 if ((dev->devnum > 1) && (usb_hub_plug_out > 0)) {
	USB_DBG_INFO(iprintf("[USB INFO] usb_get_hub_descriptor() : HUB Plug-Out !! \n" )) ;
    return -1 ;
   }
 */
 USB_DBG_HUB(iprintf("[USB HUB] usb_get_hub_descriptor() : End with call usb_control_msg() .. \n" )) ;
    
 return usb_control_msg(dev, usb_rcvctrlpipe(dev, 0),
		USB_REQ_GET_DESCRIPTOR, USB_DIR_IN | USB_RT_HUB,
		USB_DT_HUB << 8, 0, data, size,
        // HZ
        USB_HUB_TIMEOUT
        );
}

/*
 * USB 2.0 spec Section 11.24.2.1
 */
static int usb_clear_hub_feature(struct usb_device *dev, int feature)
{
 #ifdef USB_NO_VMM // no vitrtul memory management
	#ifdef USB_BS_ASSERT
		if ( !(USB_IS_UN_CACHED(feature)) ) {
				iprintf("[ASSERT][USB] usb_clear_hub_feature() : " ) ;
    		    iprintf("Error : Not Un-Cached !!\n" ) ;
			    iprintf("[ASSERT][USB] feature addr = %08XH \n", feature ) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
	#endif // USB_BS_ASSERT
 #endif // USB_NO_VMM
 
 USB_DBG_PIPE(iprintf(
    "[USB PIPE] usb_sndctrlpipe(dev, 0) = %08XH \n",
    usb_sndctrlpipe(dev, 0))) ;
 USB_DBG_PIPE(iprintf(
    "[USB PIPE] usb_pipedevice(usb_sndctrlpipe(dev, 0)) = %d \n",
    usb_pipedevice(usb_sndctrlpipe(dev, 0))
    )) ;
    
 USB_DBG_HUB(iprintf("[USB HUB] dev->devnum = %d \n",dev->devnum )) ;
 /*
 if ((dev->devnum > 1) && (usb_hub_plug_out > 0)) {
	USB_DBG_INFO(iprintf("[USB INFO] usb_clear_hub_feature() : HUB Plug-Out !! \n" )) ;
    return -1 ;
   }
 */
 USB_DBG_HUB(iprintf("[USB ROOOTHUB] usb_clear_hub_feature() : End with call usb_control_msg() .. \n" )) ;
    
 return usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
		USB_REQ_CLEAR_FEATURE, USB_RT_HUB, feature, 0,
        NULL, 0,
        // HZ
        USB_HUB_TIMEOUT
        );
}

/*
 * USB 2.0 spec Section 11.24.2.2
 * BUG: doesn't handle port indicator selector in high byte of wIndex
 */
static int usb_clear_port_feature(struct usb_device *dev, int port, int feature)
{
 USB_DBG_PIPE(iprintf(
    "[USB PIPE] usb_sndctrlpipe(dev, 0) = %08XH \n",
    usb_sndctrlpipe(dev, 0))) ;
 USB_DBG_PIPE(iprintf(
    "[USB PIPE] usb_pipedevice(usb_sndctrlpipe(dev, 0)) = %d \n",
    usb_pipedevice(usb_sndctrlpipe(dev, 0))
    )) ;
    
 USB_DBG_HUB(iprintf("[USB HUB] dev->devnum = %d \n",dev->devnum )) ;
 /*
 if ((dev->devnum > 1) && (usb_hub_plug_out > 0)) {
	USB_DBG_INFO(iprintf("[USB INFO] usb_clear_port_feature() : HUB Plug-Out !! \n" )) ;
    return -1 ;
   }
 */
 USB_DBG_HUB(iprintf("[USB HUB] usb_clear_port_feature() : End with call usb_control_msg() .. \n" )) ;
    
 return usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
		USB_REQ_CLEAR_FEATURE, USB_RT_PORT, feature,
               port, NULL, 0,
               // HZ
               USB_HUB_TIMEOUT
               );
}

/*
 * USB 2.0 spec Section 11.24.2.13
 * BUG: doesn't handle port indicator selector in high byte of wIndex
 */
// static 
int usb_set_port_feature(struct usb_device *dev, int port,
			int feature)
{
 int rc;
 
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_set_port_feature() Begin ..\n"));
 
 USB_DBG_PIPE(iprintf(
    "[USB PIPE] usb_sndctrlpipe(dev, 0) = %08XH \n",
    usb_sndctrlpipe(dev, 0))) ;
 USB_DBG_PIPE(iprintf(
    "[USB PIPE] usb_pipedevice(usb_sndctrlpipe(dev, 0)) = %d \n",
    usb_pipedevice(usb_sndctrlpipe(dev, 0))
    )) ;
    
 USB_DBG_HUB(iprintf(
	"[USB ROOTHUB] usb_set_port_feature() : Control Message .. \n"
    )) ;
    
 USB_DBG_HUB(iprintf("[USB HUB] dev->devnum = %d \n",dev->devnum )) ;
 /*
 if ((dev->devnum > 1) && (usb_hub_plug_out > 0)) {
	USB_DBG_INFO(iprintf("[USB INFO] usb_set_port_feature() : HUB Plug-Out !! \n" )) ;
    return -1 ;
   }
 */
 USB_DBG_FLOW(iprintf(
	"[USB FLOW] usb_set_port_feature() End with call usb_control_msg() ..\n" )) ;
	
 if (feature == USB_PORT_FEAT_SW_UNPLUG)  /* Charles 20040429 */
 {
	dev->sw_unplug[port-1] = 1;
	feature = USB_PORT_FEAT_RESET;
 }
	
 if (usb_control_msg(dev,usb_sndctrlpipe(dev,0),USB_REQ_SET_FEATURE,USB_RT_PORT,feature,port,NULL,0,USB_HUB_TIMEOUT) < 0)
	rc = 0;
 else
 	rc = 1;
 return rc;
}

/*
 * USB 2.0 spec Section 11.24.2.6
 */
static int usb_get_hub_status(struct usb_device *dev, void *data)
{
    
 #ifdef USB_NO_VMM // no vitrtul memory management
	#ifdef USB_BS_ASSERT
		if ( !(USB_IS_UN_CACHED(data)) ) {
				iprintf("[ASSERT][USB] usb_get_hub_status() : " ) ;
    		    iprintf("Error : Not Un-Cached !!\n" ) ;
			    iprintf("[ASSERT][USB] data addr = %08XH \n", data ) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
	#endif // USB_BS_ASSERT
 #endif // USB_NO_VMM
 
 USB_DBG_PIPE(iprintf(
    "[USB PIPE] usb_get_hub_status() : "
    "usb_rcvctrlpipe(dev, 0) = %08XH \n",
    usb_rcvctrlpipe(dev, 0))) ;
 USB_DBG_PIPE(iprintf(
    "[USB PIPE] usb_get_hub_status() : "
    "rcvctl PIPE device number = %d \n",
    usb_pipedevice(usb_rcvctrlpipe(dev, 0))
    )) ;

 USB_DBG_HUB(iprintf("[USB HUB] dev->devnum = %d \n",dev->devnum )) ;
 /*
 if ((dev->devnum > 1) && (usb_hub_plug_out > 0)) {
	USB_DBG_INFO(iprintf("[USB INFO] usb_get_hub_status() : HUB Plug-Out !! \n" )) ;
    return -1 ;
   }
 */
 USB_DBG_HUB(iprintf("[USB HUB] usb_get_hub_status() : End with call usb_control_msg() .. \n" )) ;
 return usb_control_msg(dev, usb_rcvctrlpipe(dev, 0),
		USB_REQ_GET_STATUS, USB_DIR_IN | USB_RT_HUB, 0, 0,
		data, sizeof(struct usb_hub_status),
        // HZ
        USB_HUB_TIMEOUT
        );
}

/*
 * USB 2.0 spec Section 11.24.2.7
 */
static int usb_get_port_status(struct usb_device *dev, int port, void *data)
{
 #ifdef USB_NO_VMM // no vitrtul memory management
	#ifdef USB_BS_ASSERT
		if ( !(USB_IS_UN_CACHED(data)) ) {
				iprintf("[ASSERT][USB] usb_get_port_status() : " ) ;
    		    iprintf("Error : Not Un-Cached !!\n" ) ;
			    iprintf("[ASSERT][USB] data addr = %08XH \n", data ) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
	#endif // USB_BS_ASSERT
 #endif // USB_NO_VMM
 
 USB_DBG_PIPE(iprintf(
    "[USB PIPE] usb_get_port_status() : usb_rcvctrlpipe(dev, 0) = %08XH \n",
    usb_rcvctrlpipe(dev, 0))) ;
 USB_DBG_PIPE(iprintf(
    "[USB PIPE] usb_get_port_status() : usb_pipedevice(usb_rcvctrlpipe(dev, 0)) = %d \n",
    usb_pipedevice(usb_rcvctrlpipe(dev, 0))
    )) ;
    
 USB_DBG_HUB(iprintf("[USB HUB] dev->devnum = %d \n",dev->devnum )) ;
 /*
 if ((dev->devnum > 1) && (usb_hub_plug_out > 0)) {
	USB_DBG_INFO(iprintf("[USB INFO] usb_get_port_status() : HUB Plug-Out !! \n" )) ;
    return -1 ;
   }
 */
 
 USB_DBG_HUB(iprintf("[USB HUB] usb_get_port_status() : End with call usb_control_msg() .. \n")) ;
 
 return usb_control_msg(dev, usb_rcvctrlpipe(dev, 0),
		USB_REQ_GET_STATUS, USB_DIR_IN | USB_RT_PORT, 0, port,
		data, sizeof(struct usb_hub_status),
        // HZ
        USB_HUB_TIMEOUT
        );
}

/* completion function, fires on port status changes and various faults */

#ifndef _ETCPIP
	/*
    static void hub_irq(struct urb *urb, struct pt_regs *regs)
    */
 #else
    static void hub_irq(struct urb *urb)
#endif
{
 struct usb_hub *hub = (struct usb_hub *)urb->context;
 unsigned long flags;
 int status;
 /*
 if ((hub->is_ext_hub) && (usb_hub_plug_out > 0)) {
			USB_DBG_INFO(iprintf("[USB INFO] hub_irq() : HUB Plug-Out !! \n" )) ;
            return ;
		   }
 */          
 USB_DBG_EVTLOOP(iprintf("[USB EVTLOOP] hub_irq() Begin ...\n"));
 
 USB_DBG_CALLBACK(iprintf("[USB CALLBACK] hub_irq() Begin .. \n")) ;
 USB_DBG_HUB(iprintf("[USB HUB] hub_irq() Begin .. \n")) ;

 if(! in_interrupt())
   {
	#ifdef RH_POLL_TSK 
    BLOCK_PREEMPTION;
	#endif // RH_POLL_TSK 
   }
  else
   {
	#ifdef USB_BS_ASSERT
		// 2003/7/24 B. S. Liaw
		iprintf("[ASSERT][USB] hub_irq() : " ) ;
    	iprintf("Error : Called in ISR , It's better Not do so !!\n" ) ;
	    iprintf("[ASSERT][USB] Continue ..\n") ;
	#endif
   }
   
 #ifdef USB_DBG_CALLBACK_ON
	#ifdef USB_BS_ASSERT
		// 2003/7/14 B. S. Liaw
		if (!( hub->hub_int_done == 0
    	     )
	       )
		   {
			iprintf("[ASSERT][USB] hub_irq() : " ) ;
    	    iprintf("Error : Called More than once !!\n" ) ;
	        iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif
    
	hub->hub_int_done = 1 ;
    
 	USB_DBG_CALLBACK(iprintf(
		"[USB CALLBACK] hub_irq() : Set Value : hub->hub_int_done=%d "
        "\n",
    	hub->hub_int_done
        )) ;
 #endif // USB_DBG_CALLBACK_ON

 /*    
   if(!in_interrupt())
     {
          dprintf("[TEST] hub_irq() : YIELD() ..\n") ;
      YIELD();
     }
    else
     {
      dprintf("[TEST] hub_irq() : in interrupt !!\n");
     }
 */    
 switch (urb->status)
   {
	case -ENOENT:		/* synchronous unlink */
	case -ECONNRESET:	/* async unlink */
	case -ESHUTDOWN:	/* hardware going away */
    
	 if(! in_interrupt())
	   {
	#ifdef RH_POLL_TSK 
	    UNBLOCK_PREEMPTION;
	#endif // RH_POLL_TSK 
	   }
		return;

	default:		/* presumably an error */
    
	#ifdef USB_BS_ASSERT
    if  ( !( (hub->error >= 0 ) && (hub->error < 100 )
        )  )
        {
         iprintf("[ASSERT][USB] hub_irq() 1 : " ) ;
         iprintf("Error : hub->error = %d value out of range !!\n",
                 hub->error ) ;
         iprintf("[ASSERT][USB] Continue ....\n") ;
        }
	#endif
 
	/* Cause a hub reset after 10 consecutive errors */
	dev_dbg (&hub->intf->dev, "transfer --> %d\n", urb->status);
	/*
	if (urb->status == -ENOENT)
        {
		 dprintf("[USB] -ENOENT status in hub_irq() %d \n",
               urb->status);
 		 return;
        }
	 dprintf("[USB] nonzero status in hub_irq() %d \n",
             urb->status);
	*/
	if ((++hub->nerrors < 10) || hub->error)
       {
		USB_DBG_INIT(iprintf("[USB INIT] error status in hub_irq() \n")) ;
		goto resubmit;
       }

		hub->error = urb->status;
		/* FALL THROUGH */
     
	    #ifdef USB_BS_ASSERT
        if  ( !( (hub->error >= 0 ) && (hub->error < 100 )
            )  )
            {
             iprintf("[ASSERT][USB] hub_irq() 2 : " ) ;
             iprintf("Error : hub->error = %d value out of range !!\n",
                     hub->error ) ;
             iprintf("[ASSERT][USB] Continue ....\n") ;
            }
    	#endif
     
		/* let khubd handle things */
	case 0:			/* we got data:  port status changed */
		break;
   }

 hub->nerrors = 0;

 /* Something happened, let khubd figure it out */
 #ifndef NO_SPINLOCK
 	/*
	spin_lock_irqsave(&hub_event_lock, flags);
    */
 #endif // ndef NO_SPINLOCK
 
 /*
 if( ! hub->event_empty)
   {
	dprintf("[USB] Error : HUB Event List Not Ready !!\n") ;
    
	goto resubmit ;
   }
   
 if (list_empty(&hub->event_list))
   {
	// USB_DBG_EVT(iprintf("[USB EVT] add event list ..\n"));
    USB_DBG_EVTLOOP(iprintf("[USB EVTLOOP] add event list ..\n"));
     
	// BLOCK_PREEMPTION;
	list_add(&hub->event_list, &hub_event_list);
	// UNBLOCK_PREEMPTION;
	#ifndef NO_WAITQUEUE
		wake_up(&khubd_wait);
	#endif // ndef NO_WAITQUEUE
   }
  else
   {
    dprintf("[USB] Error : HUB Event List Not Empty !!\n") ;
   }
 */
    
 #ifndef NO_SPINLOCK
	/*
	spin_unlock_irqrestore(&hub_event_lock, flags);
    */
 #endif // ndef NO_SPINLOCK
    
 // ---------------------   
 resubmit:
 // ---------------------   
 
 if(in_interrupt())
   {
	USB_DBG_HUB(iprintf( "[USB ROOTHUB] hub_irq() : in interrupt !!\n" )) ;
    
	#ifdef RH_INT_BY_POLL
    	/*
    	hub_in_submit = 0 ;
		USB_DBG_ROOTHUB(iprintf(
    		"[USB ROOTHUB] hub_irq() : Set Value : hub_in_submit=%d \n",
            hub_in_submit
            )) ;
        */
    	hub->hub_in_submit = 0 ;
		USB_DBG_HUB(iprintf(
    		"[USB ROOTHUB] hub_irq() : Set Value : hub[%08XH]->hub_in_submit=%d \n",
	    	hub,hub->hub_in_submit
		    )) ;
        /*
		if ((hub->is_ext_hub) && (usb_hub_plug_out > 0)) {
			USB_DBG_INFO(iprintf("[USB INFO] hub_irq() : HUB Plug-Out !! \n" )) ;
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
			goto    resubmit ;
		   }
       */
	 #else
	    dprintf("[TEST] hub_irq() : Not Submit Urb !!\n");
	#endif
   }
  else
   { 
	#ifdef RH_INT_BY_POLL
		// dprintf("[USB IRQ] hub_irq() : release USB Resource ..\n");
		// USB_DBG_IRQ(iprintf(
    	//		"[USB IRQ] hub_irq() : release USB Resource ..\n"));
		//USB_DBG_CALLBACK(iprintf(
    	//	"[USB CALLBACK] hub_irq() : release USB Resource ..\n"
    	//    ));
		// relres(MT_RES_USB)
		// MTqcmd_c(RELRES,MT_RES_USB) ;
    	hub->hub_in_submit = 0 ;
        
		USB_DBG_HUB(iprintf(
    		"[USB ROOTHUB] hub_irq() : Set Value : hub[%08XH]->hub_in_submit=%d \n",
	    	hub,hub->hub_in_submit
		    )) ;
        /*    
		if ((hub->is_ext_hub) && (usb_hub_plug_out > 0)) {
			USB_DBG_INFO(iprintf("[USB INFO] hub_irq() : HUB Plug-Out !! \n" )) ;
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
			goto    resubmit ;
		   }
        */
	 #else
     	/*
		USB_DBG_EVTLOOP(iprintf(
        	"[USB EVTLOOP] hub_irq() : usb_submit_urb() ...\n"
            ));
            
		#ifndef _ETCPIP
        	
			// if ((status = usb_submit_urb (hub->urb, GFP_ATOMIC)) != 0
            
		#else
			if ((status = usb_submit_urb(hub->urb)) != 0
		#endif
			// ENODEV means we raced disconnect() 
			&& status != -ENODEV)
	       {
			dev_err (&hub->intf->dev, "resubmit --> %d\n",
            		urb->status);
	       }
           
		USB_DBG_CALLBACK(iprintf(
		    "[USB CALLBACK] hub_irq() : dlytsk()/YIELD() .. \n")) ;
        // dprintf("[TEST] hub_irq() : YIELD() ..\n") ;
		#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
			// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
         #else
			YIELD();
        #endif
        */
	#endif
   }
   
 if(! in_interrupt())
   {
	#ifdef RH_POLL_TSK 
    UNBLOCK_PREEMPTION;
	#endif // RH_POLL_TSK 
   }
 /*
 if ((hub->is_ext_hub) && (usb_hub_plug_out > 0)) {
			USB_DBG_INFO(iprintf("[USB INFO] hub_irq() : HUB Plug-Out !! \n" )) ;
		   }
 */
	#ifdef USB_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
		   ) ) {
			iprintf("[ASSERT][USB] hub_irq() : " ) ;
    	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
   		   }
	#endif // USB_BS_ASSERT
 USB_DBG_HUB(iprintf("[USB HUB] hub_irq() End. \n")) ;
 USB_DBG_CALLBACK(iprintf(
    "[USB CALLBACK] hub_irq() End. \n")) ;
 USB_DBG_EVTLOOP(iprintf("[USB EVTLOOP] hub_irq() End.\n"));
} // function hub_irq()

// #ifdef USB_2_0
/* USB 2.0 spec Section 11.24.2.3 */
static inline int
hub_clear_tt_buffer(struct usb_device *hub, u16 devinfo, u16 tt)
{
 USB_DBG_HUB(iprintf("[USB HUB] hub->devnum = %d \n",hub->devnum )) ;
 // USB_DBG_INFO(iprintf("[USB INFO] hub->devnum = %d \n",hub->devnum )) ;
 /*
 if ((hub->devnum > 1) && (usb_hub_plug_out > 0)) {
	USB_DBG_INFO(iprintf("[USB INFO] hub_clear_tt_buffer() : HUB Plug-Out !! \n" )) ;
    return -1 ;
   }
 */
   
 USB_DBG_HUB(iprintf("[USB HUB] hub_clear_tt_buffer() : End with call usb_control_msg() .. \n" )) ;
 return usb_control_msg (hub, usb_rcvctrlpipe (hub, 0),
		HUB_CLEAR_TT_BUFFER, USB_DIR_IN | USB_RECIP_OTHER,
		devinfo, tt, 0, 0,
         // HZ
         USB_HUB_TIMEOUT
        ) ;
}

/*
 * enumeration blocks khubd for a long time. we use keventd instead, since
 * long blocking there is the exception, not the rule.  accordingly, HCDs
 * talking to TTs must queue control transfers (not just bulk and iso), so
 * both can talk to the same hub concurrently.
 */
static void hub_tt_kevent (void *arg)
{
 struct usb_hub		*hub = arg;
 unsigned long		flags;

 #ifndef NO_SPINLOCK
	spin_lock_irqsave (&hub->tt.lock, flags);
 #endif // ndef NO_SPINLOCK
 while (!list_empty (&hub->tt.clear_list))
   {
	struct list_head	*temp;
	struct usb_tt_clear	*clear;
	struct usb_device	*dev;
	int			status;

	temp = hub->tt.clear_list.next;
    
	BLOCK_PREEMPTION;
	clear = list_entry (temp, struct usb_tt_clear, clear_list);
	list_del (&clear->clear_list);
	UNBLOCK_PREEMPTION;

	/* drop lock so HCD can concurrently report other TT errors */
	#ifndef NO_SPINLOCK
		spin_unlock_irqrestore (&hub->tt.lock, flags);
	#endif // ndef NO_SPINLOCK
	dev = interface_to_usbdev (hub->intf);
	status = hub_clear_tt_buffer (dev, clear->devinfo, clear->tt);
	#ifndef NO_SPINLOCK
		spin_lock_irqsave (&hub->tt.lock, flags);
	#endif // ndef NO_SPINLOCK

	if (status)
       {
		#ifdef USB_DBGSTRING_ON
			err ("usb-%s-%s clear tt %d (%04x) error %d",
				dev->bus->bus_name, dev->devpath,
				clear->tt, clear->devinfo, status);
		 #else
			dprintf("[USB HUB] Error : usb clear tt %d (%04x) error %d",
				clear->tt, clear->devinfo, status);
		#endif
	   }
    #ifndef _ETCPIP
		kfree (clear);
	#elif defined(USB_DBG_MEM_C_ON)
		UsbFreeMemC(USB_MEM_C_L,(unsigned char *)clear,
        		sizeof(*clear));
	#else
		UsbFreeMemC(USB_MEM_C_L,(unsigned char *)clear);
    #endif
   }
   
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore (&hub->tt.lock, flags);
 #endif // ndef NO_SPINLOCK
}

/**
 * usb_hub_tt_clear_buffer - clear control/bulk TT state in high speed hub
 * @dev: the device whose split transaction failed
 * @pipe: identifies the endpoint of the failed transaction
 *
 * High speed HCDs use this to tell the hub driver that some split control or
 * bulk transaction failed in a way that requires clearing internal state of
 * a transaction translator.  This is normally detected (and reported) from
 * interrupt context.
 *
 * It may not be possible for that hub to handle additional full (or low)
 * speed transactions until that state is fully cleared out.
 */
void usb_hub_tt_clear_buffer (struct usb_device *dev, int pipe)
{
	struct usb_tt		*tt = dev->tt;
	unsigned long		flags;
	struct usb_tt_clear	*clear;

	/* we've got to cope with an arbitrary number of pending TT clears,
	 * since each TT has "at least two" buffers that can need it (and
	 * there can be many TTs per hub).  even if they're uncommon.
	 */
 #ifndef _ETCPIP
	if ((clear = kmalloc (sizeof *clear, SLAB_ATOMIC)) == 0) {
 #ifdef USB_DBGSTRING_ON
		err ("can't save CLEAR_TT_BUFFER state for hub at usb-%s-%s",
			dev->bus->bus_name, tt->hub->devpath);
 #else
		dprintf("[USBHUB] can't save CLEAR_TT_BUFFER state for hub ") ;
 #endif
		/* FIXME recover somehow ... RESET_TT? */
		return;
	}
  #else
	// clear = NonCache_contigmalloc(sizeof *clear) ;
    // NonCache_contigmalloc() cannot be free now !!
    
	clear = (struct usb_tt_clear *)
	#ifdef USB_DBG_MEM_C_ON
			UsbAllocMemC(USB_MEM_C_L, sizeof(*clear) ) ;
     #else
			UsbAllocMemC(USB_MEM_C_L) ;
    #endif

    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s
    USB_DBG_MEM(iprintf(
       "[USB MEM] clear = %08XH \n",clear));
    if ( clear == NULL )
       {
		dprintf("out of memory for clear \n");
		/* FIXME recover somehow ... RESET_TT? */
		return;
	   }
 #endif

	/* info that CLEAR_TT_BUFFER needs */
	clear->tt = tt->multi ? dev->ttport : 1;
	clear->devinfo = usb_pipeendpoint (pipe);
	clear->devinfo |= dev->devnum << 4;
	clear->devinfo |= usb_pipecontrol (pipe)
			? (USB_ENDPOINT_XFER_CONTROL << 11)
			: (USB_ENDPOINT_XFER_BULK << 11);
	if (usb_pipein (pipe))
		clear->devinfo |= 1 << 15;
	
	/* tell keventd to clear state for this TT */
 #ifndef NO_SPINLOCK
	spin_lock_irqsave (&tt->lock, flags);
 #endif // ndef NO_SPINLOCK
 
 BLOCK_PREEMPTION;
 list_add_tail (&clear->clear_list, &tt->clear_list);
 UNBLOCK_PREEMPTION;
 #ifndef NO_WORKQUEUE
	schedule_work (&tt->kevent);
 #endif // NO_WORKQUEUE
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore (&tt->lock, flags);
 #endif // ndef NO_SPINLOCK
}
// #endif // USB_2_0

static void usb_hub_power_on(struct usb_hub *hub)
{
 struct usb_device *dev;
 int i;

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_hub_power_on() Begin ..\n"));
 
 /* Enable power to the ports */
 
 USB_DBG_HUB(iprintf("[USB HUB] enabling power on all ports \n"));
 /*
 dev_dbg(hubdev(interface_to_usbdev(hub->intf)),
		"enabling power on all ports\n");
 */       
 dev = interface_to_usbdev(hub->intf);

 for (i = 0; i < hub->descriptor->bNbrPorts; i++)
   {
	USB_DBG_INFO(iprintf("[USB INFO] enabling power on port %d \n",i+1));
	usb_set_port_feature(dev, i + 1, USB_PORT_FEAT_POWER);
   }

 /* Wait for power to be enabled */
 USB_DBG_HUB(iprintf("[USB HUB] Wait for power to be enabled ..\n"));
 wait_ms(hub->descriptor->bPwrOn2PwrGood * 2);
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_hub_power_on() End.\n"));
}

static int usb_hub_configure(struct usb_hub *hub,
                 struct usb_endpoint_descriptor *endpoint)
{
 struct usb_device *dev = interface_to_usbdev (hub->intf);
 struct device *hub_dev;
 // struct usb_hub_status hubstatus;
 struct usb_hub_status *hubsts_mod ;		// 20021015 added by drsohn
 unsigned int pipe;
 int maxp, ret;
 char *message;

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_hub_configure() Begin ..\n"));
 
 hubsts_mod = 0x0000L ;
 
 #ifndef _ETCPIP
	//hub->descriptor = kmalloc(sizeof(*hub->descriptor), GFP_KERNEL);		// 20021012 remarked by drsohn
	hub->descriptor = kmalloc(sizeof(*hub->descriptor), GFP_KERNEL|GFP_DMA);		// 20021012 modified by drsohn
	if (!hub->descriptor) {
		message = "can't kmalloc hub descriptor";
		err("Unable to kmalloc %Zd bytes for hub descriptor", sizeof(*hub->descriptor));
		ret = -ENOMEM;
		goto fail;
	}
  #else
	#ifndef USB_MEM_LIB
		hub->descriptor = (struct usb_hub_descriptor *)
	    NonCache_contigmalloc(sizeof(*hub->descriptor)) ;
	    // NonCache_contigmalloc() cannot be free now !!
     #else
	   	// alloc_mem(sizeof(*hub->descriptor)) ;
        
		USB_DBG_MEM_C(iprintf(
		    "[USB MEM_U] usb_hub_configure() Call UsbAllocMemU() ..\n"
            ));
            
		hub->descriptor = (struct usb_hub_descriptor *)
        	UsbAllocMemU(USB_MEM_U_S
				#ifdef USB_DBG_MEM_U_ON
            	,sizeof(*hub->descriptor)
                #endif
                );
        //  if (relUsbMem(USB_MEM_U_S, mRu) != SUCCESS) {
    #endif
    
    USB_DBG_MEM(iprintf(
       "[USB MEM] hub->descriptor = %08XH \n",hub->descriptor));
    if ( hub->descriptor == NULL )
       {
		dprintf("[USB] Error : out of memory !!\n");
		message = "can't kmalloc hub descriptor";
		ret = -ENOMEM;
		goto fail;
	   }
 #endif

	/* Request the entire hub descriptor.
	 * hub->descriptor can handle USB_MAXCHILDREN ports,
	 * but the hub can/will return fewer bytes here.
	 */
	ret = usb_get_hub_descriptor(dev, hub->descriptor,
			sizeof(*hub->descriptor));
		/* <hub->descriptor> is large enough for a hub with 127 ports;
		 * the hub can/will return fewer bytes here. */
 if (ret < 0)
    {
	 err("Unable to get hub descriptor (err = %d)", ret);
		message = "can't read hub descriptor";
		goto fail;
         // NonCache_contigmalloc() cannot be free now !!
	} else if (hub->descriptor->bNbrPorts > USB_MAXCHILDREN) {
		message = "hub has too many ports!";
		ret = -ENODEV;
		goto fail;
	}
    
 hub_dev = hubdev(dev);
 
 #ifdef USB_ALL_HUB_1_PORT
    USB_DBG_INIT(iprintf(
	    "[USB INIT] Set This Hub to 1 Port Only !!\n"
        )) ;
    hub->descriptor->bNbrPorts = 1 ;
 #endif //  USB_RH_1_PORT
 
 dev->maxchild = hub->descriptor->bNbrPorts;
 
 USB_DBG_INIT(iprintf("[USB INIT] %d port%s detected \n",
    hub->descriptor->bNbrPorts, (hub->descriptor->bNbrPorts == 1)
                                ? "" : "s"
    ));
 #ifdef USB_DBG_INFO_ON
 dev_info (hub_dev, "%d port%s detected\n", dev->maxchild,
		(dev->maxchild == 1) ? "" : "s");
 #endif
        

 le16_to_cpus(&hub->descriptor->wHubCharacteristics);

 if (hub->descriptor->wHubCharacteristics & HUB_CHAR_COMPOUND)
   {
	int	i;
	char	portstr [USB_MAXCHILDREN + 1];
    
    
	USB_DBG_INIT(iprintf("[USB INIT] part of a compound device\n")) ;

	for (i = 0; i < dev->maxchild; i++)
			portstr[i] = hub->descriptor->DeviceRemovable
				    [((i + 1) / 8)] & (1 << ((i + 1) % 8))
				? 'F' : 'R';
	portstr[dev->maxchild] = 0;
	#ifdef USB_DBG_INIT_ON
		dev_dbg(hub_dev, "compound device; port removable status: %s\n", portstr);
	#endif
 
	USB_DBG_INIT(iprintf("[USB INIT] port removable status: %s\n",
    	portstr));
 
   }
  else
   {
	USB_DBG_INIT(iprintf("[USB INIT] standalone hub\n"));
	#ifdef USB_DBG_INIT_ON
		dev_dbg(hub_dev, "standalone hub\n");
	#endif
   }

 switch (hub->descriptor->wHubCharacteristics & HUB_CHAR_LPSM)
   {
	case 0x00:
			USB_DBG_INIT(iprintf("[USB INIT] ganged power switching\n"));
			#ifdef USB_DBG_INIT_ON
			dev_dbg(hub_dev, "ganged power switching\n");
			#endif
			break;
		case 0x01:
			USB_DBG_INIT(iprintf(
            "[USB INIT] individual port power switching\n"));
			#ifdef USB_DBG_INIT_ON
			dev_dbg(hub_dev, "individual port power switching\n");
			#endif
			break;
		case 0x02:
		case 0x03:
			USB_DBG_INIT(iprintf(
            "[USB INIT] unknown reserved power switching mode\n"));
			#ifdef USB_DBG_INIT_ON
			dev_dbg(hub_dev, "unknown reserved power switching mode\n");
			#endif
			break;
	}

	switch (hub->descriptor->wHubCharacteristics & HUB_CHAR_OCPM) {
		case 0x00:
			USB_DBG_INIT(iprintf(
            "[USB INIT] global over-current protection\n"));
			#ifdef USB_DBG_INIT_ON
			dev_dbg(hub_dev, "global over-current protection\n");
			#endif
			break;
		case 0x08:
			USB_DBG_INIT(iprintf(
            "[USB INIT] individual port over-current protection\n"));
			#ifdef USB_DBG_INIT_ON
			dev_dbg(hub_dev, "individual port over-current protection\n");
			#endif
			break;
		case 0x10:
		case 0x18:
			USB_DBG_INIT(iprintf(
            "[USB INIT] no over-current protection\n"));
			#ifdef USB_DBG_INIT_ON
				dev_dbg(hub_dev, "no over-current protection\n");
            #endif
         break;
	}

 #ifndef NO_SPINLOCK
	spin_lock_init (&hub->tt.lock);
 #endif // ndef NO_SPINLOCK
 
 INIT_LIST_HEAD (&hub->tt.clear_list);
 
 #ifndef NO_WORKQUEUE
	INIT_WORK (&hub->tt.kevent, hub_tt_kevent, hub);
 #endif // NO_WORKQUEUE
 switch (dev->descriptor.bDeviceProtocol)
    {
		case 0:
			break;
		case 1:
			USB_DBG_HUB(iprintf("[USB HUB] Single TT\n"));
			#ifdef USB_DBG_INIT_ON
			dev_dbg(hub_dev, "Single TT\n");
			#endif
			hub->tt.hub = dev;

			break;
		case 2:
			USB_DBG_HUB(iprintf("[USB HUB] Multiple TT\n"));
			#ifdef USB_DBG_INIT_ON
			dev_dbg(hub_dev, "TT per port\n");
			#endif
			hub->tt.hub = dev;
			hub->tt.multi = 1;

			break;
		default:
			USB_DBG_INIT(iprintf(
                "[USB INIT] Unrecognized hub protocol %d\n",
				dev->descriptor.bDeviceProtocol));
			#ifdef USB_DBG_INIT_ON
			dev_dbg(hub_dev, "Unrecognized hub protocol %d\n",
				dev->descriptor.bDeviceProtocol);
			#endif
			break;
	}

 switch (hub->descriptor->wHubCharacteristics & HUB_CHAR_TTTT)
    {
	case 0x00:
		if (dev->descriptor.bDeviceProtocol != 0)
           {
			USB_DBG_HUB(iprintf(
                "[USB INIT] TT requires at most 8 FS bit times\n"));
			#ifdef USB_DBG_INIT_ON
				dev_dbg(hub_dev, "TT requires at most 8 FS bit times\n");
			#endif
           }
		break;
	case 0x20:
		USB_DBG_HUB(iprintf(
            "[USB INIT] TT requires at most 16 FS bit times\n"));
		#ifdef USB_DBG_INIT_ON
			dev_dbg(hub_dev, "TT requires at most 16 FS bit times\n");
		#endif
		break;
	case 0x40:
		USB_DBG_HUB(iprintf(
            "[USB INIT] TT requires at most 24 FS bit times\n"));
		#ifdef USB_DBG_INIT_ON
			dev_dbg(hub_dev, "TT requires at most 24 FS bit times\n");
		#endif
		break;
	case 0x60:
		USB_DBG_HUB(iprintf(
            "[USB INIT] TT requires at most 32 FS bit times\n"));
		#ifdef USB_DBG_INIT_ON
			dev_dbg(hub_dev, "TT requires at most 32 FS bit times\n");
        #endif
		break;
   }

 USB_DBG_INFO(iprintf(
	"[USB INFO] Port indicators are %s supported\n", 
    (hub->descriptor->wHubCharacteristics & HUB_CHAR_PORTIND)
    ? "" : "not")) ;

 USB_DBG_INFO(iprintf(
       "[USB INFO] power on to power good time: %dms\n",
       hub->descriptor->bPwrOn2PwrGood * 2));
 USB_DBG_INFO(iprintf(
       "[USB INFO] hub controller current requirement: %dmA\n",
       hub->descriptor->bHubContrCurrent));
       

 
 // 2003/7/14 B. S. Liaw BugFix , Need Un-Cached
 #ifndef _ETCPIP
 	/*
	hubsts_mod = (struct usb_hub_status *)
                    kmalloc(sizeof(struct usb_hub_status),
                    GFP_KERNEL|GFP_DMA);
                    // 20021015 added by drsohn
	if (!hubsts_mod) {
		err("couldn't kmalloc struct");
		return NULL;
	}
    */
  #else
	hubsts_mod = (struct usb_hub_status *)
        	UsbAllocMemU(USB_MEM_U_S
					#ifdef USB_DBG_MEM_U_ON
    	             ,sizeof(struct usb_hub_status)
	                #endif
                );
    USB_DBG_MEM(iprintf(
       "[USB MEM] hubsts_mod = %08XH \n",hubsts_mod));
    if ( hubsts_mod == NULL )
       {
		dprintf("[USB] Error : out of memory !!\n");
		return ;
	   }
 #endif
 
 #ifdef USB_NO_VMM // no vitrtul memory management
	#ifdef USB_BS_ASSERT
		if ( !(USB_IS_UN_CACHED(hubsts_mod)) ) {
				iprintf("[ASSERT][USB] usb_hub_configure() : " ) ;
    		    iprintf("Error : Not Un-Cached !!\n" ) ;
			    iprintf("[ASSERT][USB] hubsts_mod addr = %08XH \n", hubsts_mod ) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
	#endif // USB_BS_ASSERT
 #endif // USB_NO_VMM
 
 USB_DBG_FLOW(iprintf(
    "[USB FLOW] usb_hub_configure() call usb_get_hub_status() ..\n"));
    
 //ret = usb_get_hub_status(dev, &hubstatus);	// 20021015 remarked by drsohn
 ret = usb_get_hub_status(dev, hubsts_mod);	// 20021015 modified by drsohn
 
 /*
 memcpy(&hubstatus,hubsts_mod,sizeof(struct usb_hub_status));				// 20021015 added by drsohn
    
 #ifndef _ETCPIP
	kfree(hubsts_mod);														// 20021015 added by drsohn
  #else
    // NonCache_contigmalloc() cannot be free now !!
 #endif
 */
 
 
 if (ret < 0)
    {
	 dprintf("[USB] Unable to get hub status (err = %d)", ret);
		message = "can't get hub status";
		goto fail;
	}
 /*
 le16_to_cpus(&hubstatus.wHubStatus);
 */
 // le16_to_cpus(hubsts_mod->wHubStatus);
 le16_to_cpus(&(hubsts_mod->wHubStatus));
 
 /*
 USB_DBG_INFO(iprintf("[USB INFO] local power source is %s\n",
		(hubstatus.wHubStatus & HUB_STATUS_LOCAL_POWER)
   ? "lost (inactive)" : "good"));
 */
 USB_DBG_INFO(iprintf("[USB INFO] local power source is %s\n",
		(hubsts_mod->wHubStatus & HUB_STATUS_LOCAL_POWER)
   ? "lost (inactive)" : "good"));
   
 /*
 USB_DBG_INFO(iprintf("[USB INFO] %sover-current condition exists\n",
		(hubstatus.wHubStatus & HUB_STATUS_OVERCURRENT) ? "" : "no "));
 */
 USB_DBG_INFO(iprintf("[USB INFO] %sover-current condition exists\n",
		(hubsts_mod->wHubStatus & HUB_STATUS_OVERCURRENT) ? "" : "no "));

 /* Start the interrupt endpoint */
    
 USB_DBG_INFO(iprintf("[USB INFO] Start the interrupt endpoint \n")) ;
 pipe = usb_rcvintpipe(dev, endpoint->bEndpointAddress);
 maxp = usb_maxpacket(dev, pipe, usb_pipeout(pipe));

 if (maxp > sizeof(hub->buffer))
		maxp = sizeof(hub->buffer);

 #ifndef _ETCPIP
 	/*
	hub->urb = usb_alloc_urb(0, GFP_KERNEL);
    */
  #else
	hub->urb = usb_alloc_urb(0);
 #endif
 if (!hub->urb)
    {
	 iprintf("[USB] couldn't allocate interrupt urb \n");
		message = "couldn't allocate interrupt urb";
		ret = -ENOMEM;
		goto fail;
	}
    
 if (! hub->urb->hcpriv)
   {
	USB_DBG_HUB(iprintf("[USB HUB] usb_hub_configure(1) : hcpriv Null \n" )) ;
   }

 USB_DBG_HUB(iprintf("[USB HUB] usb_hub_configure() : FILL_INT_URB \n")) ;
 USB_DBG_EVT(iprintf(
    "[USB EVT] usb_hub_configure() : hub->urb = %08XH hub_irq = %08XH "
    "hub->error = %d\n",
    hub->urb,hub_irq, hub->error)) ;
    
 #ifdef USB_BS_ASSERT
    if  ( !( (hub->error >= 0 ) && (hub->error < 100 )
        )  )
        {
         iprintf("[ASSERT][USB] usb_hub_configure() : " ) ;
         iprintf("Error : hub->error = %d value out of range !!\n",
                 hub->error ) ;
         iprintf("[ASSERT][USB] Continue ....\n") ;
        }
 #endif
 
 usb_fill_int_urb(hub->urb, dev, pipe, hub->buffer, maxp, hub_irq,
		hub, endpoint->bInterval);
        
 if (! hub->urb->hcpriv)
   {
	USB_DBG_HUB(iprintf("[USB HUB] usb_hub_configure(2) : hcpriv Null \n" )) ;
   }

 /*       
 #ifdef _ETCPIP
	for (;;) 
	   {
		// USB_DBG_INIT(iprintf("[USB INIT] get USB Resource ..\n"));
	    if (getres(MT_RES_USB, 1000) != 0)
	       {
	        iprintf("[USB] usb_hub_configure() : getres Fail !!") ;
	        continue ;
	       }
       
	    if (hub_in_submit == 0)
	       {
	        break ;
	       }
	   }
   hub_in_submit = 1 ;
 #endif
 */
 
 #ifndef RH_INT_BY_POLL
	/*
	USB_DBG_ROOTHUB(iprintf(
    	"[USB ROOTHUB] usb_hub_configure() : usb_submit_urb() .. \n"
        )) ;
	#ifndef _ETCPIP
    	
		// ret = usb_submit_urb(hub->urb, GFP_KERNEL);
        
	#else
	    ret = usb_submit_urb(hub->urb);
	#endif
	 if (ret)
       {
		iprintf("[USB] usb_hub_configure() : usb_submit_urb failed (%d) !!\n", ret);
		message = "couldn't submit status urb";
		goto fail;
	   }
    */
 #endif
 
 #ifdef USB_DBG_ABORT   
   // vvv for debug Prefetch Error !!
   dprintf("[TEST] GetTime() = %08XH\n",GetTime()) ;
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] usb_hub_configure() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
			#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
				// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
             #else
            	YIELD();
            #endif
	#ifndef NO_TIMER
	   dprintf("[TEST] jiffies = %08XH\n",jiffies) ;
	#endif // ndef NO_TIMER
   // ^^^ for debug Prefetch   
 #endif   
		
 /* Wake up khubd */
 #ifndef NO_WAITQUEUE
	wake_up(&khubd_wait);
 #endif // ndef NO_WAITQUEUE

 USB_DBG_HUB(iprintf("[USB HUB] Power On the HUB \n")) ;
 usb_hub_power_on(hub);

 // 2003/7/14 B. S. Liaw BugFix , Need Un-Cached
 if(hubsts_mod)
   {
	#ifndef USB_MEM_LIB
 		kfree(hubsts_mod);
	 #else
		// 2003/7/14 B. S. Liaw BugFix , Need Un-Cached
		UsbFreeMemU(USB_MEM_U_S, (unsigned char *)
		    	hubsts_mod
				#if defined(USB_DBG_MEM_U_ON)
	        		,sizeof(struct usb_hub_status)
    	        #endif
		       ) ;
	#endif
   }
 hubsts_mod = 0x0000L ;
 
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_hub_configure() End.\n"));
 return 0;
 
 // ---------------
 fail:
 // ---------------
	dev_err (&hub->intf->dev, "config failed, %s (err %d)\n",
			message, ret);
	/* hub_disconnect() frees urb and descriptor */
    
	// 2003/7/14 B. S. Liaw BugFix , Need Un-Cached
 if(hubsts_mod)
   {
	#ifndef USB_MEM_LIB
 		kfree(hubsts_mod);
	 #else
		// 2003/7/14 B. S. Liaw BugFix , Need Un-Cached
		UsbFreeMemU(USB_MEM_U_S, (unsigned char *)
		    	hubsts_mod
				#if defined(USB_DBG_MEM_U_ON)
	        		,sizeof(struct usb_hub_status)
    	        #endif
		       ) ;
	#endif
   }
 hubsts_mod = 0x0000L ;
   
 USB_DBG_FLOW(iprintf(
    	"[USB FLOW] usb_hub_configure() End with Fail !!\n"));
 return ret;
} // function usb_hub_configure()

static void hub_disconnect(struct usb_interface *intf)
{
 struct usb_hub *hub = usb_get_intfdata(intf);
 unsigned long flags;
 struct usb_hcd		*hcd ;


 USB_DBG_HUB(iprintf("[USB HUB] hub_disconnect() Begin .. \n" )) ;
 USB_DBG_HUB(iprintf("[USB HUB] hub_disconnect() : intf=%08XH , hub=%08XH \n",intf, hub )) ;
 if (!hub)
   {
	USB_DBG_HUB(iprintf("[USB HUB] hub_disconnect() : Error : Null HUB !! \n" )) ;
	return;
   }
   
 /*
 usb_hub_plug_out = 1 ;
 
 // USB_DBG_INFO(iprintf("[USB INFO] hub_disconnect() : wait for hub->hub_in_submit == 0 ..\n" )) ;
 
 USB_DBG_INIT(iprintf("[USB INIT] hub_disconnect() : wait for hub->hub_in_submit == 0 ..\n" )) ;
 
 while (hub->hub_in_submit)
       {
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
        NULL ;
       }
 */

 // USB_DBG_INFO(iprintf("[USB INFO] hub_disconnect() : wait by wait_ms(300) ..\n" )) ;
 USB_DBG_INIT(iprintf("[USB INIT] hub_disconnect() : wait by wait_ms(300) ..\n" )) ;
 wait_ms(300) ;
 USB_DBG_INFO(iprintf("[USB INFO] hub_disconnect() : wait End.\n" )) ;
 
 
 USB_DBG_INIT(iprintf("[USB INIT] hub_disconnect() : intf=%08XH , Set NULL ..\n",intf )) ;
 usb_set_intfdata (intf, NULL);

 #ifndef NO_SPINLOCK
	/*
	spin_lock_irqsave(&hub_event_lock, flags);
    */
 #endif // ndef NO_SPINLOCK

 USB_DBG_HUB(iprintf("[USB HUB] hub_disconnect() : event/hub list .. \n" )) ;
 /* Delete it and then reset it */
 BLOCK_PREEMPTION;
 list_del(&hub->event_list);
 INIT_LIST_HEAD(&hub->event_list);
 list_del(&hub->hub_list);
 INIT_LIST_HEAD(&hub->hub_list);
 UNBLOCK_PREEMPTION;

 #ifndef NO_SPINLOCK
 	/*
	spin_unlock_irqrestore(&hub_event_lock, flags);
    */
 #endif // ndef NO_SPINLOCK

 #ifndef NO_SEMPHORE
	down(&hub->khubd_sem); /* Wait for khubd to leave this hub alone. */
	up(&hub->khubd_sem);
 #endif // ndef NO_SEMPHORE

 /* assuming we used keventd, it must quiesce too */
 #ifndef _ETCPIP
	if (hub->tt.hub)
		flush_scheduled_work ();
 #endif // ndef _ETCPIP

 if (hub->urb)
   {
	USB_DBG_HUB(iprintf("[USB HUB] hub_disconnect() call usb_unlink_urb(hub->urb) ..\n"));
	USB_DBG_INIT(iprintf("[USB INIT] hub_disconnect() : usb_unlink_urb(hub->urb) ..\n"));
	usb_unlink_urb(hub->urb);

    // 2003/7/9 B. S. Liaw BugFix   
	hcd = hub->urb->dev->bus->hcpriv_hcd;
    
	if (  hub->urb->dev == hcd->self.root_hub )
       { 
		USB_DBG_INIT(iprintf("[USB INIT] hub_disconnect() : Root HUB !! \n" )) ;
       }
      else
       { 
		USB_DBG_INIT(iprintf("[USB INIT] hub_disconnect() : Not Root HUB \n" )) ;
		if (hub->urb->hcpriv)
           { 
			USB_DBG_INIT(iprintf("[USB INIT] hub_disconnect() : free hcpriv .. \n" )) ;
			hcd->driver->urb_free_priv(hcd, hub->urb->hcpriv);
			hub->urb->hcpriv = NULL;
	       }
          else
           {
			USB_DBG_INIT(iprintf("[USB INIT] hub_disconnect() : hcpriv Null !! \n" )) ;
           }
       }
    // End BugFix
    
    /*
    if (hub->urb) {
		USB_DBG_HUB(iprintf("[USB HUB] hub_disconnect() call usb_unlink_urb(hub->urb) ..\n"));
		USB_DBG_INIT(iprintf("[USB INIT] hub_disconnect() : usb_unlink_urb(hub->urb) ..\n"));
	usb_unlink_urb(hub->urb);
    */    
		USB_DBG_HUB(iprintf("[USB HUB] hub_disconnect() call usb_free_urb() ..\n"));
	usb_free_urb(hub->urb);
	hub->urb = NULL;
    /*
       }
	*/
   }

 if (hub->descriptor)
   {
    #ifndef _ETCPIP
		kfree(hub->descriptor);
    #else
        UsbFreeMemU(USB_MEM_U_S,(unsigned char *)hub->descriptor
					#if defined(USB_DBG_MEM_U_ON)
            		, sizeof(*hub->descriptor)
                	#endif
                  ) ;
    #endif
	hub->descriptor = NULL;
   }
   
 /*
 USB_DBG_INIT(iprintf("[USB INIT] hub_disconnect() : list_del(&global_hub_list) ..\n" )) ;
 list_del(&global_hub_list);
 */
 
 /* Free the memory */
 
 #ifndef _ETCPIP
	kfree (hub);
  #else
	UsbFreeMemU(USB_MEM_U_S, (unsigned char *) hub
				#if defined(USB_DBG_MEM_U_ON)
        			, sizeof(struct usb_hub)
                #endif
              ) ;
 #endif
 USB_DBG_HUB(iprintf("[USB HUB] hub_disconnect() End. \n" )) ;
}


// Interface descriptors advertise a "Class/Subclass/Protocol" code,
// such as "09/00/00"

// return : 1 : found , 0 : not found

static int hub_check_class(struct usb_host_interface *desc)
{
 /*
 struct usb_interface *if_alt;
 struct usb_host_interface *ifd;
 */
 /// struct usb_endpoint_descriptor *epd, *epwrite, *epread;
 // int p, i, e;
 // int i ;
 
 // struct usb_interface_descriptor *desc ;

 // desc = &intf->altsetting[intf->act_altsetting].desc;
 
 // USB_DBG_HUB(iprintf("[USB HUB] hub_check_class() : usb_dev->actconfig->desc.bNumInterfaces=%d \n", usb_dev->actconfig->desc.bNumInterfaces )) ;
 // desc->bInterfaceClass

 USB_DBG_INIT(iprintf("[USB INIT] hub_check_class() : desc->desc.bInterfaceClass=%d \n", desc->desc.bInterfaceClass)) ;

 if(desc->desc.bInterfaceClass == USB_CLASS_HUB ) { // 9
	// USB_DBG_INIT(iprintf("[USB INIT] hub_check_class() : Found HUB Class , usb_dev->actconfig->desc.bNumInterfaces=%d \n", usb_dev->actconfig->desc.bNumInterfaces)) ;
	USB_DBG_INIT(iprintf("[USB INIT] hub_check_class() : Found HUB Class \n")) ;
    return 1 ;
   }
 
 /*
 if_alt = &usb_dev->actconfig->interface[usb_dev->actconfig->desc.bNumInterfaces];
 
 // Find out what we have. 

 for(i = 0; i < if_alt->num_altsetting; i++) {
	ifd = &if_alt->altsetting[i];

	if (ifd->desc.bInterfaceClass == USB_CLASS_HUB // 9
    	// || ifd->desc.bInterfaceSubClass != 1
        )
       {
		USB_DBG_INIT(iprintf("[USB INIT] hub_check_class() : Found HUB Class , usb_dev->actconfig->desc.bNumInterfaces=%d , i=%d\n", usb_dev->actconfig->desc.bNumInterfaces, i)) ;
        return 1 ;
		// continue;
       }
	
	//if (ifd->desc.bInterfaceProtocol < USBLP_FIRST_PROTOCOL ||
	//	    ifd->desc.bInterfaceProtocol > USBLP_LAST_PROTOCOL)
    //   {
	//	continue;
    //   }
	
	// Look for bulk OUT and IN endpoints. 
    // return 1 ;
   } // for i 
 */
 
 return 0 ;
}

static int hub_probe(struct usb_interface *intf,
					const struct usb_device_id *id)
{
 struct usb_host_interface *desc;
 struct usb_endpoint_descriptor *endpoint;
 struct usb_device *usb_dev;
 struct usb_hub *hub;
 unsigned long flags;
 int x ;

 USB_DBG_FLOW(iprintf("[USB FLOW] hub_probe() Begin ..\n"));
 
 usb_hub_init_end = 0 ;

 
 desc = intf->altsetting + intf->act_altsetting;
 usb_dev = interface_to_usbdev(intf);

 if( ! hub_check_class(desc) ) {
	USB_DBG_INIT(iprintf("[USB INIT] Not HUB Class !! \n")) ;
	usb_hub_init_end = 1 ;
	return -EIO;
   }

 /* Some hubs have a subclass of 1, which AFAICT according to the */
 /*  specs is not defined, but it works */
 if(    (desc->desc.bInterfaceSubClass != 0)
     && (desc->desc.bInterfaceSubClass != 1)
   )
   {
    // ---------------------
	descriptor_error:
    // ---------------------
	dev_err (&intf->dev, "bad descriptor, ignoring hub\n");
    dprintf("[USB] bad descriptor, ignoring hub\n");
	usb_hub_init_end = 1 ;
	return -EIO;
   }

 /* Multiple endpoints? What kind of mutant ninja-hub is this? */
 if (desc->desc.bNumEndpoints != 1) {
		goto descriptor_error;
	}

 endpoint = &desc->endpoint[0].desc;

 /* Output endpoint? Curiouser and curiouser.. */
 if (!(endpoint->bEndpointAddress & USB_DIR_IN)) {
		err("Device #%d is hub class, but has output endpoint?",
			usb_dev->devnum);
		goto descriptor_error;
	}

 /* If it's not an interrupt endpoint, we'd better punt! */
 if ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
			!= USB_ENDPOINT_XFER_INT) {
		err("Device #%d is hub class, but has endpoint other than interrupt?",
			usb_dev->devnum);
		goto descriptor_error;
	usb_hub_init_end = 1 ;
		return -EIO;
	}

 /* We found a hub */
 USB_DBG_INIT(iprintf("[USB INIT] hub_probe() : USB hub found usb_dev->devnum =%d\n",
    usb_dev->devnum));
 /*
 USB_DBG_INFO(iprintf("[USB INFO] USB hub found \n"));
 */
    
 USB_DBG_MEM(iprintf("[USB MEM] hub_probe() : usb_dev=%08XH \n", usb_dev )) ;
    
 USB_DBG_MEM(iprintf("[USB MEM] hub_probe() : hubdev (usb_dev)=%08XH \n", hubdev (usb_dev) ));

 USB_DBG_MEM(iprintf("[USB MEM] hub_probe() : hubdev (usb_dev)->driver=%08XH \n", hubdev (usb_dev)->driver )) ;
 /*
 #ifdef USB_BS_ASSERT
    if  ( !(   (unsigned long) (hubdev (dev)->driver)
             > (unsigned long) 0x00001000L
        )  )
        {
         iprintf("[ASSERT][USB] hub_probe() : " ) ;
         iprintf("Error : value hubdev (dev)->driver !!\n" ) ;
         iprintf("Error : hubdev (dev)->driver=%08XH \n",
         		hubdev (dev)->driver
         		) ;
         iprintf("[ASSERT][USB] Continue ....\n") ;
        }
 #endif
 */
 #ifdef USB_DBGSTRING_ON
	USB_DBG_MEM(iprintf(
		"[USB MEM] hub_probe() : "
    	"&(hubdev (usb_dev)->driver->name)=%08XH \n",
    	&(hubdev (usb_dev)->driver->name)
    	));
 #endif // USB_DBGSTRING_ON
 /*
 USB_DBG_MEM(iprintf(
	"[USB MEM] &(hubdev (dev)->bus_id)=%08XH \n",
	&(hubdev (dev)->bus_id)
    )) ;
 USB_DBG_MEM(iprintf(
	"[USB MEM] hubdev (dev)->bus_id=%08XH \n",
	hubdev (dev)->bus_id
    )) ;
 */
 /*
 // #ifdef _ETCPIP
    // B. S. Liaw : bypass for root hub
    // if  ( hubdev (dev)->driver)
	if (  hub->urb->dev == hcd->self.root_hub )
 // #endif // ndef _ETCPIP
	   {
		dev_info (hubdev (dev), "USB hub found\n");
		USB_DBG_INFO(iprintf(
        	"[USB INFO] USB Root Hub found \n"
            ));
		dprintf(
        	"[USB] USB Root Hub found \n"
            );
	   }
      else
       {
		USB_DBG_INFO(iprintf(
        	"[USB INFO] USB Stand-alone Hub found \n"
            ));
		dprintf(
        	"[USB] USB Stand-alone Hub found \n"
            );
       }
 */
 if (usb_dev->devnum == 1) {
	USB_DBG_INFO(iprintf("[USB INFO] Root HUB found \n")) ;
	usb_device_count -- ; // >= 1 : enable ISR
    USB_DBG_INFO(iprintf("[USB INFO] hub_probe() : usb_device_count = %d \n",usb_device_count)) ;
    
    if(usb_device_count<=0) {
		#ifdef USB_DMA_ON_DEMAND
			UsbDisableDma() ;
		 #else
			#if defined(_ADM5120)
		    	// #error : ADMtek 5120  Need Define USB_DMA_ON_DEMAND !!
		    #endif
		#endif // USB_DMA_ON_DEMAND
		#if defined(USB_INT_SCHEDULING) || defined(USB_INT_ON_DEMAND)
			UsbDisableInt() ;
		 #else
			#if defined(_S3C2510)
		    	#error : SS2510 Need Define USB_INT_ON_DEMAND !!
	#elif defined(_AR7300)
    	#error : _AR7300 Need Define USB_INT_ON_DEMAND !!
			#elif defined(_ATHEROS)
				#error : _ATHEROS Need Define USB_INT_ON_DEMAND !!
			#endif
		#endif // USB_INT_ON_DEMAND
		USB_DBG_INIT(iprintf("[USB INIT] hub_probe() : reset usb_device_count to 0 \n")) ;
		usb_device_count = 0 ; // >= 1 : enable ISR
      }
   }
  else
   {
	USB_DBG_INFO(iprintf("[USB INFO] External HUB found \n" ));
	iprintf("[USB] Found External HUB , V/P = %04XH/%04XH \n" ,usb_dev->descriptor.idVendor,
					usb_dev->descriptor.idProduct);
	#ifdef _USB_SYSLOG
	    //sprintf(usbsyslog_buf,"[USB] Found External HUB , V/P = %04XH/%04XH " ,usb_dev->descriptor.idVendor,
		//			usb_dev->descriptor.idProduct);
					#if (_SYSLOG==2)
		if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef UBS_SYS_LOG_HUB_FOUND
			//sprintf_SetSystemLog("%s , V/P = %04XH/%04XH ", UBS_SYS_LOG_HUB_FOUND, usb_dev->descriptor.idVendor, usb_dev->descriptor.idProduct);
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, "%s , V/P = %04XH/%04XH ", UBS_SYS_LOG_HUB_FOUND, usb_dev->descriptor.idVendor, usb_dev->descriptor.idProduct);
#else
			//sprintf_SetSystemLog("[USB] Found External HUB , V/P = %04XH/%04XH " ,usb_dev->descriptor.idVendor, usb_dev->descriptor.idProduct);
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, "[USB] Found External HUB , V/P = %04XH/%04XH " ,usb_dev->descriptor.idVendor, usb_dev->descriptor.idProduct);
#endif // UBS_SYS_LOG_HUB_FOUND
		}
	#endif // _SYSLOG

	sprintf(usb_info_str, "External HUB , V/P = %04XH/%04XH " ,usb_dev->descriptor.idVendor,
					usb_dev->descriptor.idProduct);
                    
	// usb_device_count ++ ; // >= 1 : enable ISR
    
	iprintf("[USB] Total External Device = %d \n",usb_device_count) ;
 #ifdef _USB_SYSLOG
    //sprintf(usbsyslog_buf,"[USB] Total External Device = %d ",usb_device_count);
				#if (_SYSLOG==2)
	if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_DEV_NUM
		//sprintf_SetSystemLog(USB_SYS_LOG_DEV_NUM, usb_device_count);
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, USB_SYS_LOG_DEV_NUM, usb_device_count);
#else
		//sprintf_SetSystemLog("[USB] Total External Device = %d ",usb_device_count);
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, "[USB] Total External Device = %d ",usb_device_count);
#endif // USB_SYS_LOG_DEV_NUM
	}
 #endif // _SYSLOG
    
    /*
	usb_hub_plug_out = 0 ;
    */
   }
 
 #ifndef _ETCPIP
    /*
	// hub = kmalloc(sizeof(*hub), GFP_KERNEL);
	
    // 20021012 remarked by drsohn
	hub = kmalloc(sizeof(*hub), GFP_KERNEL|GFP_DMA);
    		// 20021012 modified by drsohn
	if (!hub) {
		err("couldn't kmalloc hub struct");
		return -ENOMEM;
	}
    */
  #else
	#ifndef USB_MEM_LIB
	hub = (struct usb_hub *)
	    NonCache_contigmalloc(sizeof(*hub)) ;
     #else
	   	// alloc_mem(sizeof(*hub)) ;
        
		USB_DBG_MEM_C(iprintf(
		    "[USB MEM_U] hub_probe() Call UsbAllocMemU ..\n"
            ));
            
		hub = (struct usb_hub *)
        	UsbAllocMemU(USB_MEM_U_S
						#ifdef USB_DBG_MEM_U_ON
            			, sizeof(struct usb_hub)
						#endif 
                     );
    #endif
    USB_DBG_MEM(iprintf(
       "[USB MEM] hub_probe() : hub = %08XH \n",hub));
    // NonCache_contigmalloc() cannot be free now !!
    
    if ( hub == NULL )
       {
		dprintf("[USB] Error : out of memory hub struct !!\n");
	usb_hub_init_end = 1 ;
		return -ENOMEM;
	   }
 #endif
 
 #ifndef USB_MEM_LIB
	// B. S. Liaw : memset is important !!
	memset(hub, 0, sizeof(*hub));
  #else
	// alloc_mem() has build-in clear ;
 #endif

 USB_DBG_EVT(iprintf(
    "[USB EVT] hub_probe() : hub->urb = %08XH hub->error = %d\n",
    hub->urb, hub->error)) ;
    
 for( x=0 ; x<USB_HUB_MAX_PORTS;x++) {
	 hub->port_enable[x] = 1 ; // for software enable/disable
   }

 hub->event_empty = 0 ;
    
 #ifdef USB_BS_ASSERT
    if  ( !( (hub->error >= 0 ) && (hub->error < 100 )
        )  )
        {
         iprintf("[ASSERT][USB] hub_probe() : " ) ;
         iprintf("Error : hub->error = %d value out of range !!\n",
                 hub->error ) ;
         iprintf("[ASSERT][USB] Continue ....\n") ;
        }
 #endif

 INIT_LIST_HEAD(&hub->event_list);
 
 hub->event_empty = 1 ;
 
 hub->intf = intf;
 
 #ifndef NO_SEMPHORE
	init_MUTEX(&hub->khubd_sem);
 #endif // ndef NO_SEMPHORE

 /* Record the new hub's existence */
 #ifndef NO_SPINLOCK
 	/*
	spin_lock_irqsave(&hub_event_lock, flags);
    */
 #endif // ndef NO_SPINLOCK
 
 BLOCK_PREEMPTION;
 INIT_LIST_HEAD(&hub->hub_list);
 list_add(&hub->hub_list, &hub_list);
 UNBLOCK_PREEMPTION;
 
 
 #ifndef NO_SPINLOCK
 	/*
	spin_unlock_irqrestore(&hub_event_lock, flags);
    */
 #endif // ndef NO_SPINLOCK

 #ifdef NO_SYS_FS
    
	hubdev(usb_dev)->driver = &hub_dev_driver ;
 #ifdef USB_DBGSTRING_ON
	// hubdev(dev)->driver->name = (char *) &hub_dev_driver_name ;
	hubdev(usb_dev)->driver->name = hub_dev_driver_name ;
 #endif // USB_DBGSTRING_ON
    
    /*
	intf->dev->driver = &hub_dev_driver ;
	intf->dev->driver->name = &hub_dev_driver_name ;
    */
 #endif // NO_SYS_FS
 
 USB_DBG_HUB(iprintf("[USB HUB] hub_probe() : intf=%08XH , hub=%08XH \n",intf, hub )) ;
 usb_set_intfdata(intf, hub);
 
 #ifdef RH_INT_BY_POLL
    hub->hub_in_submit = 0  ;
	USB_DBG_HUB(iprintf(
    	"[USB ROOTHUB] hub_probe() : Set Value : hub[%08XH]->hub_in_submit=%d \n",
	    hub,hub->hub_in_submit
	    )) ;
    /*
	global_hub = hub ;
    */
    /*
    if(global_hub_count >= USB_HUB_MAX) 
      {
       dprintf("[USB] Error : Too Many HUBs !!\n") ;
      }
	global_hub[global_hub_count] = hub ;
	USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] usb_hub_disconnect() : global_hub[%d]=%08XH \n", global_hub_count,global_hub[global_hub_count] )) ;
    global_hub_count ++ ;
    */
    /*
	if (list_empty(&global_hub_list_head))
    
	   {
        
	    USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] add global_hub_list ..\n"));
        
		list_add(&global_hub_list, &global_hub_list_head);
	   }
    
      else
       {
        
	    USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] global_hub_list_head append ..\n"));
		//  append to queue 
		// list_add_tail (&td->td_list, &td->ed->td_list);
        // list_add_tail (element, queue)
		// list_add_tail(&global_hub_list, &global_hub_list_head);
		list_add_tail(&hub->hub_list, &global_hub_list_head);
       }
    */
 #endif
 
 /*
 if (usb_dev->devnum == 1)
   {
    hub->is_ext_hub = 0 ;
	// USB_DBG_INFO(iprintf("[USB INFO] Root HUB found \n")) ;
   }
  else
   {
    hub->is_ext_hub = 1 ;
	// USB_DBG_INFO(iprintf("[USB INFO] External HUB found \n" ));
   }
 */
 
 if (usb_hub_configure(hub, endpoint) >= 0)
   {
 #ifdef USB_DBGSTRING_ON
	USB_DBG_MEM(iprintf(
		"[USB MEM] hub_probe() : "
    	"&(intf->dev.name)=%08XH , intf->dev.name=%08XH\n",
    	&(intf->dev.name) , intf->dev.name
    	));
 #endif // USB_DBGSTRING_ON
    /*     
	#ifndef NO_SYS_FS
    */
 #ifdef USB_DBGSTRING_ON
		strcpy (intf->dev.name, "Hub");
 #endif // USB_DBGSTRING_ON
    /*
      #else
      
		intf->dev.name = &hub_dev_driver_name ;
	#endif
    */
    /*
	#ifdef _ETCPIP
	for (;;) 
	   {
		// USB_DBG_INIT(iprintf("[USB INIT] get USB Resource ..\n"));
	    if (getres(MT_RES_USB, 1000) != 0)
	       {
	        iprintf("[USB] usb_hub_configure() : getres Fail !!") ;
	        continue ;
	       }
       
	    if (hub_in_submit == 1)
	       {
	        continue ;
	       }
		hub_in_submit = 1 ;
		USB_DBG_INIT(iprintf("[USB INIT] usb_submit_urb() .. \n")) ;
	    if (usb_submit_urb(hub->urb))
		   {
			iprintf("[USB] usb_submit_urb failed !!\n");
			// message = "couldn't submit status urb";
			// goto fail;
			hub_in_submit = 0 ;
		   }
	   } // infinite for 
	#endif
    */
    

	usb_hub_init_end = 1 ;
    USB_DBG_FLOW(iprintf("[USB FLOW] hub_probe() End.\n"));
	return 0;
   }
   
 dprintf("[USB] Error : hub configuration failed for device #%d \n",
         usb_dev->devnum);
 
	// free hub, but first clean up its list. 
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(&hub_event_lock, flags);
 #endif // ndef NO_SPINLOCK
 
 hub->event_empty = 0 ;

 BLOCK_PREEMPTION;
 // Delete it and then reset it 
 list_del(&hub->event_list);
 INIT_LIST_HEAD(&hub->event_list);
 list_del(&hub->hub_list);
 INIT_LIST_HEAD(&hub->hub_list);
 UNBLOCK_PREEMPTION;

 #ifndef NO_SPINLOCK
 	/*
	spin_unlock_irqrestore(&hub_event_lock, flags);
    */
 #endif // ndef NO_SPINLOCK

 /*
 #ifndef _ETCPIP
	kfree(hub);
  #else
 */
    UsbFreeMemU(USB_MEM_U_S, (unsigned char *) hub
			#ifdef USB_DBG_MEM_U_ON
        		, sizeof(struct usb_hub)
    	    #endif
            ) ; 
 /*
 #endif
 */
 
 hub_disconnect (intf);
 usb_hub_init_end = 1 ;
 
 USB_DBG_FLOW(iprintf("[USB FLOW] hub_probe() End with Fail !!\n"));
 return -ENODEV;
} // function hub_probe()

static int
hub_ioctl(struct usb_interface *intf, unsigned int code, void *user_data)
{
	struct usb_device *hub = interface_to_usbdev (intf);

	/* assert ifno == 0 (part of hub spec) */
	switch (code) {
#ifdef CONFIG_USB_DEVICEFS
	case USBDEVFS_HUB_PORTINFO: {
		struct usbdevfs_hub_portinfo *info = user_data;
		unsigned long flags;
		int i;

#ifndef NO_SPINLOCK
		spin_lock_irqsave(&hub_event_lock, flags);
#endif // ndef NO_SPINLOCK
		if (hub->devnum <= 0)
			info->nports = 0;
		else {
			info->nports = hub->maxchild;
			for (i = 0; i < info->nports; i++) {
				if (hub->children[i] == NULL)
					info->port[i] = 0;
				else
					info->port[i] =
						hub->children[i]->devnum;
			}
		}
#ifndef NO_SPINLOCK
		spin_unlock_irqrestore(&hub_event_lock, flags);
#endif // ndef NO_SPINLOCK

		return info->nports + 1;
		}
#endif // CONFIG_USB_DEVICEFS

	default:
		return -ENOSYS;
	}
}

static int usb_hub_reset(struct usb_hub *hub)
{
 struct usb_device *dev = interface_to_usbdev(hub->intf);
 int i;

 /* Disconnect any attached devices */
 for (i = 0; i < hub->descriptor->bNbrPorts; i++) {
		if (dev->children[i])
			usb_disconnect(&dev->children[i]);
	}

 /* Attempt to reset the hub */
 if(hub->urb) {
	USB_DBG_INIT(iprintf("[USB INIT] usb_hub_reset() : usb_unlink_urb(hub->urb) ..\n"));
	usb_unlink_urb(hub->urb);
   }
 else
   {
	return -1;
   }

 USB_DBG_HUB(iprintf("[USB HUB] Reset the device .. \n" )) ;
	if (usb_reset_device(dev))
		return -1;

 hub->urb->dev = dev;
 /*
 #ifdef _ETCPIP
	for (;;) 
	   {
		// USB_DBG_INIT(iprintf("[USB INIT] get USB Resource ..\n"));
	    if (getres(MT_RES_USB, 1000) != 0)
	       {
	        iprintf("[USB] usb_hub_reset() : getres Fail !!") ;
	        continue ;
	       }
       
	    if (hub_in_submit == 0)
	       {
	        break ;
	       }
	   }
   hub_in_submit = 1 ;
 #endif
 */
 
 #ifndef RH_INT_BY_POLL
 	/*
	#ifndef _ETCPIP
		if (usb_submit_urb(hub->urb, GFP_KERNEL))
	#else
		if (usb_submit_urb(hub->urb))
	#endif
    	   {
			return -1;
           }
    */
 #endif
 
 usb_hub_power_on(hub);

 return 0;
}

static void usb_hub_disconnect(struct usb_device *dev)
{
 struct usb_device *parent = dev->parent;
 int i;
 
 
 USB_DBG_INIT(iprintf("[USB INIT] usb_hub_disconnect() Begin .. \n" )) ;
 
 usb_hub_init_end = 0 ;
 
 /* Find the device pointer to disconnect */
 if (parent)
   {
	USB_DBG_HUB(iprintf("[USB HUB] usb_hub_disconnect() : parent->maxchild=%d \n", parent->maxchild )) ;
	for (i = 0; i < parent->maxchild; i++)
       {
		if (parent->children[i] == dev)
           {
			USB_DBG_INIT(iprintf("[USB INIT] usb_hub_disconnect() : usb_disconnect ..\n" )) ;
			usb_disconnect(&parent->children[i]);
			// usb_device_count -- ; // >= 1 : enable ISR
			USB_DBG_INIT(iprintf("[USB INIT] usb_hub_disconnect() End. \n" )) ;
			usb_hub_init_end = 1 ;
			return;
		   }
	   }
   }

 dprintf("[USB] Error : cannot disconnect hub %d \n",
         dev->devnum);
 #ifdef USB_DBGSTRING_ON
	err("cannot disconnect hub %s", dev->devpath);
 #else
	dprintf("[USB HUB] cannot disconnect hub ");
 #endif
 usb_hub_init_end = 1 ;
}
		/*
		 struct usb_port_status portsts;
		 struct usb_port_status *portsts_mod;					// 20021015 added by drsohn
		 unsigned short portstatus, portchange;

         USB_DBG_EVT(iprintf("[USB EVT] ==== 2 ==== port %d \n",i+1));
         
         #ifndef _ETCPIP
			portsts_mod = (struct usb_port_status *)
                          kmalloc(sizeof(struct usb_port_status),
                          GFP_KERNEL|GFP_DMA);
          #else
            
            //if (i > MAX_USB_PORT)
            //   {
            //    dprintf("[USB] i too Big !!\n") ;
            //   }
        	//portsts_mod = usb_port_status_array[i] ;
            
        	portsts_mod = (struct usb_port_status *)
                   NonCache_contigmalloc(sizeof(struct usb_port_status)) ;
            // NonCache_contigmalloc() cannot be free now !!
            USB_DBG_MEM(iprintf(
               "[USB MEM] portsts_mod = %08XH \n",portsts_mod));
            if ( portsts_mod == NULL )
               {
        		dprintf("out of memory\n");
        		return ;
        	   }
         #endif
       
         USB_DBG_FLOW(iprintf(
            "[USB FLOW] usb_hub_events() 1 "
            "call usb_get_port_status() ..\n"));
                    
       	 //ret = usb_get_port_status(dev, i + 1, &portsts);			// 20021015 remarked by drsohn
		 ret = usb_get_port_status(dev, i + 1, portsts_mod);			// 20021015 modified by drsohn
		 memcpy(&portsts,portsts_mod,sizeof(struct usb_port_status));		// 20021015 added by drsohn

    	 #ifndef _ETCPIP
			kfree(portsts_mod);				// 20021015 added by drsohn
     	  #else
       		// NonCache_contigmalloc() cannot be free now !!
    	 #endif
		 */	

static int usb_hub_port_status(struct usb_device *hub, int port,
			       u16 *status, u16 *change)
{
 struct usb_port_status *portsts;
 int ret = -ENOMEM;
 
 USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_status() Begin ..\n"));
 /*
 #ifndef _ETCPIP
	portsts = kmalloc(sizeof(*portsts), GFP_NOIO);
  #else
 */
    // NonCache_contigmalloc() cannot be free now !!
	// portsts = (struct usb_port_status *)
    //               NonCache_contigmalloc(sizeof(*portsts)) ;

	portsts = (struct usb_port_status *) // 4
        /*
		#ifdef USB_DBG_MEM_C_ON
			UsbAllocMemC( USB_MEM_C_S,
            			   sizeof(struct usb_port_status)
                         ) ;
         #else
            UsbAllocMemC(USB_MEM_C_S) ;
        #endif
        */
        	UsbAllocMemU(USB_MEM_U_S
				#ifdef USB_DBG_MEM_U_ON
                 , sizeof(struct usb_port_status)
                #endif
                 );

    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s
    
    USB_DBG_MEM(iprintf(
       "[USB MEM] usb_hub_port_status() : portsts = %08XH \n",
       portsts
       ));
    
    if ( portsts == NULL )
       {
		dprintf("[USB] Error : out of memory !! \n");
		return ;
	   }
 /*   
 #endif
 */
	if (portsts) {
		ret = usb_get_port_status(hub, port + 1, portsts);
		if (ret < 0)
           {
			dev_err (hubdev (hub),
				"%s failed (err = %d)\n", __FUNCTION__,
				ret);
           }
		else
           {
            
			*status = le16_to_cpu(portsts->wPortStatus);
			*change = le16_to_cpu(portsts->wPortChange); 
             
			USB_DBG_HUB(iprintf(
				"[USB ROOTHUB] usb_hub_port_status() : status=%08XH "
            	"change=%08XH \n" , status , change
                ));
			USB_DBG_HUB(iprintf(
				"[USB ROOTHUB] usb_hub_port_status() : *status=%04XH "
            	"*change=%04XH \n" , *status , *change
                ));
			ret = 0;
		  }
          
		#ifndef _ETCPIP
            /*
			kfree(portsts);
            */
		#else
            /*
			#ifdef USB_DBG_MEM_C_ON
				UsbFreeMemC(USB_MEM_C_S,(unsigned char *)
                			portsts,
                			 sizeof(struct usb_port_status));
			 #else
				UsbFreeMemC(USB_MEM_C_S,(unsigned char *)
                							portsts);
			#endif
            */
			UsbFreeMemU(USB_MEM_U_S, (unsigned char *)portsts
			    #if defined(USB_DBG_MEM_U_ON)
        	    , sizeof(struct usb_port_status)
				#endif
		        ) ;
		#endif
	   }
       
 USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_status() End.\n"));
 return ret;
}

#define HUB_RESET_TRIES		5
#define HUB_PROBE_TRIES		2
/*
#ifndef _ETCPIP
	#define HUB_SHORT_RESET_TIME	10
 #else
	#define HUB_SHORT_RESET_TIME	30
#endif
*/
#ifndef _ETCPIP
	#define HUB_LONG_RESET_TIME	200
 #else
	#define HUB_LONG_RESET_TIME	500
#endif

#ifdef USB_SOTA   // Sota USB 1.1 Host
	#define HUB_RESET_TIMEOUT	3500
 #else
	#define HUB_RESET_TIMEOUT	500
#endif
/* return: -1 on error, 0 on success, 1 on disconnect.  */
static int usb_hub_port_wait_reset(struct usb_device *hub, int port,
				struct usb_device *dev, unsigned int delay)
{
 int delay_time, ret;
 // struct usb_port_status portsts;
 /*
 struct usb_port_status *portsts_mod;	// 20021015 added by drsohn
 */
 u16 portstatus;
 u16 portchange;
 /*
 #ifndef _ETCPIP
	portsts_mod = (struct usb_port_status *)
                  kmalloc(sizeof(struct usb_port_status),
                  GFP_KERNEL|GFP_DMA);			// 20021015 added by drsohn
  #else
	#ifndef USB_MEM_LIB
		portsts_mod = (struct usb_port_status *)
            NonCache_contigmalloc(sizeof(struct usb_port_status)) ;
		    // NonCache_contigmalloc() cannot be free now !!
     #else
	    // 	alloc_mem(sizeof(struct usb_port_status)) ;
		USB_DBG_MEM_U(iprintf(
		    "[USB MEMPOOL] usb_hub_port_wait_reset() reqUsbMem "
            "portsts_mod ..\n"
            ));
            
		portsts_mod = (struct usb_port_status *)
		#ifdef USB_DBG_MEM_U_ON
        	reqUsbMem(USB_MEM_U_S,
            			sizeof(struct usb_port_status)
                     );
         #else
        	reqUsbMem(USB_MEM_U_S);
        #endif   
    #endif
    USB_DBG_MEM(iprintf(
       "[USB MEM] portsts_mod = %08XH \n",portsts_mod));
    if ( portsts_mod == NULL )
       {
		dprintf("out of memory\n");
		return ;
	   }
 #endif
 */
 	
 for (delay_time = 0;
		delay_time < HUB_RESET_TIMEOUT;
		delay_time += delay) 
   {
	/* wait to give the device a chance to reset */
	wait_ms(delay);

	/* read and decode port status */
	/*
	//ret = usb_get_port_status(hub, port + 1, &portsts);				// 20021015 remarked by drsohn
	ret = usb_get_port_status(hub, port + 1, portsts_mod);					// 20021015 modified by drsohn
	memcpy(&portsts,portsts_mod,sizeof(struct usb_port_status));			// 20021015 added by drsohn
	*/
	USB_DBG_HUB(iprintf(
		"[USB ROOTHUB] usb_hub_port_wait_reset() : &status=%08XH "
    	"&change=%08XH \n" , &portstatus , &portchange
        ));
	ret = usb_hub_port_status(hub, port, &portstatus, &portchange);
	if (ret < 0)
       {
		err("get_port_status(%d) failed (err = %d)", port + 1, ret);
		// return -1;
		goto end_error ;
	   }
	USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_wait_reset() : status=%04XH "
    	"change=%04XH \n" , portstatus , portchange
        ));
    /*
		portstatus = le16_to_cpu(portsts.wPortStatus);
		portchange = le16_to_cpu(portsts.wPortChange);
        
		USB_DBG_INIT(iprintf(
		 	"[USB INIT]  port %d, portstatus %x, change %x, %s \n",
            port + 1,
			portstatus, portchange, portspeed (portstatus)
            ));
     */

	/* Device went away? */
	if (!(portstatus & USB_PORT_STAT_CONNECTION))
       {
		USB_DBG_HUB(iprintf(
			"[USB ROOTHUB] usb_hub_port_wait_reset() : Device went away !!\n"
            ));
		return 1;
       }
       
	/* bomb out completely if something weird happened */
    
    
	/* if we`ve finished resetting, then break out of the loop */
	if (     !(portstatus & USB_PORT_STAT_RESET)
		// #ifndef SOTA_HUB_BUGFIX
	       && (portstatus & USB_PORT_STAT_ENABLE)
		// #endif // ndef SOTA_HUB_BUGFIX
       )
        
		#ifdef SOTA_HUB_BUGFIX
	       {
 			// break ;         
            // ----------------
			goto after_reset ;
            // ----------------
	       } 
         #else
        
	       {
			if (portstatus & USB_PORT_STAT_HIGH_SPEED)
    	       {
				dev->speed = USB_SPEED_HIGH;
				USB_DBG_INIT(iprintf("[USB INIT] usb_hub_port_wait_reset() : Set Value : dev->speed=%d , HIGH\n",dev->speed)) ;
	           }
			else if (portstatus & USB_PORT_STAT_LOW_SPEED)
	           {
				dev->speed = USB_SPEED_LOW;
				USB_DBG_INIT(iprintf("[USB INIT] usb_hub_port_wait_reset() : Set Value : dev->speed=%d , LOW\n",dev->speed)) ;
	           }
			else
        	   {
				dev->speed = USB_SPEED_FULL;
				USB_DBG_INIT(iprintf("[USB INIT] usb_hub_port_wait_reset() : Set Value : dev->speed=%d , FULL\n",dev->speed)) ;
        	   }
           
			USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_wait_reset() : OK\n" )) ;
				return 0;
	       } 
        
        #endif
        
	/* switch to the long delay after two short delay failures */
	if (delay_time >= 2 * HUB_SHORT_RESET_TIME)
       {
		delay = HUB_LONG_RESET_TIME;
	   }
       
	USB_DBG_HUB(iprintf( "[USB ROOTHUB] usb_hub_port_wait_reset() : port %d of hub %d not reset yet, waiting"
        " %dms\n",
    	port + 1,
		hub->devnum, delay
        ));
        
	#ifdef USB_DBG_HUB_ON
		dev_dbg (hubdev (hub),
			"port %d not reset yet, waiting %dms\n",
			port + 1, delay);
    #endif
   } // for delay_time
   
 // time out   
 goto end_error ;
   
 
 #ifdef SOTA_HUB_BUGFIX
 
 	// -------------
    after_reset:
    // -------------
    
 	/*
	USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] usb_hub_port_wait_reset(1) : Clear Port Reset ..\n" )) ;
    
	usb_clear_port_feature(hub,	port + 1, USB_PORT_FEAT_C_RESET) ;
 
	// wait to give the device a chance to stable 
	wait_ms(delay);
    */
	USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_wait_reset() : Enable Port ..\n" )) ;
    
    /*
	USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] usb_hub_port_wait_reset() : Clear Port Connection Change ..\n" )) ;
    */
    
	for (delay_time = 0;
		delay_time < HUB_RESET_TIMEOUT;
		delay_time += delay) 
	   {
		ret = usb_hub_port_status(hub, port, &portstatus, &portchange);
		if (ret < 0)
    	   {
			err("get_port_status(%d) failed (err = %d)", port + 1, ret);
			// return -1;
			goto end_error ;
		   }
           
		USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_wait_reset() : status=%08XH "
    		"change=%08XH \n" , portstatus , portchange
		    ));
            
		// Device went away? 
		if (!(portstatus & USB_PORT_STAT_CONNECTION))
	       {
			USB_DBG_HUB(iprintf(
				"[USB ROOTHUB] usb_hub_port_wait_reset() : Device went away !!\n"
    	        ));
			return 1;
    	   }
    
                
		if ((portchange & USB_PORT_STAT_C_CONNECTION))
    	   {
			USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_wait_reset() : USB_PORT_STAT_C_CONNECTION !!\n" ));
			USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_wait_reset() : Call usb_clear_port_feature() .. \n" )) ;
			usb_clear_port_feature(hub, port+1, USB_PORT_FEAT_C_CONNECTION);
            
			#ifdef SOTA_HUB_BUGFIX
            /*
			USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_wait_reset() : so enable the port .. \n" ));

			usb_hub_port_enable(hub, port) ;
            */
			#endif // SOTA_HUB_BUGFIX

            
			ret = usb_hub_port_status(hub, port, &portstatus, &portchange);
			if (ret < 0)
    		   {
				err("get_port_status(%d) failed (err = %d)", port + 1, ret);
				// return -1;
				goto end_error ;
			   }
			USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_wait_reset() : status=%08XH "
    			"change=%08XH \n" , portstatus , portchange
		        ));
    	   }
           
        
		if ((portchange & USB_PORT_STAT_C_ENABLE))
    	   {
			USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_wait_reset() : USB_PORT_STAT_C_ENABLE !!\n" ));
			USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_wait_reset() : Call usb_clear_port_feature() .. \n" )) ;

			usb_clear_port_feature(hub, port+1, USB_PORT_STAT_C_ENABLE) ;
            
            break ;
    	   }
        
	   } // for delay_time
    /*       
	// wait to give the device a chance to stable 
	wait_ms(delay);
    */
    
	USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_wait_reset() : Check Device Speed ..\n" )) ;
    
	ret = usb_hub_port_status(hub, port, &portstatus, &portchange);
	if (ret < 0)
       {
		err("get_port_status(%d) failed (err = %d)", port + 1, ret);
		// return -1;
		goto end_error ;
	   }
	USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_wait_reset() : status=%08XH "
    	"change=%08XH \n" , portstatus , portchange
	    ));
        
	// Device went away? 
	if (!(portstatus & USB_PORT_STAT_CONNECTION))
       {
		USB_DBG_HUB(iprintf(
			"[USB ROOTHUB] usb_hub_port_wait_reset() : Device went away !!\n"
            ));
		return 1;
       }
        
	if (portstatus & USB_PORT_STAT_HIGH_SPEED)
       {
		dev->speed = USB_SPEED_HIGH;
		USB_DBG_INIT(iprintf("[USB INIT] usb_hub_port_wait_reset() : Set Value : dev->speed=%d , HIGH\n",dev->speed)) ;
	   }
	else if (portstatus & USB_PORT_STAT_LOW_SPEED)
	   {
		dev->speed = USB_SPEED_LOW;
		USB_DBG_INIT(iprintf("[USB INIT] usb_hub_port_wait_reset() : Set Value : dev->speed=%d , LOW\n",dev->speed)) ;
	   }
	else
       {
		dev->speed = USB_SPEED_FULL;
		USB_DBG_INIT(iprintf("[USB INIT] usb_hub_port_wait_reset() : Set Value : dev->speed=%d , FULL\n",dev->speed)) ;
       }
    
	USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_wait_reset() : OK\n" )) ;
    
	return 0;
           
 #endif // SOTA_HUB_BUGFIX
 
 // ------------   
 end_error:
 // ------------   
 
 dprintf("[USB] HUB : Reset Port Fail ! \n" ) ;
 
 /*   
 #ifndef _ETCPIP
	kfree (portsts_mod);
  #else
	#ifndef USB_MEM_LIB
		// NonCache_contigmalloc() cannot be free now !!
	 #else
		USB_DBG_MEM_U(iprintf(
		    "[USB MEMPOOL] usb_hub_port_wait_reset() relUsbMem "
            "portsts_mod ..\n"
            ));
        relUsbMem(USB_MEM_U_S, portsts_mod) ;
	#endif
 #endif
 */
 return -1;
}

/* return: -1 on error, 0 on success, 1 on disconnect.  */
// static 
int usb_hub_port_reset(struct usb_device *hub, int port,
				struct usb_device *dev, unsigned int delay)
{
 int i, status;
 /*
 #ifdef _USB_SYSLOG
    char usbsyslog_buf[100];
 #endif // _SYSLOG
 */


 dprintf("[USB] HUB : (USB Device %d) Reset Port %d \n", dev->devnum ,port+1);
		#ifdef _USB_SYSLOG
		    //sprintf(usbsyslog_buf,"[USB] HUB : (USB Device %d) Reset Port %d ",
			//	 dev->devnum ,port+1);
						#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_HUB_PORT_RST
				//sprintf_SetSystemLog(USB_SYS_LOG_HUB_PORT_RST, dev->devnum ,port+1);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, USB_SYS_LOG_HUB_PORT_RST, dev->devnum ,port+1);
#else
				//sprintf_SetSystemLog("[USB] HUB : (USB Device %d) Reset Port %d ", dev->devnum ,port+1);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, "[USB] HUB : (USB Device %d) Reset Port %d ", dev->devnum ,port+1);
#endif // USB_SYS_LOG_HUB_PORT_RST
			}
		#endif // _SYSLOG
        
 /* Reset the port */
 for (i = 0; i < HUB_RESET_TRIES; i++)
   {
       
	usb_set_port_feature(hub, port + 1, USB_PORT_FEAT_RESET);

	/* return on disconnect or reset */
	status = usb_hub_port_wait_reset(hub, port, dev, delay);
	if (status != -1)
       {
		USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_reset() : Port Reset OK.\n" )) ;
		#ifndef SOTA_HUB_BUGFIX
			USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_reset(2) : Clear Port Reset ..\n" )) ;
			usb_clear_port_feature(hub,	port + 1, USB_PORT_FEAT_C_RESET) ;
		#endif // ndef SOTA_HUB_BUGFIX
		return status;
	   }
    /*   
	dprintf("[USB] HUB : Reset Fail : port %d of hub %d not enabled !! \n",
			port + 1, hub->devnum);
    break ;
    */
    
	dprintf("[USB] port %d of hub %d not enabled, trying reset again...\n",
			port + 1, hub->devnum);
	dev_dbg (hubdev (hub),
			"port %d not enabled, trying reset again...\n",
			port + 1);

	delay = HUB_LONG_RESET_TIME;
   }

 dprintf("[USB] Error : Cannot enable port %i of hub %d, disabling port. \n",
		port + 1, hub->devnum);
 dprintf("[USB] Error : Maybe the USB cable is bad? \n");
 dev_err (hubdev (hub),
		"Cannot enable port %i.  Maybe the USB cable is bad?\n",
		port + 1);

 return -1;
}

#ifdef SOTA_HUB_BUGFIX
   /*
   void usb_hub_port_enable(struct usb_device *hub, int port)
   {
	int ret;
    
	dprintf("[USB] HUB : Enable Port %d \n",port+1);
 
	ret = usb_set_port_feature(hub, port + 1, USB_PORT_FEAT_ENABLE);
	if (ret)
	    {
		 dprintf("[USB] HUB : Error : cannot enable port %d of hub %d (err = %d) \n",
			port + 1, hub->devnum, ret);
		 dev_err(hubdev(hub), "cannot enable port %d (err = %d)\n",
			port + 1, ret);
	    }
   }
   */
#endif // SOTA_HUB_BUGFIX

void usb_hub_port_disable(struct usb_device *hub, int port)
{
 int ret;
    
 dprintf("[USB] HUB : Disable Port %d \n",port+1);
 /*
 dev_dbg(hubdev(interface_to_usbdev(hub->intf)),
		"Disable Port %d,port+1\n");
 */
 
 ret = usb_clear_port_feature(hub, port + 1, USB_PORT_FEAT_ENABLE);
 if (ret)
    {
	 dprintf("[USB] Error : cannot disable port %d of hub %d (err = %d) \n",
			port + 1, hub->devnum, ret);
	 dev_err(hubdev(hub), "cannot disable port %d (err = %d)\n",
			port + 1, ret);
    }
}

/* USB 2.0 spec, 7.1.7.3 / fig 7-29:
 *
 * Between connect detection and reset signaling there must be a delay
 * of 100ms at least for debounce and power-settling. The corresponding
 * timer shall restart whenever the downstream port detects a disconnect.
 * 
 * Apparently there are some bluetooth and irda-dongles and a number
 * of low-speed devices which require longer delays of about 200-400ms.
 * Not covered by the spec - but easy to deal with.
 *
 * This implementation uses 400ms minimum debounce timeout and checks
 * every 25ms for transient disconnects to restart the delay.
 */

#define HUB_DEBOUNCE_TIMEOUT	400
#define HUB_DEBOUNCE_STEP	 25
#define HUB_DEBOUNCE_STABLE	  4

/* return: -1 on error, 0 on success, 1 on disconnect.  */
static int usb_hub_port_debounce(struct usb_device *hub, int port)
{
 int ret;
 int delay_time, stable_count;
 u16 portchange, portstatus;
 unsigned connection;


 USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_debounce() Begin .. \n" )) ;

 connection = 0;
 stable_count = 0;

 USB_DBG_HUB(iprintf(
		"[USB ROOTHUB] usb_hub_port_debounce() : HUB_DEBOUNCE_TIMEOUT=%d "
    	"HUB_DEBOUNCE_STEP=%d \n" , HUB_DEBOUNCE_TIMEOUT , HUB_DEBOUNCE_STEP
        ));
        
 for (delay_time = 0; delay_time < HUB_DEBOUNCE_TIMEOUT;
 	   delay_time += HUB_DEBOUNCE_STEP)
   {
	wait_ms(HUB_DEBOUNCE_STEP);

	USB_DBG_HUB(iprintf(
		"[USB ROOTHUB] usb_hub_port_debounce() : &status=%08XH "
    	"&change=%08XH \n" , &portstatus , &portchange
        ));
	ret = usb_hub_port_status(hub, port, &portstatus, &portchange);
    
	if (ret < 0)
       {
		USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_debounce() End with Fail !! \n" )) ;
		return -1;
       }
       
	USB_DBG_HUB(iprintf(
		"[USB ROOTHUB] usb_hub_port_debounce() : status=%08XH "
    	"change=%08XH \n" , portstatus , portchange
        ));

	if ((portstatus & USB_PORT_STAT_CONNECTION) == connection)
       {
		if (connection)
           {
			if (++stable_count == HUB_DEBOUNCE_STABLE)
               {
				USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_debounce() : Stable \n" )) ;
				break;
               }
		   }
	   }
      else
       {
		stable_count = 0;
	   }
       
	USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_debounce() : Check Connection .. \n" )) ;
    
	connection = portstatus & USB_PORT_STAT_CONNECTION;

	if ((portchange & USB_PORT_STAT_C_CONNECTION))
       {
		USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_debounce() Call usb_clear_port_feature() .. \n" )) ;
		usb_clear_port_feature(hub, port+1, USB_PORT_FEAT_C_CONNECTION);
	   }
   } // for

 USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_debounce() OK. \n" )) ;
 /* XXX Replace this with dbg() when 2.6 is about to ship. */
 /*
 dev_info (hubdev (hub),
 */
 USB_DBG_INFO(iprintf(
 	"[USB INFO] debounce: port %d: delay %dms stable %d status 0x%x\n",
	port + 1, delay_time, stable_count, portstatus)
    );

 USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_debounce() End. \n" )) ;
 return ((portstatus&USB_PORT_STAT_CONNECTION)) ? 0 : 1;
}

static void usb_hub_port_connect_change(struct usb_hub *hubstate, int port,
					u16 portstatus, u16 portchange)
{
 struct usb_device *hub_usb_device = interface_to_usbdev(hubstate->intf);
 struct usb_device *dev;
 unsigned int delay = HUB_SHORT_RESET_TIME;
 int i;
 // #ifndef _ETCPIP
	char *portstr, *tempstr;
 //  #else
 //	char portstr[1024], tempstr[1024];
 // #endif
 
 #ifdef USB_OHCI
	// 2003/3/21 B. S. Liaw
	// ohci_t * ohci;
	struct hcdx_hcd ohci;
 #endif
 /*
 #ifdef _USB_SYSLOG
    char usbsyslog_buf[100];
 #endif // _SYSLOG
 */

 USB_DBG_FLOW(iprintf(
    "[USB FLOW] usb_hub_port_connect_change() Begin ..\n"));

 portstr = 0 ;
 tempstr = 0 ;
 
 if (!hubstate) {
	dprintf("[USB] usb_hub_port_connect_change() : Null hubstate !!\n");
	goto done ;
   }
   
 if (!hubstate->intf) {
	dprintf("[USB] usb_hub_port_connect_change() : Null hubstate->intf !!\n");
	goto done ;
   }
   
 if (!hub_usb_device) {
	dprintf("[USB] usb_hub_port_connect_change() : Null hub !!\n");
	goto done ;
   }

 USB_DBG_INFO(iprintf("[USB INFO] usb_hub_port_connect_change: usb_device_count=%d, hubstate=%08XH, hub_usb_device->hub=%08XH, portstatus=%04XH\n",
 			usb_device_count, hubstate, hub_usb_device->hub, portstatus));
 if ((usb_device_count >= 2) && (hub_usb_device->hub))  /* External Hub */
 {
	if (!hub_usb_device->children[port])
	{
		USB_DBG_INFO(iprintf("Too many devices!!!\n"));
 		usb_clear_port_feature(hub_usb_device, port + 1, USB_PORT_FEAT_C_CONNECTION);
 		return;
 	}
 }
    
 #ifdef USB_DBG_ABORT   
	// vvv for debug Prefetch/Data Abort Error !!
   
	#ifndef NO_TIMER
	    dprintf("[TEST] usb_hub_port_connect_change() : jiffies = %08XH\n",jiffies) ;
	#endif // ndef NO_TIMER
   
    {
     unsigned long * __temp_ptr = 0 ;
    
	 #if !defined(_ADM5120)
        dprintf("[TEST] usb_hub_port_connect_change() :  *000000000H = %08XH \n",
		        * __temp_ptr
                );
     #endif
    }
    dprintf("[TEST] usb_hub_port_connect_change() : GetTime() = %08XH\n",
                        GetTime()) ;
	if(!in_interrupt())
       {
		dprintf("[TEST] usb_hub_port_connect_change() : dlytsk()/YIELD() ..\n") ;
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] usb_hub_port_connect_change() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
       }
      else
       {
        dprintf("[TEST] usb_hub_port_connect_change() : in interrupt !!\n");
       }
   // ^^^ for debug Prefetch/Data Abort Error !!  
 #endif // USB_DBG_ABORT
  
 USB_DBG_HUB(iprintf(
	"[USB HUB] usb_hub_port_connect_change() : portstatus = %08XH \n",
	portstatus )) ;
    
 dev_dbg (&hubstate->intf->dev,
	"port %d, status %x, change %x, %s\n",
		port + 1, portstatus, portchange, portspeed(portstatus));
        
 dprintf("[USB] Device Speed = %s (%08XH)\n", portspeed(portstatus), portstatus);
 #ifdef _USB_SYSLOG
    //sprintf(usbsyslog_buf,"[USB] Device Speed = %s ", portspeed(portstatus) );
	#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
	#endif // _SYSLOG==2
				{

		#ifdef USB_SYS_LOG_DEV_SPEED
				//sprintf_SetSystemLog(USB_SYS_LOG_DEV_SPEED, portspeed(portstatus) );
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, USB_SYS_LOG_DEV_SPEED, portspeed(portstatus) );
		#else
				//sprintf_SetSystemLog("[USB] Device Speed = %s ", portspeed(portstatus) );
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, "[USB] Device Speed = %s ", portspeed(portstatus) );
		#endif // USB_SYS_LOG_DEV_SPEED
			}
		#endif // _SYSLOG
 /*
 portstatus = le16_to_cpu(portsts->wPortStatus);
 portchange = le16_to_cpu(portsts->wPortChange);
 
 USB_DBG_INIT(iprintf(
 	"[USB INIT] port %d, portstatus %x, change %x, %s \n",
		port + 1, portstatus, portchange, portspeed (portstatus)
     )) ;
 */
 
 /* Clear the connection change status */
 /*
 usb_clear_port_feature(hub, port + 1, USB_PORT_FEAT_C_CONNECTION);
 */
 /* Disconnect any existing devices under this port */
 if (hub_usb_device->children[port]) {
	USB_DBG_INIT(iprintf(
	"[USB INIT] usb_hub_port_connect_change() : usb_disconnect()\n"));
    /*
	dprintf(
        "[USB] usb_hub_port_connect_change() : usb_disconnect()\n");
    */    
	usb_disconnect(&hub_usb_device->children[port]);
   }

 /* Return now if nothing is connected */
 if(!(portstatus & USB_PORT_STAT_CONNECTION)) {
    USB_DBG_INIT(iprintf("[USB INIT] usb_hub_port_connect_change() : nothing is connected\n"));
        
	iprintf("[USB] Device Removed !! \n");
		#ifdef _USB_SYSLOG
		    //sprintf(usbsyslog_buf,"[USB] Device Removed !! ");
						#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

			#ifdef USB_SYS_LOG_DEV_REMOVE
				//sprintf_SetSystemLog(USB_SYS_LOG_DEV_REMOVE);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, USB_SYS_LOG_DEV_REMOVE);
			#else
				//sprintf_SetSystemLog("[USB] Device Removed !! ");
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, "[USB] Device Removed !! ");
			#endif // USB_SYS_LOG_DEV_REMOVE
			}
		#endif // _SYSLOG

	sprintf(usb_info_str, "Device Removed !! ");
    
	USB_DBG_HUB(iprintf(
		"[USB HUB] usb_hub_port_connect_change() : Re-Enasble HUB Port %d .. \n",
		port+1)) ;
	dprintf("[USB] Re-Enasble HUB Port %d .. \n",
		port+1) ;
	hubstate->port_enable[port] = 1 ; // for software enable/disable

	// usb_device_count -- ; // >= 1 : enable ISR
    
    USB_DBG_INFO(iprintf(
		"[USB INFO] usb_hub_port_connect_change() : usb_device_count = %d \n",
		usb_device_count)) ;
    
    if(usb_device_count<=0) {
		#ifdef USB_DMA_ON_DEMAND
			UsbDisableDma() ;
		#endif // USB_DMA_ON_DEMAND
		#if defined(USB_INT_SCHEDULING) || defined(USB_INT_ON_DEMAND)
			UsbDisableInt() ;
		#endif // USB_INT_ON_DEMAND
		USB_DBG_INIT(iprintf("[USB INIT] usb_hub_port_connect_change() : reset usb_device_count to 0 \n")) ;
		usb_device_count = 0 ; // >= 1 : enable ISR
      }
      
    iprintf("[USB] Total External Device = %d \n",usb_device_count) ;
	#ifdef _USB_SYSLOG
	    //sprintf(usbsyslog_buf,"[USB] Total External Device = %d ",usb_device_count);
		#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
		#endif // _SYSLOG==2
				{

		#ifdef USB_SYS_LOG_DEV_NUM
			//sprintf_SetSystemLog(USB_SYS_LOG_DEV_NUM, usb_device_count);
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, USB_SYS_LOG_DEV_NUM, usb_device_count);
		#else
			//sprintf_SetSystemLog("[USB] Total External Device = %d ",usb_device_count);
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, "[USB] Total External Device = %d ",usb_device_count);
		#endif // USB_SYS_LOG_DEV_NUM
		}
	#endif // _SYSLOG
      
	if (portstatus & USB_PORT_STAT_ENABLE) {
			usb_hub_port_disable(hub_usb_device, port);
        }
       
	/* Clear the connection change status */
	
	usb_clear_port_feature(hub_usb_device, port + 1, USB_PORT_FEAT_C_CONNECTION);

	return;
   }

 if( ! hubstate->port_enable[port] ) { // for software enable/disable
	USB_DBG_HUB(iprintf(
		"[USB HUB] usb_hub_port_connect_change() : Diasbled HUB Port %d !!\n",
		port+1)) ;
	dprintf("[USB] Diasbled HUB Port %d !!\n",
		port+1) ;
	// continue ;
	goto done ;
   }

 USB_DBG_INIT(iprintf(
	"[USB INIT] usb_hub_port_connect_change() : Someone Connect on port %d\n",
	port+1
	));
 
 sprintf(usb_info_str, "Check Device Type .. ");

 iprintf("[USB] Check Device Type .. \n" ) ;

 #ifdef _USB_SYSLOG
	//sprintf(usbsyslog_buf,"[USB] Check Device Type .. ");
	#if (_SYSLOG==2)
		if (gSetting.syslog_ds.syslog_enable==1) 
	#endif // _SYSLOG==2
			{

		#ifdef USB_SYS_LOG_DEV_TYPE_CHECK
			//sprintf_SetSystemLog(USB_SYS_LOG_DEV_TYPE_CHECK);
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, USB_SYS_LOG_DEV_TYPE_CHECK);
		 #else
			//sprintf_SetSystemLog("[USB] Check Device Type .. ");
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, "[USB] Check Device Type .. ");
		#endif // USB_SYS_LOG_DEV_TYPE_CHECK
	   }
 #endif // _SYSLOG
 
 usb_device_count ++ ; // >= 1 : enable ISR

 USB_DBG_INFO(iprintf(
	"[USB INFO] usb_hub_port_connect_change() : usb_device_count = %d \n",
	usb_device_count)) ;

 if (usb_device_count <= 0) {
	usb_device_count = 1 ; // >= 1 : enable ISR
   }
 #ifdef USB_DMA_ON_DEMAND
	UsbEnableDma() ;
 #endif // USB_DMA_ON_DEMAND
 #ifdef USB_INT_ON_DEMAND
	UsbEnableInt() ;
  #else
	#if defined(_S3C2510)
    	#error : SS2510 Need Define USB_INT_ON_DEMAND !!
	#elif defined(_AR7300)
    	#error : _AR7300 Need Define USB_INT_ON_DEMAND !!
	#elif defined(_ATHEROS)
    	#error : _ATHEROS Need Define USB_INT_ON_DEMAND !!
    #endif
 #endif // USB_INT_ON_DEMAND
 
 USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_connect_change() : Debounce ..\n" ));
 if (usb_hub_port_debounce(hub_usb_device, port)) {
    /*
	dev_err (&hubstate->intf->dev,
    */
    dprintf("[USB] Connect Debounce failed, port %d disabled\n",
			port+1);
	usb_hub_port_disable(hub_usb_device, port);
	return;
   }
    
 USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_connect_change() : Set delay ..\n" ));
 /* Some low speed devices have problems with the quick delay, so */
 /*  be a bit pessimistic with those devices. RHbug #23670 */
 if (portstatus & USB_PORT_STAT_LOW_SPEED) {
	USB_DBG_INIT(iprintf("[USB INIT] usb_hub_port_connect_change() : Low Speed , wait 1 Secend \n" ));
	/*
	#ifndef _ETCPIP
		wait_ms(400);
     #else
	*/
		// wait_ms(1000);
		wait_sec(1) ;
	/*
    #endif
    */
	delay = HUB_LONG_RESET_TIME;
   }

 USB_DBG_HUB(iprintf("[USB HUB] usb_hub_port_connect_change() : Alloc ..\n" ));
 /*
 #ifndef NO_SEMPHORE
	down(&usb_address0_sem);
 #endif // ndef NO_SEMPHORE
 */
 /*
 #ifndef _ETCPIP
	tempstr = kmalloc(1024, GFP_KERNEL);
  #else
 */
 tempstr = (char *)
		UsbAllocMemU(USB_MEM_U_L
			#ifdef USB_DBG_MEM_U_ON
            		, 1024
		    #endif
			) ;

 USB_DBG_INIT(iprintf("[USB INIT] tempstr = %08XH \n",tempstr ));

 if ( tempstr == NULL ) {
		dprintf("[USB] Error : out of memory !!\n");
		return ;
	   }
 /*
 #endif
 */

 /*
 #ifndef _ETCPIP
	portstr = kmalloc(1024, GFP_KERNEL);
  #else
 */
	portstr = (char *)
		UsbAllocMemU(USB_MEM_U_L
			#ifdef USB_DBG_MEM_U_ON
            		, 1024
		    #endif
			) ;
    USB_DBG_INIT(iprintf("[USB INIT] portstr = %08XH \n",portstr ));

    if ( portstr == NULL ) {
		dprintf("[USB] Error : out of memory !!\n");
		return ;
	   }
 /*
 #endif
 */
 if (!hub_usb_device->bus) {
	dprintf("[USB] Error : Null hub->bus !!\n");
	goto done ;
   }

 for (i = 0; i < HUB_PROBE_TRIES; i++) {
	struct usb_device *pdev;
	int	len;

    // USB_DBG_INIT(iprintf("[USB INIT] usb_new_device() ==0== Probe Try : %d \n",i ));
    USB_DBG_INIT(iprintf("[USB INIT] ==== New Device ====  ==1== HUB Probe Try : %d \n", i )) ;
    
	/* Allocate a new device struct */
    BLOCK_PREEMPTION;
	dev = usb_alloc_dev(hub_usb_device, hub_usb_device->bus);
    UNBLOCK_PREEMPTION;
    
    USB_DBG_MEM(iprintf(
       "[USB MEM] new dev = %08XH , device number = %d\n",
       dev,dev->devnum
       ));
    USB_DBG_INIT(iprintf(
       "[USB INIT] new device , device number = %d \n",
       dev->devnum
       ));
       
	if (!dev) {
        /*
		err("couldn't allocate usb_device");
		dev_err (&hubstate->intf->dev,
				"couldn't allocate usb_device\n");
        */
		dprintf("[USB] Error : couldn't allocate usb_device !!\n");
		break;
	   }
       
    USB_DBG_PIPE(iprintf(
           "[USB PIPE] ?? usb_snddefctrl(dev) = %08XH \n",
           usb_snddefctrl(dev)
           )) ;
    USB_DBG_PIPE(iprintf(
           "[USB PIPE] ?? usb_pipedevice(usb_snddefctrl(dev)) = %d \n",
           usb_pipedevice(usb_snddefctrl(dev))
           )) ;

	hub_usb_device->children[port] = dev;
	dev->state = USB_STATE_POWERED;

	/* Reset the device, and detect its speed */
    
    USB_DBG_HUB(iprintf("[USB HUB] Reset the hub port .. \n" )) ;
    
	if (usb_hub_port_reset(hub_usb_device, port, dev, delay)) {
        dprintf("[USB] Error : Reset the hub port Fail !! \n" ) ;
		/*
		#ifndef _ETCPIP
        	USB_DBG_FLOW(iprintf(
           		"[USB FLOW] usb_hub_port_connect_change() Call "
                "usb_put_dev() ...\n"));
			usb_put_dev(dev);
		#endif // ndef _ETCPIP
		*/
		usb_connect(dev);
		usb_disconnect(&dev);
		break;
	   }

    USB_DBG_HUB(iprintf("[USB HUB] after Reset the hub port .. \n" )) ;

	/* Clear the connection change status */
	
	usb_clear_port_feature(hub_usb_device, port + 1, USB_PORT_FEAT_C_CONNECTION);
    
	/* Find a new address for it */
    USB_DBG_INIT(iprintf("[USB INIT] Find a new device ID for it .. \n" )) ;
	usb_connect(dev);

	/* Set up TT records, if needed  */
		if (hub_usb_device->tt) {
			dev->tt = hub_usb_device->tt;
			dev->ttport = hub_usb_device->ttport;
		} else if (dev->speed != USB_SPEED_HIGH
				&& hub_usb_device->speed == USB_SPEED_HIGH) {
			dev->tt = &hubstate->tt;
			dev->ttport = port + 1;
		}

	/* Create a readable topology string */
	// cdev = dev;

	/* Save readable and stable topology id, distinguishing
	 * devices by location for diagnostics, tools, etc.  The
	 * string is a path along hub ports, from the root.  Each
	 * device's id will be stable until USB is re-cabled, and
	 * hubs are often labeled with these port numbers.
	 *
	 * Initial size: ".NN" times five hubs + NUL = 16 bytes max
	 * (quite rare, since most hubs have 4-6 ports).
	 */
	pdev = dev->parent;

	#ifdef USB_DBGSTRING_ON
		#ifndef USB_NO_LIB
		if (pdev->devpath [0] != '0')	/* parent not root? */
			len = snprintf (dev->devpath, sizeof dev->devpath,
				"%s.%d", pdev->devpath, port + 1);
				/* root == "0", root port 2 == "2", port 3 that hub "2.3" */
			else
				len = snprintf (dev->devpath, sizeof dev->devpath,
				"%d", port + 1);
		#endif // ndef USB_NO_LIB
	#endif // USB_DBGSTRING_ON
 
	#ifdef USB_DBGSTRING_ON
		if (len == sizeof dev->devpath) {
			dev_err (&hubstate->intf->dev,
				"devpath size! usb/%03d/%03d path %s\n",
				dev->bus->busnum, dev->devnum, dev->devpath);
	       }
	#endif // USB_DBGSTRING_ON
	#ifdef USB_DBG_INFO_ON
		dev_info (&hubstate->intf->dev,
			"new USB device on port %d, assigned address %d\n",
			port + 1, dev->devnum);
    #endif

    dev->hub = hubstate ;
	dev->hub_usb_device = hub_usb_device ; // now , only tested on root hub
	dev->hub_port = port ; // Attached on root hub port

	iprintf("[USB] new USB device , assigned address %d\n",
			dev->devnum);
	#ifdef _USB_SYSLOG
	    //sprintf(usbsyslog_buf,"[USB] new USB device , assigned address %d ",
		//	dev->devnum );
		#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
		#endif // _SYSLOG==2
				{

		#ifdef USB_SYS_LOG_DEV_NEW
			//sprintf_SetSystemLog(USB_SYS_LOG_DEV_NEW, dev->devnum );
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, USB_SYS_LOG_DEV_NEW, dev->devnum );
		#else
			//sprintf_SetSystemLog("[USB] new USB device , assigned address %d ", dev->devnum );
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, "[USB] new USB device , assigned address %d ", dev->devnum );
		#endif // USB_SYS_LOG_DEV_NEW
	   }
	#endif // _SYSLOG

	/* put the device in the global device tree. the hub port
	 * is the "bus_id"; hubs show in hierarchy like bridges
	 */
	dev->dev.parent = dev->parent->dev.parent->parent;

	/* Run it through the hoops (find a driver, etc) */
	if (!usb_new_device(dev, &hub_usb_device->dev))
			goto done;

	/* Free the configuration if there was an error */
	/*
	#ifndef _ETCPIP
    	USB_DBG_FLOW(iprintf(
       		"[USB FLOW] usb_hub_port_connect_change() Call "
            "usb_put_dev() ...\n"));
			usb_put_dev(dev);
	#endif // ndef _ETCPIP
	*/
    USB_DBG_INIT(iprintf("[USB INIT] ==== New Device ====  ==9== Probe Try Next .. \n" )) ;
	/* Switch to a long reset time */
	delay = HUB_LONG_RESET_TIME;
    
   } // for i

 hub_usb_device->children[port] = NULL;
 usb_hub_port_disable(hub_usb_device, port);
 usb_device_count --;
    
 // ------------
 done:
 // ------------
 /*
 #ifndef NO_SEMPHORE
	   up(&usb_address0_sem);
 #endif // ndef NO_SEMPHORE
 */
 /*   
 #ifndef _ETCPIP
	if (portstr)
			kfree(portstr);
	if (tempstr)
			kfree(tempstr);
  #else
 */
	if (portstr) {
		UsbFreeMemU(USB_MEM_U_L, (unsigned char *) portstr
			 #if defined(USB_DBG_MEM_U_ON)
    	    	, 1024
             #endif
	        ) ;
       }     
	if (tempstr) {
		UsbFreeMemU(USB_MEM_U_L, (unsigned char *) tempstr
			 #if defined(USB_DBG_MEM_U_ON)
    	    	, 1024
             #endif
	        ) ;
       }     
 /* 
 #endif
 */
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_hub_port_connect_change() End.\n"));
}

static void usb_hub_events(void)
{
 unsigned long flags;
 struct list_head *tmp;
 struct usb_device *dev;
 struct usb_hub *hub;
 // struct usb_hub_status hubsts;
 struct usb_hub_status *hubsts_mod;		// 20021015 added by drsohn
 u16 hubstatus;
 u16 hubchange;
 u16 portstatus;
 u16 portchange;
 int i, ret;
 #ifdef USB_OHCI
	// 2003/3/21 B. S. Liaw
	struct hcdx_hcd ohci ;
 #endif
 int old_value ;
 int call_result ;
 #ifdef USB_DBG_EVT_ON
    int loop_count ;
 #endif // USB_DBG_EVT_ON
 int empty_flag ;

 USB_DBG_FLOWLOOP(iprintf("[USB FLOWLOOP] usb_hub_events() Begin ...\n"));
 #ifdef USB_DBG_EVT_ON
    loop_count = 0 ;
 #endif // USB_DBG_EVT_ON
    
 hubsts_mod = 0x0000L ;
 
 /*
  *  We restart the list everytime to avoid a deadlock with
  * deleting hubs downstream from this one. This should be
  * safe since we delete the hub from the event list.
  * Not the most efficient, but avoids deadlocks.
  */
  
 #ifdef RH_POLL_TSK 
 while (1)
 #endif
 {
	USB_DBG_EVT(iprintf("[USB EVT] ==== 0-1 ==== loop %d \n",loop_count));
    
// -------------------
start_wait:
// -------------------
	#ifdef USB_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0 ))
		{
			iprintf("[ASSERT][USB] usb_hub_events() : " ) ;
    			iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			iprintf("[ASSERT][USB] Continue ..\n") ;
    		}
	#endif // USB_BS_ASSERT

	dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
    
#if (_USB2 == 1)
	// temporary solution - bonian chen 15.04.2005
	// BLOCK_PREEMPTION takes too long, it will cause other time critical task to crash
	// see also storage\scsi.c
#if ( _USB_MS>0)
	UsbMsBlockIOLock();
#endif // _USB_MS
#endif	// defined (_USB2)
	if (!usb_hub_init_end)
	{
		goto event_end ;
	}
       
    	USB_DBG_EVT(iprintf("[USB EVT] ==== 0-2 ==== loop %d \n", loop_count));
	USB_DBG_EVT(iprintf("[USB EVT] &hub_event_list=%08XH , hub_event_list=%08XH \n",&hub_event_list, hub_event_list));
	BLOCK_PREEMPTION;
    	empty_flag = list_empty(&hub_event_list) ;
    	UNBLOCK_PREEMPTION;

	#ifdef USB_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0 ))
		{
			iprintf("[ASSERT][USB] usb_hub_events() : " ) ;
			iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    	iprintf("[ASSERT][USB] Continue ..\n") ;
		}
	#endif // USB_BS_ASSERT

	#ifdef RH_POLL_TSK // 2003/9/8 B. S. Liaw , use usb_rh_task
		USB_DBG_EVT(iprintf("[USB EVT] hub=%08XH \n",hub));
		#ifdef USB_BS_ASSERT
			if (!(  (unsigned long) hub > (unsigned long) 0x0100000L ))
			{
				iprintf("[ASSERT][USB] usb_hub_events(1) : " ) ;
		    		iprintf("Error : Invalid hub pointer : %08XH !!\n",hub) ;
		    		iprintf("[ASSERT][USB] Continue ..\n") ;
		   	}
		#endif // USB_BS_ASSERT
	#endif // RH_POLL_TSK

    	if (empty_flag)
       	{
		#ifdef RH_POLL_TSK // 2003/9/8 B. S. Liaw , use usb_rh_task
			if ((unsigned long) hub > (unsigned long) 0x0100000L ) {
				hub->event_empty = 1 ;
           		}
        		USB_DBG_FLOWLOOP(iprintf("[USB FLOWLOOP] usb_hub_events() : list_empty \n"));
        		USB_DBG_EVTLOOP(iprintf("[USB EVTLOOP] usb_hub_events() : list_empty \n"));
		#endif // RH_POLL_TSK
		goto event_end ;
       	}
	#ifdef RH_POLL_TSK // 2003/9/8 B. S. Liaw , use usb_rh_task
		if ((unsigned long) hub > (unsigned long) 0x0100000L ) {
			hub->event_empty = 0 ;
       		}
	#endif // RH_POLL_TSK

    	USB_DBG_EVT(iprintf("[USB EVT] ==== 1 ==== loop %d \n", loop_count));
        
    	#ifdef USB_DBG_EVT_ON
       		loop_count ++ ;
    	#endif // USB_DBG_EVT_ON
    
	USB_DBG_EVT(iprintf("[USB EVT] ==== 1 ==== get event list ..\n"));

	/* Grab the next entry from the beginning of the list */
    	BLOCK_PREEMPTION;
	tmp = hub_event_list.next;
	if (!usb_hub_init_end)
	{
		UNBLOCK_PREEMPTION;
#if (_USB2 == 1)
		// temporary solution - bonian chen 15.04.2005
		// BLOCK_PREEMPTION takes too long, it will cause other time critical task to crash
		// see also storage\scsi.c
#if ( _USB_MS>0)
		UsbMsBlockIOUnlock();
#endif // _USB_MS
#endif	// defined (_USB2)
		// 2003/9/9 B. S Liaw
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
		goto    start_wait ;
	}
       
	hub = list_entry(tmp, struct usb_hub, event_list);

	#ifdef USB_BS_ASSERT
		if (!(  (unsigned long) hub > (unsigned long) 0x0100000L ) )
		{
			iprintf("[ASSERT][USB] usb_hub_events(2) : " ) ;
		    	iprintf("Error : Invalid hub pointer : %08XH !!\n",hub) ;
		    	iprintf("[ASSERT][USB] Continue ..\n") ;
		}
	#endif // USB_BS_ASSERT

	#ifndef RH_POLL_TSK // 2003/9/8 B. S. Liaw , use usb_rh_task
    		if (empty_flag)
       		{
       			UNBLOCK_PREEMPTION;
			USB_DBG_FLOWLOOP(iprintf("[USB FLOWLOOP] usb_hub_events() : list_empty \n"));
        		USB_DBG_EVTLOOP(iprintf("[USB EVTLOOP] usb_hub_events() : list_empty \n"));
			goto event_end ;
       		}
       
		if ((unsigned long) hub > (unsigned long) 0x0100000L )
		{
			hub->event_empty = 0 ;
       		}
	#endif // ndef RH_POLL_TSK

	if (!usb_hub_init_end)
	{
		UNBLOCK_PREEMPTION;
#if (_USB2 == 1)
		// temporary solution - bonian chen 15.04.2005
		// BLOCK_PREEMPTION takes too long, it will cause other time critical task to crash
		// see also storage\scsi.c
#if ( _USB_MS>0)
		UsbMsBlockIOUnlock();
#endif // _USB_MS
#endif	// defined (_USB2)
		// 2003/9/9 B. S Liaw
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
		goto    start_wait ;
	}
       
	if (!hub->intf)
	{
		UNBLOCK_PREEMPTION;
#if (_USB2 == 1)
		// temporary solution - bonian chen 15.04.2005
		// BLOCK_PREEMPTION takes too long, it will cause other time critical task to crash
		// see also storage\scsi.c
#if ( _USB_MS>0)
		UsbMsBlockIOUnlock();
#endif // _USB_MS
#endif	// defined (_USB2)
        	USB_DBG_INFO(iprintf("[USB INFO] Error : hub->intf Null !!\n")) ;
		// 2003/9/9 B. S Liaw
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
		goto    start_wait ;
	}
       
	dev = interface_to_usbdev(hub->intf);

	if (!usb_hub_init_end)
	{
		UNBLOCK_PREEMPTION;
#if (_USB2 == 1)
		// temporary solution - bonian chen 15.04.2005
		// BLOCK_PREEMPTION takes too long, it will cause other time critical task to crash
		// see also storage\scsi.c
#if ( _USB_MS>0)
		UsbMsBlockIOUnlock();
#endif // _USB_MS
#endif	// defined (_USB2)
		// 2003/9/9 B. S Liaw
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
		goto    start_wait ;
	}
    
	list_del_init(tmp);
	UNBLOCK_PREEMPTION;

	#ifndef NO_SPINLOCK
		spin_unlock_irqrestore(&hub_event_lock, flags);
	#endif // ndef NO_SPINLOCK

	#ifdef USB_BS_ASSERT
       		if  ( !(    (unsigned long) hub >= (unsigned long) 0x00010000L )  )
           	{
            		iprintf("[ASSERT][USB] usb_hub_events() : " ) ;
            		iprintf("Error : hub = %08XH address out of range !!\n", hub ) ;
            		iprintf("[ASSERT][USB] Continue ....\n") ;
           	}
       		if  ( !( (hub->error >= 0 ) && (hub->error < 100 ) )  )
           	{
            		iprintf("[ASSERT][USB] usb_hub_events() : " ) ;
            		iprintf("Error : hub->error = %d value out of range !!\n", hub->error ) ;
            		iprintf("[ASSERT][USB] Continue ....\n") ;
           	}
    	#endif

	if  ( !(    (unsigned long) hub	>= (unsigned long) 0x00010000L ) )
        {
		goto event_end ;
        }

	if  ( !( (hub->error >= 0 ) && (hub->error < 100 ))  )
        {
		goto event_end ;
        }
    
	if (hub->error)
       	{
		dprintf("[USB] resetting hub %d for error %d \n", dev->devnum, hub->error);
		dev_dbg (&hub->intf->dev, "resetting for error %d\n", hub->error);

		if (usb_hub_reset(hub))
           	{
			dprintf("[USB] error resetting hub %d - disconnecting \n", dev->devnum);
			dev_dbg (&hub->intf->dev, "can't reset; disconnecting\n");
            		#ifndef NO_SEMPHORE
				up(&hub->khubd_sem);
         		#endif // ndef NO_SEMPHORE
            
			usb_hub_disconnect(dev);
			// continue;
#if (_USB2 == 1)
			// temporary solution - bonian chen 15.04.2005
			// BLOCK_PREEMPTION takes too long, it will cause other time critical task to crash
			// see also storage\scsi.c
#if ( _USB_MS>0)
			UsbMsBlockIOUnlock();
#endif // _USB_MS
#endif	// defined (_USB2)
			// 2003/9/9 B. S Liaw
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
			goto start_wait ;
		}

		hub->nerrors = 0;
		hub->error = 0;
	}

	USB_DBG_EVT(iprintf("[USB EVT] ==== 2 ==== get each port status ..\n"));
	USB_DBG_HUB(iprintf("[USB HUB] usb_hub_events() : &status=%08XH &change=%08XH \n" , &portstatus , &portchange ));
	for (i = 0; i < hub->descriptor->bNbrPorts; i++)
	{
		if (i >= USB_HUB_MAX_PORTS)
		{
			dprintf("[USB] Error : Invalid HUB Port Number : %d !!\n",i+1) ;
			continue ;
		}
		ret = usb_hub_port_status(dev, i, &portstatus, &portchange);
		if (ret < 0)
		{
			 USB_DBG_INIT(iprintf("[USB INIT] Error : get_port_status failed (err = %d) \n", ret )) ;
			 continue;
		}
		USB_DBG_HUB(iprintf("[USB HUB] usb_hub_events() : status=%08XH change=%08XH \n" , portstatus , portchange ));
        	USB_DBG_EVT(iprintf("[USB EVT] ==== 2.1 ==== port %d deal with port status changes ..\n", i+1 ));
            
                if (dev->sw_unplug[i])  /* Charles 20040429 */
                {
                	dev->sw_unplug[i] = 0;
                	portchange |= USB_PORT_STAT_C_CONNECTION;
		}
            
#ifdef _ETCPIP
#ifdef _USB2
		if (portchange & USB_PORT_STAT_C_NOT_OWNER)
		{
			// in order to avoid wrong message when USB 1.1 device connected to USB 2.0 HC
		}
		else
#endif	// defined(_USB2)
#endif	// defined(_ETCPIP)
		if (portchange & USB_PORT_STAT_C_CONNECTION)
		{
	    		USB_DBG_INFO(iprintf("[USB INFO]port %d connection change \n", i + 1 ));
        		USB_DBG_FLOW(iprintf("[USB FLOW] usb_hub_events() 2 call usb_hub_port_connect_change() ..\n"));
			usb_hub_port_connect_change(hub, i, portstatus, portchange);
			USB_DBG_FLOW(iprintf("[USB FLOW] usb_hub_events() 2 After Call usb_hub_port_connect_change(). \n"));
		}
          	else if (portchange & USB_PORT_STAT_C_ENABLE)
           	{
			dprintf("[USB] HUB port %d enable change, status %x \n", i + 1, portstatus);
			#ifdef _USB_SYSLOG
				#if (_SYSLOG==2)
					if (gSetting.syslog_ds.syslog_enable==1) 
				#endif // _SYSLOG==2
				{
					#ifdef USB_SYS_LOG_HUB_PORT_CHANGE
						sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, USB_SYS_LOG_HUB_PORT_CHANGE, i + 1, portstatus);
					#else
						sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, "[USB] HUB port %d enable change, status %x ", i + 1, portstatus);
					#endif // USB_SYS_LOG_HUB_PORT_CHANGE
				}
			#endif // _SYSLOG
			dev_dbg( hubdev (dev),"port %d enable change, status %x\n", i + 1, portstatus);
			usb_clear_port_feature(dev, i + 1, USB_PORT_FEAT_C_ENABLE);

			/*
				 * EM interference sometimes causes badly
				 * shielded USB devices to be shutdown by
				 * the hub, this hack enables them again.
			  * Works at least with mouse driver. 
			*/
			if (!(portstatus & USB_PORT_STAT_ENABLE)
				    && (portstatus & USB_PORT_STAT_CONNECTION)
				    && (dev->children[i]))
			{
				dprintf("[USB] already running port %i disabled by hub (EMI?), re-enabling... \n", i + 1);
				dev_err (&hub->intf->dev, "port %i disabled by hub (EMI?), re-enabling...", i + 1);
                                USB_DBG_FLOW(iprintf("[USB FLOW] usb_hub_events() 3 call usb_hub_port_connect_change() ..\n"));
#if 1  /* Charles 20040708 */
				usb_set_port_feature(dev, i + 1, USB_PORT_FEAT_ENABLE);
#else
				usb_hub_port_connect_change(hub, i, portstatus, portchange);
#endif
			}
		} 

	     	if (portchange & USB_PORT_STAT_C_SUSPEND)
       		{
			dbg("port %d suspend change", i + 1);
			dev_dbg (&hub->intf->dev,"suspend change on port %d\n",	i + 1);
			usb_clear_port_feature(dev,i + 1,  USB_PORT_FEAT_C_SUSPEND);
	   	}
			
		if (portchange & USB_PORT_STAT_C_OVERCURRENT)
		{
			err("port %d over-current change", i + 1);
			dev_err (&hub->intf->dev,"over-current change on port %d\n",i + 1);
			usb_clear_port_feature(dev,i + 1, USB_PORT_FEAT_C_OVER_CURRENT);
			usb_hub_power_on(hub);
		}

		if (portchange & USB_PORT_STAT_C_RESET)
           	{
			dprintf("[USB] HUB reset change on port %d\n", i + 1);
			dev_dbg (&hub->intf->dev,"reset change on port %d\n", i + 1);
			USB_DBG_HUB(iprintf("[USB HUB] clear reset change ..\n" )) ;
			usb_clear_port_feature(dev, i + 1, USB_PORT_FEAT_C_RESET);
		}
	} /* end for i */

	/* deal with hub status changes */
    	USB_DBG_EVT(iprintf("[USB EVT] ==== 3 ==== deal with hub status changes ..\n" ));
        
	// 2003/7/14 B. S. Liaw BugFix , Need Un-Cached
	#ifdef _ETCPIP
		hubsts_mod = (struct usb_hub_status *) UsbAllocMemU(USB_MEM_U_S
			#ifdef USB_DBG_MEM_U_ON
    	             		,sizeof(struct usb_hub_status)
	                #endif
                );
	    	USB_DBG_MEM(iprintf("[USB MEM] hubsts_mod = %08XH \n" , hubsts_mod )) ;
	    	if ( hubsts_mod == NULL )
    	   	{
			dprintf("[USB] Error : out of memory !!\n");
			goto event_end ;
	   	}
	#endif
    
	#ifdef USB_NO_VMM // no vitrtul memory management
		#ifdef USB_BS_ASSERT
			if ( !(USB_IS_UN_CACHED(hubsts_mod)) )
			{
				iprintf("[ASSERT][USB] usb_hub_configure() : " ) ;
    		    		iprintf("Error : Not Un-Cached !!\n" ) ;
			    	iprintf("[ASSERT][USB] hubsts_mod addr = %08XH \n", hubsts_mod ) ;
			    	iprintf("[ASSERT][USB] Continue ..\n") ;
    		   	}
		#endif // USB_BS_ASSERT
	#endif // USB_NO_VMM
 
	// 2003/7/14 B. S. Liaw BugFix , Need Un-Cached
    	if (usb_get_hub_status(dev,hubsts_mod) < 0)
       	{
		dprintf("[USB] Error : Get HUB status failed !! \n") ;
       	}
	else
       	{
		hubstatus = le16_to_cpup(&(hubsts_mod->wHubStatus));
        	hubchange = le16_to_cpup(&(hubsts_mod->wHubChange));
		if (hubchange & HUB_CHANGE_LOCAL_POWER)
           	{
			dprintf("[USB] hub power change \n");
			dev_dbg (&hub->intf->dev, "power change\n");
			usb_clear_hub_feature(dev, C_HUB_LOCAL_POWER);
		}
           
		if (hubchange & HUB_CHANGE_OVERCURRENT)
           	{
			USB_DBG_INFO(iprintf("[USB INFO] hub overcurrent change \n"));
            		dev_dbg (&hub->intf->dev, "overcurrent change\n");
			#ifndef _ETCPIP
				wait_ms(500);	/* Cool down */
             		#else
				wait_ms(1000);
            		#endif
			usb_clear_hub_feature(dev, C_HUB_OVER_CURRENT);
            		usb_hub_power_on(hub);
		}
	}
#ifdef RH_POLL_TSK
#if (_USB2 == 1)
	// temporary solution - bonian chen 15.04.2005
	// BLOCK_PREEMPTION takes too long, it will cause other time critical task to crash
	// see also storage\scsi.c
#if ( _USB_MS>0)
	UsbMsBlockIOUnlock();
#endif // _USB_MS
#endif	// defined (_USB2)
#endif
       
	// 2003/7/14 B. S. Liaw BugFix , Need Un-Cached
	if (hubsts_mod)
	{
		#ifndef USB_MEM_LIB
 			kfree(hubsts_mod);
		#else
			// 2003/7/14 B. S. Liaw BugFix , Need Un-Cached
			UsbFreeMemU(USB_MEM_U_S, (unsigned char *) hubsts_mod
				#if defined(USB_DBG_MEM_U_ON)
	    	    			,sizeof(struct usb_hub_status)
    		        	#endif
			       ) ;
		#endif
	}
	hubsts_mod = (struct usb_hub_status *)0x0000L ;
        
	#ifndef NO_SEMPHORE
		up(&hub->khubd_sem);
 	#endif // ndef NO_SEMPHORE
    	USB_DBG_EVT(iprintf("[USB EVT] ==== 4 ==== loop end \n"));
 } /* end while (1) */

// ------------------
event_end:
// ------------------

#if (_USB2 == 1)
	// temporary solution - bonian chen 15.04.2005
	// BLOCK_PREEMPTION takes too long, it will cause other time critical task to crash
	// see also storage\scsi.c
#if ( _USB_MS>0)
	UsbMsBlockIOUnlock();
#endif // _USB_MS
#endif	// defined (_USB2)
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(&hub_event_lock, flags);
 #else
    // UNBLOCK_PREEMPTION;
 #endif // ndef NO_SPINLOCK
 
 USB_DBG_FLOWLOOP(iprintf("[USB FLOWLOOP] usb_hub_events() End.\n"));
} // function usb_hub_events()

#ifndef _ETCPIP
  /*
  static int usb_hub_thread(void *__hub)
  */
#else 
  void usb_task(void)
#endif
{
#ifdef TNV105_INTBUG
 extern void TNV105_InterruptPolling(void);
#endif
#ifdef RH_INT_BY_POLL
 int time_out ;
 struct list_head *entry, *tmp ;
 struct usb_hub *hub;
 struct usb_hcd		*hcd ;
#endif
 int empty_flag = 0 ;
#ifdef USB_HUB_LOOP_EVENT // 2004/1/4 B. S. Liaw
 int i ;
#endif // USB_HUB_LOOP_EVENT
 
{
extern unsigned long get_data_segment(void);
extern void set_data_segment(unsigned long new_gp);
set_data_segment(get_data_segment());
}
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_task() Begin ...\n"));
 
 empty_flag = 0 ;
 #ifdef USB_HUB_LOOP_EVENT // 2004/1/4 B. S. Liaw
	if ( (i=period(MT_EVT_USB_HUB, USB_HUB_PERIOD*CLOCKHZ)) != SUCCESS ) {
        	dprintf("[USB] Error : HUB period call fails %d !!\n", i);
    	}
 #endif // USB_HUB_LOOP_EVENT

 /* Send me a signal to get me die (for debugging) */
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_task() infinite loop call usb_hub_events() .. \n")) ;
 do
 {
	#ifdef USB_HUB_LOOP_EVENT // 2004/1/4 B. S. Liaw
        	wteset_dec(MT_EVT_USB_HUB, 0);
        	clrevt(MT_EVT_USB_HUB);
	#endif // USB_HUB_LOOP_EVENT

// -------------------
start_wait:
// -------------------

	USB_DBG_HUB_LOOP(iprintf("[USB HUB_LOOP] usb_task() : Start .. \n" )) ;
	#ifdef USB_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0 ) )
		{
			iprintf("[ASSERT][USB] usb_task() : " ) ;
   			iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			iprintf("[ASSERT][USB] Continue ..\n") ;
   		}
	#endif // USB_BS_ASSERT

	#ifdef USB_DBG_MEM_AREA_VERIFY_ON
		USB_DBG_AREA_MSG(iprintf("[USB AREA] usb_task() : Verify Memory Area ..\n")) ;
		if (! VerifyMemArea() ) {
			iprintf("[USB] usb_task() Verify Memory Area Error !\n") ;
		}
	#endif // USB_DBG_MEM_AREA_VERIFY_ON
    
	if (!hcd_init_ok)
	{
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0 ))
			{
				iprintf("[ASSERT][USB] usb_task() : " ) ;
    		    		iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    	iprintf("[ASSERT][USB] Continue ..\n") ;
    		   	}
		#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
        	continue ;
	}
    
	if (!usb_hub_init_end)
	{
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0 ))
			{
				iprintf("[ASSERT][USB] usb_task() : " ) ;
    		    		iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    	iprintf("[ASSERT][USB] Continue ..\n") ;
    		  	}
		#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
        	continue ;
	}
	BLOCK_PREEMPTION;
	empty_flag = list_empty(&hub_list) ;
	UNBLOCK_PREEMPTION;
    	if (empty_flag || (!usb_hub_init_end))
    	{
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0 ))
			{
				iprintf("[ASSERT][USB] usb_task() : " ) ;
    		    		iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
				iprintf("[ASSERT][USB] Continue ..\n") ;
    		   	}
		#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
        	continue ;
	}
    
	BLOCK_PREEMPTION;
	entry = (&hub_list)->next ;
	tmp = entry->next ;
	UNBLOCK_PREEMPTION;
    
    	for ( ; entry != (&hub_list) ; )
    	{
		if (!usb_hub_init_end)
		{
			#ifdef USB_BS_SYSCAL_ASSERT
				if (!(  mt_busy == 0 ))
				{
					iprintf("[ASSERT][USB] usb_task() : " ) ;
    			    		iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
				    	iprintf("[ASSERT][USB] Continue ..\n") ;
	    		   	}
			#endif // USB_BS_ASSERT
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
			goto    start_wait ;
	       	}
		usb_hub_events();
     
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0 ))
			{
				iprintf("[ASSERT][USB] usb_task() : " ) ;
    		    		iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    	iprintf("[ASSERT][USB] Continue ..\n") ;
    		   	}
		#endif // USB_BS_ASSERT
		#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
		#else
            		YIELD();
		#endif
		
		if (!usb_hub_init_end)
		{
			// 2003/9/9 B. S Liaw
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
			goto    start_wait ;
	       	}
           
		BLOCK_PREEMPTION;
		hub = list_entry(entry, struct usb_hub, hub_list);
	    	UNBLOCK_PREEMPTION;

		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0 ) )
			{
				iprintf("[ASSERT][USB] usb_task() : " ) ;
    		    		iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    	iprintf("[ASSERT][USB] Continue ..\n") ;
	   		}
		#endif // USB_BS_ASSERT

		// #ifdef RH_INT_BY_POLL
		// if ( ! global_hub[hub_count])
		if (!hub)
		{
			#ifdef USB_BS_SYSCAL_ASSERT
				if (!(  mt_busy == 0 ))
				{
					iprintf("[ASSERT][USB] usb_task() : " ) ;
    			    		iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
				    	iprintf("[ASSERT][USB] Continue ..\n") ;
	    		   	}
			#endif // USB_BS_ASSERT
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
		    	continue ;
		 }
		// #endif
        
		#ifdef USB_RES_SEMPHORE    // use MT_RES_USB semphore
			if (hub->hub_in_submit == 0 ) {
				relres(MT_RES_USB) ;
			}
      
			#ifdef USB_BS_SYSCAL_ASSERT
				if (!(  mt_busy == 0 ))
				{
					iprintf("[ASSERT][USB] usb_task() : " ) ;
    			    		iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
				    	iprintf("[ASSERT][USB] Continue ..\n") ;
	    		   	}
			#endif // USB_BS_ASSERT
			#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
			#else
				YIELD();
			#endif
    
			USB_DBG_EVTLOOP(iprintf("[USB EVTLOOP] usb_task() : get USB Resource ..\n"));
    
			if (hub->hub_in_submit == 0 ) {
				relres(MT_RES_USB) ;
               		}
            
	    		if (getres(MT_RES_USB, 1000) != 0) {
		        	iprintf("[USB] usb_task() : getres Fail !!") ;
	    	    		continue ;
   			}
		#endif
    		
		if (!usb_hub_init_end) {
			// 2003/9/9 B. S Liaw
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
			goto    start_wait ;
		}
		// USB_DBG_ROOTHUB(iprintf("[USB ROOTHUB] usb_task() : Wait global_hub[%d]->hub_in_submit Free ..\n" , hub_count ));
		USB_DBG_HUB_LOOP(iprintf("[USB HUB_LOOP] usb_task() : Wait : hub[%08XH]->hub_in_submit Free .. \n", hub )) ;
            
		usb_hub_events();
            
		if (!usb_hub_init_end)
		{
			#ifdef USB_BS_SYSCAL_ASSERT
				if (!(  mt_busy == 0 ))
				{
					iprintf("[ASSERT][USB] usb_task() : " ) ;
    			    		iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
				    	iprintf("[ASSERT][USB] Continue ..\n") ;
	    		   	}
			#endif // USB_BS_ASSERT
			dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
			goto    start_wait ;
		}
               
                #ifdef TNV105_INTBUG
			TNV105_InterruptPolling();
                #endif
               
		#ifdef USB_NO_ISR_CALLBACK
#if (_USB2 == 1)
			// temporary solution - bonian chen 15.04.2005
			// BLOCK_PREEMPTION takes too long, it will cause other time critical task to crash
			// see also storage\scsi.c
#if ( _USB_MS>0)
			UsbMsBlockIOLock();
#endif // _USB_MS
#endif	// defined (_USB2)
			if (hub->urb->awd_done == ~hub->urb->awd_done_old)
			{
				if (hub->urb->callback)
				{
					USB_DBG_CALLBACK(iprintf("[USB CALLBACK] usb_task() : Call hub->urb->callback() ...\n" )) ;
					hub->urb->callback(hub->urb);
				}
				else
				{
					USB_DBG_CALLBACK(iprintf("[USB CALLBACK] usb_task() : Null hub->urb->callback() !! \n" )) ;
				}

				#ifdef USB_NO_ISR_UNLINK
				// 2003/9/12 B. S. Liaw BugFix
					hcd = hub->urb->dev->bus->hcpriv_hcd;
					if (hub->urb->dev != hcd->self.root_hub) // FixMe : B. S. Liaw , I do know why !!
				       	{
						int status_unlink ;
						status_unlink = usb_unlink_urb(hub->urb);
						if (status_unlink && status_unlink != -EINPROGRESS) {
							USB_DBG_INFO(iprintf("[USB INFO] usb_task() : usb_unlink_urb fail : %d !!\n", status_unlink));
						}
					 }
				#endif // OHCI_ED_SKEP_FIX
			} // if ok
#if (_USB2 == 1)
			// temporary solution - bonian chen 15.04.2005
			// BLOCK_PREEMPTION takes too long, it will cause other time critical task to crash
			// see also storage\scsi.c
#if ( _USB_MS>0)
			UsbMsBlockIOUnlock();
#endif // _USB_MS
#endif	// defined (_USB2)
		#endif // USB_NO_ISR_CALLBACK
            
		if (hub->hub_in_submit == 0 )
		{
			if (!usb_hub_init_end)
			{
				// 2003/9/9 B. S Liaw
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
				goto    start_wait ;
			}

			#ifdef RH_POLL_TSK    // 2003/9/8 B. S. Liaw , use usb_rh_task
				if (!hub->event_empty)
				{
					dprintf("[USB] Error : HUB Event List Not Empty !!\n") ;
    					// 2003/9/9 B. S Liaw
					dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
					goto    start_wait ;
				}
			#endif //  RH_POLL_TSK    // 2003/9/8 B. S. Liaw , use usb_rh_task

			BLOCK_PREEMPTION;
			empty_flag = list_empty(&hub->event_list) ;
			UNBLOCK_PREEMPTION;

			if (empty_flag)
			{
				USB_DBG_EVTLOOP(iprintf("[USB EVTLOOP] add event list ..\n"));
     				BLOCK_PREEMPTION;
				list_add(&hub->event_list, &hub_event_list);
				UNBLOCK_PREEMPTION;
				#ifndef NO_WAITQUEUE
					wake_up(&khubd_wait);
				#endif // ndef NO_WAITQUEUE
			}
			else
			{
				dprintf("[USB] Error : HUB Event List Not Empty !!\n") ;
			}
                   
			usb_hub_events();
			#ifdef HUB_NO_INT_PIPE
    		    		hub->hub_in_submit = 0 ;
			#else
    		    		hub->hub_in_submit = 1 ;
			#endif
			USB_DBG_HUB(iprintf("[USB HUB] usb_task() : Set Value : hub[%08XH]->hub_in_submit=%d \n", hub,hub->hub_in_submit)) ;                      
			USB_DBG_HUB(iprintf("[USB HUB] usb_task() : usb_submit_urb() .. \n" )) ;
			if (!usb_hub_init_end)
			{
				// 2003/9/9 B. S Liaw
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
				goto    start_wait ;
			}
                   
			if (!hub->urb) // hub been removed
                   	{
				USB_DBG_INIT(iprintf("[USB INIT] usb_task() : HUB removed .. \n" )) ;
				// 2003/9/9 B. S Liaw
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
				goto    start_wait ;
                   	}
                   
#if (_USB2 == 1)
			// temporary solution - bonian chen 15.04.2005
			// BLOCK_PREEMPTION takes too long, it will cause other time critical task to crash
			// see also storage\scsi.c
#if ( _USB_MS>0)
			UsbMsBlockIOLock();
#endif //_USB_MS
#endif	// defined (_USB2)
            		// 2003/7/9 B. S. Liaw BugFix   
			hcd = hub->urb->dev->bus->hcpriv_hcd;
                
			#ifdef USB_BS_SYSCAL_ASSERT
				if (!(  mt_busy == 0 ))
				{
					iprintf("[ASSERT][USB] usb_task() : " ) ;
			    		iprintf("Error : mt_busy Not Zero !! %08XH\n",mt_busy) ;
					iprintf("[ASSERT][USB] Continue ..\n") ;
			    	}
			#endif // USB_BS_ASSERT

			if (  hub->urb->dev == hcd->self.root_hub )
                   	{ 
				USB_DBG_HUB_LOOP(iprintf("[USB HUB_LOOP] usb_task() : Root HUB !! \n" )) ;
                   	}
                  	else
                   	{ 
				USB_DBG_HUB_LOOP(iprintf("[USB HUB_LOOP] usb_task() : Not Root HUB \n" )) ;
                    		if (hub->urb->hcpriv)
    	               		{ 
					USB_DBG_HUB_LOOP(iprintf("[USB HUB_LOOP] usb_task() : free hcpriv .. \n" )) ;
					hcd->driver->urb_free_priv(hcd, hub->urb->hcpriv);
					hub->urb->hcpriv = NULL;
	                   	}
				else
                       		{
					USB_DBG_HUB_LOOP(iprintf("[USB HUB_LOOP] usb_task() : hcpriv Null !! \n" )) ;
                       		}
                   	}
                   
			#ifdef USB_DBG_CALLBACK_ON  
				hub->hub_int_done = 0 ;    
			 	USB_DBG_CALLBACK(iprintf("[USB CALLBACK] hub_irq() : Set Value : hub->hub_int_done=%d \n",hub->hub_int_done)) ;
			#endif // USB_DBG_CALLBACK_ON

			#ifndef HUB_NO_INT_PIPE
				#ifdef _ETCPIP
				    if (usb_submit_urb(hub->urb))
				#endif
				{
					USB_DBG_HUB(iprintf("[USB HUB] usb_task() : usb_submit_urb failed !! \n")) ;
					iprintf("[USB] usb_submit_urb failed !!\n");
					#ifdef USB_RES_SEMPHORE    // use MT_RES_USB semphore
			        		relres(MT_RES_USB) ;
		            		#endif
				}
			#endif // nef HUB_NO_INT_PIPE
#if (_USB2 == 1)
			// temporary solution - bonian chen 15.04.2005
			// BLOCK_PREEMPTION takes too long, it will cause other time critical task to crash
			// see also storage\scsi.c
#if ( _USB_MS>0)
			UsbMsBlockIOUnlock();
#endif // _USB_MS
#endif	// defined (_USB2)
           
			#ifdef USB_BS_SYSCAL_ASSERT
				if (!(  mt_busy == 0 ))
			   	{
					iprintf("[ASSERT][USB] usb_task() : " ) ;
    		    			iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    		iprintf("[ASSERT][USB] Continue ..\n") ;
    		   		}
			#endif // USB_BS_ASSERT
			#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
		        #else
        		   	YIELD();
		        #endif
            
			#ifdef USB_RES_SEMPHORE    // use MT_RES_USB semphore
		    		if (hub->hub_in_submit == 0 )
				{
					relres(MT_RES_USB) ;
			        }
			#endif
		}
        
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0 ))
			{
				iprintf("[ASSERT][USB] usb_task() : " ) ;
    		    		iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    	iprintf("[ASSERT][USB] Continue ..\n") ;
    		  	}
		#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, CLOCKHZ/10) ;		// wait 100 ms for each hub
		BLOCK_PREEMPTION;
        	entry = tmp ;
        	tmp = entry->next ;
	    	UNBLOCK_PREEMPTION;
	} // for each hub
 } while (1) ;

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_task() infinite loop call usb_hub_events() End. \n"));

 #ifndef _ETCPIP
	dbg("usb_hub_thread exiting");
	complete_and_exit(&khubd_exited, 0);
 #endif // ndef _ETCPIP
 USB_DBG_INIT(iprintf("[USB INIT] usb_task exiting \n"));
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_task() End.\n"));
}

static struct usb_device_id hub_id_table [] = {
    { .match_flags = USB_DEVICE_ID_MATCH_DEV_CLASS,
      .bDeviceClass = USB_CLASS_HUB},
    { .match_flags = USB_DEVICE_ID_MATCH_INT_CLASS,
      .bInterfaceClass = USB_CLASS_HUB},
    { }						/* Terminating entry */
};

#ifndef _ETCPIP
   MODULE_DEVICE_TABLE (usb, hub_id_table);
#endif // ndef _ETCPIP

static struct usb_driver hub_driver = {
 #ifdef USB_DBGSTRING_ON
	.name =		"hub",
 #endif // USB_DBGSTRING_ON
 .usbdev_probe =	hub_probe,
 .disconnect =	hub_disconnect,
 .ioctl =	hub_ioctl,
 .id_table =	hub_id_table,
};

/*
 * This should be a separate module.
 */
int usb_hub_init(void)
{
 #ifndef _ETCPIP
    int pid;
  #else
    int status1 ;
 #endif


 USB_DBG_FLOW(iprintf("[USB FLOW] usb_hub_init() Begin ..\n"));
 
 hcd_init_ok = 0 ;

 #ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] usb_hub_init() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
 #endif // USB_BS_ASSERT

 USB_DBG_INIT(iprintf("[USB INIT] usb_register(&hub_driver) ..\n"));
 if (usb_register(&hub_driver) < 0)
    {
	 dprintf("[USB] Unable to register USB hub driver !!\n");
	 return -1;
	}
 USB_DBG_INIT(iprintf("[USB INIT] usb_register(&hub_driver) OK.\n"));
 /*
 #ifndef _ETCPIP
	pid = kernel_thread(usb_hub_thread, NULL,
		CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
	if (pid >= 0) {
		khubd_pid = pid;

		return 0;
	}
  #else
 */   
    // status1 = runtsk(100, usb_task, 6000); // 100 too high !!
    // status1 = runtsk(80, usb_task, 6000);
	#ifdef INT_SCHEDULING
		status1 = runtsk(100, (void (fAR *)(void))usb_task, 8000);
	#else
		status1 = runtsk(80, (void (fAR *)(void))usb_task, 8000);  /* Charles 20040426 */
	#endif // INT_SCHEDULING

    if (status1 < SUCCESS )
       {
        dprintf("RUNTASK usb_task fails\n");
        goto exi2;
       }
      else
       {
        dprintf("RUNTASK id=%d usb_task ...\n", status1);
       }
   
   USB_DBG_FLOW(iprintf("[USB FLOW] usb_hub_init() End.\n"));
   return 0;
 /*
 #endif
 */

 exi2:
 
   	/* Fall through if kernel_thread failed */
   	usb_deregister(&hub_driver);
   	dprintf("[USB] failed to start usb_task \n");

	USB_DBG_FLOW(iprintf(
    	"[USB FLOW] usb_hub_init() End With Fail.\n"));
	return -1;
}

#ifndef _ETCPIP
void usb_hub_cleanup(void)
{
 int ret;

 /* Kill the thread */
 /*
 #ifndef _ETCPIP
	ret = kill_proc(khubd_pid, SIGTERM, 1);

    // wait_for_completion() is in <linux/completion.h>
	wait_for_completion(&khubd_exited);
 #endif // ndef _ETCPIP
 */
	/*
	 * Hub resources are freed for us by usb_deregister. It calls
	 * usb_driver_purge on every device which in turn calls that
	 * devices disconnect function if it is using this driver.
	 * The hub_disconnect function takes care of releasing the
	 * individual hub resources. -greg
	 */
	usb_deregister(&hub_driver);
} /* usb_hub_cleanup() */
#endif // ndef _ETCPIP

/*
 * WARNING - If a driver calls usb_reset_device, you should simulate a
 * disconnect() and probe() for other interfaces you doesn't claim. This
 * is left up to the driver writer right now. This insures other drivers
 * have a chance to re-setup their interface.
 *
 * Take a look at proc_resetdevice in devio.c for some sample code to
 * do this.
 */
int usb_reset_device(struct usb_device *dev)
{
 struct usb_device *parent = dev->parent;
 struct usb_device_descriptor *descriptor_mod;					// 20021015 added by drsohn
 int i, ret, port = -1;

 if (!parent)
   {
	dprintf("[USB] Error : attempting to reset root hub! \n");
	return -EINVAL;
   }

 for (i = 0; i < parent->maxchild; i++)
	if (parent->children[i] == dev)
       {
		port = i;
		break;
	   }

 if (port < 0)
	return -ENOENT;

 #ifndef NO_SEMPHORE
	down(&usb_address0_sem);
 #endif // ndef NO_SEMPHORE

 /* Send a reset to the device */
 USB_DBG_HUB(iprintf("[USB HUB] Reset the device .. \n" )) ;
 if(usb_hub_port_reset(parent, port, dev, HUB_SHORT_RESET_TIME))
   {
	usb_hub_port_disable(parent, port);
    #ifndef NO_SEMPHORE
		up(&usb_address0_sem);
    #endif // ndef NO_SEMPHORE
		return(-ENODEV);
   }

 /* Reprogram the Address */
 USB_DBG_HUB(iprintf("[USB HUB] Set new address .. \n" )) ;
 ret = usb_set_address(dev);
 if (ret < 0)
   {
	dprintf("[USB] Error : HUB device not accepting new address (error=%d)", ret);
	usb_hub_port_disable(parent, port);
	#ifndef NO_SEMPHORE
		up(&usb_address0_sem);
	#endif // ndef NO_SEMPHORE
	return ret;
   }

 /* Let the SET_ADDRESS settle */
 wait_ms(10);

 #ifndef NO_SEMPHORE
	up(&usb_address0_sem);
 #endif // ndef NO_SEMPHORE

 /*
  * Now we fetch the configuration descriptors for the device and
  * see if anything has changed. If it has, we dump the current
  * parsed descriptors and reparse from scratch. Then we leave
  * the device alone for the caller to finish setting up.
  *
  * If nothing changed, we reprogram the configuration and then
  * the alternate settings.
  */

 #ifndef _ETCPIP
    /*
	// descriptor = kmalloc(sizeof *descriptor, GFP_NOIO);
	
	descriptor_mod = (struct usb_device_descriptor *)
                     kmalloc(sizeof(struct usb_device_descriptor),
                     GFP_KERNEL|GFP_DMA);
                     // 20021015 added by drsohn
	if (!descriptor_mod) {
		return -ENOMEM;
	}
    */
  #else
	// descriptor_mod = usb_device_descriptor_array ;
	#ifndef USB_MEM_LIB
	descriptor_mod = (struct usb_device_descriptor *)
            NonCache_contigmalloc(
                     	sizeof(struct usb_device_descriptor)) ;
	    	// NonCache_contigmalloc() cannot be free now !!
     #else
	    // 	alloc_mem(sizeof(struct usb_device_descriptor)) ;
		USB_DBG_MEM_U(iprintf(
		    "[USB MEM_U] usb_reset_device() Call UsbAllocMemU ..\n"
            ));
		descriptor_mod = (struct usb_device_descriptor *)
		#ifdef USB_DBG_MEM_U_ON
        	UsbAllocMemU(USB_MEM_U_M,
            			sizeof(struct usb_device_descriptor));
         #else
        	UsbAllocMemU(USB_MEM_U_M);
        #endif   
    #endif
    
    USB_DBG_MEM(iprintf(
       "[USB MEM] descriptor_mod = %08XH \n",descriptor_mod));
    if ( descriptor_mod == NULL )
       {
		dprintf("out of memory\n");
		return ret ;
	   }
 #endif
	
 //ret = usb_get_descriptor(dev, USB_DT_DEVICE, 0, &descriptor,
 //                         sizeof(descriptor));
 			// 20021015 remarked by drsohn
             
 ret = usb_get_descriptor(dev, USB_DT_DEVICE, 0, descriptor_mod,
 							sizeof(descriptor_mod));
                            // 20021015 modified by drsohn
                            
 memcpy(&descriptor_mod,descriptor_mod,
		sizeof(struct usb_device_descriptor)
       );					// 20021015 added by drsohn

 if (ret < 0)
   {
	#ifndef _ETCPIP
        /*
		kfree(descriptor_mod);
        */
	 #else
		#ifndef USB_MEM_LIB
		// NonCache_contigmalloc() cannot be free now !!
		#elif defined(USB_DBG_MEM_U_ON)
        UsbFreeMemU(USB_MEM_U_M, (unsigned char *) descriptor_mod,
        		sizeof(struct usb_device_descriptor)) ;
		#else
        UsbFreeMemU(USB_MEM_U_M, (unsigned char *) descriptor_mod) ;
		#endif
	#endif

	return ret;
   }

 le16_to_cpus(&descriptor_mod->bcdUSB);
 le16_to_cpus(&descriptor_mod->idVendor);
 le16_to_cpus(&descriptor_mod->idProduct);
 le16_to_cpus(&descriptor_mod->bcdDevice);

 if (memcmp(&dev->descriptor, descriptor_mod,
 			 sizeof(*descriptor_mod)))
   {
	#ifndef _ETCPIP
		kfree(descriptor_mod);
	 #else
		#ifndef USB_MEM_LIB
			// NonCache_contigmalloc() cannot be free now !!
		#elif defined(USB_DBG_MEM_U_ON)
	        UsbFreeMemU(USB_MEM_U_M, (unsigned char *)descriptor_mod,
            		sizeof(struct usb_device_descriptor)) ;
		#else
	        UsbFreeMemU(USB_MEM_U_M, (unsigned char *) descriptor_mod) ;
		#endif
	#endif
    
	USB_DBG_FLOW(iprintf(
		"[USB FLOW] usb_reset_device() call usb_destroy_"
        "configuration() ...\n"
		));
        
	usb_destroy_configuration(dev);

	ret = usb_get_device_descriptor(dev);
	// if (ret < 18 ) {
	if (ret < sizeof(dev->descriptor))
       {
		if (ret < 0)
           {
			#ifdef USB_DBGSTRING_ON
				err("unable to get device %s descriptor "
					"(error=%d)", dev->devpath, ret);
             #else
				dprintf("[USBHUB] unable to get device descriptor "
					"(error=%d)", ret);
            #endif
           }
		  else
           {
			#ifdef USB_DBGSTRING_ON
				err("USB device %s descriptor short read "
						"(expected %Zi, got %i)", 
						dev->devpath,
						sizeof(dev->descriptor), ret);
             #else
				dprintf("[USB] Error : device descriptor short read "
						"(expected %Zi, got %i)", 
						sizeof(dev->descriptor), ret);
            #endif
     	   }
		clear_bit(dev->devnum, dev->bus->devmap.devicemap);
		dev->devnum = -1;
		return -EIO;
	   }

	ret = usb_get_configuration(dev);
	if (ret < 0)
       {
		err("unable to get configuration (error=%d)", ret);
		USB_DBG_FLOW(iprintf(
			"[USB FLOW] usb_reset_device() call usb_destroy_configuration() ...\n"
		    ));
		usb_destroy_configuration(dev);
		clear_bit(dev->devnum, dev->bus->devmap.devicemap);
		dev->devnum = -1;
		return 1;
	   }

    USB_DBG_MEM(iprintf(
        "[USB MEM] usb_reset_device() : &(dev->actconfig)"
        " = %08XH \n",
        &(dev->actconfig)
        ));
        
	dev->actconfig = dev->config;
    
    USB_DBG_MEM(iprintf(
        "[USB MEM] usb_reset_device() : Set Value : dev->actconfig"
        " = %08XH \n",
        dev->actconfig
        ));
    USB_DBG_HUB(iprintf(
        "[USB INIT] usb_reset_device() : Set Value : dev->actconfig"
        " = %08XH \n",
        dev->actconfig
        ));
	usb_set_maxpacket(dev);

	return 1;
   }

 #ifndef _ETCPIP
	kfree(descriptor);
  #else
	#ifndef USB_MEM_LIB
		// NonCache_contigmalloc() cannot be free now !!
	#elif defined(USB_DBG_MEM_U_ON)
        UsbFreeMemU(USB_MEM_U_M, (unsigned char *) descriptor_mod,
        		sizeof(struct usb_device_descriptor)) ;
	#else
        UsbFreeMemU(USB_MEM_U_M, (unsigned char *) descriptor_mod) ;
	#endif
 #endif

 ret = usb_set_configuration(dev, dev->actconfig
 									->desc.bConfigurationValue);
 if (ret < 0)
   {
	dprintf("[USB] Error : failed to set active configuration (error=%d)", ret);
	return ret;
	}

	for (i = 0; i < dev->actconfig->desc.bNumInterfaces; i++) {
	struct usb_interface *intf = &dev->actconfig->interface[i];
		struct usb_interface_descriptor *as;

		as = &intf->altsetting[intf->act_altsetting].desc;
		ret = usb_set_interface(dev, as->bInterfaceNumber,
			as->bAlternateSetting);
	if (ret < 0) {
	#ifdef USB_DBGSTRING_ON
			err("failed to set active alternate setting "
				"for dev %s interface %d (error=%d)",
				dev->devpath, i, ret);
         #else
        	dprintf("[USBHUB] failed to set active alternate setting "
				"for interface %d (error=%d)",
				i, ret);
        #endif
			return ret;
		}
	}

 return 0;
}

