
/*
 * drivers/usb/usb.c
 *
 * (C) Copyright Linus Torvalds 1999
 * (C) Copyright Johannes Erdfelt 1999-2001
 * (C) Copyright Andreas Gal 1999
 * (C) Copyright Gregory P. Smith 1999
 * (C) Copyright Deti Fliegl 1999 (new USB architecture)
 * (C) Copyright Randy Dunlap 2000
 * (C) Copyright David Brownell 2000-2001 (kernel hotplug, usb_device_id,
 	more docs, etc)
 * (C) Copyright Yggdrasil Computing, Inc. 2000
 *     (usb_device_id matching changes by Adam J. Richter)
 * (C) Copyright Greg Kroah-Hartman 2002
 *
 * NOTE! This is not actually a driver at all, rather this is
 * just a collection of helper routines that implement the
 * generic USB things that the real drivers can use..
 *
 * Think of this as a "USB library" rather than anything else.
 * It should be considered a slave, with no callbacks. Callbacks
 * are evil.
 */
 
#ifdef _ETCPIP
   #ifdef _NEWLIB
     #include "string.h"
   #endif
	#include "feature.h"
    #define __KERNEL__
   	#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
          // for deciding whether to compile
          //  Debug Version or Release Version
    #include "etcpip.h"
    #include "UsbDbg.h"
	#include "UsbSysInit.h"
   
    #include "PoList.h" // osport_list.h
   
    #include "UsbCfg.h" // Config and Setting
    #include "UsOsport.h" // "usb_osport.h"
	#ifndef NO_TIMER
		#include "PoTimer.h" // osport_timer.h
	#endif // ndef NO_TIMER
    #ifndef NO_WAITQUEUE
		#include  "PoWait.h" // osport_wait.h
    #endif // ndef NO_WAITQUEUE
#endif // _ETCPIP
#ifndef _ETCPIP
   #include <linux/config.h>
#endif // ndef _ETCPIP

   #ifdef CONFIG_USB_DEBUG
      #define DEBUG
    #else
	  #undef DEBUG
   #endif

#ifndef _ETCPIP
   #include <linux/module.h>
   #include <linux/string.h>
   #include <linux/bitops.h>
   #include <linux/slab.h>
   #include <linux/interrupt.h>  /* for in_interrupt() */
   #include <linux/kmod.h>
   #include <linux/init.h>
   #include <linux/spinlock.h>
#endif // ndef _ETCPIP
#ifndef _ETCPIP
    #include <linux/errno.h>
 #else
	#include "PoErrno.h" // osport_errno.h
#endif
#ifndef _ETCPIP
    #include <linux/smp_lock.h>
#endif // ndef _ETCPIP
#ifndef _ETCPIP
   #include <linux/usb.h>
 #else
   #include "usb.h"
#endif
/*
#include <asm/io.h>
#include <asm/scatterlist.h>
#include <linux/mm.h>
#include <linux/dma-mapping.h>
*/

#include "UsbHcd.h"  // ../core/hcd.h
#include "usb.h"

#ifdef _ETCPIP
    #include "PoDevice.h" // osport_device.h
	#ifdef CONFIG_PCI
	    #include "UsbPoPci.h" // usb_osport_pci.h
    #endif	// CONFIG_PCI
#endif // def _ETCPIP

#ifdef USB_MEM_LIB
	#include "UsMemLib.h" // UsbMemLib.h
#endif // USB_MEM_LIB

#ifdef _USB_SYSLOG
	#include "syslog.h"
	#include "logStr.h"
	#include "config.h"
	extern struct III_Config_t gSetting;
  #else
	// #error: Please Make Sure : Not _USB_SYSLOG !!
	//#warning: Not _USB_SYSLOG !!
#endif // _SYSLOG

/*
#ifdef _ETCPIP
   struct usb_api_data awd;
#endif // def _ETCPIP
*/
/*
static const int usb_bandwidth_option =
#ifdef CONFIG_USB_BANDWIDTH
				1;
#else
				0;
#endif
*/

extern int  usb_hub_init(void);
extern void usb_hub_cleanup(void);
extern int usb_major_init(void);
extern void usb_major_cleanup(void);


int nousb;		/* Disable USB when built into kernel image */
			/* Not honored on modular build */


static int generic_probe (struct device *dev)
{
	return 0;
}
static int generic_remove (struct device *dev)
{
	return 0;
}

static struct device_driver usb_generic_driver = {
 #ifdef USB_DBGSTRING_ON
	.name =	"usb",
 #endif // USB_DBGSTRING_ON
 .bus = &usb_bus_type,
 .probe = generic_probe,
 .remove = generic_remove,
};

/* needs to be called with BKL held */
int usb_device_probe(struct device *dev)
{
	struct usb_interface * intf = to_usb_interface(dev);
	struct usb_driver * driver = to_usb_driver(dev->driver);
	const struct usb_device_id *id;
	int error = -ENODEV;

	dev_dbg(dev, "%s\n", __FUNCTION__);

	if (!driver->usbdev_probe)
		return error;

	if (!try_module_get(driver->owner)) {
		#ifdef USB_DBGSTRING_ON
			dev_err (dev, "Can't get a module reference for %s\n",
            	driver->name);
         #else
			dev_err (dev, "Can't get a module reference !!\n");
		#endif // USB_DBGSTRING_ON
		return error;
	}

	id = usb_match_id (intf, driver->id_table);
	if (id) {
		dev_dbg (dev, "%s - got id\n", __FUNCTION__);
    #ifndef NO_SEMPHORE
		down (&driver->serialize);
    #endif // ndef NO_SEMPHORE
		error = driver->usbdev_probe (intf, id);
    #ifndef NO_SEMPHORE
		up (&driver->serialize);
    #endif // ndef NO_SEMPHORE
	}
	if (!error)
		intf->driver = driver;

	module_put(driver->owner);

	return error;
}

int usb_device_remove(struct device *dev)
{
 struct usb_interface *intf;
 struct usb_driver *driver;

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_device_remove() Begin ..\n"));
 dprintf("[USB] usb_device_remove() Begin ..\n") ;
 
 intf = list_entry(dev,struct usb_interface,dev);
 driver = to_usb_driver(dev->driver);

 if (!driver)
   {
	dev_err(dev, "%s does not have a valid driver to work with!",
		    __FUNCTION__);
	return -ENODEV;
   }

 if (!try_module_get(driver->owner))
   {
	// FIXME this happens even when we just rmmod
	// drivers that aren't in active use...
	dev_err(dev, "Dieing driver still bound to device.\n");
	return -EIO;
   }

 /* if we sleep here on an umanaged driver 
  * the holder of the lock guards against 
  * module unload */
 #ifndef NO_SEMPHORE
	down(&driver->serialize);
 #endif // ndef NO_SEMPHORE

 if (intf->driver && intf->driver->disconnect)
   {
	dprintf("[USB] usb_device_remove() : intf->driver->disconnect(intf) ..\n") ;
	intf->driver->disconnect(intf);
   }

 /* if driver->disconnect didn't release the interface */
 if (intf->driver)
   {
	usb_driver_release_interface(driver, intf);
   }

 #ifndef NO_SEMPHORE
	up(&driver->serialize);
 #endif // ndef NO_SEMPHORE
 
 module_put(driver->owner);

 dprintf("[USB] usb_device_remove() End.\n") ;
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_device_remove() End.\n"));
 return 0;
}


#ifdef NO_SYS_FS
    
	// Prototypes for the device driver probing/loading functions
 
	static void usb_find_drivers(struct usb_device *);
	int  usb_find_interface_driver(struct usb_device *, unsigned int);
	static void usb_check_support(struct usb_device *);
    static void usb_scan_devices(struct usb_driver *new_driver) ;


	// We have a per-interface "registered driver" list.
 
	LIST_HEAD(usb_driver_list);
    /*
	LIST_HEAD(usb_bus_list); // in hcd.c
	*/
	#ifndef NO_SEMPHORE
	   struct semaphore usb_bus_list_lock;
	#endif // ndef NO_SEMPHORE

	#ifdef CONFIG_USB_DEVICEFS
	   devfs_handle_t usb_devfs_handle;	// /dev/usb dir. 
	#endif // CONFIG_USB_DEVICEFS
    /*
	static struct usb_busmap busmap;
    */
	static struct usb_driver *usb_minors[16];
    
#endif // NO_SYS_FS

/**
 *	usb_register - register a USB driver
 *	@new_driver: USB operations for the driver
 *
 *	Registers a USB driver with the USB core.  The list of unattached
 *	interfaces will be rescanned whenever a new driver is added, allowing
 *	the new driver to attach to any recognized devices.
 *	Returns a negative error code on failure and 0 on success.
 * 
 * NOTE: if you want your driver to use the USB major number, you must call
 * usb_register_dev() to enable that functionality.  This function no longer
 * takes care of that.
 */
int usb_register(struct usb_driver *new_driver)
{
 int retval = 0;


 USB_DBG_FLOW(iprintf("[USB FLOW] usb_register() Begin ..\n"));

 if (nousb)
   {
	return -ENODEV;
   }
   
 #ifdef USB_DBGSTRING_ON
	new_driver->driver.name = (char *)new_driver->name;
 #endif // USB_DBGSTRING_ON
 new_driver->driver.bus = &usb_bus_type;
 new_driver->driver.probe = usb_device_probe;
 new_driver->driver.remove = usb_device_remove;

 #ifdef NO_SYS_FS
 
 if (new_driver->fops != NULL)
   {
	 if (usb_minors[new_driver->minor/16])
       {
		#ifdef USB_DBGSTRING_ON
			dprintf("[USB] error registering %s driver \n",
            		new_driver->name);
         #else
			dprintf("[USB] error registering driver \n");
		#endif
		return -EINVAL;
	   }
	usb_minors[new_driver->minor/16] = new_driver;
   }
 
 #endif // NO_SYS_FS
 
 /*
 USB_DBG_INIT(iprintf("[USB INIT] registered new driver : [%s] \n",
         new_driver->name
         ));
 */
 
 #ifdef USB_DBGSTRING_ON
	USB_DBG_INIT(iprintf(
		"[USB INIT] usb_register() : registered new driver = %08XH "
    	"name = [%s] \n",
		new_driver , new_driver->name
   		));
  #else
	USB_DBG_INIT(iprintf(
		"[USB INIT] usb_register() : registered new driver = %08XH "
    	"\n",
		new_driver
   		));
 #endif

 #ifndef NO_SEMPHORE
	init_MUTEX(&new_driver->serialize);
 #endif // ndef NO_SEMPHORE

 #ifndef NO_SYS_FS
 	/*
	retval = driver_register(&new_driver->driver);
	if (!retval)
       {
		#ifdef USB_DBGSTRING_ON
			dprintf("[USB] registered new driver %s\n",
            		new_driver->name);
         #else
			dprintf("[USB] registered new driver ..\n");
		#endif
		usbfs_update_special();
	   }
     else
       {
		#ifdef USB_DBGSTRING_ON
			err("problem %d when registering driver %s",
				retval, new_driver->name);
         #else
			dprintf(
              "[USB] Error : problem %d when registering driver\n") ;
		#endif
	   }
	*/
  #else
	// Add it to the list of known drivers 

	usb_scan_devices(new_driver);
    
 #endif
 

 #ifdef USB_DBG_ABORT   
	// vvv for debug Prefetch/Data Abort Error !!
   
	#ifndef NO_TIMER
	    dprintf("[TEST] usb_register() : jiffies = %08XH\n",jiffies) ;
	#endif // ndef NO_TIMER
   
    {
     unsigned long * __temp_ptr = 0 ;
    
	 #if !defined(_ADM5120)
        dprintf("[TEST] usb_register() :  *000000000H = %08XH \n",
		        * __temp_ptr
                );
     #endif
    }
    dprintf("[TEST] usb_register() : GetTime() = %08XH\n",
                        GetTime()) ;
	if(!in_interrupt())
       {
		dprintf("[TEST] usb_register() : dlytsk()/YIELD() ..\n") ;
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] usb_register() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
       }
      else
       {
        dprintf("[TEST] usb_register() : in interrupt !!\n");
       }
   // ^^^ for debug Prefetch/Data Abort Error !!  
 #endif   
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_register() End.\n"));
 return retval;
}

/**
 *	usb_scan_devices - scans all unclaimed USB interfaces
 *
 *	Goes through all unclaimed USB interfaces, and offers them to all
 *	registered USB drivers through the 'probe' function.
 *	This will automatically be called after usb_register is called.
 *	It is called by some of the USB subsystems after one of their subdrivers
 *	are registered.
 */
#ifdef NO_SYS_FS

static void usb_scan_devices(struct usb_driver *new_driver)
{
 struct list_head *tmp;
 struct usb_bus *bus ;

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_scan_devices() Begin ..\n"));
 #ifndef NO_SPINLOCK
	down (&usb_bus_list_lock);
 #endif // ndef NO_SPINLOCK
 
 BLOCK_PREEMPTION;
 list_add_tail(&new_driver->driver_list, &usb_driver_list);
 
 tmp = usb_bus_list.next;
 
 while (tmp != &usb_bus_list)
   {
	// struct usb_bus *bus = list_entry(tmp,struct usb_bus, bus_list);
	
    USB_DBG_INIT(iprintf("[USB INIT] usb_scan_devices() : find bus ..\n"));
    
    bus = list_entry(tmp,struct usb_bus, bus_list);
    
    USB_DBG_INIT(iprintf("[USB INIT] usb_scan_devices() : bus = %08XH\n",bus));
    USB_DBG_INIT(iprintf("[USB INIT] usb_scan_devices() : bus->root_hub = %08XH\n",bus->root_hub));
	tmp = tmp->next;
	usb_check_support(bus->root_hub);
    
	UNBLOCK_PREEMPTION;
	#ifdef USB_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
		   ) ) {
			iprintf("[ASSERT][USB] usb_scan_devices() : " ) ;
    	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
   		   }
	#endif // USB_BS_ASSERT
	// 2003/7/17 B. S. Liaw BugFix    
	dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
	BLOCK_PREEMPTION;
    
    USB_DBG_INIT(iprintf("[USB INIT] usb_scan_devices() : find bus Next ..\n"));
   }
   
 USB_DBG_INIT(iprintf("[USB INIT] usb_scan_devices() : find bus End.\n"));
 
 UNBLOCK_PREEMPTION;
 
 #ifndef NO_SPINLOCK
	up (&usb_bus_list_lock);
 #endif // ndef NO_SPINLOCK
 
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_scan_devices() End.\n"));
 }
 
#endif // NO_SYS_FS
/*
 * This function is part of a depth-first search down the device tree,
 * removing any instances of a device driver.
 */
/*
static void usb_drivers_purge(struct usb_driver *driver,struct usb_device *dev)
{
	int i;

	if (!dev) {
		err("null device being purged!!!");
		return;
	}

	for (i=0; i<USB_MAXCHILDREN; i++)
		if (dev->children[i])
			usb_drivers_purge(driver, dev->children[i]);

	if (!dev->actconfig)
		return;
			
	for (i = 0; i < dev->actconfig->desc.bNumInterfaces; i++) {
		struct usb_interface *interface = &dev->actconfig->interface[i];
		
		if (interface->driver == driver) {
  #ifndef NO_SEMPHORE
			down(&driver->serialize);
  #endif // ndef NO_SEMPHORE
			driver->disconnect(dev, interface->private_data);
  #ifndef NO_SEMPHORE
			up(&driver->serialize);
  #endif // ndef NO_SEMPHORE
			// if driver->disconnect didn't release the interface 
			if (interface->driver)
				usb_driver_release_interface(driver, interface);
			
			//  This will go through the list looking for another
			//  driver that can handle the device
			 
			usb_find_interface_driver(dev, i);
		}
	}
}
*/

/**
 *	usb_deregister - unregister a USB driver
 *	@driver: USB operations of the driver to unregister
 * Context: !in_interrupt (), must be called with BKL held
 *
 *	Unlinks the specified driver from the internal USB driver list.
 * 
 * NOTE: If you called usb_register_dev(), you still need to call
 * usb_deregister_dev() to clean up your driver's allocated minor numbers,
 * this * call will no longer do it for you.
 */
void usb_deregister(struct usb_driver *driver)
{
 #ifndef _ETCPIP
	struct list_head *tmp;
 #endif // ndef _ETCPIP

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_deregister() Begin.\n"));
 #ifndef _ETCPIP
	#ifdef USB_DBGSTRING_ON
		dprintf("[USB] deregistering driver %s \n", driver->name);
      #else
		dprintf("[USB] deregistering driver ..\n");
    #endif
	driver_unregister (&driver->driver);

	usbfs_update_special();
 #endif // ndef _ETCPIP
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_deregister() End.\n"));
}

/**
 * usb_ifnum_to_if - get the interface object with a given interface number (usbcore-internal)
 * @dev: the device whose current configuration is considered
 * @ifnum: the desired interface
 *
 * This walks the device descriptor for the currently active configuration
 * and returns a pointer to the interface with that particular interface
 * number, or null.
 *
 * Note that configuration descriptors are not required to assign interface
 * numbers sequentially, so that it would be incorrect to assume that
 * the first interface in that descriptor corresponds to interface zero.
 * This routine helps device drivers avoid such mistakes.
 * However, you should make sure that you do the right thing with any
 * alternate settings available for this interfaces.
 */
struct usb_interface *usb_ifnum_to_if(struct usb_device *dev, unsigned ifnum)
{
	int i;

	for (i = 0; i < dev->actconfig->desc.bNumInterfaces; i++)
		if (dev->actconfig->interface[i].altsetting[0]
				.desc.bInterfaceNumber == ifnum)
			return &dev->actconfig->interface[i];

	return NULL;
}

/**
 * usb_epnum_to_ep_desc - get the endpoint object with a given endpoint number
 * @dev: the device whose current configuration is considered
 * @epnum: the desired endpoint
 *
 * This walks the device descriptor for the currently active configuration,
 * and returns a pointer to the endpoint with that particular endpoint
 * number, or null.
 *
 * Note that interface descriptors are not required to assign endpont
 * numbers sequentially, so that it would be incorrect to assume that
 * the first endpoint in that descriptor corresponds to interface zero.
 * This routine helps device drivers avoid such mistakes.
 */
struct usb_endpoint_descriptor *
usb_epnum_to_ep_desc(struct usb_device *dev, unsigned epnum)
{
	int i, j, k;

	for (i = 0; i < dev->actconfig->desc.bNumInterfaces; i++)
		for (j = 0; j < dev->actconfig->interface[i].num_altsetting; j++)
			for (k = 0; k < dev->actconfig->interface[i]
				.altsetting[j].desc.bNumEndpoints; k++)
				if (epnum == dev->actconfig->interface[i]
						.altsetting[j].endpoint[k]
						.desc.bEndpointAddress)
					return &dev->actconfig->interface[i]
						.altsetting[j].endpoint[k]
						.desc;

	return NULL;
}

/*
 * usb_calc_bus_time:
 *
 * returns (approximate) USB bus time in nanoseconds for a USB transaction.
 */
/*
static long usb_calc_bus_time (int low_speed, int input_dir, int isoc, int bytecount)
{
	unsigned long	tmp;

	if (low_speed)		// no isoc. here 
	{
		if (input_dir)
		{
			tmp = (67667L * (31L + 10L * BitTime (bytecount))) / 1000L;
			return (64060L + (2 * BW_HUB_LS_SETUP) + BW_HOST_DELAY + tmp);
		}
		else
		{
			tmp = (66700L * (31L + 10L * BitTime (bytecount))) / 1000L;
			return (64107L + (2 * BW_HUB_LS_SETUP) + BW_HOST_DELAY + tmp);
		}
	}

	// for full-speed: 

	if (!isoc)		// Input or Output 
	{
		tmp = (8354L * (31L + 10L * BitTime (bytecount))) / 1000L;
		return (9107L + BW_HOST_DELAY + tmp);
	} // end not Isoc 

	// for isoc: 

	tmp = (8354L * (31L + 10L * BitTime (bytecount))) / 1000L;
	return (((input_dir) ? 7268L : 6265L) + BW_HOST_DELAY + tmp);
}

/*
 * usb_check_bandwidth():
 *
 * old_alloc is from host_controller->bandwidth_allocated in microseconds;
 * bustime is from calc_bus_time(), but converted to microseconds.
 *
 * returns <bustime in us> if successful,
 * or USB_ST_BANDWIDTH_ERROR if bandwidth request fails.
 *
 * FIXME:
 * This initial implementation does not use Endpoint.bInterval
 * in managing bandwidth allocation.
 * It probably needs to be expanded to use Endpoint.bInterval.
 * This can be done as a later enhancement (correction).
 * This will also probably require some kind of
 * frame allocation tracking...meaning, for example,
 * that if multiple drivers request interrupts every 10 USB frames,
 * they don't all have to be allocated at
 * frame numbers N, N+10, N+20, etc.  Some of them could be at
 * N+11, N+21, N+31, etc., and others at
 * N+12, N+22, N+32, etc.
 * However, this first cut at USB bandwidth allocation does not
 * contain any frame allocation tracking.
 */
/*
int usb_check_bandwidth (struct usb_device *dev, struct urb *urb)
{
	int		new_alloc;
	int		old_alloc = dev->bus->bandwidth_allocated;
	unsigned int	pipe = urb->pipe;
	long		bustime;

	bustime = usb_calc_bus_time (usb_pipeslow(pipe), usb_pipein(pipe),
			usb_pipeisoc(pipe), usb_maxpacket(dev, pipe, usb_pipeout(pipe)));
	if (usb_pipeisoc(pipe))
		bustime = NS_TO_US(bustime) / urb->number_of_packets;
	else
		bustime = NS_TO_US(bustime);

	new_alloc = old_alloc + (int)bustime;
		// what new total allocated bus time would be 

	if (new_alloc > FRAME_TIME_MAX_USECS_ALLOC)
		dbg("usb-check-bandwidth %sFAILED: was %u, would be %u, bustime = %ld us",
			usb_bandwidth_option ? "" : "would have ",
			old_alloc, new_alloc, bustime);

	if (!usb_bandwidth_option)	// don't enforce it 
		return (bustime);
	return (new_alloc <= FRAME_TIME_MAX_USECS_ALLOC) ? bustime : USB_ST_BANDWIDTH_ERROR;
}
*/
/*
void usb_claim_bandwidth (struct usb_device *dev, struct urb *urb, int bustime, int isoc)
{
	dev->bus->bandwidth_allocated += bustime;
	if (isoc)
		dev->bus->bandwidth_isoc_reqs++;
	else
		dev->bus->bandwidth_int_reqs++;
	urb->bandwidth = bustime;

#ifdef USB_BANDWIDTH_MESSAGES
	dbg("bandwidth alloc increased by %d to %d for %d requesters",
		bustime,
		dev->bus->bandwidth_allocated,
		dev->bus->bandwidth_int_reqs + dev->bus->bandwidth_isoc_reqs);
#endif
}
*/

/*
 * usb_release_bandwidth():
 *
 * called to release a pipe's bandwidth (in microseconds)
 */
/*
void usb_release_bandwidth(struct usb_device *dev, struct urb *urb, int isoc)
{
	dev->bus->bandwidth_allocated -= urb->bandwidth;
	if (isoc)
		dev->bus->bandwidth_isoc_reqs--;
	else
		dev->bus->bandwidth_int_reqs--;

#ifdef USB_BANDWIDTH_MESSAGES
	dbg("bandwidth alloc reduced by %d to %d for %d requesters",
		urb->bandwidth,
		dev->bus->bandwidth_allocated,
		dev->bus->bandwidth_int_reqs + dev->bus->bandwidth_isoc_reqs);
#endif
	urb->bandwidth = 0;
}

static void usb_bus_get(struct usb_bus *bus)
{
 atomic_inc(&bus->refcnt);
}

static void usb_bus_put(struct usb_bus *bus)
{
	if (atomic_dec_and_test(&bus->refcnt))
		kfree(bus);
}
*/

/**
 *	usb_alloc_bus - creates a new USB host controller structure
 *	@op: pointer to a struct usb_operations that this bus structure should use
 *
 *	Creates a USB host controller bus structure with the specified 
 *	usb_operations and initializes all the necessary internal objects.
 *	(For use only by USB Host Controller Drivers.)
 *
 *	If no memory is available, NULL is returned.
 *
 *	The caller should call usb_free_bus() when it is finished with the structure.
 */
/*
struct usb_bus *usb_alloc_bus(struct usb_operations *op)
{
 struct usb_bus *bus;

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_alloc_bus() Begin ..\n"));
 #ifndef _ETCPIP
	bus = kmalloc(sizeof(*bus), GFP_KERNEL);
	if (!bus)
		return NULL;
  #else
	// bus = NonCache_contigmalloc(sizeof(*bus)) ;
    
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    BLOCK_PREEMPTION;

	bus = alloc_mem( sizeof(*bus) ) ;
                 
    UNBLOCK_PREEMPTION;

    USB_DBG_INIT(iprintf("[USB INIT] bus = %08XH \n",bus));

    if ( bus == NULL )
       {
		dprintf("couldn't allocate memory \n");
		return NULL;
	   }
 #endif

 #ifndef _ETCPIP
	memset(&bus->devmap, 0, sizeof(struct usb_devmap));
 #endif
 
 #ifdef DEVNUM_ROUND_ROBIN
	bus->devnum_next = 1;
 #endif // DEVNUM_ROUND_ROBIN 

 bus->op = op;
 
 USB_DBG_INIT(iprintf(
    "[USB INIT] usb_alloc_bus() : bus->op = %08XH\n",
    bus->op
    ));
    
 USB_DBG_INIT(iprintf(
    "[USB INIT] usb_alloc_bus() : bus->op->deallocate() = %08XH\n",
    bus->op->deallocate
    ));
 
 bus->root_hub = NULL;
	bus->hcpriv = NULL;
	bus->busnum = -1;
	bus->bandwidth_allocated = 0;
	bus->bandwidth_int_reqs  = 0;
	bus->bandwidth_isoc_reqs = 0;

	INIT_LIST_HEAD(&bus->bus_list);
	INIT_LIST_HEAD(&bus->inodes);

	atomic_set(&bus->refcnt, 1);

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_alloc_bus() End.\n"));
	return bus;
}
*/
/**
 *	usb_free_bus - frees the memory used by a bus structure
 *	@bus: pointer to the bus to free
 *
 *	(For use only by USB Host Controller Drivers.)
 */
/*
void usb_free_bus(struct usb_bus *bus)
{
	if (!bus)
		return;

	usb_bus_put(bus);
}
*/
/**
 *	usb_register_bus - registers the USB host controller with the usb core
 *	@bus: pointer to the bus to register
 *
 *	(For use only by USB Host Controller Drivers.)
 */
/*
void usb_register_bus(struct usb_bus *bus)
{
 int busnum;


 USB_DBG_FLOW(iprintf("[USB FLOW] usb_register_bus() Begin .. \n"));
 
 #ifndef NO_SPINLOCK
	down (&usb_bus_list_lock);
 #endif // ndef NO_SPINLOCK
 
 busnum = find_next_zero_bit(busmap.busmap, USB_MAXBUS, 1);
 if (busnum < USB_MAXBUS)
    {
		set_bit(busnum, busmap.busmap);
	 bus->busnum = busnum;
     USB_DBG_INIT(iprintf(
        "[USB INIT] usb_register_bus() : Set Value bus->busnum = %d\n",
        bus->busnum
        ));
	}
   else
    {
	 dprintf("[USB] Error : too many buses \n");
    }
    
 usb_bus_get(bus);

 // Add it to the list of buses 
	list_add(&bus->bus_list, &usb_bus_list);
 #ifndef NO_SPINLOCK
	up (&usb_bus_list_lock);
 #endif // ndef NO_SPINLOCK

 #ifdef CONFIG_USB_DEVICEFS
	usbdevfs_add_bus(bus);
 #endif // CONFIG_USB_DEVICEFS

 USB_DBG_INFO(iprintf(
    "[USB INFO] new USB bus registered, assigned bus number %d\n",
    bus->busnum
    ));
    
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_register_bus() End.\n"));
}
*/

/**
 *	usb_deregister_bus - deregisters the USB host controller
 *	@bus: pointer to the bus to deregister
 *
 *	(For use only by USB Host Controller Drivers.)
 */
/*
void usb_deregister_bus(struct usb_bus *bus)
{
 dprintf("[USB] bus %d deregistered \n", bus->busnum);

	//
	// * NOTE: make sure that all the devices are removed by the
	// * controller code, as well as having it call this when cleaning
	// * itself up
	// 
#ifndef NO_SPINLOCK
	down (&usb_bus_list_lock);
#endif // ndef NO_SPINLOCK
	list_del(&bus->bus_list);
#ifndef NO_SPINLOCK
	up (&usb_bus_list_lock);
#endif // ndef NO_SPINLOCK

#ifdef CONFIG_USB_DEVICEFS
	usbdevfs_remove_bus(bus);
#endif // CONFIG_USB_DEVICEFS

	clear_bit(bus->busnum, busmap.busmap);

	usb_bus_put(bus);
}
*/

/*
 * This function is for doing a depth-first search for devices which
 * have support, for dynamic loading of driver modules.
 */

#ifdef NO_SYS_FS
static void usb_check_support(struct usb_device *dev)
{
 int i;

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_check_support() Begin ..\n"));
 if (!dev)
    {
	 dprintf("[USB] null device being checked !!!\n");
     USB_DBG_FLOW(iprintf("[USB FLOW] usb_check_support() End with Fail !!\n"));
	 return;
	}

 for (i=0; i<USB_MAXCHILDREN; i++)
     {
	  if (dev->children[i])
         {
	      USB_DBG_INIT(iprintf(
          	"[USB INIT] usb_check_support() : children ..\n"
            ));
		  usb_check_support(dev->children[i]);
         }
     }
     
 if (!dev->actconfig)
   {
	USB_DBG_INIT(iprintf(
    	"[USB INIT] Not actconfig !!!\n"
        ));
     USB_DBG_FLOW(iprintf("[USB FLOW] usb_check_support() End with Fail !!\n"));
	 return;
    }
    
 // now we check this device 
 if (dev->devnum > 0)
   {
	USB_DBG_INIT(iprintf(
    	"[USB INIT] usb_check_support() : "
    	"dev->bus->busnum = %d , dev->devnum =  %d\n" ,
     	dev->bus->busnum , dev->devnum
    	));
     /*
     USB_DBG_INIT(iprintf(
        "[USB INIT] usb_check_support() : dev->devnum = %d \n",
        dev->devnum));
     */   
     USB_DBG_INIT(iprintf(
        "[USB INIT] usb_check_support() : "
        "dev->actconfig->desc.bNumInterfaces = %d \n",
        dev->actconfig->desc.bNumInterfaces));
        
	 for (i = 0; i < dev->actconfig->desc.bNumInterfaces; i++)
         {
		  usb_find_interface_driver(dev, i);
         }
    }
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_check_support() End.\n"));
}
#endif // NO_SYS_FS


/**
 * usb_driver_claim_interface - bind a driver to an interface
 * @driver: the driver to be bound
 * @iface: the interface to which it will be bound
 * @priv: driver data associated with that interface
 *
 * This is used by usb device drivers that need to claim more than one
 * interface on a device when probing (audio and acm are current examples).
 * No device driver should directly modify internal usb_interface or
 * usb_device structure members.
 *
 * Few drivers should need to use this routine, since the most natural
 * way to bind to an interface is to return the private data from
 * the driver's probe() method.  Any driver that does use this must
 * first be sure that no other driver has claimed the interface, by
 * checking with usb_interface_claimed().
 */
/*
void usb_driver_claim_interface(struct usb_driver *driver, 
					struct usb_interface *iface, void* priv)
*/
void usb_driver_claim_interface(struct usb_driver *driver, 
					struct usb_interface *iface)
{
 if (!iface || !driver)
		return;

 // FIXME change API to report an error in this case
 if (iface->driver)
   {
	#ifdef USB_DBGSTRING_ON
		err ("%s driver booted %s off interface %p",
	    	driver->name, iface->driver->name, iface);
     #else
		dprintf("[USB] Error : driver booted off interface %p \n",
	    	 iface);
	#endif 
   }
  else
   {
	#ifndef _ETCPIP 
	    dbg("%s driver claimed interface %p", driver->name, iface);
	 #else  
		#ifdef USB_DBGSTRING_ON
	    	USB_DBG_INIT(iprintf(
	       		"[USB INIT] driver = %08XH name = [%s] claim interface %08XH \n",
	       		driver , driver->name, iface));
	    	/*
			USB_DBG_INIT(iprintf(
			"[USB INIT] usb_register() : registered new driver = %08XH "
			"name = [%s] \n",
			new_driver , new_driver->name
		   	));
	        */
	      #else
	    	USB_DBG_INIT(iprintf(
	       		"[USB INIT] driver=%08XH claim interface %08XH \n",
	       		driver , iface));
		#endif 
	#endif
   }
 
 USB_DBG_INIT(iprintf("[USB INIT] usb_driver_claim_interface() : intf=%08XH \n",iface )) ;
 iface->driver = driver;
 /*
 USB_DBG_INIT(iprintf("[USB INIT] usb_driver_claim_interface() : intf=%08XH , priv=%08XH \n",iface, priv )) ;
 usb_set_intfdata(iface, priv);
 */
 
} /* usb_driver_claim_interface() */

/**
 * usb_interface_claimed - returns true iff an interface is claimed
 * @iface: the interface being checked
 *
 * This should be used by drivers to check other interfaces to see if
 * they are available or not.  If another driver has claimed the interface,
 * they may not claim it.  Otherwise it's OK to claim it using
 * usb_driver_claim_interface().
 *
 * Returns true (nonzero) iff the interface is claimed, else false (zero).
 */
int usb_interface_claimed(struct usb_interface *iface)
{
 USB_DBG_FLOW(iprintf(
	"[USB FLOW] usb_interface_claimed() Begin ..\n"));
 USB_DBG_MEM(iprintf("[USB MEM] iface = %08XH \n",
     iface ));
 USB_DBG_INIT(iprintf("[USB INIT]   *iface = %08XH \n",
     *iface ));
     
 if (!iface)
    {
     USB_DBG_INIT(iprintf("[USB INIT] No iface \n")) ;
	 return 0;
    }

 USB_DBG_INIT(iprintf("[USB INIT] &(iface->driver) = %08XH \n",
     &(iface->driver) ));
 USB_DBG_INIT(iprintf("[USB INIT] iface->driver = %08XH \n",
     iface->driver ));
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_interface_claimed() End with chek driver.\n"));
 return (iface->driver != NULL);
} /* usb_interface_claimed() */

/**
 * usb_driver_release_interface - unbind a driver from an interface
 * @driver: the driver to be unbound
 * @iface: the interface from which it will be unbound
 * 
 * This should be used by drivers to release their claimed interfaces.
 * It is normally called in their disconnect() methods, and only for
 * drivers that bound to more than one interface in their probe().
 *
 * When the USB subsystem disconnect()s a driver from some interface,
 * it automatically invokes this method for that interface.  That
 * means that even drivers that used usb_driver_claim_interface()
 * usually won't need to call this.
 */
void usb_driver_release_interface(struct usb_driver *driver, struct usb_interface *iface)
{
 /* this should never happen, don't release something that's not ours */
 if (!iface || iface->driver != driver)
   {
	return;
   }
 iface->driver = NULL;
 
 USB_DBG_INIT(iprintf("[USB INIT] usb_driver_release_interface() : intf=%08XH , Set NULL ..\n",iface )) ;
 usb_set_intfdata(iface, NULL);
}


/**
 * usb_match_id - find first usb_device_id matching device or interface
 * @dev: the device whose descriptors are considered when matching
 * @interface: the interface of interest
 * @id: array of usb_device_id structures, terminated by zero entry
 *
 * usb_match_id searches an array of usb_device_id's and returns
 * the first one matching the device or interface, or null.
 * This is used when binding (or rebinding) a driver to an interface.
 * Most USB device drivers will use this indirectly, through the usb core,
 * but some layered driver frameworks use it directly.
 * These device tables are exported with MODULE_DEVICE_TABLE, through
 * modutils and "modules.usbmap", to support the driver loading
 * functionality of USB hotplugging.
 *
 * What Matches:
 *
 * The "match_flags" element in a usb_device_id controls which
 * members are used.  If the corresponding bit is set, the
 * value in the device_id must match its corresponding member
 * in the device or interface descriptor, or else the device_id
 * does not match.
 *
 * "driver_info" is normally used only by device drivers,
 * but you can create a wildcard "matches anything" usb_device_id
 * as a driver's "modules.usbmap" entry if you provide an id with
 * only a nonzero "driver_info" field.  If you do this, the USB device
 * driver's probe() routine should use additional intelligence to
 * decide whether to bind to the specified interface.
 * 
 * What Makes Good usb_device_id Tables:
 *
 * The match algorithm is very simple, so that intelligence in
 * driver selection must come from smart driver id records.
 * Unless you have good reasons to use another selection policy,
 * provide match elements only in related groups, and order match
 * specifiers from specific to general.  Use the macros provided
 * for that purpose if you can.
 *
 * The most specific match specifiers use device descriptor
 * data.  These are commonly used with product-specific matches;
 * the USB_DEVICE macro lets you provide vendor and product IDs,
 * and you can also match against ranges of product revisions.
 * These are widely used for devices with application or vendor
 * specific bDeviceClass values.
 *
 * Matches based on device class/subclass/protocol specifications
 * are slightly more general; use the USB_DEVICE_INFO macro, or
 * its siblings.  These are used with single-function devices
 * where bDeviceClass doesn't specify that each interface has
 * its own class. 
 *
 * Matches based on interface class/subclass/protocol are the
 * most general; they let drivers bind to any interface on a
 * multiple-function device.  Use the USB_INTERFACE_INFO
 * macro, or its siblings, to match class-per-interface style 
 * devices (as recorded in bDeviceClass).
 *  
 * Within those groups, remember that not all combinations are
 * meaningful.  For example, don't give a product version range
 * without vendor and product IDs; or specify a protocol without
 * its associated class and subclass.
 */   
/*
#ifndef NO_SYS_FS
*/
const struct usb_device_id *
usb_match_id(struct usb_interface *interface, const struct usb_device_id *id)
/*
#else
const struct usb_device_id *
usb_match_id(struct usb_device *dev, struct usb_interface *interface,
	     const struct usb_device_id *id)
#endif
*/
{
 /*
 #ifndef NO_SYS_FS
 */
	struct usb_host_interface *intf;
 /*
 #endif
 */
 
 /*
 #ifndef NO_SYS_FS
 */
 struct usb_device *dev;
    
 #ifdef NO_SYS_FS
    /*
	struct usb_interface_descriptor	*intf = 0;
    */
 #endif


 /* proc_connectinfo in devio.c may call us with id == NULL. */
 if (id == NULL)
		return NULL;
 /*
 #ifndef NO_SYS_FS
 */
	intf = &interface->altsetting [interface->act_altsetting];
 /*
 #endif
 */
 
 dev = interface_to_usbdev(interface);

	/* It is important to check that id->driver_info is nonzero,
	   since an entry that is all zeroes except for a nonzero
	   id->driver_info is the way to create an entry that
	   indicates that the driver want to examine every
	   device and interface. */
	for (; id->idVendor || id->bDeviceClass || id->bInterfaceClass ||
	       id->driver_info; id++) {

		if ((id->match_flags & USB_DEVICE_ID_MATCH_VENDOR) &&
		    id->idVendor != dev->descriptor.idVendor)
			continue;

		if ((id->match_flags & USB_DEVICE_ID_MATCH_PRODUCT) &&
		    id->idProduct != dev->descriptor.idProduct)
			continue;

		/* No need to test id->bcdDevice_lo != 0, since 0 is never
		   greater than any unsigned number. */
		if ((id->match_flags & USB_DEVICE_ID_MATCH_DEV_LO) &&
		    (id->bcdDevice_lo > dev->descriptor.bcdDevice))
			continue;

		if ((id->match_flags & USB_DEVICE_ID_MATCH_DEV_HI) &&
		    (id->bcdDevice_hi < dev->descriptor.bcdDevice))
			continue;

		if ((id->match_flags & USB_DEVICE_ID_MATCH_DEV_CLASS) &&
		    (id->bDeviceClass != dev->descriptor.bDeviceClass))
			continue;

		if ((id->match_flags & USB_DEVICE_ID_MATCH_DEV_SUBCLASS) &&
		    (id->bDeviceSubClass!= dev->descriptor.bDeviceSubClass))
			continue;

		if ((id->match_flags & USB_DEVICE_ID_MATCH_DEV_PROTOCOL) &&
		    (id->bDeviceProtocol != dev->descriptor.bDeviceProtocol))
			continue;

 #ifdef NO_SYS_FS
        /*
		intf = &interface->altsetting [interface->act_altsetting];
        */
 #endif
		if ((id->match_flags & USB_DEVICE_ID_MATCH_INT_CLASS) &&
		    (id->bInterfaceClass != intf->desc.bInterfaceClass))
			continue;

		if ((id->match_flags & USB_DEVICE_ID_MATCH_INT_SUBCLASS) &&
		    (id->bInterfaceSubClass != intf->desc.bInterfaceSubClass))
		    continue;

		if ((id->match_flags & USB_DEVICE_ID_MATCH_INT_PROTOCOL) &&
		    (id->bInterfaceProtocol != intf->desc.bInterfaceProtocol))
		    continue;

		return id;
	}

	return NULL;
}

/**
 * usb_find_interface - find usb_interface pointer for driver and device
 * @drv: the driver whose current configuration is considered
 * @kdev: the desired device
 *
 * This walks the driver device list and returns a pointer to the interface 
 * with the matching kdev_t.
 */
struct usb_interface *usb_find_interface(struct usb_driver *drv, kdev_t kdev)
{
 struct list_head *entry;
 struct device *dev;
 struct usb_interface *intf;

 list_for_each(entry, &drv->driver.devices)
   {
	dev = container_of(entry, struct device, driver_list);

	/* can't look at usb devices, only interfaces */
	if (dev->driver == &usb_generic_driver)
			continue;

	intf = to_usb_interface(dev);
	if (!intf)
			continue;

	if (kdev_same(intf->kdev,kdev))
       {
		return intf;
	   }
   }

 /* no device found that matches */
 return NULL;	
}
/*
 * This entrypoint gets called for each new device.
 *
 * We now walk the list of registered USB drivers,
 * looking for one that will accept this interface.
 *
 * "New Style" drivers use a table describing the devices and interfaces
 * they handle.  Those tables are available to user mode tools deciding
 * whether to load driver modules for a new device.
 *
 * The probe return value is changed to be a private pointer.  This way
 * the drivers don't have to dig around in our structures to set the
 * private pointer if they only need one interface. 
 *
 * Returns: 0 if a driver accepted the interface, -1 otherwise
 */

#ifdef NO_SYS_FS
   int usb_find_interface_driver(struct usb_device *dev, unsigned ifnum)
   {
	struct list_head *tmp;
	struct usb_interface *interface;
	
	void *private;
	// 2003/7/8 B. S. Liaw : private is 
    //          usb_hub/usb_lpt/usb_mouse .. etc.
    // Now not used and is set by each driver usb_set_intfdata() ..
	/*
	int private;
	*/
	const struct usb_device_id *id;
	struct usb_driver *driver;
	int i;
 
	USB_DBG_FLOW(iprintf("[USB FLOW] usb_find_interface_driver() Begin ..\n"));
	USB_DBG_INIT(iprintf(
    	"[USB INIT] usb_find_interface_driver() : ifnum = %d \n",
		ifnum ));
	USB_DBG_INIT(iprintf(
    	"[USB INIT] usb_find_interface_driver() : "
	    "dev->actconfig->desc.bNumInterfaces = %d \n",
	    dev->actconfig->desc.bNumInterfaces));
	if ((!dev) || (ifnum >= dev->actconfig->desc.bNumInterfaces))
	   {
		dprintf("[USB] bad find_interface_driver params");
     	USB_DBG_FLOW(iprintf(
	        "[USB FLOW] usb_find_interface_driver() End with Fail : -1\n"
	        ));
		return -1;
	   }
       
	#ifndef NO_SEMPHORE
		down(&dev->serialize);
	#endif // ndef NO_SEMPHORE
	interface = dev->actconfig->interface + ifnum;
	USB_DBG_INIT(iprintf(
	    "[USB INIT] usb_find_interface_driver() : interface = %08XH \n",
	    interface
        ));
	if (usb_interface_claimed(interface))
       {
     	USB_DBG_INIT(iprintf("[USB INIT] interface_claimed\n"));
		goto out_err;
       }
	/*
	private = NULL;
	*/
	
	private = 0 ; // B. S. Liaw : 0	: OK
	// 2003/7/8 B. S. Liaw : private is 
    //          usb_hub/usb_lpt/usb_mouse .. etc.
    // Now not used and is set by each driver usb_set_intfdata() ..
	/*
	private = -ENODEV ; // B. S. Liaw : Not found
    */
	for (tmp = usb_driver_list.next; tmp != &usb_driver_list;)
	   {
		driver = list_entry(tmp, struct usb_driver, driver_list);
		#ifdef USB_DBGSTRING_ON
			USB_DBG_INIT(iprintf(
    			"[USB INIT] usb_find_drivers() : driver = %08XH "
	            "driver->name = [%s] \n",
    			driver , driver->name
	    		));
		 #else
			USB_DBG_INIT(iprintf(
    			"[USB INIT] usb_find_drivers() : driver = %08XH "
        	    "\n",
	    		driver
		    	));
		#endif
		tmp = tmp->next;
        /* 2003/7/8 B. S. Liaw : why new style driver now cannot work ??
		id = driver->id_table;
		// new style driver? 
		if (id)
    	   {
	        USB_DBG_INIT(iprintf("[USB INIT] new style driver \n"));
			for (i = 0; i < interface->num_altsetting; i++)
        	   {
				interface->act_altsetting = i;
	            
				// id = usb_match_id(dev, interface, id);
	            
				id = usb_match_id(interface, id);
				if (id)
        	       {
            	    #ifndef NO_SEMPHORE
						down(&driver->serialize);
	         		#endif // ndef NO_SEMPHORE
    	            USB_DBG_INIT(iprintf(
        	        	"[USB INIT] usb_find_interface_driver() : "
            	        "driver probe .. \n"));
	                
					// private = driver->probe(dev,ifnum,id);
        	        
					private = driver->usbdev_probe(interface,id);
                
	  				#ifndef NO_SEMPHORE
						up(&driver->serialize);
  					#endif // ndef NO_SEMPHORE
	                
					// if (!error)
					// 	intf->driver = driver;
            	    
	                
					// if (private != NULL)
        	        
					if (private != 0)
                       {
    	            	USB_DBG_INIT(iprintf(
        	        		"[USB INIT] usb_find_interface_driver() : "
            	        	"driver probe OK. \n"));
						break;
                       }
				   } // if id
			   } // for i
           
			// if driver not bound, leave defaults unchanged 
			// if (private == NULL)
			if (private == 0)
    	       {
        	    USB_DBG_INIT(iprintf("[USB INIT] driver not bound !! \n"));
				interface->act_altsetting = 0;
    	       }
		   }
    	  else
       	*/
	       { 
    	    // "old style" driver 
	        USB_DBG_INIT(iprintf("[USB INIT] old style driver \n"));
  			#ifndef NO_SEMPHORE
				down(&driver->serialize);
	  		#endif // ndef NO_SEMPHORE
    	    USB_DBG_INIT(iprintf(
        		"[USB INIT] usb_find_interface_driver() : driver probe .. \n"));
	        /*
			private = driver->probe(dev, ifnum, NULL);
        	*/
			private = (void *)(driver->usbdev_probe(interface,NULL)); // rerurn : 0 : ok
			// 2003/7/8 B. S. Liaw : private is 
            //          usb_hub/usb_lpt/usb_mouse .. etc.
     // Now not used and is set by each driver usb_set_intfdata() ..
        
			if (private != 0)
               {
    	       	USB_DBG_INIT(iprintf(
        	   		"[USB INIT] usb_find_interface_driver() : "
                   	"driver probe OK. \n"));
               }
  			#ifndef NO_SEMPHORE
				up(&driver->serialize);
	  		#endif // ndef NO_SEMPHORE
		   }
	// probe() may have changed the config on us 
	interface = dev->actconfig->interface + ifnum;
    
	// if (private)   // 2003/7/8 B. S. Liaw : private is driver 
                   //          usb_hub/usb_lpt/usb_mouse .. etc.
     // Now not used and is set by each driver usb_set_intfdata() ..
	if (private == 0) // B. S. Liaw : 0	: OK
       {
        /*
		USB_DBG_INIT(iprintf("[USB INIT] usb_find_interface_driver() : intf=%08XH , priv=%08XH \n",interface, private )) ;
		usb_driver_claim_interface(driver, interface, (void*)private);
        */
		USB_DBG_INIT(iprintf("[USB INIT] usb_find_interface_driver() : intf=%08XH \n",interface)) ;
		usb_driver_claim_interface(driver, interface );

 #ifdef USB_DBG_ABORT   
   // vvv for debug Prefetch/Data Abort Error !!
   
	#ifndef NO_TIMER
	   dprintf("[TEST] usb_find_interface_driver() : jiffies = %08XH\n",jiffies) ;
	#endif // ndef NO_TIMER
   
   {
    unsigned long * __temp_ptr = 0 ;
    
	#if !defined(_ADM5120)
        dprintf("[TEST] usb_find_interface_driver() :  *000000000H = %08XH \n",
		        * __temp_ptr
                );
    #endif
   }
   dprintf("[TEST] usb_find_interface_driver() : GetTime() = %08XH\n",
                        GetTime()) ;
   if(!in_interrupt())
     {
          dprintf("[TEST] usb_find_interface_driver() : dlytsk()/YIELD() ..\n") ;
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] usb_find_interface_driver() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
			#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
				// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
             #else
            	YIELD();
            #endif
     }
    else
     {
      dprintf("[TEST] usb_find_interface_driver() : in interrupt !!\n");
     }
   // ^^^ for debug Prefetch/Data Abort Error !!  
 #endif   
        
  		#ifndef NO_SEMPHORE
			up(&dev->serialize);
  		#endif // ndef NO_SEMPHORE
        USB_DBG_FLOW(iprintf(
        	"[USB FLOW] usb_find_interface_driver() End.\n"));
		return 0;
	   }
   } // for tmp
   
	// --------------
	out_err:
	// --------------
 
	#ifndef NO_SEMPHORE
    /*
	  up(&dev->serialize);
    */
	#endif // ndef NO_SEMPHORE
	USB_DBG_FLOW(iprintf(
       "[USB FLOW] usb_find_interface_driver() End with Fail : -1\n"));
	return -1;
   }
#endif // NO_SYS_FS

static int usb_device_match (struct device *dev, struct device_driver *drv)
{
	struct usb_interface *intf;
	struct usb_driver *usb_drv;
	const struct usb_device_id *id;

	/* check for generic driver, which we don't match any device with */
	if (drv == &usb_generic_driver)
		return 0;

	intf = to_usb_interface(dev);

	usb_drv = to_usb_driver(drv);
	id = usb_drv->id_table;
	
	id = usb_match_id (intf, usb_drv->id_table);
	if (id)
		return 1;

	return 0;
}


#ifdef	CONFIG_HOTPLUG

/*
 * USB hotplugging invokes what /proc/sys/kernel/hotplug says
 * (normally /sbin/hotplug) when USB devices get added or removed.
 *
 * This invokes a user mode policy agent, typically helping to load driver
 * or other modules, configure the device, and more.  Drivers can provide
 * a MODULE_DEVICE_TABLE to help with module loading subtasks.
 *
 * We're called either from khubd (the typical case) or from root hub
 * (init, kapmd, modprobe, rmmod, etc), but the agents need to handle
 * delays in event delivery.  Use sysfs (and DEVPATH) to make sure the
 * device (and this configuration!) are still present.

 * Some synchronization is important: removes can't start processing
 * before the add-device processing completes, and vice versa.  That keeps
 * a stack of USB-related identifiers stable while they're in use.  If we
 * know that agents won't complete after they return (such as by forking
 * a process that completes later), it's enough to just waitpid() for the
 * agent -- as is currently done.
 *
 * The reason: we know we're called either from khubd (the typical case)
 * or from root hub initialization (init, kapmd, modprobe, etc).  In both
 * cases, we know no other thread can recycle our address, since we must
 * already have been serialized enough to prevent that.
 */
static int usb_hotplug (struct device *dev, char **envp, int num_envp,
			char *buffer, int buffer_size)
{
	struct usb_interface *intf;
	struct usb_device *usb_dev;
	char *scratch;
	int i = 0;
	int length = 0;

	dbg ("%s", __FUNCTION__);
        
	if (!dev)
		return -ENODEV;
    
	if (dev->driver == &usb_generic_driver)
		return 0;
 
	intf = to_usb_interface(dev);
	if (!intf)
		return -ENODEV;
    
	usb_dev = interface_to_usbdev (intf);
	if (!usb_dev)
		return -ENODEV;

	if (usb_dev->devnum < 0) {
		dbg ("device already deleted ??");
		return -ENODEV;
	   }
	if (!usb_dev->bus) {
		dbg ("bus already removed?");
		return -ENODEV;
	}
    
	scratch = buffer;

#ifdef	CONFIG_USB_DEVICEFS
	/* If this is available, userspace programs can directly read
	 * all the device descriptors we don't tell them about.  Or
	 * even act as usermode drivers.
	 *
	 * FIXME reduce hardwired intelligence here
	 */
	envp [i++] = scratch;
 #ifndef USB_NO_LIB
	length += snprintf (scratch, buffer_size - length,
			    "DEVICE=/proc/bus/usb/%03d/%03d",
			    usb_dev->bus->busnum, usb_dev->devnum);
 #endif // ndef USB_NO_LIB
	if ((buffer_size - length <= 0) || (i >= num_envp))
		return -ENOMEM;
	++length;
	scratch += length;
#endif

	/* per-device configurations are common */
	envp [i++] = scratch;
 #ifndef USB_NO_LIB
	length += snprintf (scratch, buffer_size - length, "PRODUCT=%x/%x/%x",
			    usb_dev->descriptor.idVendor,
			    usb_dev->descriptor.idProduct,
			    usb_dev->descriptor.bcdDevice);
 #endif // ndef USB_NO_LIB
	if ((buffer_size - length <= 0) || (i >= num_envp))
		return -ENOMEM;
	++length;
	scratch += length;

	/* class-based driver binding models */
	envp [i++] = scratch;
 #ifndef USB_NO_LIB
	length += snprintf (scratch, buffer_size - length, "TYPE=%d/%d/%d",
			    usb_dev->descriptor.bDeviceClass,
			    usb_dev->descriptor.bDeviceSubClass,
			    usb_dev->descriptor.bDeviceProtocol);
 #endif // ndef USB_NO_LIB
	if ((buffer_size - length <= 0) || (i >= num_envp))
		return -ENOMEM;
	++length;
	scratch += length;

	if (usb_dev->descriptor.bDeviceClass == 0) {
		int alt = intf->act_altsetting;

		/* 2.4 only exposed interface zero.  in 2.5, hotplug
		 * agents are called for all interfaces, and can use
		 * $DEVPATH/bInterfaceNumber if necessary.
		 */
		envp [i++] = scratch;
 #ifndef USB_NO_LIB
		length += snprintf (scratch, buffer_size - length,
			    "INTERFACE=%d/%d/%d",
			    intf->altsetting[alt].desc.bInterfaceClass,
			    intf->altsetting[alt].desc.bInterfaceSubClass,
			    intf->altsetting[alt].desc.bInterfaceProtocol);
 #endif // ndef USB_NO_LIB
		if ((buffer_size - length <= 0) || (i >= num_envp))
			return -ENOMEM;
		++length;
		scratch += length;

	}
	envp [i++] = 0;

	return 0;
}

#else

	static int usb_hotplug (struct device *dev, char **envp,
			int num_envp, char *buffer, int buffer_size)
	{
	 USB_DBG_FLOW(iprintf("[USB FLOW] usb_hotplug() Not Do Anything. \n"));
	 return -ENODEV;
	} 

#endif	/* CONFIG_HOTPLUG */


/*
 * This entrypoint gets called for each new device.
 *
 * All interfaces are scanned for matching drivers.
 */

// #ifdef NO_SYS_FS
static void usb_find_drivers(struct usb_device *dev)
{
 unsigned ifnum;
 unsigned rejected = 0;
 unsigned claimed = 0;
 /*
 #ifdef _USB_SYSLOG
    char usbsyslog_buf[100];
 #endif // _SYSLOG
 */
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_find_drivers() Begin .. \n"));
 
 USB_DBG_INIT(iprintf(
    	"[USB INIT] usb_find_drivers() : &(dev->actconfig)"
     	" = %08XH \n",
     	&(dev->actconfig)
     	));
        
 USB_DBG_MEM(iprintf(
    	"[USB MEM] usb_find_drivers() : &(dev->actconfig)"
     	" = %08XH \n",
     	&(dev->actconfig)
     	));
     
 #ifdef USB_BS_ASSERT
    	if  ( !(   (unsigned long int) &(dev->actconfig)
                 > (unsigned long int)0x1000
        	)  )
           {
         	iprintf("[ASSERT][USB] usb_find_drivers() : " ) ;
         	iprintf("Error : &(dev->actconfig) = %08XH "
                 "address out of range !!\n",
                 &(dev->actconfig)
                ) ;
         	iprintf("[ASSERT][USB] Continue ....\n") ;
	       }
 #endif
 
 USB_DBG_INIT(iprintf(
	    "[USB INIT] usb_find_drivers() : dev->actconfig = %08XH \n",
	    dev->actconfig
	    ));
        
 USB_DBG_MEM(iprintf(
	    "[USB MEM] usb_find_drivers() : dev->actconfig = %08XH \n",
	    dev->actconfig
	    ));
     
 // 2003/3/28 B. S. Liaw
 if (!dev->actconfig) {
     	dprintf("[USB] FixMe : No dev->actconfig !!\n") ;
	 	USB_DBG_FLOW(iprintf("[USB FLOW] usb_find_drivers() 1 End. \n"));
	 	return;
	   }
 
 USB_DBG_INIT(iprintf(
     	"[USB INIT] usb_find_drivers() : &(dev->actconfig->desc."
        "bNumInterfaces) = %08XH \n",
     	&(dev->actconfig->desc.bNumInterfaces)
     	));
 USB_DBG_MEM(iprintf(
     	"[USB MEM] usb_find_drivers() : &(dev->actconfig->desc.bNumInterfaces)"
     	" = %08XH \n",
     	&(dev->actconfig->desc.bNumInterfaces)
     	));
     
 #ifdef USB_BS_ASSERT
    	if ( !(   (unsigned long int)
        		  &(dev->actconfig->desc.bNumInterfaces) 
        		> (unsigned long int) 0x1000
           )  )
	       {
    	    iprintf("[ASSERT][USB] usb_find_drivers() : " ) ;
        	iprintf("Error : &(dev->actconfig->desc.bNumInterfaces)  = %08XH "
                 "address out of range !!\n",
                 &(dev->actconfig->desc.bNumInterfaces) 
                ) ;
         	iprintf("[ASSERT][USB] Continue ....\n") ;
           }
 #endif

 USB_DBG_INIT(iprintf("[USB INIT] dev->actconfig->desc.bNumInterfaces = %d \n",
	     dev->actconfig->desc.bNumInterfaces
	     ));
 for (ifnum = 0; ifnum < dev->actconfig->desc.bNumInterfaces; ifnum++)
   {
	// if this interface hasn't already been claimed 
	if (!usb_interface_claimed(dev->actconfig->interface + ifnum))
       {
			USB_DBG_INIT(iprintf("[USB INIT] interface %d hasn't been claimed \n", ifnum));
		if (usb_find_interface_driver(dev, ifnum))
           {
			rejected++;
           }
		  else
           {
			claimed++;
           }
	   }
	} // for
 
 if (rejected)
    {
	 USB_DBG_INIT(iprintf("[USB INIT] usb_find_drivers() : unhandled interfaces on device \n")) ;
    }
    
 if (!claimed)
    {
	USB_DBG_INIT(iprintf("[USB INIT] device %d (vend/prod 0x%x/0x%x) is not claimed by any active driver.\n",
			dev->devnum,
			dev->descriptor.idVendor,
			dev->descriptor.idProduct));
            
	sprintf(usb_info_str, "unsupported device , V/P = %04XH/%04XH ",
			dev->descriptor.idVendor, dev->descriptor.idProduct) ;
	iprintf("[USB] unsupported device , V/P = %04XH/%04XH \n",
			dev->descriptor.idVendor, dev->descriptor.idProduct) ;

	#ifdef _USB_SYSLOG
	    //sprintf(usbsyslog_buf,"[USB] unsupported device , V/P = %04XH/%04XH ",
		//			dev->descriptor.idVendor,
		//			dev->descriptor.idProduct);
					#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_DEV_UNSUPP
			//sprintf_SetSystemLog("%s , V/P = %04XH/%04XH ", USB_SYS_LOG_DEV_UNSUPP, dev->descriptor.idVendor, dev->descriptor.idProduct);
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, "%s , V/P = %04XH/%04XH ", USB_SYS_LOG_DEV_UNSUPP, dev->descriptor.idVendor, dev->descriptor.idProduct);
#else
			//sprintf_SetSystemLog("[USB] unsupported device , V/P = %04XH/%04XH ", dev->descriptor.idVendor, dev->descriptor.idProduct);
			sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, "[USB] unsupported device , V/P = %04XH/%04XH ", dev->descriptor.idVendor, dev->descriptor.idProduct);
#endif // USB_SYS_LOG_DEV_UNSUPP
		}
	#endif // _SYSLOG

	/*
    if (usb_device_count <= 1) {
		#if defined(USB_INT_SCHEDULING) || defined(USB_INT_ON_DEMAND)
			UsbDisableInt() ;
		#endif // USB_INT_ON_DEMAND
			USB_DBG_INIT(iprintf("[USB INIT] usb_find_drivers() : reset usb_device_count to 0 \n")) ;
		usb_device_count = 0 ; // >= 1 : enable ISR
       }
     else {
		usb_device_count -- ; // >= 1 : enable ISR
       }
		*/

	    iprintf("[USB] Total External Device = %d \n",usb_device_count) ;
 #ifdef _USB_SYSLOG
    //sprintf(usbsyslog_buf,"[USB] Total External Device = %d ",usb_device_count);
				#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_DEV_NUM
		//sprintf_SetSystemLog(USB_SYS_LOG_DEV_NUM, usb_device_count);
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, USB_SYS_LOG_DEV_NUM, usb_device_count);
#else
		//sprintf_SetSystemLog("[USB] Total External Device = %d ",usb_device_count);
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, "[USB] Total External Device = %d ",usb_device_count);
#endif // USB_SYS_LOG_DEV_NUM
	}
 #endif // _SYSLOG
    
	#ifdef USB_DBG_INIT_ON
		usb_show_device(dev);
      #else
	        // #error : Please Make Sure : Not USB_DBG_INIT_ON !!
    #endif
   }
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_find_drivers() 2 End. \n"));
}
// #endif // NO_SYS_FS


/**
 * usb_alloc_dev - allocate a usb device structure (usbcore-internal)
 * @parent: hub to which device is connected
 * @bus: bus used to access the device
 * Context: !in_interrupt ()
 *
 * Only hub drivers (including virtual root hub drivers for host
 * controllers) should ever call this.
 *
 * This call is synchronous, and may not be used in an interrupt context.
 * Anybody may use usb_inc_dev_use or usb_dec_dev_use
 */
struct usb_device *usb_alloc_dev(struct usb_device *parent, 
			struct usb_bus *bus)
{
 struct usb_device *usbdev;

 USB_DBG_MEM(iprintf("[USB MEM] usb_alloc_dev() : alloc usbdev .. \n"));
 /*
 #ifndef _ETCPIP
	usbdev = kmalloc(sizeof(*dev), GFP_KERNEL);
	if (!dev)
		return NULL;
  #else
 */
	// usbdev = NonCache_contigmalloc(sizeof(*dev)) ;
    
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    // BLOCK_PREEMPTION;

	usbdev = (struct usb_device *) // size : 432
    	/*
		#ifdef USB_DBG_MEM_C_ON
			UsbAllocMemC(USB_MEM_C_M, sizeof(struct usb_device)) ;
         #else
	     	UsbAllocMemC(USB_MEM_C_M ) ;
        #endif
        */
        // 2003/7/14 B. S. Liaw BugFix , Need Un-Cached
        	UsbAllocMemU(
				    #ifdef USB_DBGSTRING_ON
						// size : 516
						USB_MEM_U_L
					 #else
						// size : 508  /* Charles 20040429 */
						USB_MEM_U_M
					#endif
					#ifdef USB_DBG_MEM_U_ON
    	             ,sizeof(struct usb_device)
	                #endif
                );
                 
    // UNBLOCK_PREEMPTION;

    USB_DBG_MEM(iprintf(
		"[USB MEM] usb_alloc_dev() : usbdev = %08XH \n",usbdev));

    if ( usbdev == NULL ) {
        USB_DBG_INIT(iprintf("[USB INIT] usb_alloc_dev() : Error : couldn't allocate usbdev !!\n"));
		dprintf("[USB] Error : Out of Memory !!\n");
		return NULL;
	   }
 /*
 #endif
 */

 /*
 #ifndef _ETCPIP
	memset(usbdev, 0, sizeof(*usbdev));
 #endif
 */

 device_initialize(&usbdev->dev);
 usbdev->state = USB_STATE_ATTACHED;
 
 usb_bus_get(bus);

 if (!parent)
    {
	#ifdef USB_DBGSTRING_ON
		usbdev->devpath [0] = '0';
	#endif
    }
 usbdev->bus = bus;
 usbdev->parent = parent;
 INIT_LIST_HEAD(&usbdev->filelist);

 #ifndef NO_SEMPHORE
 	/*
	init_MUTEX(&dev->serialize);
    */
 #endif // ndef NO_SEMPHORE

 if (usbdev->bus->op->allocate) {
     usbdev->bus->op->allocate(usbdev);
    }

 return usbdev;
}

/**
 * usb_get_dev - increments the reference count of the usb device structure
 * @dev: the device being referenced
 *
 * Each live reference to a device should be refcounted.
 *
 * Drivers for USB interfaces should normally record such references in
 * their probe() methods, when they bind to an interface, and release
 * them by calling usb_put_dev(), in their disconnect() methods.
 *
 * A pointer to the device with the incremented reference counter is returned.
 */
struct usb_device *usb_get_dev (struct usb_device *dev)
{
 struct device *tmp;

 if (!dev)
	return NULL;

 tmp = get_device(&dev->dev);
 if (tmp)        
	return to_usb_device(tmp);
 else
	return NULL;
}

/**
 * usb_put_dev - release a use of the usb device structure
 * @dev: device that's been disconnected
 *
 * Must be called when a user of a device is finished with it.  When the last
 * user of the device calls this function, the memory of the device is freed.
 */
#ifndef _ETCPIP
	/*
	void usb_put_dev(struct usb_device *dev)
	{
	 if (dev)
		put_device(&dev->dev);
	}
	*/
#endif // ndef _ETCPIP

/**
 * usb_release_dev - free a usb device structure when all users of 
   it are finished.
 * @dev: device that's been disconnected
 *
 * Will be called only by the device core when all users of this 
   usb device are
 * done.
 */
static void usb_release_dev(struct device *dev)
{
 struct usb_device *usbdev;

 USB_DBG_INIT(iprintf("[USB INIT] usb_release_dev() .. \n"));
 
 usbdev = to_usb_device(dev);
 
 if (!usbdev)
   {
	return;
   }

 if (usbdev->bus && usbdev->bus->op && usbdev->bus->op->deallocate)
   {
	USB_DBG_INIT(iprintf("[USB INIT] usbdev->bus->op->deallocate() .. \n"));
	usbdev->bus->op->deallocate(usbdev);
   }
  else
   {
	USB_DBG_INIT(iprintf("[USB INIT] Error : No udev->bus->op->deallocate() !! \n"));
   }
   
 USB_DBG_FLOW(iprintf(
	"[USB FLOW] usb_release_dev() call usb_destroy_configuration() ...\n"
    ));
 usb_destroy_configuration(usbdev);
 usb_bus_put(usbdev->bus);
 
 USB_DBG_MEM(iprintf("[USB MEM] usb_release_dev() : free usbdev .. \n"));

 /*
 	kfree (usbdev);
 */
 // 2003/7/14 B. S. Liaw BugFix , Need Un-Cached
 UsbFreeMemU(
				#ifdef USB_DBGSTRING_ON
					// size : 516
					USB_MEM_U_L ,
				 #else
					// size : 432
					USB_MEM_U_M ,
				#endif
				(unsigned char *) usbdev
				#if defined(USB_DBG_MEM_U_ON)
	        		, sizeof(struct usb_device)
    	        #endif
		       ) ;
}


/**
 * usb_get_current_frame_number - return current bus frame number
 * @dev: the device whose bus is being queried
 *
 * Returns the current frame number for the USB host controller
 * used with the given USB device.  This can be used when scheduling
 * isochronous requests.
 *
 * Note that different kinds of host controller have different
 * "scheduling horizons".  While one type might support scheduling only
 * 32 frames into the future, others could support scheduling up to
 * 1024 frames into the future.
 */
int usb_get_current_frame_number(struct usb_device *dev)
{
	return dev->bus->op->get_frame_number (dev);
}

/*-------------------------------------------------------------------*/
/*
 * __usb_get_extra_descriptor() finds a descriptor of specific type in the
 * extra field of the interface and endpoint descriptor structs.
 */

int __usb_get_extra_descriptor(char *buffer, unsigned size, unsigned char type, void **ptr)
{
	struct usb_descriptor_header *header;

	while (size >= sizeof(struct usb_descriptor_header)) {
		header = (struct usb_descriptor_header *)buffer;

		if (header->bLength < 2) {
			err("invalid descriptor length of %d", header->bLength);
			return -1;
		}

		if (header->bDescriptorType == type) {
			*ptr = header;
			return 0;
		}

		buffer += header->bLength;
		size -= header->bLength;
	}
	return -1;
}

/**
 * usb_disconnect - disconnect a device (usbcore-internal)
 * @pdev: pointer to device being disconnected
 * Context: !in_interrupt ()
 *
 * Something got disconnected. Get rid of it, and all of its children.
 *
 * Only hub drivers (including virtual root hub drivers for host
 * controllers) should ever call this.
 *
 * This call is synchronous, and may not be used in an interrupt context.
 */
void usb_disconnect(struct usb_device **pdev)
{
 struct usb_device * dev = *pdev;
 int i;
 /*
 #ifdef _USB_SYSLOG
    char usbsyslog_buf[100];
 #endif // _SYSLOG
 */

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_disconnect() Begin ..\n"));
 
 if (!dev)
    {
 	 return;
    }

 *pdev = NULL;

 iprintf("[USB] disconnect on device %d\n", dev->devnum);
		#ifdef _USB_SYSLOG
		    //sprintf(usbsyslog_buf,"[USB] disconnect on device %d ", dev->devnum);
						#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_DEV_DISCONN
				//sprintf_SetSystemLog(USB_SYS_LOG_DEV_DISCONN, dev->devnum);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, USB_SYS_LOG_DEV_DISCONN, dev->devnum);
#else
				//sprintf_SetSystemLog("[USB] disconnect on device %d ", dev->devnum);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_CORE, "[USB] disconnect on device %d ", dev->devnum);
#endif // USB_SYS_LOG_DEV_DISCONN
			}
		#endif // _SYSLOG
 
 #ifdef USB_BS_ASSERT
    if  ( !( dev->devnum > 0
        )  )
        {
         iprintf("[ASSERT][USB] usb_disconnect() : " ) ;
         iprintf("Error : dev->devnum !!\n" ) ;
         iprintf("[ASSERT][USB] Continue ....\n") ;
        }
 #endif

 /* mark the device as inactive, so any further urb submissions for
  * this device will fail.
  */
 dev->state = USB_STATE_NOTATTACHED;

 dev_info (&dev->dev, "USB disconnect, address %d\n", dev->devnum);

 /*
 usb_hub_plug_out = 1 ;

 USB_DBG_INFO(iprintf("[USB INFO] hub_disconnect() : wait for hub->hub_in_submit == 0 ..\n" )) ;
 while (hub->hub_in_submit)
       {
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
        NULL ;
       }
 USB_DBG_INFO(iprintf("[USB INFO] hub_disconnect() : wait End.\n" )) ;
 */
 
 /* Free up all the children before we remove this device */
 for (i = 0; i < USB_MAXCHILDREN; i++)
   {
	struct usb_device **child = dev->children + i;
	if (*child)
       {
		USB_DBG_INFO(iprintf("[USB INFO] usb_disconnect() : usb_disconnect(child) .. \n"));
		usb_disconnect(child);
       }
   }

 #ifdef USB_DBG_INFO_ON
	dev_dbg (&dev->dev, "unregistering interfaces\n");
 #endif
 if (dev->actconfig)
   {
	for (i = 0; i < dev->actconfig->desc.bNumInterfaces; i++)
       {
		struct usb_interface *interface = &dev->actconfig->interface[i];

		/* remove this interface */
		#ifndef _ETCPIP
			device_unregister(&interface->dev);
         #else
			// usb_device_remove(struct device *dev)
			if (interface->driver && interface->driver->disconnect)
			   {
	    // #ifdef USB_DBG_INIT_ON
		#ifdef USB_DBG_MEM_COUNT_ON
			iprintf("[USB MEM_COUNT] usb_disconnect(1) : show USB Mem Count .. \n");
			showUsbMemC() ;
			showUsbMemU() ;
		#endif
				USB_DBG_INIT(iprintf("[USB INIT] usb_disconnect() : "
                	"interface->driver->disconnect(interface) ..\n")) ;
				interface->driver->disconnect(interface);
				// usb_destroy_configuration(dev);
	    // #ifdef USB_DBG_INIT_ON
		#ifdef USB_DBG_MEM_COUNT_ON
			iprintf("[USB MEM_COUNT] usb_disconnect(2) : show USB Mem Count .. \n");
			showUsbMemC() ;
			showUsbMemU() ;
		#endif
			   }
              else
               {
				USB_DBG_INIT(iprintf("[USB INIT] usb_disconnect() : Cannot run "
                	"interface->driver->disconnect(intf) !!\n")) ;
				USB_DBG_INFO(iprintf("[USB INFO] usb_disconnect() : Null interface->driver->disconnect(intf) !!\n")) ;
				/*
				dprintf("[USB INIT] usb_disconnect() : Cannot run "
                	"interface->driver->disconnect(intf) !!\n") ;
				*/
               }
		#endif // ndef _ETCPIP
	   }
   }

 dev_dbg (&dev->dev, "unregistering device\n");
 
 /* Free the device number and remove the /proc/bus/usb entry */
 if (dev->devnum > 0)
   {
	#ifdef _ETCPIP
		USB_DBG_INIT(iprintf(
			"[USB INIT] usb_disconnect() Not call usb_destroy_configuration() ...\n"
    		));
        /*
		USB_DBG_FLOW(iprintf(
			"[USB FLOW] usb_disconnect() call usb_destroy_configuration() ...\n"
    		));
        // extern void usb_destroy_configuration(struct usb_device *dev);
		usb_destroy_configuration(dev);
        */
        /*
		// static void usb_release_dev(struct device *dev)
		usb_release_dev(dev) ;
        */
	#endif // _ETCPIP
        
	clear_bit(dev->devnum, dev->bus->devmap.devicemap);
	#ifdef CONFIG_USB_DEVICEFS
		usbfs_remove_device(dev);
	#endif // CONFIG_USB_DEVICEFS
   }
   
 #ifndef _ETCPIP
	device_unregister(&dev->dev);
  #else
	USB_DBG_FLOW(iprintf(
		"[USB FLOW] usb_disconnect() call usb_release_dev() ...\n"
    	));
	// static void usb_release_dev(struct device *dev)
	usb_release_dev(&dev->dev) ;
 #endif // ndef _ETCPIP

 /* Decrement the reference count, it'll auto free everything when */
 /* it hits 0 which could very well be now */
 #ifndef _ETCPIP
	USB_DBG_FLOW(iprintf(
    	"[USB FLOW] usb_disconnect() Call usb_free_dev() ...\n"
        ));
	usb_put_dev(dev);
 #endif // ndef _ETCPIP
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_disconnect() End.\n"));
}

/**
 * usb_connect - pick device address (usbcore-internal)
 * @dev: newly detected device (in DEFAULT state)
 *
 * Picks a device address.  It's up to the hub (or root hub) driver
 * to handle and manage enumeration, starting from the DEFAULT state.
 * Only hub drivers (including virtual root hub drivers for host
 * controllers) should ever call this.
 */
void usb_connect(struct usb_device *dev)
{
 int devnum;


 USB_DBG_FLOW(iprintf("[USB FLOW] usb_connect() Begin ..\n"));
 
 // FIXME needs locking for SMP!!
	/* why? this is called only from the hub thread, 
	 * which hopefully doesn't run on multiple CPU's simultaneously 8-)
	 * ... it's also called from modprobe/rmmod/apmd threads as part
	 * of virtual root hub init/reinit.  In the init case, the hub code 
	 * won't have seen this, but not so for reinit ... 
	 */
	dev->descriptor.bMaxPacketSize0 = 8;  /* Start off at 8 bytes  */
 #ifndef DEVNUM_ROUND_ROBIN
	devnum = find_next_zero_bit(dev->bus->devmap.devicemap, 128, 1);
  #else	/* round_robin alloc of devnums */
	/* Try to allocate the next devnum beginning at bus->devnum_next. */
	devnum = find_next_zero_bit(dev->bus->devmap.devicemap, 128, dev->bus->devnum_next);
	if (devnum >= 128)
		devnum = find_next_zero_bit(dev->bus->devmap.devicemap, 128, 1);

	dev->bus->devnum_next = ( devnum >= 127 ? 1 : devnum + 1);
 #endif	/* round_robin alloc of devnums */

 if (devnum < 128)
    {
	 set_bit(devnum, dev->bus->devmap.devicemap);
	 dev->devnum = devnum;
     USB_DBG_INIT(iprintf(
       "[USB INIT] usb_connect() : Set Value dev->devnum = %d \n",
       dev->devnum
       )) ;
	}
    
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_connect() End.\n"));
}


// hub-only!! ... and only exported for reset/reinit path.
// otherwise used internally, for usb_new_device()
int usb_set_address(struct usb_device *dev)
{
 int retval;

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_set_address() Begin .. \n")) ;
 
 #ifdef USB_RES_SEMPHORE    // use MT_RES_USB semphore
	if (getres(MT_RES_USB, 1000) != 0)
    	   {
	        iprintf("[USB] usb_set_address() : getres Fail !!") ;
    	    // continue ;
	       }
 #endif
 
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
       ) ) {
				iprintf("[ASSERT][USB] usb_set_address() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT

 #ifdef USB_DBG_ABORT   
   // vvv for debug Prefetch Error !!
   dprintf("[TEST] GetTime() = %08XH\n",GetTime()) ;
			#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
				// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
             #else
            	YIELD();
            #endif
   // ^^^ for debug Prefetch   
 #endif   
 
 USB_DBG_INIT(iprintf(
    "[USB INIT] usb_set_address() : "
    "dev->bus->busnum = %d , dev->devnum =  %d\n" ,
     dev->bus->busnum , dev->devnum
    ));
    
 USB_DBG_INFO(iprintf("[USB INFO] dev->speed = %d \n",dev->speed)) ;
 
 // #define usb_snddefctrl(dev)		((PIPE_CONTROL << 30) | __default_pipe(dev))
 USB_DBG_PIPE(iprintf(
    "[USB PIPE] usb_snddefctrl(dev) = %08XH \n",usb_snddefctrl(dev))) ;
 USB_DBG_PIPE(iprintf(
    "[USB PIPE] usb_pipedevice(pipe) = %d \n",
    usb_pipedevice(usb_snddefctrl(dev))
    )) ;
 if (dev->devnum == 0)
		return -EINVAL;
        
 if (dev->state != USB_STATE_DEFAULT && dev->state != USB_STATE_ADDRESS)
		return -EINVAL;
 
 USB_DBG_INIT(iprintf(
	"[USB INIT] usb_set_address() : Control Message .. \n"
    )) ;
 /*
 #ifdef USB_NO_VMM // no vitrtul memory management
	#ifdef USB_BS_ASSERT
		#ifdef _ADM5120
			if (!(  ((unsigned long)data & (unsigned long)0x20000000L)
    	          ==((unsigned long)0x20000000L)
    		     )
		       )
			   {
				iprintf("[ASSERT][USB] usb_set_address() : " ) ;
    		    iprintf("Error : Not Un-Cached !!\n" ) ;
			    iprintf("[ASSERT][USB] data addr = %08XH \n", data ) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // _ADM5120
	#endif // USB_BS_ASSERT
 #endif // USB_NO_VMM
 */   
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_set_address() Call "
    "usb_control_msg() ... \n")) ;
 
 #ifndef _ETCPIP
 	/*
	//int usb_control_msg(struct usb_device *dev, unsigned int pipe,
	//   __u8 request, __u8 requesttype,
	//   __u16 value, __u16 index, void *data, __u16 size, int timeout)
 
	retval = usb_control_msg(dev, usb_snddefctrl(dev), USB_REQ_SET_ADDRESS,
		0, dev->devnum, 0, NULL, 0, HZ * USB_CTRL_SET_TIMEOUT);
    */
  #else
	retval = usb_control_msg(dev, usb_snddefctrl(dev), 
    	USB_REQ_SET_ADDRESS,
		0, dev->devnum, 0, NULL, 0, USB_CTRL_SETADDR_TIMEOUT);
 #endif
 
 USB_DBG_CTLMSG(iprintf("[USB CTLMSG] usb_set_address() : retval = %d \n" , retval )) ;
    
 if (retval == 0)
   {
	dev->state = USB_STATE_ADDRESS;
   }
 
 /*       
 dprintf("[USB] new device connect on bus number %d [%s] , "
         "assigned device number %d \n",
		dev->bus->busnum, portstr, dev->devnum);
 */       
 #ifdef USB_RES_SEMPHORE    // use MT_RES_USB semphore
	relres(MT_RES_USB) ;
 #endif
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_set_address() End.\n")) ;
 return retval;
}


/* improve on the default device description, if we can ... and
 * while we're at it, maybe show the vendor and product strings.
 */
static void set_device_description (struct usb_device *dev)
{
 void    *buf;
 int	mfgr = dev->descriptor.iManufacturer;
 int	prod = dev->descriptor.iProduct;
 int	vendor_id = dev->descriptor.idVendor;
 int	product_id = dev->descriptor.idProduct;
 char	*mfgr_str, *prod_str;

	/* set default; keep it if there are no strings, or kmalloc fails */
 #ifdef USB_DBGSTRING_ON
	sprintf (dev->dev.name, "USB device %04x:%04x",
		 vendor_id, product_id);
 #endif // USB_DBGSTRING_ON

 #ifndef _ETCPIP
	if (!(buf = kmalloc(256 * 2, GFP_KERNEL)))
		return;
  #else
    // B. S. Liaw : alloc_mem is on mt_??/USFSup1.c
    //              But now on ap/upnp/upnpport.c
    //              void *alloc_mem(int size) 
    //              this will fill all 0s

    // BLOCK_PREEMPTION;

	buf = (void *)
		#ifdef USB_DBG_MEM_C_ON
			UsbAllocMemC(USB_MEM_C_M, 256 * 2 ) ;
         #else
          UsbAllocMemC(USB_MEM_C_M) ;
        #endif
                 
    // UNBLOCK_PREEMPTION;

    USB_DBG_MEM(iprintf("[USB MEM] buf = %08XH \n",buf));

    if ( buf == NULL )
       {
		dprintf("couldn't allocate buf memory \n");
		return ;
	   }
 #endif

	
	prod_str = (char *) buf;
	mfgr_str = (char *) buf + 256;

 #ifdef USB_DBGSTRING_ON
	if (prod && usb_string (dev, prod, prod_str, 256) > 0)
  #else
	if (prod && usb_string (dev, prod) > 0)
 #endif
     {
#ifdef DEBUG
  #ifndef _ETCPIP
		dev_printk (KERN_INFO, &dev->dev, "Product: %s\n", prod_str);
   #else
		dev_printk (&dev->dev, "Product: %s\n", prod_str);
  #endif
#endif
	} else {
		prod_str = 0;
	}

 #ifdef USB_DBGSTRING_ON
	if (mfgr && usb_string (dev, mfgr, mfgr_str, 256) > 0)
  #else
	if (mfgr && usb_string (dev, mfgr) > 0)
 #endif
	   {
		// #ifdef DEBUG
		#ifdef USB_DBG_INFO_ON
			#ifndef _ETCPIP
				dev_printk (KERN_INFO, &dev->dev, "Manufacturer: %s\n", mfgr_str);
			 #else
				dev_printk (&dev->dev, "Manufacturer: %s\n", mfgr_str);
			#endif
		#endif
	   }
      else
       {
		mfgr_str = 0;
	   }

	/* much like pci ... describe as either:
	 * - both strings:   'product descr (vendor descr)'
	 * - product only:   'product descr (USB device vvvv:pppp)'
	 * - vendor only:    'USB device vvvv:pppp (vendor descr)'
	 * - neither string: 'USB device vvvv:pppp'
	 */

 #ifdef USB_DBGSTRING_ON
	#ifndef USB_NO_LIB
	if (prod_str && mfgr_str) {
		snprintf(dev->dev.name, sizeof dev->dev.name,
			 "%s (%s)", prod_str, mfgr_str);
	} else if (prod_str) {
		snprintf(dev->dev.name, sizeof dev->dev.name,
			 "%s (USB device %04x:%04x)",
			 prod_str, vendor_id, product_id);
	} else if (mfgr_str) {
		snprintf(dev->dev.name, sizeof dev->dev.name,
			 "USB device %04x:%04x (%s)",
			 vendor_id, product_id, mfgr_str);
	}
	#endif // ndef USB_NO_LIB
 #endif // USB_DBGSTRING_ON

 USB_DBG_MEM(iprintf(
	"[USB MEM] set_device_description(): free buf .. \n"
    ));
 #ifndef USB_MEM_LIB
 	kfree(buf);
 #elif defined USB_DBG_MEM_C_ON
    UsbFreeMemC(USB_MEM_C_M,(unsigned char *)buf,256*2) ;
 #else
    UsbFreeMemC(USB_MEM_C_M,(unsigned char *)buf) ;
 #endif
}

/*
 * By the time we get here, we chose a new device address
 * and is in the default state. We need to identify the thing and
 * get the ball rolling..
 *
 * Returns 0 for success, != 0 for error.
 *
 * This call is synchronous, and may not be used in an interrupt context.
 *
 * Only hub drivers (including virtual root hub drivers for host
 * controllers) should ever call this.
 */
#define NEW_DEVICE_RETRYS	2
#define SET_ADDRESS_RETRYS	2

int usb_new_device(struct usb_device *dev, struct device *parent)
{
 int err = 0;
 int i;
 int j;


 USB_DBG_FLOW(iprintf("[USB FLOW] usb_new_device() Begin ..\n"));
 
 // #ifdef USB_DBG_INIT_ON
 #ifdef USB_DBG_MEM_COUNT_ON
	iprintf("[USB MEM_COUNT] usb_new_device(0) : show USB Mem Count .. \n");
	showUsbMemC() ;
	showUsbMemU() ;
 #endif
        
 USB_DBG_MEM(iprintf(
    "[USB MEM] usb_new_device() dev = %08XH \n",dev));
    
 #ifdef USB_DBG_ABORT   
   // vvv for debug Prefetch Error !!
   dprintf("[TEST] GetTime() = %08XH\n",GetTime()) ;
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] usb_new_device() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
			#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
				// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
             #else
            	YIELD();
            #endif
   // ^^^ for debug Prefetch   
 #endif   

 /*
  * Set the driver for the usb device to point to the "generic" driver.
  * This prevents the main usb device from being sent to the usb bus
  * probe function.  Yes, it's a hack, but a nice one :)
  *
  * Do it asap, so more driver model stuff (like the device.h message
  * utilities) can be used in hcd submit/unlink code paths.
  */
 usb_generic_driver.bus = &usb_bus_type;
 dev->dev.parent = parent;
 dev->dev.driver = &usb_generic_driver;
 dev->dev.bus = &usb_bus_type;
 dev->dev.release = usb_release_dev;
 usb_get_dev(dev);
 if(dev->dev.bus_id[0] == 0)
   {
	#ifdef USB_DBGSTRING_ON
		sprintf (&dev->dev.bus_id[0], "%d-%s",
			 dev->bus->busnum, dev->devpath);
	 #else
		sprintf (&dev->dev.bus_id[0], "%d",
			 dev->bus->busnum);
	#endif
   }

 /* dma masks come from the controller; readonly, except to hcd */
 dev->dev.dma_mask = parent->dma_mask;

 /* it's not usable yet */
 dev->state = USB_STATE_DEFAULT;

 /* USB 2.0 section 5.5.3 talks about ep0 maxpacket ...
  * it's fixed size except for full speed devices.
  */
 /* USB v1.1 5.5.3 */
 /* We read the first 8 bytes from the device descriptor to get to */
 /*  the bMaxPacketSize0 field. Then we set the maximum packet size */
 /*  for the control pipe, and retrieve the rest */
 
 USB_DBG_INIT(iprintf("[USB INIT] dev->speed = %d \n",dev->speed)) ;
 
 switch (dev->speed) {
	case USB_SPEED_HIGH:		/* fixed at 64 */
		USB_DBG_INFO(iprintf("[USB INFO] USB_SPEED_HIGH !!\n")) ;
		i = 64;
		break;
	case USB_SPEED_FULL:		/* 8, 16, 32, or 64 */
		USB_DBG_INFO(iprintf("[USB INFO] USB_SPEED_FULL !!\n")) ;
		/* to determine the ep0 maxpacket size, read the first 8
		 * bytes from the device descriptor to get bMaxPacketSize0;
		 * then correct our initial (small) guess.
		 */
		// FALLTHROUGH
		i = 8;
		break;
        
	case USB_SPEED_LOW:		/* fixed at 8 */
		USB_DBG_INFO(iprintf("[USB INFO] USB_SPEED LOW \n")) ;
		i = 8;
		break;
        
	default:
		USB_DBG_INIT(iprintf(
        	"[USB INIT] Error : USB_SPEED INVALID !! \n"
            )) ;
		return -EINVAL;
	}
 dev->epmaxpacketin [0] = i;
 dev->epmaxpacketout[0] = i;

 for (i = 0; i < NEW_DEVICE_RETRYS; ++i)
   {
 
	for (j = 0; j < SET_ADDRESS_RETRYS; ++j)
       {

		#ifdef USB_MORE_DELAY
			USB_DBG_INIT(iprintf(
    		"[USB INIT] ==== New Device ====  ==1-3== (USB Device %d) Wait 200 ms \n",dev->devnum));
			// wait_ms(30) ; 
			// wait_ms(30) ; 
			// wait_ms(30) ; 
			wait_ms(200) ; 
		#endif
 
		USB_DBG_INIT(iprintf(
		   "[USB INIT] ==== New Device ====  ==2== (USB Device %d) Set Address \n",dev->devnum));
           
	    // #ifdef USB_DBG_INIT_ON
		#ifdef USB_DBG_MEM_COUNT_ON
			iprintf(
    			"[USB MEM_COUNT] usb_new_device(2) : show USB Mem Count .. \n");
			showUsbMemC() ;
			showUsbMemU() ;
		#endif
 
		err = usb_set_address(dev);
 
		USB_DBG_INIT(iprintf(
		   "[USB INIT] ==== New Device ====  ==2-1== (USB Device %d) after Set Address \n",dev->devnum));
    
		// dprintf("[USB] err = %d (-%02XH) \n", err , (0-err));
		USB_DBG_INIT(iprintf("[USB INIT] usb_new_device() : err=%d (-%02XH) \n",  err , (0-err) ));
           
		#ifndef TIMEOUT_LOOSER_CHECK
			if (err >= 0)
		  #else
			if ( (err >= 0) || (err == -EINPROGRESS) )  // -115
		#endif
			   {
				// dprintf("[USB] Error : device not accepting new address (err=%d) !!\n", err ) ;
				break;
       		   }
               
		USB_DBG_INFO(iprintf("[USB INFO]  Error : device not accepting new address (err=%d) !!\n", err )) ;
		USB_DBG_INIT(iprintf(
			"[USB INIT] ==== New Device ====  ==2-2== (USB Device %d) Wait 200 ms \n",dev->devnum));
		wait_ms(200);
	   } // for j
       
	#ifndef TIMEOUT_LOOSER_CHECK
    	if (err < 0)
	  #else
		if ( (err < 0) && (err != -EINPROGRESS) ) // -115
	#endif
	   {
		dprintf("[USB] Error : device not accepting new address=%d "
             "(err=%d -%02XH) \n",
			    dev->devnum, err , (0-err));
		dev_err(&dev->dev, "USB device not accepting new address=%d (error=%d)\n",
				dev->devnum, err);
		dev->state = USB_STATE_DEFAULT;
			clear_bit(dev->devnum, dev->bus->devmap.devicemap);
		dev->devnum = -1;
	    USB_DBG_INIT(iprintf(
    	   "[USB INIT] usb_new_device() : Set Value dev->devnum = %d \n",
	      dev->devnum
	       )) ;
		return 1;
	   }
 
	USB_DBG_INIT(iprintf(
    	"[USB INIT] New Device  ==3== Wait 30 ms \n"));
	#ifndef _ETCPIP
    	
		// wait_ms(10);	// Let the SET_ADDRESS settle 
        
	  #else
		wait_ms(30) ; 
	#endif
 
	// #ifndef USB_DBG_TEMP_DISABLE  // for debug Prefetch Error !!
    /*
	if ( (dev->speed == USB_SPEED_HIGH) ||(dev->speed == USB_SPEED_LOW))
       {
		USB_DBG_INIT(iprintf("[USB INIT] ==== New Device ====  ==4== (USB Device %d) bypass Get Descriptor !!\n",dev->devnum));
        break ;
       }
	  else
    */
       {
	USB_DBG_INIT(iprintf("[USB INIT] ==== New Device ====  ==4== (USB Device %d) Get Descriptor \n",dev->devnum));
    
	/* high and low speed devices don't need this... */
    
	err = usb_get_descriptor(dev, USB_DT_DEVICE, 0, &dev->descriptor, 8);
    
	USB_DBG_INIT(iprintf("[USB INIT] usb_new_device() : err=%d (-%02XH) \n",  err , (0-err) ));
    
	#ifdef USB_DBG_ABORT   
	   // vvv for debug Prefetch Error !!
	   dprintf("[TEST] GetTime() = %08XH\n",GetTime()) ;
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] usb_new_device() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
			#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
				// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
             #else
            	YIELD();
            #endif
	   // ^^^ for debug Prefetch   
	#endif
    
	if (err >= 8) // 2003/7/16 B. S. Liaw
       {
		USB_DBG_INIT(iprintf("[USB INIT] usb_new_device() : Get device descriptor OK. \n" )) ;
		break;
       }
       
	/* 
	if (err >= 8)
    */
	if (err < 8) // 2003/7/15 B. S. Liaw , 8 : OK , < 8 : Fail
       {
		USB_DBG_INFO(iprintf("[USB INFO] usb_new_device() : Error : Get device descriptor fail !! (err=%d) \n", err )) ;
		// break;
       }
       
	USB_DBG_INIT(iprintf("[USB INIT] ==== New Device ====  ==4-2== (USB Device %d) Wait 100 ms \n",dev->devnum));
	wait_ms(100);
       }
   } // for i
    /* 
	if ( (dev->speed == USB_SPEED_HIGH) ||(dev->speed == USB_SPEED_LOW))
       {
		// USB_DBG_INIT(iprintf("[USB INIT] ==== New Device ====  ==4== (USB Device %d) bypass Get Descriptor !!\n",dev->devnum));
        // break ;
       }
	  else
    */
       {
 #ifndef TIMEOUT_LOOSER_CHECK
    if (err < 8)
  #else
	if ( (err < 8) && (err != -EINPROGRESS) )  // -115
 #endif
    {
	 if(err < 0)
       {
		dprintf("[USB] Error : Get device descriptor fail !! (err=%d) \n", err ) ;
		dprintf("[USB] Error : device not responding, giving up (err=%d)\n", err);
       }
	  else
       {
		dprintf("[USB] Error : device descriptor short read "
                "(expected %i, got %i ) \n", 8, err);
       }
		clear_bit(dev->devnum, dev->bus->devmap.devicemap);
	 dev->devnum = -1;
     USB_DBG_INIT(iprintf(
       "[USB INIT] usb_new_device() : Set Value dev->devnum = %d \n",
       dev->devnum
       )) ;
     USB_DBG_FLOW(iprintf("[USB FLOW] usb_new_device() End with Fail.\n"));
	 return 1;
	}
       }

 #ifdef USB_DBG_ABORT   
   // vvv for debug Prefetch Error !!
   dprintf("[TEST] GetTime() = %08XH\n",GetTime()) ;
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] usb_new_device() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
			#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
				// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
             #else
            	YIELD();
            #endif
   // ^^^ for debug Prefetch   
 #endif   
 
 // #ifndef USB_DBG_TEMP_DISABLE  // for debug Prefetch Error !!

 #ifdef USB_MORE_DELAY
	USB_DBG_INIT(iprintf(
    	"[USB INIT] ==== New Device ====  ==4-3== (USB Device %d) Wait 30 ms \n",dev->devnum));
	wait_ms(30) ; 
 #endif

 if (dev->speed == USB_SPEED_FULL)
 //  if ( (dev->speed == USB_SPEED_HIGH) ||(dev->speed == USB_SPEED_LOW))
   {
	dev->epmaxpacketin [0] = dev->descriptor.bMaxPacketSize0;
	dev->epmaxpacketout[0] = dev->descriptor.bMaxPacketSize0;
   }

 /* USB device state == addressed ... still not usable */
 /*
 USB_DBG_INIT(iprintf("[USB INIT] ==== New Device ====  ==5-0== (USB Device %d) Wait 100 ms \n",dev->devnum));
 wait_ms(100) ; 
 */
 USB_DBG_INIT(iprintf("[USB INIT] ==== New Device ====  ==5== (USB Device %d) Get Device Descriptor \n",dev->devnum));
    
 err = usb_get_device_descriptor(dev);
 
 USB_DBG_INIT(iprintf("[USB INIT] usb_new_device() : err=%d (-%02XH) \n",  err , (0-err) ));
 #ifdef USB_DBG_INFO_ON
 // #ifdef USB_DBG_INIT_ON
 // #ifdef DEBUG
	// usb_show_device(dev);
	usb_show_device_descriptor(&dev->descriptor);
 #endif
 
 #ifndef TIMEOUT_LOOSER_CHECK
    // #error : Please Make Sure : Not TIMEOUT_LOOSER_CHECK !!
    if (err < (signed)sizeof(dev->descriptor))
  #else
	if (   (err < (signed)sizeof(dev->descriptor))
        && (err != -EINPROGRESS)  // -115
       )
 #endif
    {
	 if (err < 0)
       {
        // -110 : connection time out
		dprintf("[USB] unable to get device descriptor (error=%d) \n",
        		err);
       }
	  else
       {
		dprintf("[USB] device descriptor short read (expected %Zi, got"
        		" %i) \n",
				sizeof(dev->descriptor), err);
       }
	
		clear_bit(dev->devnum, dev->bus->devmap.devicemap);
	 dev->devnum = -1;
     USB_DBG_INIT(iprintf(
       "[USB INIT] usb_new_device() : Set Value dev->devnum = %d \n",
       dev->devnum
       )) ;
     USB_DBG_FLOW(iprintf("[USB FLOW] usb_new_device() End with Fail.\n"));
	 return 1;
	}
 
 #ifdef USB_MORE_DELAY
	USB_DBG_INIT(iprintf(
    	"[USB INIT] ==== New Device ====  ==5-3== (USB Device %d) Wait 30 ms \n",dev->devnum));
	wait_ms(30) ; 
 #endif
 
 USB_DBG_INIT(iprintf(
    "[USB INIT] ==== New Device ====  ==6== (USB Device %d) Get Configuration \n",dev->devnum));
 err = usb_get_configuration(dev);
 
 USB_DBG_INIT(iprintf("[USB INIT] usb_new_device() : err=%d (-%02XH) \n",  err , (0-err) ));
 
 // #ifdef DEBUG
 // #ifdef USB_DBG_INIT_ON
 #ifdef USB_DBG_INFO_ON
	usb_show_device(dev);
 #endif
 
 #ifdef USB_DBG_ABORT   
   // vvv for debug Prefetch Error !!
   dprintf("[TEST] GetTime() = %08XH\n",GetTime()) ;
   if(!in_interrupt())
     {
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] usb_new_device() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
			#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
				// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
             #else
            	YIELD();
            #endif
     }
    else
     {
      dprintf("[TEST] in interrupt !!\n");
     }
   // ^^^ for debug Prefetch   
 #endif
 
 #ifndef TIMEOUT_LOOSER_CHECK
    if (err < 0)
  #else
	if ( (err < 0) && (err != -EINPROGRESS) )  // -115
 #endif
    {
	 dprintf("[USB] unable to get device %d configuration (error=%d) \n",
			dev->devnum, err);
		clear_bit(dev->devnum, dev->bus->devmap.devicemap);
	 dev->devnum = -1;
     USB_DBG_INIT(iprintf(
       "[USB INIT] usb_new_device() : Set Value dev->devnum = %d \n",
       dev->devnum
       )) ;
     USB_DBG_INIT(iprintf(
        "[USB INIT] ==== New Device ====  ==6-2== (USB Device %d) after Get Configuration fail\n",dev->devnum));
     USB_DBG_FLOW(iprintf("[USB FLOW] usb_new_device() End with Fail.\n"));
	 return 1;
	}
 USB_DBG_INIT(iprintf(
    "[USB INIT] ==== New Device ====  ==6-1== (USB Device %d) after Get Configuration \n",dev->devnum));

 #ifdef USB_DBG_ABORT   
   // vvv for debug Prefetch Error !!
   dprintf("[TEST] GetTime() = %08XH\n",GetTime()) ;
   if(!in_interrupt())
     {
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] usb_new_device() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
			#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
				// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
             #else
            	YIELD();
            #endif
     }
    else
     {
      dprintf("[TEST] in interrupt !!\n");
     }
   // ^^^ for debug Prefetch   
 #endif   
 
 #ifdef USB_MORE_DELAY
	USB_DBG_INIT(iprintf(
    	"[USB INIT] ==== New Device ====  ==6-3== (USB Device %d) Wait 30 ms \n",dev->devnum));
	wait_ms(30) ; 
 #endif
 
 // #ifndef USB_DBG_TEMP_DISABLE  // for debug Prefetch Error !!
 USB_DBG_INIT(iprintf(
    "[USB INIT] ==== New Device ====  ==7== (USB Device %d) Set Configuration \n",dev->devnum));
 /* we set the default configuration here */
 err = usb_set_configuration(dev, dev->config[0].desc.bConfigurationValue);
 
 USB_DBG_INIT(iprintf("[USB INIT] usb_new_device() : err=%d (-%02XH) \n",  err , (0-err) ));
 
 #ifndef TIMEOUT_LOOSER_CHECK
    if (err)
  #else
    if ( (err < 0) && (err != -EINPROGRESS) )  // -115
 #endif
   {
	dprintf("[USB] failed to set device %d default configuration "
          "(error=%d) \n",
			 dev->devnum, err);
		dev_err(&dev->dev, "failed to set device %d default configuration (error=%d)\n",
			dev->devnum, err);
		clear_bit(dev->devnum, dev->bus->devmap.devicemap);
	dev->devnum = -1;
    USB_DBG_INIT(iprintf(
       "[USB INIT] usb_new_device() : Set Value dev->devnum = %d \n",
       dev->devnum
       )) ;
       
    USB_DBG_INIT(iprintf(
        "[USB INIT] ==== New Device ====  ==7-2== (USB Device %d) after Set Configuration Fail\n",dev->devnum));
    USB_DBG_FLOW(iprintf("[USB FLOW] usb_new_device() End with Fail.\n"));
	return 1;
   }

 USB_DBG_INIT(iprintf(
    "[USB INIT] ==== New Device ====  ==7-1== (USB Device %d) after Set Configuration \n",dev->devnum));

	/* USB device state == configured ... tell the world! */
    
 #ifdef USB_DBG_ABORT   
   // vvv for debug Prefetch Error !!
   dprintf("[TEST] GetTime() = %08XH\n",GetTime()) ;
   if(!in_interrupt())
     {
		#ifdef USB_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] usb_new_device() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_SYSCAL_ASSERT
			#ifdef USB_WAIT_DLY_TSK // use dlytsk() to replace YIELD()
				// dlytsk(cur_task, DLY_TICKS, 1+TIME_PLUS) ;
				dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
             #else
            	YIELD();
            #endif
     }
    else
     {
      dprintf("[TEST] in interrupt !!\n");
     }
   // ^^^ for debug Prefetch   
 #endif   
 
 /* USB device state == configured ... tell the world! */
 // #ifndef USB_DBG_TEMP_DISABLE  // for debug Prefetch Error !!
 USB_DBG_INFO(iprintf(
	"[USB INFO] new device strings: Mfr=%d, Product=%d, "
    "SerialNumber=%d \n",
	dev->descriptor.iManufacturer, dev->descriptor.iProduct,
    dev->descriptor.iSerialNumber
    ));
    
 set_device_description (dev);
        
 // #ifdef DEBUG
 #ifdef USB_DBG_INIT_ON
    /*
	if (dev->descriptor.iManufacturer)
		usb_show_string(dev, "Manufacturer", dev->descriptor.iManufacturer);
	if (dev->descriptor.iProduct)
		usb_show_string(dev, "Product", dev->descriptor.iProduct);
	*/
	if (dev->descriptor.iSerialNumber)
		usb_show_string(dev, "SerialNumber", dev->descriptor.iSerialNumber);
 #endif

 /* now that the basic setup is over, add a /proc/bus/usb entry */
	/* put into sysfs, with device and config specific files */
 /*
 #ifdef CONFIG_USB_DEVICEFS
	err = device_add (&dev->dev);
	if (err)
		return err;
    USB_DBG_FLOW(iprintf(
       "[USB FLOW] usb_new_device() call usbdevfs_add_device() ... \n"));
	usb_create_driverfs_dev_files (dev);
 #endif // CONFIG_USB_DEVICEFS
 */

 /* Register all of the interfaces for this device with the driver core.
  * Remember, interfaces get bound to drivers, not devices. */
  
 for (i = 0; i < dev->actconfig->desc.bNumInterfaces; i++)
   {
	struct usb_interface *interface = &dev->actconfig->interface[i];
	struct usb_interface_descriptor *desc;

	desc = &interface->altsetting [interface->act_altsetting].desc;
	interface->dev.parent = &dev->dev;
	interface->dev.driver = NULL;
	interface->dev.bus = &usb_bus_type;
	interface->dev.dma_mask = parent->dma_mask;
	#ifdef USB_DBGSTRING_ON
		sprintf (&interface->dev.bus_id[0], "%d-%s:%d",
			 dev->bus->busnum, dev->devpath,
			 desc->bInterfaceNumber);
	 #else
		sprintf (&interface->dev.bus_id[0], "%d:%d",
			 dev->bus->busnum,
			 desc->bInterfaceNumber);
	#endif
	if (!desc->iInterface
	#ifdef USB_DBGSTRING_ON
		|| usb_string (dev, desc->iInterface,
			interface->dev.name,
			sizeof interface->dev.name) <= 0
	 #else
		|| usb_string(dev, desc->iInterface) <= 0
	#endif 
       )
       {
		/* typically devices won't bother with interface
		 * descriptions; this is the normal case.  an
		 * interface's driver might describe it better.
		 * (also: iInterface is per-altsetting ...)
		 */
		#ifdef USB_DBGSTRING_ON
			sprintf (&interface->dev.name[0],
				"usb-%s-%s interface %d",
				dev->bus->bus_name, dev->devpath,
				desc->bInterfaceNumber);
		 #else
         	/*
			sprintf (&interface->dev.name[0],
				"usb interface %d",
				desc->bInterfaceNumber);
            */
		#endif
	   }
 #ifdef USB_DBG_INFO_ON
	dev_dbg (&dev->dev, "%s - registering interface bus_id[%s]\n",
         __FUNCTION__, interface->dev.bus_id);
 #endif // USB_DBG_INFO_ON
	/*
	#ifdef CONFIG_USB_DEVICEFS
		device_add (&interface->dev);
		usb_create_driverfs_intf_files (interface);
	#endif // CONFIG_USB_DEVICEFS
	*/
   } // for i
   
 // #ifndef USB_DBG_TEMP_DISABLE  // for debug Prefetch Error !!
 #ifdef NO_SYS_FS
	/* find drivers willing to handle this device */
	USB_DBG_INIT(iprintf(
    	"[USB INIT] ==== New Device ====  ==8== (USB Device %d) find drivers \n",dev->devnum));
        
    // #ifdef USB_DBG_INIT_ON
	#ifdef USB_DBG_MEM_COUNT_ON
		iprintf(
    		"[USB MEM_COUNT] usb_new_device(8) : show USB Mem Count .. \n");
		showUsbMemC() ;
		showUsbMemU() ;
	#endif
 
	USB_DBG_FLOW(iprintf(
    	"[USB FLOW] usb_new_device() call usb_find_drivers() ... \n"));
	usb_find_drivers(dev);
	USB_DBG_INIT(iprintf("[USB INIT] ==== New Device ====  ==8-1== (USB Device %d) after find drivers \n",dev->devnum));
    // #ifdef USB_DBG_INIT_ON
	#ifdef USB_DBG_MEM_COUNT_ON
		iprintf("[USB MEM_COUNT] usb_new_device(8-1) : show USB Mem Count .. \n");
		showUsbMemC() ;
		showUsbMemU() ;
	#endif
 #endif // NO_SYS_FS
 

 // #ifndef USB_DBG_TEMP_DISABLE  // for debug Prefetch Error !!
 /* userspace may load modules and/or configure further */
 /*
 USB_DBG_FLOW(iprintf(
    "[USB FLOW] usb_new_device() call call_policy() ... \n"));
 call_policy("add", dev);
 */
 // #endif // ndef USB_DBG_TEMP_DISABLE for debug Prefetch Error !!
 
 /* add a /proc/bus/usb entry */
 usbfs_add_device(dev);
 
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_new_device() End ..\n"));
 return 0;
}

/**
 * usb_buffer_alloc - allocate dma-consistent buffer for URB_NO_DMA_MAP
 * @dev: device the buffer will be used with
 * @size: requested buffer size
 * @mem_flags: affect whether allocation may block
 * @dma: used to return DMA address of buffer
 *
 * Return value is either null (indicating no buffer could be allocated), or
 * the cpu-space pointer to a buffer that may be used to perform DMA to the
 * specified device.  Such cpu-space buffers are returned along with the DMA
 * address (through the pointer provided).
 *
 * These buffers are used with URB_NO_DMA_MAP set in urb->transfer_flags to
 * avoid behaviors like using "DMA bounce buffers", or tying down I/O mapping
 * hardware for long idle periods.  The implementation varies between
 * platforms, depending on details of how DMA will work to this device.
 * Using these buffers also helps prevent cacheline sharing problems on
 * architectures where CPU caches are not DMA-coherent.
 *
 * When the buffer is no longer used, free it with usb_buffer_free().
 */
void *usb_buffer_alloc (
	struct usb_device *dev,
	size_t size,
	int mem_flags,
	dma_addr_t *dma
)
{
 if(    !dev || !dev->bus || !dev->bus->op 
     || !dev->bus->op->buffer_alloc
   )
   {
	#ifdef USB_BS_ASSERT
	    iprintf("[ASSERT][USB] usb_buffer_alloc() : Error !!\n") ;
	#endif // USB_BS_ASSERT
	return 0;
   }
 return dev->bus->op->buffer_alloc (dev->bus, size, mem_flags, dma);
}

/**
 * usb_buffer_free - free memory allocated with usb_buffer_alloc()
 * @dev: device the buffer was used with
 * @size: requested buffer size
 * @addr: CPU address of buffer
 * @dma: DMA address of buffer
 *
 * This reclaims an I/O buffer, letting it be reused.  The memory must have
 * been allocated using usb_buffer_alloc(), and the parameters must match
 * those provided in that allocation request. 
 */
void usb_buffer_free (
	struct usb_device *dev,
	size_t size,
	void *addr,
	dma_addr_t dma
)
{
 if (!dev || !dev->bus || !dev->bus->op || !dev->bus->op->buffer_free) {
	#ifdef USB_BS_ASSERT
	    iprintf("[ASSERT][USB] usb_buffer_free() : Error !!\n") ;
	#endif // USB_BS_ASSERT
   	return;
   }
 dev->bus->op->buffer_free (dev->bus, size, addr, dma);
}

/**
 * usb_buffer_map - create DMA mapping(s) for an urb
 * @urb: urb whose transfer_buffer will be mapped
 *
 * Return value is either null (indicating no buffer could be mapped), or
 * the parameter.  URB_NO_DMA_MAP is added to urb->transfer_flags if the
 * operation succeeds.  If the device is connected to this system through
 * a non-DMA controller, this operation always succeeds.
 *
 * This call would normally be used for an urb which is reused, perhaps
 * as the target of a large periodic transfer, with usb_buffer_dmasync()
 * calls to synchronize memory and dma state.  It may not be used for
 * control requests.
 *
 * Reverse the effect of this call with usb_buffer_unmap().
 */
struct urb *usb_buffer_map (struct urb *urb)
{
 struct usb_bus		*bus;
 struct device		*controller;

 if(    !urb
	 || usb_pipecontrol (urb->pipe)
	 || !urb->dev
	 || !(bus = urb->dev->bus)
	 || !(controller = bus->controller)
   )
   {
	return 0;
   }

 if (controller->dma_mask)
   {
	#ifndef USB_NO_VMM // no vitrtul memory management
		urb->transfer_dma = dma_map_single (controller,
			urb->transfer_buffer, urb->transfer_buffer_length,
			usb_pipein (urb->pipe)
				? DMA_FROM_DEVICE : DMA_TO_DEVICE);
	 #else
     	/*
		// 2003/6/30 B. S. Liaw
		urb->transfer_dma = urb->transfer_buffer ;
		#ifdef USB_BS_ASSERT
		#ifdef _ADM5120
			if (!(  ((unsigned long)urb->transfer_dma & (unsigned long)0x20000000L)
                  ==((unsigned long)0x20000000L)
    		     )
	    	   )
			   {
				iprintf("[ASSERT][USB] usb_buffer_map() : " ) ;
    		    iprintf("Error : Not Un-Cached !!\n" ) ;
		        iprintf("[ASSERT][USB] urb->transfer_dma addr = %08XH \n", urb->transfer_dma) ;
		        iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
	    #endif // _ADM5120
		#endif // USB_BS_ASSERT
        */
	#endif    
	// FIXME generic api broken like pci, can't report errors
	// if (urb->transfer_dma == DMA_ADDR_INVALID) return 0;
   }
  else
   {
	#ifndef USB_NO_VMM
		urb->transfer_dma = ~0;
	#endif // ndef USB_NO_VMM
   }
 urb->transfer_flags |= URB_NO_DMA_MAP;
 return urb;
}

/**
 * usb_buffer_dmasync - synchronize DMA and CPU view of buffer(s)
 * @urb: urb whose transfer_buffer will be synchronized
 */
void usb_buffer_dmasync (struct urb *urb)
{
	struct usb_bus		*bus;
	struct device		*controller;

	if (!urb
			|| !(urb->transfer_flags & URB_NO_DMA_MAP)
			|| !urb->dev
			|| !(bus = urb->dev->bus)
			|| !(controller = bus->controller))
		return;

 if (controller->dma_mask)
   {
	#ifndef USB_NO_VMM
		dma_sync_single(controller,
			urb->transfer_dma, urb->transfer_buffer_length,
			usb_pipein (urb->pipe)
				? DMA_FROM_DEVICE : DMA_TO_DEVICE);
	#endif // ndef USB_NO_VMM
   }
}

/**
 * usb_buffer_unmap - free DMA mapping(s) for an urb
 * @urb: urb whose transfer_buffer will be unmapped
 *
 * Reverses the effect of usb_buffer_map().
 */
void usb_buffer_unmap (struct urb *urb)
{
 struct usb_bus		*bus;
 struct device		*controller;

	if (!urb
			|| !(urb->transfer_flags & URB_NO_DMA_MAP)
			|| !urb->dev
			|| !(bus = urb->dev->bus)
			|| !(controller = bus->controller))
		return;

	if (controller->dma_mask)
       {
		#ifndef USB_NO_VMM
			dma_unmap_single(controller,
				urb->transfer_dma, urb->transfer_buffer_length,
				usb_pipein (urb->pipe)
					? DMA_FROM_DEVICE : DMA_TO_DEVICE);
		#endif // ndef USB_NO_VMM
       }
	urb->transfer_flags &= ~URB_NO_DMA_MAP;
}

/**
 * usb_buffer_map_sg - create scatterlist DMA mapping(s) for an endpoint
 * @dev: device to which the scatterlist will be mapped
 * @pipe: endpoint defining the mapping direction
 * @sg: the scatterlist to map
 * @nents: the number of entries in the scatterlist
 *
 * Return value is either < 0 (indicating no buffers could be mapped), or
 * the number of DMA mapping array entries in the scatterlist.
 *
 * The caller is responsible for placing the resulting DMA addresses from
 * the scatterlist into URB transfer buffer pointers, and for setting the
 * URB_NO_DMA_MAP transfer flag in each of those URBs.
 *
 * Top I/O rates come from queuing URBs, instead of waiting for each one
 * to complete before starting the next I/O.   This is particularly easy
 * to do with scatterlists.  Just allocate and submit one URB for each DMA
 * mapping entry returned, stopping on the first error or when all succeed.
 * Better yet, use the usb_sg_*() calls, which do that (and more) for you.
 *
 * This call would normally be used when translating scatterlist requests,
 * rather than usb_buffer_map(), since on some hardware (with IOMMUs) it
 * may be able to coalesce mappings for improved I/O efficiency.
 *
 * Reverse the effect of this call with usb_buffer_unmap_sg().
 */

/*
#ifndef NO_SCATTER_LIST
   int usb_buffer_map_sg (struct usb_device *dev, unsigned pipe,
		struct scatterlist *sg, int nents)
   {
	struct usb_bus		*bus;
	struct device		*controller;

	if (!dev
			|| usb_pipecontrol (pipe)
			|| !(bus = dev->bus)
			|| !(controller = bus->controller)
			|| !controller->dma_mask)
		return -1;

	// FIXME generic api broken like pci, can't report errors
	return dma_map_sg (controller, sg, nents,
			usb_pipein (pipe) ? DMA_FROM_DEVICE : DMA_TO_DEVICE);
   }
#endif // ndef NO_SCATTER_LIST
*/
/**
 * usb_buffer_dmasync_sg - synchronize DMA and CPU view of scatterlist buffer(s)
 * @dev: device to which the scatterlist will be mapped
 * @pipe: endpoint defining the mapping direction
 * @sg: the scatterlist to synchronize
 * @n_hw_ents: the positive return value from usb_buffer_map_sg
 *
 * Use this when you are re-using a scatterlist's data buffers for
 * another USB request.
 */
/*
#ifndef NO_SCATTER_LIST
void usb_buffer_dmasync_sg (struct usb_device *dev, unsigned pipe,
		struct scatterlist *sg, int n_hw_ents)
{
	struct usb_bus		*bus;
	struct device		*controller;

	if (!dev
			|| !(bus = dev->bus)
			|| !(controller = bus->controller)
			|| !controller->dma_mask)
		return;

	dma_sync_sg (controller, sg, n_hw_ents,
			usb_pipein (pipe) ? DMA_FROM_DEVICE : DMA_TO_DEVICE);
}
#endif // ndef NO_SCATTER_LIST
*/
/**
 * usb_buffer_unmap_sg - free DMA mapping(s) for a scatterlist
 * @dev: device to which the scatterlist will be mapped
 * @pipe: endpoint defining the mapping direction
 * @sg: the scatterlist to unmap
 * @n_hw_ents: the positive return value from usb_buffer_map_sg
 *
 * Reverses the effect of usb_buffer_map_sg().
 */

/*
#ifndef NO_SCATTER_LIST
	// struct scatterlist is on include/asm-xxx/scatterlist.h
	void usb_buffer_unmap_sg (struct usb_device *dev, unsigned pipe,
		struct scatterlist *sg,
         int n_hw_ents)
{
	struct usb_bus		*bus;
	struct device		*controller;

	if (!dev
			|| !(bus = dev->bus)
			|| !(controller = bus->controller)
			|| !controller->dma_mask)
		return;

	dma_unmap_sg (controller, sg, n_hw_ents,
			usb_pipein (pipe) ? DMA_FROM_DEVICE : DMA_TO_DEVICE);
}
#endif // ndef NO_SCATTER_LIST
*/

struct bus_type usb_bus_type = {
 #ifdef USB_DBGSTRING_ON
	.name =		"usb",
 #endif // USB_DBGSTRING_ON
 .match =	usb_device_match,
 .hotplug =	usb_hotplug,
};

#ifndef _ETCPIP
   #ifndef MODULE

   static int __init usb_setup_disable(char *str)
   {
	nousb = 1;
	return 1;
   }

   /* format to disable USB on kernel command line is: nousb */
   __setup("nousb", usb_setup_disable);

   #endif
#endif // ndef _ETCPIP

/*
 * for external read access to <nousb>
 */
int usb_disabled(void)
{
	return nousb;
}

/*
 * Init
 */
#ifndef _ETCPIP
  static 
#endif
int __init 
usb_init(void)
{
 // #ifdef _ETCPIP
 #ifndef NO_TIMER
	int status1 ;
 #endif // ndef NO_TIMER
 // #endif // _ETCPIP

 USB_DBG_FLOW(iprintf("[USB FLOW] usb_init() Begin ..\n"));
 
 if (nousb) {
	 info("USB support disabled\n");
	 return 0;
	}
    
 // #ifdef USB_MEM_LIB
	// initialize fixed size memory block pool
    // this need Non-Cache Area
 if ( ! UsbMem_init())
   {
	USB_DBG_INIT(iprintf(
    	"[USB INIT] usb_init() : Call UsbMem_init() Failed !!\n")
        ) ;
	dprintf("[USB] Error : Memory Initialize Failed !!\n") ;
    return 0 ;
   }
 /*  
 #ifdef USB_DBG_MEM_C_ON
   showUsbMemC() ;
 #endif
 */
 if ( ! initUsbPool())
   {
	USB_DBG_INIT(iprintf(
    	"[USB INIT] usb_init() : Call initUsbPool() Failed !!\n")
        ) ;
	dprintf("[USB] Error : Memory Initialize Failed !!\n") ;
    return 0 ;
   }
 // #endif // USB_MEM_LIB
    
 #ifdef _ETCPIP
	#ifndef NO_TIMER
 
    in_isr = 0 ;
    init_timervecs();
    
    // status1 = runtsk(100, usb_timer_task, 6000); // 100 too high !!
    // status1 = runtsk(80, usb_timer_task, 6000);
	#ifdef INT_SCHEDULING
	    status1 = runtsk(100, (void (fAR *)(void))usb_timer_task, 6000);
	#else
	    status1 = runtsk(80, (void (fAR *)(void))usb_timer_task, 6000); 
        				// 100 too high !!
	#endif // INT_SCHEDULING

    if (status1 < SUCCESS )
       {
        dprintf("RUNTASK usb_timer_task fails\n");
		return 0;
       }
      else
       {
        dprintf("RUNTASK id=%d usb_timer_task ...\n", status1);
       }
	#endif // ndef NO_TIMER
 #endif // _ETCPIP
         
 
 #ifndef NO_KSET
	bus_register(&usb_bus_type);
  #else
	bus_register(&usb_bus_type,&usb_driver_list,&usb_bus_list) ;
 #endif

 usb_major_init();
 #ifdef CONFIG_USB_DEVICEFS
	usbfs_init();
 #endif // CONFIG_USB_DEVICEFS
 
 usb_hub_init();
 /*
 #ifndef NO_SYS_FS
 	driver_register(&usb_generic_driver);
 #endif // ndef NO_SYS_FS
 */
 USB_DBG_FLOW(iprintf("[USB FLOW] usb_init() End.\n"));
 return 0;
}

/*
 * Cleanup
 */
#ifndef _ETCPIP
   /*
   static void __exit usb_exit(void)
   {
	// This will matter if shutdown/reboot does exitcalls.
	if (nousb)
		return;

	driver_unregister(&usb_generic_driver);
	usb_major_cleanup();
    #ifdef CONFIG_USB_DEVICEFS
	usbfs_cleanup();
    #endif // CONFIG_USB_DEVICEFS
	usb_hub_cleanup();
	bus_unregister(&usb_bus_type);
   }
   */
#endif // ndef _ETCPIP

#ifndef _ETCPIP
/*
subsys_initcall(usb_init);
module_exit(usb_exit);
*/
/*
 * USB may be built into the kernel or be built as modules.
 * These symbols are exported for device (or host controller)
 * driver modules to use.
 */
/*
EXPORT_SYMBOL(usb_epnum_to_ep_desc);

EXPORT_SYMBOL(usb_register);
EXPORT_SYMBOL(usb_deregister);
EXPORT_SYMBOL(usb_disabled);

EXPORT_SYMBOL(usb_device_probe);
EXPORT_SYMBOL(usb_device_remove);

EXPORT_SYMBOL(usb_alloc_dev);
EXPORT_SYMBOL(usb_put_dev);
EXPORT_SYMBOL(usb_get_dev);
EXPORT_SYMBOL(usb_hub_tt_clear_buffer);

EXPORT_SYMBOL(usb_driver_claim_interface);
EXPORT_SYMBOL(usb_interface_claimed);
EXPORT_SYMBOL(usb_driver_release_interface);
EXPORT_SYMBOL(usb_match_id);
EXPORT_SYMBOL(usb_find_interface);
EXPORT_SYMBOL(usb_ifnum_to_if);

EXPORT_SYMBOL(usb_new_device);
EXPORT_SYMBOL(usb_reset_device);
EXPORT_SYMBOL(usb_connect);
EXPORT_SYMBOL(usb_disconnect);

EXPORT_SYMBOL(__usb_get_extra_descriptor);

EXPORT_SYMBOL(usb_get_current_frame_number);

EXPORT_SYMBOL (usb_buffer_alloc);
EXPORT_SYMBOL (usb_buffer_free);

EXPORT_SYMBOL (usb_buffer_map);
EXPORT_SYMBOL (usb_buffer_dmasync);
EXPORT_SYMBOL (usb_buffer_unmap);

EXPORT_SYMBOL (usb_buffer_map_sg);
EXPORT_SYMBOL (usb_buffer_dmasync_sg);
EXPORT_SYMBOL (usb_buffer_unmap_sg);

MODULE_LICENSE("GPL");
*/
#endif // ndef _ETCPIP


// B. S. Liaw TEST for compiler warning !!


