

// Name of file : UsbLpt.c
// ====================================================================

/********************************************************************/
/*       Printer Hardware Layer driver for USB LPT                  */
/*------------------------------------------------------------------*/
/*      Copyright (C) 2003 BRNT			                    */
/*------------------------------------------------------------------*/
/*      Modified, programmed by B. S. Liaw                          */
/*                                                                  */
/*      Description : 2003/4/3 first edited for SS2510 USB          */
/*                             (from w83877.h)                      */
/********************************************************************/


// ====================================================================

//U8 Prt_Init(void);    //Init. & check the port connection
//void Prt_Open(void);  //Send Cmd. to the printer for kick off signal
//U8 Prt_Buf_Write(U8 *, U32, U32 *);   //Send data to the Super I/O data port, hardware
//void Prt_Close(void);
//
// In the following, only 4 functions can be called:
//
// if( Prt_Init() )
// {
//      Prt_Open();
//      ...
//      .
//      Prt_Buf_Write(buf, length, &byte_sent);
//      .
//      Prt_Buf_Write(buf, length, &byte_sent);
//      ...
//      Prt_Close();
// }
//Then, the printer can print out.

// These three functions shown previously, Prt_Open(), Prt_Buf_Write(),
//                                  Prt_Close(), are formed as a group.
// I.e., during a printing session,
//     Prt_Open() must be executed first, (called once)
//     then Prt_Buf_Write(), can be called many times to send user data
//                           out to the printer
//                              finally, call Prt_Close() to close this printing session
// ======================================================================


// #ifdef _ETCPIP
#ifdef _NEWLIB
  #include "sys_type.h"
  #include "time.h"
#endif
#include "feature.h"
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "UsbLptDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
// #endif // _ETCPIP

#ifndef _ETCPIP
   #include <linux/lp.h>
 #else
   #include "lp.h"
#endif
#ifndef _ETCPIP
   #include <linux/devfs_fs_kernel.h>
#endif // ndef _ETCPIP

/*
#ifndef _ETCPIP
   #undef DEBUG
 #else
   #ifdef CONFIG_USB_DEBUG
      #define DEBUG
    #else
	  #undef DEBUG
   #endif
#endif
*/
#ifndef _ETCPIP
   #include <linux/usb.h>
 #else
   #include "usb.h"
#endif

#ifdef _ETCPIP
	#include "prdrv.h"

    #include "UsbLpt.h" // it is modified from w83877tf.h
                        // Printer Hardware Layer driver Header for USB
    #include "UsbLptIo.h"
#endif // _ETCPIP
#ifdef _USB_SYSLOG
	#include "syslog.h"
	#include "logStr.h"
	#include "config.h"
	extern struct III_Config_t gSetting;
 #else
	// #error: Please Make Sure : Not _USB_SYSLOG !!
	//#warning: Not _USB_SYSLOG !!
#endif // _SYSLOG

// ====================================================================

#ifndef USBLPT_READ_BY_CHECK
	static int read_from_printer(char *buf, int bufsize) ;
#endif // ndef USBLPT_READ_BY_CHECK
/*
static usb_init_string(struct usblp *usblp) ;
*/

// ====================================================================

#ifdef _ETCPIP
	int volatile usblp_plug_out = 1 ; // 1 : the USB Cable Plug Out 
	int volatile usb_lpt_ok = 0 ;     // 1 : Driver Init Ok
#endif // _ETCPIP
/*
struct usblp *usblp  = usblp_table[USBLP_MINOR_BASE];
*/
/*
struct usblp volatile *usblp  ;
*/
/*
struct usblp usblp_o  ;
*/
// struct usblp *usblp  ;

// ====================================================================

/**********************************************************************
Name: U8 Prt_Init(void)
Purpose: Init. & check the port connection
Input: none
Output: 1: everything is O.K.; otherwise, return 0
Note:
**********************************************************************/

unsigned char Prt_Init(void)
{
 /*
 #ifdef _USB_SYSLOG
    char usbsyslog_buf[100];
 #endif // _SYSLOG
 */


 USBLPT_DBG_FLOW(iprintf(
	"[USBLPT FLOW] Prt_Init() Begin ..\n"
    ));
 USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Init() Begin ..\n"));
 USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Init() Printer Initialize .... \n"));
 // dprintf("[USBLPT] Printer Initialize .... \n");

 /*
 #if (SUPER_IO==0) // WinBond
   Prt_Config_83877_PPM(D_EF_CR0_ECP_EPP);      // Set Parallel Port Mode
 #elif (SUPER_IO==1) // SMSC
   Prt_Config_83877_PPM(D_EF_CR1_ECP_EPP);      // Set Parallel Port Mode
 #endif

   D_ECP_ECR=0x10;  // Standard Parallel port, disable nFault int. 
   D_ECP_DCR=0x0c;  // Out, disbale nACK int., pin: nSLIN=0, INIT=1, Autofd=1, nStrobe=1
 */
 // if(( usb_lpt_ok != 1 ) || ( usblp_plug_out >= 1 ))
 if( usblp_plug_out >= 1 )
   {
	USBLPT_DBG_INFO(iprintf(
		"[USBLPT INFO] Prt_Init() : Usb Lpt Plug Out !!\n"
        ));
	dprintf("[USBLPT] Printer Initialize Fail : Plug Out !!\n");
    return(D_Fail); 
   }
 /*  
 BLOCK_PREEMPTION;
 // usblp  = usblp_table[USBLP_MINOR_BASE];
 // usblp  = &usblp_o ;
 usblp = global_usblp ;
 UNBLOCK_PREEMPTION;
 */
 
 // Prt_Addr_Byte_Write( 0x76, 0);
 // Prt_Addr_Write(0xe8);

 if(( usb_lpt_ok <= 0 ) || ( usblp_plug_out >= 1 ))
   {
    return(D_Fail); 
   }
 if (!global_usblp) {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Init() : Null global_usblp !!\n"));
    return(D_Fail); 
   }

 /* 2003/90/27 B. S. Liaw , comment out
 if(usblp_open(global_usblp) != 0 )
   {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Init() : Open Fail !!\n"
        ));
    
	dprintf("[USBLPT] Printer Initialize : Open Fail !!\n");
    
    return(D_Fail); 
   }
 */

 if(global_usblp->opened != 1) {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Init() : Open Fail !!\n"
        ));
    
	dprintf("[USBLPT] Printer Initialize : Open Fail !!\n");

	dprintf("[USBLPT] Printer Initialize Fail (Open) !!\n");
	USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Init(1) End with Fail !!\n"));
    return(D_Fail); 
   }
  else {
    USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Init() : Usb LPT already Opened !!\n")) ;
   }

 /* 2003/7/7 B. S. Liaw , Move to UsbLptIo.c
 
 if (usblp->quirks & USBLP_QUIRK_USB_INIT)
   {
	USBLPT_DBG_INFO(iprintf(
		"[USBLPT INFO] QUIRK USB INIT ..\n"
        ));
	USBLPT_DBG_FLOW(iprintf(
		"[USBLPT FLOW] Prt_Init() call usb_init_string() ..\n"
        ));
	if (usb_init_string(usblp) != D_OK)
       {
	    dprintf("[USBLPT] init string Fail !!\n");
       }
   }
 */
 // B. S. Liaw : Donot Reset here , otherwise the old print data will loss !!
 // USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Init() : Prt_Reset() .. \n"));
 // Prt_Reset() ;
 
 // Check the Printer port if it's powered & connected.
 if( Prt_Check_Port_Pwr_Connect() != D_OK)
   {
	dprintf("[USBLPT] Printer Initialize Fail (Power) !!\n");
	USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Init(2) End with Fail !!\n"));
    return(D_Fail);     //Check the Printer port if it's powered & connected.
   }
    
 /*
 // ??? 2003/10/01 B. S. Liaw BugFix
 // --> 2003/10/02 : Comment out this !!

 USBLPT_DBG_INFO(iprintf(
	"[USBLPT INFO] Prt_Init() : Write FF (Form Feed) (0CH) ..\n"));
 
 USBLPT_DBG_FLOW(iprintf(
		"[USBLPT FLOW] Prt_Init() call Prt_Write_Byte() ..\n"));

 // 0CH : FF , Form Feed
 if ( Prt_Write_Byte(0x0C) == D_False) {
	
	//		USBLPT_DBG_FLOW(iprintf(
    //    		"[USBLPT FLOW] Prt_Buf_Write() End with Fail !!\n"
    //        	));
	//		return(D_False);
	
	USBLPT_DBG_INFO(iprintf(
		"[USBLPT INFO] Prt_Init() : Write FF (Form Feed) (0CH) Fail !! \n"));
	dprintf("[USBLPT] Printer Form Feed Fail !! \n");
   }
 */      
 /*
   D_ECP_DCR=0x0c;   // Out, disbale nACK int., pin: nSLIN=0, INIT=1, Autofd=1, nStrobe=1 
 */
 iprintf("[USBLPT] Printer Initialize Ok. \n");

 #ifdef _USB_SYSLOG
    //sprintf(usbsyslog_buf,"[USBLPT] Printer Initialize Ok. ") ;
				#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_INIT_OK
		//sprintf_SetSystemLog(USB_SYS_LOG_PR_INIT_OK);
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_PR_INIT_OK);
#else
		//sprintf_SetSystemLog("[USBLPT] Printer Initialize Ok. ");
		sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Printer Initialize Ok. ");
#endif // USB_SYS_LOG_PR_INIT_OK
	}
 #endif // _SYSLOG
 
 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Init() End. \n"));
 return(D_OK);
} // function Prt_Init()

/**********************************************************************
Name: U8 Prt_Check_Port_Pwr_Connect(void)
Purpose: Check the Printer port if it's powered & connected.
Input: none
Output: 1: everything is O.K.; otherwise, return 0
Note:
**********************************************************************/

unsigned char Prt_Check_Port_Pwr_Connect(void)
{
 /*
 U8 LV_tmp;
 
 LV_tmp = D_Prt_Mask_S_EPP_sts_2_0 & D_S_EPP_sts_prt;
  // Mask b2-0 of D_S_EPP_sts_prt
 
 #if NBDEBUG >= 1
   dprintf("Prt_Get_S_EPP_Sts()=%x\n", LV_tmp);
 #endif
   if(LV_tmp == D_Prt_No_Connect)
 */
 
 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Check_Port_Pwr_Connect() Begin ..\n"));
 
 // if(( usb_lpt_ok != 1 ) || ( usblp_plug_out >= 1 ))
 if( usblp_plug_out >= 1 )
   {
    /*
	dprintf("[USBLPT] Port not connected or no power applied !\n");
	dprintf("[USBLPT] Check the Printer Power and USB Cable, please !\n");
    */
    USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Prt_Check_Port_Pwr_Connect()--1  usblp_plug_out >= 1 !!\n")) ;
    USBLPT_DBG_INIT(
        iprintf("[USBLPT INIT] Port not connected or no power applied !!\n") ;
        iprintf("[USBLPT INIT] Check the Port Connection , and Power on the Printer, please !\n");
       ) ;
	#ifdef USBLPT_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
		   ) ) {
			iprintf("[ASSERT][USB] Prt_Check_Port_Pwr_Connect() : " ) ;
   		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
		    iprintf("[ASSERT][USB] Continue ..\n") ;
   		   }
	#endif // USBLPT_BS_SYSCAL_ASSERT

	#ifdef USBLPT_READ_BY_URBTASK
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
	#endif // USBLPT_READ_BY_URBTASK
	USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Check_Port_Pwr_Connect() End With Faul !!\n"));
    return(D_Fail);
   }
   
 /*
 if(   ((LV_tmp & D_Prt_No_power_Mask) == D_Prt_No_power)
    || ((LV_tmp & D_Prt_No_power_Mask)== D_Prt_No_power_Mask )
   )
        {
         iprintf("Port is connected, but no power applied !\tPower on the Printer, please !\n");
         return(D_Fail);
        }
 */
 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Check_Port_Pwr_Connect() End.\n"));
 return(D_OK);
} // function U8 Prt_Check_Port_Pwr_Connect(void)

#if defined(_NB_SHARE) || (_LPD >= 1)

/*************************************************/
unsigned char Prt_Get_S_EPP_Sts(void)
{
 unsigned char LV_ret;

 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Get_S_EPP_Sts() Begin ..\n"));
 /*
 LV_ret= D_S_EPP_sts_prt;
 */
 if(( usb_lpt_ok != 1 ) || ( usblp_plug_out >= 1 ))
   {
	LV_ret= 0 ;
   }
  else
   {
	LV_ret=   PARPORT_STATUS_SELECT | PARPORT_STATUS_ERROR
    		| PARPORT_STATUS_BUSY & ~PARPORT_STATUS_PAPEREND;
   }
 
 USBLPT_DBG_FLOW(iprintf(
	"[USBLPT FLOW] Prt_Get_S_EPP_Sts() End.\n"));
 return(LV_ret);
}
 #else
	// #error : Please Make Sure !!
#endif // defined(_NB_SHARE) || (_LPD >=1)

/**********************************************************************
Name:void Prt_Open( void)
Purpose: Send Cmd. to the printer for kick off signal
Input: Set or clear Auto feed function
Output: none
Note:
**********************************************************************/

void Prt_Open(void)
{
 USBLPT_DBG_FLOW(iprintf(
	"[USBLPT FLOW] Prt_Open() Begin ..\n"));
 if(( usb_lpt_ok != 1 ) || ( usblp_plug_out >= 1 ))
   {
	dprintf("[USBLPT] Printer Open Fail !!\n");
	USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Open() End with Fail !!\n"));
	return ;
   }
   
 /*   
 // usblp  = usblp_table[USBLP_MINOR_BASE];
 // usblp  = &usblp_o ;
 usblp = global_usblp ;
 
 // Prt_Addr_Byte_Write( 0x76, 0);
 // Prt_Addr_Write(0xe8);
 usblp_open(usblp) ;
 
 if (usblp->quirks & USBLP_QUIRK_USB_INIT)
   {
	USBLPT_DBG_FLOW(iprintf(
		"[USBLPT FLOW] Prt_Open() call usb_init_string() ..\n"));
	if (usb_init_string(usblp) != D_OK)
       {
	    dprintf("[USBLPT] init string Fail !!\n");
       }
   }
 */
 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Open() End.\n"));
}


// ====================================================================

#ifndef USBLPT_READ_BY_CHECK
   /*
   static int
   read_from_printer(int fd, char *buf, int bufsize)
   */
   static int read_from_printer(char *buf, int bufsize)
   {
	#ifdef HAVE_POLL
	  struct pollfd ufds;
	#endif
	int status;
	int retry = 5;
	/* 
	status = 0 ;
	dprintf("[USBLPT] read_from_printer() : Not Do Anything !! \n") ;
	*/
 
	memset(buf, 0, bufsize);
 
	do
	   {
		#ifdef HAVE_POLL
	      ufds.fd = fd;
	      ufds.events = POLLIN;
	      ufds.revents = 0;
	      (void) poll(&ufds, 1, 1000);
		#endif
    
 if(( usb_lpt_ok <= 0 ) || ( usblp_plug_out >= 1 ))
   {
    return(D_Fail); 
   }
		if (!global_usblp) {
			USBLPT_DBG_INFO(iprintf("[USBLPT INFO] read_from_printer() : Null global_usblp !!\n"));
		    return(D_Fail); 
		   }
 if(global_usblp->opened != 1) {
    return(D_Fail); 
   }

	  	USBLPT_DBG_INIT(iprintf("[USBLPT INIT] read_from_printer() : read ..\n" )) ;
    
		// status = read(fd, buf, bufsize - 1);
    
	    status = usblp_read(global_usblp , buf, bufsize - 1);
        
		#ifndef USBLPT_READ_NO_DELAY
		#ifdef USBLPT_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] wait_ms() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
		#endif // ndef USBLPT_READ_NO_DELAY
    
	    if (status <= 0)
    	   {
			USBLPT_DBG_INIT(iprintf("[USBLPT INIT] read_from_printer() : Error : Cannot read from printer \n" )) ;
			USBLPT_DBG_INIT(iprintf("[USBLPT INIT] read_from_printer() : sleep .. \n" )) ;
        	sleep(1);
			USBLPT_DBG_INIT(iprintf("[USBLPT INIT] read_from_printer() : after sleep \n" )) ;
    	   }
	   }
	  while ((status == 0) && (--retry != 0));
 
	 return status;
   }
#endif // ndef USBLPT_READ_BY_CHECK

// ====================================================================


/**********************************************************************
Name:void Prt_Buf_Write(U8 *LV_Ptr, U32 LV_Legth, U32 *byte_sent)
Purpose: Send data to the Super I/O data port, hardware
Input: data pointer & length
Output: D_OK: everything is O.K.;
        D_False: timeout has been reached.
Note: If there is any error during sending to the printer,
      number of byte sent will be reported to (*byte_sent).
**********************************************************************/

unsigned char Prt_Buf_Write(unsigned char *LV_Ptr,
    unsigned long LV_Legth, unsigned long *byte_sent)
{
 // #ifndef NO_PRT_WRITE_BYTE
    // 2003/7/7 B. S. Liaw BugFix , remove unsigned for j
    long j ;
    unsigned long j_ul ;
 // #endif // ndef NO_PRT_WRITE_BYTE
 int i ;
 int status;
 char buf[1024];
 // char buf[128];
 int retry ;
 #ifdef USBLPT_READ_BEFORE_WRITE     // ok for Epson Stylus CX5100
	int read_retry ;
 #endif // USBLPT_READ_BEFORE_WRITE
 /*
 #ifdef _USB_SYSLOG
    char usbsyslog_buf[100];
 #endif // _SYSLOG
 */

 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Buf_Write() Begin ..\n"));

 if(( usb_lpt_ok <= 0 ) || ( usblp_plug_out >= 1 ))
   {
    return(D_Fail); 
   }
 if (!global_usblp) {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Buf_Write() : Null global_usblp !!\n"));
    return(D_Fail); 
   }
 if(global_usblp->opened != 1) {
    return(D_Fail); 
   }

 USBLPT_DBG_SPOOL(iprintf(
	"[USBLPT SPOOL] Prt_Buf_Write(1) : LV_Legth = %lu\n",
	LV_Legth )) ;

 // ----------------------------------------------------  

 #ifdef USBLPT_READ_BEFORE_WRITE     // ok for Epson Stylus CX5100
	#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
		if (global_usblp->bidir) {
		    for(read_retry=0;read_retry<3;read_retry++)
		    // for (read_retry=0;read_retry<30;read_retry++)
	    	   {
				if(( usb_lpt_ok <= 0 ) || ( usblp_plug_out >= 1 ))
				   {
				    return(D_Fail); 
				   }
				if (!global_usblp) {
					USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Buf_Write() : Null global_usblp !!\n"));
				    return(D_Fail); 
				   }
				 if(global_usblp->opened != 1) {
			    	return(D_Fail); 
				   }
				for(i=0; i < 12 ; i++)
				   {
    				buf[i] = 0x00 ;
				   }
 	
				USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Prt_Buf_Write() : read ..\n")) ;
 
				#ifndef USBLPT_READ_BY_CHECK
					status = read_from_printer(buf, 1024);
				 #else
					// status = usblp_read(usblp , buf, 128-1);
					status = usblp_read(global_usblp , buf, 1024-1);
				#endif
        
				#ifdef USBLPT_BS_SYSCAL_ASSERT
					if (!(  mt_busy == 0
				         )
					   )
					   {
						iprintf("[ASSERT][USBLPT] Prt_Buf_Write() : " ) ;
    				    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
					    iprintf("[ASSERT][USBLPT] Continue ..\n") ;
	    			   }
				#endif // USBLPT_BS_SYSCAL_ASSERT

				#ifndef USBLPT_READ_NO_DELAY
					dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
				#endif // ndef USBLPT_READ_NO_DELAY
				/* 
				for (i=0; i < 12 ; i++)
				   {
					if (buf[i])
	    			   {
						dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
						USBLPT_DBG_INFO(iprintf("[USBLPT INIT] Prt_Buf_Write() : READ %d : %02X \n",i ,buf[i])) ;
			    	   }
				   } // for i
				// dprintf("\n",buf[i]) ;
    			*/
     
				if (status < 0)
    			   {
				    USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Buf_Write() : Error : Cannot read from printer \n" )) ;
					// dprintf("[USBLPT] Error : Cannot read from printer \n") ;
	                  goto read_ok ;
				   }
				for (i=0; i < 12 ; i++)
				   {
					if (buf[i])
		    		   {
        	    	    /*
						dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
    	        	    */
						USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Buf_Write() : READ %d : %02X \n",i ,buf[i])) ;
						// USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Prt_Buf_Write() : READ %d : %02X \n",i ,buf[i])) ;
    				   }
				   } // for i
				if (!buf[0])
			       {
        		    /*
					dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
		            */
					// USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Buf_Write() : ReRead Ok ! \n")) ;
        	        if (global_usblp->quirks & USBLPT_QUIRK_INIT_EPSON)
            		   {
 			    	    USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Prt_Buf_Write() : ReRead Ok ! \n")) ;
	                   	goto read_ok ;
		    	       }
					USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Buf_Write() : Re Read .. \n" )) ;
    			   }
	    	      else
    	    	   {
	                /*
					if (usblp->quirks & USBLP_QUIRK_INIT_HP1000)
        	    	   {
						USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Prt_Buf_Write() : ReRead Ok ! \n")) ;
    		    	    goto read_ok ;
		               }
    	            */
					USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Buf_Write() : Re Read .. \n" )) ;
	    	       }
		
		       } // for read_retry
      
		    // ----------------
			read_ok:
		    // ----------------
    	  
		   } // if (usblp->bidir)
	#endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw
 #endif // USBLPT_READ_BEFORE_WRITE

 // ----------------------------------------------------  
 
 #ifdef USBLPT_CHECK_BEFORE_WRITE
	 // Calvin
	 if(Prt_Check_Port_Pwr_Connect() != D_OK) {
		*byte_sent = (unsigned long) 0L ;
		return D_False;
	   }

	#ifdef USBLPT_BS_SYSCAL_ASSERT
		if (!(  mt_busy == 0
		   ) ) {
			iprintf("[ASSERT][USB] Prt_Buf_Write() : " ) ;
    		iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			iprintf("[ASSERT][USB] Continue ..\n") ;
    	   }
	#endif // USB_BS_ASSERT

	#ifndef USBLPT_READ_NO_DELAY
		dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
	#endif // ndef USBLPT_READ_NO_DELAY
 #endif // USBLPT_CHECK_BEFORE_WRITE
   
 // ----------------------------------------------------  

 led_usb_flash(1);

 USBLPT_DBG_WRITE(iprintf(
	"[USBLPT WRITE] Prt_Buf_Write() : before LV_Legth = %d \n",	LV_Legth ));
 // USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Prt_Buf_Write() : before LV_Legth = %d \n",	LV_Legth ));
 /*    
 #ifdef NO_PRT_WRITE_BYTE
 */
 for (retry=3 ; retry > 0 ; retry -- ) // 3-1 = 2 times
   {
	if(( usb_lpt_ok <= 0 ) || ( usblp_plug_out >= 1 ))
	   {
	   	led_usb_flash(0);
	    return(D_Fail); 
	   }
	if (!global_usblp) {
		USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Buf_Write() : Null global_usblp !!\n"));
		led_usb_flash(0);
	    return(D_Fail); 
	   }
	if(global_usblp->opened != 1) {
		led_usb_flash(0);
	    return(D_Fail); 
	   }

	USBLPT_DBG_INIT(iprintf(
		"[USBLPT INIT] Prt_Buf_Write() : Write , LV_Legth = %d .. \n",
		LV_Legth )) ;
	USBLPT_DBG_FLOW(iprintf(
		"[USBLPT FLOW] Prt_Buf_Write() call usblp_write() ..\n"));
    
    /*
	usblp_write( usblp , LV_Ptr , LV_Legth ) ;
    */
	// if (usblp_write( usblp , LV_Ptr , LV_Legth ) < LV_Legth)
	j = usblp_write( global_usblp, LV_Ptr , LV_Legth ) ;

	// dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
    
	USBLPT_DBG_WRITE(iprintf(
		"[USBLPT WRITE] Prt_Buf_Write() : after LV_Legth = %d , j = %d\n",
		LV_Legth, j));
	USBLPT_DBG_SPOOL(iprintf(
		"[USBLPT SPOOL] Prt_Buf_Write() : after LV_Legth = %d , j = %d\n",
		LV_Legth, j));
	// USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Prt_Buf_Write() : after LV_Legth = %d , j = %d\n",LV_Legth, j));
        
	if (j < 0 )
       { 
		iprintf("[USBLPT] Error : Cannot write to printer \n") ;
		#ifdef _USB_SYSLOG
		    //sprintf(usbsyslog_buf,"[USBLPT] Error : Cannot write to printer ");
						#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_WR_FAIL
				//sprintf_SetSystemLog(USB_SYS_LOG_PR_WR_FAIL);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_PR_WR_FAIL);
#else
				//sprintf_SetSystemLog("[USBLPT] Error : Cannot write to printer ");
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Error : Cannot write to printer ");
#endif // USB_SYS_LOG_PR_WR_FAIL
			}
		#endif // _SYSLOG

		if(( usb_lpt_ok <= 0 ) || ( usblp_plug_out >= 1 ))
		   {
		   	led_usb_flash(0);
		    return(D_Fail); 
		   }
		if (!global_usblp) {
			USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Buf_Write() : Null global_usblp !!\n"));
			led_usb_flash(0);
		    return(D_Fail); 
		   }
		 if(global_usblp->opened != 1) {
		 	led_usb_flash(0);
		    return(D_Fail); 
		   }

		iprintf("[USBLPT] Retry .. \n") ;

		#ifdef _USB_SYSLOG
		    //sprintf(usbsyslog_buf,"[USBLPT] Retry .. ");
						#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_RETRY
				//sprintf_SetSystemLog(USB_SYS_LOG_PR_RETRY);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_PR_RETRY);
#else
				//sprintf_SetSystemLog("[USBLPT] Retry .. ");
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Retry .. ");
#endif // USB_SYS_LOG_PR_RETRY
			}
		#endif // _SYSLOG

		#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
			if (global_usblp->bidir)
			   {
				for (i=0; i < 12 ; i++)
				   {
	    			buf[i] = 0x00 ;
				   }
 
				USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Prt_Buf_Write() : read ..\n")) ;
    	    
				#ifndef USBLPT_READ_BY_CHECK
					status = read_from_printer(buf, 1024) ;
				 #else
					// status = usblp_read(usblp , buf, 128-1);
					status = usblp_read(global_usblp , buf, 1024-1);
				#endif
        
				#ifdef USBLPT_BS_SYSCAL_ASSERT
				if (!(  mt_busy == 0
			         )
				   )
				   {
					iprintf("[ASSERT][USB] Prt_Buf_Write() : " ) ;
	    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
				    iprintf("[ASSERT][USB] Continue ..\n") ;
    			   }
				#endif // USB_BS_ASSERT

				#ifndef USBLPT_READ_NO_DELAY
					dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
				#endif // ndef USBLPT_READ_NO_DELAY
 
				for (i=0; i < 12 ; i++)
				   {
					if (buf[i])
			    	   {
						#ifdef USBLPT_BS_SYSCAL_ASSERT
						if (!(  mt_busy == 0
					         )
						   )
						   {
							iprintf("[ASSERT][USB] Prt_Buf_Write() : " ) ;
		    			    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
						    iprintf("[ASSERT][USB] Continue ..\n") ;
			    		   }
						#endif // USB_BS_ASSERT
						dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
						USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Prt_Buf_Write() : READ %d : %02X \n",i ,buf[i])) ;
			    	   }
				   } // for i
				// dprintf("\n",buf[i]) ;
			   } // if (usblp->bidir)
		#endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw

		if (retry >= 1 )
           {
			USBLPT_DBG_INFO(iprintf("[USBLPT INFO] ==== Retry %d .. \n",retry));
			wait_ms(200) ; 
			// #ifdef USBLPT_READ_BY_URBTASK
			// dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
			// #endif // USBLPT_READ_BY_URBTASK
            continue ;
           }
        
		// Report back No byte sent to the printer
		*byte_sent = (unsigned long) 0L ;
		USBLPT_DBG_SPOOL(iprintf(
			"[USBLPT SPOOL] Prt_Buf_Write(2) : *byte_sent = %lu\n",
			*byte_sent )) ;
        
		USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Buf_Write() End with Fail !!\n"));
		led_usb_flash(0);
		return(D_False);
       }
	// else if (j < LV_Legth)
	else if (j <= LV_Legth)
       { 
        j_ul = (unsigned long) j ;

		// Report back the number of byte sent to the printer
//if (j < LV_Legth)
//	iprintf("Prt_Buf_Write:LV=%ld, j=%ld\n",LV_Legth, j_ul);
		*byte_sent=j_ul;

		USBLPT_DBG_SPOOL(iprintf(
			"[USBLPT SPOOL] Prt_Buf_Write(3) : *byte_sent = %lu\n",
			*byte_sent )) ;
        /*
		USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Buf_Write() End with Fail !!\n"));
		return(D_False);
        */
		USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Buf_Write(1) End.\n"));
		led_usb_flash(0);
		return(D_OK);
       }
   } // for retry
  /*
  #else
  */
    /*
	// ndef NO_PRT_WRITE_BYTE
	USBLPT_DBG_PRINT(iprintf("[USBLPT PRINT] Prt_Buf_Write() : LV_Legth = %lu\n"
                       , LV_Legth));
	for(j=0; j< LV_Legth; j++)
       {
     	USBLPT_DBG_SPOOL(
			if( (j%1000) == 0 )
                      iprintf("[USBLPT SPOOL] Wrote Prt_Buf_Write(): %lu\n", j)
                  ) ;
    	
 		//for (i=0 ; i< 10 ; i++)   
        //   {
      	//	if ( usblp_check_status(usblp, 0)== 0)
        //       {
		//		break ;
	    //       }
       	//   }
    	
		USBLPT_DBG_INIT(iprintf(
	   		"[USBLPT INIT] Prt_Buf_Write() : (byte) LV_Legth = %d , j = %d\n",
			LV_Legth, j
        	));
 
		USBLPT_DBG_FLOW(iprintf(
		"[USBLPT FLOW] Prt_Buf_Write() call Prt_Write_Byte() ..\n"));
        
		if ( Prt_Write_Byte( *(LV_Ptr+j) ) == D_False)
	       {
			// Report back teh number of byte sent to the printer
			*byte_sent=j;
        
			USBLPT_DBG_FLOW(iprintf(
        		"[USBLPT FLOW] Prt_Buf_Write() End with Fail !!\n"
            	));
			return(D_False);
           }
       } // for j
    */   
	// ndef NO_PRT_WRITE_BYTE
 /*
 #endif 
 */
   
 // Calvin
 *byte_sent= (unsigned long) LV_Legth;
 USBLPT_DBG_SPOOL(iprintf(
	"[USBLPT SPOOL] Prt_Buf_Write(4) : *byte_sent = %lu\n",
	*byte_sent )) ;
 
 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Buf_Write(2) End.\n"));
 led_usb_flash(0);
 return(D_OK);
}
/**********************************************************************
Name: void Prt_Close(void)
Purpose: close the printing session
Input:
Output:
Note:
**********************************************************************/

void Prt_Close(void)
{
 //      Prt_Addr_Byte_Write( 0x76, 0);
 /*
 D_ECP_DCR=0x0c; 
 // Out, disbale nACK int., pin: nSLIN=0, INIT=1, Autofd=1, nStrobe=1
 */
}


// #ifdef _NB_SHARE
/**********************************************************************
Name: U8 Prt_Sts_Check_On_Line(void)
Purpose: Check if the printer is in the on-line state
Input: none
Output: A logic 1 means the printer is on line;
        a logic 0 means it is not selected.
Note:
**********************************************************************/

unsigned char Prt_Sts_Check_On_Line(void)
{
 if(( usb_lpt_ok <= 0 ) || ( usblp_plug_out >= 1 ))
   {
    return(D_Fail); 
   }
 if (!global_usblp) {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Sts_Check_On_Line() : Null global_usblp !!\n"));
    return(D_Fail); 
   }
 if(global_usblp->opened != 1) {
    return(D_Fail); 
   }
 /*
  if(    (PARPORT_STATUS_SELECT & Prt_Get_S_EPP_Sts())
      == PARPORT_STATUS_SELECT
    )
 */
 if(( usb_lpt_ok == 1 ) && ( usblp_plug_out == 0 ))
    {
	USBLPT_DBG_INIT(iprintf(
		"[USBLPT INIT] Prt_Sts_Check_On_Line() : Check Status ..\n"
		));
        
	/* the followings are on lp.h
	// 1 : 	LP_POUTPA
	#define LP_POUTPA	0x20  // unchanged input, active high 
						// USB : PAPER_EMPTY 1 : paper empty
                        
	// 2 :  LP_PSELECD
	#define LP_PSELECD	0x10  // unchanged input, active high 
						// USB : SELECT , 1 : selected
                        
	// 3 :  LP_PERRORP
	#define LP_PERRORP	0x08  // unchanged input, active low 
						// USB : NOT_ERROR , 1 : no error
	*/
    
	if ( usblp_check_status(global_usblp, 0) & 2)
       {
	    USBLPT_DBG_INIT(dprintf("[USBLPT INIT] Prt_Sts_Check_On_Line() : Printer Not On-Line.\n") );
	    dprintf("[USBLPT] Printer Not On-Line.\n") ;
    	return( D_False);
       }
      else
       {
	    USBLPT_DBG_PRINT(dprintf("[USBLPT PRINT] Printer On Line.\n") );
	    return( D_True);
       }
    }
  else
    {
     return( D_False);
    }
}
// #endif // _NB_SHARE



// #if (_LPD >= 1) || defined(_NB_SHARE)
//  return : D_False : Bit Value 0 : Error
unsigned char Prt_Sts_Check_Error(void)
{
 	
 if(( usb_lpt_ok <= 0 ) || ( usblp_plug_out >= 1 ))
   {
    return(D_Fail); 
   }
 if (!global_usblp) {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Sts_Check_Error() : Null global_usblp !!\n"));
    return(D_Fail); 
   }
 if(global_usblp->opened != 1) {
    return(D_Fail); 
   }

 // if( (PARPORT_STATUS_ERROR & Prt_Get_S_EPP_Sts()
 //     ) == PARPORT_STATUS_ERROR
 //   )
 
 if(( usb_lpt_ok != 1 ) || ( usblp_plug_out != 0 ))
   {
    /*
	dprintf("[USBLPT] Port not connected !\tCheck the Port Connection, please !\n");
    */
    USBLPT_DBG_INIT(
      iprintf("[USBLPT INIT] Prt_Sts_Check_Error()--1 usb_lpt_ok != 1 or usblp_plug_out != 0 !!\n")
      ) ;
    USBLPT_DBG_INIT(
        iprintf("[USBLPT INIT] Port not connected or no power applied !! \n") ;
        iprintf("[USBLPT INIT] Check the Port Connection , and Power on the Printer, please !\n");
       ) ;
	#ifdef USBLPT_READ_BY_URBTASK
		#ifdef USBLPT_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] wait_ms() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
	#endif // USBLPT_READ_BY_URBTASK
    return(D_Fail); // Error
   }
  else
   {
	/* 
	// 1 : 	LP_POUTPA , paper empty
	// 2 :  LP_PSELECD , not selected
	// 3 :  LP_PERRORP , error 
	*/
    
	if ( usblp_check_status(global_usblp, 0) != 0)
       {
	    USBLPT_DBG_INIT(iprintf(
			"[USBLPT INIT] Prt_Sts_Check_Error() : Printer Error !!\n") );
	    USBLPT_DBG_INFO(iprintf(
			"[USBLPT INFO] Prt_Sts_Check_Error() : Printer Error !!\n") );
    	return( D_False);
       }
      else
       {
	    USBLPT_DBG_PRINT(dprintf("[USBLPT PRINT] Printer Not Error.\n") );
		return( D_True);
       }
   }
   
}


/*
//  return : D_False : Bit Value 0 : Error
U8 Prt_Sts_Check_Busy(void)
{
	
  if( (PARPORT_STATUS_BUSY & Prt_Get_S_EPP_Sts()
      ) == PARPORT_STATUS_BUSY
    )
    {
     LPT_DBG_PRINT(dprintf("[LPT PRINT] Printer Not Busy.\n") );
     return( D_True);
    }
  else
    return( D_False); // Bit Value 0 : Error
}
*/
// #endif // _LPD



// #ifdef _NB_SHARE
/************************************************************************
Name: U8 Prt_Sts_Check_Paper_End(void)
Purpose: Check if the printer has run out of the paper
Input: none
Output: D_False indicates a paper end;
        D_True means the presence of paper.
Note: for SuperI/O, 1 indicates a paper end;
                    0 means the presence of paper.
************************************************************************/
unsigned char Prt_Sts_Check_Paper_End(void)
{
 USBLPT_DBG_FLOW(iprintf(
	"[USBLPT FLOW] Prt_Sts_Check_Paper_End() Begin ..\n"));
 
 if(( usb_lpt_ok <= 0 ) || ( usblp_plug_out >= 1 ))
   {
    // return(D_Fail); 
	return( D_True);
   }
 if (!global_usblp) {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Sts_Check_Paper_End() : Null global_usblp !!\n"));
    return(D_Fail); 
   }
 if(global_usblp->opened != 1) {
    return(D_Fail); 
   }

 /*
 if((PARPORT_STATUS_PAPEREND & Prt_Get_S_EPP_Sts())
    != PARPORT_STATUS_PAPEREND
   )
 */
 if(( usb_lpt_ok == 1 ) && ( usblp_plug_out == 0 ))
   {
	USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Prt_Sts_Check_Paper_End() : Check Status ..\n" )) ;
        
	/* 
	// 1 : 	LP_POUTPA // USB : PAPER_EMPTY 1 : paper empty
	// 2 :  LP_PSELECD	// USB : SELECT , 1 : selected
	// 3 :  LP_PERRORP	// USB : NOT_ERROR , 1 : no error
	*/
    
	if ( usblp_check_status(global_usblp, 0) & 1)
       {
	    USBLPT_DBG_INIT(dprintf("[USBLPT INIT] Printer Paper End.\n") );
    	return( D_False);
       }
      else
       {
      
	    //LPT_DBG_INIT(dprintf("[LPT INIT] Printer Paper Ok"
	    //                     " (Not paper end).\n") );
	    USBLPT_DBG_PRINT(dprintf("[USBLPT PRINT] Printer Paper Ok"
                         " (Not paper end).\n") );
	    return( D_True);
       }
   }    
  else
   {
    USBLPT_DBG_INIT(dprintf("[USBLPT INIT] Printer Paper End (Error).\n") );
    return( D_False);
   }
}

// #endif // _NB_SHARE


/**********************************************************************
Name: U8 Prt_Sts_wait_nBusy(void)
Purpose: wait for nBusy to be released, i.e., going to high
Input: none
Output: D_OK: everything is O.K.; D_False: timeout has been reached.
Note: A logic 0 in this bit, hardware pin, means that the printer is
                busy and cannot accept a new character.
      A logic 1 means that it is ready to accept the next character.
**********************************************************************/

unsigned char Prt_Sts_wait_nBusy(void)
{
 USBLPT_DBG_FLOW(iprintf(
	"[USBLPT FLOW] Prt_Sts_wait_nBusy() Begin ..\n"));
 if(( usb_lpt_ok == 1 ) && ( usblp_plug_out == 0 ))
   {
	USBLPT_DBG_FLOW(iprintf(
		"[USBLPT FLOW] Prt_Sts_wait_nBusy() End.\n"));
    return( D_OK);
   }    
  else
   {
	USBLPT_DBG_FLOW(iprintf(
		"[USBLPT FLOW] Prt_Sts_wait_nBusy() End with Fail !!\n"));
    return( D_False);
   }
   
 /*
 U8 LV_ret;
 U32 LV_Tm_Cnt, timeSt, timeDiff = 0L;

 timeSt = GetTime();
 LV_Tm_Cnt=0L;
 
 while(   (LV_Tm_Cnt < D_Prt_Wait_TimeOut)
       && (timeDiff < D_Prt_wait_TimeOut_ms)
      )
   {
    
	// LV_ret= D_S_EPP_sts_prt;
    
     USBLPT_DBG_PRINT_CHAR(iprintf("[USBLPT PRINT CHAR] Prt_Sts_wait_nBusy() : "
                           "D_S_EPP_sts_prt = %02XH\n", LV_ret));

    LV_Tm_Cnt++;
    
    // if((PARPORT_STATUS_BUSY & LV_ret) == PARPORT_STATUS_BUSY)
    
       {
         USBLPT_DBG_PRINT_CHAR(
                if( LV_Tm_Cnt > 0x1000 )
              {
                  iprintf("[USBLPT PRINT CHAR] Prt_Sts_wait_nBusy() waited "
                          "= %x, %lu\n", LV_Tm_Cnt, LV_Tm_Cnt);
                  iprintf("[USBLPT PRINT] Prt_Sts_wait_nBusy()> Time "
                          "Diff: %lu ms\n", timeDiff);
              }
            ) ;
        return(D_OK);
       }
        timeDiff = GetTime() - timeSt;
   } // while
 
 USBLPT_DBG_PRINT(
   iprintf("[USBLPT PRINT] Prt_Sts_wait_nBusy() wait busy Time Out\n");
   iprintf("[USBLPT PRINT] Prt_Sts_wait_nBusy() Return with D_False ,"
           "LV_Tm_Cnt: %x , timeDiff: %lu\n", LV_Tm_Cnt, timeDiff) );

  return(D_False);
 */
}


#if defined(_NB_SHARE) || (_LPD >=1)

/**********************************************************************
Name: Prt_Reset(void)
Purpose: Send a hardware INIT-low signal to the printer to reset to its
         power-up state
Input: none
Output: none
Note: When INIT becomes LOW, the printer controller is reset to its
        power-up state and the print buffer is cleared.
************************************************************************/
void Prt_Reset(void)
{
 //int i;
 int error;
 /*
 #ifdef _USB_SYSLOG
    char usbsyslog_buf[100];
 #endif // _SYSLOG
 */

 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Reset() Begin ..\n"));

 if(( usb_lpt_ok <= 0 ) || ( usblp_plug_out >= 1 ))
   {
    return ; 
   }
 if (!global_usblp) {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Reset() : Null global_usblp !!\n"));
    return ; 
   }
 if(global_usblp->opened != 1) {
    return ; 
   }

 // dprintf("[USBLPT] Printer Reset ..\n"));

 //D_ECP_DCR=0x00;   // Out, disbale nACK int., pin: nSLIN=1, INIT=0,
 //                  // Autofd=1, nStrobe=1
 
 // for (i=0; i<200; i++);   
 // for (i=0; i<1000; ) i++;   
     //Around 20 uS, this pulse width must be more than 50 microseconds for meeting the init. spec.
      //When INIT becomes LOW, the printer controller is reset to its power-up state and the print buffer is cleared.
 
 //  D_ECP_DCR=0x0c;    // Out, disbale nACK int., pin: nSLIN=0, INIT=1, 
                      //  Autofd=1, nStrobe=1 
 
 if((usb_lpt_ok == 1) && (usblp_plug_out == 0))
   {
	#ifdef USBLPT_RESET_CMD // 2003/10/14 B. S. Liaw

	// dprintf("[USBLPT] Printer Reset ..\n");
    
	error = usblp_reset(global_usblp);
 
	if (error < 0)
	   {
		USBLPT_DBG_INFO(iprintf(
			"[USBLPT INFO] Error : Cannot Reset Printer : %d !! \n",
			error));
		iprintf("[USBLPT] Error : Cannot Reset Printer : %d !! \n",
			error);

		#ifdef _USB_SYSLOG
		    //sprintf(usbsyslog_buf,"[USBLPT] Error : Cannot Reset Printer : %d !! \n",
			//	error);
						#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{

#ifdef USB_SYS_LOG_PR_RESET_FAIL
				//sprintf_SetSystemLog("%s : %d !! \n", USB_SYS_LOG_PR_RESET_FAIL, error);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "%s : %d !! \n", USB_SYS_LOG_PR_RESET_FAIL, error);
#else
				//sprintf_SetSystemLog("[USBLPT] Error : Cannot Reset Printer : %d !! \n",error);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Error : Cannot Reset Printer : %d !! \n",error);
#endif // USB_SYS_LOG_PR_RESET_FAIL
			}
		#endif // _SYSLOG
    	return ;
	   }
      else
       {
		wait_ms(USBLPT_INIT_WAIT_MS);
		iprintf("[USBLPT] Printer Reset OK.\n");
		#ifdef _USB_SYSLOG
		    //sprintf(usbsyslog_buf,"[USBLPT] Printer Reset OK. ");
			#if (_SYSLOG==2)
			if (gSetting.syslog_ds.syslog_enable==1) 
			#endif // _SYSLOG==2
				{
#ifdef USB_SYS_LOG_PR_RESET_OK
				//sprintf_SetSystemLog(USB_SYS_LOG_PR_RESET_OK);
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, USB_SYS_LOG_PR_RESET_OK);
#else
				//sprintf_SetSystemLog("[USBLPT] Printer Reset OK. ");
				sprintf_SetSystemLog(LOG_TYPE_HW_USB_DEVICE, "[USBLPT] Printer Reset OK. ");
#endif // USB_SYS_LOG_PR_RESET_OK
			}
		#endif // _SYSLOG
	    return ;
       }
	 #else
		USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Reset() : Not Do Anything !! \n"));
	    return ;
	#endif
    }
  else
   {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] No Printer to Reset !! \n"));
	return ;
   }
}
#endif // defined(_NB_SHARE) || (_LPD >=1)


#ifdef _NB_SHARE
/*************************************************/
void Prt_nStrobe(void)  //Cause nStrobe a low pulse, 19092001
{
 /*
    D_S_EPP_Ctl_prt = 0x08+0x04; // Out, disbale nACK int., pin: nSLIN=0, INIT=1, Autofd=1, nStrobe=1 
    D_S_EPP_Ctl_prt = 0x08+0x05; //  Out, disbale nACK int., pin: nSLIN=0, INIT=1, Autofd=1, nStrobe=0 
    D_S_EPP_Ctl_prt = 0x08+0x04; // Out, disbale nACK int., pin: nSLIN=0, INIT=1, Autofd=1, nStrobe=1 
 */
}


/*************************************************/
void Prt_Config_83877_PPM(unsigned char LV_ppm)    /* Set Parallel Port Mode */
{
 /*
 LPT_DBG_INIT(dprintf("[LPT INIT] Prt_Config_83877_PPM() :"
      " W83877tf_EFER_chip = %04XH\n", W83877tf_EFER_chip )) ;
 LPT_DBG_INIT(dprintf("[LPT INIT] Prt_Config_83877_PPM() :"
      " W83877tf_EFER_CPU = %08XH\n", W83877tf_EFER_CPU )) ;

 LPT_DBG_INIT(dprintf("[LPT INIT] Prt_Config_83877_PPM() :"
      " W83877tf_EFIR_chip = %04XH\n", W83877tf_EFIR_chip )) ;
 LPT_DBG_INIT(dprintf("[LPT INIT] Prt_Config_83877_PPM() :"
      " W83877tf_EFIR_CPU = %08XH\n", W83877tf_EFIR_CPU )) ;

 LPT_DBG_INIT(dprintf("[LPT INIT] Prt_Config_83877_PPM() :"
      " W83877tf_EFDR_chip = %04XH\n", W83877tf_EFDR_chip )) ;
 LPT_DBG_INIT(dprintf("[LPT INIT] Prt_Config_83877_PPM() :"
      " W83877tf_EFDR_CPU = %08XH\n", W83877tf_EFDR_CPU )) ;

 LPT_DBG_INIT(
      dprintf("[LPT INIT] Prt_Config_83877_PPM() : D_EPP_ECP_Base_Addr ="
      " %04XH\n", D_EPP_ECP_Base_Addr)) ;

 LPT_DBG_INIT(
      dprintf("[LPT INIT] Prt_Config_83877_PPM() : D_S_EPP_dat_chip ="
      " %08XH\n", D_S_EPP_dat_chip)) ;
 LPT_DBG_INIT(
      dprintf("[LPT INIT] Prt_Config_83877_PPM() : D_S_EPP_dat ="
      " %08XH\n", D_S_EPP_dat)) ;
 LPT_DBG_INIT(
      dprintf("[LPT INIT] Prt_Config_83877_PPM() : D_S_EPP_sts_chip ="
      " %08XH\n", D_S_EPP_sts_chip)) ;
 LPT_DBG_INIT(
      dprintf("[LPT INIT] Prt_Config_83877_PPM() : D_S_EPP_sts ="
      " %08XH\n", D_S_EPP_sts)) ;
 LPT_DBG_INIT(
      dprintf("[LPT INIT] Prt_Config_83877_PPM() : D_S_EPP_Ctl_chip ="
      " %08XH\n", D_S_EPP_Ctl_chip)) ;
 LPT_DBG_INIT(
      dprintf("[LPT INIT] Prt_Config_83877_PPM() : D_S_EPP_Ctl ="
      " %08XH\n", D_S_EPP_Ctl)) ;
 #if (SUPER_IO==0) // WinBond
        //Winbond SuperI/O: W83877tf is used
  W83877tf_EFER=D_Key_W83877tf_EFER;    // Enable EFER with specific keyword. 
  W83877tf_EFIR=D_Key_W83877tf_EFER;    // Extended Function Index Reg. 
  
 #if (SUPERIO_CLK_48MHZ==1)
  W83877tf_EFIR =0x2c;
  W83877tf_EFDR =0x04;   
 #endif // SUPERIO_CLK_48MHZ

  W83877tf_EFIR =0x09;  // Point to CR9 
  W83877tf_EFDR =0x80;  // PRT_MODE2=1, Enable R/W CR0-45 

  W83877tf_EFIR =0x00;  // Point to CR0 
  W83877tf_EFDR =LV_ppm;  // Select mode:SPP/EPP/ECP 

  W83877tf_EFIR =0x03;  // Point to CR3 
  W83877tf_EFDR =0x00;  // Select EPP V1.9. 

  W83877tf_EFIR =0x05;  // Point to CR5 
  W83877tf_EFDR = 0x08; // Set FIFO threshold for ECP mode. 
   // 2002/06/26 B. S. Liaw
   LPT_DBG_INIT(dprintf("[LPT INIT] Prt_Config_83877_PPM() : Set CR23"
      " to %02XH\n", D_default_EF_CR23)) ;
   
   //LPT_DBG_INIT(
   //   dprintf("[LPT INIT] Prt_Config_83877_PPM() : So , the address ="
   //   " %04XH\n", D_EPP_ECP_Base_Addr)) ;
   //LPT_DBG_INIT(
   //   dprintf("[LPT INIT] Prt_Config_83877_PPM() : Par_PORT_ADDR(0) ="
   //   " %04XH\n", Par_PORT_ADDR(0))) ;
   //LPT_DBG_INIT(
   //   dprintf("[LPT INIT] Prt_Config_83877_PPM() : Par_PORT_ADDR(1) ="
   //   " %04XH\n", Par_PORT_ADDR(1))) ;
   

   // ------------------------------------------------------------------


  W83877tf_EFER = 0xaa;   // Exit Configuration Mode. 
 #elif (SUPER_IO==1) // SMSC
        //SMSC SuperI/O: SP37E760 is uesd
  W83877tf_EFER=D_Key_SP37E760_EFER;     // Enable EFER with specific keyword. 

  W83877tf_EFIR =0x01;
  W83877tf_EFDR =LV_ppm;

  W83877tf_EFIR =0x04;
  W83877tf_EFDR =0x03;  // Select mode:EPP/ECP, EPP V1.9. 

  W83877tf_EFIR =0x0a;  // Point to CRA 
  W83877tf_EFDR = 0x08; // Set FIFO threshold for ECP mode. 

  W83877tf_EFIR =0x23;  // point to CR23 (Parallel port Base Address)
  W83877tf_EFDR =D_default_EF_CR23;     // Set Base Addr. of Parallel port for SP37E760 as 0x378 

  W83877tf_EFER =0xaa;     // Exit Configuration Mode.
  
 #endif  //_HNP
 */
}


#endif // _NB_SHARE

// ====================================================================

#ifndef NO_PRT_WRITE_BYTE

   /*******************************************************************
   Name: U8 Prt_Write_Byte(U8 LV_BYTE)
   Purpose: write data to printer port
   Input: Data Byte
   Output: D_OK: everything is O.K.; D_False: timeout has been reached.
   Note:
   *******************************************************************/


   unsigned char Prt_Write_Byte(unsigned char LV_BYTE) {
	unsigned char buffer[5] ;
	/*
	struct usblp *usblp;
	*/
	USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Write_Byte() Begin ..\n"));

 if(( usb_lpt_ok <= 0 ) || ( usblp_plug_out >= 1 ))
   {
    return(D_Fail); 
   }
 if (!global_usblp) {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_Write_Byte() : Null global_usblp !!\n"));
    return(D_Fail); 
   }
 if(global_usblp->opened != 1) {
    return(D_Fail); 
   }

	/*
	D_S_EPP_Ctl_prt=0x08+0x04; // Out, disbale nACK int., pin: nSLIN=0,
                // INIT=1, Autofd=1, nStrobe=1 
	D_S_EPP_dat_prt=LV_BYTE;
	*/
	/* 
	usblp  = usblp_table[USBLP_MINOR_BASE];
	*/
	buffer[0] = LV_BYTE ;
 
	/*
	if ((retval = usblp_check_status(usblp, 0)))
       {
		retval = retval > 1 ? -EIO : -ENOSPC;
		goto out;
	   }
	*/
  
 USBLPT_DBG_PRINT_CHAR( iprintf("[USBLPT PRINT CHAR] write data byte %02XH , [%c]\n",LV_BYTE,LV_BYTE));
 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Prt_Write_Byte() : Write ..\n")) ;
 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_Write_Byte() call usblp_write() ..\n"));
	usblp_write( global_usblp , &buffer[0] , 1 ) ;

 // dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;

 /*                               
 if( D_S_EPP_dat_prt != LV_BYTE )
   {
    iprintf("[USBLPT] Prt_Write_Byte() "
         "verify data Error !!\n") ;
    LPT_DBG_PRINT_CHAR( iprintf("[LPT PRINT CHAR] Prt_Write_Byte() "
         "verify data Error !!\n") );
    return(D_False);
   }
  */
	if(Prt_Sts_wait_nBusy()==D_False)	// Check nBUSY 
      {
		USBLPT_DBG_FLOW(iprintf(
			"[USBLPT FLOW] Prt_Write_Byte() End with Fail !!\n"));
		return(D_False);
      }
 USBLPT_DBG_PRINT_CHAR( iprintf("[USBLPT PRINT CHAR] Strobe Low\n"));
	/*
	D_S_EPP_Ctl_prt=0x08+0x05; // Out, disbale nACK int., pin: nSLIN=0,
	// INIT=1, Autofd=1, nStrobe=0
 // Out, disbale nACK int., pin: nSLIN=0, INIT=1, Autofd=1, nStrobe=0 
 #if defined(ADMtek)
   // for (x=0;x<100;x++) // B. S. Liaw : for some printer
   for (x=0;x<10;x++)
       {
         // Empty Descript
        //if(Prt_Sts_wait_nBusy()==D_False)   // Check nBUSY 
        //   return(D_False);
       } 
 #endif

 LPT_DBG_PRINT_CHAR( iprintf("[LPT PRINT CHAR] Strobe High\n"));
	D_S_EPP_Ctl_prt=0x08+0x04; // Out, disbale nACK int., pin: nSLIN=0,
	// INIT=1, Autofd=1, nStrobe=1
 // Out, disbale nACK int., pin: nSLIN=0, INIT=1, Autofd=1, nStrobe=1

	*/
 /* 2002/07/20 Pert , Donot need to delay for upper layer do many thing
 #if defined(ADMtek)
   for (x=0;x<100;x++)
       {
        // Empty Descript
        //if(Prt_Sts_wait_nBusy()==D_False)   // Check nBUSY
        //   return(D_False);
       } 
 #endif
 */
	USBLPT_DBG_FLOW(iprintf(
		"[USBLPT FLOW] Prt_Write_Byte() End.\n"));
	return(D_OK);
   }// function Prt_Write_Byte

#endif // ndef NO_PRT_WRITE_BYTE

// ====================================================================

#ifdef _NB_SHARE
// **********************************************************************
// FUNCTION:  printerConnect
// PURPOSE:  get printer state (can be called from outside this module)
// RETURN:
//			0: off,  1: on,  2: printing
// **********************************************************************

unsigned char printerConnect(void)
{
 unsigned char retVar, mix;
	
    
 USBLPT_DBG_FLOW(iprintf(
	"[USBLPT FLOW] printerConnect() Begin ..\n"));
        
	retVar = Prt_Get_S_EPP_Sts();
	mix = retVar & PARPORT_STATUS_SELECT;
	dprintf("printerConnect()> pr on-line:%02XH & %02XH = %02XH (%s)\n", retVar, PARPORT_STATUS_SELECT, 
		mix, ((mix == PARPORT_STATUS_SELECT) ? "On" : "Off"));
	mix = retVar & PARPORT_STATUS_ERROR;
	dprintf("printerConnect()> pr error  :%02XH & %02XH = %02XH (%s)\n", retVar, PARPORT_STATUS_ERROR, 
		mix, ((mix == PARPORT_STATUS_ERROR) ? "Normal" : "Error"));
	mix = retVar & PARPORT_STATUS_BUSY;
	dprintf("printerConnect()> pr busy   :%02XH & %02XH = %02XH (%s)\n", retVar, PARPORT_STATUS_BUSY, 
		mix, ((mix == PARPORT_STATUS_BUSY) ? "Normal" : "Busy"));
	mix = retVar & PARPORT_STATUS_PAPEREND;
	dprintf("printerConnect()> pr on-line:%02XH & %02XH = %02XH (%s)\n", retVar, PARPORT_STATUS_PAPEREND, 
		mix, ((mix == PARPORT_STATUS_PAPEREND) ? "End" : "Normal"));
	
	retVar = Prt_Check_Port_Pwr_Connect();
	dprintf("printerConnect()> Summary: %u\n", retVar);
	
 USBLPT_DBG_FLOW(iprintf(
	"[USBLPT FLOW] printerConnect() End.\n"));
 return retVar;
}

#endif // _NB_SHARE


/**********************************************************************
Name: Prt_GetId(void)
Purpose: Get the IEEE 1284 Device ID String of USB Line Printer
Input: none
Output: Pointer to the string , unsigned char *
Note: Before Printer Init Ok , return null pointer as invalid value
************************************************************************/
#ifdef _NB_SHARE

// 2003/11/12 B. S. Liaw add this function

unsigned char * Prt_GetId(void) {
 unsigned char *ret_ptr ;


 USBLPT_DBG_FLOW(iprintf("[USBLPT FLOW] Prt_GetId() Begin ..\n"));

 if(( usb_lpt_ok <= 0 ) || ( usblp_plug_out >= 1 ))
   {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_GetId() : No Printer ID Now !! \n"));
    return (unsigned char *)(0x00); 
   }
 if (!global_usblp) {
	USBLPT_DBG_INFO(iprintf(
		"[USBLPT INFO] Prt_GetId() : Null global_usblp !!\n"));
    return (unsigned char *)(0x00); 
   }
 if(global_usblp->opened != 1) {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_GetId() : No Printer ID Now !! \n"));
    return (unsigned char *)(0x00); 
   }
 
 if((usb_lpt_ok == 1) && (usblp_plug_out == 0))
   {
	ret_ptr = (unsigned char *)(&global_usblp->device_id_string[2]);

	USBLPT_DBG_INFO(iprintf(
		"[USBLPT INFO] Prt_GetId() : Device ID =[%s] \n",
		ret_ptr
	    ));
    return ret_ptr ; 
    }
  else
   {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] Prt_GetId() : No Printer ID Now !! \n"));
    return (unsigned char *)(0x00); 
   }
} // function Prt_GetId()

#endif // _NB_SHARE


// ====================================================================

// ====================================================================
/*
// 2003/4/14 B. S. Liaw , This is important !!
// vendor USB init string 
// at least EPSON STYLUS C41UX Need this for exit packet mode

static usb_init_string(struct usblp *usblp)
// exit_packet_mode(void)
{
 // 00 00 00 1b 01 40 45 4a 4c 20 31 32 38 34 2e 34 ..
 static char hdr[] 
	= "\000\000\000\033\001@EJL 1284.4\n@EJL     \n\033@";
 unsigned long i;
 
    
 USBLPT_DBG_FLOW(iprintf(
 	"[USBLPT FLOW] usb_init_string() Begin ..\n"
 	));
 
// memcpy(printer_cmd + bufpos, hdr, sizeof(hdr) - 1); 
 // DON'T include null! 
 
// bufpos += sizeof(hdr) - 1;
 
 USBLPT_DBG_FLOW(iprintf(
	"[USBLPT FLOW] usb_init_string() call usblp_write() ..\n"));
    
 i = usblp_write( usblp , hdr , sizeof(hdr)-1 ) ;
    
 USBLPT_DBG_INIT(iprintf(
	"[USBLPT INIT] usb_init_string() : sizeof(hdr)-1 = %d , i = %d\n",
	sizeof(hdr)-1, i
    ));
        
 if (i < sizeof(hdr)-1)
   { 
	USBLPT_DBG_FLOW(iprintf(
    	"[USBLPT FLOW] usb_init_string() End with Fail !!\n"
    	));
	return(D_False);
   }
       
 USBLPT_DBG_FLOW(iprintf(
 	"[USBLPT FLOW] usb_init_string() End.\n"
 	));
 return(D_OK);
}
*/

// ====================================================================


// #ifdef _NB_SHARE
// #ifdef _ETCPIP

// #endif // _ETCPIP


// ====================================================================

// ====================================================================
// ====================================================================

#ifdef EPSON_LPT_DIAG   // 2003/4/14 B. S. Liaw

char *printer = NULL;
char *raw_device = NULL;
char *printer_model = NULL;
char printer_cmd[1025];
int bufpos = 0;

/*
#ifdef NEW_EPSON_LPT
	int isnew = 1;
 #else
	#error : Please Make Sure : Not NEW_EPSON_LPT !!
	int isnew = 0;
#endif
*/

/*
static void
print_models(void)
{
  stp_printer_t *printer = &printer_list[0];
  while (printer->short_name)
    {
      printf("%10s      %s\n", printer->short_name, _(printer->long_name));
      printer++;
    }
}

void
do_help(int code)
{
  printf("%s", _(help_msg));
  exit(code);
}
*/

#ifdef NEW_EPSON_LPT
   #error : Please Make Sure : NEW_EPSON_LPT !!
   static void
   exit_packet_mode(void)
   {
    // 00 00 00 1b 01 40 45 4a 4c 20 31 32 38 34 2e 34 ..
	static char hdr[] = "\000\000\000\033\001@EJL 1284.4\n@EJL     \n\033@";
    
	memcpy(printer_cmd + bufpos, hdr, sizeof(hdr) - 1); /* DON'T include null! */
	bufpos += sizeof(hdr) - 1;
   }
 #else
	// #error : Please Make Sure : Not NEW_EPSON_LPT !!
#endif

static void initialize_print_cmd(void)
{
 bufpos = 0;
 #ifdef NEW_EPSON_LPT
	/*
    // if (isnew)
    */ {
		// 2003/7/16 B. S. Liaw : This need done by driver ..
        /*
		exit_packet_mode();
        */
		USBLPT_DBG_INIT(iprintf("[USBLPT INIT] initialize_print_cmd() : Not Call exit_packet_mode !! \n")) ;
       }
 #endif // NEW_EPSON_LPT
}

int do_print_cmd(void)
{
  /*
  FILE *pfile;
  */
  int bytes = 0;
  int retries = 0;
  char command[1024];
  memcpy(printer_cmd + bufpos, "\f\033\000\033\000", 5);
  bufpos += 5;
  /*
  if (raw_device)
    {
      pfile = fopen(raw_device, "wb");
      if (!pfile)
        {
          fprintf(stderr, _("Cannot open device %s: %s\n"), raw_device,
                  strerror(errno));
          return 1;
        }
    }
  else
    {
      if (!access("/bin/lpr", X_OK) ||
          !access("/usr/bin/lpr", X_OK) ||
          !access("/usr/bsd/lpr", X_OK))
        {
        if (printer == NULL)
          strcpy(command, "lpr -l");
        else
          snprintf(command, 1023, "lpr -P%s -l", printer);
        }
      else if (printer == NULL)
        strcpy(command, "lp -s -oraw");
      else
        snprintf(command, 1023, "lp -s -oraw -d%s", printer);

      if ((pfile = popen(command, "w")) == NULL)
        {
          fprintf(stderr, _("Cannot print to printer %s with %s\n"), printer,
                  command);
          return 1;
        }
    }
  */
  while (bytes < bufpos)
   {
    /*
    int status = fwrite(printer_cmd + bytes, 1, bufpos - bytes, pfile);
    */
    int status ;
    
      
 if(( usb_lpt_ok <= 0 ) || ( usblp_plug_out >= 1 ))
   {
    return(D_Fail); 
   }
 if (!global_usblp) {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] do_print_cmd() : Null global_usblp !!\n"));
    return(D_Fail); 
   }
 if(global_usblp->opened != 1) {
    return(D_Fail); 
   }

	USBLPT_DBG_INIT(iprintf("[USBLPT INIT] do_print_cmd() : Write ..\n")) ;
    
    status = usblp_write( global_usblp , printer_cmd + bytes ,
                                        bufpos - bytes ) ;
	// dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
    if (status == 0)
        {
          retries++;
          if (retries > 2)
           {
            /*
            fprintf(stderr, _("Unable to send command to printer\n"));
            */
			dprintf("[USBLPT] Error : Unable to send command to printer\n") ;
            /*
              if (raw_device)
                fclose(pfile);
              else
                pclose(pfile);
              */
            return 1;
           }
        }
      else if (status == -1)
        {
          /*
          fprintf(stderr, _("Unable to send command to printer\n"));
          */
      	dprintf("[USBLPT] Error : Unable to send command to printer\n") ;
          /*
          if (raw_device)
            fclose(pfile);
          else
            pclose(pfile);
          */
          return 1;
        }
      else
        {
          bytes += status;
          retries = 0;
        }
    }
  /*
  if (raw_device)
    fclose(pfile);
  else
    pclose(pfile);
  */
  return 0;
}

// ====================================================================

static void
do_remote_cmd(const char *cmd, int nargs, ...)
{
  static char remote_hdr[] = "\033@\033(R\010\000\000REMOTE1";
  static char remote_trailer[] = "\033\000\000\000\033\000";
  int i;
  va_list args;
  va_start(args, nargs);

  memcpy(printer_cmd + bufpos, remote_hdr, sizeof(remote_hdr) - 1);
  bufpos += sizeof(remote_hdr) - 1;
  memcpy(printer_cmd + bufpos, cmd, 2);
  bufpos += 2;
  printer_cmd[bufpos] = nargs % 256;
  printer_cmd[bufpos + 1] = (nargs >> 8) % 256;
  if (nargs > 0)
    for (i = 0; i < nargs; i++)
      printer_cmd[bufpos + 2 + i] = va_arg(args, int);
  bufpos += 2 + nargs;
  memcpy(printer_cmd + bufpos, remote_trailer, sizeof(remote_trailer) - 1);
  bufpos += sizeof(remote_trailer) - 1;
}

static void
add_newlines(int count)
{
  int i;
  for (i = 0; i < count; i++)
    {
      printer_cmd[bufpos++] = '\r';
      printer_cmd[bufpos++] = '\n';
    }
}

static void
add_resets(int count)
{
  int i;
  for (i = 0; i < count; i++)
    {
      printer_cmd[bufpos++] = '\033';
      printer_cmd[bufpos++] = '\000';
    }
}
/*
const char *colors[] =
{
  N_("Black"),
  N_("Cyan"),
  N_("Magenta"),
  N_("Yellow"),
  N_("Light Cyan"),
  N_("Light Magenta"),
  N_("Black/Dark Yellow"),
  0
};
*/

const char *colors[8] =
{
  "Black",
  "Cyan",
  "Magenta",
  "Yellow",
  "Light Cyan",
  "Light Magenta",
  "Black/Dark Yellow",
  0
};


// return : 1 : ok
int do_ink_level(void)
{
  /*
  int fd;
  */
  int status;
  char buf[1024];
  char *ind;
  int i;
  /*
  if (!raw_device)
    {
      fprintf(stderr,_("Obtaining ink levels requires using a raw device.\n"));
      exit(1);
    }
  fd = open(raw_device, O_RDWR, 0666);
  if (fd == -1)
    {
      fprintf(stderr, _("Cannot open %s read/write: %s\n"), raw_device,
              strerror(errno));
      exit(1);
    }
  */
 if(( usb_lpt_ok <= 0 ) || ( usblp_plug_out >= 1 ))
   {
    return(D_Fail); 
   }
 if (!global_usblp) {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] do_ink_level() : Null global_usblp !!\n"));
    return(D_Fail); 
   }
 if(global_usblp->opened != 1) {
    return(D_Fail); 
   }

  initialize_print_cmd();
  do_remote_cmd("ST", 2, 0, 1);
  add_resets(2);

 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] do_ink_level(1) : Write ..\n")) ;
 // USBLPT_DBG_INFO(iprintf("[USBLPT INIT] do_ink_level(1) : Write ..\n")) ;

  /*
  if (write(fd, printer_cmd, bufpos) < bufpos)
  */
  if (usblp_write( global_usblp , printer_cmd , bufpos)  < bufpos)
    {
      /*
      fprintf(stderr, _("Cannot write to %s: %s\n"), raw_device,
              strerror(errno));
      */
		#ifdef USBLPT_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] wait_ms() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
	  dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
      dprintf("[USBLPT] Error : Cannot write to printer \n") ;
      /*
      exit(1);
      */
      return 0 ;
    }
 
 // dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
    
 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw

	USBLPT_DBG_INIT(iprintf("[USBLPT INIT] do_ink_level(1) : read ..\n")) ;
	#ifndef USBLPT_READ_BY_CHECK
		/*
		status = read_from_printer(fd, buf, 1024);
		*/
		status = read_from_printer(buf, 1024);
	#else
		status = usblp_read(global_usblp , buf, 1024-1);
	#endif

	#ifndef USBLPT_READ_NO_DELAY
		#ifdef USBLPT_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] wait_ms() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
	#endif // ndef USBLPT_READ_NO_DELAY

	if (status < 0) {
		/*
	      fprintf(stderr, _("Cannot read from %s: %s\n"),
              raw_device,strerror(errno));
    	*/
		dprintf("[USBLPT] Error : Cannot read from printer \n") ;
		/*
	      exit(1);
    	*/
		return 0 ;
	   }
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw


 ind = buf;
  
 do{
	ind = strchr(ind, 'I');
   }
	while (    ind && ind[1] != 'Q' 
            && (ind[1] != '\0' && ind[2] != ':')
          );
  
  if (!ind || ind[1] != 'Q' || ind[2] != ':')
    { 
     /*
      fprintf(stderr, _("Cannot parse output from printer\n"));
     */
     USBLPT_DBG_INIT(iprintf("[USBLPT INIT] do_ink_level(1) : Cannot parse output from printer\n")) ;
     /*
      exit(1);
     */
     return 0 ;
    }
  ind += 3;

 #ifndef LPT_DIAG_NOT_SHOW
  /*
  printf("%20s    %s\n", _("Ink color"), _("Percent remaining"));
  */
  dprintf("%20s    %s\n", "Ink color", "Percent remaining");
  for (i = 0; i < 7; i++)
    {
      int val, j;
      if (!ind[0] || ind[0] == ';')
        {
         /*
         exit(0);
         */
         return 1 ;
        }
      for (j = 0; j < 2; j++)
        {
          if (ind[j] >= '0' && ind[j] <= '9')
            ind[j] -= '0';
          else if (ind[j] >= 'A' && ind[j] <= 'F')
            ind[j] = ind[j] - 'A' + 10;
          else if (ind[j] >= 'a' && ind[j] <= 'f')
            ind[j] = ind[j] - 'a' + 10;
          else
           {
            /*
            exit(1);
            */
            return 1 ;
           }
        }
      val = (ind[0] << 4) + ind[1];
      /*
      printf("%20s    %3d\n", _(colors[i]), val);
      */
      dprintf("%20s    %3d\n", colors[i], val);
      ind += 2;
    }
 #endif // ndef LPT_DIAG_NOT_SHOW

  initialize_print_cmd();
  do_remote_cmd("ST", 2, 0, 0);
  add_resets(2);

 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] do_ink_level(2) : Write ..\n")) ;
 // USBLPT_DBG_INFO(iprintf("[USBLPT INFO] do_ink_level(2) : Write ..\n")) ;

 /*
 (void) write(fd, printer_cmd, bufpos);
 */
  usblp_write( global_usblp , printer_cmd , bufpos)  ;

 // dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;


 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	USBLPT_DBG_INIT(iprintf("[USBLPT INIT] do_ink_level(2) : read ..\n")) ;
	 // USBLPT_DBG_INFO(iprintf("[USBLPT INFO] do_ink_level(2) : read ..\n")) ;

	/*
	(void) read(fd, buf, 1024);
	*/
	status = usblp_read(global_usblp , buf, 1024);

	#ifndef USBLPT_READ_NO_DELAY
		#ifdef USBLPT_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] wait_ms() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
	#endif // ndef USBLPT_READ_NO_DELAY

	if (status == 0)
	   {
		USBLPT_DBG_INIT(iprintf("[USBLPT INIT] do_ink_level(1) : Error : Zero Read from printer \n" )) ;
	   }
   
	if (status < 0)
	   {
		USBLPT_DBG_INFO(iprintf("[USBLPT INFO] do_ink_level(2) : Error : Cannot read from printer \n" )) ;
	   }
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw

 /*
 (void) close(fd);
 */
 /*
 exit(0);
 */
 return 1 ;
}

// return : 1 : ok
int do_status(void)
{
 /*
 int fd;
 */
 int status;
 char buf[1024];
 char *where;
 memset(buf, 0, 1024);
 
 /*
 if (!raw_device)
   {
     fprintf(stderr, _("Printer status requires using a raw device.\n"));
     exit(1);
   }
 fd = open(raw_device, O_RDWR, 0666);
 if (fd == -1)
   {
     fprintf(stderr, _("Cannot open %s read/write: %s\n"), raw_device,
             strerror(errno));
     exit(1);
   }
 */
 if(( usb_lpt_ok <= 0 ) || ( usblp_plug_out >= 1 ))
   {
    return(D_Fail); 
   }
 if (!global_usblp) {
	USBLPT_DBG_INFO(iprintf("[USBLPT INFO] do_status() : Null global_usblp !!\n"));
    return(D_Fail); 
   }
 if(global_usblp->opened != 1) {
    return(D_Fail); 
   }

 bufpos = 0;
 initialize_print_cmd();
 do_remote_cmd("ST", 2, 0, 1);
 
 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] do_status(1) : Write ..\n")) ;
 // USBLPT_DBG_INFO(iprintf("[USBLPT INFO] do_status(1) : Write ..\n")) ;
 /*
 if (write(fd, printer_cmd, bufpos) < bufpos)
 */
  if (usblp_write( global_usblp , printer_cmd , bufpos) < bufpos)
    { 
      /*
      fprintf(stderr, _("Cannot write to %s: %s\n"),
              raw_device, strerror(errno));
      */
		#ifdef USBLPT_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] wait_ms() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
	  dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;
      dprintf("[USBLPT] Error : Cannot write to printer \n") ;
      /*
      exit(1);
      */
      return 0 ;
    }

 // dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
    
 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	USBLPT_DBG_INIT(iprintf("[USBLPT INIT] do_status(1) : read ..\n")) ;
 
	#ifndef USBLPT_READ_BY_CHECK
		/*
		status = read_from_printer(fd, buf, 1024);
		*/
		status = read_from_printer(buf, 1024);
	 #else
		status = usblp_read(global_usblp , buf, 1024-1);
	#endif

	#ifndef USBLPT_READ_NO_DELAY
		#ifdef USBLPT_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] wait_ms() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
	#endif // ndef USBLPT_READ_NO_DELAY

	if (status < 0) {
      /*
      fprintf(stderr, _("Cannot read from %s: %s\n"),
              raw_device, strerror(errno));
      */
      dprintf("[USBLPT] Error : Cannot read from printer \n") ;
      /*
      exit(1);
      */
		return 0 ;
	   }
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw

 initialize_print_cmd();
 do_remote_cmd("ST", 2, 0, 0);
 add_resets(2);

 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] do_status(2) : Write ..\n")) ;
 // USBLPT_DBG_INFO(iprintf("[USBLPT INFO] do_status(2) : Write ..\n")) ;

 /*
 (void) write(fd, printer_cmd, bufpos);
 */
  usblp_write( global_usblp , printer_cmd , bufpos)  ;
 // Prt_Buf_Write() ;
 
 // dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;

  /*
  (void) read(fd, buf, 1024);
  */

 #ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
	USBLPT_DBG_INIT(iprintf("[USBLPT INIT] do_status(2) : read ..\n")) ;

	status =  usblp_read(global_usblp , buf, 1024);

	#ifndef USBLPT_READ_NO_DELAY
		#ifdef USBLPT_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] wait_ms() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
		dlytsk(cur_task, DLY_TICKS, USBLPT_DLY_TSK_TICK) ;
	#endif // ndef USBLPT_READ_NO_DELAY

	if (status <= 0) {
		USBLPT_DBG_INIT(iprintf("[USBLPT INIT] do_status(2) : Error : Cannot read from printer \n" )) ;
	   }
 #endif // USBLPT_BIDIR // 2003/10/13 B. S. Liaw
  


 USBLPT_DBG_INIT(iprintf("[USBLPT INIT] do_status(2) : Check Status .. \n" )) ;
 while ((where = strchr(buf, ';')) != NULL)
    {
	 *where = '\n';
    }
    
 #ifndef LPT_DIAG_NOT_SHOW
	dprintf("[USBLPT] Status :  %s\n", buf);
 #endif // ndef LPT_DIAG_NOT_SHOW
  
  /*
  (void) close(fd);
  */
 USBLPT_DBG_FLOW(iprintf("[USBLPT INIT] do_status() End.\n")) ;
  /*
  exit(0);
  */
  return 1 ;
}

void
do_nozzle_check(void)
{
  do_remote_cmd("VI", 2, 0, 0);
  do_remote_cmd("NC", 2, 0, 0);
  /*
  printf(_("Running nozzle check, please ensure paper is in the printer.\n"));
  */
  dprintf("[USBLPT] Running nozzle check, please ensure paper is in the"
  		" printer.\n");
  /*
  exit(do_print_cmd());
  */
  do_print_cmd();
}

#endif // EPSON_LPT_DIAG


// ====================================================================

unsigned char lpt_test_str[] = "This is USB LPT port test...\n";

void ParallelTest(void)
{
 unsigned short  ch;
 unsigned char   buf[26];
 int i;
 unsigned long   byte_sent;
 extern void Delay(int s);
 char a = 0x0a;
 #ifdef LPT_DIAG_SPEED   // 2003/10/20 B. S. Liaw
	unsigned long start_printing_time = 0L ;
	unsigned long test_times = 0L ;
 #endif // LPT_DIAG_SPEED

 dprintf("[USBLPT] ==== Parallel Test ====\n") ;
 
 /*
	// wait 5 seconds
	dprintf("[USBLPT] wait 5 seconds ..\n") ;
	dlytsk(cur_task, DLY_SECS, 5);
 */
 
 // ----------------------
 re_test :
 // ----------------------
 
		#ifdef USBLPT_BS_SYSCAL_ASSERT
			if (!(  mt_busy == 0
		         )
			   )
			   {
				iprintf("[ASSERT][USB] wait_ms() : " ) ;
    		    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
			    iprintf("[ASSERT][USB] Continue ..\n") ;
    		   }
		#endif // USB_BS_ASSERT
 dlytsk(cur_task, DLY_TICKS, USB_DLY_TSK_TICK) ;

 dprintf("[USBLPT] Start Test ..\n") ;
 
 
 if ( ! Prt_Check_Port_Pwr_Connect())
   {
	#ifdef LPT_DIAG_LOOP		// 2003/6/12 B. S. Liaw
		goto re_test ;
	#endif // LPT_DIAG_LOOP
   }
   
 /*  
 if ( ! Prt_Sts_Check_Error())
   {
	USBLPT_DBG_INIT(dprintf("[USBLPT INIT] Printer Error !!\n") );
	goto re_test ;
   }
   
 if ( ! Prt_Sts_Check_On_Line())
   {
	USBLPT_DBG_INIT(dprintf("[USBLPT INIT] Printer Off-Line !!\n") );
	goto re_test ;
   }
 */
 if( Prt_Init() ) {
	
	dprintf("[USBLPT] Open ..\n") ;
    Prt_Open();
    
    
	#ifndef LPT_DIAG_NOT_CHECK  // 2003/9/4 B. S. Liaw
	if ( ! Prt_Sts_Check_Error())
	   {
		USBLPT_DBG_INIT(dprintf("[USBLPT INIT] ParallelTest() : Printer Error !!\n") );
		#ifdef LPT_DIAG_LOOP		// 2003/6/12 B. S. Liaw
			goto re_test ;
		#endif // LPT_DIAG_LOOP
	   }
   
	 if ( ! Prt_Sts_Check_On_Line())
	   {
		USBLPT_DBG_INIT(dprintf("[USBLPT INIT] ParallelTest() : Printer Off-Line !!\n") );
		#ifdef LPT_DIAG_LOOP		// 2003/6/12 B. S. Liaw
			goto re_test ;
		#endif // LPT_DIAG_LOOP
	   }
	#endif // ndef LPT_DIAG_NOT_CHECK 
    
    /*
	#ifdef EPSON_LPT_DIAG   // 2003/4/14 B. S. Liaw
		dprintf("[USBLPT] Initialize for Epson USB Printer ..\n") ;
    	initialize_print_cmd();
	#endif // EPSON_LPT_DIAG
    */
    
	#ifdef LPT_DIAG_SPEED   // 2003/10/20 B. S. Liaw
		dprintf("[USBLPT] ParallelTest() : Test Speed ..\n") ;
		test_times = 0L ;
		start_printing_time = get_sys_time()/CLOCKHZ;
	#endif // LPT_DIAG_SPEED

	#ifdef LPT_DIAG_LOOP		// 2003/6/12 B. S. Liaw
		while(1)
	#endif // LPT_DIAG_LOOP
	  {
		/*
		#ifdef EPSON_LPT_DIAG   // 2003/4/14 B. S. Liaw
			if (usb_init_string(usblp) != D_OK) {
			    dprintf("[USBLPT] Error : init string Fail !!\n");
		       }
			continue ;
		#endif // EPSON_LPT_DIAG
		*/
		#ifdef EPSON_LPT_DIAG   // 2003/4/14 B. S. Liaw
        
			#ifndef LPT_DIAG_NOT_SHOW
				dprintf("[USBLPT] ParallelTest() : get ink level 1 ..\n") ;
			#endif // ndef LPT_DIAG_NOT_SHOW
	    	if ( !do_ink_level() )
               {
				#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
					dprintf("[USBLPT] ParallelTest() : get ink level 1 Fail !!\n") ;
				#endif // USBLPT_BIDIR
				#ifdef LPT_DIAG_LOOP		// 2003/6/12 B. S. Liaw
	                // goto re_test ;
				#endif // LPT_DIAG_LOOP
               }
        
			#ifndef LPT_DIAG_NOT_SHOW
				dprintf("[USBLPT] ParallelTest() : get status 1 ..\n") ;
			#endif // ndef LPT_DIAG_NOT_SHOW
	    	if ( ! do_status() )
               {
				dprintf("[USBLPT] ParallelTest() : Error get status 1 Fail !!\n") ;
				#ifdef LPT_DIAG_LOOP		// 2003/6/12 B. S. Liaw
            	    // goto re_test ;
				#endif // LPT_DIAG_LOOP
               }
        
			// dprintf("[USBLPT] print nozzle check ..\n") ;
	    	// do_nozzle_check();
        
		#endif // EPSON_LPT_DIAG
    
   
		#ifndef LPT_DIAG_NOT_PRINT   // 2003/6/9 B. S. Liaw
			while (! Prt_Sts_Check_Paper_End() )
               {
				USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Paper End !! \n ")) ;
                NULL ;
               }
			dprintf("[USBLPT] Print Out 1 [%s] ..\n",lpt_test_str) ;
		    if (   Prt_Buf_Write( lpt_test_str, strlen(lpt_test_str),
            				      &byte_sent)
		        != D_OK)
	    	    {
        		 return ;
		        }
                
			while (! Prt_Sts_Check_Paper_End() )
               {
				USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Paper End !! \n ")) ;
                NULL ;
               }
			dprintf("[USBLPT] Print Out 2 LF ..\n") ;
	    	if (Prt_Buf_Write(&a, 1, &byte_sent)
    	    	!= D_OK)
		        {
    		     return ;
        		}

			while (! Prt_Sts_Check_Paper_End() )
               {
				USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Paper End !! \n ")) ;
                NULL ;
               }
			dprintf("[USBLPT] Print Out 3 abc..z ..\n") ;
	    	for (i=0; i<26; i++)
	    	   {
	    	    buf[i] = 'a' + i;
		       }
    		if (Prt_Buf_Write(buf, 26, &byte_sent)
	        	!= D_OK)
		        {
	    	     return ;
        		}

			while (! Prt_Sts_Check_Paper_End() )
               {
				USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Paper End !! \n ")) ;
                NULL ;
               }
			dprintf("[USBLPT] Print Out 4 ABC..Z\n") ;
	    	for (i=0; i<26; i++) {
        		    buf[i] = 'A' + i;
		        }
	    	if (Prt_Buf_Write(buf, 26, &byte_sent)
    	    	!= D_OK)
		        {
    		     return ;
    	    	}

			while (! Prt_Sts_Check_Paper_End() )
               {
				USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Paper End !! \n ")) ;
                NULL ;
               }
			dprintf("[USBLPT] Print Out 5 0..9\n") ;
    		for (i=0; i<10; i++) {
    	    	    buf[i] = '0' + i;
		        }
    		if (Prt_Buf_Write(buf, 10, &byte_sent)
    	    	!= D_OK)
		        {
    		     return ;
		        }

			while (! Prt_Sts_Check_Paper_End() )
               {
				USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Paper End !! \n ")) ;
                NULL ;
               }
			dprintf("[USBLPT] Print Out 6 CR ..\n") ;
	    	buf[0] = '\n';
    		if (Prt_Buf_Write(buf, 1, &byte_sent)
	        	!= D_OK)
	    	    {
    		     return ;
	        	}
    
		#endif // ndef LPT_DIAG_NOT_PRINT
    
    
		#ifdef EPSON_LPT_DIAG   // 2003/4/14 B. S. Liaw
    
	    	initialize_print_cmd();
    
			#ifndef LPT_DIAG_NOT_SHOW
				dprintf("[USBLPT] ParallelTest() : get ink level 2 ..\n") ;
			#endif // ndef LPT_DIAG_NOT_SHOW
	    	if ( !do_ink_level() )
               {
				#ifdef USBLPT_BIDIR // 2003/10/13 B. S. Liaw
					dprintf("[USBLPT] ParallelTest() : get ink level 2 Fail !!\n") ;
				#endif // USBLPT_BIDIR
				#ifdef LPT_DIAG_LOOP		// 2003/6/12 B. S. Liaw
	                // goto re_test ;
				#endif // LPT_DIAG_LOOP
               }
    
			#ifndef LPT_DIAG_NOT_SHOW
				dprintf("[USBLPT] ParallelTest() : get status 2..\n") ;
			#endif // ndef LPT_DIAG_NOT_SHOW
	    	if ( ! do_status() )
               {
				dprintf("[USBLPT] ParallelTest() : Error get status 2 Fail !!\n") ;
				#ifdef LPT_DIAG_LOOP		// 2003/6/12 B. S. Liaw
    	            // goto re_test ;
				#endif // LPT_DIAG_LOOP
               }
        
			#ifndef LPT_DIAG_NOT_PRINT   // 2003/6/9 B. S. Liaw
				while (! Prt_Sts_Check_Paper_End() )
	               {
					USBLPT_DBG_INIT(iprintf("[USBLPT INIT] Paper End !! \n ")) ;
	                NULL ;
    	           }
				dprintf("[USBLPT] print nozzle check ..\n") ;
	    		do_nozzle_check();
			#endif // ndef LPT_DIAG_NOT_PRINT
    
		#endif // EPSON_LPT_DIAG

		#ifdef LPT_DIAG_SPEED   // 2003/10/20 B. S. Liaw
			test_times ++ ;
	
			if ( (start_printing_time != 0) &&
			     ((get_sys_time()/CLOCKHZ)-start_printing_time >= 30)) {
				dprintf("[USBLPT] ParallelTest() : speed=%d/30sec \n" , test_times) ;
				dprintf("[USBLPT] ParallelTest() : Test Speed End.\n") ;
				break ;
			   }
		#endif // LPT_DIAG_SPEED
       } // infinte while for LPT_DIAG_LOOP
    
	dprintf("[USBLPT] Close ..\n") ;
	Prt_Close();
   } //  if Prt_Init() 
 dprintf("[USBLPT] End of Test.\n") ;

 dprintf("[USBLPT] ============================\n") ;
}

// ====================================================================




// ====================================================================
// End of file : UsbLpt.c

