
// Name of File : usb_osport_pci.h
// ====================================================================

// ====================================================================
// include/linux/pci.h
// ====================================================================
/* This defines the direction arg to the DMA mapping routines. */

#define PCI_DMA_BIDIRECTIONAL   0
#define PCI_DMA_TODEVICE        1
#define PCI_DMA_FROMDEVICE      2
#define PCI_DMA_NONE            3

#define DEVICE_COUNT_COMPATIBLE 4
#define DEVICE_COUNT_IRQ        2
#define DEVICE_COUNT_DMA        2
#define DEVICE_COUNT_RESOURCE   12


#define PCI_ANY_ID (~0)

#define pci_present pcibios_present


#define pci_for_each_dev_reverse(dev) \
        for(dev = pci_dev_g(pci_devices.prev); dev != pci_dev_g(&pci_devices); dev = pci_dev_g(dev->global_list.prev))

#define pci_for_each_bus(bus) \
for(bus = pci_bus_b(pci_root_buses.next); bus != pci_bus_b(&pci_root_buses); bus = pci_bus_b(bus->node.next))


#define pci_for_each_dev(dev) \
        for(dev = pci_dev_g(pci_devices.next); dev != pci_dev_g(&pci_devices); dev = pci_dev_g(dev->global_list.next))




/*
 * The pci_dev structure is used to describe both PCI and ISAPnP devices.
 */
struct pci_dev {
        
 struct list_head global_list;   // node in list of all PCI devices 
 struct list_head bus_list;      // node in per-bus list 
 /*
 struct pci_bus  *bus;           // bus this device is on 
 struct pci_bus  *subordinate;   // bus this device bridges to 

 void            *sysdata;       // hook for sys-specific extension 
 struct proc_dir_entry *procent; // device entry in /proc/bus/pci 

 unsigned int    devfn;          // encoded device & function index 
 unsigned short  vendor;
 unsigned short  device;
 unsigned short  subsystem_vendor;
 unsigned short  subsystem_device;
 unsigned int    class;          // 3 bytes: (base,sub,prog-if) 
 u8              hdr_type;       // PCI header type (`multi' flag masked out) 
 u8              rom_base_reg;   // which config register controls the ROM 
 */
 struct pci_driver *driver;      // which driver has allocated this device 
 void            *driver_data;   // data private to the driver 
 /*
 u64             dma_mask;       // Mask of the bits of bus address this
                                 //   device implements.  Normally this is
                                 //   0xffffffff.  You only need to change
                                 //   this if your device has broken DMA
                                 //   or supports 64-bit transfers.  

 u32             current_state;  // Current operating state. In ACPI-speak,
                                 //   this is D0-D3, D0 being fully functional,
                                 //   and D3 being off. 
 */
 
 struct device  dev;            /* Generic device interface */

 /*
 // device is compatible with these IDs 
 unsigned short vendor_compatible[DEVICE_COUNT_COMPATIBLE];
 unsigned short device_compatible[DEVICE_COUNT_COMPATIBLE];

 
  // Instead of touching interrupt line and base address registers
  // directly, use the values stored here. They might be different!
 */ 
 unsigned int    irq;
 /*
 struct resource resource[DEVICE_COUNT_RESOURCE]; // I/O and memory regions + expansion ROMs 
 
 struct resource dma_resource[DEVICE_COUNT_DMA];
 struct resource irq_resource[DEVICE_COUNT_IRQ];
 */
 #ifdef USB_DBGSTRING_ON
	char            name[80];       // device name 
	char            slot_name[8];   // slot name 
 #endif
 /*
 int             active;         // ISAPnP: device is active 
 int             ro;             // ISAPnP: read only 
 unsigned short  regs;           // ISAPnP: supported registers 

 int (*prepare)(struct pci_dev *dev);    // ISAPnP hooks 
 int (*activate)(struct pci_dev *dev);
 int (*deactivate)(struct pci_dev *dev);
 */
};

#define pci_dev_g(n) list_entry(n, struct pci_dev, global_list)
#define pci_dev_b(n) list_entry(n, struct pci_dev, bus_list)

/* Now on "mod_devicetable.h"
struct pci_device_id {
        unsigned int vendor, device;            // Vendor and device ID or PCI_ANY_ID 
        unsigned int subvendor, subdevice;      // Subsystem ID's or PCI_ANY_ID 
        unsigned int class, class_mask;         // (class,subclass,prog-if) triplet 
        unsigned long driver_data;              // Data private to the driver 
};
*/
struct pci_driver {
        
 struct list_head node;
 
 #ifdef USB_DBGSTRING_ON
	char *name;
 #endif // USB_DBGSTRING_ON 
 
 const struct pci_device_id *id_table;   /* NULL if wants all devices */
 #ifndef _ETCPIP
	int (*probe) (struct pci_dev *dev, const struct pci_device_id *id);
         /* New device inserted */
 #else
	int  (*probe)( struct pci_dev *dev, const struct pci_device_id *id,
			int irq , void * mem_base ) ;
 #endif
 
 void (*remove) (struct pci_dev *dev);   /* Device removed (NULL if not a hot-plug capable driver) */
        
        //int  (*save_state) (struct pci_dev *dev, u32 state);    // Save Device Context 
 #ifdef	CONFIG_PM
        int  (*suspend)(struct pci_dev *dev, u32 state);        // Device suspended 
        int  (*resume) (struct pci_dev *dev);                   // Device woken up 
 #endif
        //int  (*enable_wake) (struct pci_dev *dev, u32 state, int enable);   // Enable wake event 
        
};

extern int
pci_register_driver(struct pci_driver *drv) ;


/*
 * a helper function which helps ensure correct pci_driver
 * setup and cleanup for commonly-encountered hotplug/modular cases
 *
 * This MUST stay in a header, as it checks for -DMODULE
 */
#ifndef _ETCPIP
static inline int pci_module_init(struct pci_driver *drv)
{
 // int rc = pci_register_driver (drv);
 int rc ;
 
 USB_DBG_FLOW(iprintf("[USB FLOW] pci_module_init() Begin ..\n"));

 rc = pci_register_driver (drv);
 if (rc > 0)
    {
     USB_DBG_FLOW(iprintf("[USB FLOW] pci_module_init() End. \n"));
     return 0;
    }
 /* iff CONFIG_HOTPLUG and built into kernel, we should
  * leave the driver around for future hotplug events.
  * For the module case, a hotplug daemon of some sort
  * should load a module in response to an insert event. */
 #if defined(CONFIG_HOTPLUG) && !defined(MODULE)
    if (rc == 0)
       return 0;
  #else
    if (rc == 0)
       {
        USB_DBG_FLOW(iprintf(
           "[USB FLOW] pci_module_init() End with Fail -ENODEV !! \n"
           ));
        rc = -ENODEV;
       }
 #endif

 /* if we get here, we need to clean up pci driver instance
  * and return some sort of error */
 
 #ifndef _ETCPIP
        pci_unregister_driver (drv);
 #endif // ndef _ETCPIP
 USB_DBG_FLOW(iprintf("[USB FLOW] pci_module_init() End with Fail !! \n"));
 return rc;
}
#endif // ndef _ETCPIP



/* these helpers provide future and backwards compatibility
 * for accessing popular PCI BAR info */
 /*
#define pci_resource_start(dev,bar)   ((dev)->resource[(bar)].start)
#define pci_resource_end(dev,bar)     ((dev)->resource[(bar)].end)
#define pci_resource_flags(dev,bar)   ((dev)->resource[(bar)].flags)
#define pci_resource_len(dev,bar) \
        ((pci_resource_start((dev),(bar)) == 0 &&       \
          pci_resource_end((dev),(bar)) ==              \
          pci_resource_start((dev),(bar))) ? 0 :        \
                                                        \
         (pci_resource_end((dev),(bar)) -               \
          pci_resource_start((dev),(bar)) + 1))
*/

/* Similar to the helpers above, these manipulate per-pci_dev
 * driver-specific data.  Currently stored as pci_dev::driver_data,
 * a void pointer, but it is not present on older kernels.
 */
static inline void *pci_get_drvdata (struct pci_dev *pdev)
{
        return pdev->driver_data;
}

static inline void pci_set_drvdata (struct pci_dev *pdev, void *data)
{
 pdev->driver_data = data;
}


// ====================================================================
// include/linux/pci_ids.h
// ====================================================================

#define PCI_BASE_CLASS_SERIAL           0x0c
#define PCI_CLASS_SERIAL_FIREWIRE       0x0c00
#define PCI_CLASS_SERIAL_ACCESS         0x0c01
#define PCI_CLASS_SERIAL_SSA            0x0c02
#define PCI_CLASS_SERIAL_USB            0x0c03
#define PCI_CLASS_SERIAL_FIBER          0x0c04
#define PCI_CLASS_SERIAL_SMBUS          0x0c05

// ====================================================================
// include/asm-armnommu/pci.h
// ====================================================================

#ifdef USB_MEM_LIB
	#include "UsMemLib.h" // UsbMemLib.h
#endif

#ifdef CONFIG_PCI
	static inline void
	pci_free_consistent(struct pci_dev *hwdev, size_t size,
    		void *vaddr, dma_addr_t dma_handle)
#else
  #error : Please CONFIG_PCI
  /*
	static inline void
	mem_free_consistent(size_t size,
    		void *vaddr, dma_addr_t dma_handle)
  */
#endif
{
 #ifndef _ETCPIP
	#warning : No consistent_free() !!
	consistent_free(vaddr, size, dma_handle);
 #elif !defined(USB_MEM_LIB)
	#warning : No consistent_free() !!
	// NonCache_contigmalloc() cannot be free now !!
 #elif !defined(USB_NO_DMA_POOL)
	UsbFreeMemU(USB_MEM_U_L, (unsigned char *)dma_handle) ;
 #else
    UsbFreeMemU(USB_MEM_U_M, (unsigned char *)dma_handle
			#if defined(USB_DBG_MEM_U_ON)
		    	, size
    		#endif
            ) ;
 #endif
}


/* Map a single buffer of the indicated size for DMA in streaming mode.
 * The 32-bit bus address to use is returned.
 *
 * Once the device is given the dma address, the device owns this memory
 * until either pci_unmap_single or pci_dma_sync_single is performed.
 */
#ifdef CONFIG_PCI
	static inline dma_addr_t
	pci_map_single(struct pci_dev *hwdev, void *ptr, size_t size,
    			 int direction)
	{
	 consistent_sync(ptr, size, direction);
	 return virt_to_bus(ptr);
	}
#endif


/* Unmap a single streaming mode DMA translation.  The dma_addr and size
 * must match what was provided for in a previous pci_map_single call.  All
 * other usages are undefined.
 *
 * After this call, reads by the cpu to the buffer are guarenteed to see
 * whatever the device wrote there.
 */
#ifdef CONFIG_PCI
	static inline void
	pci_unmap_single(struct pci_dev *hwdev, dma_addr_t dma_addr,
    				 size_t size, int direction)
	{
	 /* nothing to do */
	}
#endif



// ====================================================================

#ifdef CONFIG_PCI
	// in usb_osport.c
	extern void *pci_alloc_consistent(struct pci_dev *hwdev, size_t size,
          dma_addr_t *handle) ;
#endif

// ====================================================================
// End of File : usb_osport_pci.h

