
// Name of File : FatBlockIo.c
// ==================================================================== 

/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001-2003, Salvatore Isaja                               *
 *                                                                        *
 * This is "blockio.c" - Services to access the hosting block device      *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

// ==================================================================== 

#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
	// for deciding whether to compile
	//  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsbMsInit.h"

#include "UsOsport.h" // "usb_osport.h"

#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h


#ifdef _USB_SYSLOG
	#include "syslog.h"
	#include "logStr.h"

	#include "config.h"
	extern struct III_Config_t gSetting;
 #else
	// #error: Please Make Sure : Not _USB_SYSLOG !!
	// #warning: Not _USB_SYSLOG !!
#endif // _USB_SYSLOG


// ====================================================================

/* 
//#include <sys/types.h>
//#include <sockapi.h>

#include "nglobal.h"

#include "fat.h"
#include "fatfs_global.h"
*/
// ====================================================================

// 2003/12/17 B. S. Liaw : 
// usb_block_read() and usb_block_write() may be replaced by functions 
// on hardware/usb/storage/FatBlockIo.c
// so , need include blockio.h
// on /kernel/fatfs/blockio.h :
// int usb_block_read(unsigned long VolumeID, unsigned long SectorNo, void *Buffer);
// int usb_block_write(unsigned long VolumeID, unsigned long SectorNo, void *Buffer);
#include "blockio.h"      // on /kernel/fatfs

#include "blkdev.h"


// ====================================================================
// Note : PartitionID : 0    : whole Physical Disk
//                      >= 1 : Partition (Logical Disk )
//                      i.e. count from 1
//        VolumeID    : for upper FAT Layer , count from 0
//
//        PartitionID = VolumeID + 1
// ====================================================================


// int UsbDiskOpened = FALSE;
// int UsbDiskOpened[USBMS_MAX_MINORS+1] ;

// B. S. Liaw : Sector_buf need uncached
// unsigned char *Sector_buf = NULL ;

// struct block_device *disk_bdev = NULL ;

// ====================================================================
/*
#ifdef USBMS_DBG_EMU_FLOPPY
	// only for testing
	BYTE Floppy144_Boot[512] = {0xEB, 0x4C, 0x90, 0x2B, 0x24, 0x2B, 0x63, 0x79, 0x49, 0x48, 0x43, 0x00, 0x02, 0x01, 0x01, 0x00,  // kL.+$+cyIHC.....
                            0x02, 0xE0, 0x00, 0x40, 0x0B, 0xF0, 0x09, 0x00, 0x12, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00,  // .`.@.p..........
                            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x29, 0x78, 0x56, 0x34, 0x12, 0x4E, 0x4F, 0x20, 0x4E, 0x41,  // ......).vpdNO NA
                            0x4D, 0x45, 0x2D, 0x2D, 0x2D, 0x2D, 0x46, 0x41, 0x54, 0x31, 0x32, 0x20, 0x20, 0x20, 0x33, 0xC9,  // ME----FAT12   3I
                            0x8E, 0xD1, 0xBC, 0xF0, 0x7B, 0x8E, 0xD9, 0xB8, 0x00, 0x20, 0x8E, 0xC0, 0xFC, 0xBD, 0xFC, 0x31,  // .Q<p{.Y8. .@|=|1
                            0xC0, 0x8E, 0xD8, 0xBD, 0x00, 0x7C, 0x8E, 0xD0, 0x8D, 0x66, 0xE0, 0x8A, 0x56, 0x24, 0xCD, 0x13,  // @.X=.|.P.f`.V$M.
                            0xB8, 0xE0, 0x1F, 0x8E, 0xC0, 0x89, 0xEE, 0x89, 0xEF, 0xB9, 0x00, 0x01, 0xF3, 0xA5, 0x06, 0xBB,  // 8`..@.n.o9..s%.;
                            0x74, 0x7C, 0x53, 0xCB, 0x8E, 0xD8, 0x8E, 0xD0, 0xE8, 0xEA, 0x00, 0x4C, 0x6F, 0x61, 0x64, 0x69,  // t|SK.X.Phj.Loadi
                            0x6E, 0x67, 0x20, 0x46, 0x72, 0x65, 0x65, 0x44, 0x4F, 0x53, 0x2E, 0x2E, 0x2E, 0x0D, 0x0A, 0x52,  // ng FreeDOS.....R
                            0x4F, 0x4F, 0x54, 0x00, 0xCD, 0x03, 0x8B, 0x46, 0x46, 0x8B, 0x56, 0x48, 0x8B, 0x7E, 0x40, 0x31,  // OOT.M..FF.VH.~@1
                            0xDB, 0xC7, 0x46, 0x3E, 0x60, 0x00, 0x8E, 0x46, 0x3E, 0xE8, 0xC9, 0x00, 0x72, 0x38, 0x31, 0xFF,  // [GF>`..F>hI.r81.
                            0xB9, 0x0B, 0x00, 0xBE, 0xC1, 0x7D, 0x57, 0xF3, 0xA6, 0x5F, 0x26, 0x8B, 0x45, 0x1A, 0x74, 0x0B,  // 9..>A}Ws&_&.E.t.
                            0x83, 0xC7, 0x20, 0x26, 0x80, 0x3D, 0x00, 0x75, 0xE7, 0x72, 0x6F, 0x50, 0xE8, 0x96, 0x00, 0x20,  // .G &.=.ugroPh..
                            0x46, 0x41, 0x54, 0x00, 0x8E, 0x46, 0x3E, 0x31, 0xDB, 0x8B, 0x7E, 0x16, 0x8B, 0x46, 0x42, 0x8B,  // FAT..F>1[.~..FB.
                            0x56, 0x44, 0xE8, 0x90, 0x00, 0x58, 0x72, 0x52, 0x1E, 0x06, 0x1F, 0x07, 0xBF, 0x00, 0x20, 0xAB,  // VDh..XrR....?. +
                            0x89, 0xC6, 0x01, 0xF6, 0x01, 0xC6, 0xD1, 0xEE, 0xAD, 0x73, 0x04, 0xB1, 0x04, 0xD3, 0xE8, 0x80,  // .F.v.FQn-s.1.Sh.
                            0xE4, 0x0F, 0x3D, 0xFF, 0x0F, 0x72, 0xE8, 0x31, 0xC0, 0xAB, 0x0E, 0x1F, 0xE8, 0x56, 0x00, 0x20,  // d.=..rh1@+..hV.
                            0x4B, 0x45, 0x52, 0x4E, 0x45, 0x4C, 0x00, 0x8E, 0x46, 0x3E, 0x31, 0xDB, 0xBE, 0x00, 0x20, 0xAD,  // KERNEL..F>1[>. -
                            0x09, 0xC0, 0x74, 0x2F, 0x48, 0x48, 0x8B, 0x7E, 0x0D, 0x81, 0xE7, 0xFF, 0x00, 0xF7, 0xE7, 0x03,  // .@t/HH.~..g..wg.
                            0x46, 0x4A, 0x13, 0x56, 0x4C, 0xE8, 0x3D, 0x00, 0x73, 0xE5, 0xE8, 0x28, 0x00, 0x0D, 0x0A, 0x42,  // FJ.VLh=.seh(...B
                            0x4F, 0x4F, 0x54, 0x20, 0x65, 0x72, 0x72, 0x6F, 0x72, 0x21, 0x0D, 0x0A, 0x00, 0x30, 0xE4, 0xCD,  // OOT error!...0dM
                            0x16, 0xCD, 0x19, 0xE8, 0x0F, 0x00, 0x20, 0x47, 0x4F, 0x21, 0x0D, 0x0A, 0x00, 0x8A, 0x5E, 0x24,  // .M.h.. GO!....^$
                            0xEA, 0x00, 0x00, 0x60, 0x00, 0x5E, 0x31, 0xDB, 0xB4, 0x0E, 0xAC, 0x3C, 0x00, 0x74, 0x04, 0xCD,  // j..`.^1[4.,<.t.M
                            0x10, 0xEB, 0xF7, 0x56, 0xC3, 0x56, 0x52, 0x50, 0xF7, 0x76, 0x18, 0x89, 0xD1, 0x31, 0xD2, 0xF7,  // .kwVCVRPwv..Q1Rw
                            0x76, 0x1A, 0x88, 0xD6, 0xD0, 0xCC, 0xD0, 0xCC, 0x86, 0xC4, 0x8A, 0x56, 0x18, 0x28, 0xCA, 0xFE,  // v..VPLPL.D.V.(J~
                            0xC1, 0x09, 0xC1, 0x88, 0xD0, 0xB8, 0x01, 0x02, 0x8A, 0x56, 0x24, 0xCD, 0x13, 0x73, 0x08, 0x30,  // A.A.P8...V$M.s.0
                            0xE4, 0xCD, 0x13, 0x58, 0x5A, 0xEB, 0xCF, 0x03, 0x5E, 0x0B, 0x73, 0x07, 0x8C, 0xC0, 0x80, 0xC4,  // dM.XZkO.^.s..@.D
                            0x10, 0x8E, 0xC0, 0x58, 0x5A, 0x05, 0x01, 0x00, 0x83, 0xD2, 0x00, 0x4F, 0x75, 0xB8, 0xF8, 0x5E,  // ..@XZ....R.Ou8x^
                            0xC3, 0x4B, 0x45, 0x52, 0x4E, 0x45, 0x4C, 0x20, 0x20, 0x53, 0x59, 0x53, 0x00, 0x00, 0x00, 0x00,  // CKERNEL  SYS....
                            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  // ................
                            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  // ................
                            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x55, 0xAA}; // ..............U*

#endif // USBMS_DBG_EMU_FLOPPY
*/

#ifdef USBMS_SCSI_SHORTCUT  /* Charles 20040526 */
#include "ms_usb.h"
#include "scsi_cmnd.h"

struct us_data *MSUsb_us;
unsigned long serial_number;
#endif
unsigned char USBDeviceRW=0;

// for runtime debug from console
#define USBMS_LOG_FAT_BLK_ENTRIES		256
/*
static unsigned char usb_ms_log_init = 0, usb_ms_log_entry;
unsigned char usb_ms_log_status[USBMS_LOG_FAT_BLK_ENTRIES];
static unsigned char usb_ms_log_volumn[USBMS_LOG_FAT_BLK_ENTRIES];
static unsigned char usb_ms_log_nr_sector[USBMS_LOG_FAT_BLK_ENTRIES];
static unsigned long usb_ms_log_sector[USBMS_LOG_FAT_BLK_ENTRIES];
*/
unsigned char usb_ms_log_init = 0, usb_ms_log_entry;
unsigned char usb_ms_log_status[USBMS_LOG_FAT_BLK_ENTRIES];
unsigned char usb_ms_log_volumn[USBMS_LOG_FAT_BLK_ENTRIES];
unsigned char usb_ms_log_nr_sector[USBMS_LOG_FAT_BLK_ENTRIES];
unsigned long usb_ms_log_sector[USBMS_LOG_FAT_BLK_ENTRIES];

static
void usb_storage_query_inbound_log_init(void)
{
	if (usb_ms_log_init == 0) {
		usb_ms_log_entry = 0;
		usb_ms_log_init = 1;
	}
}

unsigned long usb_storage_query_inbound_index(void)
{
	unsigned long	index;
	MASK_INTS();
	index = usb_ms_log_entry;
	if (usb_ms_log_entry < (USBMS_LOG_FAT_BLK_ENTRIES-1))	usb_ms_log_entry ++;
		else												usb_ms_log_entry = 0;
	UNMASK_INTS();
	return	index;
}

static inline
void usb_storage_query_inbound_log_status(unsigned long index, unsigned char status)
{
	usb_ms_log_status[index] = status;
}

void usb_storage_query_inbound_log(unsigned long index, unsigned char status
									, unsigned char volume, unsigned long sector_no)
{
	usb_storage_query_inbound_log_status(index, status);
	usb_ms_log_volumn[index] = volume;
	usb_ms_log_nr_sector[index] = 1;
	usb_ms_log_sector[index] = sector_no;
}

static
unsigned long usb_storage_query_inbound_log_compress(unsigned long index, unsigned char status)
{
	unsigned long rt_index, next_index, prev_index;

	rt_index = index;
	if (index < (USBMS_LOG_FAT_BLK_ENTRIES-1))	next_index = index + 1;
		else									next_index = 0;
	if (index > 0)	prev_index = index - 1;
		else		prev_index = (USBMS_LOG_FAT_BLK_ENTRIES-1);
	MASK_INTS();
	if ( (next_index == usb_ms_log_entry)
		&& (status == usb_ms_log_status[prev_index])
		&& (usb_ms_log_volumn[index] == usb_ms_log_volumn[prev_index])
		&& (usb_ms_log_sector[index]
			== (usb_ms_log_sector[prev_index]+((unsigned long)(usb_ms_log_nr_sector[prev_index]) & 0xff))) ) {
		usb_ms_log_entry = index;
		usb_ms_log_nr_sector[prev_index] ++;
		rt_index = prev_index;
	}
	UNMASK_INTS();

	return	rt_index;
}

void usb_storage_query_inbound_display(void)
{
	unsigned long	start_idx, index, i;

dprintf("bnchen test x%08x x%08x x%08x\n", index, usb_ms_log_volumn, usb_ms_log_sector);
	MASK_INTS();
	start_idx = usb_ms_log_entry;
	UNMASK_INTS();
	index = start_idx;
	i = 0;
	dprintf("inbound requests : w/r q ca hit mis lok syn v");
	do {
		if ((i & 0x01) == 0)
			dprintf("\n%04 -", i);
		dprintf(" x%02x:%08x(x%02x) x%02x"
				, usb_ms_log_volumn[index], usb_ms_log_sector[index]
				, usb_ms_log_nr_sector[index], usb_ms_log_status[index]);
		if (index < (USBMS_LOG_FAT_BLK_ENTRIES-1))	index++;
			else							index = 0;
		i ++;
	} while (index != start_idx);
	dprintf("\n");
}

// ====================================================================

int OpenUsbDisk(unsigned long PartitionID)
{
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] OpenUsbDisk() %d.. \n", PartitionID));
 USBMS_DBG_INIT(iprintf("[USBMS INIT] OpenUsbDisk() .. \n"));

 if(UsbDiskOpened[PartitionID]) {
	USBMS_DBG_INIT(iprintf("[USBMS INIT] OpenUsbDisk() : Already Opened !! \n"));
    return 1 ;
   }

 /*
 open_by_devnum((dev_t)(0x100001),FMODE_READ, BDEV_RAW) ;
 */
 /*
 #ifndef USBMS_ONLY_ONE_DISK
	disk_bdev = open_by_devnum((dev_t)(0x100001),FMODE_READ, BDEV_RAW) ;
  #else
 */
 if ( PartitionID>0 ) {
		// if(!disk_global_inode || !disk_global_bdev) {
		// if(!disk_global_bdev || !disk_global_bdev->bd_inode)
		// disk_bdev = open_by_devnum((dev_t)(0x0001),FMODE_READ, BDEV_RAW) ;
		disk_global_bdev[PartitionID]
			= open_by_devnum( (dev_t)(MKDEV(0,0)+ PartitionID),
			// = open_by_devnum( (dev_t)(MKDEV(0,0)+ PartitionID+5),
			// = open_by_devnum( (dev_t)(MKDEV(0,0)+ PartitionID+4),   // 1+4=5
			// = open_by_devnum( (dev_t)(MKDEV(0,0)+ PartitionID+2),   // 1+2=3
			// = open_by_devnum( (dev_t)(MKDEV(0,0)+ PartitionID+1),   // 1+1=2
											FMODE_READ, BDEV_RAW) ;
	   }
 /*
 #endif // USBMS_ONLY_ONE_DISK
 */

 if(!disk_global_bdev[PartitionID] || !disk_global_bdev[PartitionID]->bd_inode) {
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] OpenUsbDisk() : Error : Disk Not Init !! \n"));
	return -2;
   }

 /*
 // B. S. Liaw : Sector_buf need uncached
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] OpenUsbDisk() : Sector_buf = UsbAllocMemU() size=%d .. \n",
	USBMS_SECTOR_SIZE
	));
 Sector_buf = (unsigned char *) UsbAllocMemU(USB_MEM_U_M
				#ifdef USB_DBG_MEM_U_ON
					, USBMS_SECTOR_SIZE   // size : 512
		        #endif
           		) ;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] OpenUsbDisk() : Sector_buf=%08XH \n",
	Sector_buf
	));
 */
 if(!disk_global_Sector_buf) {
	// USBMS_DBG_INFO(iprintf("[USBMS INFO] OpenUsbDisk() : Error : Out of Memory !! \n"));
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] OpenUsbDisk() : Error : Null disk_global_Sector_buf !! \n"));
	USBMS_DBG_FLOW(iprintf(
		"[USBMS FLOW] OpenUsbDisk(1) End with Failed !! \n"));
	return -2;
   }

 UsbDiskOpened[PartitionID] = TRUE;
 usb_storage_query_inbound_log_init();
 return 1;

}

int CloseUsbDisk(unsigned long PartitionID)
{
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] CloseUsbDisk(): UsbDiskOpened[PartitionID] = %d \n",
	UsbDiskOpened[PartitionID]
	));
 USBMS_DBG_INFO(iprintf("[USBMS INFO] CloseUsbDisk(): UsbDiskOpened[PartitionID] = %d \n",
	UsbDiskOpened[PartitionID]
	));

 if(! UsbDiskOpened[PartitionID]){
	USBMS_DBG_INIT(iprintf("[USBMS INIT] CloseUsbDisk() : Disk Already Plug-Out !! \n"));
	return 0;
   }
 /*
 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] CloseUsbDisk() : UsbFreeMemU(Sector_buf) , size=%d .. \n" ,
	USBMS_SECTOR_SIZE   // size : 512
	));
 UsbFreeMemU(USB_MEM_U_M,(unsigned char *) Sector_buf
			#if defined(USB_DBG_MEM_U_ON)
				, USBMS_SECTOR_SIZE   // size : 512
			#endif
			) ;
 Sector_buf = NULL ;
 */
 UsbDiskOpened[PartitionID] = FALSE;
 return 0;
}

#ifdef USBMS_SCSI_SHORTCUT  /* Charles 20040528 */
int read_dev_sectors(struct inode *inode, struct block_device *bdev, sector_t n,
			unsigned int num_sectors, unsigned char *Sector_buf, sector_t diskSector)
{
 struct scsi_cmnd *srb;	

 srb = MSUsb_us->srb;
 
 srb->sc_data_direction = SCSI_DATA_READ;
 srb->request_buffer = Sector_buf;
 srb->request_bufflen = num_sectors * ReadCacheSectorSize;
 srb->serial_number = ++serial_number;
 srb->cmd_len = 10;
 srb->cmnd[0] = READ_10;
 srb->cmnd[2] = (unsigned char) (diskSector >> 24) & 0xff;
 srb->cmnd[3] = (unsigned char) (diskSector >> 16) & 0xff;
 srb->cmnd[4] = (unsigned char) (diskSector >> 8) & 0xff;
 srb->cmnd[5] = (unsigned char) diskSector & 0xff;
 srb->cmnd[6] = srb->cmnd[9] = 0;
 srb->cmnd[7] = (unsigned char) (num_sectors >> 8) & 0xff;
 srb->cmnd[8] = (unsigned char) num_sectors & 0xff;
 
 if ( MSUsb_us->transport(srb, MSUsb_us) == -ENODEV )
    return read_dev_sector(inode,bdev,n,num_sectors,Sector_buf);

 return 1;	
}

int write_dev_sectors(struct inode *inode, struct block_device *bdev, sector_t n,
			unsigned int num_sectors, unsigned char *Sector_buf, sector_t diskSector)
{
 struct scsi_cmnd *srb;	

 srb = MSUsb_us->srb;
 
 srb->sc_data_direction = SCSI_DATA_WRITE;
 srb->request_buffer = Sector_buf;
 srb->request_bufflen = num_sectors * ReadCacheSectorSize;
 srb->serial_number = ++serial_number;
 srb->cmd_len = 10;
 srb->cmnd[0] = WRITE_10;
 srb->cmnd[2] = (unsigned char) (diskSector >> 24) & 0xff;
 srb->cmnd[3] = (unsigned char) (diskSector >> 16) & 0xff;
 srb->cmnd[4] = (unsigned char) (diskSector >> 8) & 0xff;
 srb->cmnd[5] = (unsigned char) diskSector & 0xff;
 srb->cmnd[6] = srb->cmnd[9] = 0;
 srb->cmnd[7] = (unsigned char) (num_sectors >> 8) & 0xff;
 srb->cmnd[8] = (unsigned char) num_sectors & 0xff;
 
 if ( MSUsb_us->transport(srb, MSUsb_us) == -ENODEV )
    return write_dev_sector(inode,bdev,n,num_sectors,Sector_buf);

 return 1;	
}
#endif


// ==================================================================== 
// ==================================================================== 

// fat_read()						drivers\fat\readwrit.c
//	fat_readbuf()					drivers\fat\blockio.c
//		biosdisk_request(FD32_BLOCKREAD)	drivers\biosdisk\request.c
//			biosdisk_read()			drivers\biosdisk\common.c
//				retry
//				biosdisk_extread()	drivers\biosdisk\extdisk.c
//				biosdisk_stdread()	drivers\biosdisk\stddisk.c

// ==================================================================== 
// usb_block_read() and usb_block_write() are origial on 
// hardware/usb/storage/FatBlockIo.c
// and replaced by the following ..
// on /kernel/fatfs/blockio.c
// return value : 1 : OK , 0 : Fail , -1 I/O Error or Plug-Out
// Charles 20040518: Modify for Read/Write Cache check before calling lower layer functions!
// Bonian 20050328: Use background task to perform read operation
// ==================================================================== 

// special copy function for blocks. (from non-cacheable memory to cacheable memory)
// address should be alignment to 4 and length should be multiply of 4
void usb_block_copy(void *dest, void *src, unsigned long length)
{
	unsigned long	i, len;
	len = length >> 2;
	for (i = 0; i < len; i++)
		((unsigned long *)dest)[i] = ((unsigned long *)src)[i];
}

int usb_block_readex(unsigned long VolumeID, unsigned long SectorNo, void *Buffer, unsigned short max_prefetch_sectors)
{
	unsigned long	log_index;
 struct block_device *bdev;
	
 USBMS_DBG_SECTOR(iprintf("[USBMS SECTOR] usb_block_read(): VolumeID=%ld SectorNo=%ld Buffer=%08XH \n",
	VolumeID, SectorNo,Buffer));
 USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_block_read(): VolumeID=%ld SectorNo=%ld Buffer=%08XH \n",
	VolumeID, SectorNo,Buffer));
 #ifndef USBMS_DBG_DISABLE_FATIF
	USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_block_read(): UsbDiskOpened[VolumeID+1]=%d\n",
		UsbDiskOpened[VolumeID+1]));

	#ifdef USB_DBG_MEM_COUNT_ON
	#ifdef USBMS_DBG_INIT_ON
		iprintf("[USBMS INIT] usb_block_read(1) ====== show USB Mem Count .. \n") ;
		showUsbMemC() ;
		showUsbMemU() ;
	#endif
	#endif // USB_DBG_MEM_COUNT_ON

	log_index = usb_storage_query_inbound_index();
	usb_storage_query_inbound_log(log_index, 0x01, VolumeID, SectorNo);

	if(! UsbDiskOpened[VolumeID+1]) {
		USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_block_read() : OpenUsbDisk() .. \n"));
		if ( !OpenUsbDisk(VolumeID+1) ) {
			USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_block_read(1) : Error : Open Usb Disk Failed !! \n"));
			return -1;
		   }
		USBMS_DBG_PLUG(iprintf("[USBMS PLUG] usb_block_read() : UsbDiskOpened[VolumeID+1]=%d\n",
			UsbDiskOpened[VolumeID+1]));
		USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_block_read() : UsbDiskOpened[VolumeID+1]=%d\n",
			UsbDiskOpened[VolumeID+1]));
		if (! UsbDiskOpened[VolumeID+1]) {
			USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_block_read(2) : Error : Open Usb Disk Failed !! \n"));
			return -1;
		   }
	   }

	bdev = disk_global_bdev[VolumeID+1];
	if (bdev && bdev->bd_inode)
	{
		if (SectorNo > bdev->bd_part->nr_sects)
		{
			iprintf("[USB] Read sector %u is larger than the total number of this HD's sectors(%u).\n",
				SectorNo, bdev->bd_part->nr_sects);
			return 0;
		}

	   		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_block_read() : read_dev_sector() .. \n"));
	   		if (USBDeviceRW++)
	   			return 3;
			usb_storage_query_inbound_log_status(log_index, 0x41);
			if (read_dev_sector(bdev->bd_inode,bdev,SectorNo,1,disk_global_Sector_buf) <= 0)
			{
				USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_block_read(2) : Error : Read Sector Failed !! \n"));
				dprintf("[USBMS] Error : Read Sector Failed !! \n");
				USBDeviceRW = 0;
				usb_storage_query_inbound_log_status(log_index, 0x49);
				return 0 ;
		   	}
	   		usb_block_copy(Buffer,disk_global_Sector_buf,bdev_hardsect_size(bdev));
	   		USBDeviceRW = 0;
			usb_storage_query_inbound_log_status(log_index, 0x51);
	}
	else
	{
		USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_block_read(6) : Error : Disk Not Init !! \n"));
		dprintf("[USBMS] Sector Read Error : No Device !! \n");
		return -1;
	}
	
	#ifdef USB_DBG_MEM_COUNT_ON
	#ifdef USBMS_DBG_INIT_ON
		iprintf("[USBMS INIT] usb_block_read(2) ====== show USB Mem Count .. \n"
    		) ;
		showUsbMemC() ;
		showUsbMemU() ;
	#endif
	#endif // USB_DBG_MEM_COUNT_ON
 #else
	iprintf("[USBMS DISABLE_FATIF] usb_block_read() : Not Read !! \n") ;
	return 0 ;
 #endif // ndef USBMS_DBG_DISABLE_FATIF

 return 1;
}

int usb_block_read(unsigned long VolumeID, unsigned long SectorNo, void *Buffer)
{
	usb_block_readex(VolumeID, SectorNo, Buffer, 0);
}
// ==================================================================== 

int usb_block_write(unsigned long VolumeID, unsigned long SectorNo, void *Buffer)
{
	unsigned long	log_index;
 struct block_device *bdev, *bdev_prev;

 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] usb_block_write: VolumeID=%ld SectorNo=%ld Buffer=%08XH \n",
	VolumeID, SectorNo,Buffer
	));
 USBMS_DBG_WR_CACHE(iprintf(
	"[USBMS WR_CACHE] usb_block_write: VolumeID=%ld SectorNo=%ld Buffer=%08XH \n",
	VolumeID, SectorNo,Buffer
	));
 #ifndef USBMS_DBG_DISABLE_FATIF
	USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_block_write: UsbDiskOpened[VolumeID+1] = %d\n",
		UsbDiskOpened[VolumeID+1]
		));

	log_index = usb_storage_query_inbound_index();
	usb_storage_query_inbound_log(log_index, 0x81, VolumeID, SectorNo);
	#ifdef USB_DBG_MEM_COUNT_ON
	#ifdef USBMS_DBG_INIT_ON
		iprintf("[USBMS INIT] usb_block_write ====== show USB Mem Count .. \n") ;
		showUsbMemC() ;
		showUsbMemU() ;
	#endif
	#endif // USB_DBG_MEM_COUNT_ON

	if(! UsbDiskOpened[VolumeID+1])
	{
		USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_block_write: Error : Usb Disk Not Opened !! \n"));
		return -1;
	}

	bdev = disk_global_bdev[VolumeID+1];
	if(bdev && bdev->bd_inode)
	{
		if (SectorNo > bdev->bd_part->nr_sects)
		{
			iprintf("[USB] Write sector %u is larger than the total number of this HD's sectors(%u).\n",
				SectorNo, bdev->bd_part->nr_sects);
			return 0;
		}

			if (USBDeviceRW++)
				return 3;
			usb_block_copy(disk_global_Sector_buf,Buffer,bdev_hardsect_size(bdev));
			usb_storage_query_inbound_log_status(log_index, 0xC1);
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_block_write() : write_dev_sector() .. \n"));
			if (write_dev_sector(bdev->bd_inode,bdev,SectorNo,1,disk_global_Sector_buf) <= 0)
			{
				USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_block_write(3): Error : Write Sector Failed!!\n"));
				dprintf("[USBMS] Error : Write Sector Failed !! \n");
				USBDeviceRW = 0;
				usb_storage_query_inbound_log_status(log_index, 0xC9);
				return 0 ;
		   	}
		   	USBDeviceRW = 0;
		   	usb_storage_query_inbound_log_status(log_index, 0xD1);
	}
	else
	{
		USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_block_write: Error : Disk Not Init!!\n"));
		dprintf("[USBMS] Sector Write Error : No Device!!\n");
		return -1;
	}

	#ifdef USB_DBG_MEM_COUNT_ON
	#ifdef USBMS_DBG_INIT_ON
		iprintf("[USBMS INIT] usb_block_write(2) ====== show USB Mem Count .. \n") ;
		showUsbMemC() ;
		showUsbMemU() ;
	#endif
	#endif // USB_DBG_MEM_COUNT_ON
 #else
	iprintf("[USBMS DISABLE_FATIF] usb_block_write() : Not Write !! \n") ;
	return 0 ;
 #endif // ndef USBMS_DBG_DISABLE_FATIF

 return 1 ;
}

// ====================================================================
// ==================================================================== 
/*
int usb_cluster_read(unsigned long VolumeID, unsigned long SectorNo,
		unsigned int num_sectors , void *Buffer)
{
 unsigned int remain_sectors = 0 ;
 unsigned int now_sectors = 0 ;
 unsigned long now_sector_num = 0 ;

 //dprintf("usb_cluster_read(): SectorNo = %ld, UsbDiskOpened = %d\n", SectorNo, UsbDiskOpened);
 USBMS_DBG_SECTOR(iprintf("[USBMS SECTOR] usb_cluster_read(): VolumeID=%ld SectorNo=%ld Buffer=%08XH \n",
	VolumeID, SectorNo,Buffer));
 USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_cluster_read(): VolumeID=%ld SectorNo=%ld Buffer=%08XH \n",
	VolumeID, SectorNo,Buffer));

 #ifndef USBMS_DBG_DISABLE_FATIF
	//USBMS_DBG_PLUG(iprintf("[USBMS PLUG] usb_cluster_read(): UsbDiskOpened[VolumeID+1]=%d\n",
	//	UsbDiskOpened[VolumeID+1]));
	USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_cluster_read(): UsbDiskOpened[VolumeID+1]=%d\n",
		UsbDiskOpened[VolumeID+1]));
	
	//dprintf("[USBMS] usb_cluster_read(): VolumeID=%ld SectorNo=%ld Buffer=%08XH \n",
	//	VolumeID, SectorNo,Buffer);
	

	#ifdef USB_DBG_MEM_COUNT_ON
	#ifdef USBMS_DBG_INIT_ON
		iprintf("[USBMS INIT] usb_cluster_read(1) ====== show USB Mem Count .. \n") ;
		showUsbMemC() ;
		showUsbMemU() ;
	#endif
	#endif // USB_DBG_MEM_COUNT_ON
    
	if(! UsbDiskOpened[VolumeID+1]) {
		USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_cluster_read() : OpenUsbDisk() .. \n"));
		if ( !OpenUsbDisk(VolumeID+1) ) {
			USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_cluster_read(1) : Error : Open Usb Disk Failed !! \n"));
			return -1;
		   }
		USBMS_DBG_PLUG(iprintf("[USBMS PLUG] usb_cluster_read() : UsbDiskOpened[VolumeID+1]=%d\n",
			UsbDiskOpened[VolumeID+1]));
		USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_cluster_read() : UsbDiskOpened[VolumeID+1]=%d\n",
			UsbDiskOpened[VolumeID+1]));
		if (! UsbDiskOpened[VolumeID+1]) {
			USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_cluster_read(2) : Error : Open Usb Disk Failed !! \n"));
			return -1;
		   }
	   }

	remain_sectors = num_sectors ;
	now_sector_num = SectorNo ;

	for(;remain_sectors>0;) {

		if(!disk_global_Sector_buf) {
			USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_cluster_read(3) : Error : Null disk_global_Sector_buf !! \n"));
			dprintf("[USBMS] Sector Read Error : No Device !! \n");
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_cluster_read(1) End with Failed !! \n"));
			return -1;
		   }

		if (remain_sectors >= USBMS_RW_MAX_SECTORS) {
			now_sectors = USBMS_RW_MAX_SECTORS ;
		   }
		  else {
			now_sectors = remain_sectors ;
		   }

		if (   disk_global_bdev[VolumeID+1] 
			&& disk_global_bdev[VolumeID+1]->bd_inode
		   ) {
			USBMS_DBG_FLOW(iprintf(
				"[USBMS FLOW] usb_cluster_read() : read_dev_sector() .. \n"
				));
			if (! read_dev_sector(disk_global_bdev[VolumeID+1]->bd_inode,disk_global_bdev[VolumeID+1],
						now_sector_num, now_sectors ,
						disk_global_Sector_buf)) {
				USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_cluster_read(5) : Error : Read Sector Failed !! \n"));
				dprintf("[USBMS] Error : Read Sector Failed !! \n");
				return 0 ;
			   }
		   }
		  else {
			USBMS_DBG_INFO(iprintf(
				"[USBMS INFO] usb_cluster_read(6) : Error : Disk Not Init !! \n"));
			// dprintf("[USBMS] Error : Disk Not Init before read !! \n");
			dprintf("[USBMS] Sector Read Error : No Device !! \n");
			return -1;
		   }
	
			memcpy(Buffer, disk_global_Sector_buf, USBMS_SECTOR_SIZE*now_sectors);

		if (remain_sectors >= USBMS_RW_MAX_SECTORS) {
			remain_sectors -= USBMS_RW_MAX_SECTORS ;
		   }
		  else {
			remain_sectors = 0 ;
		   }
		now_sector_num += USBMS_RW_MAX_SECTORS ;
	   } // for remain_sectors

	#ifdef USB_DBG_MEM_COUNT_ON
	#ifdef USBMS_DBG_INIT_ON
		iprintf("[USBMS INIT] usb_cluster_read(2) ====== show USB Mem Count .. \n"
    		) ;
		showUsbMemC() ;
		showUsbMemU() ;
	#endif
	#endif // USB_DBG_MEM_COUNT_ON

  #else
	iprintf("[USBMS DISABLE_FATIF] usb_cluster_read() : Not Read !! \n") ;
	return 0 ;
 #endif // ndef USBMS_DBG_DISABLE_FATIF
 return 1;
}
*/

// ==================================================================== 
/*
int usb_cluster_write(unsigned long VolumeID, unsigned long SectorNo,
		unsigned int num_sectors , void *Buffer)
{
 unsigned int remain_sectors = 0 ;
 unsigned int now_sectors = 0 ;
 unsigned long now_sector_num = 0 ;

 
 //USBMS_DBG_INFO(iprintf(
 //	"[USBMS INFO] usb_cluster_write(1): VolumeID=%ld SectorNo=%ld Buffer=%08XH \n",
 //	VolumeID, SectorNo,Buffer
 //	));
 
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] usb_cluster_write(1): VolumeID=%ld SectorNo=%ld Buffer=%08XH \n",
	VolumeID, SectorNo,Buffer
	));
 USBMS_DBG_WR_CACHE(iprintf(
	"[USBMS WR_CACHE] usb_cluster_write(1): VolumeID=%ld SectorNo=%ld Buffer=%08XH \n",
	VolumeID, SectorNo,Buffer
	));

#ifdef	USBMS_BLK_CACHE
 blk_sync_cache();
#endif	// defined(USBMS_BLK_CACHE)

 #ifndef USBMS_DBG_DISABLE_FATIF
	//USBMS_DBG_PLUG(iprintf("[USBMS PLUG] usb_cluster_write(): UsbDiskOpened[VolumeID+1] = %d\n",
	//	UsbDiskOpened[VolumeID+1]
	//	));
	USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_cluster_write(2): UsbDiskOpened[VolumeID+1] = %d\n",
		UsbDiskOpened[VolumeID+1]
		));

	#ifdef USB_DBG_MEM_COUNT_ON
	#ifdef USBMS_DBG_INIT_ON
		iprintf("[USBMS INIT] usb_cluster_write(1) ====== show USB Mem Count .. \n") ;
		showUsbMemC() ;
		showUsbMemU() ;
	#endif
	#endif // USB_DBG_MEM_COUNT_ON
	if(! UsbDiskOpened[VolumeID+1]) {
	    
		//USBMS_DBG_INIT(iprintf("[USBMS INIT] usb_cluster_write(3) : OpenUsbDisk() .. \n"));
		//if ( !OpenUsbDisk() ) {
		//	USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_cluster_write(2) : Error : Open Usb Disk Failed !! \n"));
		//	return -1;
		//   }
		//USBMS_DBG_INIT(iprintf("[USBMS INOT] usb_cluster_write() : UsbDiskOpened=%d\n",
		//	UsbDiskOpened));
		//if (! UsbDiskOpened)
    	
		   {
			USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_cluster_write(3) : Error : Usb Disk Not Opened !! \n"));
			return -1;
		   }
	   }

	remain_sectors = num_sectors ;
	now_sector_num = SectorNo ;

	for(;remain_sectors>0;) {

		if(!disk_global_Sector_buf) {
			USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_cluster_write(4) : Error : Null disk_global_Sector_buf !! \n"));
			dprintf("[USBMS] Sector Write Error : No Device !! \n");
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_cluster_write(1) End with Failed !! \n"));
			return -1;
		   }
	
		if (remain_sectors >= USBMS_RW_MAX_SECTORS) {
			now_sectors = USBMS_RW_MAX_SECTORS ;
		   }
		  else {
			now_sectors = remain_sectors ;
		   }

		#ifndef USBMS_BLK_CACHE // 2004/1/15 B. S. Liaw
			memcpy(disk_global_Sector_buf, Buffer, USBMS_SECTOR_SIZE*now_sectors);
		#endif // ndef USBMS_BLK_CACHE
	
		if(disk_global_bdev[VolumeID+1] && disk_global_bdev[VolumeID+1]->bd_inode) {
			USBMS_DBG_FLOW(iprintf(
				"[USBMS FLOW] usb_cluster_write() : write_dev_sector() .. \n"
				));
			if (! write_dev_sector(disk_global_bdev[VolumeID+1]->bd_inode,
						disk_global_bdev[VolumeID+1],
						now_sector_num, now_sectors ,
				#ifndef USBMS_BLK_CACHE // 2004/1/15 B. S. Liaw
						disk_global_Sector_buf)) {
				 #else
						Buffer)) {
				#endif // USBMS_BLK_CACHE
				USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_cluster_write(5) : Error : Write Sector Failed !! \n"));
				dprintf("[USBMS] Error : Write Sector Failed !! \n");
				return 0 ;
			   }
		   }
		  else {
			USBMS_DBG_INFO(iprintf(
				"[USBMS INFO] usb_cluster_write(6) : Error : Disk Not Init !! \n"));
			dprintf("[USBMS] Sector Write Error : No Device !! \n");
			return -1;
		   }

		if (remain_sectors >= USBMS_RW_MAX_SECTORS) {
			remain_sectors -= USBMS_RW_MAX_SECTORS ;
		   }
		  else {
			remain_sectors = 0 ;
		   }
		now_sector_num += USBMS_RW_MAX_SECTORS ;
	   } // for remain_sectors

	#ifdef USB_DBG_MEM_COUNT_ON
	#ifdef USBMS_DBG_INIT_ON
		iprintf("[USBMS INIT] usb_cluster_write(2) ====== show USB Mem Count .. \n") ;
		showUsbMemC() ;
		showUsbMemU() ;
	#endif
	#endif // USB_DBG_MEM_COUNT_ON
  #else
	iprintf("[USBMS DISABLE_FATIF] usb_cluster_write() : Not Write !! \n") ;
	return 0 ;
 #endif // ndef USBMS_DBG_DISABLE_FATIF
 return 1 ;
}
*/

// ====================================================================
// ====================================================================

unsigned char UsbRWTest;
unsigned char UsbRWShowInfo;

void sector_read_verify(void) {
#if 1
 unsigned char backup[USBMS_SECTOR_SIZE];
 struct block_device *bdev;
 unsigned long sector;
 unsigned long testSec[3] = { 1027940, 2964745, 490313 };
 int i;
  
 bdev = disk_global_bdev[1];
 UsbRWTest = 1;
 UsbRWShowInfo = 1;
 i = 0;
 randomize();
 while (UsbRWTest)
 {
 	if (i < 3)
 		sector = testSec[i++];
 	else
 		sector = random(bdev->bd_part->nr_sects-1);
 	if (UsbRWShowInfo)
 		iprintf("Test Sector %u:", sector);
 	if (usb_block_read(0,sector,backup) >= 1)
	{
 		if (usb_block_write(0,sector,backup) == 1)
 		{
 			if (UsbRWShowInfo)
 				iprintf("OK!\n");
 		}
 		else
 			iprintf("Test Sector %u write error!\n", sector);
 	}
 	else
 		iprintf("Test Sector %u read error!\n", sector);
 }
 
 blk_sync_cache();
 
#else
 unsigned long start_test_time = 0L ;
 unsigned long test_times = 0L ;
 unsigned char BackUpBuffer[USBMS_SECTOR_SIZE] ;

 dprintf("[USBMS] sector_read_verify(1): VolumeID=%ld SectorNo=%ld \n",
	0, 63);

 memset(BackUpBuffer, 0, USBMS_SECTOR_SIZE) ;

 #ifdef USB_CODE_TYPE_DEBUG
	iprintf("[USBMS] sector_read_verify(2) : Please turn off USB_CODE_TYPE_DEBUG !! \n");
 #endif // USB_CODE_TYPE_DEBUG
 #ifdef USBMS_BLK_CACHE // 2004/1/15 B. S. Liaw
	iprintf("[USBMS] sector_read_verify(3) : Please turn off USBMS_BLK_CACHE for Real Read !! \n");
 #endif // USBMS_BLK_CACHE

 if(! UsbDiskOpened[0+1]) {
	USBMS_DBG_INIT(iprintf("[USBMS INIT] sector_read_verify() : OpenUsbDisk() .. \n"));
	if ( !OpenUsbDisk(0+1) ) {
		USBMS_DBG_INFO(iprintf("[USBMS INFO] sector_read_verify() : Error : Open Usb Disk Failed !! \n"));
		return ;
	   }
	USBMS_DBG_PLUG(iprintf("[USBMS PLUG] sector_read_verify() : UsbDiskOpened[PartitionID]=%d\n",
		UsbDiskOpened[0+1]));
	USBMS_DBG_INIT(iprintf("[USBMS INIT] sector_read_verify() : UsbDiskOpened[PartitionID]=%d\n",
		UsbDiskOpened[0+1]));
	if (! UsbDiskOpened[0+1]) {
		dprintf("[USBMS] sector_read_verify(3) : Error : Open Usb Disk Failed !! \n");
		return ;
	   }
   }

 if(!disk_global_Sector_buf) {
	dprintf("[USBMS] sector_read_verify(4) : Error : Null disk_global_Sector_buf !! \n");
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sector_read_verify(1) End with Failed !! \n"));
	return ;
   }


 if(disk_global_bdev[0+1] && disk_global_bdev[0+1]->bd_inode) {

	dprintf("[USBMS] sector_read_verify(5) : First Read ..\n") ;
	memset(disk_global_Sector_buf, 0, USBMS_SECTOR_SIZE) ;

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sector_read_verify() : read_dev_sector() .. \n"));
	if (read_dev_sector(disk_global_bdev[0+1]->bd_inode,disk_global_bdev[0+1],
						63, 1 , 
						disk_global_Sector_buf) <= 0) {
		USBMS_DBG_INFO(iprintf("[USBMS INFO] sector_read_verify() : Error : Read Sector Failed !! \n"));
		dprintf("[USBMS] Error : Read Sector Failed !! \n");
		goto test_end ;
	   }

	memcpy(BackUpBuffer, disk_global_Sector_buf, USBMS_SECTOR_SIZE) ;

	BLOCK_PREEMPTION;
 	   {
	    unsigned char *buf_ptr ;

		buf_ptr =	(unsigned char *)(BackUpBuffer) ;
		USBMS_DBG_BUFFER(iprintf(
			"[USBMS BUFFER] sector_read_verify() : Dump Sector 512 Bytes : \n")) ;
			usb_DumpMem((char *) &(buf_ptr[0]), USBMS_SECTOR_SIZE ,USB_DUMP_OFFSET);
	   }
	UNBLOCK_PREEMPTION;

	dprintf("[USBMS] sector_read_verify(6) : Test ..\n") ;
	test_times = 0L ;
	start_test_time = get_sys_time()/CLOCKHZ;
	for(test_times=0;;test_times++) {

		dprintf("[USBMS] sector_read_verify(7) : test %ld .. \n",
				test_times ) ;

		BLOCK_PREEMPTION;
		if(!disk_global_Sector_buf) {
			UNBLOCK_PREEMPTION;
			dprintf("[USBMS] sector_read_verify(8) : Error : Null disk_global_Sector_buf !! \n");
			USBMS_DBG_FLOW(iprintf(
				"[USBMS FLOW] sector_read_speed(2) End with Failed !! \n"
				));
			return ;
		   }

		if(!disk_global_bdev[0+1] || !disk_global_bdev[0+1]->bd_inode) {
			UNBLOCK_PREEMPTION;
			dprintf("[USBMS] sector_read_verify(9) : Error : Null disk_global_bdev !! \n");
			USBMS_DBG_FLOW(iprintf(
				"[USBMS FLOW] sector_read_verify(3) End with Failed !! \n"
				));
			return ;
		   }

		memset(disk_global_Sector_buf, 0, USBMS_SECTOR_SIZE) ;
		UNBLOCK_PREEMPTION;

		USBMS_DBG_FLOW(iprintf(
			"[USBMS FLOW] sector_read_verify() : read_dev_sector() .. \n"
			));
		if (read_dev_sector(disk_global_bdev[0+1]->bd_inode,
							disk_global_bdev[0+1],
							63, 1 ,
							disk_global_Sector_buf) <= 0) {
			USBMS_DBG_INFO(iprintf("[USBMS INFO] sector_read_verify() : Error : Read Sector Failed !! \n"));
			dprintf("[USBMS] Error : Read Sector Failed !! \n");
			break ;
		   }

		BLOCK_PREEMPTION;
		if(!disk_global_Sector_buf) {
			dprintf("[USBMS] sector_read_verify(10) : Error : Null disk_global_Sector_buf !! \n");
			UNBLOCK_PREEMPTION;
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sector_read_speed(2) End with Failed !! \n"));
			return ;
		   }
		if ( memcmp(BackUpBuffer, disk_global_Sector_buf, USBMS_SECTOR_SIZE) != 0 ) {
			// BLOCK_PREEMPTION;
			dprintf("[USBMS] sector_read_verify(11) : Error : Verify Failed !! \n") ;
			// UNBLOCK_PREEMPTION;
			// break ;
		   }
		UNBLOCK_PREEMPTION;
		/*
		if ( (start_test_time != 0) &&
		     ((get_sys_time()/CLOCKHZ)-start_test_time >= 30)) {

			BLOCK_PREEMPTION;
			dprintf("[USBMS] sector_read_verify(12) : speed=%d sector / 30 sec \n" , test_times) ;
			UNBLOCK_PREEMPTION;

			break ;
		   }
		*/
	   } // for test_times

	// -----------
	test_end :
	// -----------

	dprintf("[USBMS] sector_read_verify(13): End test. \n");

   }
  else {
	dprintf("[USBMS] sector_read_verify(14) : Error : Disk Not Init !! \n");
	return ;
   }

 /*
 memcpy(Buffer, disk_global_Sector_buf, 512);

 */
 return ;
#endif
} // function sector_read_verify()


// ====================================================================

// int sector_read_test(unsigned long VolumeID, unsigned long SectorNo, void *Buffer)
void sector_read_speed(void) {
 unsigned long start_test_time = 0L ;
 unsigned long test_times = 0L ;


 /*
 USBMS_DBG_INFO(iprintf("[USBMS INFO] sector_read_test(): VolumeID=%ld SectorNo=%ld Buffer=%08XH \n",
	VolumeID, SectorNo,Buffer));
 */
 /*
 USBMS_DBG_INFO(iprintf("[USBMS INFO] sector_read_test(): VolumeID=%ld SectorNo=%ld \n",
	0, 63));
 */
 dprintf("[USBMS] sector_read_speed(1): VolumeID=%ld SectorNo=%ld \n",
	0, 63);

 #ifdef USB_CODE_TYPE_DEBUG
	iprintf("[USBMS] sector_read_speed(2) : Please turn off USB_CODE_TYPE_DEBUG !! \n");
 #endif // USB_CODE_TYPE_DEBUG
 #ifdef USBMS_BLK_CACHE // 2004/1/15 B. S. Liaw
	iprintf("[USBMS] sector_read_speed(3) : Please turn off USBMS_BLK_CACHE for Real Speed !! \n");
 #endif // USBMS_BLK_CACHE

 /*
 USBMS_DBG_INIT(iprintf("[USBMS INIT] sector_read_speed(): VolumeID=%ld SectorNo=%ld Buffer=%08XH \n",
	VolumeID, SectorNo,Buffer));
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] sector_read_speed(): UsbDiskOpened[VolumeID+1] = %d\n",
	UsbDiskOpened[VolumeID+1]));
 USBMS_DBG_INIT(iprintf("[USBMS INIT] sector_read_speed(): UsbDiskOpened[VolumeID+1] = %d\n",
	UsbDiskOpened[VolumeID+1]));
 
 #ifdef USBMS_DBG_INIT_ON
	iprintf("[USBMS INIT] sector_read_test(1) ====== show USB Mem Count .. \n") ;
	showUsbMemC() ;
	showUsbMemU() ;
 #endif
 */
 if(! UsbDiskOpened[0+1]) {
	USBMS_DBG_INIT(iprintf("[USBMS INIT] sector_read_speed() : OpenUsbDisk() .. \n"));
	if ( !OpenUsbDisk(0+1) ) {
		USBMS_DBG_INFO(iprintf("[USBMS INFO] sector_read_speed() : Error : Open Usb Disk Failed !! \n"));
		return ;
	   }
	USBMS_DBG_PLUG(iprintf("[USBMS PLUG] sector_read_speed() : UsbDiskOpened[PartitionID]=%d\n",
		UsbDiskOpened[0+1]));
	USBMS_DBG_INIT(iprintf("[USBMS INIT] sector_read_speed() : UsbDiskOpened[PartitionID]=%d\n",
		UsbDiskOpened[0+1]));
	if (! UsbDiskOpened[0+1]) {
		dprintf("[USBMS] sector_read_speed(4) : Error : Open Usb Disk Failed !! \n");
		return ;
	   }
   }
 
 if(!disk_global_Sector_buf) {
	dprintf("[USBMS] sector_read_speed(5) : Error : Null disk_global_Sector_buf !! \n");
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sector_read_speed(1) End with Failed !! \n"));
	return ;
   }


 if(disk_global_bdev[0+1] && disk_global_bdev[0+1]->bd_inode) {
	dprintf("[USBMS] sector_read_speed(6) : Test Speed ..\n") ;
	test_times = 0L ;
	start_test_time = get_sys_time()/CLOCKHZ;
	for(test_times=0;;test_times++) {

		if(!disk_global_Sector_buf) {
			dprintf("[USBMS] sector_read_speed(7) : Error : Null disk_global_Sector_buf !! \n");
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sector_read_speed(2) End with Failed !! \n"));
			return ;
		   }

		if(!disk_global_bdev[0+1] || !disk_global_bdev[0+1]->bd_inode) {
			dprintf("[USBMS] sector_read_speed(8) : Error : Null disk_global_bdev !! \n");
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sector_read_speed(3) End with Failed !! \n"));
			return ;
		   }
	

		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sector_read_speed(4) : read_dev_sector() .. \n"));
		if (read_dev_sector(disk_global_bdev[0+1]->bd_inode,disk_global_bdev[0+1],
						63, 1 , 
						disk_global_Sector_buf) <= 0) {
			USBMS_DBG_INFO(iprintf("[USBMS INFO] sector_read_speed() : Error : Read Sector Failed !! \n"));
			dprintf("[USBMS] Error : Read Sector Failed !! \n");
			break ;
		   }

		if ( (start_test_time != 0) &&
		     ((get_sys_time()/CLOCKHZ)-start_test_time >= 30)) {

			BLOCK_PREEMPTION;
			dprintf("[USBMS] sector_read_speed(9) : speed=%d sector / 30 sec \n" , test_times) ;
			UNBLOCK_PREEMPTION;

			break ;
		   }
	   } // for test_times

	dprintf("[USBMS] sector_read_speed(10): End test. \n");

   }
  else {
	dprintf("[USBMS] sector_read_speed(11) : Error : Disk Not Init !! \n");
	return ;
   }

 /*
 memcpy(Buffer, disk_global_Sector_buf, 512);

 #ifdef USBMS_DBG_INIT_ON
	iprintf("[USBMS INIT] sector_read_speed(2) ====== show USB Mem Count .. \n"
    	) ;
	showUsbMemC() ;
	showUsbMemU() ;
 #endif
 */
 return ;
} // function sector_read_speed()


// ==================================================================== 

// int DumpSector(unsigned long VolumeID, unsigned long SectorNo, void *Buffer)
void DumpSector(unsigned long PartitionID,unsigned long SectorNo) {

 dprintf("[USBMS] DumpSector(): PartitionID = %lu, SectorNo = %08XH (%ld) \n",
	PartitionID, SectorNo,SectorNo);

 /*
 USBMS_DBG_INIT(iprintf("[USBMS INIT] DumpSector(): VolumeID=%ld SectorNo=%ld Buffer=%08XH \n",
	VolumeID, SectorNo,Buffer));
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] DumpSector(): UsbDiskOpened[PartitionID] = %d\n",
	UsbDiskOpened[PartitionID]));
 USBMS_DBG_INIT(iprintf("[USBMS INIT] DumpSector(): UsbDiskOpened[PartitionID] = %d\n",
	UsbDiskOpened[PartitionID]));
 
 #ifdef USBMS_DBG_INIT_ON
	iprintf("[USBMS INIT] DumpSector(1) ====== show USB Mem Count .. \n") ;
	showUsbMemC() ;
	showUsbMemU() ;
 #endif
 */
 if(! UsbDiskOpened[PartitionID]) {
	USBMS_DBG_INIT(iprintf("[USBMS INIT] DumpSector() : OpenUsbDisk() .. \n"));
	if ( !OpenUsbDisk(PartitionID) ) {
		USBMS_DBG_INFO(iprintf("[USBMS INFO] DumpSector() : Error : Open Usb Disk Failed !! \n"));
		return ;
	   }
	USBMS_DBG_PLUG(iprintf("[USBMS PLUG] DumpSector() : UsbDiskOpened[PartitionID]=%d\n",
		UsbDiskOpened[PartitionID]));
	USBMS_DBG_INIT(iprintf("[USBMS INIT] DumpSector() : UsbDiskOpened[PartitionID]=%d\n",
		UsbDiskOpened[PartitionID]));
	if (! UsbDiskOpened[PartitionID]) {
		dprintf("[USBMS] DumpSector() : Error : Open Usb Disk Failed !! \n");
		return ;
	   }
   }

 if(!disk_global_Sector_buf) {
	dprintf("[USBMS] DumpSector() : Error : Null disk_global_Sector_buf !! \n");
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] DumpSector(1) End with Failed !! \n"));
	return ;
   }


 if(disk_global_bdev[PartitionID] && disk_global_bdev[PartitionID]->bd_inode) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] DumpSector() : read_dev_sector() .. \n"));
	if (read_dev_sector(disk_global_bdev[PartitionID]->bd_inode,disk_global_bdev[PartitionID],
						SectorNo, 1 ,
						disk_global_Sector_buf) <= 0) {
		USBMS_DBG_INFO(iprintf("[USBMS INFO] DumpSector() : Error : Read Sector Failed !! \n"));
		dprintf("[USBMS] Error : Read Sector Failed !! \n");
		return ;
	   }
   }
  else
   {
	dprintf("[USBMS] DumpSector() : Error : Disk Not Init !! \n");
	return ;
   }

#ifdef USBMS_DBG_BUFFER_ON   // I/O Buffer
 BLOCK_PREEMPTION;
   {
    unsigned char *buf_ptr ;

	buf_ptr =	(unsigned char *)(disk_global_Sector_buf) ;
	USBMS_DBG_BUFFER(iprintf(
		"[USBMS BUFFER] DumpSector() : Dump Sector 512 Bytes : \n")) ;
			usb_DumpMem((char *) &(buf_ptr[0]), USBMS_SECTOR_SIZE , (USB_DUMP_OFFSET | USB_DUMP_ASCII));
   }
 UNBLOCK_PREEMPTION;
#endif // USBMS_DBG_BUFFER_ON

 /*
 #ifdef USBMS_DBG_INIT_ON
	iprintf("[USBMS INIT] DumpSector(2) ====== show USB Mem Count .. \n"
    	) ;
	showUsbMemC() ;
	showUsbMemU() ;
 #endif
 */
 return ;
} // function DumpSector()

// ==================================================================== 
// End of File : FatBlockIo.c

