
// drivers/block/genhd.c

/*
 *  gendisk handling
 */

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsbMsInit.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP

/*
#include <linux/config.h>
#include <linux/module.h>
*/
/*
#include <linux/fs.h>
*/
#include "PoFile.h" // include/linux/file.h
#include "radix-tree.h" // include/linux/radix-tree.h
#include "PoFs.h" // <linux/fs.h>
/*
#include <linux/genhd.h>
*/
#include "block/genhd.h"
/*
#include <linux/kernel.h>
*/
#include "PoKernel.h"
/*
#include <linux/blkdev.h>
*/
#include "bitops.h" // B. S. Liaw add for test_bit()
#include "blkdev.h"
/*
#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/seq_file.h>
#include <linux/slab.h>
#include <linux/kmod.h>
#include <linux/kobj_map.h>
*/
#include "kobj_map.h"

#ifndef USBMS_ONLY_ONE_DISK
	#ifndef USBMS_REDUCE_RESOURCE
		#define MAX_PROBE_HASH 255	/* random */
	 #else
		#define MAX_PROBE_HASH 7
	#endif // USBMS_REDUCE_RESOURCE
#endif // ndef USBMS_ONLY_ONE_DISK


static struct subsystem block_subsys;

/*
 * Can be deleted altogether. Later.
 *
 * Modified under both block_subsys.rwsem and major_names_lock.
 */
static struct blk_major_name {
 #ifndef USBMS_ONLY_ONE_DISK
	struct blk_major_name *next;
	int major;
 #endif // ndef USBMS_ONLY_ONE_DISK
 #ifdef USB_DBGSTRING_ON
	char name[16];
 #endif // USB_DBGSTRING_ON
}
	#ifndef USBMS_ONLY_ONE_DISK
		 *major_names[MAX_PROBE_HASH];
	 #else
		 *major_names = (struct blk_major_name *) NULL ;
	#endif // USBMS_ONLY_ONE_DISK

/*
#ifndef NO_SPINLOCK
	static spinlock_t major_names_lock = SPIN_LOCK_UNLOCKED;
#endif // ndef NO_SPINLOCK
*/
/* index in the above - for now: assume no multimajor ranges */
#ifndef USBMS_ONLY_ONE_DISK
   static inline int major_to_index(int major)
   {
	return major % MAX_PROBE_HASH;
   }
#endif // ndef USBMS_ONLY_ONE_DISK

/* get block device names in somewhat random order */
/*
int get_blkdev_list(char *p)
{
 struct blk_major_name *n;
 int i, len;

 len = sprintf(p, "\nBlock devices:\n");

 #ifndef NO_SEMPHORE
	down_read(&block_subsys.rwsem);
 #endif // ndef NO_SEMPHORE
	for (i = 0; i < ARRAY_SIZE(major_names); i++) {
		for(n = major_names[i]; n; n = n->next) {
		#ifdef USB_DBGSTRING_ON
			len += sprintf(p+len, "%3d %s\n",
				       n->major, n->name);
		 #else
			len += sprintf(p+len, "%3d \n",
				       n->major );
		#endif // ndef USB_DBGSTRING_ON
		   }
	}
 #ifndef NO_SEMPHORE
	up_read(&block_subsys.rwsem);
 #endif // ndef NO_SEMPHORE

 return len;
}
*/
#ifndef USBMS_ONLY_ONE_DISK
	int register_blkdev(unsigned int major, const char *name)
 #else
	int register_blkdev(const char *name)
#endif // USBMS_ONLY_ONE_DISK
{
 #ifndef USBMS_ONLY_ONE_DISK
	struct blk_major_name **n, *p;
	int index, ret = 0;
  #else
	struct blk_major_name *p;
	int ret = 0;
 #endif // USBMS_ONLY_ONE_DISK
 unsigned long flags;

 /*
 #ifndef NO_SEMPHORE
	down_write(&block_subsys.rwsem);
 #endif // ndef NO_SEMPHORE
 */
 /* temporary */
 #ifndef USBMS_ONLY_ONE_DISK
	if (major == 0) {
		for (index = ARRAY_SIZE(major_names)-1; index > 0; index--) {
			if (major_names[index] == NULL)
				break;
		   }

		if (index == 0) {
			printk("register_blkdev: failed to get major for %s\n",
			       name);
			ret = -EBUSY;
			goto out;
		   }
		major = index;
		ret = major;
	   }
 #else
	/*
	if (major_names == NULL) {
		 USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] register_blkdev(1) : Error : failed to get major for %s\n",
			       name)) ;
		ret = -EBUSY;
		goto out;
	   }
	ret = 0 ;
	*/
 #endif // USBMS_ONLY_ONE_DISK

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] register_blkdev() : p = UsbAllocMemC() , size=%d.. \n",
	sizeof(struct blk_major_name)
	));

 /*
 p = kmalloc(sizeof(struct blk_major_name), GFP_KERNEL);
 */
 p = (struct blk_major_name *) UsbAllocMemC(USB_MEM_C_S
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct blk_major_name)
		        #endif
           		) ;

 if(p == NULL) {
	ret = -ENOMEM;
	goto out;
   }

 #ifndef USBMS_ONLY_ONE_DISK
	p->major = major;
	USBMS_DBG_INFO(iprintf("[USBMS INFO] register_blkdev() : p->major=%d \n",
		p->major )) ;
 #endif // ndef USBMS_ONLY_ONE_DISK

 #ifdef USB_DBGSTRING_ON
	strlcpy(p->name, name, sizeof(p->name));
 #endif // USB_DBGSTRING_ON

 #ifndef USBMS_ONLY_ONE_DISK
	p->next = 0;
	index = major_to_index(major);
   #else
	// index = 0 ;
 #endif // ndef USBMS_ONLY_ONE_DISK
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(&major_names_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 #ifndef USBMS_ONLY_ONE_DISK
	for (n = &major_names[index]; *n; n = &(*n)->next) {
		if ((*n)->major == major)
			break;
		}
	if (!*n)
		*n = p;
	else
		ret = -EBUSY;
  #else
	if (!major_names) {
		major_names = p ;
	   }
	  else {
		 USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] register_blkdev(2) : Error : failed to get major for [%s] \n",
			       name)) ;
		ret = -EBUSY;
	   }
 #endif // ndef USBMS_ONLY_ONE_DISK
 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(&major_names_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 if (ret < 0) {
	/*
	printk("register_blkdev: cannot get major %d for %s\n",
	*/
	#ifndef USBMS_ONLY_ONE_DISK
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] register_blkdev(3) : Error : cannot get major %d for [%s] !!\n",
			major, name
			));
	 #else
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] register_blkdev(4) : Error : cannot get for [%s] !!\n",
			name
			));
	#endif // USBMS_ONLY_ONE_DISK
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] register_blkdev() : UsbFreeMemC(p) , size=%d .. \n",
		sizeof(struct blk_major_name)
		));

	/*
	kfree(p);
	*/
	UsbFreeMemC(USB_MEM_C_S,(unsigned char *) p
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(struct blk_major_name)
					#endif
					);
   }

 // ---------
 out:
 // ---------

 /*
 #ifndef NO_SEMPHORE
	up_write(&block_subsys.rwsem);
 #endif // ndef NO_SEMPHORE
 */
 return ret;
}
/*
EXPORT_SYMBOL(register_blkdev);
*/
/* todo: make void - error printk here */
#ifndef USBMS_ONLY_ONE_DISK
	int unregister_blkdev(unsigned int major, const char *name)
 #else
	int unregister_blkdev(const char *name)
#endif // USBMS_ONLY_ONE_DISK
{
 #ifndef USBMS_ONLY_ONE_DISK
	struct blk_major_name **n;
 #endif // ndef USBMS_ONLY_ONE_DISK
 struct blk_major_name *p = NULL;
 #ifndef USBMS_ONLY_ONE_DISK
	int index = major_to_index(major);
  #else
	// int index = 0 ;
 #endif // USBMS_ONLY_ONE_DISK
 unsigned long flags;
 int ret = 0;

 /*
 #ifndef NO_SEMPHORE
	down_write(&block_subsys.rwsem);
 #endif // ndef NO_SEMPHORE
 */
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(&major_names_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 #ifndef USBMS_ONLY_ONE_DISK
	for (n = &major_names[index]; *n; n = &(*n)->next)
		if ((*n)->major == major)
			break;
	if (!*n
		#ifdef USB_DBGSTRING_ON
			|| strcmp((*n)->name, name)
		#endif // USB_DBGSTRING_ON
	   )
		ret = -EINVAL;
	else {
		p = *n;
		*n = p->next;
	}
  #else
	if (! major_names
		#ifdef USB_DBGSTRING_ON
			|| strcmp(major_names->name, name)
		#endif // USB_DBGSTRING_ON
	   ) {
		 USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] unregister_blkdev() : Error : failed to get major for %s\n",
			       name)) ;
		ret = -EINVAL;
	   }
	  else {
		p= major_names ;
	   }
 #endif // ndef USBMS_ONLY_ONE_DISK
 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(&major_names_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 /*
 #ifndef NO_SEMPHORE
	up_write(&block_subsys.rwsem);
 #endif // ndef NO_SEMPHORE
 */
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] unregister_blkdev() : UsbFreeMemC(p) , size=%d.. \n" ,
	sizeof(struct blk_major_name)
	));
 /*
 kfree(p);
 */
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) p
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(struct blk_major_name)
					#endif
					);

 return ret;
}
/*
EXPORT_SYMBOL(unregister_blkdev);
*/
// #ifndef _ETCPIP
	static struct kobj_map *bdev_map;
// #endif // ndef _ETCPIP

/*
 * Register device numbers dev..(dev+range-1)
 * range must be nonzero
 * The hash chain is sorted on range, so that subranges can override.
 */
void blk_register_region(dev_t dev, unsigned long range, struct module *module,
			 struct kobject *(*probe)(dev_t, int *, void *),
			 int (*lock)(dev_t, void *), void *data)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_register_region() Begin .. \n"));
 // #ifndef _ETCPIP
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_register_region() : kobj_map() .. \n"));
	kobj_map(bdev_map, dev, range, module, probe, lock, data);
 // #endif // ndef _ETCPIP
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_register_region() End. \n"));
}
/*
EXPORT_SYMBOL(blk_register_region);
*/
void blk_unregister_region(dev_t dev, unsigned long range)
{
 
 // #ifndef _ETCPIP
	kobj_unmap(bdev_map, dev, range);
 // #endif // ndef _ETCPIP
 
}
/*
EXPORT_SYMBOL(blk_unregister_region);
*/
static struct kobject *exact_match(dev_t dev, int *part, void *data)
{
	struct gendisk *p = data;
	return &p->kobj;
}

static int exact_lock(dev_t dev, void *data)
{
	struct gendisk *p = data;

	if (!get_disk(p))
		return -1;
	return 0;
}

/**
 * add_gendisk - add partitioning information to kernel list
 * @disk: per-device partitioning information
 *
 * This function registers the partitioning information in @disk
 * with the kernel.
 */
int add_disk(struct gendisk *disk)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] add_disk() Begin .. \n"));
 disk->flags |= GENHD_FL_UP;
 #ifndef USBMS_ONLY_ONE_DISK
	USBMS_DBG_INIT(iprintf("[USBMS INIT] add_disk() : MKDEV()=%04XH \n",
		MKDEV(disk->major, disk->first_minor)
		));
  #else
	USBMS_DBG_INIT(iprintf("[USBMS INIT] add_disk() : MKDEV()=%04XH \n",
		MKDEV(0, 0)
		));
 #endif // USBMS_ONLY_ONE_DISK
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] add_disk() : blk_register_region() .. \n"));
 #ifndef USBMS_ONLY_ONE_DISK
	blk_register_region(MKDEV(disk->major, disk->first_minor),
			    disk->minors, NULL, exact_match, exact_lock, disk);
  #else
	blk_register_region(MKDEV(0, 0),
			    disk->minors, NULL, exact_match, exact_lock, disk);
 #endif // USBMS_ONLY_ONE_DISK

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] add_disk() : register_disk() .. \n"));
 // register_disk() on fs/partitions/check.c
 if(register_disk(disk)<=0) {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] register_disk() : I/O Error !! \n"));
	dprintf("[USBMS] Error : check Disk Failed !! \n");

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] add_disk() End with Failed !! \n"));
	return -EIO ;
   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] add_disk() : blk_register_queue() .. \n"));
 blk_register_queue(disk);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] add_disk() End. \n"));
 return 1 ; 
}
/*
EXPORT_SYMBOL(add_disk);
EXPORT_SYMBOL(del_gendisk);	// in partitions/check.c
*/
void unlink_gendisk(struct gendisk *disk)
{
 blk_unregister_queue(disk);
 #ifndef USBMS_ONLY_ONE_DISK
	USBMS_DBG_INIT(iprintf("[USBMS INIT] unlink_gendisk() : MKDEV()=%04XH \n",
		MKDEV(disk->major, disk->first_minor)
		));
	blk_unregister_region(MKDEV(disk->major, disk->first_minor),
			      disk->minors);
  #else
	USBMS_DBG_INIT(iprintf("[USBMS INIT] unlink_gendisk() : MKDEV()=%04XH \n",
		MKDEV(0, 0)
		));
	blk_unregister_region(MKDEV(0, 0), disk->minors);
 #endif // USBMS_ONLY_ONE_DISK
}

#define to_disk(obj) container_of(obj,struct gendisk,kobj)

/**
 * get_gendisk - get partitioning information for a given device
 * @dev: device to get partitioning information for
 *
 * This function gets the structure containing partitioning
 * information for the given device @dev.
 */


// #ifndef _ETCPIP
struct gendisk *get_gendisk(dev_t dev, int *part)
{
 struct kobject *kobj = kobj_lookup(bdev_map, dev, part);


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] get_gendisk() Called .. \n"));

 return  kobj ? to_disk(kobj) : NULL;
}
// #endif // ndef _ETCPIP


#ifdef CONFIG_PROC_FS
/* iterator */
/*
static void *part_start(struct seq_file *part, loff_t *pos)
{
	struct list_head *p;
	loff_t l = *pos;

 #ifndef NO_SEMPHORE
	down_read(&block_subsys.rwsem);
 #endif // ndef NO_SEMPHORE
	list_for_each(p, &block_subsys.kset.list)
		if (!l--)
			return list_entry(p, struct gendisk, kobj.entry);
	return NULL;
}

static void *part_next(struct seq_file *part, void *v, loff_t *pos)
{
	struct list_head *p = ((struct gendisk *)v)->kobj.entry.next;
	++*pos;
	return p==&block_subsys.kset.list ? NULL : 
		list_entry(p, struct gendisk, kobj.entry);
}

static void part_stop(struct seq_file *part, void *v)
{
 #ifndef NO_SEMPHORE
	up_read(&block_subsys.rwsem);
 #endif // ndef NO_SEMPHORE
}

static int show_partition(struct seq_file *part, void *v)
{
	struct gendisk *sgp = v;
	int n;
	char buf[BDEVNAME_SIZE];

	if (&sgp->kobj.entry == block_subsys.kset.list.next)
		seq_puts(part, "major minor  #blocks  name\n\n");

	// Don't show non-partitionable devices or empty devices 
	if (!get_capacity(sgp) || sgp->minors == 1)
		return 0;

	// show the full disk and all non-0 size partitions of it 
	
	// seq_printf(part, "%4d  %4d %10llu %s\n",
	
	dprintf("[USBMS] show_partition() : %4d  %4d %10llu %s\n",
		sgp->major, sgp->first_minor,
		(unsigned long long)get_capacity(sgp) >> 1,
		disk_name(sgp, 0, buf));
	for (n = 0; n < sgp->minors - 1; n++) {
		if (!sgp->part[n])
			continue;
		if (sgp->part[n]->nr_sects == 0)
			continue;
		
		// seq_printf(part, "%4d  %4d %10llu %s\n",
		
		dprintf("[USBMS] show_partition() : %4d  %4d %10llu %s\n",
			sgp->major, n + 1 + sgp->first_minor,
			(unsigned long long)sgp->part[n]->nr_sects >> 1 ,
			disk_name(sgp, n + 1, buf));
	}

	return 0;
}
*/
/*
struct seq_operations partitions_op = {
	.start =part_start,
	.next =	part_next,
	.stop =	part_stop,
	.show =	show_partition
};
*/
#endif


extern int blk_dev_init(void);

static struct kobject *base_probe(dev_t dev, int *part, void *data)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] base_probe() Begin .. \n"));
 /*
 request_module("block-major-%d", MAJOR(dev));
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] base_probe() End. \n"));
 return NULL;
}
/*
int __init device_init(void)
*/
int device_init(void)
{
 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] device_init() Begin .. \n"
	));

 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] device_init() call kobj_map_init() .. \n"
	));
 bdev_map = kobj_map_init(base_probe, &block_subsys);
 
 blk_dev_init();

 #ifndef _ETCPIP
	subsystem_register(&block_subsys);
 #endif // ndef _ETCPIP

 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] device_init() End. \n"
	));
 return 0;
}
/*
#ifndef _ETCPIP
subsys_initcall(device_init);
#endif // ndef _ETCPIP
*/


/*
 * kobject & sysfs bindings for block devices
 */
/*
#ifndef NO_SYS_FS
// struct attribute on sysfs.h
struct disk_attribute {
	struct attribute attr;
	ssize_t (*show)(struct gendisk *, char *);
};

static ssize_t disk_attr_show(struct kobject *kobj, struct attribute *attr,
			      char *page)
{
	struct gendisk *disk = to_disk(kobj);
	struct disk_attribute *disk_attr =
		container_of(attr,struct disk_attribute,attr);
	ssize_t ret = 0;

	if (disk_attr->show)
		ret = disk_attr->show(disk,page);
	return ret;
}

static struct sysfs_ops disk_sysfs_ops = {
	.show	= &disk_attr_show,
};
#endif // ndef NO_SYS_FS
*/
/*
static ssize_t disk_dev_read(struct gendisk * disk, char *page)
{
 dev_t base = MKDEV(disk->major, disk->first_minor); 

 USBMS_DBG_INFO(iprintf("[USBMS INFO] disk_dev_read() : MKDEV()=%04XH \n",
	MKDEV(disk->major, disk->first_minor)
	));
 USBMS_DBG_INFO(iprintf("[USBMS INFO] disk_dev_read() : base=%08XH \n", base));
 
 return print_dev_t(page, base);
}
*/
/*
static ssize_t disk_range_read(struct gendisk * disk, char *page)
{
	return sprintf(page, "%d\n", disk->minors);
}

static ssize_t disk_size_read(struct gendisk * disk, char *page)
{
	return sprintf(page, "%llu\n", (unsigned long long)get_capacity(disk));
}
*/
/*
#ifndef NO_TIMER
static inline unsigned jiffies_to_msec(unsigned jif)
{
#if 1000 % HZ == 0
	return jif * (1000 / HZ);
#elif HZ % 1000 == 0
	return jif / (HZ / 1000);
#else
	return (jif / HZ) * 1000 + (jif % HZ) * 1000 / HZ;
#endif
}
#endif // ndef NO_TIMER
*/
/*
static ssize_t disk_stats_read(struct gendisk * disk, char *page)
{
	disk_round_stats(disk);
 #ifndef NO_TIMER
	return sprintf(page,
		"%8u %8u %8llu %8u "
		"%8u %8u %8llu %8u "
		"%8u %8u %8u"
		"\n",
		disk_stat_read(disk, reads), disk_stat_read(disk, read_merges),
		(unsigned long long)disk_stat_read(disk, read_sectors),
		jiffies_to_msec(disk_stat_read(disk, read_ticks)),
		disk_stat_read(disk, writes), 
		disk_stat_read(disk, write_merges),
		(unsigned long long)disk_stat_read(disk, write_sectors),
		jiffies_to_msec(disk_stat_read(disk, write_ticks)),
		disk->in_flight,
		jiffies_to_msec(disk_stat_read(disk, io_ticks)),
		jiffies_to_msec(disk_stat_read(disk, time_in_queue)));
	#endif // ndef NO_TIMER
}
*/
/*
#ifndef NO_SYS_FS
// struct attribute on sysfs.h
static struct disk_attribute disk_attr_dev = {
	.attr = {.name = "dev", .mode = S_IRUGO },
	.show	= disk_dev_read
};
static struct disk_attribute disk_attr_range = {
	.attr = {.name = "range", .mode = S_IRUGO },
	.show	= disk_range_read
};
static struct disk_attribute disk_attr_size = {
	.attr = {.name = "size", .mode = S_IRUGO },
	.show	= disk_size_read
};
static struct disk_attribute disk_attr_stat = {
	.attr = {.name = "stat", .mode = S_IRUGO },
	.show	= disk_stats_read
};
#endif // ndef NO_SYS_FS
*/
/*
#ifndef NO_SYS_FS
// struct attribute on sysfs.h
static struct attribute * default_attrs[] = {
	&disk_attr_dev.attr,
	&disk_attr_range.attr,
	&disk_attr_size.attr,
	&disk_attr_stat.attr,
	NULL,
};
#endif // ndef NO_SYS_FS
*/
/*
#ifndef _ETCPIP
static void disk_release(struct kobject * kobj)
#else
*/
void disk_release(struct gendisk *disk)
/*
#endif
*/
{
 /*
 #ifndef _ETCPIP
	struct gendisk *disk = to_disk(kobj);
 #endif
 */
		
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] disk_release() Begin .. \n")) ;
 // drivers/char/random.c 
 /*
 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] disk_release() : UsbFreeMemC(disk->random) .. \n"
		));
 */
 /*
 kfree(disk->random);
 */
 /*
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) disk->random
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(struct timer_rand_state)
					#endif
					);
 */

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] disk_release() : UsbFreeMemC(disk->part) .. \n"
		));
 /*
 kfree(disk->part);
 */
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) disk->part
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(struct hd_struct *)
					#endif
					);
 free_disk_stats(disk);

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] disk_release() : UsbFreeMemC(disk) .. \n"
		));
 /*
 kfree(disk);
 */
 UsbFreeMemC(USB_MEM_C_M,(unsigned char *) disk
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(struct gendisk) // size : 172
					#endif
					);
}
/*
static struct kobj_type ktype_block = {
	.release	= disk_release,
#ifndef NO_SYS_FS
	.sysfs_ops	= &disk_sysfs_ops,
	.default_attrs	= default_attrs,
#endif // ndef NO_SYS_FS
};
*/
extern struct kobj_type ktype_part;

static int block_hotplug_filter(struct kset *kset, struct kobject *kobj)
{
 /*
 #ifndef NO_SYS_FS
	struct kobj_type *ktype = get_ktype(kobj);

	return ((ktype == &ktype_block) || (ktype == &ktype_part));
 #endif // ndef NO_SYS_FS
 */
}


static struct kset_hotplug_ops block_hotplug_ops = {
	.filter	= block_hotplug_filter,
};


/* declare block_subsys. */
/*
static decl_subsys(block, &ktype_block, &block_hotplug_ops);
*/
/*
static decl_subsys(block,&ktype_block);
*/
/*
 * aggregate disk stat collector.  Uses the same stats that the sysfs
 * entries do, above, but makes them available through one seq_file.
 * Watching a few disks may be efficient through sysfs, but watching
 * all of them will be more efficient through this interface.
 *
 * The output looks suspiciously like /proc/partitions with a bunch of
 * extra fields.
 */

/* iterator */
/*
static void *diskstats_start(struct seq_file *part, loff_t *pos)
{
	loff_t k = *pos;
	struct list_head *p;

 #ifndef NO_SEMPHORE
	down_read(&block_subsys.rwsem);
 #endif // ndef NO_SEMPHORE
	list_for_each(p, &block_subsys.kset.list)
		if (!k--)
			return list_entry(p, struct gendisk, kobj.entry);
	return NULL;
}
*/
/*
static void *diskstats_next(struct seq_file *part, void *v, loff_t *pos)
{
	struct list_head *p = ((struct gendisk *)v)->kobj.entry.next;
	++*pos;
	return p==&block_subsys.kset.list ? NULL :
		list_entry(p, struct gendisk, kobj.entry);
}
*/
/*
static void diskstats_stop(struct seq_file *part, void *v)
{
 #ifndef NO_SEMPHORE
	up_read(&block_subsys.rwsem);
 #endif // ndef NO_SEMPHORE
}
*/
/*
static int diskstats_show(struct seq_file *s, void *v)
{
 struct gendisk *gp = v;
 char buf[BDEVNAME_SIZE];
 int n = 0;

 
 //if (&sgp->kobj.entry == block_subsys.kset.list.next)
 //		seq_puts(s,	"major minor name"
 //				"     rio rmerge rsect ruse wio wmerge "
 // 				"wsect wuse running use aveq"
 //				"\n\n");
 
 
 disk_round_stats(gp);
 #ifndef NO_TIMER
	seq_printf(s, "%4d %4d %s %u %u %llu %u %u %u %llu %u %u %u %u\n",
		gp->major, n + gp->first_minor, disk_name(gp, n, buf),
		disk_stat_read(gp, reads), disk_stat_read(gp, read_merges),
		(unsigned long long)disk_stat_read(gp, read_sectors),
		jiffies_to_msec(disk_stat_read(gp, read_ticks)),
		disk_stat_read(gp, writes), disk_stat_read(gp, write_merges),
		(unsigned long long)disk_stat_read(gp, write_sectors),
		jiffies_to_msec(disk_stat_read(gp, write_ticks)),
		gp->in_flight,
		jiffies_to_msec(disk_stat_read(gp, io_ticks)),
		jiffies_to_msec(disk_stat_read(gp, time_in_queue)));
  #else
	dprintf("[USBMS] diskstats_show() : %4d %4d %s %u %u %llu %u %u %llu %u \n",
		gp->major, n + gp->first_minor, disk_name(gp, n, buf),
		disk_stat_read(gp, reads), disk_stat_read(gp, read_merges),
		(unsigned long long)disk_stat_read(gp, read_sectors),
		disk_stat_read(gp, writes), disk_stat_read(gp, write_merges),
		(unsigned long long)disk_stat_read(gp, write_sectors),
		gp->in_flight
		);
 #endif // ndef NO_TIMER

	// now show all non-0 size partitions of it 
	for (n = 0; n < gp->minors - 1; n++) {
		struct hd_struct *hd = gp->part[n];

		if (hd && hd->nr_sects)
			
			// seq_printf(s, "%4d %4d %s %u %u %u %u\n",
			
			dprintf("%4d %4d %s %u %u %u %u\n",
				gp->major, n + gp->first_minor + 1,
				disk_name(gp, n + 1, buf),
				hd->reads, hd->read_sectors,
				hd->writes, hd->write_sectors);
	}
 
	return 0;
}
*/
/*
struct seq_operations diskstats_op = {
	.start	= diskstats_start,
	.next	= diskstats_next,
	.stop	= diskstats_stop,
	.show	= diskstats_show
};
*/

struct gendisk *alloc_disk(int minors)
{
 /*
 struct gendisk *disk = kmalloc(sizeof(struct gendisk), GFP_KERNEL);
 */
 struct gendisk *disk;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] alloc_disk() Begin .. \n")) ;

 USBMS_DBG_INIT(iprintf("[USBMS INIT] alloc_disk() : minors=%d .. \n" ,
	minors
	));

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] alloc_disk() : disk = UsbAllocMemC() .. \n"
	));

 disk = (struct gendisk *) UsbAllocMemC(USB_MEM_C_M
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct gendisk) // size : 172
		        #endif
           		) ;

 if(disk) {
	/*
	memset(disk, 0, sizeof(struct gendisk));
	*/
	if (!init_disk_stats(disk)) {
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] alloc_disk() : UsbFreeMemC(disk) .. \n"
			));
		/*
		kfree(disk);
		*/
		UsbFreeMemC(USB_MEM_C_M,(unsigned char *) disk
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(struct gendisk) // size : 172
					#endif
					);
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] alloc_disk() End with Fail !! \n")) ;
		return NULL;
	   }
	if (minors > 1) {
		int size = (minors - 1) * sizeof(struct hd_struct *);

		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] alloc_disk() : disk->part = UsbAllocMemC() , size=%d .. \n" ,
			size
			));
		/*
		disk->part = kmalloc(size, GFP_KERNEL);
		*/
		disk->part = (struct hd_struct **) UsbAllocMemC(USB_MEM_C_S
				#ifdef USB_DBG_MEM_C_ON
					, size
		        #endif
           		) ;

		if (!disk->part) {
			USBMS_DBG_MEM(iprintf(
				"[USBMS MEM] alloc_disk() : UsbFreeMemC(disk) .. \n"
				));
			/*
			kfree(disk);
			*/
			UsbFreeMemC(USB_MEM_C_M,(unsigned char *) disk
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(struct gendisk) // size : 172
					#endif
					);

			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] alloc_disk() End with Fail !! \n")) ;
			return NULL;
		   }
		memset(disk->part, 0, size);
	   }
	disk->minors = minors;
	kobj_set_kset_s(disk,block_subsys);
	kobject_init(&disk->kobj);
	/*
	// drivers/char/random.c
	rand_initialize_disk(disk);
	*/
   }
  else
   {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] alloc_disk() : Error : Out of Memory for disk !! \n")) ;
   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] alloc_disk() End. \n")) ;
 return disk;
}
/*
EXPORT_SYMBOL(alloc_disk);
*/
struct kobject *get_disk(struct gendisk *disk)
{
	/*
	#ifndef _ETCPIP
	struct module *owner;
	#endif // ndef _ETCPIP
	*/
	struct kobject *kobj;

	if (!disk->fops)
		return NULL;
	/*
	#ifndef _ETCPIP
	owner = disk->fops->owner;
	
	if (owner && !try_module_get(owner))
		return NULL;
	#endif // ndef _ETCPIP
	*/
	kobj = kobject_get(&disk->kobj);
	if (kobj == NULL) {
	/*
	#ifndef _ETCPIP
		module_put(owner);
	#endif // ndef _ETCPIP
	*/
		return NULL;
	}
	return kobj;

}
/*
EXPORT_SYMBOL(get_disk);
*/
void put_disk(struct gendisk *disk)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] put_disk() Begin .. \n"));
 // #ifndef _ETCPIP
	if (disk)
		kobject_put(&disk->kobj);
 // #endif // ndef _ETCPIP
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] put_disk() End. \n"));
}
/*
EXPORT_SYMBOL(put_disk);
*/
/*
void set_device_ro(struct block_device *bdev, int flag)
{
	if (bdev->bd_contains != bdev)
		bdev->bd_part->policy = flag;
	else
		bdev->bd_disk->policy = flag;
}
*/
/*
EXPORT_SYMBOL(set_device_ro);
*/
/*
void set_disk_ro(struct gendisk *disk, int flag)
{
	int i;
	disk->policy = flag;
	for (i = 0; i < disk->minors - 1; i++)
		if (disk->part[i]) disk->part[i]->policy = flag;
}
*/
/*
EXPORT_SYMBOL(set_disk_ro);
*/
/*
int bdev_read_only(struct block_device *bdev)
{
	if (!bdev)
		return 0;
	else if (bdev->bd_contains != bdev)
		return bdev->bd_part->policy;
	else
		return bdev->bd_disk->policy;
}
*/
/*
EXPORT_SYMBOL(bdev_read_only);
*/
int invalidate_partition(struct gendisk *disk, int index)
{
 int res = 0;
 /*
 struct block_device *bdev = bdget_disk(disk, index);
 */
 struct block_device *bdev ;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] invalidate_partition() Begin .. \n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] invalidate_partition() : bdget_disk() .. \n"));
 bdev = bdget_disk(disk, index);
 USBMS_DBG_MEM(iprintf("[USBMS MEM] invalidate_partition() : bdev=%08XH \n",bdev));
 #ifndef NO_INODE
	// __invalidate_device() is on fs/inode.c
	if (bdev)
       {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] invalidate_partition() : __invalidate_device() .. \n"));
		res = __invalidate_device(bdev, 1);
       }
	// bdput() is on fs/block_dev.c
	bdput(bdev);
 #endif // ndef NO_INODE
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] invalidate_partition() End. \n"));
 return res;
}
/*
EXPORT_SYMBOL(invalidate_partition);
*/

// B. S. Liaw TEST for compiler warning !!

