
/*
 *  linux/drivers/block/ll_rw_blk.c
 *
 * Copyright (C) 1991, 1992 Linus Torvalds
 * Copyright (C) 1994,      Karl Keyte: Added support for disk statistics
 * Elevator latency, (C) 2000  Andrea Arcangeli <andrea@suse.de> SuSE
 * Queue request tables / lock, selectable elevator, Jens Axboe <axboe@suse.de>
 * kernel-doc documentation started by NeilBrown <neilb@cse.unsw.edu.au> -  July2000
 * bio rewrite, highmem i/o, etc, Jens Axboe <axboe@suse.de> - may 2001
 */

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
	// for deciding whether to compile
	//  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP

/*
 * This handles all read/write requests to block devices
 */

/*
#include <linux/config.h>
*/
/*
#include <linux/kernel.h>
*/
#include "PoKernel.h"
/*
#include <linux/module.h>
*/
/*
#include <linux/backing-dev.h>
*/
#include "backing-dev.h"
/*
#include <linux/bio.h>
*/
#include "bio.h"
/*
#include <linux/blkdev.h>
*/
#include "bitops.h" // B. S. Liaw add for test_bit()
#include "blkdev.h"
/*
#include <linux/highmem.h>
#include <linux/mm.h>
#include <linux/kernel_stat.h>
*/
/*
#include <linux/string.h>
*/
#include "PoString.h" // linux/string.h
/*
#include <linux/init.h>
#include <linux/bootmem.h>	// for max_pfn/max_low_pfn 
#include <linux/completion.h>
#include <linux/slab.h>
#include <linux/swap.h>
*/

static void blk_unplug_work(void *data);
static void blk_unplug_timeout(unsigned long data);

/*
 * For the allocated request tables
 */
/* kmem_cache_t is on include/linux/slab.h
static kmem_cache_t *request_cachep;
*/

/*
 * plug management
 */
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	static LIST_HEAD(blk_plug_list);
#endif // ndef NO_ELEVATOR
*/
/*
#ifndef NO_SPINLOCK
	static spinlock_t blk_plug_lock __cacheline_aligned_in_smp = SPIN_LOCK_UNLOCKED;
#endif // ndef NO_SPINLOCK
*/
#ifndef NO_WAITQUEUE
	static wait_queue_head_t congestion_wqh[2];
#endif // ndef NO_WAITQUEUE

/*
 * Controlling structure to kblockd
 */
static struct workqueue_struct *kblockd_workqueue; 

unsigned long blk_max_low_pfn, blk_max_pfn;

/*
EXPORT_SYMBOL(blk_max_low_pfn);
EXPORT_SYMBOL(blk_max_pfn);
*/

/* Amount of time in which a process may batch requests */
#define BLK_BATCH_TIME	(HZ/50UL)

/* Number of requests a "batching" process may submit */
#define BLK_BATCH_REQ	32


// #ifdef _ETCPIP
// B. S. Liaw Add these ..
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	static LIST_HEAD(tags_busy_list);
	static LIST_HEAD(rq_queuelist);
 #else
*/
	request_queue_t global_request_queue ;
	struct request  *global_rq=NULL ;
/*
#endif // NO_ELEVATOR
*/
// #endif // _ETCPIP




/*
 * Return the threshold (number of used requests) at which the queue is
 * considered to be congested.  It include a little hysteresis to keep the
 * context switch rate down.
 */
static inline int queue_congestion_on_threshold(struct request_queue *q)
{
	int ret;

	ret = q->nr_requests - (q->nr_requests / 8) + 1;

	if (ret > q->nr_requests)
		ret = q->nr_requests;

	return ret;
}

/*
 * The threshold at which a queue is considered to be uncongested
 */
static inline int queue_congestion_off_threshold(struct request_queue *q)
{
 int ret;

 ret = q->nr_requests - (q->nr_requests / 8) - 1;

 if (ret < 1)
		ret = 1;

 return ret;
}

/*
 * A queue has just exitted congestion.  Note this in the global counter of
 * congested queues, and wake up anyone who was waiting for requests to be
 * put back.
 */
static void clear_queue_congested(request_queue_t *q, int rw)
{
 enum bdi_state bit;
 #ifndef NO_WAITQUEUE
	wait_queue_head_t *wqh = &congestion_wqh[rw];
 #endif // ndef NO_WAITQUEUE


 bit = (rw == WRITE) ? BDI_write_congested : BDI_read_congested;
 #ifndef NO_BACKDEV 
	clear_bit(bit, &q->backing_dev_info.state);
 #endif // ndef NO_BACKDEV 
 #ifndef NO_WAITQUEUE
	if (waitqueue_active(wqh))
		wake_up(wqh);
 #endif // ndef NO_WAITQUEUE
}

/*
 * A queue has just entered congestion.  Flag that in the queue's VM-visible
 * state flags and increment the global gounter of congested queues.
 */
static void set_queue_congested(request_queue_t *q, int rw)
{
	enum bdi_state bit;

	bit = (rw == WRITE) ? BDI_write_congested : BDI_read_congested;
 #ifndef NO_BACKDEV 
	set_bit(bit, &q->backing_dev_info.state);
 #endif // ndef NO_BACKDEV 
}

/**
 * blk_get_backing_dev_info - get the address of a queue's backing_dev_info
 * @bdev:	device
 *
 * Locates the passed device's request queue and returns the address of its
 * backing_dev_info
 *
 * Will return NULL if the request queue cannot be located.
 */
/*
#ifndef NO_BACKDEV 
struct backing_dev_info *blk_get_backing_dev_info(struct block_device *bdev)
{
	struct backing_dev_info *ret = NULL;
	request_queue_t *q = bdev_get_queue(bdev);

	if (q)
		ret = &q->backing_dev_info;
	return ret;
}
#endif // ndef NO_BACKDEV 
*/
void blk_queue_activity_fn(request_queue_t *q, activity_fn *fn, void *data)
{
	q->activity_fn = fn;
	q->activity_data = data;
}

/**
 * blk_queue_prep_rq - set a prepare_request function for queue
 * @q:		queue
 * @pfn:	prepare_request function
 *
 * It's possible for a queue to register a prepare_request callback which
 * is invoked before the request is handed to the request_fn. The goal of
 * the function is to prepare a request for I/O, it can be used to build a
 * cdb from the request data for instance.
 *
 */
void blk_queue_prep_rq(request_queue_t *q, prep_rq_fn *pfn)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_prep_rq() Begin .. \n"));
 q->prep_rq_fn = pfn;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_prep_rq() End. \n"));
}
/*
EXPORT_SYMBOL(blk_queue_prep_rq);
*/
/**
 * blk_queue_merge_bvec - set a merge_bvec function for queue
 * @q:		queue
 * @mbfn:	merge_bvec_fn
 *
 * Usually queues have static limitations on the max sectors or segments that
 * we can put in a request. Stacking drivers may have some settings that
 * are dynamic, and thus we have to query the queue whether it is ok to
 * add a new bio_vec to a bio at a given offset or not. If the block device
 * has such limitations, it needs to register a merge_bvec_fn to control
 * the size of bio's sent to it. Per default now merge_bvec_fn is defined for
 * a queue, and only the fixed limits are honored.
 *
 */
#ifndef USB_NO_VMM // vitrtul memory management
void blk_queue_merge_bvec(request_queue_t *q, merge_bvec_fn *mbfn)
{
	q->merge_bvec_fn = mbfn;
}
#endif // ndef USB_NO_VMM
/*
EXPORT_SYMBOL(blk_queue_merge_bvec);
*/
/**
 * blk_queue_make_request - define an alternate make_request function for a device
 * @q:  the request queue for the device to be affected
 * @mfn: the alternate make_request function
 *
 * Description:
 *    The normal way for &struct bios to be passed to a device
 *    driver is for them to be collected into requests on a request
 *    queue, and then to allow the device driver to select requests
 *    off that queue when it is ready.  This works well for many block
 *    devices. However some block devices (typically virtual devices
 *    such as md or lvm) do not benefit from the processing on the
 *    request queue, and are served best by having the requests passed
 *    directly to them.  This can be achieved by providing a function
 *    to blk_queue_make_request().
 *
 * Caveat:
 *    The driver that does this *must* be able to deal appropriately
 *    with buffers in "highmemory". This can be accomplished by either calling
 *    __bio_kmap_atomic() to get a temporary kernel mapping, or by calling
 *    blk_queue_bounce() to create a buffer in normal memory.
 **/
// #ifndef USB_NO_VMM // vitrtul memory management
void blk_queue_make_request(request_queue_t * q, make_request_fn * mfn)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_make_request() Begin .. \n")) ;
 /*
  * set defaults
  */
 q->nr_requests = BLKDEV_MAX_RQ;
 q->max_phys_segments = MAX_PHYS_SEGMENTS;
 q->max_hw_segments = MAX_HW_SEGMENTS;
 USBMS_DBG_BLOCK(iprintf("[USBMS BLOCK] blk_queue_make_request() : make_request_fn = mfn .. \n")) ;
 q->make_request_fn = mfn;
 /*
 #ifndef NO_BACKDEV 
	q->backing_dev_info.ra_pages = (VM_MAX_READAHEAD * 1024) / PAGE_CACHE_SIZE;
	q->backing_dev_info.state = 0;
	q->backing_dev_info.memory_backed = 0;
 #endif // ndef NO_BACKDEV 
 */
 // #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		USBMS_DBG_INFO(iprintf("[USBMS INFO] blk_queue_make_request() : Bug on q->req !! \n"));
		// q->req = req ;
 // #endif // NO_ELEVATOR
 blk_queue_max_sectors(q, MAX_SECTORS);
 blk_queue_hardsect_size(q, 512);
 blk_queue_dma_alignment(q, 511);

 q->unplug_thresh = 4;		/* hmm */
 /*
 #ifndef NO_TIMER
	q->unplug_delay = (3 * HZ) / 1000;	// 3 milliseconds 
	if (q->unplug_delay == 0)
		q->unplug_delay = 1;
 #endif // ndef NO_TIMER
 */
 /*
 #ifndef NO_WORKQUEUE
	INIT_WORK(&q->unplug_work, blk_unplug_work, q);
 #endif // ndef NO_WORKQUEUE
 */
 /*
 #ifndef NO_TIMER
	q->unplug_timer.function = blk_unplug_timeout;
	q->unplug_timer.data = (unsigned long)q;
 #endif // ndef NO_TIMER
 */
 /*
  * by default assume old behaviour and bounce for any highmem page
  */
 /*
 #ifdef CONFIG_HIGHMEM
	blk_queue_bounce_limit(q, BLK_BOUNCE_HIGH);
 #endif // CONFIG_HIGHMEM
 */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_BLOCK(iprintf(
		"[USBMS BLOCK] blk_queue_make_request() : INIT_LIST_HEAD() .. \n"
		)) ;
	INIT_LIST_HEAD(&q->plug_list);
 #endif // ndef NO_ELEVATOR
 */
 blk_queue_activity_fn(q, NULL, NULL);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_make_request() End. \n")) ;
}
// #endif // ndef USB_NO_VMM
/*
EXPORT_SYMBOL(blk_queue_make_request);
*/

/**
 * blk_queue_bounce_limit - set bounce buffer limit for queue
 * @q:  the request queue for the device
 * @dma_addr:   bus address limit
 *
 * Description:
 *    Different hardware can have different requirements as to what pages
 *    it can do I/O directly to. A low level driver can call
 *    blk_queue_bounce_limit to have lower memory pages allocated as bounce
 *    buffers for doing I/O to pages residing above @page. By default
 *    the block layer sets this to the highest numbered "low" memory page.
 **/
#ifdef CONFIG_HIGHMEM
void blk_queue_bounce_limit(request_queue_t *q, u64 dma_addr)
{
	unsigned long bounce_pfn = dma_addr >> PAGE_SHIFT;
	unsigned long mb = dma_addr >> 20;
	static request_queue_t *last_q;

	/*
	 * set appropriate bounce gfp mask -- unfortunately we don't have a
	 * full 4GB zone, so we have to resort to low memory for any bounces.
	 * ISA has its own < 16MB zone.
	 */
	if (bounce_pfn < blk_max_low_pfn) {
		/*
		BUG_ON(dma_addr < BLK_BOUNCE_ISA);
		*/
		if (dma_addr < BLK_BOUNCE_ISA) {
			dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
		   }
		init_emergency_isa_pool();
		q->bounce_gfp = GFP_NOIO | GFP_DMA;
	} else
		q->bounce_gfp = GFP_NOIO;

	/*
	 * keep this for debugging for now...
	 */
	if (dma_addr != BLK_BOUNCE_HIGH && q != last_q) {
		printk("blk: queue %p, ", q);
		if (dma_addr == BLK_BOUNCE_ANY)
			printk("no I/O memory limit\n");
		else
			printk("I/O limit %luMb (mask 0x%Lx)\n", mb, (long long) dma_addr);
	}

	q->bounce_pfn = bounce_pfn;
	last_q = q;
}
#endif // CONFIG_HIGHMEM
/*
EXPORT_SYMBOL(blk_queue_bounce_limit);
*/

/**
 * blk_queue_max_sectors - set max sectors for a request for this queue
 * @q:  the request queue for the device
 * @max_sectors:  max sectors in the usual 512b unit
 *
 * Description:
 *    Enables a low level driver to set an upper limit on the size of
 *    received requests.
 **/
void blk_queue_max_sectors(request_queue_t *q, unsigned short max_sectors)
{
 #ifndef USB_NO_VMM // no vitrtul memory management
	if ((max_sectors << 9) < PAGE_CACHE_SIZE) {
		max_sectors = 1 << (PAGE_CACHE_SHIFT - 9);
		printk("%s: set to minimum %d\n", __FUNCTION__, max_sectors);
	}
 #endif // ndef USB_NO_VMM

 q->max_sectors = max_sectors;
}
/*
EXPORT_SYMBOL(blk_queue_max_sectors);
*/

/**
 * blk_queue_max_phys_segments - set max phys segments for a request for this queue
 * @q:  the request queue for the device
 * @max_segments:  max number of segments
 *
 * Description:
 *    Enables a low level driver to set an upper limit on the number of
 *    physical data segments in a request.  This would be the largest sized
 *    scatter list the driver could handle.
 **/
void blk_queue_max_phys_segments(request_queue_t *q, unsigned short max_segments)
{
	if (!max_segments) {
		max_segments = 1;
		printk("%s: set to minimum %d\n", __FUNCTION__, max_segments);
	}

	q->max_phys_segments = max_segments;
}
/*
EXPORT_SYMBOL(blk_queue_max_phys_segments);
*/

/**
 * blk_queue_max_hw_segments - set max hw segments for a request for this queue
 * @q:  the request queue for the device
 * @max_segments:  max number of segments
 *
 * Description:
 *    Enables a low level driver to set an upper limit on the number of
 *    hw data segments in a request.  This would be the largest number of
 *    address/length pairs the host adapter can actually give as once
 *    to the device.
 **/
void blk_queue_max_hw_segments(request_queue_t *q, unsigned short max_segments)
{
	if (!max_segments) {
		max_segments = 1;
		printk("%s: set to minimum %d\n", __FUNCTION__, max_segments);
	}

	q->max_hw_segments = max_segments;
}
/*
EXPORT_SYMBOL(blk_queue_max_hw_segments);
*/

/**
 * blk_queue_max_segment_size - set max segment size for blk_rq_map_sg
 * @q:  the request queue for the device
 * @max_size:  max size of segment in bytes
 *
 * Description:
 *    Enables a low level driver to set an upper limit on the size of a
 *    coalesced segment
 **/
void blk_queue_max_segment_size(request_queue_t *q, unsigned int max_size)
{
 #ifndef USB_NO_VMM // no vitrtul memory management
	if (max_size < PAGE_CACHE_SIZE) {
		max_size = PAGE_CACHE_SIZE;
		printk("%s: set to minimum %d\n", __FUNCTION__, max_size);
	}
 #endif // ndef USB_NO_VMM

	q->max_segment_size = max_size;
}
/*
EXPORT_SYMBOL(blk_queue_max_segment_size);
*/

/**
 * blk_queue_hardsect_size - set hardware sector size for the queue
 * @q:  the request queue for the device
 * @size:  the hardware sector size, in bytes
 *
 * Description:
 *   This should typically be set to the lowest possible sector size
 *   that the hardware can operate on (possible without reverting to
 *   even internal read-modify-write operations). Usually the default
 *   of 512 covers most hardware.
 **/
void blk_queue_hardsect_size(request_queue_t *q, unsigned short size)
{
	q->hardsect_size = size;
}
/*
EXPORT_SYMBOL(blk_queue_hardsect_size);
*/

/*
 * Returns the minimum that is _not_ zero, unless both are zero.
 */
#define min_not_zero(l, r) (l == 0) ? r : ((r == 0) ? l : min(l, r))

/**
 * blk_queue_stack_limits - inherit underlying queue limits for stacked drivers
 * @t:	the stacking driver (top)
 * @b:  the underlying device (bottom)
 **/
void blk_queue_stack_limits(request_queue_t *t, request_queue_t *b)
{
	/* zero is "infinity" */
	t->max_sectors = min_not_zero(t->max_sectors,b->max_sectors);

	t->max_phys_segments = min(t->max_phys_segments,b->max_phys_segments);
	t->max_hw_segments = min(t->max_hw_segments,b->max_hw_segments);
	t->max_segment_size = min(t->max_segment_size,b->max_segment_size);
	t->hardsect_size = max(t->hardsect_size,b->hardsect_size);
}
/*
EXPORT_SYMBOL(blk_queue_stack_limits);
*/
/**
 * blk_queue_segment_boundary - set boundary rules for segment merging
 * @q:  the request queue for the device
 * @mask:  the memory boundary mask
 **/
void blk_queue_segment_boundary(request_queue_t *q, unsigned long mask)
{
 #ifndef USB_NO_VMM // no vitrtul memory management
	if (mask < PAGE_CACHE_SIZE - 1) {
		mask = PAGE_CACHE_SIZE - 1;
		printk("%s: set to minimum %lx\n", __FUNCTION__, mask);
	}
 #endif // ndef USB_NO_VMM

	q->seg_boundary_mask = mask;
}
/*
EXPORT_SYMBOL(blk_queue_segment_boundary);
*/
/**
 * blk_queue_dma_alignment - set dma length and memory alignment
 * @q:     the request queue for the device
 * @mask:  alignment mask
 *
 * description:
 *    set required memory and length aligment for direct dma transactions.
 *    this is used when buiding direct io requests for the queue.
 *
 **/
void blk_queue_dma_alignment(request_queue_t *q, int mask)
{
	q->dma_alignment = mask;
}
/*
EXPORT_SYMBOL(blk_queue_dma_alignment);
*/
/**
 * blk_queue_find_tag - find a request by its tag and queue
 *
 * @q:	 The request queue for the device
 * @tag: The tag of the request
 *
 * Notes:
 *    Should be used when a device returns a tag and you want to match
 *    it with a request.
 *
 *    no locks need be held.
 **/
struct request *blk_queue_find_tag(request_queue_t *q, int tag)
{
	struct blk_queue_tag *bqt = q->queue_tags;
/*
	if (unlikely(bqt == NULL || tag >= bqt->real_max_depth))
*/
	if (bqt == NULL || tag >= bqt->real_max_depth)
		return NULL;

	return bqt->tag_index[tag];
}
/*
EXPORT_SYMBOL(blk_queue_find_tag);
*/
/**
 * blk_queue_free_tags - release tag maintenance info
 * @q:  the request queue for the device
 *
 *  Notes:
 *    blk_cleanup_queue() will take care of calling this function, if tagging
 *    has been used. So there's usually no need to call this directly, unless
 *    tagging is just being disabled but the queue remains in function.
 **/
void blk_queue_free_tags(request_queue_t *q)
{
	struct blk_queue_tag *bqt = q->queue_tags;

	if (!bqt)
		return;

	if (atomic_dec_and_test(&bqt->refcnt)) {
		/*
		BUG_ON(bqt->busy);
		*/
 if (bqt->busy) {
		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }
		/*
		BUG_ON(!list_empty(&bqt->busy_list));
		*/
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	/*
	#ifndef _ETCPIP
	 if (!((!list_empty(&bqt->busy_list))!=0)) {
	 #else
	*/
	 if (!list_empty(bqt->busy_list)) {
	/*
	#endif
	*/
		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }
 #endif // ndef NO_ELEVATOR 
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] blk_queue_free_tags() : UsbFreeMemC(bqt->tag_index) .. \n"
			));
		/*
		kfree(bqt->tag_index);
		*/
		UsbFreeMemC(USB_MEM_C_S,(unsigned char *) bqt->tag_index
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(struct request *)
					#endif
					);

		bqt->tag_index = NULL;

		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] blk_queue_free_tags() : UsbFreeMemC(bqt->tag_map) .. \n"
			));
		/*
		kfree(bqt->tag_map);
		*/
		UsbFreeMemC(USB_MEM_C_S,(unsigned char *) bqt->tag_map
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(unsigned long)
					#endif
					);

		bqt->tag_map = NULL;

		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] blk_queue_free_tags() : UsbFreeMemC(bqt) .. \n"
			));
		/*
		kfree(bqt);
		*/
		UsbFreeMemC(USB_MEM_C_S,(unsigned char *) bqt
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(struct blk_queue_tag)
					#endif
					);
	}

	q->queue_tags = NULL;
	q->queue_flags &= ~(1 << QUEUE_FLAG_QUEUED);
}
/*
EXPORT_SYMBOL(blk_queue_free_tags);
*/
static int
init_tag_map(request_queue_t *q, struct blk_queue_tag *tags, int depth)
{
 int bits, i;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] init_tag_map() Begin .. \n"));

 if (depth > q->nr_requests * 2) {
		depth = q->nr_requests * 2;
		/*
		printk(KERN_ERR "%s: adjusted depth to %d\n",
				__FUNCTION__, depth);
		*/
		dprintf("[USBMS] Error :%s: adjusted depth to %d\n",
				__FUNCTION__, depth);
	}

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] init_tag_map() : tags->tag_index = UsbAllocMemC() .. \n"
	));

    /*
	tags->tag_index = kmalloc(depth * sizeof(struct request *), GFP_ATOMIC);
	*/
	tags->tag_index = (struct request **) UsbAllocMemC(USB_MEM_C_S
				#ifdef USB_DBG_MEM_C_ON
					, depth * sizeof(struct request *)
		        #endif
           		) ;

	if (!tags->tag_index)
		goto fail;

	bits = (depth / BLK_TAGS_PER_LONG) + 1;

	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] init_tag_map() : tags->tag_map = UsbAllocMemC() .. \n"
		));

	/*
	tags->tag_map = kmalloc(bits * sizeof(unsigned long), GFP_ATOMIC);
	*/
	tags->tag_map = (unsigned long *) UsbAllocMemC(USB_MEM_C_S
				#ifdef USB_DBG_MEM_C_ON
					, bits * sizeof(unsigned long)
		        #endif
           		) ;

	if (!tags->tag_map)
		goto fail;

	memset(tags->tag_index, 0, depth * sizeof(struct request *));
	memset(tags->tag_map, 0, bits * sizeof(unsigned long));
	tags->max_depth = depth;
	tags->real_max_depth = bits * BITS_PER_LONG;

	/*
	 * set the upper bits if the depth isn't a multiple of the word size
	 */
	for (i = depth; i < bits * BLK_TAGS_PER_LONG; i++)
		{
		/*
		__set_bit(i, tags->tag_map);
		*/
		set_bit(i, tags->tag_map);
		}

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
 #ifndef _ETCPIP
	INIT_LIST_HEAD(&tags->busy_list);
  #else
	tags->busy_list = &tags_busy_list ;
 #endif
 #endif // ndef NO_ELEVATOR 
 tags->busy = 0;
 atomic_set(&tags->refcnt, 1);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] init_tag_map() End. \n"));
 return 0;

 fail:

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] init_tag_map() : UsbFreeMemC(tags->tag_index) .. \n"
		));
 /*
 kfree(tags->tag_index);
 */
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) tags->tag_index
					#if defined(USB_DBG_MEM_C_ON)
						,depth * sizeof(struct request *)
					#endif
					);
 tags->tag_index = NULL ; // B. S. Liaw add this

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] init_tag_map() End with Fail !! \n"));
 return -ENOMEM;
}

/**
 * blk_queue_init_tags - initialize the queue tag info
 * @q:  the request queue for the device
 * @depth:  the maximum queue depth supported
 **/
int blk_queue_init_tags(request_queue_t *q, int depth,
			struct blk_queue_tag *tags)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_resize_tags() Begin .. \n"));
 if (!tags) {

	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] blk_queue_init_tags() : tags = UsbAllocMemC() .. \n"
		));

	/*
	tags = kmalloc(sizeof(struct blk_queue_tag), GFP_ATOMIC);
	*/
	tags = (struct blk_queue_tag *) UsbAllocMemC(USB_MEM_C_S
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct blk_queue_tag)
		        #endif
           		) ;

		if (!tags)
			goto fail;

		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_resize_tags() : init_tag_map() .. \n"));
		if (init_tag_map(q, tags, depth))
			goto fail;
	} else
		atomic_inc(&tags->refcnt);

	/*
	 * assign it, all done
	 */
	q->queue_tags = tags;
	q->queue_flags |= (1 << QUEUE_FLAG_QUEUED);
	return 0;
 fail:

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] blk_queue_init_tags() : UsbFreeMemC(tags) .. \n"
		));
 /*
 kfree(tags);
 */
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) tags
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(struct blk_queue_tag)
					#endif
					);

 return -ENOMEM;
}
/*
EXPORT_SYMBOL(blk_queue_init_tags);
*/
/**
 * blk_queue_resize_tags - change the queueing depth
 * @q:  the request queue for the device
 * @new_depth: the new max command queueing depth
 *
 *  Notes:
 *    Must be called with the queue lock held.
 **/
int blk_queue_resize_tags(request_queue_t *q, int new_depth)
{
 struct blk_queue_tag *bqt = q->queue_tags;
 struct request **tag_index;
 unsigned long *tag_map;
 int bits, max_depth;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_resize_tags() Begin .. \n"));
	if (!bqt)
		return -ENXIO;

	/*
	 * don't bother sizing down
	 */
	if (new_depth <= bqt->real_max_depth) {
		bqt->max_depth = new_depth;
		return 0;
	}

	/*
	 * save the old state info, so we can copy it back
	 */
 tag_index = bqt->tag_index;
 tag_map = bqt->tag_map;
 max_depth = bqt->real_max_depth;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_resize_tags() : init_tag_map() .. \n"));
 if (init_tag_map(q, bqt, new_depth))
		return -ENOMEM;

 memcpy(bqt->tag_index, tag_index, max_depth * sizeof(struct request *));
 bits = max_depth / BLK_TAGS_PER_LONG;
 memcpy(bqt->tag_map, tag_map, bits * sizeof(unsigned long));

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] blk_queue_resize_tags() : UsbFreeMemC(tag_index) .. \n"
		));
 /*
 kfree(tag_index);
 */
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) tag_index
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(struct request *)
					#endif
					);

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] blk_queue_resize_tags() : UsbFreeMemC(tag_map) .. \n"
		));
 /*
 kfree(tag_map);
 */
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) tag_map
					#if defined(USB_DBG_MEM_C_ON)
						,bits * sizeof(unsigned long)
					#endif
					);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_resize_tags() End. \n"));
 return 0;
}

/**
 * blk_queue_end_tag - end tag operations for a request
 * @q:  the request queue for the device
 * @rq: the request that has completed
 *
 *  Description:
 *    Typically called when end_that_request_first() returns 0, meaning
 *    all transfers have been done for a request. It's important to call
 *    this function before end_that_request_last(), as that will put the
 *    request back on the free list thus corrupting the internal tag list.
 *
 *  Notes:
 *   queue lock must be held.
 **/
void blk_queue_end_tag(request_queue_t *q, struct request *rq)
{
	struct blk_queue_tag *bqt = q->queue_tags;
	int tag = rq->tag;
	/*
	BUG_ON(tag == -1);
	*/
 if (tag == -1) {
		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }
 /*
 if (unlikely(tag >= bqt->real_max_depth))
 */
 if (tag >= bqt->real_max_depth)
		return;
 /*
 if (unlikely(!__test_and_clear_bit(tag, bqt->tag_map))) {
 */
 if (!__test_and_clear_bit(tag, bqt->tag_map)) {
		printk("attempt to clear non-busy tag (%d)\n", tag);
		return;
	}

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_end_tag() : list_del_init() .. \n"));
 /*
 #ifndef _ETCPIP
 */
	list_del_init(&rq->queuelist);
 /*
 #else
	list_del_init(rq->queuelist);
 #endif
 */
 #endif // ndef NO_ELEVATOR
 rq->flags &= ~REQ_QUEUED;
 rq->tag = -1;
 /*
	if (unlikely(bqt->tag_index[tag] == NULL))
 */
 if (bqt->tag_index[tag] == NULL)
		printk("tag %d is missing\n", tag);

 bqt->tag_index[tag] = NULL;
 bqt->busy--;
}

/*
EXPORT_SYMBOL(blk_queue_end_tag);
*/

/**
 * blk_queue_start_tag - find a free tag and assign it
 * @q:  the request queue for the device
 * @rq:  the block request that needs tagging
 *
 *  Description:
 *    This can either be used as a stand-alone helper, or possibly be
 *    assigned as the queue &prep_rq_fn (in which case &struct request
 *    automagically gets a tag assigned). Note that this function
 *    assumes that any type of request can be queued! if this is not
 *    true for your device, you must check the request type before
 *    calling this function.  The request will also be removed from
 *    the request queue, so it's the drivers responsibility to readd
 *    it if it should need to be restarted for some reason.
 *
 *  Notes:
 *   queue lock must be held.
 **/
int blk_queue_start_tag(request_queue_t *q, struct request *rq)
{
 struct blk_queue_tag *bqt = q->queue_tags;
 unsigned long *map = bqt->tag_map;
 int tag = 0;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_start_tag() Begin .. \n"));
 /*
 if (unlikely((rq->flags & REQ_QUEUED))) {
 */
 if(rq->flags & REQ_QUEUED) {
	#ifdef USB_DBGSTRING_ON
		/*
		printk(KERN_ERR 
		       "request %p for device [%s] already tagged %d",
		       rq, rq->rq_disk ? rq->rq_disk->disk_name : "?", rq->tag);
		BUG();
		*/
		dprintf("[USBMS] Error : request %p for device [%s] already tagged %d",
		       rq, rq->rq_disk ? rq->rq_disk->disk_name : "?", rq->tag);
	  #else
		dprintf("[USBMS] Error : request %p for device already tagged %d",
		       rq, rq->tag);
	#endif // ndef USB_DBGSTRING_ON
   }

 for (map = bqt->tag_map; *map == -1UL; map++) {
		tag += BLK_TAGS_PER_LONG;

		if (tag >= bqt->max_depth)
			return 1;
	}

	tag += ffz(*map);
	/*
	__set_bit(tag, bqt->tag_map);
	*/
	set_bit(tag, bqt->tag_map);

	rq->flags |= REQ_QUEUED;
	rq->tag = tag;
	bqt->tag_index[tag] = rq;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_start_tag() : blkdev_dequeue_request() .. \n"));
	blkdev_dequeue_request(rq);
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	 #ifndef _ETCPIP
		list_add(&rq->queuelist, &bqt->busy_list);
	  #else
		list_add(&rq->queuelist, bqt->busy_list);
	 #endif
 #endif // ndef NO_ELEVATOR
 bqt->busy++;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_start_tag() End. \n"));
 return 0;
}
/*
EXPORT_SYMBOL(blk_queue_start_tag);
*/
/**
 * blk_queue_invalidate_tags - invalidate all pending tags
 * @q:  the request queue for the device
 *
 *  Description:
 *   Hardware conditions may dictate a need to stop all pending requests.
 *   In this case, we will safely clear the block side of the tag queue and
 *   readd all requests to the request queue in the right order.
 *
 *  Notes:
 *   queue lock must be held.
 **/
void blk_queue_invalidate_tags(request_queue_t *q)
{
 struct blk_queue_tag *bqt = q->queue_tags;
 struct list_head *tmp, *n;
 struct request *rq;


 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
 #ifndef _ETCPIP
	list_for_each_safe(tmp, n, &bqt->busy_list) {
  #else
	list_for_each_safe(tmp, n, bqt->busy_list) {
 #endif
		rq = list_entry_rq(tmp);

		if (rq->tag == -1) {
			printk("bad tag found on list\n");
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_invalidate_tags() : list_del_init() .. \n"));
		 /*
		 #ifndef _ETCPIP
		 */
			list_del_init(&rq->queuelist);
		 /*
		 #else
			list_del_init(rq->queuelist);
		 #endif
		 */
		#endif // ndef NO_ELEVATOR
			rq->flags &= ~REQ_QUEUED;
		} else
			blk_queue_end_tag(q, rq);

		rq->flags &= ~REQ_STARTED;
 	
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		// __elv_add_request() is on include/linux/elevator.h
		__elv_add_request(q, rq, ELEVATOR_INSERT_BACK, 0);
	 #else
		USBMS_DBG_INFO(iprintf("[USBMS INFO] blk_queue_invalidate_tags() : Bug on req !! \n"));

		USBMS_DBG_MEM(iprintf("[USBMS MEM] blk_queue_invalidate_tags() : q->req=%08XH \n",
			q->req)) ;

		#ifdef USBMS_BS_ASSERT
			if (!(   (unsigned long) q->req
			   > (unsigned long) 0x200000L
    	     )
		   )
		   {
			iprintf("[ASSERT][USBMS] blk_queue_invalidate_tags() : " ) ;
		    iprintf("Error : Invlid address q->req=%08XH \n",
				q->req ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
		#endif // USBMS_BS_ASSERT

		req = q->req;
		/*
		if ((rq->flags & REQ_DONTPREP) || !q->prep_rq_fn)
			break;
		*/
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_invalidate_tags() : q->prep_rq_fn() .. \n"));
		ret = q->prep_rq_fn(q, req);
		if (ret == BLKPREP_OK) {
			// break;
		   } 
		/*
		  else if (ret == BLKPREP_DEFER) {
			req = NULL;
			break;
		   }
		
		  else if (ret == BLKPREP_KILL) {
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_queue_invalidate_tags() : blkdev_dequeue_request() .. \n"));
			blkdev_dequeue_request(req);
			rq->flags |= REQ_QUIET;
			while (end_that_request_first(req, 0, rq->nr_sectors))
				;
			end_that_request_last(req);
		   }
		*/
		 else {
			printk("%s: bad return=%d\n", __FUNCTION__, ret);
			// break;
		   }
		
	#endif // ndef NO_ELEVATOR
   }
 #endif // ndef NO_ELEVATOR 
}
/*
EXPORT_SYMBOL(blk_queue_invalidate_tags);
*/
static char *rq_flags[] = {
	"REQ_RW",
	"REQ_FAILFAST",
	"REQ_SOFTBARRIER",
	"REQ_HARDBARRIER",
	"REQ_CMD",
	"REQ_NOMERGE",
	"REQ_STARTED",
	"REQ_DONTPREP",
	"REQ_QUEUED",
	"REQ_PC",
	"REQ_BLOCK_PC",
	"REQ_SENSE",
	"REQ_FAILED",
	"REQ_QUIET",
	"REQ_SPECIAL",
	"REQ_DRIVE_CMD",
	"REQ_DRIVE_TASK",
	"REQ_DRIVE_TASKFILE",
	"REQ_PREEMPT",
	"REQ_PM_SUSPEND",
	"REQ_PM_RESUME",
	"REQ_PM_SHUTDOWN",
};

#ifdef USBMS_DBG_BLOCK_ON
void blk_dump_rq_flags(struct request *rq, char *msg)
{
 int bit;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_dump_rq_flags() Begin ..\n"));
 #ifdef USB_DBGSTRING_ON
	/*
	printk("%s: dev %s: flags = ", msg,
	*/
	iprintf("[USBMS BLOCK] %s: dev %s: flags = ", msg,
		rq->rq_disk ? rq->rq_disk->disk_name : "?");
  #else
	iprintf("[USBMS BLOCK] %s  ", msg );
 #endif // ndef USB_DBGSTRING_ON

 bit = 0;
 do {
	if (rq->flags & (1 << bit))
			printk("%s ", rq_flags[bit]);
		bit++;
	} while (bit < __REQ_NR_BITS);
 /*
 printk("\nsector %llu, nr/cnr %lu/%u\n", (unsigned long long)rq->sector,
 */
 iprintf("\n") ;
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	iprintf("[USBMS BLOCK] sector %lu, nr/cnr %lu/%u\n", rq->sector,
		       rq->nr_sectors,
		       rq->current_nr_sectors);
  #else
 */
	iprintf("[USBMS BLOCK] sector %lu \n", rq->sector	) ;
 /*
 #endif
 */

 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	// printk("bio %p, biotail %p, buffer %p, data %p, len %u\n", rq->bio, rq->biotail, rq->buffer, rq->data, rq->data_len);
	iprintf("[USBMS BLOCK] bio=%p, biotail=%p, buffer=%p, data=%08XH, len=%u\n",
		rq->bio, rq->biotail, rq->buffer, rq->data, rq->data_len);
 #else
 */
	iprintf("[USBMS BLOCK] bio=%p, buffer=%p, data=%08XH, len=%u\n",
		rq->bio, rq->buffer, rq->data, rq->data_len);
 /*
 #endif // NO_ELEVATOR
 */

 if (rq->flags & (REQ_BLOCK_PC | REQ_PC)) {
		printk("cdb: ");
		for (bit = 0; bit < sizeof(rq->cmd); bit++)
			printk("%02x ", rq->cmd[bit]);
		printk("\n");
	}
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_dump_rq_flags() End.\n"));
}
#endif // USBMS_DBG_BLOCK_ON
/*
EXPORT_SYMBOL(blk_dump_rq_flags);
*/
/*
#ifndef _ETCPIP
void blk_recount_segments(request_queue_t *q, struct bio *bio)
{
	struct bio_vec *bv, *bvprv = NULL;
	int i, nr_phys_segs, nr_hw_segs, seg_size, cluster;
	
 #ifndef BLOCK_IO_REDUCED
	// if (unlikely(!bio->bi_io_vec))
	
	if (!bio->bi_io_vec)
		return;
 #endif // ndef BLOCK_IO_REDUCED

	cluster = q->queue_flags & (1 << QUEUE_FLAG_CLUSTER);
	seg_size = nr_phys_segs = nr_hw_segs = 0;
	bio_for_each_segment(bv, bio, i) {
		if (bvprv && cluster) {
			if (seg_size + bv->bv_len > q->max_segment_size)
				goto new_segment;
			if (!BIOVEC_PHYS_MERGEABLE(bvprv, bv))
				goto new_segment;
			if (!BIOVEC_SEG_BOUNDARY(q, bvprv, bv))
				goto new_segment;

			seg_size += bv->bv_len;
			bvprv = bv;
			continue;
		}
 new_segment:
	#ifndef USB_NO_VMM // vitrtul memory management
		if (!bvprv || !BIOVEC_VIRT_MERGEABLE(bvprv, bv))
			nr_hw_segs++;
	#endif // ndef USB_NO_VMM

		nr_phys_segs++;
		bvprv = bv;
		seg_size = bv->bv_len;
   }

	bio->bi_phys_segments = nr_phys_segs;
	bio->bi_hw_segments = nr_hw_segs;
	bio->bi_flags |= (1 << BIO_SEG_VALID);
}
#endif // ndef _ETCPIP
*/


// #ifndef _ETCPIP
int blk_phys_contig_segment(request_queue_t *q, struct bio *bio,
				   struct bio *nxt)
{
	if (!(q->queue_flags & (1 << QUEUE_FLAG_CLUSTER)))
		return 0;

 #ifndef USB_NO_VMM // vitrtul memory management
	if (!BIOVEC_PHYS_MERGEABLE(__BVEC_END(bio), __BVEC_START(nxt)))
		return 0;
 #endif // ndef USB_NO_VMM
	if (bio->bi_size + nxt->bi_size > q->max_segment_size)
		return 0;

	/*
	 * bio and nxt are contigous in memory, check if the queue allows
	 * these two to be merged into one
	 */
 #ifndef USB_NO_VMM // vitrtul memory management
	if (BIO_SEG_BOUNDARY(q, bio, nxt))
		return 1;
 #endif // ndef USB_NO_VMM

 return 0;
}
// #endif // ndef _ETCPIP
/*
EXPORT_SYMBOL(blk_phys_contig_segment);
*/

// #ifndef _ETCPIP
int blk_hw_contig_segment(request_queue_t *q, struct bio *bio,
				 struct bio *nxt)
{
	if (!(q->queue_flags & (1 << QUEUE_FLAG_CLUSTER)))
		return 0;

 #ifndef USB_NO_VMM // vitrtul memory management
	if (!BIOVEC_VIRT_MERGEABLE(__BVEC_END(bio), __BVEC_START(nxt)))
		return 0;
 #endif // ndef USB_NO_VMM
	if (bio->bi_size + nxt->bi_size > q->max_segment_size)
		return 0;

	/*
	 * bio and nxt are contigous in memory, check if the queue allows
	 * these two to be merged into one
	 */
 #ifndef USB_NO_VMM // vitrtul memory management
	if (BIO_SEG_BOUNDARY(q, bio, nxt))
		return 1;
 #endif // ndef USB_NO_VMM

 return 0;
}
// #endif // ndef _ETCPIP
/*
EXPORT_SYMBOL(blk_hw_contig_segment);
*/

/*
 * map a request to scatterlist, return number of sg entries setup. Caller
 * must make sure sg can hold rq->nr_phys_segments entries
 */
/*
#ifndef NO_SCATTER_LIST
int blk_rq_map_sg(request_queue_t *q, struct request *rq, struct scatterlist *sg)
{
 // #ifndef USB_NO_VMM // vitrtul memory management
	struct bio_vec *bvec, *bvprv;
	struct bio *bio;
 // #endif // ndef USB_NO_VMM
	int nsegs, i, cluster;

	nsegs = 0;
	cluster = q->queue_flags & (1 << QUEUE_FLAG_CLUSTER);

	
	// for each bio in rq
	 
	bvprv = NULL;
 // #ifndef USB_NO_VMM // vitrtul memory management
	rq_for_each_bio(bio, rq) {
		
		// for each segment in bio
		
		bio_for_each_segment(bvec, bio, i) {
			int nbytes = bvec->bv_len;

			if (bvprv && cluster) {
				if (sg[nsegs - 1].length + nbytes > q->max_segment_size)
					goto new_segment;

				if (!BIOVEC_PHYS_MERGEABLE(bvprv, bvec))
					goto new_segment;
				if (!BIOVEC_SEG_BOUNDARY(q, bvprv, bvec))
					goto new_segment;

				sg[nsegs - 1].length += nbytes;
			} else {
 new_segment:
				memset(&sg[nsegs],0,sizeof(struct scatterlist));
				sg[nsegs].page = bvec->bv_page;
				sg[nsegs].length = nbytes;
				sg[nsegs].offset = bvec->bv_offset;

				nsegs++;
			}
			bvprv = bvec;
		} // segments in bio 
	} // bios in rq *

 // #endif // ndef USB_NO_VMM
	return nsegs;
}
#endif // ndef NO_SCATTER_LIST
*/
/*
EXPORT_SYMBOL(blk_rq_map_sg);
*/

/*
 * the standard queue merge functions, can be overridden with device
 * specific ones if so desired
 */

// #ifndef USB_NO_VMM // vitrtul memory management
static inline int ll_new_mergeable(request_queue_t *q,
				   struct request *req,
				   struct bio *bio)
{
 #ifndef NO_SCATTER_LIST
	int nr_phys_segs = bio_phys_segments(q, bio);
 #endif // ndef NO_SCATTER_LIST

 #ifndef NO_SCATTER_LIST
	if (req->nr_phys_segments + nr_phys_segs > q->max_phys_segments) {
		req->flags |= REQ_NOMERGE;
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		q->last_merge = NULL;
 #endif // ndef NO_ELEVATOR 
		return 0;
	}
 #endif // ndef NO_SCATTER_LIST

	/*
	 * A hw segment is just getting larger, bump just the phys
	 * counter.
	 */
 #ifndef NO_SCATTER_LIST
	req->nr_phys_segments += nr_phys_segs;
 #endif // ndef NO_SCATTER_LIST
	return 1;
}
// #endif // ndef USB_NO_VMM

// #ifndef USB_NO_VMM // vitrtul memory management
static inline int ll_new_hw_segment(request_queue_t *q,
				    struct request *req,
				    struct bio *bio)
{
 #ifndef NO_SCATTER_LIST
	int nr_hw_segs = bio_hw_segments(q, bio);
	int nr_phys_segs = bio_phys_segments(q, bio);
 #endif // ndef NO_SCATTER_LIST


 #ifndef NO_SCATTER_LIST
	if (req->nr_hw_segments + nr_hw_segs > q->max_hw_segments
		#ifndef NO_SCATTER_LIST
		    || req->nr_phys_segments + nr_phys_segs > q->max_phys_segments
		#endif // ndef NO_SCATTER_LIST
       ) {
		req->flags |= REQ_NOMERGE;
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		q->last_merge = NULL;
 #endif // ndef NO_ELEVATOR 
		return 0;
	   }
 #endif // ndef NO_SCATTER_LIST

	/*
	 * This will form the start of a new hw segment.  Bump both
	 * counters.
	 */
 #ifndef NO_SCATTER_LIST
	req->nr_hw_segments += nr_hw_segs;
	req->nr_phys_segments += nr_phys_segs;
 #endif // ndef NO_SCATTER_LIST
 return 1;
}
// #endif // ndef USB_NO_VMM
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
static int ll_back_merge_fn(request_queue_t *q, struct request *req, 
			    struct bio *bio)
{
	if (req->nr_sectors + bio_sectors(bio) > q->max_sectors) {
		req->flags |= REQ_NOMERGE;
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		q->last_merge = NULL;
 #endif // ndef NO_ELEVATOR 
		return 0;
	}

 #ifndef USB_NO_VMM // vitrtul memory management
	if (BIOVEC_VIRT_MERGEABLE(__BVEC_END(req->biotail), __BVEC_START(bio)))
		return ll_new_mergeable(q, req, bio);
 #endif // ndef USB_NO_VMM

 return ll_new_hw_segment(q, req, bio);
}


static int ll_front_merge_fn(request_queue_t *q, struct request *req, 
			     struct bio *bio)
{
	if (req->nr_sectors + bio_sectors(bio) > q->max_sectors) {
		req->flags |= REQ_NOMERGE;
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		q->last_merge = NULL;
 #endif // ndef NO_ELEVATOR 
		return 0;
	}
 
 #ifndef USB_NO_VMM // vitrtul memory management
	if (BIOVEC_VIRT_MERGEABLE(__BVEC_END(bio), __BVEC_START(req->bio)))
		return ll_new_mergeable(q, req, bio);
 #endif // ndef USB_NO_VMM
 
 return ll_new_hw_segment(q, req, bio);
}


static int ll_merge_requests_fn(request_queue_t *q, struct request *req,
				struct request *next)
{
 #ifndef NO_SCATTER_LIST
	int total_phys_segments = req->nr_phys_segments +next->nr_phys_segments;
	int total_hw_segments = req->nr_hw_segments + next->nr_hw_segments;
 #endif // ndef NO_SCATTER_LIST

 
 // First check if the either of the requests are re-queued
 // requests.  Can't merge them if they are.
 
 if (req->special || next->special)
		return 0;

	
 //	Will it become to large?
	
 if ((req->nr_sectors + next->nr_sectors) > q->max_sectors)
		return 0;

 #ifndef NO_SCATTER_LIST
	total_phys_segments = req->nr_phys_segments + next->nr_phys_segments;
 #endif // ndef NO_SCATTER_LIST
 #ifndef NO_SCATTER_LIST
	if (blk_phys_contig_segment(q, req->biotail, next->bio))
		total_phys_segments--;
 #endif // ndef NO_SCATTER_LIST

 #ifndef NO_SCATTER_LIST
	if (total_phys_segments > q->max_phys_segments)
		return 0;

	total_hw_segments = req->nr_hw_segments + next->nr_hw_segments;
 #endif // ndef NO_SCATTER_LIST
 #ifndef NO_SCATTER_LIST
	if (blk_hw_contig_segment(q, req->biotail, next->bio))
		total_hw_segments--;
 #endif // ndef NO_SCATTER_LIST

 #ifndef NO_SCATTER_LIST
	if (total_hw_segments > q->max_hw_segments)
		return 0;
 #endif // ndef NO_SCATTER_LIST

 // Merge is OK...
 #ifndef NO_SCATTER_LIST
	req->nr_phys_segments = total_phys_segments;
	req->nr_hw_segments = total_hw_segments;
 #endif // ndef NO_SCATTER_LIST
 return 1;
}
#endif // ndef NO_ELEVATOR 
*/

/*
 * "plug" the device if there are no outstanding requests: this will
 * force the transfer to start only after we have put all the requests
 * on the list.
 *
 * This is called with interrupts off and no requests on the queue and
 * with the queue lock held.
 */
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
void blk_plug_device(request_queue_t *q)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_plug_device() Begin .. \n"));
 // WARN_ON(!irqs_disabled());
 
 // if (!((!irqs_disabled())!=0)) {
 //		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
 //	   }

 // * don't plug a stopped queue, it must be paired with blk_start_queue()
 // * which will restart the queueing

 if (    !blk_queue_plugged(q)
	  && !test_bit(QUEUE_FLAG_STOPPED, &q->queue_flags)) {
	#ifndef NO_SPINLOCK
		spin_lock(&blk_plug_lock);
	#endif // ndef NO_SPINLOCK
		
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		USBMS_DBG_MEM(iprintf("[USBMS MEM] blk_plug_device() : blk_plug_list=%08XH \n",
			blk_plug_list));
		#ifdef USBMS_BS_ASSERT
			if (!(   (unsigned long) &blk_plug_list
				   > (unsigned long) 0x200000L
		         )
			   )
			   {
				iprintf("[ASSERT][USBMS] blk_plug_device() : " ) ;
    		    iprintf("Error : Invlid address &blk_plug_list=%08XH \n",
						&blk_plug_list) ;
			    iprintf("[ASSERT][USBMS] Continue ..\n") ;
    		   }
		#endif // USBMS_BS_ASSERT
	#endif // ndef NO_ELEVATOR
		
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_plug_device() : list_add_tail() .. \n"));
		list_add_tail(&q->plug_list, &blk_plug_list);
	#endif // ndef NO_ELEVATOR
	#ifndef NO_TIMER
		mod_timer(&q->unplug_timer, jiffies + q->unplug_delay);
	#endif // ndef NO_TIMER
	#ifndef NO_SPINLOCK
		spin_unlock(&blk_plug_lock);
	#endif // ndef NO_SPINLOCK
   }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_plug_device() End. \n"));
}
#endif // ndef NO_ELEVATOR
*/
/*
EXPORT_SYMBOL(blk_plug_device);
*/

/*
 * remove the queue from the plugged list, if present. called with
 * queue lock held and interrupts disabled.
 */
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
int blk_remove_plug(request_queue_t *q)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_remove_plug() Begin .. \n"));

 // WARN_ON(!irqs_disabled());
 // if(!((!irqs_disabled())!=0)) {
 //    dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
 //   }

 if (blk_queue_plugged(q)) {
	#ifndef NO_SPINLOCK
		spin_lock(&blk_plug_lock);
	#endif // ndef NO_SPINLOCK
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_remove_plug() : list_del_init() .. \n"));
		list_del_init(&q->plug_list);
	#endif // ndef NO_ELEVATOR
	#ifndef NO_TIMER
		del_timer(&q->unplug_timer);
	#endif // ndef NO_TIMER
	#ifndef NO_SPINLOCK
		spin_unlock(&blk_plug_lock);
	#endif // ndef NO_SPINLOCK
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_remove_plug() End with 1 \n"));
	return 1;
   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_remove_plug() End with 0 \n"));
 return 0;
}
#endif // ndef NO_ELEVATOR
*/
/*
EXPORT_SYMBOL(blk_remove_plug);
*/
/*
 * remove the plug and let it rip..
 */
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
static inline void __generic_unplug_device(request_queue_t *q)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __generic_unplug_device() Begin .. \n"));
 if (test_bit(QUEUE_FLAG_STOPPED, &q->queue_flags))
		return;

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __generic_unplug_device() : blk_remove_plug() .. \n"));
	if (!blk_remove_plug(q))
		return;
 #endif // ndef NO_ELEVATOR
 /*
 #ifndef NO_TIMER
	del_timer(&q->unplug_timer);
 #endif // ndef NO_TIMER
 */
	/*
	 * was plugged, fire request_fn if queue has stuff to do
	 */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	// elv_next_request() is on include/linux/elevator.h
	if (elv_next_request(q))
       {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __generic_unplug_device() : q->request_fn() .. \n"));
		q->request_fn(q);
       }
 
  #else
 */
	/*	
		USBMS_DBG_INFO(iprintf("[USBMS INFO] __generic_unplug_device() : Bug on req !! \n"));

		USBMS_DBG_MEM(iprintf("[USBMS MEM] __generic_unplug_device() : q->req=%08XH \n",
			q->req)) ;

		#ifdef USBMS_BS_ASSERT
			if (!(   (unsigned long) q->req
			   > (unsigned long) 0x200000L
    	     )
		   )
		   {
			iprintf("[ASSERT][USBMS] __generic_unplug_device() : " ) ;
		    iprintf("Error : Invlid address q->req=%08XH \n",
				q->req ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
		#endif // USBMS_BS_ASSERT

		req = q->req;
		
		//if ((rq->flags & REQ_DONTPREP) || !q->prep_rq_fn)
		//	break;
		
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __generic_unplug_device() : q->prep_rq_fn() .. \n"));
		
		// ret = q->prep_rq_fn(q, req);
		// if (ret == BLKPREP_OK) {
		
		if (q->prep_rq_fn(q, req) == BLKPREP_OK) {
			// break;
		   } 
		
		//  else if (ret == BLKPREP_DEFER) {
		//	req = NULL;
		//	break;
		//   }
		
		//  else if (ret == BLKPREP_KILL) {
		//	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __generic_unplug_device() : blkdev_dequeue_request() .. \n"));
		//	blkdev_dequeue_request(req);
		//	rq->flags |= REQ_QUIET;
		//	while (end_that_request_first(req, 0, rq->nr_sectors))
		//		;
		//	end_that_request_last(req);
		//   }
		
		 else {
			// printk("%s: bad return=%d\n", __FUNCTION__, ret);
			printk("%s: q->prep_rq_fn() Failed !! \n", __FUNCTION__);
			// break;
		   }
	*/
 /*
 #endif // ndef NO_ELEVATOR 
 */
}
#endif // ndef NO_ELEVATOR 

/**
 * generic_unplug_device - fire a request queue
 * @data:    The &request_queue_t in question
 *
 * Description:
 *   Linux uses plugging to build bigger requests queues before letting
 *   the device have at them. If a queue is plugged, the I/O scheduler
 *   is still adding and merging requests on the queue. Once the queue
 *   gets unplugged (either by manually calling this function, or by
 *   calling blk_run_queues()), the request_fn defined for the
 *   queue is invoked and transfers started.
 **/
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
void generic_unplug_device(void *data)
{
 request_queue_t *q = data;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] generic_unplug_device() Begin .. \n"));
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irq(q->queue_lock);
 #endif // ndef NO_SPINLOCK
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] generic_unplug_device() : __generic_unplug_device() .. \n"));
 __generic_unplug_device(q);
 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irq(q->queue_lock);
 #endif // ndef NO_SPINLOCK
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] generic_unplug_device() End. \n"));
}
#endif // ndef NO_ELEVATOR
/*
EXPORT_SYMBOL(generic_unplug_device);
*/
/*
#ifndef NO_WORKQUEUE
static void blk_unplug_work(void *data)
{
 request_queue_t *q = data;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_unplug_work() : q->unplug_fn() .. \n"));
 q->unplug_fn(q);
}
#endif // ndef NO_WORKQUEUE
*/
static void blk_unplug_timeout(unsigned long data)
{
	request_queue_t *q = (request_queue_t *)data;

 #ifndef NO_WORKQUEUE
	kblockd_schedule_work(&q->unplug_work);
 #endif // ndef NO_WORKQUEUE
}

/**
 * blk_start_queue - restart a previously stopped queue
 * @q:    The &request_queue_t in question
 *
 * Description:
 *   blk_start_queue() will clear the stop flag on the queue, and call
 *   the request_fn for the queue if it was in a stopped state when
 *   entered. Also see blk_stop_queue(). Must not be called from driver
 *   request function due to recursion issues. Queue lock must be held.
 **/
void blk_start_queue(request_queue_t *q)
{
	clear_bit(QUEUE_FLAG_STOPPED, &q->queue_flags);
 #ifndef NO_WORKQUEUE
	schedule_work(&q->unplug_work);
 #endif // ndef NO_WORKQUEUE
}
/*
EXPORT_SYMBOL(blk_start_queue);
*/
/**
 * blk_stop_queue - stop a queue
 * @q:    The &request_queue_t in question
 *
 * Description:
 *   The Linux block layer assumes that a block driver will consume all
 *   entries on the request queue when the request_fn strategy is called.
 *   Often this will not happen, because of hardware limitations (queue
 *   depth settings). If a device driver gets a 'queue full' response,
 *   or if it simply chooses not to queue more I/O at one point, it can
 *   call this function to prevent the request_fn from being called until
 *   the driver has signalled it's ready to go again. This happens by calling
 *   blk_start_queue() to restart queue operations. Queue lock must be held.
 **/
void blk_stop_queue(request_queue_t *q)
{
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_stop_queue() : blk_remove_plug() .. \n"));
	blk_remove_plug(q);
 #endif // ndef NO_ELEVATOR
 set_bit(QUEUE_FLAG_STOPPED, &q->queue_flags);
}
/*
EXPORT_SYMBOL(blk_stop_queue);
*/
/**
 * blk_run_queue - run a single device queue
 * @q:	The queue to run
 */
void blk_run_queue(struct request_queue *q)
{
 unsigned long flags;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_run_queue() Begin .. \n"));
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(q->queue_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_run_queue() : blk_remove_plug() .. \n"));
	blk_remove_plug(q);
 #endif // ndef NO_ELEVATOR
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_run_queue() : q->request_fn() .. \n"));
 if( ! q->request_fn(q) ) {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] blk_run_queue() : q->request_fn() Failed !! \n"));
   }
 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(q->queue_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_run_queue() End. \n"));
}
/*
EXPORT_SYMBOL(blk_run_queue);
*/
/**
 * blk_run_queues - fire all plugged queues
 *
 * Description:
 *   Start I/O on all plugged queues known to the block layer. Queues that
 *   are currently stopped are ignored. This is equivalent to the older
 *   tq_disk task queue run.
 **/
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	#define blk_plug_entry(entry) list_entry((entry), request_queue_t, plug_list)
#endif // ndef NO_ELEVATOR
*/
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	void blk_run_queues(void)
 // #else
 //  	void blk_run_queues(request_queue_t *q)
 // #endif
{
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	LIST_HEAD(local_plug_list);
 #endif // ndef NO_ELEVATOR


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_run_queues() Begin .. \n"));
 #ifndef NO_SPINLOCK
	spin_lock_irq(&blk_plug_lock);
 #endif // ndef NO_SPINLOCK
 // * this will happen fairly often
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if (list_empty(&blk_plug_list))
		goto out;

	list_splice_init(&blk_plug_list, &local_plug_list);
 #endif // ndef NO_ELEVATOR
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	while (!list_empty(&local_plug_list))
 #endif // ndef NO_ELEVATOR
	   {
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			request_queue_t *q = blk_plug_entry(local_plug_list.next);
		#endif // ndef NO_ELEVATOR
		#ifndef NO_SPINLOCK
			spin_unlock_irq(&blk_plug_lock);
		#endif // ndef NO_SPINLOCK
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_run_queues() : q->unplug_fn() .. \n"));
		q->unplug_fn(q);
		#ifndef NO_SPINLOCK
			spin_lock_irq(&blk_plug_lock);
		#endif // ndef NO_SPINLOCK
	   }
 out:

 #ifndef NO_SPINLOCK
	spin_unlock_irq(&blk_plug_lock);
 #endif // ndef NO_SPINLOCK
}
#endif // ndef NO_ELEVATOR
*/
/*
EXPORT_SYMBOL(blk_run_queues);
*/
/**
 * blk_cleanup_queue: - release a &request_queue_t when it is no longer needed
 * @q:    the request queue to be released
 *
 * Description:
 *     blk_cleanup_queue is the pair to blk_init_queue() or
 *     blk_queue_make_request().  It should be called when a request queue is
 *     being released; typically when a block device is being de-registered.
 *     Currently, its primary task it to free all the &struct request
 *     structures that were allocated to the queue and the queue itself.
 *
 * Caveat:
 *     Hopefully the low level driver will have finished any
 *     outstanding requests first...
 **/
void blk_cleanup_queue(request_queue_t * q)
{
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	struct request_list *rl = &q->rq;
 #endif // ndef NO_ELEVATOR


 if (!atomic_dec_and_test(&q->refcnt))
		return;
 /*
 elevator_exit(q);
 */
 /*
 #ifndef NO_TIMER
	del_timer_sync(&q->unplug_timer);
 #endif // ndef NO_TIMER
 */
 kblockd_flush();
 /*
 #ifndef USB_NO_VMM // no vitrtul memory management
	if (rl->rq_pool)
		mempool_destroy(rl->rq_pool);
 #endif // ndef USB_NO_VMM
 */
 if (blk_queue_tagged(q))
		blk_queue_free_tags(q);
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] blk_cleanup_queue() : UsbFreeMemC(q) , size=%d.. \n" ,
		sizeof(request_queue_t)
		));
	
	// kfree(q);

	UsbFreeMemC(USB_MEM_C_M,(unsigned char *) q
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(request_queue_t) // size : 148
					#endif
					);
 #endif // ndef NO_ELEVATOR
 */
}

/*
EXPORT_SYMBOL(blk_cleanup_queue);
*/

static int blk_init_free_list(request_queue_t *q)
{
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	struct request_list *rl = &q->rq;
 #endif // ndef NO_ELEVATOR

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	rl->count[READ] = rl->count[WRITE] = 0;
 #endif // ndef NO_ELEVATOR
 /*
 #ifndef NO_WAITQUEUE
	init_waitqueue_head(&rl->wait[READ]);
	init_waitqueue_head(&rl->wait[WRITE]);
 #endif // ndef NO_WAITQUEUE
 */
 /*
 #ifndef USB_NO_VMM // no vitrtul memory management
	// kmem_cache_t is on include/linux/slab.h
	rl->rq_pool = mempool_create(BLKDEV_MIN_RQ, mempool_alloc_slab, mempool_free_slab, request_cachep);
	if (!rl->rq_pool)
		return -ENOMEM;
 #endif // ndef USB_NO_VMM
 */
 return 0;
}

// #ifndef USB_NO_VMM // vitrtul memory management
	static int __make_request(request_queue_t *, struct bio *);
// #endif // ndef USB_NO_VMM

#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	static elevator_t *chosen_elevator =
	#if defined(CONFIG_IOSCHED_AS)
		&iosched_as;
	#elif defined(CONFIG_IOSCHED_DEADLINE)
		&iosched_deadline;
	#elif defined(CONFIG_IOSCHED_NOOP)
		&elevator_noop;
	#else
		NULL;
		#error "You must have at least 1 I/O scheduler selected"
	#endif
#endif // ndef NO_ELEVATOR 

#if defined(CONFIG_IOSCHED_AS) || defined(CONFIG_IOSCHED_DEADLINE) || defined (CONFIG_IOSCHED_NOOP)
static int __init elevator_setup(char *str)
{
#ifdef CONFIG_IOSCHED_DEADLINE
	if (!strcmp(str, "deadline"))
		chosen_elevator = &iosched_deadline;
#endif
#ifdef CONFIG_IOSCHED_AS
	if (!strcmp(str, "as"))
		chosen_elevator = &iosched_as;
#endif
#ifdef CONFIG_IOSCHED_NOOP
	if (!strcmp(str, "noop"))
		chosen_elevator = &elevator_noop;
#endif
	return 1;
}

__setup("elevator=", elevator_setup);
#endif /* CONFIG_IOSCHED_AS || CONFIG_IOSCHED_DEADLINE || CONFIG_IOSCHED_NOOP */

/*
#ifndef _ETCPIP
	request_queue_t *blk_alloc_queue(int gfp_mask)
 #else
*/
	request_queue_t *blk_alloc_queue(void)
/*
#endif
*/
{
 /*
 request_queue_t *q = kmalloc(sizeof(*q), gfp_mask);
 */
 request_queue_t *q ;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_alloc_queue() Begin .. \n"));

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] blk_alloc_queue() : q = UsbAllocMemC() , size=%d .. \n" ,
		sizeof(request_queue_t)
		));

	q = (request_queue_t *) UsbAllocMemC(USB_MEM_C_M
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(request_queue_t) // size : 148
		        #endif
           		) ;
  #else
	q = &global_request_queue ;
 #endif // NO_ELEVATOR

 USBMS_DBG_INIT(iprintf("[USBMS INIT] blk_alloc_queue() : q=%08XH \n",q)) ;
 if (!q)
		return NULL;
 /*
 memset(q, 0, sizeof(*q));
 */
 #ifndef NO_TIMER
	init_timer(&q->unplug_timer);
 #endif // ndef NO_TIMER
 atomic_set(&q->refcnt, 1);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_alloc_queue() End. \n"));
 return q;
}
/*
EXPORT_SYMBOL(blk_alloc_queue);
*/
/**
 * blk_init_queue  - prepare a request queue for use with a block device
 * @rfn:  The function to be called to process requests that have been
 *        placed on the queue.
 * @lock: Request queue spin lock
 *
 * Description:
 *    If a block device wishes to use the standard request handling procedures,
 *    which sorts requests and coalesces adjacent requests, then it must
 *    call blk_init_queue().  The function @rfn will be called when there
 *    are requests on the queue that need to be processed.  If the device
 *    supports plugging, then @rfn may not be called immediately when requests
 *    are available on the queue, but may be called at some time later instead.
 *    Plugged queues are generally unplugged when a buffer belonging to one
 *    of the requests on the queue is needed, or due to memory pressure.
 *
 *    @rfn is not required, or even expected, to remove all requests off the
 *    queue, but only as many as it can handle at a time.  If it does leave
 *    requests on the queue, it is responsible for arranging that the requests
 *    get dealt with eventually.
 *
 *    The queue spin lock must be held while manipulating the requests on the
 *    request queue.
 *
 *    Function returns a pointer to the initialized request queue, or NULL if
 *    it didn't succeed.
 *
 * Note:
 *    blk_init_queue() must be paired with a blk_cleanup_queue() call
 *    when the block device is deactivated (such as at module unload).
 **/
request_queue_t *blk_init_queue(request_fn_proc *rfn
	/*
	#ifndef NO_SPINLOCK
		, spinlock_t *lock
	#endif // ndef NO_SPINLOCK
	*/
)
{
 request_queue_t *q;
 static int printed;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_init_queue() Begin .. \n"));

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_init_queue() : blk_alloc_queue() .. \n"));
 /*
 #ifndef _ETCPIP
	q = blk_alloc_queue(GFP_KERNEL);
 #else
 */
 q = blk_alloc_queue();
 /*
 #endif
 */
 USBMS_DBG_INIT(iprintf("[USBMS INIT] blk_init_queue() : q=%08XH \n",q)) ;
 if (!q)
		return NULL;

 if (blk_init_free_list(q))
		goto out_init;

 if (!printed) {
	printed = 1;
	/*
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		printk("Using %s io scheduler\n", chosen_elevator->elevator_name);
	#endif // ndef NO_ELEVATOR
	*/
	USBMS_DBG_INFO(iprintf("[USBMS INFO] blk_init_queue() : Using io scheduler \n"));
   }
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if (elevator_init(q, chosen_elevator))
		goto out_elv;
 #endif // ndef NO_ELEVATOR 
 */
 USBMS_DBG_INIT(iprintf("[USBMS INIT] blk_init_queue() : q->request_fn = rfn .. \n")) ;
 q->request_fn		= rfn;
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	q->back_merge_fn       	= ll_back_merge_fn;
	q->front_merge_fn      	= ll_front_merge_fn;
	q->merge_requests_fn	= ll_merge_requests_fn;
 #endif // ndef NO_ELEVATOR 
 */
 q->prep_rq_fn		= NULL;
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	q->unplug_fn		= generic_unplug_device;
 #endif // ndef NO_ELEVATOR 
 */
 q->queue_flags		= (1 << QUEUE_FLAG_CLUSTER);
 /*
 #ifndef NO_SPINLOCK
	q->queue_lock		= lock;
 #endif // ndef NO_SPINLOCK
 */
 blk_queue_segment_boundary(q, 0xffffffff);

 // #ifndef USB_NO_VMM // vitrtul memory management
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_init_queue() : blk_queue_make_request() .. \n")) ;
 blk_queue_make_request(q, __make_request);
 // #endif // ndef USB_NO_VMM
 /*
 #ifdef _ETCPIP
	rq->queuelist = &rq_queuelist ;
 #endif // _ETCPIP
 */
 blk_queue_max_segment_size(q, MAX_SEGMENT_SIZE);

 blk_queue_max_hw_segments(q, MAX_HW_SEGMENTS);
 blk_queue_max_phys_segments(q, MAX_PHYS_SEGMENTS);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_init_queue() End. \n"));
 return q;

 out_elv:
	blk_cleanup_queue(q);
 out_init:
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] blk_init_queue() : UsbFreeMemC(q) .. \n"
		));
	// kfree(q);
	UsbFreeMemC(USB_MEM_C_M,(unsigned char *) q
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(request_queue_t) // size : 148
					#endif
					);
 #endif // ndef NO_ELEVATOR
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_init_queue() End with Fail !! \n"));
 return NULL;
}

/*
EXPORT_SYMBOL(blk_init_queue);
*/

int blk_get_queue(request_queue_t *q)
{
	if (!test_bit(QUEUE_FLAG_DEAD, &q->queue_flags)) {
		atomic_inc(&q->refcnt);
		return 0;
	}

	return 1;
}
/*
EXPORT_SYMBOL(blk_get_queue);
*/
static inline void blk_free_request(request_queue_t *q, struct request *rq)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_free_request() Begin .. \n"));
 
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	elv_put_request(q, rq);
	/* 
	#ifndef USB_NO_VMM // no vitrtul memory management
		// mempool_t is on include/linux/mempool.h
		mempool_free(rq, q->rq.rq_pool);
	#endif // ndef USB_NO_VMM
	*/
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] blk_free_request() : UsbFreeMemC(rq) , size=%d .. \n",
		sizeof(struct request)
		));
	UsbFreeMemC(USB_MEM_C_M,(unsigned char *) rq
			#if defined(USB_DBG_MEM_C_ON)
				, sizeof(struct request)  // size : 140
			#endif
			) ;
 #endif // ndef NO_ELEVATOR 
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_free_request() End. \n"));
}

static inline struct request *blk_alloc_request(request_queue_t *q,int gfp_mask)
{
 /*
 #ifndef USB_NO_VMM // no vitrtul memory management
 // mempool_t is on include/linux/mempool.h
 struct request *rq = mempool_alloc(q->rq.rq_pool, gfp_mask);
 #endif // ndef USB_NO_VMM
 */
 struct request *rq ;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_alloc_request() Begin .. \n"));

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] blk_alloc_request() : rq = UsbAllocMemC() , size=%d .. \n",
		sizeof(struct request)
		));

	rq = (struct request *) UsbAllocMemC(USB_MEM_C_M
			#ifdef USB_DBG_MEM_C_ON
 		  		,sizeof(struct request)  // size : 140
			#endif
			) ;
 #else
	rq = global_rq ;
 #endif // NO_ELEVATOR
 
 if(!rq)
   {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_alloc_request() End with Failed !! \n"));
	return NULL;
   }

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if (!elv_set_request(q, rq, gfp_mask))
			return rq;
	/*
	#ifndef USB_NO_VMM // no vitrtul memory management
	// mempool_t is on include/linux/mempool.h
	mempool_free(rq, q->rq.rq_pool);
	#endif // ndef USB_NO_VMM
	*/
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] blk_alloc_request() : UsbFreeMemC(rq) , size=%d .. \n",
		sizeof(struct request)
		));
	UsbFreeMemC(USB_MEM_C_M,(unsigned char *) rq
			#if defined(USB_DBG_MEM_C_ON)
				, sizeof(struct request)  // size : 140
			#endif
			) ;

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_alloc_request() End with Failed !! \n"));
	return NULL;
  #else
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_alloc_request() End. \n"));
	return rq;
 #endif // ndef NO_ELEVATOR 
}

/*
 * ioc_batching returns true if the ioc is a valid batching request and
 * should be given priority access to a request.
 */
/*
#ifndef _ETCPIP
static inline int ioc_batching(struct io_context *ioc)
{
	if (!ioc)
		return 0;

	
 //	 * Make sure the process is able to allocate at least 1 request
 //	 * even if the batch times out, otherwise we could theoretically
 //	 * lose wakeups.
	 
 #ifndef NO_TIMER
	return ioc->nr_batch_requests == BLK_BATCH_REQ ||
		(ioc->nr_batch_requests > 0
		&& time_before(jiffies, ioc->last_waited + BLK_BATCH_TIME));
 #endif // ndef NO_TIMER
}
#endif // ndef _ETCPIP
*/

/*
 * ioc_set_batching sets ioc to be a new "batcher" if it is not one. This
 * will cause the process to be a "batcher" on all queues in the system. This
 * is the behaviour we want though - once it gets a wakeup it should be given
 * a nice run.
 */
/*
#ifndef _ETCPIP
void ioc_set_batching(struct io_context *ioc)
{
	if (!ioc || ioc_batching(ioc))
		return;

	ioc->nr_batch_requests = BLK_BATCH_REQ;
 #ifndef NO_TIMER
	ioc->last_waited = jiffies;
 #endif // ndef NO_TIMER
}
#endif // ndef _ETCPIP
*/

/*
 * A request has just been released.  Account for it, update the full and
 * congestion status, wake up any waiters.   Called under q->queue_lock.
 */
static void freed_request(request_queue_t *q, int rw)
{
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	struct request_list *rl = &q->rq;
 #endif // ndef NO_ELEVATOR

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	rl->count[rw]--;
 #endif // ndef NO_ELEVATOR

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if (rl->count[rw] < queue_congestion_off_threshold(q))
 #endif // ndef NO_ELEVATOR
	   {
		clear_queue_congested(q, rw);
	   }

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if (rl->count[rw]+1 <= q->nr_requests) {
		/*
		#ifndef _ETCPIP
			smp_mb();
		#endif // ndef _ETCPIP
		*/
		/*
		#ifndef NO_WAITQUEUE
			if (waitqueue_active(&rl->wait[rw]))
				wake_up(&rl->wait[rw]);
			if (!waitqueue_active(&rl->wait[rw]))
				blk_clear_queue_full(q, rw);
		#endif // ndef NO_WAITQUEUE
		*/
	   }
 #endif // ndef NO_ELEVATOR
}

#define blkdev_free_rq(list) list_entry((list)->next, struct request, queuelist)
/*
 * Get a free request, queue_lock must not be held
 */
static struct request *get_request(request_queue_t *q, int rw, int gfp_mask)
{
 /*
 struct request *rq = NULL;
 struct request_list *rl = &q->rq;
 #ifndef _ETCPIP
 struct io_context *ioc = get_io_context(gfp_mask);
 #endif // ndef _ETCPIP
 */
 struct request *rq ;
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	struct request_list *rl ;
 #endif // ndef NO_ELEVATOR
 /*
 #ifndef _ETCPIP
	struct io_context *ioc ;
 #endif // ndef _ETCPIP
 */


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] get_request() Begin .. \n"));
 rq = NULL;
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	rl = &q->rq;
 #endif // ndef NO_ELEVATOR
 /*
 #ifndef _ETCPIP
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] get_request() : get_io_context() .. \n"));
	ioc = get_io_context(gfp_mask);
 #endif // ndef _ETCPIP
 */
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irq(q->queue_lock);
 #endif // ndef NO_SPINLOCK
 */
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
 if(rl->count[rw]+1 >= q->nr_requests) {
	/*
	 * The queue will fill after this allocation, so set it as
	 * full, and mark this process as "batching". This process
	 * will be allowed to complete a batch of requests, others
	 * will be blocked.
	 */
	USBMS_DBG_INFO(iprintf("[USBMS INFO] get_request() : Error : Queue Full !! \n")) ;
	/*
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if (!blk_queue_full(q, rw)) {
		#ifndef _ETCPIP
			ioc_set_batching(ioc);
		#endif // ndef _ETCPIP
		blk_set_queue_full(q, rw);
	   }
	#endif // ndef NO_ELEVATOR
	*/
   }
 #endif // ndef NO_ELEVATOR
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
 if (blk_queue_full(q, rw)
	
	#ifndef _ETCPIP
			&& !ioc_batching(ioc)
	#endif // ndef _ETCPIP
	
	
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		 && !elv_may_queue(q, rw)
	#endif // ndef NO_ELEVATOR 
	
   ) {
	//  * The queue is full and the allocating process is not a
	// * "batcher", and not exempted by the IO scheduler
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] get_request() : Error : blk_queue_full() !! \n"
		));
	#ifndef NO_SPINLOCK
		spin_unlock_irq(q->queue_lock);
	#endif // ndef NO_SPINLOCK
	goto out;
   }
 #endif // ndef NO_ELEVATOR
 */

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	rl->count[rw]++;
	if (rl->count[rw] >= queue_congestion_on_threshold(q))
		set_queue_congested(q, rw);
 #endif // ndef NO_ELEVATOR
 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irq(q->queue_lock);
 #endif // ndef NO_SPINLOCK
 */

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] get_request() : blk_alloc_request() .. \n"));
 rq = blk_alloc_request(q, gfp_mask);
 if (!rq) {
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] get_request() : Error : blk_alloc_request() Failed !! \n"
		));
	/*
	 * Allocation failed presumably due to memory. Undo anything
	 * we might have messed up.
	 *
	 * Allocating task should really be put onto the front of the
	 * wait queue, but this is pretty rare.
	 */
	/*
	#ifndef NO_SPINLOCK
		spin_lock_irq(q->queue_lock);
	#endif // ndef NO_SPINLOCK
	*/
	freed_request(q, rw);
	/*
	#ifndef NO_SPINLOCK
		spin_unlock_irq(q->queue_lock);
	#endif // ndef NO_SPINLOCK
	*/
	goto out;
   }

 /*
 #ifndef _ETCPIP
 if (ioc_batching(ioc))
		ioc->nr_batch_requests--;
 #endif // ndef _ETCPIP
 */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	// #ifndef _ETCPIP
	USBMS_DBG_INIT(iprintf("[USBMS INIT] get_request() : INIT_LIST_HEAD(&rq->queuelist) .. \n"));
	INIT_LIST_HEAD(&rq->queuelist);
	//#else
	//	rq->queuelist = &rq_queuelist ;
	//#endif
 #endif // ndef NO_ELEVATOR
 */
 /*
  * first three bits are identical in rq->flags and bio->bi_rw,
  * see bio.h and blkdev.h
  */
 rq->flags = rw;

 rq->errors = 0;
 rq->rq_status = RQ_ACTIVE;
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	rq->bio = rq->biotail = NULL;
  #else
	rq->bio = NULL;
 #endif // NO_ELEVATOR 
 // #endif // ndef USB_NO_VMM
 rq->buffer = NULL;
 rq->ref_count = 1;
 rq->q = q;
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	rq->rl = rl;
 #endif // ndef NO_ELEVATOR
 /*
 #ifndef _ETCPIP
    // struct completion is in <linux/completion.h>
	rq->waiting = NULL;
  #else
 */
	// 2003/12/02 B. S. Liaw :
	// req->waiting replaced by MT_EVT_USBMS_RQ/MT_EVT_USBMS_RQBI on kernel/local.h
	rq->waiting_evt = (uint)NULL_VALUE ;
 /*
 #endif
 */
 rq->special = NULL;
 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] get_request() : Set Value : rq->special=%08XH \n",rq->special
		)) ;
 rq->data = NULL;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] get_request() : Set Value : rq->data=%08XH \n",
	rq->data
	));
 rq->sense = NULL;

 out:
 /*
 #ifndef _ETCPIP
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] get_request() : put_io_context() .. \n"));
 put_io_context(ioc);
 #endif // ndef _ETCPIP
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] get_request() End. \n"));
 return rq;
}

/*
 * No available requests for this queue, unplug the device and wait for some
 * requests to become available.
 */
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
   static struct request *get_request_wait(request_queue_t *q, int rw)
   {
	DEFINE_WAIT(wait);
	struct request *rq;

	generic_unplug_device(q);
	do {
		struct request_list *rl = &q->rq;

		prepare_to_wait_exclusive(&rl->wait[rw], &wait,
				TASK_UNINTERRUPTIBLE);

		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] get_request_wait() : get_request() .. \n"));
		rq = get_request(q, rw, GFP_NOIO);

		if (!rq) {
			struct io_context *ioc;

			io_schedule();

			//
			// * After sleeping, we become a "batching" process and
			// * will be able to allocate at least one request, and
			// * up to a big batch of them for a small period time.
			// * See ioc_batching, ioc_set_batching
			

		#ifndef _ETCPIP
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] get_request_wait() : get_io_context() .. \n"));
			ioc = get_io_context(GFP_NOIO);
			ioc_set_batching(ioc);
			put_io_context(ioc);
		#endif // ndef _ETCPIP
		}
		finish_wait(&rl->wait[rw], &wait);
	} while (!rq);

	return rq;
}
#endif // ndef NO_ELEVATOR
*/

struct request *blk_get_request(request_queue_t *q, int rw, int gfp_mask)
{
 struct request *rq;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_get_request() Begin .. \n"));
 /*
 BUG_ON(rw != READ && rw != WRITE);
 */
 if (rw != READ && rw != WRITE) {
		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if (gfp_mask & __GFP_WAIT)
	   {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_get_request() : get_request_wait() .. \n"));
		rq = get_request_wait(q, rw);
	   }
	else
 #endif // ndef NO_ELEVATOR
	   {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_get_request() : get_request() .. \n"));
		rq = get_request(q, rw, gfp_mask);
	   }

 return rq;
}
/*
EXPORT_SYMBOL(blk_get_request);
*/
/**
 * blk_requeue_request - put a request back on queue
 * @q:		request queue where request should be inserted
 * @rq:		request to be inserted
 *
 * Description:
 *    Drivers often keep queueing requests until the hardware cannot accept
 *    more, when that condition happens we need to put the request back
 *    on the queue. Must be called with queue lock held.
 */
void blk_requeue_request(request_queue_t *q, struct request *rq)
{
 if (blk_rq_tagged(rq))
	blk_queue_end_tag(q, rq);
 
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	elv_requeue_request(q, rq);
 #endif // ndef NO_ELEVATOR 
 
}
/*
EXPORT_SYMBOL(blk_requeue_request);
*/
/**
 * blk_insert_request - insert a special request in to a request queue
 * @q:		request queue where request should be inserted
 * @rq:		request to be inserted
 * @at_head:	insert request at head or tail of queue
 * @data:	private data
 * @reinsert:	true if request it a reinsertion of previously processed one
 *
 * Description:
 *    Many block devices need to execute commands asynchronously, so they don't
 *    block the whole kernel from preemption during request execution.  This is
 *    accomplished normally by inserting aritficial requests tagged as
 *    REQ_SPECIAL in to the corresponding request queue, and letting them be
 *    scheduled for actual execution by the request queue.
 *
 *    We have the option of inserting the head or the tail of the queue.
 *    Typically we use the tail for new ioctls and so forth.  We use the head
 *    of the queue for things like a QUEUE_FULL message from a device, or a
 *    host that is unable to accept a particular command.
 */
/*
void blk_insert_request(request_queue_t *q, struct request *rq,
*/
int blk_insert_request(request_queue_t *q, struct request *rq,
			int at_head, void *data, int reinsert)
{
 unsigned long flags;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_insert_request() Begin .. \n"));
 /*
  * tell I/O scheduler that this isn't a regular read/write (ie it
  * must not attempt merges on this) and that it acts as a soft
  * barrier
  */
 rq->flags |= REQ_SPECIAL | REQ_SOFTBARRIER;

 rq->special = data;
 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] blk_insert_request() : Set Value : rq->special=%08XH \n",rq->special
		)) ;
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(q->queue_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 // #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_INFO(iprintf("[USBMS INFO] blk_insert_request() : q->req = rq .. \n"));
	q->req = rq ;
 // #endif // NO_ELEVATOR

 /*
  * If command is tagged, release the tag
  */
 if(reinsert) {
		blk_requeue_request(q, rq);
	} else {
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		int where = ELEVATOR_INSERT_BACK;

		if (at_head)
			where = ELEVATOR_INSERT_FRONT;
	#endif // ndef NO_ELEVATOR 

	if (blk_rq_tagged(rq))
			blk_queue_end_tag(q, rq);

	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		drive_stat_acct(rq, rq->nr_sectors, 1);
     #else
		drive_stat_acct(rq, 1, 1);
    #endif
	
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		// __elv_add_request() is on include/linux/elevator.h
		__elv_add_request(q, rq, where, 0);
	 #else
		USBMS_DBG_INFO(iprintf("[USBMS INFO] blk_insert_request() : Bug on req !! \n"));

		USBMS_DBG_MEM(iprintf("[USBMS MEM] blk_insert_request() : q->req=%08XH \n",
			q->req)) ;

		#ifdef USBMS_BS_ASSERT
			if (!(   (unsigned long) q->req
			   > (unsigned long) 0x200000L
    	     )
		   )
		   {
			iprintf("[ASSERT][USBMS] blk_insert_request() : " ) ;
		    iprintf("Error : Invlid address q->req=%08XH \n",
				q->req ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
		#endif // USBMS_BS_ASSERT

		rq = q->req;
		
		//if ((rq->flags & REQ_DONTPREP) || !q->prep_rq_fn)
		//	break;
		
		/*
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_insert_request(1) : q->prep_rq_fn() .. \n"));
		
		// ret = q->prep_rq_fn(q, rq);
		// if (ret == BLKPREP_OK) {
		
		if (q->prep_rq_fn(q, rq) == BLKPREP_OK) {
			// break;
		   } 
		
		
		//  else if (ret == BLKPREP_DEFER) {
		//	req = NULL;
		//	break;
		//   }
		
		//  else if (ret == BLKPREP_KILL) {
		//	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_insert_request(2) : blkdev_dequeue_request() .. \n"));
		//	blkdev_dequeue_request(rq);
		//	rq->flags |= REQ_QUIET;
		//	while (end_that_request_first(req, 0, rq->nr_sectors))
		//		;
		//	end_that_request_last(rq);
		//   }
		
		 else {
			// printk("%s: bad return=%d\n", __FUNCTION__, ret);
			printk("%s: q->prep_rq_fn() Failed !! \n", __FUNCTION__);
			// break;
		   }
		*/
	#endif // NO_ELEVATOR 
   }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_insert_request(3) : q->request_fn() .. \n"));
 if( ! q->request_fn(q) ) {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] blk_insert_request() : q->request_fn() Failed !! \n"));
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_insert_request() End with Failed !! \n"));
	return 0 ;
   }

 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(q->queue_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_insert_request() End. \n"));
 return 1 ;
}
/*
EXPORT_SYMBOL(blk_insert_request);
*/
void drive_stat_acct(struct request *rq, int nr_sectors, int new_io)
{
	int rw = rq_data_dir(rq);

	if (!blk_fs_request(rq) || !rq->rq_disk)
		return;

	if (rw == READ) {
		disk_stat_add(rq->rq_disk, read_sectors, nr_sectors);
		if (!new_io)
			disk_stat_inc(rq->rq_disk, read_merges);
	} else if (rw == WRITE) {
		disk_stat_add(rq->rq_disk, write_sectors, nr_sectors);
		if (!new_io)
			disk_stat_inc(rq->rq_disk, write_merges);
	}
	if (new_io) {
		disk_round_stats(rq->rq_disk);
		rq->rq_disk->in_flight++;
	}
}

/*
 * add-request adds a request to the linked list.
 * queue lock is held and interrupts disabled, as we muck with the
 * request queue list.
 */
static inline void add_request(request_queue_t * q, struct request * req)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] add_request() Begin .. \n"));

 #ifdef USBMS_BS_ASSERT
	if (!(   (unsigned long) req
		   > (unsigned long) 0x200000L
         )
	   )
	   {
		iprintf("[ASSERT][USBMS] add_request() : " ) ;
	    iprintf("Error : Invlid address req=%08XH \n",
				req ) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	drive_stat_acct(req, req->nr_sectors, 1);
  #else
	drive_stat_acct(req, 1, 1);
 #endif

	if (q->activity_fn)
		q->activity_fn(q->activity_data, rq_data_dir(req));

	/*
	 * elevator indicated where it wants this request to be
	 * inserted at elevator_merge time
	 */
 
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	// __elv_add_request() is on include/linux/elevator.h
	__elv_add_request(q, req, ELEVATOR_INSERT_SORT, 0);
  #else
	/*
	// USBMS_DBG_INFO(iprintf("[USBMS INFO] add_request() : Bug on req !! \n"));

	//USBMS_DBG_MEM(iprintf("[USBMS MEM] add_request() : q->req=%08XH \n",
	//		q->req)) ;
	
	#ifdef USBMS_BS_ASSERT
			if (!(   (unsigned long) q->req
			   > (unsigned long) 0x200000L
    	     )
		   )
		   {
			iprintf("[ASSERT][USBMS] add_request() : " ) ;
		    iprintf("Error : Invlid address q->req=%08XH \n",
				q->req ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	#endif // USBMS_BS_ASSERT
	
	// req = q->req;
	
	q->req = req;
	USBMS_DBG_MEM(iprintf("[USBMS MEM] add_request() : Set Value : q->req=%08XH \n",
			q->req)) ;
	
	//if ((rq->flags & REQ_DONTPREP) || !q->prep_rq_fn)
	//		break;
	
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] add_request() : q->prep_rq_fn() .. \n"));
	
	//ret = q->prep_rq_fn(q, req);
	//if (ret == BLKPREP_OK) {
	
	if (q->prep_rq_fn(q, req) == BLKPREP_OK) {
		// break;
	   } 
	
	//  else if (ret == BLKPREP_DEFER) {
	//	req = NULL;
	//	break;
	//   }
		
	//  else if (ret == BLKPREP_KILL) {
	//		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] add_request() : blkdev_dequeue_request() .. \n"));
	//		blkdev_dequeue_request(req);
	//		rq->flags |= REQ_QUIET;
	//		while (end_that_request_first(req, 0, rq->nr_sectors))
	//			;
	//		end_that_request_last(req);
	//	   }
	
	 else {
			// printk("%s: bad return=%d\n", __FUNCTION__, ret);
			dprintf("[USBMS] Error : %s: q->prep_rq_fn() Failed !! \n", __FUNCTION__);
			// break;
		   }
	*/
 #endif // NO_ELEVATOR 

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] add_request() End. \n"));
}
 
/*
 * disk_round_stats()	- Round off the performance stats on a struct
 * disk_stats.
 *
 * The average IO queue length and utilisation statistics are maintained
 * by observing the current state of the queue length and the amount of
 * time it has been in this state for.
 *
 * Normally, that accounting is done on IO completion, but that can result
 * in more than a second's worth of IO being accounted for within any one
 * second, leading to >100% utilisation.  To deal with that, we call this
 * function to do a round-off before returning the results when reading
 * /proc/diskstats.  This accounts immediately for all queue usage up to
 * the current jiffies and restarts the counters again.
 */
void disk_round_stats(struct gendisk *disk)
{
 #ifndef NO_TIMER
	unsigned long now = jiffies;
 #endif // ndef NO_TIMER

 #ifndef NO_TIMER
	disk_stat_add(disk, time_in_queue, 
			disk->in_flight * (now - disk->stamp));
	disk->stamp = now;

	if (disk->in_flight)
		disk_stat_add(disk, io_ticks, (now - disk->stamp_idle));
	disk->stamp_idle = now;
 #endif // ndef NO_TIMER
}

/*
 * queue lock must be held
 */
void __blk_put_request(request_queue_t *q, struct request *req)
{
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	struct request_list *rl = req->rl;
 #endif // ndef NO_ELEVATOR
 /*
	if (unlikely(!q))
 */
 if (!q)
		return;
 /*
	if (unlikely(--req->ref_count))
 */
 if (--req->ref_count)
		return;
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	elv_completed_request(req->q, req);
 #endif // ndef NO_ELEVATOR 
 */
 req->rq_status = RQ_INACTIVE;
 req->q = NULL;
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	req->rl = NULL;
 #endif // ndef NO_ELEVATOR

	/*
	 * Request may not have originated from ll_rw_blk. if not,
	 * it didn't come out of our reserved rq pools
	 */
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
 if (rl) 
 #endif // ndef NO_ELEVATOR
	{
	int rw = rq_data_dir(req);

	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		/*
		BUG_ON(!list_empty(&req->queuelist));
		*/
		/*
		 #ifndef _ETCPIP
		 */
		 if (!list_empty(&req->queuelist)) {
		 /*
		 #else
		 if (!((!list_empty(req->queuelist))!=0)) {
 
		 #endif
		 */
		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }
	#endif // ndef NO_ELEVATOR

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __blk_put_request() : blk_free_request() .. \n"));
	blk_free_request(q, req);
	freed_request(q, rw);
   }
}

void blk_put_request(struct request *req)
{
 request_queue_t *q = req->q;
	
	/*
	 * if req->q isn't set, this request didnt originate from the
	 * block layer, so it's safe to just disregard it
	 */
 if (q) {
		unsigned long flags;

	#ifndef NO_SPINLOCK
		spin_lock_irqsave(q->queue_lock, flags);
	#endif // ndef NO_SPINLOCK
		__blk_put_request(q, req);
	#ifndef NO_SPINLOCK
		spin_unlock_irqrestore(q->queue_lock, flags);
	#endif // ndef NO_SPINLOCK
   }
}
/*
EXPORT_SYMBOL(blk_put_request);
*/
/**
 * blk_congestion_wait - wait for a queue to become uncongested
 * @rw: READ or WRITE
 * @timeout: timeout in jiffies
 *
 * Waits for up to @timeout jiffies for a queue (any queue) to exit congestion.
 * If no queues are congested then just wait for the next request to be
 * returned.
 */
/*
#ifndef _ETCPIP
void blk_congestion_wait(int rw, long timeout)
{
	DEFINE_WAIT(wait);
 #ifndef NO_WAITQUEUE
	wait_queue_head_t *wqh = &congestion_wqh[rw];
 #endif // ndef NO_WAITQUEUE

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_congestion_wait() : blk_run_queues() .. \n"));
 blk_run_queues();
 #ifndef NO_WAITQUEUE
	prepare_to_wait(wqh, &wait, TASK_UNINTERRUPTIBLE);
 #endif // ndef NO_WAITQUEUE
	io_schedule_timeout(timeout);
 #ifndef NO_WAITQUEUE
	finish_wait(wqh, &wait);
 #endif // ndef NO_WAITQUEUE
}
#endif // ndef _ETCPIP
*/
/*
EXPORT_SYMBOL(blk_congestion_wait);
*/

/*
 * Has to be called with the request spinlock acquired
 */
static int attempt_merge(request_queue_t *q, struct request *req,
			  struct request *next)
{
 if (!rq_mergeable(req) || !rq_mergeable(next))
		return 0;

 /*
  * not contigious
  */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if (req->sector + req->nr_sectors != next->sector)
		return 0;
 #endif // ndef NO_ELEVATOR
 */
	if (rq_data_dir(req) != rq_data_dir(next)
	    || req->rq_disk != next->rq_disk
 /*
 #ifndef _ETCPIP
    // struct completion is in <linux/completion.h>
	    || next->waiting 
  #else
 */
		// 2003/12/02 B. S. Liaw :
		// req->waiting replaced by MT_EVT_USBMS_RQ/MT_EVT_USBMS_RQBI on kernel/local.h
		|| next->waiting_evt
 /*
 #endif // ndef _ETCPIP
 */
        || next->special)
		return 0;

	/*
	 * If we are allowed to merge, then append bio list
	 * from next to rq and release next. merge_requests_fn
	 * will have updated segment counts, update sector
	 * counts here.
 	 */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if (!q->merge_requests_fn(q, req, next))
		return 0;
 #endif // ndef NO_ELEVATOR 
 */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	req->biotail->bi_next = next->bio;
	req->biotail = next->biotail;
 #endif // ndef NO_ELEVATOR 
 */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	req->nr_sectors = req->hard_nr_sectors += next->hard_nr_sectors;
  #else
 */
	req->hard_nr_sectors += next->hard_nr_sectors;
 /*
 #endif
 */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		elv_merge_requests(q, req, next);
 #endif // ndef NO_ELEVATOR 
 */
 if (req->rq_disk) {
		disk_round_stats(req->rq_disk);
		req->rq_disk->in_flight--;
   }

 __blk_put_request(q, next);
 return 1;
}

static inline int attempt_back_merge(request_queue_t *q, struct request *rq)
{
 
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	struct request *next = elv_latter_request(q, rq);

	if (next)
		return attempt_merge(q, rq, next);
 #endif // ndef NO_ELEVATOR 
 return 0;
}

static inline int attempt_front_merge(request_queue_t *q, struct request *rq)
{
 
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	struct request *prev = elv_former_request(q, rq);

	if (prev)
		return attempt_merge(q, prev, rq);
 #endif // ndef NO_ELEVATOR 
 
 return 0;
}

/**
 * blk_attempt_remerge  - attempt to remerge active head with next request
 * @q:    The &request_queue_t belonging to the device
 * @rq:   The head request (usually)
 *
 * Description:
 *    For head-active devices, the queue can easily be unplugged so quickly
 *    that proper merging is not done on the front request. This may hurt
 *    performance greatly for some devices. The block layer cannot safely
 *    do merging on that first request for these queues, but the driver can
 *    call this function and make it happen any way. Only the driver knows
 *    when it is safe to do so.
 **/
void blk_attempt_remerge(request_queue_t *q, struct request *rq)
{
	unsigned long flags;

 #ifndef NO_SPINLOCK
	spin_lock_irqsave(q->queue_lock, flags);
 #endif // ndef NO_SPINLOCK
	attempt_back_merge(q, rq);
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(q->queue_lock, flags);
 #endif // ndef NO_SPINLOCK
}
/*
EXPORT_SYMBOL(blk_attempt_remerge);
*/
/*
 * Non-locking blk_attempt_remerge variant.
 */
void __blk_attempt_remerge(request_queue_t *q, struct request *rq)
{
 attempt_back_merge(q, rq);
}
/*
EXPORT_SYMBOL(__blk_attempt_remerge);
*/
// #ifndef USB_NO_VMM // vitrtul memory management
// void blk_queue_make_request(request_queue_t * q, make_request_fn * mfn)
// B. S. Liaw : blk_queue_make_request(q, __make_request);
// B. S. Liaw : q->make_request_fn = mfn;
static int __make_request(request_queue_t *q, struct bio *bio)
{
 struct request *req, *freereq = NULL;
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	int el_ret, rw, nr_sectors, cur_nr_sectors, barrier, ra;
  #else
 */
	int rw, nr_sectors, cur_nr_sectors, barrier, ra;
 /*
 #endif
 */
 sector_t sector;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __make_request() Begin .. \n"));
 sector = bio->bi_sector;
 nr_sectors = bio_sectors(bio);

 /* 
 // #if (USBMS_RW_MAX_SECTORS >= 8)
 #if (USBMS_RD_SECTORS >= 8)
	if (bio->bi_rw == READ) {
		// nr_sectors = USBMS_RW_MAX_SECTORS ; // >= 8 
		nr_sectors = USBMS_RD_SECTORS ;
	   }
 #endif // USBMS_RD_SECTORS >= 8
 */

 cur_nr_sectors = bio_cur_sectors(bio);

 rw = bio_data_dir(bio);

 /*
  * low level driver can indicate that it wants pages above a
  * certain limit bounced to low memory (ie for highmem, or even
  * ISA dma in theory)
  */
 /*
 #ifdef CONFIG_HIGHMEM
	blk_queue_bounce(q, &bio);
 #endif // CONFIG_HIGHMEM
 */
 /*
 #ifndef NO_SPINLOCK
	spin_lock_prefetch(q->queue_lock);
 #endif // ndef NO_SPINLOCK
 */
 barrier = test_bit(BIO_RW_BARRIER, &bio->bi_rw);
 
 // #ifndef _ETCPIP
	ra = bio->bi_rw & (1 << BIO_RW_AHEAD);
 // #endif // ndef _ETCPIP
 
 again:
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irq(q->queue_lock);
 #endif // ndef NO_SPINLOCK
 */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
 if (elv_queue_empty(q)) {
 #endif // ndef NO_ELEVATOR 
 */
	/*
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __make_request(1) : blk_plug_device() .. \n"));
		blk_plug_device(q);
	#endif // ndef NO_ELEVATOR
	*/
	goto get_rq;
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
   }
 #endif // ndef NO_ELEVATOR
 */
 if (barrier)
		goto get_rq;
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	el_ret = elv_merge(q, &req, bio);

	switch (el_ret) {
		case ELEVATOR_BACK_MERGE:
			
			// BUG_ON(!rq_mergeable(req));
			
		 if (!rq_mergeable(req)) {
				dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
			   }

		 // #ifndef USB_NO_VMM // vitrtul memory management
			if (!q->back_merge_fn(q, req, bio))
				break;
		 // #endif // ndef USB_NO_VMM

			req->biotail->bi_next = bio;
			req->biotail = bio;
			req->nr_sectors = req->hard_nr_sectors += nr_sectors;
			drive_stat_acct(req, nr_sectors, 0);
			if (!attempt_back_merge(q, req))
			   {
				// #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
					elv_merged_request(q, req);
				// #endif // ndef NO_ELEVATOR 
			   }
			goto out;

		case ELEVATOR_FRONT_MERGE:
			
			// BUG_ON(!rq_mergeable(req));
			
			if (!rq_mergeable(req)) {
				dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
			   }

			if (!q->front_merge_fn(q, req, bio))
				break;

		// #ifndef USB_NO_VMM // vitrtul memory management
			bio->bi_next = req->bio;
			req->cbio = req->bio = bio;
			req->nr_cbio_segments = bio_segments(bio);
			req->nr_cbio_sectors = bio_sectors(bio);
		// #endif // ndef USB_NO_VMM

			
			// may not be valid. if the low level driver said
			// it didn't need a bounce buffer then it better
			// not touch req->buffer either...
			 
	//	#ifndef USB_NO_VMM // vitrtul memory management
		// struct page is on include/linux/mm.h
			req->buffer = bio_data(bio);
	//	#endif // ndef USB_NO_VMM
			req->current_nr_sectors = cur_nr_sectors;
			req->hard_cur_sectors = cur_nr_sectors;
			req->sector = req->hard_sector = sector;
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] __make_request() : Set Value : req->sector=%08XH \n" ,
	req->sector
	));
			req->nr_sectors = req->hard_nr_sectors += nr_sectors;
			drive_stat_acct(req, nr_sectors, 0);
			if (!attempt_front_merge(q, req))
				elv_merged_request(q, req);
			goto out;

		
		//elevator says don't/can't merge. get new request
		 
		case ELEVATOR_NO_MERGE:
			break;

		default:
			
			// printk("elevator returned crap (%d)\n", el_ret);
			
			dprintf("[USBMS] Error : elevator returned crap (%d) !!\n", el_ret);
			BUG();
	   }
 #endif // ndef NO_ELEVATOR 
 */

 /*
  * Grab a free request from the freelist - if that is empty, check
  * if we are doing read ahead and abort instead of blocking for
  * a free slot.
  */
 get_rq:

 if(freereq) {
	req = freereq;
	freereq = NULL;
   } else {
	/*
	#ifndef NO_SPINLOCK
		spin_unlock_irq(q->queue_lock);
	#endif // ndef NO_SPINLOCK
	*/
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __make_request(2) : get_request() .. \n"));
	if ((freereq = get_request(q, rw, GFP_ATOMIC)) == NULL) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] __make_request() : Error : get_request() Failed !! \n"
			));
		/*
		 * READA bit set
		 */
		if (ra)
				goto end_io;
		/*
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			freereq = get_request_wait(q, rw);
		#endif // ndef NO_ELEVATOR
		*/
	   }
	/*
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		goto again;
	 #else
	*/
		req = freereq;
		freereq = NULL;
	/*
	#endif // NO_ELEVATOR 
	*/
   }

 USBMS_DBG_BLOCK(iprintf("[USBMS BLOCK] __make_request() : Set Up Request .. \n"));
 /*
  * first three bits are identical in rq->flags and bio->bi_rw,
  * see bio.h and blkdev.h
  */
 // REQ_CMD : regular fs rw request
 req->flags = (bio->bi_rw & 7) | REQ_CMD;

 /*
  * REQ_BARRIER implies no merging, but lets make it explicit
  */
 if (barrier)
		req->flags |= (REQ_HARDBARRIER | REQ_NOMERGE);

 /*
  * don't stack up retries for read ahead
  */
 if (ra)
		req->flags |= REQ_FAILFAST;

 req->errors = 0;
 req->hard_sector = req->sector = sector;
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] __make_request() : Set Value : req->sector=%08XH \n" ,
	req->sector
	));
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	req->hard_nr_sectors = req->nr_sectors = nr_sectors;
	req->current_nr_sectors = req->hard_cur_sectors = cur_nr_sectors;
  #else
 */
	req->hard_nr_sectors = nr_sectors;
	req->hard_cur_sectors = cur_nr_sectors;
 /*
 #endif
 */
 // #ifndef USB_NO_VMM // vitrtul memory management
 /*
 #ifndef NO_SCATTER_LIST
	req->nr_phys_segments = bio_phys_segments(q, bio);
	req->nr_hw_segments = bio_hw_segments(q, bio);
 #endif // ndef NO_SCATTER_LIST
 */
 req->nr_cbio_segments = bio_segments(bio);
 req->nr_cbio_sectors = bio_sectors(bio);
 req->buffer = bio_data(bio);	/* see ->buffer comment above */
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] __make_request() : Set Value : req->buffer=%08XH \n",
	req->buffer
	));
 // #endif // ndef USB_NO_VMM
 
 // #ifdef NO_SCATTER_LIST
	// B. S. Liaw add this !!
	req->data = bio_data(bio);
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] __make_request() : Set Value : req->data=%08XH .. \n",
		req->data
		));
 // #endif // NO_SCATTER_LIST
 

 /*
 #ifndef _ETCPIP
    // struct completion is in <linux/completion.h>
	req->waiting = NULL;
  #else
 */
	// 2003/12/02 B. S. Liaw :
	// req->waiting replaced by MT_EVT_USBMS_RQ/MT_EVT_USBMS_RQBI on kernel/local.h
	req->waiting_evt = (uint)NULL_VALUE ;
 /*
 #endif // ndef _ETCPIP
 */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	req->cbio = req->bio = req->biotail = bio;
 #else
 */
	req->cbio = req->bio = bio;
 /*
 #endif // NO_ELEVATOR
 */
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] __make_request() : bio->bi_bdev=%08XH \n",
	bio->bi_bdev));
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] __make_request() : bio->bi_bdev->bd_disk=%08XH \n",
	bio->bi_bdev->bd_disk));

 #ifdef USBMS_BS_ASSERT
	if ( ! (IS_VALID_ADDR(bio->bi_bdev->bd_disk)) ) {
		iprintf("[ASSERT][USBMS] __make_request() : " ) ;
	    iprintf("Error : Invalid Address !!\n" ) ;
        iprintf("[ASSERT][USBMS] bio->bi_bdev->bd_disk=%08XH \n",
				bio->bi_bdev->bd_disk ) ; ;
        iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT

 req->rq_disk = bio->bi_bdev->bd_disk;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] __make_request() : Set Value : req->rq_disk=%08XH \n",
	req->rq_disk));
 /*
 #ifndef NO_TIMER
	req->start_time = jiffies;
 #endif // ndef NO_TIMER
 */
 USBMS_DBG_BLOCK(iprintf("[USBMS BLOCK] __make_request() : Add Request .. \n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __make_request(3) : add_request() .. \n"));
 add_request(q, req);
 /*
 #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
 */

#ifndef USBMS_NO_USB_TASK  /* Charles 20040520 */
 #ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
	// B. S. Liaw add this !!
	// 2003/12/12 B. S. Liaw :
	// req->waiting replaced by MT_EVT_USBMS_RQ/MT_EVT_USBMS_RQBI on kernel/local.h
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] __make_request() : clrevt MT_EVT_USBMS_RQ=%d ..\n",
		MT_EVT_USBMS_RQ
		)) ;
    if (clrevt(MT_EVT_USBMS_RQ) != SUCCESS) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] __make_request() : Error : clrevt fail: Invalid MT_EVT_USBMS_RQ=%d\n",
			MT_EVT_USBMS_RQ
			)) ;
        dprintf("[USBMS] Error : clrevt fail: Invalid MT_EVT_USBMS_RQ %d\n",
				MT_EVT_USBMS_RQ);
	   }
  #else
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] __make_request() : Set Value : req->evt_blk_rq=0 .. \n"
		)) ;
	// BLOCK_PREEMPTION;
	req->evt_blk_rq = 0 ;
	// UNBLOCK_PREEMPTION;
 #endif // ndef USBMS_USB_WAIT_BY_EVENT
#endif
 // sreq->sr_request->waiting_evt = MT_EVT_USBMS_RQ ;
 req->waiting_evt = MT_EVT_USBMS_RQ ;

 USBMS_DBG_MEM(iprintf("[USBMS MEM] __make_request() : req=%08XH \n",
	req));
	
 // USBMS_DBG_MEM(iprintf("[USBMS MEM] __make_request() : sreq->sr_request=%08XH \n",
 //	sreq->sr_request));

 // sreq->sr_request->rq_status = RQ_SCSI_BUSY;
 req->rq_status = RQ_SCSI_BUSY;


 q->req = req ;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] __make_request() : q->req=%08XH \n",
	q->req)) ;

 #ifdef USBMS_BS_ASSERT
	if (!(   (unsigned long) q->req
		   > (unsigned long) 0x200000L
	   ) ) {
		iprintf("[ASSERT][USBMS] __make_request() : " ) ;
	    iprintf("Error : Invlid address q->req=%08XH \n",
				q->req ) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT
 /*
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __make_request(4) : q->prep_rq_fn() .. \n"));
 if (q->prep_rq_fn(q, req) == BLKPREP_OK) {
	// break;
   } 
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __make_request(5) : q->request_fn() .. \n"));
//iprintf("__make_request1: %u.%u\n", get_sys_time(), USB_CURRENT_TICKS); 
 if( ! q->request_fn(q) ) {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] __make_request() : q->request_fn() Failed !! \n"));
	goto end_fail ;
   }

 USBMS_DBG_INIT(iprintf("[USBMS INIT] __make_request() *** task sleeping \n")) ;
 USBMS_DBG_BLOCK(iprintf("[USBMS BLOCK] __make_request() *** task sleeping \n")) ;

#ifndef USBMS_NO_USB_TASK  /* Charles 20040520 */
 #ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
		// 2003/12/12 B. S. Liaw :
		// req->waiting replaced by MT_EVT_USBMS_RQ/MT_EVT_USBMS_RQBI on kernel/local.h
	    (void)wteset_dec(MT_EVT_USBMS_RQ, 0); // 0 : no timeout

	    if (clrevt(MT_EVT_USBMS_RQ) != SUCCESS) {
			USBMS_DBG_INFO(iprintf(
				"[USBMS INFO] __make_request() : Error : clrevt fail: Invalid MT_EVT_USBMS_RQ=%d\n",
				MT_EVT_USBMS_RQ
				)) ;
        	dprintf("[USBMS] Error : clrevt fail: Invalid MT_EVT_USBMS_RQ %d\n",
					MT_EVT_USBMS_RQ);
		   }
  #else
	for(;;) {
		// NULL ;
		if (req->evt_blk_rq) {
			// NULL ;
			// break ;
			goto awakened ;
		   }
		// NULL ;
	   }

	// -----------
	awakened:
	// -----------

	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] __make_request() : Set Value : req->evt_blk_rq=0 .. \n"
		)) ;
	// BLOCK_PREEMPTION;
	req->evt_blk_rq = 0 ;
	// UNBLOCK_PREEMPTION;
 #endif // ndef USBMS_USB_WAIT_BY_EVENT
#endif
 USBMS_DBG_BLOCK(iprintf("[USBMS BLOCK] __make_request() *** task awakened \n")) ;
 USBMS_DBG_INIT(iprintf("[USBMS INIT] __make_request() *** task awakened \n")) ;
 // USBMS_DBG_INFO(iprintf("[USBMS INFO] __make_request() *** task awakened \n")) ;
 /*
 #endif /// NO_ELEVATOR
 */

 // USBMS_DBG_INFO(iprintf("[USBMS INFO] __make_request() : q=%08XH \n",
 //	    q)) ;
 // USBMS_DBG_INFO(iprintf("[USBMS INFO] __make_request() : q->req=%08XH \n",
 // 	q->req)) ;

 // -------------
 out:
 // -------------
//iprintf("__make_request2: %u.%u\n", get_sys_time(), USB_CURRENT_TICKS); 
 if (freereq)
		__blk_put_request(q, freereq);

 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	 if(blk_queue_plugged(q)) {
		int nr_queued = q->rq.count[READ] + q->rq.count[WRITE];

		if(nr_queued == q->unplug_thresh) {
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __make_request(6) : __generic_unplug_device() .. \n"));
			__generic_unplug_device(q);
		   }
	   }
 #endif // ndef NO_ELEVATOR
 */

 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irq(q->queue_lock);
 #endif // ndef NO_SPINLOCK
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __make_request(7) End. \n"));
 
 if (test_bit(BIO_UPTODATE, &bio->bi_flags))  /* Check if there is any error! Charles */
    return 1;
 else
 {
    set_bit(BIO_UPTODATE, &bio->bi_flags);
    return 0;
 }

 // -------------
 end_io:
 // -------------

 // #ifndef USB_NO_VMM
	bio_endio(bio, nr_sectors * bio->bi_hardsect_size, -EWOULDBLOCK);
 // #endif // ndef USB_NO_VMM

 // -------------
 end_fail:
 // -------------

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __make_request(8) End with Failed !! \n"));
 // return 0;
 return -1 ;
}
// #endif // ndef USB_NO_VMM

/*
 * If bio->bi_dev is a partition, remap the location
 */

static inline void blk_partition_remap(struct bio *bio)
{
 /*
 struct block_device *bdev = bio->bi_bdev;
 */
 struct block_device *bdev ;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_partition_remap() Begin .. \n"));

 USBMS_DBG_MEM(iprintf("[USBMS MEM] blk_partition_remap() : bio->bi_bdev=%08XH \n",
	bio->bi_bdev));
 bdev = bio->bi_bdev;

 USBMS_DBG_MEM(iprintf("[USBMS MEM] blk_partition_remap() : bdev->bd_contains=%08XH \n",
	bdev->bd_contains));

 #ifdef USBMS_BS_ASSERT
	if ( ! (IS_VALID_ADDR(bdev->bd_contains)) ) {
		iprintf("[ASSERT][USBMS] blk_partition_remap() : " ) ;
    	iprintf("Error : Invalid Address !!\n" ) ;
        iprintf("[ASSERT][USBMS] bdev->bd_contains=%08XH \n",
					bdev->bd_contains ) ; ;
        iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT
 // if(bdev->bd_contains) {
 if (IS_VALID_ADDR(bdev->bd_contains)) {

	if(bdev != bdev->bd_contains) {
		/*
		struct hd_struct *p = bdev->bd_part;
		*/
		struct hd_struct *p;

		USBMS_DBG_INIT(iprintf("[USBMS INIT] blk_partition_remap() : Re-Map Partition .. \n"));
		USBMS_DBG_MEM(iprintf("[USBMS MEM] blk_partition_remap() : bdev->bd_part=%08XH \n",
			bdev->bd_part));
		p = bdev->bd_part;

		switch (bio->bi_rw) {
			case READ:
				p->read_sectors += bio_sectors(bio);
				p->reads++;
				break;
			case WRITE:
				p->write_sectors += bio_sectors(bio);
				p->writes++;
				break;
		   }

		USBMS_DBG_BLOCK(iprintf(
			"[USBMS BLOCK] blk_partition_remap() : Before , bio->bi_sector=%08XH \n",
			bio->bi_sector));
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] blk_partition_remap() : p->start_sect=%08XH \n",
			p->start_sect));
		USBMS_DBG_BLOCK(iprintf(
			"[USBMS BLOCK] blk_partition_remap() : p->start_sect=%08XH \n",
			p->start_sect));

		bio->bi_sector += p->start_sect;

		USBMS_DBG_BLOCK(iprintf(
			"[USBMS BLOCK] blk_partition_remap() : After , bio->bi_sector=%08XH \n",
			bio->bi_sector));

		bio->bi_bdev = bdev->bd_contains;
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] blk_partition_remap() : Set Value : bio->bi_bdev=%08XH \n",
			bio->bi_bdev));

		#ifdef USBMS_ONLY_ONE_DISK
			bio->bi_bdev->bd_disk = bdev->bd_disk ;
			USBMS_DBG_MEM(iprintf(
				"[USBMS MEM] blk_partition_remap() : Set Value : bio->bi_bdev->bd_disk=%08XH \n",
				bio->bi_bdev->bd_disk));
		#endif // USBMS_ONLY_ONE_DISK
	   }
   } // if bdev->bd_contains
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blk_partition_remap() End. \n"));
}

// ====================================================================

// for runtime debug from console
extern unsigned char usb_ms_log_status[];
extern unsigned long usb_storage_query_inbound_index(void);
extern void usb_storage_query_inbound_log(unsigned long index, unsigned char status
									, unsigned char volume, unsigned long sector_no);

static inline
void usb_storage_query_inbound_log_status(unsigned long index, unsigned char status)
{
	usb_ms_log_status[index] = status;
}

#define USBMS_LOG_LL_BLK_ENTRIES		128
static unsigned char usb_msca_log_init = 0, usb_msca_log_entry;
static unsigned char usb_msca_log_status[USBMS_LOG_LL_BLK_ENTRIES];
static unsigned char usb_msca_log_cmd[USBMS_LOG_LL_BLK_ENTRIES];
static unsigned char usb_msca_log_volumn[USBMS_LOG_LL_BLK_ENTRIES];
static unsigned char usb_msca_log_nr_sector[USBMS_LOG_LL_BLK_ENTRIES];
static unsigned long usb_msca_log_sector[USBMS_LOG_LL_BLK_ENTRIES];

static
void usb_storage_inner_log_init(void)
{
	if (usb_msca_log_init == 0) {
		usb_msca_log_entry = 0;
		usb_msca_log_init = 1;
	}
}

static
unsigned long usb_storage_inner_index(void)
{
	unsigned long	index;
	MASK_INTS();
	index = usb_msca_log_entry;
	if (usb_msca_log_entry < (USBMS_LOG_LL_BLK_ENTRIES-1))	usb_msca_log_entry ++;
		else												usb_msca_log_entry = 0;
	UNMASK_INTS();
	return	index;
}

static inline
void usb_storage_inner_log_status(unsigned long index, unsigned char status)
{
	usb_msca_log_status[index] = status;
}

static inline
void usb_storage_inner_log(unsigned long index, unsigned char status, unsigned char cmd
									, unsigned char volume, unsigned long sector_no, unsigned char nr_sector)
{
	usb_storage_inner_log_status(index, status);
	usb_msca_log_cmd[index] = cmd;
	usb_msca_log_volumn[index] = volume;
	usb_msca_log_nr_sector[index] = nr_sector;
	usb_msca_log_sector[index] = sector_no;
}

void usb_storage_inner_display(void)
{
}

void usb_storage_cache_display(void)
{
}
// ------------------------------

void UsbMsBlockIOLock(void)
{
	while (getres(MT_RES_USBMS_IO, 0xffffffff) != SUCCESS) { scdtsk(); }
}

void UsbMsBlockIOUnlock(void)
{
	relres(MT_RES_USBMS_IO);
}

int blk_sync_cache(void)
{
	return	1;
}

// ====================================================================

/**
 * generic_make_request: hand a buffer to its device driver for I/O
 * @bio:  The bio describing the location in memory and on the device.
 *
 * generic_make_request() is used to make I/O requests of block
 * devices. It is passed a &struct bio, which describes the I/O that needs
 * to be done.
 *
 * generic_make_request() does not return any status.  The
 * success/failure status of the request, along with notification of
 * completion, is delivered asynchronously through the bio->bi_end_io
 * function described (one day) else where.
 *
 * The caller of generic_make_request must make sure that bi_io_vec
 * are set to describe the memory buffer, and that bi_dev and bi_sector are
 * set to describe the device address, and the
 * bi_end_io and optionally bi_private are set to describe how
 * completion notification should be signaled.
 *
 * generic_make_request and the drivers it calls may use bi_next if this
 * bio happens to be merged with someone else, and may change bi_dev and
 * bi_sector for remaps as it sees fit.  So the values of these fields
 * should NOT be depended on after the call to generic_make_request.
 */
// #ifndef USB_NO_VMM // vitrtul memory management
// void generic_make_request(struct bio *bio)
int generic_make_request(struct bio *bio)
{
 request_queue_t *q;
 sector_t maxsector;
 /*
 int ret, nr_sectors = bio_sectors(bio);
 */
 int ret, nr_sectors ;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] generic_make_request() Begin .. \n"));

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] generic_make_request() : bio=%08XH \n",
	bio));

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] generic_make_request() : bio_sectors() .. \n"));
 nr_sectors = bio_sectors(bio);

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] generic_make_request() : bio=%08XH \n",
	bio));

 /* Test device or partition size, when known. */
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] generic_make_request() : bio->bi_bdev=%08XH \n",
	bio->bi_bdev));

 #ifdef USBMS_BS_ASSERT
	if ( ! (IS_VALID_ADDR(bio->bi_bdev)) ) {
		iprintf("[ASSERT][USBMS] generic_make_request() : " ) ;
	    iprintf("Error : Invalid Address !!\n" ) ;
        iprintf("[ASSERT][USBMS] bio->bi_bdev=%08XH \n",
				bio->bi_bdev ) ; ;
        iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT

 #ifndef NO_INODE
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] generic_make_request() : bio->bi_bdev->bd_inode=%08XH \n",
		bio->bi_bdev->bd_inode));
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] generic_make_request(1) : bd_inode->i_size=%ld \n",
		bio->bi_bdev->bd_inode->i_size));
	/*
	#ifndef _ETCPIP
		maxsector = bio->bi_bdev->bd_inode->i_size >> 9;
	 #else
	*/
		maxsector = bio->bi_bdev->bd_inode->i_size ;
	/*
	#endif
	*/
 #endif // ndef NO_INODE

 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] generic_make_request(2) : maxsector=%lu (%08XH) \n",
	maxsector, maxsector ));
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] generic_make_request(2) : nr_sectors=%lu (%08XH) \n",
	nr_sectors, nr_sectors ));

 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] generic_make_request(2) : bio->bi_sector=%lu (%08XH) \n",
	bio->bi_sector, bio->bi_sector ));

 if(maxsector) {
	sector_t sector = bio->bi_sector;

	if (maxsector < nr_sectors ||
		    maxsector - nr_sectors < sector) {
		char b[BDEVNAME_SIZE];
		/* This may well happen - the kernel calls
		 * bread() without checking the size of the
		 * device, e.g., when mounting a device. */
		/*
		printk(KERN_INFO
		       "attempt to access beyond end of device\n");
		*/
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] generic_make_request(1) : Error : attempt to access beyond end of device\n"
			));
		dprintf("[USBMS] Error : Access beyond end of device\n");
		/*
		printk(KERN_INFO "%s: rw=%ld, want=%Lu, limit=%Lu\n",
			       bdevname(bio->bi_bdev, b),
			       bio->bi_rw,
			       (unsigned long long) sector + nr_sectors,
			       (long long) maxsector);
		*/
		dprintf("[USBMS] generic_make_request: rw=%ld, want=%u, limit=%u\n",
			       bio->bi_rw,
			       (unsigned long) sector + nr_sectors,
			       (unsigned long) maxsector);
		set_bit(BIO_EOF, &bio->bi_flags);
		ret = -EIO ;
		goto end_io;
	   }
   }

 /*
  * Resolve the mapping until finished. (drivers are
  * still free to implement/resolve their own stacking
  * by explicitly returning 0)
  *
  * NOTE: we don't repeat the blk_size check for each new device.
  * Stacking drivers are expected to know what they are doing.
  */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
 do
 #endif // ndef NO_ELEVATOR
 */
   {
	char b[BDEVNAME_SIZE];

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] generic_make_request() : bdev_get_queue() .. \n"));
	q = bdev_get_queue(bio->bi_bdev);
	if (!q) {
		/*
		printk(KERN_ERR
			       "generic_make_request: Trying to access "
				"nonexistent block-device %s (%Lu)\n",
				bdevname(bio->bi_bdev, b),
				(long long) bio->bi_sector);
		*/
		dprintf("[USBMS] Error :generic_make_request: Trying to access "
				"nonexistent block-device %s (%Lu)\n",
				bdevname(bio->bi_bdev, b),
				(long long) bio->bi_sector);
		// -------------
		end_io:
		// -------------
		// #ifndef USB_NO_VMM
			bio_endio(bio, bio->bi_size, -EIO);
		// #endif // ndef USB_NO_VMM
		/*
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			break;
		 #else
		*/
			// return 0 ;
			return -EIO ;
		/*
		#endif
		*/
	   }
	/*
	if (unlikely(bio_sectors(bio) > q->max_sectors)) {
	*/
	if (bio_sectors(bio) > q->max_sectors) {
			printk("bio too big device %s (%u > %u)\n", 
				bdevname(bio->bi_bdev, b),
				bio_sectors(bio),
				q->max_sectors);
			goto end_io;
		}

	if (test_bit(QUEUE_FLAG_DEAD, &q->queue_flags))
			goto end_io;

	/*
	 * If this device has partitions, remap block n
	 * of partition p to block n+start(p) of the disk.
	 */
	USBMS_DBG_FLOW(iprintf(
		"[USBMS FLOW] generic_make_request() : blk_partition_remap() .. \n"
		));
	blk_partition_remap(bio);
	
	ret = q->make_request_fn(q, bio);
   }
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	 while (ret);
 #endif // ndef NO_ELEVATOR
 */

 // ---------------
 end_while:
 // ---------------

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] generic_make_request() End. \n"));
 // #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	return ret ;
 //  #else
 //	return 1 ;
 // #endif
}
// #endif // ndef USB_NO_VMM
/*
EXPORT_SYMBOL(generic_make_request);
*/
/**
 * submit_bio: submit a bio to the block device layer for I/O
 * @rw: whether to %READ or %WRITE, or maybe to %READA (read ahead)
 * @bio: The &struct bio which describes the I/O
 *
 * submit_bio() is very similar in purpose to generic_make_request(), and
 * uses that function to do most of the work. Both are fairly rough
 * interfaces, @bio must be presetup and ready for I/O.
 *
 */
// #ifndef USB_NO_VMM // vitrtul memory management
// B. S. Liaw : None Call this ??
int submit_bio(int rw, struct bio *bio)
{
 /*
 int count = bio_sectors(bio);
 */
// int count ;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] submit_bio() Begin .. \n"));

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] submit_bio() : bio=%08XH \n",
	bio
	));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] submit_bio() : bio_sectors() .. \n"));

// count = bio_sectors(bio);

 /*
 BIO_BUG_ON(!bio->bi_size);
 */
 if(!bio->bi_size) {
	dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
   }
 // #ifndef BLOCK_IO_REDUCED
 /*
 BIO_BUG_ON(!bio->bi_io_vec);
 */
 if(!bio->bi_io_vec) {
	dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
   }
 // #endif // ndef BLOCK_IO_REDUCED

 bio->bi_rw = rw;
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	if (rw & WRITE)
		mod_page_state(pgpgout, count);
	else
		mod_page_state(pgpgin, count);
 #endif // ndef USB_NO_VMM
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] submit_bio() : generic_make_request() .. \n"));
 if (generic_make_request(bio)) {
	return 1;
   }
  else {
	return 0 ;
   }
}
// #endif // ndef USB_NO_VMM
/*
EXPORT_SYMBOL(submit_bio);
*/
/**
 * blk_rq_next_segment
 * @rq:		the request being processed
 *
 * Description:
 *	Points to the next segment in the request if the current segment
 *	is complete. Leaves things unchanged if this segment is not over
 *	or if no more segments are left in this request.
 *
 *	Meant to be used for bio traversal during I/O submission
 *	Does not affect any I/O completions or update completion state
 *	in the request, and does not modify any bio fields.
 *
 *	Decrementing rq->nr_sectors, rq->current_nr_sectors and
 *	rq->nr_cbio_sectors as data is transferred is the caller's
 *	responsibility and should be done before calling this routine.
 **/
void blk_rq_next_segment(struct request *rq)
{
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if (rq->current_nr_sectors > 0)
 #endif // ndef NO_ELEVATOR
		return;
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	if (rq->nr_cbio_sectors > 0) {
		--rq->nr_cbio_segments;
		rq->current_nr_sectors = blk_rq_vec(rq)->bv_len >> 9;
	} else {
		if ((rq->cbio = rq->cbio->bi_next)) {
			rq->nr_cbio_segments = bio_segments(rq->cbio);
			rq->nr_cbio_sectors = bio_sectors(rq->cbio);
 			rq->current_nr_sectors = bio_cur_sectors(rq->cbio);
		}
 	}
 #endif // ndef USB_NO_VMM
 */
 /* remember the size of this segment before we start I/O */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	rq->hard_cur_sectors = rq->current_nr_sectors;
 #endif // ndef NO_ELEVATOR
 */
}

/**
 * process_that_request_first	-	process partial request submission
 * @req:	the request being processed
 * @nr_sectors:	number of sectors I/O has been submitted on
 *
 * Description:
 *	May be used for processing bio's while submitting I/O without
 *	signalling completion. Fails if more data is requested than is
 *	available in the request in which case it doesn't advance any
 *	pointers.
 *
 *	Assumes a request is correctly set up. No sanity checks.
 *
 * Return:
 *	0 - no more data left to submit (not processed)
 *	1 - data available to submit for this request (processed)
 **/
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
int process_that_request_first(struct request *req, unsigned int nr_sectors)
{
 unsigned int nsect;

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if (req->nr_sectors < nr_sectors)
 #endif // ndef NO_ELEVATOR
		return 0;

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	req->nr_sectors -= nr_sectors;
	req->sector += nr_sectors;
	while (nr_sectors) {
		nsect = min_t(unsigned, req->current_nr_sectors, nr_sectors);
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			req->current_nr_sectors -= nsect;
		#endif // ndef NO_ELEVATOR
		nr_sectors -= nsect;
		#ifndef USB_NO_VMM // vitrtul memory management
			if (req->cbio) {
				req->nr_cbio_sectors -= nsect;
				blk_rq_next_segment(req);
			}
		#endif // ndef USB_NO_VMM
	   }
	return 1;
 #endif // ndef NO_ELEVATOR
}
#endif // ndef NO_ELEVATOR
*/
/*
EXPORT_SYMBOL(process_that_request_first);
*/
void blk_recalc_rq_segments(struct request *rq)
{
 // #ifndef USB_NO_VMM // vitrtul memory management
	struct bio *bio;
 // #endif // ndef USB_NO_VMM
 int nr_phys_segs, nr_hw_segs;

 // #ifndef USB_NO_VMM // vitrtul memory management
	if (!rq->bio)
		return;

	nr_phys_segs = nr_hw_segs = 0;

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	rq_for_each_bio(bio, rq) 
  #else
	bio = rq->bio ;
 #endif
	   {
		/* Force bio hw/phys segs to be recalculated. */
		bio->bi_flags &= ~(1 << BIO_SEG_VALID);
		/*
		#ifndef NO_SCATTER_LIST
			nr_phys_segs += bio_phys_segments(rq->q, bio);
			nr_hw_segs += bio_hw_segments(rq->q, bio);
		#endif // ndef NO_SCATTER_LIST
		*/
	   }
 // #endif // ndef USB_NO_VMM
 /*
 #ifndef NO_SCATTER_LIST
	rq->nr_phys_segments = nr_phys_segs;
	rq->nr_hw_segments = nr_hw_segs;
 #endif // ndef NO_SCATTER_LIST
 */
}
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
void blk_recalc_rq_sectors(struct request *rq, int nsect)
{
 if (blk_fs_request(rq)) {
	rq->hard_sector += nsect;
	rq->hard_nr_sectors -= nsect;

	
	// * Move the I/O submission pointers ahead if required,
	// * i.e. for drivers not aware of rq->cbio.
	 
	if ((rq->nr_sectors >= rq->hard_nr_sectors) &&
	    (rq->sector <= rq->hard_sector)) {
		rq->sector = rq->hard_sector;
		USBMS_DBG_INIT(iprintf(
			"[USBMS INIT] blk_recalc_rq_sectors() : Set Value : rq->sector=%08XH \n" ,
			rq->sector
			));
		rq->nr_sectors = rq->hard_nr_sectors;
		#ifndef USB_NO_VMM // vitrtul memory management
			rq->hard_cur_sectors = bio_cur_sectors(rq->bio);
		#endif // ndef USB_NO_VMM
		rq->current_nr_sectors = rq->hard_cur_sectors;
		// #ifndef USB_NO_VMM // vitrtul memory management
			rq->nr_cbio_segments = bio_segments(rq->bio);
			rq->nr_cbio_sectors = bio_sectors(rq->bio);
		// #ifndef USB_NO_VMM // vitrtul memory management
			// struct page is on include/linux/mm.h
			rq->buffer = bio_data(rq->bio);
		// #endif // ndef USB_NO_VMM

			rq->cbio = rq->bio;
		// #endif // ndef USB_NO_VMM
	   }

	
	// * if total number of sectors is less than the first segment
	// * size, something has gone terribly wrong
	 
	if (rq->nr_sectors < rq->current_nr_sectors) {
		// printk("blk: request botched\n");
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] blk_recalc_rq_sectors() : Error : request botched !! \n"
			));
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] blk_recalc_rq_sectors() : rq->nr_sectors=%d \n" ,
			rq->nr_sectors
			));
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] blk_recalc_rq_sectors() : rq->current_nr_sectors=%d \n" ,
			rq->current_nr_sectors
			));
		rq->nr_sectors = rq->current_nr_sectors;
	   }
   }
}
#endif // ndef NO_ELEVATOR
*/

static int __end_that_request_first(struct request *req, int uptodate,
				    int nr_bytes)
{
 int total_bytes, bio_nbytes, error = 0, next_idx = 0;
 // #ifndef USB_NO_VMM // vitrtul memory management
	struct bio *bio;
 // #endif // ndef USB_NO_VMM

 /*
  * for a REQ_BLOCK_PC request, we want to carry any eventual
  * sense key with us all the way through
  */
 if (!blk_pc_request(req))
		req->errors = 0;

 if(!uptodate) {
	error = -EIO;
	if (!(req->flags & REQ_QUIET)) {
		#ifdef USB_DBGSTRING_ON
				printk("end_request: I/O error, dev %s, sector %llu\n",
					req->rq_disk ? req->rq_disk->disk_name : "?",
					(unsigned long long)req->sector);
		 #else
				dprintf("[USBMS] Error : I/O error, sector %lu !! \n",
					(unsigned long long)req->sector);
		#endif // ndef USB_DBGSTRING_ON

	   }
   }

 total_bytes = bio_nbytes = 0;

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	while ((bio = req->bio))
  #else
	bio = req->bio ;
 #endif // ndef NO_ELEVATOR
   {
	int nbytes;
	if (nr_bytes >= bio->bi_size) {
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			req->bio = bio->bi_next;
		#endif // ndef NO_ELEVATOR
			nbytes = bio->bi_size;
		// #ifndef USB_NO_VMM
			bio_endio(bio, nbytes, error);
		// #endif // ndef USB_NO_VMM
			next_idx = 0;
			bio_nbytes = 0;
		}
	   else {
		int idx = bio->bi_idx + next_idx;
		/*
		if (unlikely(bio->bi_idx >= bio->bi_vcnt)) {
		*/
		if (bio->bi_idx >= bio->bi_vcnt) {
			#ifdef USBMS_DBG_BLOCK_ON
				USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __end_that_request_first() : blk_dump_rq_flags() \n"));
				blk_dump_rq_flags(req, "__end_that");
			#endif // USBMS_DBG_BLOCK_ON
			printk("%s: bio idx %d >= vcnt %d\n",
						__FUNCTION__,
						bio->bi_idx, bio->bi_vcnt);

			#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
				break;
			  #else
				goto end_while ;
			#endif // ndef NO_ELEVATOR
		   }

		nbytes = bio_iovec_idx(bio, idx)->bv_len;
		/*
		BIO_BUG_ON(nbytes > bio->bi_size);
		*/
		/*
		if (nbytes > bio->bi_size) {
			dprintf("[USBMS] BUG : Badness in %s at %s:%d\n",
					__FUNCTION__, __FILE__, __LINE__);
		   }
		*/
		/*
		 * not a complete bvec done
		 */
		/*
		if (unlikely(nbytes > nr_bytes)) {
		*/
		if (nbytes > nr_bytes) {
			if (!nr_bytes)  /* No byte is read/written! Charles */
			   bio_endio(bio, nr_bytes, -EIO);
			else
			{
			   bio_iovec_idx(bio, idx)->bv_offset += nr_bytes;
			   bio_iovec_idx(bio, idx)->bv_len -= nr_bytes;
			   bio_nbytes += nr_bytes;
			   total_bytes += nr_bytes;
			}
			#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
				break;
			  #else
				goto end_while ;
			#endif // ndef NO_ELEVATOR
		   }

		/*
		 * advance to the next vector
		 */
		next_idx++;
		bio_nbytes += nbytes;
	   }

	total_bytes += nbytes;
	nr_bytes -= nbytes;

	if ((bio = req->bio)) {
		/*
		 * end more in this run, or just return 'not-done'
		 */
		/*
		if (unlikely(nr_bytes <= 0))
		*/
		if (nr_bytes <= 0) {
			#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
				break;
			  #else
				goto end_while ;
			 #endif // ndef NO_ELEVATOR
		   }
	   }
	} // while

 // --------
 end_while:
 // --------
 /*
  * completely done
  */
 if (!req->bio)
		return 0;

 /*
  * if the request wasn't completed, update state
  */
 if (bio_nbytes) {
	// #ifndef USB_NO_VMM
		bio_endio(bio, bio_nbytes, error);
	// #endif // ndef USB_NO_VMM
	req->bio->bi_idx += next_idx;
   }

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	blk_recalc_rq_sectors(req, total_bytes / bio->bi_hardsect_size);
 #endif // ndef NO_ELEVATOR
 blk_recalc_rq_segments(req);
 return 1;
}

/**
 * end_that_request_first - end I/O on a request
 * @req:      the request being processed
 * @uptodate: 0 for I/O error
 * @nr_sectors: number of sectors to end I/O on
 *
 * Description:
 *     Ends I/O on a number of sectors attached to @req, and sets it up
 *     for the next range of segments (if any) in the cluster.
 *
 * Return:
 *     0 - we are done with this request, call end_that_request_last()
 *     1 - still buffers pending for this request
 **/
int end_that_request_first(struct request *req, int uptodate, int nr_sectors)
{
	return __end_that_request_first(req, uptodate, nr_sectors * req->q->hardsect_size);
}
/*
EXPORT_SYMBOL(end_that_request_first);
*/
/**
 * end_that_request_chunk - end I/O on a request
 * @req:      the request being processed
 * @uptodate: 0 for I/O error
 * @nr_bytes: number of bytes to complete
 *
 * Description:
 *     Ends I/O on a number of bytes attached to @req, and sets it up
 *     for the next range of segments (if any). Like end_that_request_first(),
 *     but deals with bytes instead of sectors.
 *
 * Return:
 *     0 - we are done with this request, call end_that_request_last()
 *     1 - still buffers pending for this request
 **/
/* B. S. Liaw : None Call this ??
int end_that_request_chunk(struct request *req, int uptodate, int nr_bytes)
{
 return __end_that_request_first(req, uptodate, nr_bytes);
}
*/
/*
EXPORT_SYMBOL(end_that_request_chunk);
*/
/*
 * queue lock must be held
 */
void end_that_request_last(struct request *req)
{
 struct gendisk *disk = req->rq_disk;
 /*
 #ifndef _ETCPIP
    // struct completion is in <linux/completion.h>
	struct completion *waiting = req->waiting;
  #else
 */
 // #ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
	// 2003/12/02 B. S. Liaw :
	// req->waiting replaced by MT_EVT_USBMS_RQ/MT_EVT_USBMS_RQBI on kernel/local.h
	uint waiting_evt = req->waiting_evt ;
 // #endif // USBMS_USB_WAIT_BY_EVENT
 /*
 #endif // ndef _ETCPIP
 */


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] end_that_request_last() Begin .. \n"));
 if(disk && blk_fs_request(req)) {
	#ifndef NO_TIMER
		unsigned long duration = jiffies - req->start_time;
	#endif // ndef NO_TIMER
	switch (rq_data_dir(req)) {
	    case WRITE:
			disk_stat_inc(disk, writes);
			#ifndef NO_TIMER
				disk_stat_add(disk, write_ticks, duration);
			#endif // ndef NO_TIMER
			break;
	    case READ:
			disk_stat_inc(disk, reads);
			#ifndef NO_TIMER
				disk_stat_add(disk, read_ticks, duration);
			#endif // ndef NO_TIMER
			break;
		}
	disk_round_stats(disk);
	disk->in_flight--;
   }

 __blk_put_request(req->q, req);
 /* Do this LAST! The structure may be freed immediately afterwards */
 /*
 #ifndef _ETCPIP
    // struct completion is in <linux/completion.h>
	if (waiting)
		complete(waiting);
  #else
 */
 // 2003/12/02 B. S. Liaw :
 // req->waiting replaced by MT_EVT_USBMS_RQ/MT_EVT_USBMS_RQBI on kernel/local.h
 if (waiting_evt) {
	/*
	#ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
		USBMS_DBG_BLOCK(iprintf(
			"[USBMS BLOCK] end_that_request_last() : SIGNAL(%d) .. \n",
			waiting_evt)) ;
		USBMS_DBG_INIT(iprintf(
			"[USBMS INIT] end_that_request_last() : SIGNAL(%d) .. \n",
			waiting_evt)) ;
		// kernel/mtmacro.h : #define SIGNAL(signo) setevt(signo)
		if (SIGNAL(waiting_evt) != SUCCESS) {
			USBMS_DBG_INFO(iprintf(
				"[USBMS INFO] end_that_request_last() : Error : SIGNAL waiting_evt=%d fail !! \n",
				waiting_evt
				)) ;
			dprintf("[USBMS] Error : SIGNAL %d fail !! \n",waiting_evt);
		   }
	#else
		switch(waiting_evt) {
			case MT_EVT_USBMS_RQ :
				USBMS_DBG_INIT(iprintf(
					"[USBMS INIT] end_that_request_last() : Set Value : req->evt_blk_rq=1 .. \n"
					)) ;
				// BLOCK_PREEMPTION;
				req->evt_blk_rq = 1 ;
				// UNBLOCK_PREEMPTION;
				// iprintf("[USBMS] end_that_request_last() : FixMe : MT_EVT_USBMS_RQ !! \n") ;
				break ;
				#ifndef USBMS_NO_IOCTL
					case MT_EVT_USBMS_RQBI :
						USBMS_DBG_INIT(iprintf(
						"[USBMS INIT] end_that_request_last() : Set Value : req->evt_blk_rq_bi=1 .. \n"
						)) ;
					// BLOCK_PREEMPTION;
					req->evt_blk_rq_bi = 1 ;
					// UNBLOCK_PREEMPTION;
					break ;
				#endif // ndef USBMS_NO_IOCTL
			   }
	#endif // ndef USBMS_USB_WAIT_BY_EVENT
	*/
   }
 /*
 #endif
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] end_that_request_last() End. \n"));
}
/*
EXPORT_SYMBOL(end_that_request_last);
*/
/*
void end_request(struct request *req, int uptodate)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] end_request() Begin .. \n"));
 if (!end_that_request_first(req, uptodate, req->hard_cur_sectors)) {
	
	// B. S. Liaw : on drivers/char/random.c 
	// add_disk_randomness(req->rq_disk);
	
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] end_request() : blkdev_dequeue_request() .. \n"));
	blkdev_dequeue_request(req);
	end_that_request_last(req);
  }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] end_request() End. \n"));
}
*/
/*
EXPORT_SYMBOL(end_request);
*/
// #ifndef USB_NO_VMM // vitrtul memory management
void blk_rq_bio_prep(request_queue_t *q, struct request *rq, struct bio *bio)
{
 /* first three bits are identical in rq->flags and bio->bi_rw */
 rq->flags |= (bio->bi_rw & 7);
 /*
 #ifndef NO_SCATTER_LIST
	rq->nr_phys_segments = bio_phys_segments(q, bio);
	rq->nr_hw_segments = bio_hw_segments(q, bio);
 #endif // ndef NO_SCATTER_LIST
 */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	rq->current_nr_sectors = bio_cur_sectors(bio);
	rq->hard_cur_sectors = rq->current_nr_sectors;
	rq->hard_nr_sectors = rq->nr_sectors = bio_sectors(bio);
  #else
 */
	rq->hard_cur_sectors = bio_cur_sectors(bio) ;
	rq->hard_nr_sectors =  bio_sectors(bio);
 /*
 #endif
 */
 rq->nr_cbio_segments = bio_segments(bio);
 rq->nr_cbio_sectors = bio_sectors(bio);
 // #ifndef USB_NO_VMM // vitrtul memory management
 // struct page is on include/linux/mm.h
 rq->buffer = bio_data(bio);
 // #endif // ndef USB_NO_VMM
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	rq->cbio = rq->bio = rq->biotail = bio;
 #else
 */
	rq->cbio = rq->bio = bio;
 /*
 #endif // NO_ELEVATOR
 */
}
// #endif // ndef USB_NO_VMM
/*
EXPORT_SYMBOL(blk_rq_bio_prep);
*/
/* B. S. Liaw None Call this ??
void blk_rq_prep_restart(struct request *rq)
{
 // #ifndef USB_NO_VMM // vitrtul memory management
	struct bio *bio;

	bio = rq->cbio = rq->bio;
	if (bio) {
		rq->nr_cbio_segments = bio_segments(bio);
		rq->nr_cbio_sectors = bio_sectors(bio);
		rq->hard_cur_sectors = bio_cur_sectors(bio);
	// #ifndef USB_NO_VMM // vitrtul memory management
		// struct page is on include/linux/mm.h
		rq->buffer = bio_data(bio);
	// #endif // ndef USB_NO_VMM
	}
 // #endif // ndef USB_NO_VMM
 rq->sector = rq->hard_sector;
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] blk_rq_prep_restart() : Set Value : rq->sector=%08XH \n" ,
	rq->sector
	));
 rq->nr_sectors = rq->hard_nr_sectors;
 rq->current_nr_sectors = rq->hard_cur_sectors;
}
*/
/*
EXPORT_SYMBOL(blk_rq_prep_restart);
*/
#ifndef NO_WORKQUEUE
int kblockd_schedule_work(struct work_struct *work)
{
return queue_work(kblockd_workqueue, work);
}
#endif // ndef NO_WORKQUEUE

void kblockd_flush(void)
{
 /*
	flush_workqueue(kblockd_workqueue);
 */
}
/*
int __init blk_dev_init(void)
*/
int blk_dev_init(void)
{
 int i;
 /*
 kblockd_workqueue = create_workqueue("kblockd");
 if (!kblockd_workqueue)
		panic("Failed to create kblockd\n");
 */

 /* kmem_cache_t is on include/linux/slab.h
 #ifndef USB_NO_VMM // vitrtul memory management
 request_cachep = kmem_cache_create("blkdev_requests",
			sizeof(struct request), 0, 0, NULL, NULL); // size : 140
 if (!request_cachep)
		panic("Can't create request pool slab cache\n");
 #endif // ndef USB_NO_VMM
 */

 #ifndef USB_NO_VMM // vitrtul memory management
	blk_max_low_pfn = max_low_pfn;
	blk_max_pfn = max_pfn;
 #endif // ndef USB_NO_VMM

 #ifndef NO_WAITQUEUE
	for (i = 0; i < ARRAY_SIZE(congestion_wqh); i++)
		init_waitqueue_head(&congestion_wqh[i]);
 #endif // ndef NO_WAITQUEUE
 return 0;
}

/*
 * IO Context helper functions
 */
/*
#ifndef _ETCPIP
void put_io_context(struct io_context *ioc)
{
	if (ioc == NULL)
		return;
 
 // BUG_ON(atomic_read(&ioc->refcount) == 0);
 
 if (atomic_read(&ioc->refcount) == 0) {
		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }

	if (atomic_dec_and_test(&ioc->refcount)) {
		if (ioc->aic && ioc->aic->dtor)
			ioc->aic->dtor(ioc->aic);
 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] put_io_context() : UsbFreeMemC(ioc) .. \n"
		));
		
		// kfree(ioc);
		
		UsbFreeMemC(USB_MEM_C_S,(unsigned char *) ioc
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(struct io_context)
					#endif
					);
	}
}

// Called by the exitting task 
void exit_io_context(void)
{
	unsigned long flags;
 #ifndef _ETCPIP
	struct io_context *ioc;
 #endif // ndef _ETCPIP

 #ifndef _ETCPIP
	local_irq_save(flags);
	ioc = current->io_context;
 #endif // ndef _ETCPIP
	if (ioc) {
		if (ioc->aic && ioc->aic->exit)
			ioc->aic->exit(ioc->aic);
		put_io_context(ioc);
 #ifndef _ETCPIP
		current->io_context = NULL;
 #endif // ndef _ETCPIP
	} else
   {
	
	//	WARN_ON(1);
	
	if (!((1)!=0)) {
		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }
   }
 #ifndef _ETCPIP
 local_irq_restore(flags);
 #endif // ndef _ETCPIP
}
#endif // ndef _ETCPIP
*/

/*
 * If the current task has no IO context then create one and initialise it.
 * If it does have a context, take a ref on it.
 *
 * This is always called in the context of the task which submitted the I/O.
 * But weird things happen, so we disable local interrupts to ensure exclusive
 * access to *current.
 */
/*
#ifndef _ETCPIP
struct io_context *get_io_context(int gfp_flags)
{
 #ifndef _ETCPIP
	struct task_struct *tsk = current;
 #endif // ndef _ETCPIP
 unsigned long flags;
 struct io_context *ret;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] get_io_context() Begin .. \n"));
 
 #ifndef _ETCPIP
	local_irq_save(flags);

	ret = tsk->io_context;
	if (ret == NULL) {

		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] get_io_context() : ret = UsbAllocMemC() .. \n"
			));

		
		// ret = kmalloc(sizeof(*ret), GFP_ATOMIC);
		
		ret = (struct io_context *) UsbAllocMemC(USB_MEM_C_S
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct io_context)
		        #endif
           		) ;

		if (ret) {
			atomic_set(&ret->refcount, 1);
			ret->pid = tsk->pid;
		#ifndef NO_TIMER
			ret->last_waited = jiffies; // doesn't matter... 
		#endif // ndef NO_TIMER
			ret->nr_batch_requests = 0; // because this is 0 
			ret->aic = NULL;
			tsk->io_context = ret;
		}
	}
	if (ret)
		atomic_inc(&ret->refcount);
	local_irq_restore(flags);
 #endif // ndef _ETCPIP
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] get_io_context() End. \n"));
 return ret;
}

void copy_io_context(struct io_context **pdst, struct io_context **psrc)
{
	struct io_context *src = *psrc;
	struct io_context *dst = *pdst;

 if(src) {
	
	// BUG_ON(atomic_read(&src->refcount) == 0);
	
	if (atomic_read(&src->refcount) == 0) {
		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }
	atomic_inc(&src->refcount);
	put_io_context(dst);
	*pdst = src;
   }
}

void swap_io_context(struct io_context **ioc1, struct io_context **ioc2)
{
	struct io_context *temp;
	temp = *ioc1;
	*ioc1 = *ioc2;
	*ioc2 = temp;
}
#endif // ndef _ETCPIP
*/


/*
 * sysfs parts below
 */
/*
#ifndef NO_SYS_FS
struct queue_sysfs_entry {
 // #ifndef NO_SYS_FS
    // struct attribute on sysfs.h
	struct attribute attr;
 // #endif // ndef NO_SYS_FS
	ssize_t (*show)(struct request_queue *, char *);
	ssize_t (*store)(struct request_queue *, const char *, size_t);
};
#endif // ndef NO_SYS_FS
*/
/*
static ssize_t
queue_var_show(unsigned int var, char *page)
{
	return sprintf(page, "%d\n", var);
}

static ssize_t
queue_var_store(unsigned long *var, const char *page, size_t count)
{
	char *p = (char *) page;

	*var = simple_strtoul(p, &p, 10);
	return count;
}

static ssize_t queue_requests_show(struct request_queue *q, char *page)
{
	return queue_var_show(q->nr_requests, (page));
}

static ssize_t
queue_requests_store(struct request_queue *q, const char *page, size_t count)
{
	struct request_list *rl = &q->rq;

	int ret = queue_var_store(&q->nr_requests, page, count);
	if (q->nr_requests < BLKDEV_MIN_RQ)
		q->nr_requests = BLKDEV_MIN_RQ;

	if (rl->count[READ] >= queue_congestion_on_threshold(q))
		set_queue_congested(q, READ);
	else if (rl->count[READ] < queue_congestion_off_threshold(q))
		clear_queue_congested(q, READ);

	if (rl->count[WRITE] >= queue_congestion_on_threshold(q))
		set_queue_congested(q, WRITE);
	else if (rl->count[WRITE] < queue_congestion_off_threshold(q))
		clear_queue_congested(q, WRITE);

	if (rl->count[READ] >= q->nr_requests) {
		blk_set_queue_full(q, READ);
	} else if (rl->count[READ]+1 <= q->nr_requests) {
		blk_clear_queue_full(q, READ);
 #ifndef NO_WAITQUEUE
		wake_up(&rl->wait[READ]);
 #endif // ndef NO_WAITQUEUE
	}

	if (rl->count[WRITE] >= q->nr_requests) {
		blk_set_queue_full(q, WRITE);
	} else if (rl->count[WRITE]+1 <= q->nr_requests) {
		blk_clear_queue_full(q, WRITE);
 #ifndef NO_WAITQUEUE
		wake_up(&rl->wait[WRITE]);
 #endif // ndef NO_WAITQUEUE
	}
	return ret;
}
*/
/*
#ifndef NO_SYS_FS
static struct queue_sysfs_entry queue_requests_entry = {
 // #ifndef NO_SYS_FS
	// struct attribute on sysfs.h
	.attr = {.name = "nr_requests", .mode = S_IRUGO | S_IWUSR },
 // #endif // ndef NO_SYS_FS
 .show = queue_requests_show,
 .store = queue_requests_store,
};
#endif // ndef NO_SYS_FS
*/
/*
#ifndef NO_SYS_FS
	// struct attribute on sysfs.h
	static struct attribute *default_attrs[] = {
		&queue_requests_entry.attr,
		NULL,
	};
#define to_queue(atr) container_of((atr), struct queue_sysfs_entry, attr)
#endif // ndef NO_SYS_FS
*/
/*
#ifndef NO_SYS_FS
// struct attribute on sysfs.h
static ssize_t
queue_attr_show(struct kobject *kobj, struct attribute *attr, char *page)
{
	struct queue_sysfs_entry *entry = to_queue(attr);
	struct request_queue *q;

	q = container_of(kobj, struct request_queue, kobj);
	if (!entry->show)
		return 0;

	return entry->show(q, page);
}
#endif // ndef NO_SYS_FS

#ifndef NO_SYS_FS
// struct attribute on sysfs.h
static ssize_t
queue_attr_store(struct kobject *kobj, struct attribute *attr,
		    const char *page, size_t length)
{
	struct queue_sysfs_entry *entry = to_queue(attr);
	struct request_queue *q;

	q = container_of(kobj, struct request_queue, kobj);
	if (!entry->store)
		return -EINVAL;

	return entry->store(q, page, length);
}
#endif // ndef NO_SYS_FS
*/
/*
#ifndef NO_SYS_FS
// struct attribute on sysfs.h
static struct sysfs_ops queue_sysfs_ops = {
	.show	= queue_attr_show,
	.store	= queue_attr_store,
};

struct kobj_type queue_ktype = {
	.sysfs_ops	= &queue_sysfs_ops,
	.default_attrs	= default_attrs,
};
#endif // ndef NO_SYS_FS
*/

int blk_register_queue(struct gendisk *disk)
{
	int ret;

	request_queue_t *q = disk->queue;

	if (!q)
		return -ENXIO;

	q->kobj.parent = kobject_get(&disk->kobj);
	if (!q->kobj.parent)
		return -EBUSY;

 #ifdef USB_DBGSTRING_ON
	snprintf(q->kobj.name, KOBJ_NAME_LEN, "%s", "queue");
 #endif // USB_DBGSTRING_ON

 #ifndef NO_SYS_FS
	q->kobj.ktype = &queue_ktype;
 #endif // ndef NO_SYS_FS

	ret = kobject_register(&q->kobj);
	if (ret < 0)
		return ret;

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	ret = elv_register_queue(q);
	if (ret) {
		kobject_unregister(&q->kobj);
		return ret;
	}
 #endif // ndef NO_ELEVATOR 

	return 0;
}

void blk_unregister_queue(struct gendisk *disk)
{
 request_queue_t *q = disk->queue;

 if (q) {
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		elv_unregister_queue(q);
	#endif // ndef NO_ELEVATOR 

	#ifndef _ETCPIP
		kobject_unregister(&q->kobj);
	#endif // ndef _ETCPIP
	//#ifndef _ETCPIP
		kobject_put(&disk->kobj);
	//#endif // ndef _ETCPIP
   }
}

// B. S. Liaw TEST for compiler warning !!

