
// fs/bio.c

/*
 * Copyright (C) 2001 Jens Axboe <axboe@suse.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public Licens
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-
 *
 */

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsbMsInit.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP

/*
#include <linux/mm.h>
#include <linux/swap.h>
#include <linux/bio.h>
*/
#include "bio.h"
/*
#include <linux/blkdev.h>
*/
#include "bitops.h" // B. S. Liaw add for test_bit()
#include "blkdev.h"
/*
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/kernel.h>
*/
#include "PoKernel.h"
/*
#include <linux/module.h>
#include <linux/mempool.h>
#include <linux/workqueue.h>
*/



#define BIO_POOL_SIZE 256
/*
static mempool_t *bio_pool;
// kmem_cache_t is on include/linux/slab.h
static kmem_cache_t *bio_slab;
*/

#define BIOVEC_NR_POOLS 6

/*
 * a small number of entries is fine, not going to be performance critical.
 * basically we just need to survive
 */
#define BIO_SPLIT_ENTRIES 8	
/* kmem_cache_t is on include/linux/slab.h
mempool_t *bio_split_pool;

struct biovec_pool {
	int nr_vecs;
	char *name; 
	kmem_cache_t *slab;
	mempool_t *pool;
};
*/
/*
 * if you change this list, also change bvec_alloc or things will
 * break badly! cannot be bigger than what you can fit into an
 * unsigned short
 */

/* kmem_cache_t is on include/linux/slab.h
#define BV(x) { .nr_vecs = x, .name = "biovec-" #x }
static struct biovec_pool bvec_array[BIOVEC_NR_POOLS] = {
	BV(1), BV(4), BV(16), BV(64), BV(128), BV(BIO_MAX_PAGES),
};
*/
#undef BV

/*
#ifndef _ETCPIP
	static inline struct bio_vec *bvec_alloc(int gfp_mask, int nr, unsigned long *idx)
 #else
*/
static inline struct bio_vec *bvec_alloc(int nr, unsigned long *idx)
/*
#endif
*/
{
 /* kmem_cache_t is on include/linux/slab.h
	struct biovec_pool *bp;
 */
 struct bio_vec *bvl;

 USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] bvec_alloc() : nr=%d \n",
		nr
		));
 /*
  * see comment near bvec_array define!
  */
 switch (nr) {
	case   1        : *idx = 0; break;
	case   2 ...   4: *idx = 1; break;
	case   5 ...  16: *idx = 2; break;
	case  17 ...  64: *idx = 3; break;
	case  65 ... 128: *idx = 4; break;
	case 129 ... BIO_MAX_PAGES: *idx = 5; break;
	default:
		return NULL;
   }
 /*
  * idx now points to the pool we want to allocate from
  */
 /* kmem_cache_t is on include/linux/slab.h
	bp = bvec_array + *idx;
 */
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] bvec_alloc() : bvl = UsbAllocMemC() , size=%d .. \n",
	nr*sizeof(struct bio_vec)
	));
 /* kmem_cache_t is on include/linux/slab.h
	bvl = mempool_alloc(bp->pool, gfp_mask);
 */
 bvl = (struct bio_vec *) UsbAllocMemC(USB_MEM_C_S
			#ifdef USB_DBG_MEM_C_ON
 		  		,nr*sizeof(struct bio_vec) // size : 12
			#endif
			) ;
 if (bvl)
   {
	/*
	memset(bvl, 0, bp->nr_vecs * sizeof(struct bio_vec));
	*/
	memset(bvl, 0, nr * sizeof(struct bio_vec));
   }
 return bvl;
}

/*
 * default destructor for a bio allocated with bio_alloc()
 */
void bio_destructor(struct bio *bio)
{
 const int pool_idx = BIO_POOL_IDX(bio);
/* kmem_cache_t is on include/linux/slab.h
 struct biovec_pool *bp = bvec_array + pool_idx;
 */
 /*
 BIO_BUG_ON(pool_idx >= BIOVEC_NR_POOLS);
 */
 if(pool_idx >= BIOVEC_NR_POOLS) {
	dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
   }

 /*
  * cloned bio doesn't own the veclist
  */
 if (!bio_flagged(bio, BIO_CLONED))
   {
	/*
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		// kmem_cache_t is on include/linux/slab.h
		// mempool_free(bio->bi_io_vec, bp->pool);
		
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] bio_destructor() : UsbFreeMemC(bio->bi_io_vec) , size=%d .. \n",
			sizeof(struct bio_vec)
			));
		UsbFreeMemC(USB_MEM_C_S,(unsigned char *) bio->bi_io_vec
			#if defined(USB_DBG_MEM_C_ON)
				,sizeof(struct bio_vec) // size : 12
			#endif
			) ;
	#endif // ndef NO_ELEVATOR
	*/
   }

 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	// kmem_cache_t is on include/linux/slab.h
	mempool_free(bio, bio_pool);
	
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] bio_destructor() : UsbFreeMemC(bio) , size=%d .. \n",
		sizeof(struct bio)
		));
	UsbFreeMemC(USB_MEM_C_S,(unsigned char *) bio
			#if defined(USB_DBG_MEM_C_ON)
				,sizeof(struct bio)
			#endif
			) ;
 #endif // ndef NO_ELEVATOR
 */
}

inline void bio_init(struct bio *bio)
{
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	bio->bi_next = NULL;
 #endif // ndef NO_ELEVATOR
 */
 bio->bi_flags = 1 << BIO_UPTODATE;
 bio->bi_rw = 0;
 bio->bi_vcnt = 0;
 bio->bi_idx = 0;
 bio->bi_phys_segments = 0;
 bio->bi_hw_segments = 0;
 bio->bi_size = 0;
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] bio_init() : Set Value : bio->bi_size=%d \n",
	bio->bi_size )) ;
 #ifndef BLOCK_IO_REDUCED
	bio->bi_max_vecs = 0;
 #endif // ndef BLOCK_IO_REDUCED
 bio->bi_end_io = NULL;
 #ifndef BLOCK_IO_REDUCED
	atomic_set(&bio->bi_cnt, 1);
 #endif // ndef BLOCK_IO_REDUCED
 bio->bi_private = NULL;
}

/**
 * bio_alloc - allocate a bio for I/O
 * @gfp_mask:   the GFP_ mask given to the slab allocator
 * @nr_iovecs:	number of iovecs to pre-allocate
 *
 * Description:
 *   bio_alloc will first try it's on mempool to satisfy the allocation.
 *   If %__GFP_WAIT is set then we will block on the internal pool waiting
 *   for a &struct bio to become free.
 **/
/*
#ifndef _ETCPIP
	struct bio *bio_alloc(int gfp_mask, int nr_iovecs)
#else
*/
struct bio *bio_alloc(int nr_iovecs)
/*
#endif
*/
{
 struct bio_vec *bvl = NULL;
 unsigned long idx;
 struct bio *bio;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] bio_alloc() Begin .. \n"));
 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] bio_alloc() : bio = UsbAllocMemC() , size=%d .. \n",
		sizeof(struct bio)
		));
 /* kmem_cache_t is on include/linux/slab.h
 bio = mempool_alloc(bio_pool, gfp_mask);
 */
 bio = (struct bio *) UsbAllocMemC(USB_MEM_C_S
			#ifdef USB_DBG_MEM_C_ON
 		  		,sizeof(struct bio)
			#endif
			) ;
 /*
 if (unlikely(!bio))
 */
 if (!bio)
		goto out;

 bio_init(bio);
 /*
	if (unlikely(!nr_iovecs))
 */
 if (!nr_iovecs)
		goto noiovec;

 /*
 #ifndef _ETCPIP
	bvl = bvec_alloc(gfp_mask, nr_iovecs, &idx);
  #else
 */
	bvl = bvec_alloc(nr_iovecs, &idx);
 /*
 #endif
 */
 if(bvl) {
	bio->bi_flags |= idx << BIO_POOL_OFFSET;
	#ifndef BLOCK_IO_REDUCED
		bio->bi_max_vecs = bvec_array[idx].nr_vecs;
	#endif // ndef BLOCK_IO_REDUCED
	noiovec:
		bio->bi_io_vec = bvl;
		bio->bi_destructor = bio_destructor;
	out:
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] bio_alloc() End. \n"));
	return bio;
   }

 /* kmem_cache_t is on include/linux/slab.h
 mempool_free(bio, bio_pool);
 */
 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] bio_alloc() : UsbFreeMemC(bio) , size=%d .. \n",
		sizeof(struct bio)
		));
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) bio
			#if defined(USB_DBG_MEM_C_ON)
				,sizeof(struct bio)
			#endif
			) ;
 bio = NULL;
 goto out;
}

/**
 * bio_put - release a reference to a bio
 * @bio:   bio to release reference to
 *
 * Description:
 *   Put a reference to a &struct bio, either one you have gotten with
 *   bio_alloc or bio_get. The last put of a bio will free it.
 **/
void bio_put(struct bio *bio)
{
 /*
 #ifndef BLOCK_IO_REDUCED
	// BIO_BUG_ON(!atomic_read(&bio->bi_cnt));
	if(!atomic_read(&bio->bi_cnt)) {
		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }
 #endif // ndef BLOCK_IO_REDUCED
 */
 /*
  * last put frees it
  */
 /*
 #ifndef BLOCK_IO_REDUCED
	if(atomic_dec_and_test(&bio->bi_cnt))
 #endif // ndef BLOCK_IO_REDUCED
 */
	   {
		/*
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			bio->bi_next = NULL;
		#endif // ndef NO_ELEVATOR
		*/
		bio->bi_destructor(bio);
	   }
}
/*
inline int bio_phys_segments(request_queue_t *q, struct bio *bio)
{
 
 //	if (unlikely(!bio_flagged(bio, BIO_SEG_VALID)))
 
 if (!bio_flagged(bio, BIO_SEG_VALID))
   {
	#ifndef _ETCPIP
		blk_recount_segments(q, bio);
	#endif // ndef _ETCPIP
   }

 return bio->bi_phys_segments;
}
*/
/*
inline int bio_hw_segments(request_queue_t *q, struct bio *bio)
{
 
 // if (unlikely(!bio_flagged(bio, BIO_SEG_VALID)))
 
 if(!bio_flagged(bio, BIO_SEG_VALID))
   {
	#ifndef _ETCPIP
		blk_recount_segments(q, bio);
	#endif // ndef _ETCPIP
   }

 return bio->bi_hw_segments;
}
*/
/**
 * 	__bio_clone	-	clone a bio
 * 	@bio: destination bio
 * 	@bio_src: bio to clone
 *
 *	Clone a &bio. Caller will own the returned bio, but not
 *	the actual data it points to. Reference count of returned
 * 	bio will be one.
 */
/*
inline void __bio_clone(struct bio *bio, struct bio *bio_src)
{
	bio->bi_io_vec = bio_src->bi_io_vec;

	bio->bi_sector = bio_src->bi_sector;
	bio->bi_bdev = bio_src->bi_bdev;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] __bio_clone() : Set Value : bio->bi_bdev=%08XH \n",
		bio->bi_bdev));
	bio->bi_flags |= 1 << BIO_CLONED;
	bio->bi_rw = bio_src->bi_rw;

	
	// notes -- maybe just leave bi_idx alone. assume identical mapping
	// for the clone
	
	bio->bi_vcnt = bio_src->bi_vcnt;
	bio->bi_idx = bio_src->bi_idx;
	if (bio_flagged(bio, BIO_SEG_VALID)) {
		bio->bi_phys_segments = bio_src->bi_phys_segments;
		bio->bi_hw_segments = bio_src->bi_hw_segments;
		bio->bi_flags |= (1 << BIO_SEG_VALID);
	}
	bio->bi_size = bio_src->bi_size;
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] __bio_clone() : Set Value : bio->bi_size=%d \n",
		bio->bi_size )) ;

	
	// cloned bio does not own the bio_vec, so users cannot fiddle with
	// it. clear bi_max_vecs and clear the BIO_POOL_BITS to make this
	// apparent
	 
 #ifndef BLOCK_IO_REDUCED
	bio->bi_max_vecs = 0;
 #endif // ndef BLOCK_IO_REDUCED
	bio->bi_flags &= (BIO_POOL_MASK - 1);
}
*/
/**
 *	bio_clone	-	clone a bio
 *	@bio: bio to clone
 *	@gfp_mask: allocation priority
 *
 * 	Like __bio_clone, only also allocates the returned bio
 */
/*
struct bio *bio_clone(struct bio *bio, int gfp_mask)
{
 
 #ifndef _ETCPIP
	struct bio *b = bio_alloc(gfp_mask, 0);
  #else
 
	struct bio *b = bio_alloc(0);
 
 #endif
 

	if (b)
		__bio_clone(b, bio);

	return b;
}
*/
/**
 *	bio_get_nr_vecs		- return approx number of vecs
 *	@bdev:  I/O target
 *
 *	Return the approximate number of pages we can send to this target.
 *	There's no guarantee that you will be able to fit this number of pages
 *	into a bio, it does not account for dynamic restrictions that vary
 *	on offset.
 */
/*
int bio_get_nr_vecs(struct block_device *bdev)
{
	request_queue_t *q = bdev_get_queue(bdev);
	int nr_pages;

	nr_pages = ((q->max_sectors << 9) + PAGE_SIZE - 1) >> PAGE_SHIFT;
	if (nr_pages > q->max_phys_segments)
		nr_pages = q->max_phys_segments;
	if (nr_pages > q->max_hw_segments)
		nr_pages = q->max_hw_segments;

	return nr_pages;
}
*/
/**
 *	bio_add_page	-	attempt to add page to bio
 *	@bio: destination bio
 *	@page: page to add
 *	@len: vec entry length
 *	@offset: vec entry offset
 *
 *	Attempt to add a page to the bio_vec maplist. This can fail for a
 *	number of reasons, such as the bio being full or target block
 *	device limitations.
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
// struct page is on include/linux/mm.h
int bio_add_page(struct bio *bio, struct page *page, unsigned int len,
		 unsigned int offset)
{
	request_queue_t *q = bdev_get_queue(bio->bi_bdev);
	int retried_segments = 0;
	struct bio_vec *bvec;

	
 // cloned bio must not modify vec list
	 
 
 //	if (unlikely(bio_flagged(bio, BIO_CLONED)))
 
	if (bio_flagged(bio, BIO_CLONED))
		return 0;

 #ifndef BLOCK_IO_REDUCED
	if (bio->bi_vcnt >= bio->bi_max_vecs)
		return 0;
 #endif // ndef BLOCK_IO_REDUCED

	if (((bio->bi_size + len) >> 9) > q->max_sectors)
		return 0;

	
 //	 * we might lose a segment or two here, but rather that than
 //	 * make this too complex.
	 

	while (bio_phys_segments(q, bio) >= q->max_phys_segments
	    || bio_hw_segments(q, bio) >= q->max_hw_segments) {

		if (retried_segments)
			return 0;

		bio->bi_flags &= ~(1 << BIO_SEG_VALID);
		retried_segments = 1;
	}

	
	 // setup the new entry, we might clear it again later if we
	 // cannot add the page
	 
	bvec = &bio->bi_io_vec[bio->bi_vcnt];
 #ifndef USB_NO_VMM // vitrtul memory management
	bvec->bv_page = page;
 #endif // ndef USB_NO_VMM
	bvec->bv_len = len;
	bvec->bv_offset = offset;

	
	// * if queue has other restrictions (eg varying max sector size
	// * depending on offset), it can specify a merge_bvec_fn in the
	// * queue to get further control
	 
	if (q->merge_bvec_fn) {
		
		// * merge_bvec_fn() returns number of bytes it can accept
		// * at this offset
		 
		if (q->merge_bvec_fn(q, bio, bvec) < len) {
		#ifndef USB_NO_VMM // vitrtul memory management
			bvec->bv_page = NULL;
		#endif // ndef USB_NO_VMM
			bvec->bv_len = 0;
			bvec->bv_offset = 0;
			return 0;
		}
	}

	bio->bi_vcnt++;
	bio->bi_phys_segments++;
	bio->bi_hw_segments++;
	bio->bi_size += len;
	return len;
}
#endif // ndef USB_NO_VMM
*/
/*
#ifndef USB_NO_VMM // vitrtul memory management
static struct bio *__bio_map_user(struct block_device *bdev,
				  unsigned long uaddr, unsigned int len,
				  int write_to_vm)
{
	unsigned long end = (uaddr + len + PAGE_SIZE - 1) >> PAGE_SHIFT;
	unsigned long start = uaddr >> PAGE_SHIFT;
	const int nr_pages = end - start;
	request_queue_t *q = bdev_get_queue(bdev);
	int ret, offset, i;
	struct page **pages;
	struct bio *bio;

	
	// transfer and buffer must be aligned to at least hardsector
	// size for now, in the future we can relax this restriction
	 
	if ((uaddr & queue_dma_alignment(q)) || (len & queue_dma_alignment(q)))
		return NULL;

	bio = bio_alloc(GFP_KERNEL, nr_pages);
	if (!bio)
		return NULL;

	pages = kmalloc(nr_pages * sizeof(struct page *), GFP_KERNEL);
	if (!pages)
		goto out;

	down_read(&current->mm->mmap_sem);
	ret = get_user_pages(current, current->mm, uaddr, nr_pages,
						write_to_vm, 0, pages, NULL);
	up_read(&current->mm->mmap_sem);

	if (ret < nr_pages)
		goto out;

	bio->bi_bdev = bdev;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] __bio_map_user() : Set Value : bio->bi_bdev=%08XH \n",
		bio->bi_bdev));

	offset = uaddr & ~PAGE_MASK;
	for (i = 0; i < nr_pages; i++) {
		unsigned int bytes = PAGE_SIZE - offset;

		if (len <= 0)
			break;

		if (bytes > len)
			bytes = len;

		
		// sorry...
		 
		if (bio_add_page(bio, pages[i], bytes, offset) < bytes)
			break;

		len -= bytes;
		offset = 0;
	}


 // release the pages we didn't map into the bio, if any
	
	while (i < nr_pages)
		page_cache_release(pages[i++]);

	kfree(pages);

	
	// set data direction, and check if mapped pages need bouncing
	 
	if (!write_to_vm)
		bio->bi_rw |= (1 << BIO_RW);

	blk_queue_bounce(q, &bio);
	return bio;
 out:
	kfree(pages);
	bio_put(bio);
	return NULL;
}
#endif // ndef USB_NO_VMM
*/
/**
 *	bio_map_user	-	map user address into bio
 *	@bdev: destination block device
 *	@uaddr: start of user address
 *	@len: length in bytes
 *	@write_to_vm: bool indicating writing to pages or not
 *
 *	Map the user space address into a bio suitable for io to a block
 *	device.
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
struct bio *bio_map_user(struct block_device *bdev, unsigned long uaddr,
			 unsigned int len, int write_to_vm)
{
	struct bio *bio;

	bio = __bio_map_user(bdev, uaddr, len, write_to_vm);

	if (bio) {
		
		// * subtle -- if __bio_map_user() ended up bouncing a bio,
		// * it would normally disappear when its bi_end_io is run.
		// * however, we need it for the unmap, so grab an extra
		// * reference to it
		
		bio_get(bio);

		if (bio->bi_size < len) {
			bio_endio(bio, bio->bi_size, 0);
		// #ifndef USB_NO_VMM
			bio_unmap_user(bio, 0);
		// #endif // ndef USB_NO_VMM
			return NULL;
		}
	}

	return bio;
}
#endif // ndef USB_NO_VMM
*/
/*
#ifndef USB_NO_VMM // vitrtul memory management
static void __bio_unmap_user(struct bio *bio, int write_to_vm)
{
	struct bio_vec *bvec;
	int i;

	
	// find original bio if it was bounced
	 
	if (bio->bi_private) {
		
		// someone stole our bio, must not happen
		
		
		// BUG_ON(!bio_flagged(bio, BIO_BOUNCED));
		
 if (!bio_flagged(bio, BIO_BOUNCED)) {
		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }
	
		bio = bio->bi_private;
	}

	
	// make sure we dirty pages we wrote to
	 
	__bio_for_each_segment(bvec, bio, i, 0) {
		if (write_to_vm)
			set_page_dirty_lock(bvec->bv_page);

		page_cache_release(bvec->bv_page);
	}

	bio_put(bio);
}
#endif // ndef USB_NO_VMM
*/
/**
 *	bio_unmap_user	-	unmap a bio
 *	@bio:		the bio being unmapped
 *	@write_to_vm:	bool indicating whether pages were written to
 *
 *	Unmap a bio previously mapped by bio_map_user(). The @write_to_vm
 *	must be the same as passed into bio_map_user(). Must be called with
 *	a process context.
 *
 *	bio_unmap_user() may sleep.
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
void bio_unmap_user(struct bio *bio, int write_to_vm)
{
	__bio_unmap_user(bio, write_to_vm);
	bio_put(bio);
}
#endif // ndef USB_NO_VMM
*/
/*
 * bio_set_pages_dirty() and bio_check_pages_dirty() are support functions
 * for performing direct-IO in BIOs.
 *
 * The problem is that we cannot run set_page_dirty() from interrupt context
 * because the required locks are not interrupt-safe.  So what we can do is to
 * mark the pages dirty _before_ performing IO.  And in interrupt context,
 * check that the pages are still dirty.   If so, fine.  If not, redirty them
 * in process context.
 *
 * We special-case compound pages here: normally this means reads into hugetlb
 * pages.  The logic in here doesn't really work right for compound pages
 * because the VM does not uniformly chase down the head page in all cases.
 * But dirtiness of compound pages is pretty meaningless anyway: the VM doesn't
 * handle them at all.  So we skip compound pages here at an early stage.
 *
 * Note that this code is very hard to test under normal circumstances because
 * direct-io pins the pages with get_user_pages().  This makes
 * is_page_cache_freeable return false, and the VM will not clean the pages.
 * But other code (eg, pdflush) could clean the pages if they are mapped
 * pagecache.
 *
 * Simply disabling the call to bio_set_pages_dirty() is a good way to test the
 * deferred bio dirtying paths.
 */

/*
 * bio_set_pages_dirty() will mark all the bio's pages as dirty.
 */
/*
void bio_set_pages_dirty(struct bio *bio)
{
	struct bio_vec *bvec = bio->bi_io_vec;
	int i;

	for (i = 0; i < bio->bi_vcnt; i++) {
		#ifndef USB_NO_VMM // vitrtul memory management
			// struct page is on include/linux/mm.h
		struct page *page = bvec[i].bv_page;

		if (page && !PageCompound(page))
			set_page_dirty_lock(page);
		#endif // ndef USB_NO_VMM
	}
}

static void bio_release_pages(struct bio *bio)
{
	struct bio_vec *bvec = bio->bi_io_vec;
	int i;

	for (i = 0; i < bio->bi_vcnt; i++) {
		#ifndef USB_NO_VMM // vitrtul memory management
			// struct page is on include/linux/mm.h
		struct page *page = bvec[i].bv_page;

		if (page)
			put_page(page);
		#endif // ndef USB_NO_VMM
	}
}
*/
/*
 * bio_check_pages_dirty() will check that all the BIO's pages are still dirty.
 * If they are, then fine.  If, however, some pages are clean then they must
 * have been written out during the direct-IO read.  So we take another ref on
 * the BIO and the offending pages and re-dirty the pages in process context.
 *
 * It is expected that bio_check_pages_dirty() will wholly own the BIO from
 * here on.  It will run one page_cache_release() against each page and will
 * run one bio_put() against the BIO.
 */

static void bio_dirty_fn(void *data);
/*
#ifndef NO_WORKQUEUE
	static DECLARE_WORK(bio_dirty_work, bio_dirty_fn, NULL);
#endif // ndef NO_WORKQUEUE
*/
/*
#ifndef NO_SPINLOCK
	static spinlock_t bio_dirty_lock = SPIN_LOCK_UNLOCKED;
#endif // ndef NO_SPINLOCK
*/
/*
static struct bio *bio_dirty_list = NULL;
*/
/*
 * This runs in process context
 */
/*
static void bio_dirty_fn(void *data)
{
	unsigned long flags;
	struct bio *bio;

 
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(&bio_dirty_lock, flags);
 #endif // ndef NO_SPINLOCK
 
	bio = bio_dirty_list;
	bio_dirty_list = NULL;
 
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(&bio_dirty_lock, flags);
 #endif // ndef NO_SPINLOCK
 

	while (bio) {
		struct bio *next = bio->bi_private;

		bio_set_pages_dirty(bio);
		bio_release_pages(bio);
		bio_put(bio);
		bio = next;
	}
}
*/
/*
void bio_check_pages_dirty(struct bio *bio)
{
 struct bio_vec *bvec = bio->bi_io_vec;
 int nr_clean_pages = 0;
 int i;

 #ifndef USB_NO_VMM // vitrtul memory management
	for (i = 0; i < bio->bi_vcnt; i++) {
			// struct page is on include/linux/mm.h
			struct page *page = bvec[i].bv_page;

		if (PageDirty(page) || PageCompound(page)) {
			page_cache_release(page);
			bvec[i].bv_page = NULL;
		} else {
			nr_clean_pages++;
		}
	}
 #endif // ndef USB_NO_VMM

	if (nr_clean_pages) {
		unsigned long flags;

 
 #ifndef NO_SPINLOCK
		spin_lock_irqsave(&bio_dirty_lock, flags);
 #endif // ndef NO_SPINLOCK
 
		bio->bi_private = bio_dirty_list;
		bio_dirty_list = bio;
 
 #ifndef NO_SPINLOCK
		spin_unlock_irqrestore(&bio_dirty_lock, flags);
 #endif // ndef NO_SPINLOCK
 
	#ifndef NO_WORKQUEUE
		schedule_work(&bio_dirty_work);
	#endif // ndef NO_WORKQUEUE
	} else {
		bio_put(bio);
	}
}
*/
/**
 * bio_endio - end I/O on a bio
 * @bio:	bio
 * @bytes_done:	number of bytes completed
 * @error:	error, if any
 *
 * Description:
 *   bio_endio() will end I/O on @bytes_done number of bytes. This may be
 *   just a partial part of the bio, or it may be the whole bio. bio_endio()
 *   is the preferred way to end I/O on a bio, it takes care of decrementing
 *   bi_size and clearing BIO_UPTODATE on error. @error is 0 on success, and
 *   and one of the established -Exxxx (-EIO, for instance) error values in
 *   case something went wrong. Noone should call bi_end_io() directly on
 *   a bio unless they own it and thus know that it has an end_io function.
 **/
void bio_endio(struct bio *bio, unsigned int bytes_done, int error)
{
 if(error) {
	clear_bit(BIO_UPTODATE, &bio->bi_flags);
   }
 /*
 if (unlikely(bytes_done > bio->bi_size)) {
 */
 if (bytes_done > bio->bi_size) {
	printk("%s: want %u bytes done, only %u left\n", __FUNCTION__,
			bytes_done, bio->bi_size);
	bytes_done = bio->bi_size;
   }

 bio->bi_size -= bytes_done;
 bio->bi_sector += (bytes_done / bio->bi_hardsect_size);

 if(bio->bi_end_io) {
	bio->bi_end_io(bio, bytes_done, error);
   }
  else {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] bio_endio() : Null bio->bi_end_io() !! \n"));
   }
}
/*
void bio_pair_release(struct bio_pair *bp)
{
 if (atomic_dec_and_test(&bp->cnt)) {
	struct bio *master = bp->bio1.bi_private;

	bio_endio(master, master->bi_size, bp->error);
	// kmem_cache_t is on include/linux/slab.h
	// mempool_free(bp, bp->bio2.bi_private);
	
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] bio_pair_release() : UsbFreeMemC(bp) , size=%d .. \n",
		sizeof(struct bio_pair)
		));
	UsbFreeMemC(USB_MEM_C_S,(unsigned char *) bp
			#if defined(USB_DBG_MEM_C_ON)
				,sizeof(struct bio_pair)
			#endif
			) ;
   }
}

static int bio_pair_end_1(struct bio * bi, unsigned int done, int err)
{
	struct bio_pair *bp = container_of(bi, struct bio_pair, bio1);

	if (bi->bi_size)
		return 1;
	if (err)
		bp->error = err;

	bio_pair_release(bp);
	return 0;
}

static int bio_pair_end_2(struct bio * bi, unsigned int done, int err)
{
	struct bio_pair *bp = container_of(bi, struct bio_pair, bio2);

	if (bi->bi_size)
		return 1;
	if (err)
		bp->error = err;

	bio_pair_release(bp);
	return 0;
}
*/
/*
 * split a bio - only worry about a bio with a single page
 * in it's iovec
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
// mempool_t is on include/linux/mempool.h
struct bio_pair *bio_split(struct bio *bi, mempool_t *pool, int first_sectors)
{
 // kmem_cache_t is on include/linux/slab.h
 // struct bio_pair *bp = mempool_alloc(pool, GFP_NOIO);
 
 struct bio_pair *bp ;


 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] bio_split() : bp = UsbAllocMemC() , size=%d .. \n",
		sizeof(struct bio_pair)
		));
 bp = (struct bio_pair *) UsbAllocMemC(USB_MEM_C_S
			#ifdef USB_DBG_MEM_C_ON
 		  		,sizeof(struct bio_pair)
			#endif
			) ;

	if (!bp)
		return bp;
 
 //	BUG_ON(bi->bi_vcnt != 1);
 
 if (bi->bi_vcnt != 1) {
		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }
 
 //	BUG_ON(bi->bi_idx != 0);
 
 if (bi->bi_idx != 0) {
		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }
	atomic_set(&bp->cnt, 3);
	bp->error = 0;
	bp->bio1 = *bi;
	bp->bio2 = *bi;
	bp->bio2.bi_sector += first_sectors;
	bp->bio2.bi_size -= first_sectors << 9;
	bp->bio1.bi_size = first_sectors << 9;
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] bio_split() : Set Value : bp->bio1.bi_size=%d \n",
		bp->bio1.bi_size )) ;

	bp->bv1 = bi->bi_io_vec[0];
	bp->bv2 = bi->bi_io_vec[0];
	bp->bv2.bv_offset += first_sectors << 9;
	bp->bv2.bv_len -= first_sectors << 9;
	bp->bv1.bv_len = first_sectors << 9;

	bp->bio1.bi_io_vec = &bp->bv1;
	bp->bio2.bi_io_vec = &bp->bv2;

	bp->bio1.bi_end_io = bio_pair_end_1;
	bp->bio2.bi_end_io = bio_pair_end_2;

	bp->bio1.bi_private = bi;
	bp->bio2.bi_private = pool;

	return bp;
}
#endif // ndef USB_NO_VMM
*/
/*
static void *bio_pair_alloc(int gfp_flags, void *data)
{

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] bio_pair_alloc() : return = UsbAllocMemC() , size=%d .. \n" ,
	sizeof(struct bio_pair)
	));
 
 // return kmalloc(sizeof(struct bio_pair), gfp_flags);
 
 return (void *) UsbAllocMemC(USB_MEM_C_S
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct bio_pair)
		        #endif
           		) ;
}

static void bio_pair_free(void *bp, void *data)
{
	kfree(bp);
}
*/
/*
static void __init biovec_init_pools(void)
{
	int i, size, megabytes, pool_entries = BIO_POOL_SIZE;
	int scale = BIOVEC_NR_POOLS;

	megabytes = nr_free_pages() >> (20 - PAGE_SHIFT);

	
 //	find out where to start scaling
	
	if (megabytes <= 16)
		scale = 0;
	else if (megabytes <= 32)
		scale = 1;
	else if (megabytes <= 64)
		scale = 2;
	else if (megabytes <= 96)
		scale = 3;
	else if (megabytes <= 128)
		scale = 4;

	
	// scale number of entries
	 
	pool_entries = megabytes * 2;
	if (pool_entries > 256)
		pool_entries = 256;

	for (i = 0; i < BIOVEC_NR_POOLS; i++) {
		struct biovec_pool *bp = bvec_array + i;

		size = bp->nr_vecs * sizeof(struct bio_vec);

		bp->slab = kmem_cache_create(bp->name, size, 0,
						SLAB_HWCACHE_ALIGN, NULL, NULL);
		if (!bp->slab)
			panic("biovec: can't init slab cache\n");

		if (i >= scale)
			pool_entries >>= 1;

		bp->pool = mempool_create(pool_entries, mempool_alloc_slab,
					mempool_free_slab, bp->slab);
		if (!bp->pool)
			panic("biovec: can't init mempool\n");
	}
}
*/
/*
static int __init init_bio(void)
*/
int init_bio(void)
{
 /*
 // kmem_cache_t is on include/linux/slab.h
 bio_slab = kmem_cache_create("bio", sizeof(struct bio), 0,
					SLAB_HWCACHE_ALIGN, NULL, NULL);
 if (!bio_slab)
		panic("bio: can't create slab cache\n");
 bio_pool = mempool_create(BIO_POOL_SIZE, mempool_alloc_slab, mempool_free_slab, bio_slab);
 if (!bio_pool)
		panic("bio: can't create mempool\n");

 biovec_init_pools();

 bio_split_pool = mempool_create(BIO_SPLIT_ENTRIES, bio_pair_alloc, bio_pair_free, NULL);
 if (!bio_split_pool)
		panic("bio: can't create split pool\n");
 */
 return 0;
}
/*
subsys_initcall(init_bio);

EXPORT_SYMBOL(bio_alloc);
EXPORT_SYMBOL(bio_put);
EXPORT_SYMBOL(bio_endio);
EXPORT_SYMBOL(bio_init);
EXPORT_SYMBOL(__bio_clone);
EXPORT_SYMBOL(bio_clone);
EXPORT_SYMBOL(bio_phys_segments);
EXPORT_SYMBOL(bio_hw_segments);
EXPORT_SYMBOL(bio_add_page);
EXPORT_SYMBOL(bio_get_nr_vecs);
EXPORT_SYMBOL(bio_map_user);
EXPORT_SYMBOL(bio_unmap_user);
EXPORT_SYMBOL(bio_pair_release);
EXPORT_SYMBOL(bio_split);
EXPORT_SYMBOL(bio_split_pool);
*/

// B. S. Liaw TEST for compiler warning !!

