/*
 *  linux/fs/block_dev.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *  Copyright (C) 2001  Andrea Arcangeli <andrea@suse.de> SuSE
 */

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsbMsInit.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP


/*
#include <linux/config.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/fcntl.h>
*/
#include "posix_types.h" // B. S. Liaw add this
#include "fcntl.h"
/*
#include <linux/slab.h>
#include <linux/kmod.h>
#include <linux/major.h>
*/
#ifndef USBMS_ONLY_ONE_DISK
	#include "major.h"
#endif // ndef USBMS_ONLY_ONE_DISK
/*
#include <linux/devfs_fs_kernel.h>
#include <linux/smp_lock.h>
#include <linux/highmem.h>
#include <linux/blkdev.h>
*/
#include "bitops.h" // B. S. Liaw add for test_bit()
#include "blkdev.h"
/*
#include <linux/module.h>
#include <linux/blkpg.h>
#include <linux/buffer_head.h>
*/
#include "buffer_head.h"
/*
#include <linux/mpage.h>
#include <linux/mount.h>
#include <linux/uio.h>
#include <linux/namei.h>
#include <asm/uaccess.h>
*/

static sector_t max_block(struct block_device *bdev)
{
 sector_t retval = ~((sector_t)0);
 #ifndef NO_INODE
	loff_t sz = i_size_read(bdev->bd_inode);
 #endif // ndef NO_INODE 

 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] max_block() : sz=%d (%08XH)\n",
	sz, sz ));

 #ifndef NO_INODE
	if (sz) {
		/*
		#ifndef _ETCPIP
			unsigned int size = block_size(bdev);
			unsigned int sizebits = blksize_bits(size);
			USBMS_DBG_INIT(iprintf(
				"[USBMS INIT] max_block() : size=%d (%08XH)\n",
				size, size ));

			retval = (sz >> sizebits);
		 #else
		*/
			retval = sz ;
		/*
		#endif // _ETCPIP
		*/
	   }
 #endif // ndef NO_INODE 
 USBMS_DBG_INIT(iprintf("[USBMS INIT] max_block() : retval=%08XH \n",
	retval));
 return retval;
}

/* Kill _all_ buffers, dirty or not.. */
/*
#ifndef USB_NO_VMM // vitrtul memory management
static void kill_bdev(struct block_device *bdev)
{
 // invalidate_bdev() is on fs/buffer.c
 // invalidate_bdev(bdev, 1);
 #ifndef USB_NO_VMM // vitrtul memory management
	truncate_inode_pages(bdev->bd_inode->i_mapping, 0);
 #endif // ndef USB_NO_VMM
}	
#endif // ndef USB_NO_VMM
*/
int set_blocksize(struct block_device *bdev, int size)
{
 #ifndef INODE_REDUCED  
	int oldsize;
 #endif // ndef INODE_REDUCED  


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] set_blocksize() Begin .. \n"));
 /* Size must be a power of two, and between 512 and PAGE_SIZE */
 if (size > PAGE_SIZE || size < 512 || (size & (size-1)))
		return -EINVAL;

	/* Size cannot be smaller than the size supported by the device */
	if (size < bdev_hardsect_size(bdev))
		return -EINVAL;

 #ifndef INODE_REDUCED  
	oldsize = bdev->bd_block_size;
	if (oldsize == size)
 #endif // ndef INODE_REDUCED  
	   {
		return 0;
	   }

	/* Ok, we're actually changing the blocksize.. */
 // sync_blockdev() is on fs/buffer.c
	sync_blockdev(bdev);
 
 #ifndef INODE_REDUCED  
	bdev->bd_block_size = size;
 #endif // ndef INODE_REDUCED  
 // #ifndef NO_INODE
 #ifndef INODE_REDUCED  
	bdev->bd_inode->i_blkbits = blksize_bits(size);
 #endif // ndef INODE_REDUCED  
 // #endif // ndef NO_INODE
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	kill_bdev(bdev);
 #endif // ndef USB_NO_VMM
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] set_blocksize() End. \n"));
 return 0;
}
/*
EXPORT_SYMBOL(set_blocksize);
*/
int sb_set_blocksize(struct super_block *sb, int size)
{
 #ifndef INODE_REDUCED
	int bits;
 #endif // ndef INODE_REDUCED


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sb_set_blocksize() Begin .. \n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sb_set_blocksize() : set_blocksize() .. \n"));
 if (set_blocksize(sb->s_bdev, size) < 0)
		return 0;
	sb->s_blocksize = size;
 #ifndef INODE_REDUCED
	for (bits = 9, size >>= 9; size >>= 1; bits++)
		;
	sb->s_blocksize_bits = bits;
 #endif // ndef INODE_REDUCED

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sb_set_blocksize() End. \n"));
 return sb->s_blocksize;
}
/*
EXPORT_SYMBOL(sb_set_blocksize);
*/
/*
// B. S. Liaw : None Call this ??
int sb_min_blocksize(struct super_block *sb, int size)
{
 int minsize = bdev_hardsect_size(sb->s_bdev);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sb_min_blocksize() Begin .. \n"));
 if (size < minsize)
		size = minsize;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sb_min_blocksize() End with Call sb_set_blocksize() .. \n"));
 return sb_set_blocksize(sb, size);
}
*/
/*
EXPORT_SYMBOL(sb_min_blocksize);
*/
#ifndef NO_INODE
#if _USB2 > 0	// reduce function call layer
int
#else	// _USB2
static int
#endif	// _USB2
blkdev_get_block(struct inode *inode, sector_t iblock,
		struct buffer_head *bh, int create)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blkdev_get_block() Begin .. \n"));
 USBMS_DBG_MEM(iprintf("[USBMS MEM] blkdev_get_block() : iblock=%08XH \n",
	iblock));

 
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blkdev_get_block() : max_block() .. \n"));
 if(iblock >= max_block(inode->i_bdev)) {
	#ifdef USBMS_BS_ASSERT
		iprintf("[ASSERT][USBMS] blkdev_get_block() : " ) ;
	    iprintf("Error : iblock>max_block !!\n" ) ;
	    iprintf("[ASSERT][USBMS] iblock=%08XH \n", iblock ) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	#endif // USBMS_BS_ASSERT
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blkdev_get_block() End with Failed !! \n"));
	return -EIO;
   }

 bh->b_bdev = inode->i_bdev;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] blkdev_get_block() : Set Value : bh->b_bdev=%08XH \n",
	bh->b_bdev
	));

 bh->b_blocknr = iblock;
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] blkdev_get_block() : Set Value : bh->b_blocknr=%d \n",
	bh->b_blocknr));
 /*
 set_buffer_mapped(bh);
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blkdev_get_block() End. \n"));
 // return 0;
 return 1 ;
}
#endif // ndef NO_INODE


/*
#ifndef NO_INODE
static int
blkdev_get_blocks(struct inode *inode, sector_t iblock,
		unsigned long max_blocks, struct buffer_head *bh, int create)
{
 if ((iblock + max_blocks) > max_block(inode->i_bdev))
		return -EIO;

 bh->b_bdev = inode->i_bdev;
 bh->b_blocknr = iblock;
 // #ifndef NO_INODE
	bh->b_size = max_blocks << inode->i_blkbits;
 // #endif // ndef NO_INODE
 
 // set_buffer_mapped(bh);
 
 return 0;
}
#endif // ndef NO_INODE
*/
/*
static int
blkdev_direct_IO(int rw, struct kiocb *iocb, const struct iovec *iov,
			loff_t offset, unsigned long nr_segs)
{
 #ifndef _ETCPIP
	struct file *file = iocb->ki_filp;
 #endif // ndef _ETCPIP
 #ifndef NO_INODE
	struct inode *inode = file->f_dentry->d_inode->i_mapping->host;

	return blockdev_direct_IO(rw, iocb, inode, inode->i_bdev, iov, offset,
				nr_segs, blkdev_get_blocks, NULL);
 #endif // ndef NO_INODE
}
*/
#if _USB2 > 0	// reduce function call layer
#else	// _USB2
/*
#ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	static int blkdev_writepage(struct page *page, struct writeback_control *wbc)
 #else
*/
static int blkdev_writepage(struct inode *inode,
				sector_t iblock , // block number on Disk Partition
				unsigned int num_sectors ,
				unsigned char *buf )
/*
#endif
*/
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blkdev_writepage() Begin .. \n"));

 #ifndef USBMS_BLK_CACHE // 2004/1/15 B. S. Liaw
	#ifdef USBMS_BS_ASSERT
		if ( !(USB_IS_UN_CACHED(buf)) ) {
			iprintf("[ASSERT][USBMS] blkdev_writepage() : " ) ;
		    iprintf("Error : Not Un-Cached !!\n" ) ;
		    iprintf("[ASSERT][USBMS] buf=%08XH \n", buf ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	#endif // USBMS_BS_ASSERT
 #endif // ndef USBMS_BLK_CACHE

 USBMS_DBG_MEM(iprintf("[USBMS MEM] blkdev_writepage() : iblock=%08XH \n",
	iblock));

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blkdev_writepage() End with Call block_write_full_page() .. \n"));
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	return block_write_full_page(page, blkdev_get_block, wbc);
  #else
 */
 return block_write_full_page(inode,iblock,
			num_sectors ,
			buf, blkdev_get_block
			/*
			#ifndef NO_VFS
				, wbc
			#endif // ndef NO_VFS
			*/
			);
 /*
 #endif
 */
}
/*
#ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	static int blkdev_readpage(struct file * file, struct page * page)
 #else
*/
static int blkdev_readpage(struct inode *inode,
				sector_t iblock , // block number on Disk Partition
				unsigned int num_sectors ,
				unsigned char *buf )
/*
#endif
*/
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blkdev_readpage() Begin .. \n"));

 #ifndef USBMS_BLK_CACHE // 2004/1/15 B. S. Liaw
	 #ifdef USBMS_BS_ASSERT
		if ( !(USB_IS_UN_CACHED(buf)) ) {
			iprintf("[ASSERT][USBMS] blkdev_readpage() : " ) ;
		    iprintf("Error : Not Un-Cached !!\n" ) ;
	    	iprintf("[ASSERT][USBMS] buf=%08XH \n", buf ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	#endif // USBMS_BS_ASSERT
 #endif // ndef USBMS_BLK_CACHE

 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] blkdev_readpage() End with Call block_read_full_page() .. \n"
	));
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	return block_read_full_page(page, blkdev_get_block);
  #else
 */
 return block_read_full_page(inode,iblock,
							num_sectors ,
							buf, blkdev_get_block);
 /*
 #endif
 */
}
#endif	// _USB2
/*
#ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	static int blkdev_prepare_write(struct file *file, struct page *page, unsigned from, unsigned to)
 #else

	static int blkdev_prepare_write(struct file *file, unsigned char *buf,
			unsigned from, unsigned to)

#endif

{
 #ifndef INODE_REDUCED  
	return block_prepare_write(page, from, to, blkdev_get_block);
 #endif // ndef INODE_REDUCED  
}
*/
/*
#ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	static int blkdev_commit_write(struct file *file, struct page *page, unsigned from, unsigned to)
 #else
	static int blkdev_commit_write(struct file *file, unsigned char *buf,
			unsigned from, unsigned to)
#endif
{
 #ifndef USB_NO_VMM // vitrtul memory management
	return block_commit_write(page, from, to);
  #else
	// return block_commit_write(buf, from, to);
 #endif
}
*/
/*
 * private llseek:
 * for a block special file file->f_dentry->d_inode->i_size is zero
 * so we compute the size by hand (just as in block_read/write above)
 */
/*
static loff_t block_llseek(struct file *file, loff_t offset, int origin)
{
 #ifndef NO_INODE
	struct inode *bd_inode;
 #endif // ndef NO_INODE
 loff_t size;
 loff_t retval;
 // for f_dentry
 #ifndef NO_INODE
	bd_inode = file->f_dentry->d_inode->i_bdev->bd_inode;
 #endif // ndef NO_INODE
 
 #ifndef INODE_REDUCED  
	down(&bd_inode->i_sem);
 #endif // ndef INODE_REDUCED  
 #ifndef NO_INODE
	size = i_size_read(bd_inode);
 #endif // ndef NO_INODE

	switch (origin) {
		case 2:
			offset += size;
			break;
		case 1:
			offset += file->f_pos;
	}
	retval = -EINVAL;
	if (offset >= 0 && offset <= size) {
		if (offset != file->f_pos) {
			file->f_pos = offset;
		}
		retval = offset;
	}
 #ifndef INODE_REDUCED  
	up(&bd_inode->i_sem);
 #endif // ndef INODE_REDUCED  
	return retval;
}
*/	
/*
 *	Filp may be NULL when we are called by an msync of a vma
 *	since the vma has no handle.
 */
/* 
static int block_fsync(struct file *filp, struct dentry *dentry, int datasync)
{
 #ifndef NO_INODE
	struct inode * inode = dentry->d_inode;

	// sync_blockdev() is on fs/buffer.c
	return sync_blockdev(inode->i_bdev);
 #endif // ndef NO_INODE
}
*/
/*
 * pseudo-fs
 */
/*
#ifndef NO_SPINLOCK
	static spinlock_t bdev_lock __cacheline_aligned_in_smp = SPIN_LOCK_UNLOCKED;
#endif // ndef NO_SPINLOCK
*/
/* kmem_cache_t is on include/linux/slab.h
static kmem_cache_t * bdev_cachep;
*/
#ifndef NO_INODE
struct bdev_inode {
	struct block_device bdev;
	struct inode vfs_inode;
};

static inline struct bdev_inode *BDEV_I(struct inode *inode)
{
	return container_of(inode, struct bdev_inode, vfs_inode);
}

static struct inode *bdev_alloc_inode(struct super_block *sb)
{
 /* kmem_cache_t is on include/linux/slab.h
 struct bdev_inode *ei = kmem_cache_alloc(bdev_cachep, SLAB_KERNEL);
 */
 struct bdev_inode *ei ;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] bdev_alloc_inode() Begin .. \n"));
 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] bdev_alloc_inode() : ei = UsbAllocMemC() .. \n"
		));

 ei = (struct bdev_inode *) UsbAllocMemC(USB_MEM_C_S
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct bdev_inode) 
		        #endif
           		) ;

 if (!ei)
		return NULL;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] bdev_alloc_inode() End. \n"));
 return &ei->vfs_inode;
}
#endif // ndef NO_INODE

#ifndef NO_INODE
static void bdev_destroy_inode(struct inode *inode)
{
 /* kmem_cache_t is on include/linux/slab.h
	kmem_cache_free(bdev_cachep, BDEV_I(inode));
 */


 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] bdev_destroy_inode() : UsbFreeMemC(BDEV_I(inode)) .. \n"
		));
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) BDEV_I(inode)
		#if defined(USB_DBG_MEM_C_ON)
			, sizeof(struct bdev_inode)
		#endif
		);
}
#endif // ndef NO_INODE

/* kmem_cache_t is on include/linux/slab.h
static void init_once(void * foo, kmem_cache_t * cachep, unsigned long flags)
{
 #ifndef INODE_REDUCED  
	struct bdev_inode *ei = (struct bdev_inode *) foo;
 #endif // ndef INODE_REDUCED  
	struct block_device *bdev = &ei->bdev;

	if ((flags & (SLAB_CTOR_VERIFY|SLAB_CTOR_CONSTRUCTOR)) ==
	    SLAB_CTOR_CONSTRUCTOR)
	{
		memset(bdev, 0, sizeof(*bdev));
 #ifndef NO_SEMPHORE
		sema_init(&bdev->bd_sem, 1);
 #endif // ndef NO_SEMPHORE
 #ifndef INODE_REDUCED  
		INIT_LIST_HEAD(&bdev->bd_inodes);
 #endif // ndef INODE_REDUCED  
		INIT_LIST_HEAD(&bdev->bd_list);
 #ifndef INODE_REDUCED  
		inode_init_once(&ei->vfs_inode);
 #endif // ndef INODE_REDUCED  
	}
}
*/

#ifndef NO_INODE
   static inline void __bd_forget(struct inode *inode)
   {
	#ifndef INODE_REDUCED  
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __bd_forget() : list_del_init() .. \n"));
		list_del_init(&inode->i_devices);
	#endif // ndef INODE_REDUCED  
	inode->i_bdev = NULL;
	USBMS_DBG_MEM(iprintf("[USBMS MEM] __bd_forget() : Set Value : inode->i_bdev=%08XH \n",
		inode->i_bdev));
	#ifndef NO_INODE
		inode->i_mapping = &inode->i_data;
	#endif // ndef NO_INODE
   }
#endif // ndef NO_INODE

#ifndef NO_INODE
static void bdev_clear_inode(struct inode *inode)
{
 struct block_device *bdev = &BDEV_I(inode)->bdev;
 struct list_head *p;
 /*
 #ifndef NO_SPINLOCK
	spin_lock(&bdev_lock);
 #endif // ndef NO_SPINLOCK
 */
 #ifndef INODE_REDUCED  
	while ( (p = bdev->bd_inodes.next) != &bdev->bd_inodes ) {
		__bd_forget(list_entry(p, struct inode, i_devices));
	}
 #endif // ndef INODE_REDUCED  
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] bdev_clear_inode() : list_del_init() .. \n"));
 list_del_init(&bdev->bd_list);
 /*
 #ifndef NO_SPINLOCK
	spin_unlock(&bdev_lock);
 #endif // ndef NO_SPINLOCK
 */
}

static struct super_operations bdev_sops = {
 #ifndef INODE_REDUCED  
	.statfs = simple_statfs,
 #endif // ndef INODE_REDUCED  
 .alloc_inode = bdev_alloc_inode,
 .destroy_inode = bdev_destroy_inode,
 #ifndef INODE_REDUCED  
	.drop_inode = generic_delete_inode,
 #endif // ndef INODE_REDUCED  
 .clear_inode = bdev_clear_inode,
};
#endif // ndef NO_INODE
/*
static struct super_block *bd_get_sb(struct file_system_type *fs_type,
	int flags, const char *dev_name, void *data)
{
	return get_sb_pseudo(fs_type, "bdev:", &bdev_sops, 0x62646576);
}

static struct file_system_type bd_type = {
 #ifdef USB_DBGSTRING_ON
	.name		= "bdev",
 #endif // USB_DBGSTRING_ON 
	.get_sb		= bd_get_sb,
	.kill_sb	= kill_anon_super,
};
*/
/*
#ifndef NO_VFS  // struct vfsmount
	static struct vfsmount *bd_mnt;
	struct super_block *blockdev_superblock;
#else
*/
	struct super_block blockdev_superblock;
/*
#endif
*/
/*
void __init bdev_cache_init(void)
*/
void bdev_cache_init(void)
{
 /*
 int err;
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] bdev_cache_init() Begin .. \n"));
 /*
 // kmem_cache_t is on include/linux/slab.h
 bdev_cachep = kmem_cache_create("bdev_cache",
					sizeof(struct bdev_inode),
					0,
					SLAB_HWCACHE_ALIGN|SLAB_RECLAIM_ACCOUNT,
					init_once,
					NULL);
 // kmem_cache_t is on include/linux/slab.h
 if (!bdev_cachep)
		panic("Cannot create bdev_cache SLAB cache");
 */
 /* 
	err = register_filesystem(&bd_type);

	if (err)
		panic("Cannot register bdev pseudo-fs");
 */
 /*
 #ifndef NO_VFS  // struct vfsmount
	bd_mnt = kern_mount(&bd_type);
	err = PTR_ERR(bd_mnt);
	if (IS_ERR(bd_mnt))
		panic("Cannot create bdev pseudo-fs");
	blockdev_superblock = bd_mnt->mnt_sb;	// For writeback 
  #else
 */
	// B. S. Liaw : for sb->s_op->alloc_inode
	blockdev_superblock.s_op = &bdev_sops ;
 /*
 #endif
 */

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] bdev_cache_init() End. \n"));
}

/*
 * Most likely _very_ bad one - but then it's hardly critical for small
 * /dev and can be fixed when somebody will need really large one.
 * Keep in mind that it will be fed through icache hash function too.
 */
static inline unsigned long hash(dev_t dev)
{
 return MAJOR(dev)+MINOR(dev);
}

#ifndef NO_INODE
static int bdev_test(struct inode *inode, void *data)
{
	return BDEV_I(inode)->bdev.bd_dev == *(dev_t *)data;
}
#endif // ndef NO_INODE

#ifndef NO_INODE
static int bdev_set(struct inode *inode, void *data)
{
	BDEV_I(inode)->bdev.bd_dev = *(dev_t *)data;
	return 0;
}
#endif // ndef NO_INODE

static LIST_HEAD(all_bdevs);

struct block_device *bdget(dev_t dev)
{
 struct block_device *bdev;
 #ifndef NO_INODE
	struct inode *inode;
 #endif // ndef NO_INODE


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] bdget() Begin .. \n"));

 #ifndef NO_INODE
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] bdget() : iget5_locked() .. \n"));
	#ifndef NO_VFS  // struct vfsmount
		inode = iget5_locked(bd_mnt->mnt_sb, hash(dev),
			bdev_test, bdev_set, &dev);
	 #else
		inode = iget5_locked(&blockdev_superblock, hash(dev),
			bdev_test, bdev_set, &dev);
	#endif

	USBMS_DBG_MEM(iprintf("[USBMS MEM] bdget() : inode=%08XH \n",inode));
	if (!inode)
		return NULL;
 #endif // ndef NO_INODE

 #ifndef NO_INODE
	bdev = &BDEV_I(inode)->bdev;
 #endif // ndef NO_INODE
 USBMS_DBG_MEM(iprintf("[USBMS MEM] bdget() : bdev=%08XH \n",bdev));

 #ifndef NO_INODE
	if (inode->i_state & I_NEW) {
		USBMS_DBG_BLOCK(iprintf("[USBMS BLOCK] bdget() : I_NEW .. \n"));
		bdev->bd_contains = NULL;
		bdev->bd_inode = inode;
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] bdget() : Set Value : bdev->bd_inode=%08XH \n",
			bdev->bd_inode));
		// #ifndef NO_INODE
		#ifndef INODE_REDUCED  
			bdev->bd_block_size = (1 << inode->i_blkbits);
		#endif // ndef INODE_REDUCED  
		// #endif // ndef NO_INODE
		bdev->bd_part_count = 0;
		bdev->bd_invalidated = 0;
		inode->i_mode = S_IFBLK;
		inode->i_rdev = dev;
		inode->i_bdev = bdev;
	USBMS_DBG_MEM(iprintf("[USBMS MEM] bdget() : Set Value : inode->i_bdev=%08XH \n",
		inode->i_bdev));
		inode->i_data.a_ops = &def_blk_aops;
		#ifndef USB_NO_VMM // vitrtul memory management
			mapping_set_gfp_mask(&inode->i_data, GFP_USER);
		#endif // ndef USB_NO_VMM
		#ifndef NO_BACKDEV 
			inode->i_data.backing_dev_info = &default_backing_dev_info;
		#endif // ndef NO_BACKDEV 
		/*
		#ifndef NO_SPINLOCK
			spin_lock(&bdev_lock);
		#endif // ndef NO_SPINLOCK
		*/
		list_add(&bdev->bd_list, &all_bdevs);
		/*
		#ifndef NO_SPINLOCK
			spin_unlock(&bdev_lock);
		#endif // ndef NO_SPINLOCK
		*/
		// #ifndef INODE_REDUCED  
			unlock_new_inode(inode);
		// #endif // ndef INODE_REDUCED  
	   }
 #endif // ndef NO_INODE

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] bdget() End. \n"));
 return bdev;
}
/*
EXPORT_SYMBOL(bdget);
*/
/*
long nr_blockdev_pages(void)
{
	struct list_head *p;
	long ret = 0;

 #ifndef NO_SPINLOCK
	spin_lock(&bdev_lock);
 #endif // ndef NO_SPINLOCK

	list_for_each(p, &all_bdevs) {
		struct block_device *bdev;
		bdev = list_entry(p, struct block_device, bd_list);
 #ifndef USB_NO_VMM // vitrtul memory management
		ret += bdev->bd_inode->i_mapping->nrpages;
 #endif // ndef USB_NO_VMM
	}

 #ifndef NO_SPINLOCK
	spin_unlock(&bdev_lock);
 #endif // ndef NO_SPINLOCK
 
	return ret;
}
*/
void bdput(struct block_device *bdev)
{
 /*
 #ifndef NO_INODE
	iput(bdev->bd_inode);
 #endif // ndef NO_INODE
 */
}
/*
EXPORT_SYMBOL(bdput);
*/ 
/*
#ifndef NO_INODE
int bd_acquire(struct inode *inode)
{
 struct block_device *bdev;
 
 #ifndef NO_SPINLOCK
	spin_lock(&bdev_lock);
 #endif // ndef NO_SPINLOCK
 
 #ifndef INODE_REDUCED
	// INODE_REDUCED : igrab()
	if (inode->i_bdev && igrab(inode->i_bdev->bd_inode)) {
		
		#ifndef NO_SPINLOCK
			spin_unlock(&bdev_lock);
		#endif // ndef NO_SPINLOCK
		
		return 0;
	}
 #endif // ndef INODE_REDUCED  
 
 #ifndef NO_SPINLOCK
	spin_unlock(&bdev_lock);
 #endif // ndef NO_SPINLOCK
 
 #ifndef NO_INODE
	bdev = bdget(inode->i_rdev);
	if (!bdev)
		return -ENOMEM;
 #endif // ndef NO_INODE

 #ifndef NO_SPINLOCK
	spin_lock(&bdev_lock);
 #endif // ndef NO_SPINLOCK
 
 #ifndef NO_INODE
	if (inode->i_bdev)
		__bd_forget(inode);
	inode->i_bdev = bdev;
 #endif // ndef NO_INODE
 #ifndef NO_INODE
	inode->i_mapping = bdev->bd_inode->i_mapping;
 #endif // ndef NO_INODE
 #ifndef INODE_REDUCED  
	list_add(&inode->i_devices, &bdev->bd_inodes);
 #endif // ndef INODE_REDUCED  

#ifndef NO_SPINLOCK
	spin_unlock(&bdev_lock);
#endif // ndef NO_SPINLOCK

	return 0;
}
#endif // ndef NO_INODE
*/
/* Call when you free inode */

#ifndef NO_INODE
void bd_forget(struct inode *inode)
{
/*
#ifndef NO_SPINLOCK
	spin_lock(&bdev_lock);
#endif // ndef NO_SPINLOCK
*/
 #ifndef NO_INODE
	if (inode->i_bdev)
		__bd_forget(inode);
 #endif // ndef NO_INODE
/*
#ifndef NO_SPINLOCK
	spin_unlock(&bdev_lock);
#endif // ndef NO_SPINLOCK
*/
}
#endif // ndef NO_INODE
/*
int bd_claim(struct block_device *bdev, void *holder)
{
	int res;

 #ifndef NO_SPINLOCK
	spin_lock(&bdev_lock);
 #endif // ndef NO_SPINLOCK


	// first decide result 
	if (bdev->bd_holder == holder)
		res = 0;	 // already a holder 
	else if (bdev->bd_holder != NULL)
		res = -EBUSY; 	 // held by someone else 
	else if (bdev->bd_contains == bdev)
		res = 0;  	 // is a whole device which isn't held 

	else if (bdev->bd_contains->bd_holder == bd_claim)
		res = 0; 	 // is a partition of a device that is being partitioned 
	else if (bdev->bd_contains->bd_holder != NULL)
		res = -EBUSY;	 // is a partition of a held device 
	else
		res = 0;	 // is a partition of an un-held device 

	// now impose change 
	if (res==0) {
		// * note that for a whole device bd_holders
		// * will be incremented twice, and bd_holder will
		// * be set to bd_claim before being set to holder
		 
		bdev->bd_contains->bd_holders ++;
		bdev->bd_contains->bd_holder = bd_claim;
		bdev->bd_holders++;
		bdev->bd_holder = holder;
	}

#ifndef NO_SPINLOCK
	spin_unlock(&bdev_lock);
#endif // ndef NO_SPINLOCK

	return res;
}
*/
/*
EXPORT_SYMBOL(bd_claim);
*/
void bd_release(struct block_device *bdev)
{
/*
#ifndef NO_SPINLOCK
	spin_lock(&bdev_lock);
#endif // ndef NO_SPINLOCK
*/
	if (!--bdev->bd_contains->bd_holders)
		bdev->bd_contains->bd_holder = NULL;
	if (!--bdev->bd_holders)
		bdev->bd_holder = NULL;
/*
#ifndef NO_SPINLOCK
	spin_unlock(&bdev_lock);
#endif // ndef NO_SPINLOCK
*/
}
/*
EXPORT_SYMBOL(bd_release);
*/
/*
 * Tries to open block device by device number.  Use it ONLY if you
 * really do not have anything better - i.e. when you are behind a
 * truly sucky interface and all you are given is a device number.  _Never_
 * to be used for internal purposes.  If you ever need it - reconsider
 * your API.
 */

struct block_device *open_by_devnum(dev_t dev, unsigned mode, int kind)
{
 struct block_device *bdev = bdget(dev);
 int err = -ENOMEM;
 int flags = mode & FMODE_WRITE ? O_RDWR : O_RDONLY;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] open_by_devnum() Begin .. \n"));
 if (bdev)
   {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] open_by_devnum(1) : blkdev_get() .. \n"));
	err = blkdev_get(bdev, mode, flags, kind);
   }
 
 // return err ? ERR_PTR(err) : bdev;
 // 2003/12/26 B. S. Liaw BugFix ..
 if(err) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] open_by_devnum(2) End with Failed !! \n"));
	dprintf("[USBMS] Error : Open Logical Disk Failed !! \n") ;
	return (struct block_device *) NULL ;
   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] open_by_devnum() End. \n"));
 return bdev;
}

/*
EXPORT_SYMBOL(open_by_devnum);
*/
/*
 * This routine checks whether a removable media has been changed,
 * and invalidates all buffer-cache-entries in that case. This
 * is a relatively slow routine, so we have to try to minimize using
 * it. Thus it is called only upon a 'mount' or 'open'. This
 * is the best way of combining speed and utility, I think.
 * People changing diskettes in the middle of an operation deserve
 * to lose :-)
 */
#ifndef USBMS_NO_REMOVAL
int check_disk_change(struct block_device *bdev)
{
 struct gendisk *disk = bdev->bd_disk;
 struct block_device_operations * bdops = disk->fops;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] check_disk_change() Begin .. \n"));
 if (!bdops->media_changed)
		return 0;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] check_disk_change() : bdops->media_changed() .. \n"));
 if (!bdops->media_changed(bdev->bd_disk))
		return 0;

 #ifndef NO_INODE
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] check_disk_change() : __invalidate_device() .. \n"));
	// __invalidate_device() is on fs/inode.c
	if (__invalidate_device(bdev, 0))
		printk("VFS: busy inodes on changed media.\n");
 #endif // ndef NO_INODE

 if (bdops->revalidate_disk)
   {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] check_disk_change() : bdops->revalidate_disk() .. \n"));
	bdops->revalidate_disk(bdev->bd_disk);
   }
 if (bdev->bd_disk->minors > 1)
		bdev->bd_invalidated = 1;
	return 1;
}
#endif // ndef USBMS_NO_REMOVAL
/*
EXPORT_SYMBOL(check_disk_change);
*/
static void bd_set_size(struct block_device *bdev, loff_t size)
{
 unsigned bsize = bdev_hardsect_size(bdev);


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] bd_set_size() Begin .. \n"));
 USBMS_DBG_INIT(iprintf("[USBMS INIT] bd_set_size() : size=%08XH \n",
	size));
 USBMS_DBG_INIT(iprintf("[USBMS INIT] bd_set_size() : size=%08lXH \n",
	size));
 USBMS_DBG_INIT(iprintf("[USBMS INIT] bd_set_size() : size=%u \n",
	size));
 #ifndef NO_INODE
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] bd_set_size() : i_size_write() .. \n")) ;
	i_size_write(bdev->bd_inode, size);
 #endif // ndef NO_INODE
 /*
	while (bsize < PAGE_CACHE_SIZE) {
		if (size & bsize)
			break;
		bsize <<= 1;
	}
 */
 #ifndef INODE_REDUCED  
	bdev->bd_block_size = bsize;
 #endif // ndef INODE_REDUCED  
 // #ifndef NO_INODE
 #ifndef INODE_REDUCED  
	bdev->bd_inode->i_blkbits = blksize_bits(bsize);
 #endif // ndef INODE_REDUCED  
 // #endif // ndef NO_INODE
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] bd_set_size() End. \n"));
}

#ifndef NO_INODE

static int do_open(struct block_device *bdev, struct inode *inode, struct file *file)
{
 /*
 #ifndef _ETCPIP
	struct module *owner = NULL;
 #endif // ndef _ETCPIP
 */
 
 // #ifndef _ETCPIP
	struct gendisk *disk;
 // #endif
 
 int ret = -ENXIO;
 int part;
 loff_t bd_size ; // B. S. Liaw add this
 #ifdef USBMS_ONLY_ONE_DISK
	int x ;
 #endif // USBMS_ONLY_ONE_DISK


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open() Begin .. \n"));
 /*
 #ifndef _ETCPIP
	 lock_kernel();
 #endif
 */
 
 USBMS_DBG_MEM(iprintf("[USBMS MEM] do_open(1) : inode=%08XH \n",
	inode));
 /*
 #ifndef _ETCPIP
 */
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(1) : get_gendisk() .. \n"));
	disk = get_gendisk(bdev->bd_dev, &part);
 /*
  #else
	part = 0 ;
 #endif
 */
 if(!disk) {
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] do_open() : Error : get_gendisk() Failed !! \n"));
	/*
	#ifndef _ETCPIP
		unlock_kernel();
	#endif // ndef _ETCPIP
	*/
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(2) : bdput() .. \n"));
	bdput(bdev);
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(3) End with Failed !! \n"));
	return ret;
   }

 /*
 #ifndef _ETCPIP
	owner = disk->fops->owner;
 #endif // ndef _ETCPIP
 */
 /*
 #ifndef NO_SEMPHORE
	down(&bdev->bd_sem);
 #endif // ndef NO_SEMPHORE
 */
 USBMS_DBG_INIT(iprintf("[USBMS INIT] do_open() : bdev->bd_openers=%d \n",
	bdev->bd_openers
	));

 if (!bdev->bd_openers) {
	bdev->bd_disk = disk;
	USBMS_DBG_MEM(iprintf("[USBMS MEM] do_open(2) : Set Value : bdev->bd_disk=%08XH \n",bdev->bd_disk));

	bdev->bd_contains = bdev;

	USBMS_DBG_INIT(iprintf("[USBMS INIT] do_open() : part=%d \n",part));
	if (!part) {
		/*
		#ifndef NO_BACKDEV 
			struct backing_dev_info *bdi;
		#endif // ndef NO_BACKDEV 
		*/
		USBMS_DBG_INFO(iprintf("[USBMS INFO] do_open() : Not Check Partition yet !! \n"));

		#ifdef USBMS_ONLY_ONE_DISK
			// on fs/check.c
			for(x=0;x<USBMS_MAX_MINORS+1;x++) {
				disk_global_bdev[x] = NULL ;
			   }

		#endif // USBMS_ONLY_ONE_DISK


		USBMS_DBG_MEM(iprintf("[USBMS MEM] do_open(3) : disk->fops=%08XH \n",
			disk->fops));
		USBMS_DBG_MEM(iprintf("[USBMS MEM] do_open(4) : disk->fops->open=%08XH \n",
			disk->fops->open));
		if (disk->fops->open) {
			#ifndef NO_INODE
				USBMS_DBG_FLOW(iprintf(
					"[USBMS FLOW] do_open(4) : disk->fops->open() .. \n"
					));
				ret = disk->fops->open(inode, file);
				if (ret)
					goto out_first;
			#endif // ndef NO_INODE
		   }
		USBMS_DBG_INIT(iprintf(
			"[USBMS INIT] do_open() : bdev->bd_openers=%d \n",
			bdev->bd_openers
			));
		if (!bdev->bd_openers) {
			USBMS_DBG_INIT(iprintf(
				"[USBMS INIT] do_open() : get_capacity(disk)=%d \n",
				get_capacity(disk)
				));
			USBMS_DBG_INIT(iprintf(
				"[USBMS INIT] do_open() : get_capacity(disk) * %d =%lu\n",
				disk->queue->hardsect_size, get_capacity(disk) * disk->queue->hardsect_size
				));
			/*
			// B. S. Liaw : I donot know why parameter size become 0 ??
			// B. S. Liaw : this is because original loff_t is long long
			//              Now change to long on UsOsport.h line # 31
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(5) : bd_set_size() .. \n"));
			bd_set_size(bdev,(loff_t)get_capacity(disk)<<9);
			*/
			// B. S. Liaw : So, I rewrite into the following ..
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(6) : get_capacity() .. \n"));
			bd_size = (loff_t) get_capacity(disk); // B. S. Liaw
			// bd_size = get_capacity(disk) ; // B. S. Liaw
			USBMS_DBG_INIT(iprintf("[USBMS INIT] do_open() : bd_size=%ld \n",
				bd_size
				));
/*			bd_size *= disk->queue->hardsect_size; // B. S. Liaw
			USBMS_DBG_INIT(iprintf("[USBMS INIT] do_open() : bd_size=%lu \n",
				bd_size
				));
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(7) : bd_set_size() .. \n"));
 */
			bd_set_size(bdev,bd_size);

			/*
			#ifndef NO_BACKDEV 
				bdi = blk_get_backing_dev_info(bdev);
				if (bdi == NULL)
					bdi = &default_backing_dev_info;
				bdev->bd_inode->i_data.backing_dev_info = bdi;
			#endif // ndef NO_BACKDEV 
			*/
		   }
		#ifndef USBMS_ONLY_ONE_DISK
			if (bdev->bd_invalidated) {
				USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(8) : rescan_partitions() .. \n"));
				/*
				#ifndef USB_NO_VMM // vitrtul memory management
					rescan_partitions(disk, bdev);
				 #else
				*/
					rescan_partitions(bdev->bd_inode,disk, bdev);
				/*
				#endif
				*/
			   }
		#endif // ndef USBMS_ONLY_ONE_DISK
		//  no part
	   } else {
		//  has part
		struct hd_struct *p;
		#ifndef USBMS_ONLY_ONE_DISK
			struct block_device *whole;
		#endif // ndef USBMS_ONLY_ONE_DISK

		USBMS_DBG_INIT(iprintf("[USBMS INIT] do_open() : Partition Of One Disk .. \n"));
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(9) : bdget_disk() .. \n"));
		whole = bdget_disk(disk, 0);
		USBMS_DBG_MEM(iprintf("[USBMS MEM] do_open(5) : whole=%08XH \n",whole));
		ret = -ENOMEM;
		if (!whole)
				goto out_first;
		/*
		#ifndef INODE_REDUCED  
			// B. S. Liaw : Run this lead to recursive
			//				call : blkdev_get()->do_open()->blkdev_get()
			USBMS_DBG_INFO(iprintf("[USBMS INFO] do_open() FixMe : Eror : Infinite Recursive Call !! \n"));
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(10) : blkdev_get() .. \n"));
			// #ifndef _ETCPIP
				ret = blkdev_get(whole, file->f_mode, file->f_flags, BDEV_RAW);
			// #else
			//	ret = blkdev_get(whole, file->f_mode, file->f_flags, BDEV_RAW,disk);
			// #endif
			if (ret)
				goto out_first;
		#endif // ndef INODE_REDUCED
		*/
		bdev->bd_contains = whole;
		/*
		#ifndef NO_SEMPHORE
			down(&whole->bd_sem);
		#endif // ndef NO_SEMPHORE
		*/
		whole->bd_part_count++;

		p = disk->part[part - 1];
		/*
		#ifndef NO_BACKDEV 
				bdev->bd_inode->i_data.backing_dev_info =
				   whole->bd_inode->i_data.backing_dev_info;
		#endif // ndef NO_BACKDEV 
		*/
		if (!(disk->flags & GENHD_FL_UP) || !p || !p->nr_sects) {
				whole->bd_part_count--;
				#ifndef NO_SEMPHORE
					up(&whole->bd_sem);
				#endif // ndef NO_SEMPHORE
				ret = -ENXIO;
				goto out_first;
			   }
		kobject_get(&p->kobj);
		bdev->bd_part = p;
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(11) : bd_set_size() .. \n"));
#if 1  /* Charles 20040524 */
		USBMS_DBG_INIT(iprintf("[USBMS INIT] do_open(11) : bd_size=%ld \n",
				p->nr_sects
				));
		bd_set_size(bdev, (loff_t) p->nr_sects);
#else
		bd_set_size(bdev, (loff_t) p->nr_sects << 9);
#endif
		/*
		#ifndef NO_SEMPHORE
			up(&whole->bd_sem);
		#endif // ndef NO_SEMPHORE
		*/
		//  has part
	   }
	// no bdev->bd_openers
   } else {
	// has bdev->bd_openers
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(12) : put_disk() .. \n"));
	put_disk(disk);
	/*
	#ifndef _ETCPIP
		module_put(owner);
	#endif // ndef _ETCPIP
	*/
	USBMS_DBG_INIT(iprintf("[USBMS INIT] do_open() : bdev=%08XH \n",
		bdev
		));
	USBMS_DBG_INIT(iprintf("[USBMS INIT] do_open() : bdev->bd_contains=%08XH \n",
		bdev->bd_contains
		));
	if (bdev->bd_contains == bdev) {
		USBMS_DBG_INIT(iprintf(
			"[USBMS INIT] do_open() : bdev->bd_disk=%08XH \n",
			bdev->bd_disk
			));
		USBMS_DBG_INIT(iprintf(
			"[USBMS INIT] do_open() : bdev->bd_disk->fops=%08XH \n",
			bdev->bd_disk->fops
			));
		USBMS_DBG_INIT(iprintf(
			"[USBMS INIT] do_open() : bdev->bd_disk->fops->open=%08XH \n",
			bdev->bd_disk->fops->open
			));
		if (bdev->bd_disk->fops->open) {
			#ifndef NO_INODE
				USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(13) : fops->open() .. \n"));
				ret = bdev->bd_disk->fops->open(inode, file);
				if (ret)
					goto out;
			#endif // ndef NO_INODE
		   }
		USBMS_DBG_INIT(iprintf("[USBMS INIT] do_open() : bdev->bd_invalidated=%d \n",
			bdev->bd_invalidated
			));
		#ifndef USBMS_ONLY_ONE_DISK
			if (bdev->bd_invalidated) {
				USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(14) : rescan_partitions() .. \n"));
				/*
				#ifndef USB_NO_VMM // vitrtul memory management
					rescan_partitions(bdev->bd_disk, bdev);
				 #else
				*/
					rescan_partitions(bdev->bd_inode,bdev->bd_disk, bdev);
				/*
				#endif
				*/
			   }
		#endif // ndef USBMS_ONLY_ONE_DISK
			   }
			  else {
				/*
				#ifndef NO_SEMPHORE
					down(&bdev->bd_contains->bd_sem);
				#endif // ndef NO_SEMPHORE
				*/
				USBMS_DBG_INIT(iprintf("[USBMS INIT] do_open() : bdev->bd_contains->bd_part_count=%d \n",
					bdev->bd_contains->bd_part_count
					));
				bdev->bd_contains->bd_part_count++;
				/*
				#ifndef NO_SEMPHORE
					up(&bdev->bd_contains->bd_sem);
				#endif // ndef NO_SEMPHORE
				*/
			   }
   }

 USBMS_DBG_INIT(iprintf("[USBMS INIT] do_open() : bdev->bd_openers=%d \n",
	bdev->bd_openers
	));
 bdev->bd_openers++;
 /*
 #ifndef NO_SEMPHORE
	up(&bdev->bd_sem);
 #endif // ndef NO_SEMPHORE
 */
 /*
 #ifndef _ETCPIP
	unlock_kernel();
 #endif // ndef _ETCPIP
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open() End. \n"));
 return 0;

 // ----------------
 out_first:
 // ----------------

 #ifndef USBMS_ONLY_ONE_DISK
	bdev->bd_disk = NULL;
	USBMS_DBG_MEM(iprintf("[USBMS MEM] do_open(6) : Set Value : bdev->bd_disk=%08XH \n",bdev->bd_disk));
 #endif // ndef USBMS_ONLY_ONE_DISK
 /*
 #ifndef NO_BACKDEV 
	bdev->bd_inode->i_data.backing_dev_info = &default_backing_dev_info;
 #endif // ndef NO_BACKDEV
 */
 if(bdev != bdev->bd_contains)
   {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(15) : blkdev_put() .. \n"));
	blkdev_put(bdev->bd_contains, BDEV_RAW);
   }
 bdev->bd_contains = NULL;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(16) : put_disk() .. \n"));
 put_disk(disk);
 /*
 #ifndef _ETCPIP
	module_put(owner);
 #endif // ndef _ETCPIP
 */

 // ----------------
 out:
 // ----------------
 /*
 #ifndef NO_SEMPHORE
	up(&bdev->bd_sem);
 #endif // ndef NO_SEMPHORE
 */
 /*
 #ifndef _ETCPIP
	unlock_kernel();
 #endif // ndef _ETCPIP
 */
 if (ret) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(17) : bdput() .. \n"));
	bdput(bdev);
   }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] do_open(18) End with Fail !! \n"));
 return ret;
}
#endif // ndef NO_INODE

/*
#ifndef _ETCPIP
*/
int blkdev_get(struct block_device *bdev, mode_t mode, unsigned flags, int kind)
/*
 #else
	int blkdev_get(struct block_device *bdev, mode_t mode, unsigned flags, int kind ,
	struct gendisk *disk)
#endif
*/
{
 /*
  * This crockload is due to bad choice of ->open() type.
  * It will go away.
  * For now, block device ->open() routine must _not_
  * examine anything in 'inode' argument except ->i_rdev.
  */
 
 struct file fake_file;
 memset(&fake_file, 0, sizeof(struct file));


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blkdev_get() Begin .. \n"));

 fake_file.f_mode = mode;
 fake_file.f_flags = flags;
 /* for f_dentry
 fake_file.f_dentry = &fake_dentry;
 */
 /*
 #ifndef INODE_REDUCED  
	fake_dentry.d_inode = bdev->bd_inode;
 #endif // ndef INODE_REDUCED  
 */
 USBMS_DBG_MEM(iprintf("[USBMS MEM] blkdev_get() : bdev->bd_inode=%08XH \n",
	bdev->bd_inode));
 #ifndef NO_INODE
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blkdev_get() End by Call do_open() .. \n"));
	/*
	#ifndef _ETCPIP
	*/
		return do_open(bdev, bdev->bd_inode, &fake_file);
	/*
	 #else
		return do_open(bdev, bdev->bd_inode, &fake_file,disk);
	#endif
	*/
 #endif // ndef NO_INODE
 
}
/*
EXPORT_SYMBOL(blkdev_get);
*/
/* B. S. Liaw : None Call This ??
#ifndef NO_INODE
int blkdev_open(struct inode * inode, struct file * filp)
{
 struct block_device *bdev;
 int res;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blkdev_open() Begin .. \n"));
	
 // Preserve backwards compatibility and allow large file access
 // even if userspace doesn't ask for it explicitly. Some mkfs
 // binary needs it. We might want to drop this workaround
 // during an unstable branch.
	 
 filp->f_flags |= O_LARGEFILE;

 // #ifndef NO_INODE
	bd_acquire(inode);
	bdev = inode->i_bdev;

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blkdev_open() : do_open() .. \n"));
	res = do_open(bdev, inode, filp);
	if (res)
		return res;
 // #endif // ndef NO_INODE

	if (!(filp->f_flags & O_EXCL) )
		return 0;

	if (!(res = bd_claim(bdev, filp)))
		return 0;

 blkdev_put(bdev, BDEV_FILE);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] blkdev_open() End. \n"));
 return res;
}
#endif // ndef NO_INODE
*/

/*
EXPORT_SYMBOL(blkdev_open);
*/
int blkdev_put(struct block_device *bdev, int kind)
{
 int ret = 0;
 #ifndef NO_INODE
	struct inode *bd_inode = bdev->bd_inode;
 #endif // ndef NO_INODE
 struct gendisk *disk = bdev->bd_disk;

 /*
 #ifndef NO_SEMPHORE
	down(&bdev->bd_sem);
 #endif // ndef NO_SEMPHORE
 */
 /*
 lock_kernel();
 */
 USBMS_DBG_INIT(iprintf("[USBMS INIT] blkdev_put() : bdev->bd_openers=%d \n",
	bdev->bd_openers
	));

 if (!--bdev->bd_openers) {
	switch (kind) {
		case BDEV_FILE:
		case BDEV_FS:
			#ifndef NO_INODE
				sync_blockdev(bd_inode->i_bdev);
			#endif // ndef NO_INODE
			break;
		}
	/*
	#ifndef USB_NO_VMM // vitrtul memory management
		kill_bdev(bdev);
	#endif // ndef USB_NO_VMM
	*/
   }

 if (bdev->bd_contains == bdev) {
	#ifndef NO_INODE
		if (disk->fops->release)
			ret = disk->fops->release(bd_inode, NULL);
	#endif // ndef NO_INODE
   } else {
	#ifndef NO_SEMPHORE
		down(&bdev->bd_contains->bd_sem);
	#endif // ndef NO_SEMPHORE
		bdev->bd_contains->bd_part_count--;
	#ifndef NO_SEMPHORE
		up(&bdev->bd_contains->bd_sem);
	#endif // ndef NO_SEMPHORE
   }
 USBMS_DBG_INIT(iprintf("[USBMS INIT] blkdev_put() : bdev->bd_openers=%d \n",
	bdev->bd_openers
	));
 if (!bdev->bd_openers) {
	/*
	#ifndef _ETCPIP
		struct module *owner = disk->fops->owner;
	#endif // ndef _ETCPIP
	*/

	put_disk(disk);
	/*
	#ifndef _ETCPIP
		module_put(owner);
	#endif // ndef _ETCPIP
	*/

	if (bdev->bd_contains != bdev) {
		// #ifndef _ETCPIP
			kobject_put(&bdev->bd_part->kobj);
		// #endif // ndef _ETCPIP
		bdev->bd_part = NULL;
	   }
	#ifndef USBMS_ONLY_ONE_DISK
		bdev->bd_disk = NULL;
		USBMS_DBG_MEM(iprintf("[USBMS MEM] blkdev_put() : Set Value : bdev->bd_disk=%08XH \n",bdev->bd_disk));
	#endif // ndef USBMS_ONLY_ONE_DISK
	#ifndef NO_BACKDEV 
		bdev->bd_inode->i_data.backing_dev_info = &default_backing_dev_info;
	#endif // ndef NO_BACKDEV 
	if (bdev != bdev->bd_contains) {
		blkdev_put(bdev->bd_contains, BDEV_RAW);
	   }
	bdev->bd_contains = NULL;
   }
 /*
 #ifndef _ETCPIP
 unlock_kernel();
 #endif // ndef _ETCPIP
 */
 #ifndef NO_SEMPHORE
	up(&bdev->bd_sem);
 #endif // ndef NO_SEMPHORE
 bdput(bdev);
 return ret;
}

/*
EXPORT_SYMBOL(blkdev_put);
*/

#ifndef NO_INODE
   int blkdev_close(struct inode * inode, struct file * filp)
   {
	if (inode->i_bdev->bd_holder == filp)
		bd_release(inode->i_bdev);
	return blkdev_put(inode->i_bdev, BDEV_FILE);
   }
#endif // ndef NO_INODE

/*
static ssize_t blkdev_file_write(struct file *file, const char __user *buf,
				   size_t count, loff_t *ppos)
{
	struct iovec local_iov = { .iov_base = (void __user *)buf, .iov_len = count };

	return generic_file_write_nolock(file, &local_iov, 1, ppos);
}
*/
/*
static ssize_t blkdev_file_aio_write(struct kiocb *iocb, const char __user *buf,
				   size_t count, loff_t pos)
{
 #ifndef _ETCPIP
	struct iovec local_iov = { .iov_base = (void __user *)buf, .iov_len = count };

	return generic_file_aio_write_nolock(iocb, &local_iov, 1, &iocb->ki_pos);
 #endif // ndef _ETCPIP
}
*/

struct address_space_operations def_blk_aops = {
#if _USB2 > 0	// reduce function call layer
 .readpage	= NULL,
 .writepage	= NULL,
#else	// _USB2
 .readpage	= blkdev_readpage,
 .writepage	= blkdev_writepage,
#endif	// _USB2
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	.sync_page	= block_sync_page,
 #endif // ndef USB_NO_VMM
 */
 /*
 .prepare_write	= blkdev_prepare_write,
 */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	.commit_write	= blkdev_commit_write,
	.writepages	= generic_writepages,
	.direct_IO	= blkdev_direct_IO,
 #endif // ndef USB_NO_VMM
 */
};

/*
struct file_operations def_blk_fops = {
	.open		= blkdev_open,
	.release	= blkdev_close,
	.llseek		= block_llseek,
	.read		= generic_file_read,
	.write		= blkdev_file_write,
  	.aio_read	= generic_file_aio_read,
  	.aio_write	= blkdev_file_aio_write, 
 #ifndef USB_NO_VMM // vitrtul memory management
	.mmap		= generic_file_mmap,
 #endif // ndef USB_NO_VMM
	.fsync		= block_fsync,
	.ioctl		= blkdev_ioctl,
	.readv		= generic_file_readv,
	.writev		= generic_file_writev,
	.sendfile	= generic_file_sendfile,
};
*/
/*
EXPORT_SYMBOL(def_blk_fops);
*/
/*
#ifndef USBMS_NO_IOCTL
int ioctl_by_bdev(struct block_device *bdev, unsigned cmd, unsigned long arg)
{
 int res;
 
 #ifndef _ETCPIP
	// mm_segment_t is on include/asm-mips/processor.h
	mm_segment_t old_fs = get_fs();

	set_fs(KERNEL_DS);
 #endif // ndef _ETCPIP
 
 #ifndef NO_INODE
		res = blkdev_ioctl(bdev->bd_inode, NULL, cmd, arg);
 #endif // ndef NO_INODE
 
 #ifndef _ETCPIP
	set_fs(old_fs);
 #endif // ndef _ETCPIP
 
 return res;
}
#endif // ndef USBMS_NO_IOCTL
*/
/*
EXPORT_SYMBOL(ioctl_by_bdev);
*/
/**
 * lookup_bdev  - lookup a struct block_device by name
 *
 * @path:	special file representing the block device
 *
 * Get a reference to the blockdevice at @path in the current
 * namespace if possible and return it.  Return ERR_PTR(error)
 * otherwise.
 */
/*
struct block_device *lookup_bdev(const char *path)
{
	struct block_device *bdev;
 #ifndef NO_INODE  
	struct inode *inode;
 #endif // ndef NO_INODE  
	struct nameidata nd;
	int error;

	if (!path || !*path)
		return ERR_PTR(-EINVAL);

	error = path_lookup(path, LOOKUP_FOLLOW, &nd);
	if (error)
		return ERR_PTR(error);

 #ifndef NO_INODE  
	inode = nd.dentry->d_inode;
	error = -ENOTBLK;
	if (!S_ISBLK(inode->i_mode))
		goto fail;
 #endif // ndef NO_INODE  
	error = -EACCES;
	if (nd.mnt->mnt_flags & MNT_NODEV)
		goto fail;
 #ifndef NO_INODE  
	error = bd_acquire(inode);
	if (error)
		goto fail;
	bdev = inode->i_bdev;
 #endif // ndef NO_INODE  

out:
	path_release(&nd);
	return bdev;
fail:
	bdev = ERR_PTR(error);
	goto out;
}
*/
/**
 * open_bdev_excl  -  open a block device by name and set it up for use
 *
 * @path:	special file representing the block device
 * @flags:	%MS_RDONLY for opening read-only
 * @kind:	usage (same as the 4th paramter to blkdev_get)
 * @holder:	owner for exclusion
 *
 * Open the blockdevice described by the special file at @path, claim it
 * for the @holder and properly set it up for @kind usage.
 */
/*
struct block_device *open_bdev_excl(const char *path, int flags,
				    int kind, void *holder)
{
	struct block_device *bdev;
	mode_t mode = FMODE_READ;
	int error = 0;

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] open_bdev_excl() Begin .. \n"));
	bdev = lookup_bdev(path);
	if (IS_ERR(bdev))
		return bdev;

	if (!(flags & MS_RDONLY))
		mode |= FMODE_WRITE;
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] open_bdev_excl() : blkdev_get() .. \n"));
	error = blkdev_get(bdev, mode, 0, kind);
	if (error)
		return ERR_PTR(error);
	error = -EACCES;
	if (!(flags & MS_RDONLY) && bdev_read_only(bdev))
		goto blkdev_put;
	error = bd_claim(bdev, holder);
	if (error)
		goto blkdev_put;

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] open_bdev_excl() End. \n"));
	return bdev;
	
blkdev_put:
	blkdev_put(bdev, BDEV_FS);
	return ERR_PTR(error);
}
*/
/*
EXPORT_SYMBOL(open_bdev_excl);
*/
/**
 * close_bdev_excl  -  release a blockdevice openen by open_bdev_excl()
 *
 * @bdev:	blockdevice to close
 * @kind:	usage (same as the 4th paramter to blkdev_get)
 *
 * This is the counterpart to open_bdev_excl().
 */
/*
void close_bdev_excl(struct block_device *bdev, int kind)
{
 bd_release(bdev);
 blkdev_put(bdev, kind);
}
*/
/*
EXPORT_SYMBOL(close_bdev_excl);
*/

// B. S. Liaw TEST for compiler warning !!
