
/*
 *  linux/fs/buffer.c
 *
 *  Copyright (C) 1991, 1992, 2002  Linus Torvalds
 */

/*
 * Start bdflush() with kernel_thread not syscall - Paul Gortmaker, 12/95
 *
 * Removed a lot of unnecessary code and simplified things now that
 * the buffer cache isn't our primary cache - Andrew Tridgell 12/96
 *
 * Speed up hash, lru, and free list operations.  Use gfp() for allocating
 * hash table, use SLAB cache for buffer heads. SMP threading.  -DaveM
 *
 * Added 32k buffer block sizes - these are required older ARM systems. - RMK
 *
 * async buffer flushing, 1999 Andrea Arcangeli <andrea@suse.de>
 */

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
        #include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsbMsInit.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
        #include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP

/*
#include <linux/config.h>
#include <linux/kernel.h>
*/
#include "PoKernel.h"
/*
#include <linux/fs.h>
*/
#include "PoFile.h" // include/linux/file.h
#include "radix-tree.h" // include/linux/radix-tree.h
#include "PoFs.h" // <linux/fs.h>
/*
#include <linux/mm.h>
#include <linux/percpu.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/blkdev.h>
*/
#include "bitops.h" // B. S. Liaw add for test_bit()
#include "blkdev.h"
/*
#include <linux/file.h>
*/
#include "PoFile.h"
/*
#include <linux/quotaops.h>
#include <linux/highmem.h>
#include <linux/module.h>
#include <linux/writeback.h>
*/
#include "writeback.h"
/*
#include <linux/hash.h>
#include <linux/suspend.h>
#include <linux/buffer_head.h>
*/
#include "buffer_head.h"
/*
#include <linux/bio.h>
*/
#include "bio.h"
/*
#include <linux/notifier.h>
#include <linux/cpu.h>
#include <asm/bitops.h>
*/
#include "bitops.h"

// #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	struct buffer_head *global_bh = NULL ;
	struct bio *global_bio = NULL ;
// #endif // NO_ELEVATOR

/*
#ifndef USB_NO_VMM // vitrtul memory management
        static void invalidate_bh_lrus(void);

        #define BH_ENTRY(list) list_entry((list), struct buffer_head, b_assoc_buffers)
#endif // ndef USB_NO_VMM
*/
/*
 * Hashed waitqueue_head's for wait_on_buffer()
 */
/*
#ifndef NO_WAITQUEUE
	#define BH_WAIT_TABLE_ORDER     7
        static struct bh_wait_queue_head {
                wait_queue_head_t wqh;
        } ____cacheline_aligned_in_smp bh_wait_queue_heads[1<<BH_WAIT_TABLE_ORDER];
#endif // ndef NO_WAITQUEUE
*/

/*
 * Debug/devel support stuff
 */

void __buffer_error(char *file, int line)
{
 static int enough;

 if (enough > 10)
                return;
 enough++;
 /*
 printk("buffer layer error at %s:%d\n", file, line);
 */
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] __buffer_error() : Error : buffer layer error !! \n"));
 /*
 #ifndef CONFIG_KALLSYMS
	printk("Pass this trace through ksymoops for reporting\n");
 
	dump_stack();
 
 #endif
 */
}
/*
EXPORT_SYMBOL(__buffer_error);
*/
inline void
init_buffer(struct buffer_head *bh, bh_end_io_t *handler, void *private)
{
 bh->b_end_io = handler;
 bh->b_private = private;
}

/*
 * Return the address of the waitqueue_head to be used for this
 * buffer_head
 */
/*
#ifndef NO_WAITQUEUE
wait_queue_head_t *bh_waitq_head(struct buffer_head *bh)
{
        return &bh_wait_queue_heads[hash_ptr(bh, BH_WAIT_TABLE_ORDER)].wqh;
}
#endif // ndef NO_WAITQUEUE
*/
/*
EXPORT_SYMBOL(bh_waitq_head);
*/
/*
void wake_up_buffer(struct buffer_head *bh)
{
 #ifndef NO_WAITQUEUE
        wait_queue_head_t *wq = bh_waitq_head(bh);
 #endif // ndef NO_WAITQUEUE

        smp_mb();
 #ifndef NO_WAITQUEUE
        if (waitqueue_active(wq))
                wake_up_all(wq);
 #endif // ndef NO_WAITQUEUE
}
*/
/*
EXPORT_SYMBOL(wake_up_buffer);
*/
void unlock_buffer(struct buffer_head *bh)
{
 /*
  * unlock_buffer against a zero-count bh is a bug, if the page
  * is not locked.  Because then nothing protects the buffer's
  * waitqueue, which is used here. (Well.  Other locked buffers
  * against the page will pin it.  But complain anyway).
  */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if(atomic_read(&bh->b_count) == 0 
		//#ifndef USB_NO_VMM // vitrtul memory management
    	//    // struct page is on include/linux/mm.h
	    //    &&
    	//    !PageLocked(bh->b_page) &&
		//    !PageWriteback(bh->b_page)
		//#endif 
	   )
	   {
		USBMS_DBG_INIT(iprintf(
			"[USBMS INIT] unlock_buffer() : Error : atomic_read(&bh->b_count) == 0 !! \n"
			));
		buffer_error();
	   }
 #endif // ndef NO_ELEVATOR
 */
 /*
 clear_buffer_locked(bh);
 smp_mb__after_clear_bit();
 
 wake_up_buffer(bh);
 */
}

/*
 * Block until a buffer comes unlocked.  This doesn't stop it
 * from becoming locked again - you have to lock it yourself
 * if you want to preserve its state.
 */
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	void __wait_on_buffer(struct buffer_head * bh)
 //#else
 //	void __wait_on_buffer(request_queue_t *q)
 //#endif
{
 #ifndef NO_WAITQUEUE
	wait_queue_head_t *wqh = bh_waitq_head(bh);
	DEFINE_WAIT(wait);
 #endif // ndef NO_WAITQUEUE

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __wait_on_buffer() Begin .. \n"));
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
 if (atomic_read(&bh->b_count) == 0 
        #ifndef USB_NO_VMM // vitrtul memory management
                // struct page is on include/linux/mm.h
                &&
                        (!bh->b_page || !PageLocked(bh->b_page))
         #else
                &&
                        (!bh->b_buf)
        #endif 
           )
           {
                USBMS_DBG_INFO(iprintf("
                        [USBMS INFO] __wait_on_buffer() : Error : atomic_read(&bh->b_count) == 0 !! \n"
                        ));
                buffer_error();
           }
 #endif // ndef NO_ELEVATOR
 #ifndef USB_NO_VMM // vitrtul memory management
	do 
 #endif // ndef USB_NO_VMM
   {
	#ifndef NO_WAITQUEUE
                prepare_to_wait(wqh, &wait, TASK_UNINTERRUPTIBLE);
	#endif // ndef NO_WAITQUEUE
	#ifndef USB_NO_VMM // vitrtul memory management
		if (buffer_locked(bh))
	#endif // ndef USB_NO_VMM
		   {
	        USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __wait_on_buffer() : blk_run_queues() .. \n"));
			#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
				blk_run_queues();
			 #else
				blk_run_queues(q);
			#endif
			#ifndef _ETCPIP
                        io_schedule();
			#endif // ndef _ETCPIP
		   }
   }
 #ifndef USB_NO_VMM // vitrtul memory management
	while (buffer_locked(bh));
 #endif // ndef USB_NO_VMM
 #ifndef NO_WAITQUEUE
	finish_wait(wqh, &wait);
 #endif // ndef NO_WAITQUEUE
}
#endif // ndef NO_WAITQUEUE
*/
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
static void
__set_page_buffers(struct page *page, struct buffer_head *head)
{
        if (page_has_buffers(page))
                buffer_error();
        page_cache_get(page);
        SetPagePrivate(page);
        page->private = (unsigned long)head;
}
#endif // ndef USB_NO_VMM
*/
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
static void
__clear_page_buffers(struct page *page)
{
        ClearPagePrivate(page);
        page->private = 0;
        page_cache_release(page);
}
#endif // ndef USB_NO_VMM
*/
static void buffer_io_error(struct buffer_head *bh)
{
 char b[BDEVNAME_SIZE];
 /*
 printk(KERN_ERR "Buffer I/O error on device %s, logical block %Lu\n",
 */
 dprintf("[USBMS] Error : Buffer I/O error on device %s, logical block %Lu\n",
                        bdevname(bh->b_bdev, b),
                        (unsigned long long)bh->b_blocknr);
}

/*
 * Default synchronous end-of-IO handler..  Just mark it up-to-date and
 * unlock the buffer. This is what ll_rw_block uses too.
 */
/*
void end_buffer_read_sync(struct buffer_head *bh, int uptodate)
{
 if (uptodate) {
        
        //set_buffer_uptodate(bh);
        
   } else {
        // This happens, due to failed READA attempts. 
        
        #ifndef USB_NO_VMM // vitrtul memory management
                clear_buffer_uptodate(bh);
        #endif // ndef USB_NO_VMM
        
   }
 unlock_buffer(bh);
 put_bh(bh);
}

void end_buffer_write_sync(struct buffer_head *bh, int uptodate)
{
        char b[BDEVNAME_SIZE];

        if (uptodate) {
 
                // set_buffer_uptodate(bh);
 
        } else {
                buffer_io_error(bh);
                printk(KERN_WARNING "lost page write due to I/O error on %s\n",
                       bdevname(bh->b_bdev, b));
        
        #ifndef USB_NO_VMM // vitrtul memory management
                set_buffer_write_io_error(bh);
 
                clear_buffer_uptodate(bh);
        #endif // ndef USB_NO_VMM
        
   }
 unlock_buffer(bh);
 put_bh(bh);
}
*/
/*
 * Write out and wait upon all the dirty data associated with a block
 * device via its mapping.  Does not take the superblock lock.
 */
int sync_blockdev(struct block_device *bdev)
{
 int ret = 0;

 if(bdev) {
        int err;

        #ifndef NO_INODE
        #ifndef USB_NO_VMM // vitrtul memory management
                ret = filemap_fdatawrite(bdev->bd_inode->i_mapping);
                err = filemap_fdatawait(bdev->bd_inode->i_mapping);
        #endif // ndef USB_NO_VMM
        #endif // ndef NO_INODE
        if (!ret)
                        ret = err;
   }
 return ret;
}
/*
EXPORT_SYMBOL(sync_blockdev);
*/
/*
 * Write out and wait upon all dirty data associated with this
 * superblock.  Filesystem data as well as the underlying block
 * device.  Takes the superblock lock.
 */
/*
int fsync_super(struct super_block *sb)
{
 
        sync_inodes_sb(sb, 0);
 
        DQUOT_SYNC(sb);
        lock_super(sb);
        if (sb->s_dirt && sb->s_op->write_super)
                sb->s_op->write_super(sb);
        unlock_super(sb);
        if (sb->s_op->sync_fs)
                sb->s_op->sync_fs(sb, 1);
        sync_blockdev(sb->s_bdev);
 
        sync_inodes_sb(sb, 1);
 
        return sync_blockdev(sb->s_bdev);
}
*/

/*
 * Write out and wait upon all dirty data associated with this
 * device.   Filesystem data as well as the underlying block
 * device.  Takes the superblock lock.
 */

int fsync_bdev(struct block_device *bdev)
{
 
 #ifndef INODE_REDUCED  
        struct super_block *sb = get_super(bdev);
        if (sb) {
                int res = fsync_super(sb);
                drop_super(sb);
                return res;
        }
 #endif // ndef INODE_REDUCED  

 return sync_blockdev(bdev);
}

/*
 * sync everything.  Start out by waking pdflush, because that writes back
 * all queues in parallel.
 */
#ifndef _ETCPIP
static void do_sync(unsigned long wait)
{
        wakeup_bdflush(0);
        sync_inodes(0);         /* All mappings, inodes and their blockdevs */
        DQUOT_SYNC(NULL);
        sync_supers();          /* Write the superblocks */
        sync_filesystems(0);    /* Start syncing the filesystems */
        sync_filesystems(wait); /* Waitingly sync the filesystems */
        sync_inodes(wait);      /* Mappings, inodes and blockdevs, again. */
        if (!wait)
                printk("Emergency Sync complete\n");
}
#endif // ndef _ETCPIP

/*
#ifndef _ETCPIP
asmlinkage long sys_sync(void)
{
        do_sync(1);
        return 0;
}

void emergency_sync(void)
{
        pdflush_operation(do_sync, 0);
}
#endif // ndef _ETCPIP
*/

/*
 * Generic function to fsync a file.
 *
 * filp may be NULL if called via the msync of a vma.
 */
/*
int file_fsync(struct file *filp, struct dentry *dentry, int datasync)
{
        struct inode * inode = dentry->d_inode;
        struct super_block * sb;
        int ret;

        // sync the inode to buffers 
        write_inode_now(inode, 0);

        // sync the superblock to buffers 
 // #ifndef INODE_REDUCED
        sb = inode->i_sb;
 // #endif // ndef INODE_REDUCED
        lock_super(sb);
        if (sb->s_op->write_super)
                sb->s_op->write_super(sb);
        unlock_super(sb);

        // .. finally sync the buffers to disk
        ret = sync_blockdev(sb->s_bdev);
        return ret;
}
*/
/*
// asmlinkage 
long sys_fsync(unsigned int fd)
{
        struct file * file;
        struct dentry * dentry;
        struct inode * inode;
        int ret, err;

        ret = -EBADF;
        file = fget(fd);
        if (!file)
                goto out;

        dentry = file->f_dentry;
        inode = dentry->d_inode;

        ret = -EINVAL;
        if (!file->f_op || !file->f_op->fsync) {
                // Why?  We can still call filemap_fdatawrite 
                goto out_putf;
        }

        // We need to protect against concurrent writers.. 
 #ifndef INODE_REDUCED  
        down(&inode->i_sem);
        #endif // ndef NO_SEMPHORE
 #ifndef _ETCPIP
        current->flags |= PF_SYNCWRITE;
 #endif // ndef _ETCPIP
 #ifndef NO_INODE
        ret = filemap_fdatawrite(inode->i_mapping);
 #endif // ndef NO_INODE
        err = file->f_op->fsync(file, dentry, 0);
        if (!ret)
                ret = err;
 #ifndef NO_INODE
        #ifndef USB_NO_VMM // vitrtul memory management
        err = filemap_fdatawait(inode->i_mapping);
        #endif // ndef USB_NO_VMM
 #endif // ndef NO_INODE
        if (!ret)
                ret = err;
 #ifndef _ETCPIP
        current->flags &= ~PF_SYNCWRITE;
 #endif // ndef _ETCPIP
 #ifndef INODE_REDUCED  
        up(&inode->i_sem);
        #endif // ndef NO_SEMPHORE

out_putf:
        fput(file);
out:
        return ret;
}
*/
/*
// asmlinkage
long sys_fdatasync(unsigned int fd)
{
        struct file * file;
        struct dentry * dentry;
        struct inode * inode;
        int ret, err;

        ret = -EBADF;
        file = fget(fd);
        if (!file)
                goto out;

        dentry = file->f_dentry;
        inode = dentry->d_inode;

        ret = -EINVAL;
        if (!file->f_op || !file->f_op->fsync)
                goto out_putf;

 #ifndef INODE_REDUCED  
        down(&inode->i_sem);
        #endif // ndef NO_SEMPHORE
 #ifndef _ETCPIP
        current->flags |= PF_SYNCWRITE;
 #endif // ndef _ETCPIP
 #ifndef NO_INODE
        ret = filemap_fdatawrite(inode->i_mapping);
 #endif // ndef NO_INODE
        err = file->f_op->fsync(file, dentry, 1);
        if (!ret)
                ret = err;
 #ifndef NO_INODE
        #ifndef USB_NO_VMM // vitrtul memory management
        err = filemap_fdatawait(inode->i_mapping);
        #endif // ndef USB_NO_VMM
 #endif // ndef NO_INODE
        if (!ret)
                ret = err;
 #ifndef _ETCPIP
        current->flags &= ~PF_SYNCWRITE;
 #endif // ndef _ETCPIP
 #ifndef INODE_REDUCED  
        up(&inode->i_sem);
        #endif // ndef NO_SEMPHORE

out_putf:
        fput(file);
out:
        return ret;
}
*/
/*
 * Various filesystems appear to want __find_get_block to be non-blocking.
 * But it's the page lock which protects the buffers.  To get around this,
 * we get exclusion from try_to_free_buffers with the blockdev mapping's
 * private_lock.
 *
 * Hack idea: for the blockdev mapping, i_bufferlist_lock contention
 * may be quite high.  This code could TryLock the page, and if that
 * succeeds, there is no need to take private_lock. (But if
 * private_lock is contended then so is mapping->page_lock).
 */
/*
static struct buffer_head *
__find_get_block_slow(struct block_device *bdev, sector_t block, int unused)
{
        struct inode *bd_inode = bdev->bd_inode;
 #ifndef NO_INODE
        struct address_space *bd_mapping = bd_inode->i_mapping;
 #endif // ndef NO_INODE
        struct buffer_head *ret = NULL;
        unsigned long index;
        struct buffer_head *bh;
        struct buffer_head *head;
 #ifndef USB_NO_VMM // vitrtul memory management
        struct page *page;
 #endif // ndef USB_NO_VMM

 #ifndef USB_NO_VMM // vitrtul memory management
        index = block >> (PAGE_CACHE_SHIFT - bd_inode->i_blkbits);
        page = find_get_page(bd_mapping, index);
        if (!page)
                goto out;
 #endif // ndef USB_NO_VMM
 
 #ifndef NO_SPINLOCK
        spin_lock(&bd_mapping->private_lock);
 #endif // ndef NO_SPINLOCK
 
 #ifndef USB_NO_VMM // vitrtul memory management
        if (!page_has_buffers(page))
                goto out_unlock;
        head = page_buffers(page);
 #endif // ndef USB_NO_VMM
        bh = head;
        do {
                if (bh->b_blocknr == block) {
                        ret = bh;
                        get_bh(bh);
                        goto out_unlock;
                }
 #ifndef USB_NO_VMM // vitrtul memory management
                bh = bh->b_this_page;
 #endif // ndef USB_NO_VMM
        } while (bh != head);
        buffer_error();
        printk("block=%llu, b_blocknr=%llu\n",
                (unsigned long long)block, (unsigned long long)bh->b_blocknr);
        printk("b_state=0x%08lx, b_size=%u\n", bh->b_state, bh->b_size);
out_unlock:
 
 #ifndef NO_SPINLOCK
        spin_unlock(&bd_mapping->private_lock);
 #endif // ndef NO_SPINLOCK
 
        page_cache_release(page);
out:
        return ret;
}
*/
/* If invalidate_buffers() will trash dirty buffers, it means some kind
   of fs corruption is going on. Trashing dirty data always imply losing
   information that was supposed to be just stored on the physical layer
   by the user.

   Thus invalidate_buffers in general usage is not allwowed to trash
   dirty buffers. For example ioctl(FLSBLKBUF) expects dirty data to
   be preserved.  These buffers are simply skipped.
  
   We also skip buffers which are still in use.  For example this can
   happen if a userspace program is reading the block device.

   NOTE: In the case where the user removed a removable-media-disk even if
   there's still dirty data not synced on disk (due a bug in the device driver
   or due an error of the user), by not destroying the dirty buffers we could
   generate corruption also on the next media inserted, thus a parameter is
   necessary to handle this case in the most safe way possible (trying
   to not corrupt also the new disk inserted with the data belonging to
   the old now corrupted disk). Also for the ramdisk the natural thing
   to do in order to release the ramdisk memory is to destroy dirty buffers.

   These are two special cases. Normal usage imply the device driver
   to issue a sync on the device (without waiting I/O completion) and
   then an invalidate_buffers call that doesn't trash dirty buffers.

   For handling cache coherency with the blkdev pagecache the 'update' case
   is been introduced. It is needed to re-read from disk any pinned
   buffer. NOTE: re-reading from disk is destructive so we can do it only
   when we assume nobody is changing the buffercache under our I/O and when
   we think the disk contains more recent information than the buffercache.
   The update == 1 pass marks the buffers we need to update, the update == 2
   pass does the actual I/O. */
void invalidate_bdev(struct block_device *bdev, int destroy_dirty_buffers)
{

 #ifndef USB_NO_VMM // vitrtul memory management
        invalidate_bh_lrus();
 #endif // ndef USB_NO_VMM

 /*
  * FIXME: what about destroy_dirty_buffers?
  * We really want to use invalidate_inode_pages2() for
  * that, but not until that's cleaned up.
  */
 #ifndef NO_INODE
 #ifndef USB_NO_VMM // vitrtul memory management
        invalidate_inode_pages(bdev->bd_inode->i_mapping);
 #endif // ndef USB_NO_VMM
 #endif // ndef NO_INODE
}

/*
 * Kick pdflush then try to free up some ZONE_NORMAL memory.
 */
/*
static void free_more_memory(void)
{
        struct zone *zone;
        pg_data_t *pgdat;

        wakeup_bdflush(1024);
        USBMS_DBG_FLOW(iprintf("[USBMS FLOW] free_more_memory() : blk_run_queues() .. \n"));
        blk_run_queues();
        yield();

        for_each_pgdat(pgdat) {
                zone = pgdat->node_zonelists[GFP_NOFS&GFP_ZONEMASK].zones[0];
                if (zone)
                        try_to_free_pages(zone, GFP_NOFS, 0);
        }
}
*/
/*
 * I/O completion handler for block_read_full_page() - pages
 * which come unlocked at the end of I/O.
 */

static void end_buffer_async_read(struct buffer_head *bh, int uptodate)
{
 /*
 #ifndef NO_SPINLOCK
        static spinlock_t page_uptodate_lock = SPIN_LOCK_UNLOCKED;
 #endif // ndef NO_SPINLOCK
 */
 unsigned long flags;
 struct buffer_head *tmp;
 #ifndef USB_NO_VMM // vitrtul memory management
        struct page *page;
  #else
        unsigned char *buf ;
 #endif
 int page_uptodate = 1;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] end_buffer_async_read() Begin .. \n"));
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        BUG_ON(!buffer_async_read(bh));
 #endif // ndef USB_NO_VMM
 */
 #ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        page = bh->b_page;
  #else
        buf = bh->b_buf ;
 #endif 
 if (uptodate) {
                /*
                #ifndef USB_NO_VMM // vitrtul memory management
                        set_buffer_uptodate(bh);
                #endif // ndef USB_NO_VMM
                */
   } else {
        /*
        #ifndef USB_NO_VMM // vitrtul memory management
                clear_buffer_uptodate(bh);
        #endif // ndef USB_NO_VMM
        */
        buffer_io_error(bh);
        #ifndef USB_NO_VMM // vitrtul memory management
                SetPageError(page);
        #endif // ndef USB_NO_VMM
   }

        
 // Be _very_ careful from here on. Bad things can happen if
 // two buffer heads end IO at almost the same time and both
 // decide that the page is now completely done.

 /*
 #ifndef NO_SPINLOCK
        spin_lock_irqsave(&page_uptodate_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        clear_buffer_async_read(bh);
 #endif // ndef USB_NO_VMM
 */
 unlock_buffer(bh);
 tmp = bh;
 #ifndef USB_NO_VMM // vitrtul memory management
        do {
                /*
                #ifndef USB_NO_VMM // vitrtul memory management
                        
                        if (!buffer_uptodate(tmp))
                                page_uptodate = 0;
                        
                        if (buffer_async_read(tmp)) {
                                        BUG_ON(!buffer_locked(tmp));
                                goto still_busy;
                        }
                #endif // ndef USB_NO_VMM
                */
                tmp = tmp->b_this_page;
           } while (tmp != bh);
 #endif // ndef USB_NO_VMM
 /*
 #ifndef NO_SPINLOCK
        spin_unlock_irqrestore(&page_uptodate_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
        
 // If none of the buffers had errors and they are all
 // uptodate then we can set the page uptodate.
         
 #ifndef USB_NO_VMM // vitrtul memory management
        if (page_uptodate && !PageError(page))
                SetPageUptodate(page);
        unlock_page(page);
 #endif // ndef USB_NO_VMM
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] end_buffer_async_read() End. \n"));
 return;

 still_busy:
 /*
 #ifndef NO_SPINLOCK
        spin_unlock_irqrestore(&page_uptodate_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] end_buffer_async_read() End with Failed !! \n"));
 return;
}

/*
 * Completion handler for block_write_full_page() - pages which are unlocked
 * during I/O, and which have PageWriteback cleared upon I/O completion.
 */

void end_buffer_async_write(struct buffer_head *bh, int uptodate)
{
 char b[BDEVNAME_SIZE];
 /*
 #ifndef NO_SPINLOCK
        static spinlock_t page_uptodate_lock = SPIN_LOCK_UNLOCKED;
 #endif // ndef NO_SPINLOCK
 */
 unsigned long flags;
 struct buffer_head *tmp;
 #ifndef USB_NO_VMM // vitrtul memory management
        struct page *page;
  #else
        unsigned char *buf ;
 #endif


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] end_buffer_async_write() Begin .. \n"));
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
 BUG_ON(!buffer_async_write(bh));
 if(!buffer_async_write(bh)) {
        dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
   }
 #endif // ndef USB_NO_VMM
 */

 #ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        page = bh->b_page;
  #else
        buf = bh->b_buf ;
 #endif 
 if(uptodate) {
        /*
        #ifndef USB_NO_VMM // vitrtul memory management
        set_buffer_uptodate(bh);
        #endif // ndef USB_NO_VMM
        */
   } else {
        buffer_io_error(bh);
        /*
        printk(KERN_WARNING "lost page write due to I/O error on %s\n",
        */
        dprintf("[USBMS] Error : lost page write due to I/O error on %s\n",
                       bdevname(bh->b_bdev, b));
        /*
        #ifndef USB_NO_VMM // vitrtul memory management
                set_bit(AS_EIO, &page->mapping->flags);
                clear_buffer_uptodate(bh);
                SetPageError(page);
        #endif // ndef USB_NO_VMM
        */
        }
 /*
 #ifndef NO_SPINLOCK
        spin_lock_irqsave(&page_uptodate_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
 clear_buffer_async_write(bh);
 #endif // ndef USB_NO_VMM
 */
 unlock_buffer(bh);
 #ifndef USB_NO_VMM // vitrtul memory management
        tmp = bh->b_this_page;
        while (tmp != bh) {
                /*
                #ifndef USB_NO_VMM // vitrtul memory management
                if (buffer_async_write(tmp)) {
                
                        BUG_ON(!buffer_locked(tmp));
                
                        goto still_busy;
                   }
                #endif // ndef USB_NO_VMM
                */
                tmp = tmp->b_this_page;
           }
 #endif // ndef USB_NO_VMM
 /*
 #ifndef NO_SPINLOCK
        spin_unlock_irqrestore(&page_uptodate_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        end_page_writeback(page);
 #endif // ndef USB_NO_VMM
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] end_buffer_async_write() End. \n"));
 return;

 // -------------
 still_busy:
 // -------------
 /*
 #ifndef NO_SPINLOCK
        spin_unlock_irqrestore(&page_uptodate_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] end_buffer_async_write() End with Failed. \n"));
 return;
}

/*
 * If a page's buffers are under async readin (end_buffer_async_read
 * completion) then there is a possibility that another thread of
 * control could lock one of the buffers after it has completed
 * but while some of the other buffers have not completed.  This
 * locked buffer would confuse end_buffer_async_read() into not unlocking
 * the page.  So the absence of BH_Async_Read tells end_buffer_async_read()
 * that this buffer is not under async I/O.
 *
 * The page comes unlocked when it has no locked buffer_async buffers
 * left.
 *
 * PageLocked prevents anyone starting new async I/O reads any of
 * the buffers.
 *
 * PageWriteback is used to prevent simultaneous writeout of the same
 * page.
 *
 * PageLocked prevents anyone from starting writeback of a page which is
 * under read I/O (PageWriteback is only ever set against a locked page).
 */

void mark_buffer_async_read(struct buffer_head *bh)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] mark_buffer_async_read() Begin .. \n"));
 bh->b_end_io = end_buffer_async_read;
 #ifndef USB_NO_VMM // vitrtul memory management
        set_buffer_async_read(bh);
 #endif // ndef USB_NO_VMM
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] mark_buffer_async_read() End. \n"));
}
/*
EXPORT_SYMBOL(mark_buffer_async_read);
*/

#ifdef _ETCPIP
static inline
#endif
void mark_buffer_async_write(struct buffer_head *bh)
{
        bh->b_end_io = end_buffer_async_write;
 #ifndef USB_NO_VMM // vitrtul memory management
        set_buffer_async_write(bh);
 #endif // ndef USB_NO_VMM
}

/*
EXPORT_SYMBOL(mark_buffer_async_write);
*/

/*
 * fs/buffer.c contains helper functions for buffer-backed address space's
 * fsync functions.  A common requirement for buffer-based filesystems is
 * that certain data from the backing blockdev needs to be written out for
 * a successful fsync().  For example, ext2 indirect blocks need to be
 * written back and waited upon before fsync() returns.
 *
 * The functions mark_buffer_inode_dirty(), fsync_inode_buffers(),
 * inode_has_buffers() and invalidate_inode_buffers() are provided for the
 * management of a list of dependent buffers at ->i_mapping->private_list.
 *
 * Locking is a little subtle: try_to_free_buffers() will remove buffers
 * from their controlling inode's queue when they are being freed.  But
 * try_to_free_buffers() will be operating against the *blockdev* mapping
 * at the time, not against the S_ISREG file which depends on those buffers.
 * So the locking for private_list is via the private_lock in the address_space
 * which backs the buffers.  Which is different from the address_space 
 * against which the buffers are listed.  So for a particular address_space,
 * mapping->private_lock does *not* protect mapping->private_list!  In fact,
 * mapping->private_list will always be protected by the backing blockdev's
 * ->private_lock.
 *
 * Which introduces a requirement: all buffers on an address_space's
 * ->private_list must be from the same address_space: the blockdev's.
 *
 * address_spaces which do not place buffers at ->private_list via these
 * utility functions are free to use private_lock and private_list for
 * whatever they want.  The only requirement is that list_empty(private_list)
 * be true at clear_inode() time.
 *
 * FIXME: clear_inode should not call invalidate_inode_buffers().  The
 * filesystems should do that.  invalidate_inode_buffers() should just go
 * BUG_ON(!list_empty).
 *
 * FIXME: mark_buffer_dirty_inode() is a data-plane operation.  It should
 * take an address_space, not an inode.  And it should be called
 * mark_buffer_dirty_fsync() to clearly define why those buffers are being
 * queued up.
 *
 * FIXME: mark_buffer_dirty_inode() doesn't need to add the buffer to the
 * list if it is already on a list.  Because if the buffer is on a list,
 * it *must* already be on the right one.  If not, the filesystem is being
 * silly.  This will save a ton of locking.  But first we have to ensure
 * that buffers are taken *off* the old inode's list when they are freed
 * (presumably in truncate).  That requires careful auditing of all
 * filesystems (do it inside bforget()).  It could also be done by bringing
 * b_inode back.
 */
/*
void buffer_insert_list(spinlock_t *lock,
                struct buffer_head *bh, struct list_head *list)
{
        spin_lock(lock);
        list_move_tail(&bh->b_assoc_buffers, list);
        spin_unlock(lock);
}
*/
/*
 * The buffer's backing address_space's private_lock must be held
 */
/*
static inline void __remove_assoc_queue(struct buffer_head *bh)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __remove_assoc_queue() Begn .. \n"));
 #ifndef USB_NO_VMM // vitrtul memory management
        USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __remove_assoc_queue() : list_del_init() .. \n"));
        list_del_init(&bh->b_assoc_buffers);
 #endif // ndef USB_NO_VMM
}
*/
int inode_has_buffers(struct inode *inode)
{
 #ifndef USB_NO_VMM // vitrtul memory management
        return !list_empty(&inode->i_data.private_list);
  #else
        return 1 ;
 #endif
}

/*
 * osync is designed to support O_SYNC io.  It waits synchronously for
 * all already-submitted IO to complete, but does not queue any new
 * writes to the disk.
 *
 * To do O_SYNC writes, just queue the buffer writes with ll_rw_block as
 * you dirty the buffers, and then use osync_inode_buffers to wait for
 * completion.  Any other dirty buffers which are not yet queued for
 * write will not be flushed to disk by the osync.
 */
/*
static int osync_buffers_list(spinlock_t *lock, struct list_head *list)
{
        struct buffer_head *bh;
        struct list_head *p;
        int err = 0;

        spin_lock(lock);
repeat:
        list_for_each_prev(p, list) {
                bh = BH_ENTRY(p);
 #ifndef USB_NO_VMM // vitrtul memory management
                if (buffer_locked(bh)) {
                        get_bh(bh);
                        spin_unlock(lock);
                        wait_on_buffer(bh);
                        if (!buffer_uptodate(bh))
                                err = -EIO;
                        brelse(bh);
                        spin_lock(lock);
                        goto repeat;
                }
 #endif // ndef USB_NO_VMM
        }
        spin_unlock(lock);
        return err;
}
*/
/**
 * sync_mapping_buffers - write out and wait upon a mapping's "associated"
 *                        buffers
 * @buffer_mapping - the mapping which backs the buffers' data
 * @mapping - the mapping which wants those buffers written
 *
 * Starts I/O against the buffers at mapping->private_list, and waits upon
 * that I/O.
 *
 * Basically, this is a convenience function for fsync().  @buffer_mapping is
 * the blockdev which "owns" the buffers and @mapping is a file or directory
 * which needs those buffers to be written for a successful fsync().
 */
/*
int sync_mapping_buffers(struct address_space *mapping)
{
        struct address_space *buffer_mapping = mapping->assoc_mapping;

 #ifndef USB_NO_VMM // vitrtul memory management
        if (buffer_mapping == NULL || list_empty(&mapping->private_list))
                return 0;
 #endif // ndef USB_NO_VMM

 #ifndef NO_SPINLOCK
        return fsync_buffers_list(&buffer_mapping->private_lock,
                                        &mapping->private_list);
 #endif // ndef NO_SPINLOCK
}
*/
/*
EXPORT_SYMBOL(sync_mapping_buffers);
*/

/*
 * Called when we've recently written block `bblock', and it is known that
 * `bblock' was for a buffer_boundary() buffer.  This means that the block at
 * `bblock + 1' is probably a dirty indirect block.  Hunt it down and, if it's
 * dirty, schedule it for IO.  So that indirects merge nicely with their data.
 */
/*
void write_boundary_block(struct block_device *bdev,
                        sector_t bblock, unsigned blocksize)
{
 struct buffer_head *bh = __find_get_block(bdev, bblock + 1, blocksize);
 if (bh) {
                if (buffer_dirty(bh))
                        ll_rw_block(WRITE, 1, &bh);
                put_bh(bh);
        }
}
*/
/*
void mark_buffer_dirty_inode(struct buffer_head *bh, struct inode *inode)
{
 #ifndef NO_INODE
        struct address_space *mapping = inode->i_mapping;
 #endif // ndef NO_INODE
 #ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        struct address_space *buffer_mapping = bh->b_page->mapping;
 #endif 

        mark_buffer_dirty(bh);
        if (!mapping->assoc_mapping) {
                mapping->assoc_mapping = buffer_mapping;
        } else {
                if (mapping->assoc_mapping != buffer_mapping)
                        BUG();
        }
 #ifndef NO_SPINLOCK
        if (list_empty(&bh->b_assoc_buffers))
                buffer_insert_list(&buffer_mapping->private_lock,
                                bh, &mapping->private_list);
 #endif // ndef NO_SPINLOCK
}
EXPORT_SYMBOL(mark_buffer_dirty_inode);
*/
/*
 * Add a page to the dirty page list.
 *
 * It is a sad fact of life that this function is called from several places
 * deeply under spinlocking.  It may not sleep.
 *
 * If the page has buffers, the uptodate buffers are set dirty, to preserve
 * dirty-state coherency between the page and the buffers.  It the page does
 * not have buffers then when they are later attached they will all be set
 * dirty.
 *
 * The buffers are dirtied before the page is dirtied.  There's a small race
 * window in which a writepage caller may see the page cleanness but not the
 * buffer dirtiness.  That's fine.  If this code were to set the page dirty
 * before the buffers, a concurrent writepage caller could clear the page dirty
 * bit, see a bunch of clean buffers and we'd end up with dirty buffers/clean
 * page on the dirty page list.
 *
 * There is also a small window where the page is dirty, and not on dirty_pages.
 * Also a possibility that by the time the page is added to dirty_pages, it has
 * been set clean.  The page lists are somewhat approximate in this regard.
 * It's better to have clean pages accidentally attached to dirty_pages than to
 * leave dirty pages attached to clean_pages.
 *
 * We use private_lock to lock against try_to_free_buffers while using the
 * page's buffer list.  Also use this to protect against clean buffers being
 * added to the page after it was set dirty.
 *
 * FIXME: may need to call ->reservepage here as well.  That's rather up to the
 * address_space though.
 *
 * For now, we treat swapper_space specially.  It doesn't use the normal
 * block a_ops.
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
int __set_page_dirty_buffers(struct page *page)
{
        struct address_space * const mapping = page->mapping;
        int ret = 0;

        if (mapping == NULL) {
                SetPageDirty(page);
                goto out;
        }

 #ifndef NO_SPINLOCK
        spin_lock(&mapping->private_lock);
 #endif // ndef NO_SPINLOCK
        if (page_has_buffers(page)) {
                struct buffer_head *head = page_buffers(page);
                struct buffer_head *bh = head;

                do {
                        if (buffer_uptodate(bh))
                                set_buffer_dirty(bh);
                        else
                                buffer_error();
 #ifndef USB_NO_VMM // vitrtul memory management
                        bh = bh->b_this_page;
 #endif // ndef USB_NO_VMM
                } while (bh != head);
        }
 #ifndef NO_SPINLOCK
        spin_unlock(&mapping->private_lock);
 #endif // ndef NO_SPINLOCK

        if (!TestSetPageDirty(page)) {
                spin_lock(&mapping->page_lock);
                if (page->mapping) {    // Race with truncate? 
        #ifndef NO_BACKDEV 
                        if (!mapping->backing_dev_info->memory_backed)
                                inc_page_state(nr_dirty);
        #endif // ndef NO_BACKDEV
                        USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __set_page_dirty_buffers() : list_del() .. \n"));
                        list_del(&page->list);
                        list_add(&page->list, &mapping->dirty_pages);
                }
                spin_unlock(&mapping->page_lock);
                __mark_inode_dirty(mapping->host, I_DIRTY_PAGES);
        }
        
out:
        return ret;
}
#endif // ndef USB_NO_VMM
*/
/*
EXPORT_SYMBOL(__set_page_dirty_buffers);
*/
/*
 * Write out and wait upon a list of buffers.
 *
 * We have conflicting pressures: we want to make sure that all
 * initially dirty buffers get waited on, but that any subsequently
 * dirtied buffers don't.  After all, we don't want fsync to last
 * forever if somebody is actively writing to the file.
 *
 * Do this in two main stages: first we copy dirty buffers to a
 * temporary inode list, queueing the writes as we go.  Then we clean
 * up, waiting for those writes to complete.
 * 
 * During this second stage, any subsequent updates to the file may end
 * up refiling the buffer on the original inode's dirty list again, so
 * there is a chance we will end up with a buffer queued for write but
 * not yet completed on that list.  So, as a final cleanup we go through
 * the osync code to catch these locked, dirty buffers without requeuing
 * any newly dirty buffers for write.
 */
/*
int fsync_buffers_list(spinlock_t *lock, struct list_head *list)
{
        struct buffer_head *bh;
        struct list_head tmp;
        int err = 0, err2;

        INIT_LIST_HEAD(&tmp);

        spin_lock(lock);
        while (!list_empty(list)) {
                bh = BH_ENTRY(list->next);
                USBMS_DBG_FLOW(iprintf("[USBMS FLOW] fsync_buffers_list() : list_del_init() .. \n"));
                list_del_init(&bh->b_assoc_buffers);
 #ifndef USB_NO_VMM // vitrtul memory management
                if (buffer_dirty(bh) || buffer_locked(bh)) {
                        list_add(&bh->b_assoc_buffers, &tmp);
                        if (buffer_dirty(bh)) {
                                get_bh(bh);
                                spin_unlock(lock);
                                //
                                // Ensure any pending I/O completes so that
                                // ll_rw_block() actually writes the current
                                // contents - it is a noop if I/O is still in
                                // flight on potentially older contents.
                                
                                wait_on_buffer(bh);
                                ll_rw_block(WRITE, 1, &bh);
                                brelse(bh);
                                spin_lock(lock);
                        }
                }
 #endif // ndef USB_NO_VMM
        }

        while (!list_empty(&tmp)) {
                bh = BH_ENTRY(tmp.prev);
                __remove_assoc_queue(bh);
                get_bh(bh);
                spin_unlock(lock);
                wait_on_buffer(bh);
                if (!buffer_uptodate(bh))
                        err = -EIO;
                brelse(bh);
                spin_lock(lock);
        }
        
        spin_unlock(lock);
        err2 = osync_buffers_list(lock, list);
        if (err)
                return err;
        else
                return err2;
}
*/
/*
 * Invalidate any and all dirty buffers on a given inode.  We are
 * probably unmounting the fs, but that doesn't mean we have already
 * done a sync().  Just drop the buffers from the inode list.
 *
 * NOTE: we take the inode's blockdev's mapping's private_lock.  Which
 * assumes that all the buffers are against the blockdev.  Not true
 * for reiserfs.
 */

void invalidate_inode_buffers(struct inode *inode)
{
 if (inode_has_buffers(inode)) {
        #ifndef NO_INODE
                struct address_space *mapping = &inode->i_data;
                #ifndef USB_NO_VMM // vitrtul memory management
                        struct list_head *list = &mapping->private_list;
                        struct address_space *buffer_mapping = mapping->assoc_mapping;
                #endif // ndef USB_NO_VMM
                /*
                #ifndef NO_SPINLOCK
                        spin_lock(&buffer_mapping->private_lock);
                #endif // ndef NO_SPINLOCK
                */
                #ifndef USB_NO_VMM // vitrtul memory management
                while (!list_empty(list))
                        __remove_assoc_queue(BH_ENTRY(list->next));
                #endif // ndef USB_NO_VMM
                /*
                #ifndef NO_SPINLOCK
                        spin_unlock(&buffer_mapping->private_lock);
                #endif // ndef NO_SPINLOCK
                */
        #endif // ndef NO_INODE
        }
}

/*
 * Remove any clean buffers from the inode's buffer list.  This is called
 * when we're trying to free the inode itself.  Those buffers can pin it.
 *
 * Returns true if all buffers were removed.
 */
/*
int remove_inode_buffers(struct inode *inode)
{
 int ret = 1;

 if (inode_has_buffers(inode)) {
        #ifndef NO_INODE
                struct address_space *mapping = &inode->i_data;
                #ifndef USB_NO_VMM // vitrtul memory management
                        struct list_head *list = &mapping->private_list;
                        struct address_space *buffer_mapping = mapping->assoc_mapping;
                #endif // ndef USB_NO_VMM
                
                #ifndef NO_SPINLOCK
                spin_lock(&buffer_mapping->private_lock);
                #endif // ndef NO_SPINLOCK
                
                #ifndef USB_NO_VMM // vitrtul memory management
                        while (!list_empty(list)) {
                                struct buffer_head *bh = BH_ENTRY(list->next);
                                if (buffer_dirty(bh)) {
                                        ret = 0;
                                        break;
                                   }
                                __remove_assoc_queue(bh);
                           }
                #endif // ndef USB_NO_VMM
                
                #ifndef NO_SPINLOCK
                spin_unlock(&buffer_mapping->private_lock);
                #endif // ndef NO_SPINLOCK
                
        #endif // ndef NO_INODE
   }
 return ret;
}
*/
/*
 * Create the appropriate buffers when given a page for data area and
 * the size of each buffer.. Use the bh->b_this_page linked list to
 * follow the buffers created.  Return NULL if unable to create more
 * buffers.
 *
 * The retry flag is used to differentiate async IO (paging, swapping)
 * which may not fail from ordinary buffer allocations.
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        static struct buffer_head *
        create_buffers(struct page * page, unsigned long size, int retry)
#else
*/
static struct buffer_head * create_buffers(unsigned char *buf)
/*
#endif
*/
{
 struct buffer_head *bh, *head;
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        long offset;
 #endif // ndef USB_NO_VMM
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] create_buffers() Begin .. \n"));
 try_again:

 head = NULL;
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
 offset = PAGE_SIZE;
 while ((offset -= size) >= 0) {
 #endif // ndef USB_NO_VMM
 */
	/*
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
        USBMS_DBG_FLOW(iprintf("[USBMS FLOW] create_buffers() : alloc_buffer_head() .. \n"));
        // bh = alloc_buffer_head(GFP_NOFS);
		bh = alloc_buffer_head();
	 #else
	*/
	bh = global_bh ;
	/*
	#endif // NO_ELEVATOR
	*/

    USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] create_buffers() : bh=%08XH \n",
        bh )) ;
    if (!bh)
        goto no_grow;

    bh->b_bdev = NULL;
	USBMS_DBG_MEM(iprintf(
        "[USBMS MEM] create_buffers() : Clear Value : bh->b_bdev=%08XH \n",
   	    bh->b_bdev
       	));
	/*
    #ifndef USB_NO_VMM // vitrtul memory management
           bh->b_this_page = head;
    #endif // ndef USB_NO_VMM
	*/
    bh->b_blocknr = -1;
    head = bh;

    bh->b_state = 0;
	/*
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
        atomic_set(&bh->b_count, 0);
	#endif // ndef NO_ELEVATOR
	*/

	/*
	#ifndef USB_NO_VMM // vitrtul memory management
        bh->b_size = size;
	    USBMS_DBG_INIT(iprintf(
			"[USBMS INIT] create_buffers() : Set Value : bh->b_size=%d \n",
	         bh->b_size )) ;
	#endif // ndef USB_NO_VMM
	*/

    /* Link the buffer to its page */
    USBMS_DBG_FLOW(iprintf("[USBMS FLOW] create_buffers() : set_bh_page() .. \n"));
	/*
    #ifndef USB_NO_VMM // vitrtul memory management
		set_bh_page(bh, page, offset);
     #else
	*/

    set_bh_page(bh, buf);
	/*
    #endif
	*/
    bh->b_end_io = NULL;
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
   }
 #endif // ndef USB_NO_VMM
 */
 USBMS_DBG_MEM(iprintf("[USBMS MEM] create_buffers() : head=%08XH \n",
        head));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] create_buffers() End. \n"));
 return head;

 /*
  * In case anything failed, we just free everything we got.
  */
 no_grow:

 if (head) {
	/*
    #ifndef USB_NO_VMM // vitrtul memory management
		do {
    #endif // ndef USB_NO_VMM
	*/
			bh = head;
			/*
            #ifndef USB_NO_VMM // vitrtul memory management
    			head = head->b_this_page;
            #endif // ndef USB_NO_VMM
			*/
			USBMS_DBG_FLOW(iprintf(
				"[USBMS FLOW] create_buffers() : free_buffer_head() .. \n"
                ));
			free_buffer_head(bh);
	/*
	#ifndef USB_NO_VMM // vitrtul memory management
	       } while (head);
	#endif // ndef USB_NO_VMM
	*/
   }

 /*
  * Return failure for non-async IO requests.  Async IO requests
  * are not allowed to fail, so we have to wait until buffer heads
  * become available.  But we don't want tasks sleeping with 
  * partially complete buffers, so all were released above.
  */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	if (!retry) {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] create_buffers() End with Failed !! \n"));
		return NULL;
	   }
 #endif // ndef USB_NO_VMM
 */
 /* We're _really_ low on memory. Now we just
  * wait for old buffer heads to become free due to
  * finishing IO.  Since this is an async request and
  * the reserve list is empty, we're sure there are 
  * async buffer heads in use.
  */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        free_more_memory();
 #endif // ndef USB_NO_VMM
 */
 goto try_again;
}
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
static inline void
link_dev_buffers(struct page *page, struct buffer_head *head)
{
        struct buffer_head *bh, *tail;

        bh = head;
        do {
                tail = bh;
 #ifndef USB_NO_VMM // vitrtul memory management
                bh = bh->b_this_page;
 #endif // ndef USB_NO_VMM
        } while (bh);
 #ifndef USB_NO_VMM // vitrtul memory management
        tail->b_this_page = head;
 #endif // ndef USB_NO_VMM
        __set_page_buffers(page, head);
}
#endif // ndef USB_NO_VMM
*/

/*
 * Initialise the state of a blockdev page's buffers.
 */ 
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
static void
init_page_buffers(struct page *page, struct block_device *bdev,
                        int block, int size)
{
        struct buffer_head *head = page_buffers(page);
        struct buffer_head *bh = head;
        unsigned int b_state;

        b_state = 1 << BH_Mapped;
        if (PageUptodate(page))
                b_state |= 1 << BH_Uptodate;

        do {
                if (!(bh->b_state & (1 << BH_Mapped))) {
                        init_buffer(bh, NULL, NULL);
                        bh->b_bdev = bdev;
                        bh->b_blocknr = block;
                        bh->b_state = b_state;
                }
                block++;
 #ifndef USB_NO_VMM // vitrtul memory management
                bh = bh->b_this_page;
 #endif // ndef USB_NO_VMM
        } while (bh != head);
}
#endif // ndef USB_NO_VMM
*/
/*
 * Create the page-cache page that contains the requested block.
 *
 * This is user purely for blockdev mappings.
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
static struct page *
grow_dev_page(struct block_device *bdev, unsigned long block,
                        unsigned long index, int size)
{
 #ifndef NO_INODE
        struct inode *inode = bdev->bd_inode;
 #endif // ndef NO_INODE
 #ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        struct page *page;
 #endif
        struct buffer_head *bh;

 #ifndef NO_INODE
        page = find_or_create_page(inode->i_mapping, index, GFP_NOFS);
        if (!page)
                return NULL;
 #endif // ndef NO_INODE

        if (!PageLocked(page))
                BUG();

        if (page_has_buffers(page)) {
                bh = page_buffers(page);
                if (bh->b_size == size)
                        return page;
                if (!try_to_free_buffers(page))
                        goto failed;
        }

 
 //     Allocate some buffers for this page
         
        bh = create_buffers(page, size, 0);
        if (!bh)
                goto failed;

 
 //     Link the page to the buffers and initialise them.  Take the
 //     lock to be atomic wrt __find_get_block(), which does not
 //     run under the page lock.
         

 #ifndef NO_SPINLOCK
        spin_lock(&inode->i_mapping->private_lock);
 #endif // ndef NO_SPINLOCK
 
 link_dev_buffers(page, bh);
 init_page_buffers(page, bdev, block, size);
 
 #ifndef NO_SPINLOCK
        spin_unlock(&inode->i_mapping->private_lock);
 #endif // ndef NO_SPINLOCK
 
 return page;

 failed:
        buffer_error();
        unlock_page(page);
        page_cache_release(page);
        return NULL;
}
#endif // ndef USB_NO_VMM
*/
/*
 * Create buffers for the specified block device block's page.  If
 * that page was dirty, the buffers are set dirty also.
 *
 * Except that's a bug.  Attaching dirty buffers to a dirty
 * blockdev's page can result in filesystem corruption, because
 * some of those buffers may be aliases of filesystem data.
 * grow_dev_page() will go BUG() if this happens.
 */
/*
static inline int
grow_buffers(struct block_device *bdev, unsigned long block, int size)
{
 #ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        struct page *page;
 #endif // ndef USB_NO_VMM
        unsigned long index;
        int sizebits;

        // Size must be multiple of hard sectorsize 
        if (size & (bdev_hardsect_size(bdev)-1))
                BUG();
        if (size < 512 || size > PAGE_SIZE)
                BUG();

        sizebits = -1;
        do {
                sizebits++;
        } while ((size << sizebits) < PAGE_SIZE);

        index = block >> sizebits;
        block = index << sizebits;

        // Create a page with the proper size buffers.. 
        page = grow_dev_page(bdev, block, index, size);
        if (!page)
                return 0;
 #ifndef USB_NO_VMM // vitrtul memory management
        unlock_page(page);
 #endif // ndef USB_NO_VMM
        page_cache_release(page);
        return 1;
}
*/
/*
struct buffer_head *
__getblk_slow(struct block_device *bdev, sector_t block, int size)
{
 for (;;) {
                struct buffer_head * bh;

                bh = __find_get_block(bdev, block, size);
                if (bh)
                        return bh;

        #ifndef USB_NO_VMM // vitrtul memory management
                if (!grow_buffers(bdev, block, size))
                        free_more_memory();
        #endif // ndef USB_NO_VMM
        }
}
*/
/*
 * The relationship between dirty buffers and dirty pages:
 *
 * Whenever a page has any dirty buffers, the page's dirty bit is set, and
 * the page appears on its address_space.dirty_pages list.
 *
 * At all times, the dirtiness of the buffers represents the dirtiness of
 * subsections of the page.  If the page has buffers, the page dirty bit is
 * merely a hint about the true dirty state.
 *
 * When a page is set dirty in its entirety, all its buffers are marked dirty
 * (if the page has buffers).
 *
 * When a buffer is marked dirty, its page is dirtied, but the page's other
 * buffers are not.
 *
 * Also.  When blockdev buffers are explicitly read with bread(), they
 * individually become uptodate.  But their backing page remains not
 * uptodate - even if all of its buffers are uptodate.  A subsequent
 * block_read_full_page() against that page will discover all the uptodate
 * buffers, will set the page uptodate and will perform no I/O.
 */

/**
 * mark_buffer_dirty - mark a buffer_head as needing writeout
 *
 * mark_buffer_dirty() will set the dirty bit against the buffer,
 * then set its backing page dirty, then attach the page to its
 * address_space's dirty_pages list and then attach the address_space's
 * inode to its superblock's dirty inode list.
 *
 * mark_buffer_dirty() is atomic.  It takes bh->b_page->mapping->private_lock,
 * mapping->page_lock and the global inode_lock.
 */
/*
void mark_buffer_dirty(struct buffer_head *bh)
{
 
        if (!buffer_uptodate(bh))
                buffer_error();
 #ifndef USB_NO_VMM // vitrtul memory management
        if (!buffer_dirty(bh) && !test_set_buffer_dirty(bh))
        // struct page is on include/linux/mm.h
                __set_page_dirty_nobuffers(bh->b_page);
 #endif 
}
*/
/*
 * Decrement a buffer_head's reference count.  If all buffers against a page
 * have zero reference count, are clean and unlocked, and if the page is clean
 * and unlocked then try_to_free_buffers() may strip the buffers from the page
 * in preparation for freeing it (sometimes, rarely, buffers are removed from
 * a page but it ends up not being freed, and buffers may later be reattached).
 */
/*
void __brelse(struct buffer_head * buf)
{
 if (atomic_read(&buf->b_count)) {
                put_bh(buf);
                return;
        }
 
 // printk(KERN_ERR "VFS: brelse: Trying to free free buffer\n");
 
 dprintf("[USBMS] VFS: Error : Trying to free free buffer !! \n");
 buffer_error();                // For the stack backtrace 
}
*/
/*
 * bforget() is like brelse(), except it discards any
 * potentially dirty data.
 */
/*
void __bforget(struct buffer_head *bh)
{
        clear_buffer_dirty(bh);
        if (!list_empty(&bh->b_assoc_buffers)) {
 #ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
                struct address_space *buffer_mapping = bh->b_page->mapping;
 #endif // ndef USB_NO_VMM

 #ifndef NO_SPINLOCK
                spin_lock(&buffer_mapping->private_lock);
 #endif // ndef NO_SPINLOCK
                USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __bforget() : list_del_init() .. \n"));
                list_del_init(&bh->b_assoc_buffers);
                spin_unlock(&buffer_mapping->private_lock);
        }
        __brelse(bh);
}
*/
/*
static struct buffer_head *__bread_slow(struct buffer_head *bh)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __bread_slow() Begin .. \n"));
 
 //       lock_buffer(bh);

 //        if (buffer_uptodate(bh)) {
 //               unlock_buffer(bh);
 //               return bh;
 //       } else {
 //               if (buffer_dirty(bh))
 //                       buffer_error();
 
                get_bh(bh);
                bh->b_end_io = end_buffer_read_sync;
                USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __bread_slow() : submit_bh() .. \n"));
                submit_bh(READ, bh);
                wait_on_buffer(bh);
        
        //        if (buffer_uptodate(bh))
        
                        return bh;
 
 //       }
 
        brelse(bh);
        return NULL;
}
*/
/*
 * Per-cpu buffer LRU implementation.  To reduce the cost of __find_get_block().
 * The bhs[] array is sorted - newest buffer is at bhs[0].  Buffers have their
 * refcount elevated by one when they're in an LRU.  A buffer can only appear
 * once in a particular CPU's LRU.  A single buffer can be present in multiple
 * CPU's LRUs at the same time.
 *
 * This is a transparent caching front-end to sb_bread(), sb_getblk() and
 * sb_find_get_block().
 *
 * The LRUs themselves only need locking against invalidate_bh_lrus.  We use
 * a local interrupt disable for that.
 */
/*
#define BH_LRU_SIZE     8

struct bh_lru {
        struct buffer_head *bhs[BH_LRU_SIZE];
};
*/
/*
#ifndef _ETCPIP
static DEFINE_PER_CPU(struct bh_lru, bh_lrus) = {{0}};
#endif // ndef _ETCPIP
*/

#ifdef CONFIG_SMP
#define bh_lru_lock()   local_irq_disable()
#define bh_lru_unlock() local_irq_enable()
#else
/*
#define bh_lru_lock()   preempt_disable()
#define bh_lru_unlock() preempt_enable()
*/
#endif
/*
static inline void check_irqs_on(void)
{
 
 #ifdef irqs_disabled
        BUG_ON(irqs_disabled());
 #endif
 
}
*/
/*
 * The LRU management algorithm is dopey-but-simple.  Sorry.
 */
/*
static void bh_lru_install(struct buffer_head *bh)
{
        struct buffer_head *evictee = NULL;
        struct bh_lru *lru;

        check_irqs_on();
        bh_lru_lock();
        lru = &__get_cpu_var(bh_lrus);
        if (lru->bhs[0] != bh) {
                struct buffer_head *bhs[BH_LRU_SIZE];
                int in;
                int out = 0;

                get_bh(bh);
                bhs[out++] = bh;
                for (in = 0; in < BH_LRU_SIZE; in++) {
                        struct buffer_head *bh2 = lru->bhs[in];

                        if (bh2 == bh) {
                                __brelse(bh2);
                        } else {
                                if (out >= BH_LRU_SIZE) {
                                        BUG_ON(evictee != NULL);
                                        evictee = bh2;
                                } else {
                                        bhs[out++] = bh2;
                                }
                        }
                }
                while (out < BH_LRU_SIZE)
                        bhs[out++] = NULL;
                memcpy(lru->bhs, bhs, sizeof(bhs));
        }
        // bh_lru_unlock();

        if (evictee)
                __brelse(evictee);
}
*/
/*
 * Look up the bh in this cpu's LRU.  If it's there, move it to the head.
 */
/*
static inline struct buffer_head *
lookup_bh_lru(struct block_device *bdev, sector_t block, int size)
{
        struct buffer_head *ret = NULL;
        struct bh_lru *lru;
        int i;

        check_irqs_on();
        bh_lru_lock();
        lru = &__get_cpu_var(bh_lrus);
        for (i = 0; i < BH_LRU_SIZE; i++) {
                struct buffer_head *bh = lru->bhs[i];

                if (bh && bh->b_bdev == bdev &&
                                bh->b_blocknr == block && bh->b_size == size) {
                        if (i) {
                                while (i) {
                                        lru->bhs[i] = lru->bhs[i - 1];
                                        i--;
                                }
                                lru->bhs[0] = bh;
                        }
                        get_bh(bh);
                        ret = bh;
                        break;
                }
        }
        // bh_lru_unlock();
        return ret;
}
*/
/*
 * Perform a pagecache lookup for the matching buffer.  If it's there, refresh
 * it in the LRU and mark it as accessed.  If it is not present then return
 * NULL
 */

struct buffer_head *
__find_get_block(struct block_device *bdev, sector_t block, int size)
{
 
 #ifndef USB_NO_VMM // vitrtul memory management
         struct buffer_head *bh = lookup_bh_lru(bdev, block, size);
 #else
         struct buffer_head *bh ;
 #endif

 #ifdef USB_NO_VMM
        USBMS_DBG_MEM(iprintf(
                "[USBMS MEM] __find_get_block() : bh = UsbAllocMemC() .. \n"
                ));

        bh = (struct buffer_head *) UsbAllocMemC(USB_MEM_C_S
							#ifdef USB_DBG_MEM_C_ON
								, sizeof(struct buffer_head) // size : 36
                            #endif
						   ) ;
 #endif

 if (bh == NULL) {
        #ifndef USB_NO_VMM // vitrtul memory management

                bh = __find_get_block_slow(bdev, block, size);
                if (bh)
                        bh_lru_install(bh);
     #else
                USBMS_DBG_INFO(iprintf(
                        "[USBMS INFO] __find_get_block() : Error : out of memory for bh !! \n"
                        ));
        
    #endif
        }
 #ifndef USB_NO_VMM // vitrtul memory management
        if (bh)
                touch_buffer(bh);
 #endif // ndef USB_NO_VMM
 return bh;
}

/*
EXPORT_SYMBOL(__find_get_block);
*/
/*
 * __getblk will locate (and, if necessary, create) the buffer_head
 * which corresponds to the passed block_device, block and size. The
 * returned buffer has its reference count incremented.
 *
 * __getblk() cannot fail - it just keeps trying.  If you pass it an
 * illegal block number, __getblk() will happily return a buffer_head
 * which represents the non-existent block.  Very weird.
 *
 * __getblk() will lock up the machine if grow_dev_page's try_to_free_buffers()
 * attempt is failing.  FIXME, perhaps?
 */
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
struct buffer_head *
__getblk(struct block_device *bdev, sector_t block, int size)
{
 struct buffer_head *bh = __find_get_block(bdev, block, size);

 #ifndef USB_NO_VMM // vitrtul memory management
        if (bh == NULL)
                bh = __getblk_slow(bdev, block, size);
 #endif // ndef USB_NO_VMM
 return bh;
}
#endif // ndef NO_ELEVATOR
*/
/*
EXPORT_SYMBOL(__getblk);
*/
/*
 * Do async read-ahead on a buffer..
 */
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
void __breadahead(struct block_device *bdev, sector_t block, int size)
{
 struct buffer_head *bh = __getblk(bdev, block, size);

 ll_rw_block(READA, 1, &bh);
 brelse(bh);
}
#endif // ndef NO_ELEVATOR
*/
/*
EXPORT_SYMBOL(__breadahead);
*/
/**
 *  __bread() - reads a specified block and returns the bh
 *  @block: number of block
 *  @size: size (in bytes) to read
 * 
 *  Reads a specified block, and returns buffer head that contains it.
 *  It returns NULL if the block was unreadable.
 */
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
struct buffer_head *
__bread(struct block_device *bdev, sector_t block, int size)
{
 struct buffer_head *bh = __getblk(bdev, block, size);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __bread() Begin .. \n"));
 
 // if (!buffer_uptodate(bh))
 
   {
        USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __bread() : __bread_slow() .. \n"));
                bh = __bread_slow(bh);
   }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __bread() End. \n"));
 return bh;
}
#endif // ndef NO_ELEVATOR
*/
/*
EXPORT_SYMBOL(__bread);
*/
/*
 * invalidate_bh_lrus() is called rarely - at unmount.  Because it is only for
 * unmount it only needs to ensure that all buffers from the target device are
 * invalidated on return and it doesn't need to worry about new buffers from
 * that device being added - the unmount code has to prevent that.
 */
/*
static void invalidate_bh_lru(void *arg)
{
        struct bh_lru *b = &get_cpu_var(bh_lrus);
        int i;

        for (i = 0; i < BH_LRU_SIZE; i++) {
                brelse(b->bhs[i]);
                b->bhs[i] = NULL;
        }
        put_cpu_var(bh_lrus);
}
*/
/*
static void invalidate_bh_lrus(void)
{
        on_each_cpu(invalidate_bh_lru, NULL, 1, 1);
}
*/
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        void set_bh_page(struct buffer_head *bh,
                struct page *page, unsigned long offset)
 #else
*/
void set_bh_page(struct buffer_head *bh,
                unsigned char *buf)
/*
#endif
*/
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] set_bh_page() Begin .. \n"));
 USBMS_DBG_MEM(iprintf("[USBMS MEM] set_bh_page() : bh=%08XH \n",
	bh
	));

 #ifndef USBMS_BLK_CACHE // 2004/1/15 B. S. Liaw
	#ifdef USBMS_BS_ASSERT
		if ( !(USB_IS_UN_CACHED(buf)) ) {
			iprintf("[ASSERT][USBMS] set_bh_page() : " ) ;
			iprintf("Error : Not Un-Cached !!\n" ) ;
			iprintf("[ASSERT][USBMS] buf=%08XH \n", buf ) ;
			iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	#endif // USBMS_BS_ASSERT
 #endif // ndef USBMS_BLK_CACHE

 bh->b_buf = buf ;

 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] set_bh_page() : Set Value : bh->b_buf=%08XH  \n",
	bh->b_buf));

 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	 USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] set_bh_page() : offset=%08XH  \n",
		offset));

	if (offset >= PAGE_SIZE)
		BUG();
 #endif // ndef USB_NO_VMM
 */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	if (PageHighMem(page))
		// This catches illegal uses and preserves the offset:
		bh->b_data = (char *)(0 + offset);
       else
		bh->b_data = page_address(page) + offset;
  #else
 */
	bh->b_data = buf ;
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] set_bh_page() : Set Value : bh->b_data=%08XH  \n",
		bh->b_data
		));
 /*
 #endif // ndef USB_NO_VMM
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] set_bh_page() End. \n"));
}

/*
EXPORT_SYMBOL(set_bh_page);
*/
/*
 * Called when truncating a buffer on a page completely.
 */
/*
static inline void discard_buffer(struct buffer_head * bh)
{
        lock_buffer(bh);
        clear_buffer_dirty(bh);
        bh->b_bdev = NULL;
        clear_buffer_mapped(bh);
 
        clear_buffer_req(bh);
        clear_buffer_new(bh);
        clear_buffer_delay(bh);
        unlock_buffer(bh);
}
*/
/**
 * try_to_release_page() - release old fs-specific metadata on a page
 *
 * @page: the page which the kernel is trying to free
 * @gfp_mask: memory allocation flags (and I/O mode)
 *
 * The address_space is to try to release any data against the page
 * (presumably at page->private).  If the release was successful, return `1'.
 * Otherwise return zero.
 *
 * The @gfp_mask argument specifies whether I/O may be performed to release
 * this page (__GFP_IO), and whether the call may block (__GFP_WAIT).
 *
 * NOTE: @gfp_mask may go away, and this function may become non-blocking.
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        int try_to_release_page(struct page *page, int gfp_mask)
 #else
        int try_to_release_page(unsigned char *buf, int gfp_mask)
#endif
{
        struct address_space * const mapping = page->mapping;

        if (!PageLocked(page))
                BUG();
        if (PageWriteback(page))
                return 0;
        
        if (mapping && mapping->a_ops->releasepage)
                return mapping->a_ops->releasepage(page, gfp_mask);
        return try_to_free_buffers(page);
}
*/
/**
 * block_invalidatepage - invalidate part of all of a buffer-backed page
 *
 * @page: the page which is affected
 * @offset: the index of the truncation point
 *
 * block_invalidatepage() is called when all or part of the page has become
 * invalidatedby a truncate operation.
 *
 * block_invalidatepage() does not have to release all buffers, but it must
 * ensure that no dirty buffer is left outside @offset and that no I/O
 * is underway against any of the blocks which are outside the truncation
 * point.  Because the caller is about to free (and possibly reuse) those
 * blocks on-disk.
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        int block_invalidatepage(struct page *page, unsigned long offset)
 #else
        int block_invalidatepage(unsigned char *buf, unsigned long offset)
#endif
{
        struct buffer_head *head, *bh, *next;
        unsigned int curr_off = 0;
        int ret = 1;

        BUG_ON(!PageLocked(page));
        if (!page_has_buffers(page))
                goto out;

        head = page_buffers(page);
        bh = head;
        do {
                unsigned int next_off = curr_off + bh->b_size;
 #ifndef USB_NO_VMM // vitrtul memory management
                next = bh->b_this_page;
 #endif // ndef USB_NO_VMM

                
                // is this block fully invalidated?
                 
                if (offset <= curr_off)
                        discard_buffer(bh);
                curr_off = next_off;
                bh = next;
        } while (bh != head);

        
        // * We release buffers only if the entire page is being invalidated.
        // * The get_block cached value has been unconditionally invalidated,
        // * so real IO is not possible anymore.
         
        if (offset == 0)
                ret = try_to_release_page(page, 0);
out:
        return ret;
}
*/
/*
EXPORT_SYMBOL(block_invalidatepage);
*/
/*
 * We attach and possibly dirty the buffers atomically wrt
 * __set_page_dirty_buffers() via private_lock.  try_to_free_buffers
 * is already excluded via the page lock.
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	void create_empty_buffers(struct page *page,
			unsigned long blocksize, unsigned long b_state)
#else
*/
struct buffer_head * create_empty_buffers(unsigned char *buf)
/*
#endif
*/
{
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        struct buffer_head *bh, *head, *tail;
  #else
 */
 struct buffer_head *bh ;
 /*
 #endif
 */


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] create_empty_buffers() Begin .. \n"));

 #ifndef USBMS_BLK_CACHE // 2004/1/15 B. S. Liaw
	#ifdef USBMS_BS_ASSERT
		if ( !(USB_IS_UN_CACHED(buf)) ) {
			iprintf("[ASSERT][USBMS] create_empty_buffers() : " ) ;
			iprintf("Error : Not Un-Cached !!\n" ) ;
			iprintf("[ASSERT][USBMS] buf=%08XH \n", buf ) ;
			iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	#endif // USBMS_BS_ASSERT
 #endif // ndef USBMS_BLK_CACHE

 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] create_empty_buffers() : create_buffers() .. \n"
	));
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        head = create_buffers(page, blocksize, 1);
        USBMS_DBG_MEM(iprintf("[USBMS MEM] create_empty_buffers() : head=%08XH \n",
                head));
        bh = head;
        do {
                bh->b_state |= b_state;
                tail = bh;
                bh = bh->b_this_page;
           } while (bh);
        tail->b_this_page = head;
  #else
 */
  // bh = create_buffers(buf, blocksize, 1);
  bh = create_buffers(buf);
  USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] create_empty_buffers() : bh=%08XH \n",
	bh
	));
 /*
 // bh->b_size = blocksize;
 bh->b_size = 1 ;
 */
 /*
 bh->b_state |= b_state;
 */
 /*
 #endif
 */

 /* 
 #ifndef NO_SPINLOCK
        spin_lock(&page->mapping->private_lock);
 #endif // ndef NO_SPINLOCK
 */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        if (PageUptodate(page) || PageDirty(page)) {
                bh = head;
                do {
                        if (PageDirty(page))
                                set_buffer_dirty(bh);
                        if (PageUptodate(page))
                                set_buffer_uptodate(bh);
                        bh = bh->b_this_page;
                   } while (bh != head);
        }
 #endif // ndef USB_NO_VMM
 */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        USBMS_DBG_FLOW(iprintf("[USBMS FLOW] create_empty_buffers() : __set_page_buffers() .. \n"));
        __set_page_buffers(page, head);
 #endif // ndef USB_NO_VMM
 */
 /* 
 #ifndef NO_SPINLOCK
        spin_unlock(&page->mapping->private_lock);
 #endif // ndef NO_SPINLOCK
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] create_empty_buffers() End. \n"));
 
 // #ifdef USB_NO_VMM // vitrtul memory management
	return bh ;
 // #endif // USB_NO_VMM

}

/*
EXPORT_SYMBOL(create_empty_buffers);
*/

/*
 * We are taking a block for data and we don't want any output from any
 * buffer-cache aliases starting from return from that function and
 * until the moment when something will explicitly mark the buffer
 * dirty (hopefully that will not happen until we will free that block ;-)
 * We don't even need to mark it not-uptodate - nobody can expect
 * anything from a newly allocated buffer anyway. We used to used
 * unmap_buffer() for such invalidation, but that was wrong. We definitely
 * don't want to mark the alias unmapped, for example - it would confuse
 * anyone who might pick it with bread() afterwards...
 *
 * Also..  Note that bforget() doesn't lock the buffer.  So there can
 * be writeout I/O going on against recently-freed buffers.  We don't
 * wait on that I/O in bforget() - it's more efficient to wait on the I/O
 * only if we really need to.  That happens here.
 */
/*
void unmap_underlying_metadata(struct block_device *bdev, sector_t block)
{
 struct buffer_head *old_bh;

 old_bh = __find_get_block_slow(bdev, block, 0);
 if (old_bh) {
        #if 0   // This happens.  Later. 
                if (buffer_dirty(old_bh))
                        buffer_error();
        #endif
                clear_buffer_dirty(old_bh);
                wait_on_buffer(old_bh);
                clear_buffer_req(old_bh);
                __brelse(old_bh);
        }
}
*/
/*
EXPORT_SYMBOL(unmap_underlying_metadata);
*/
/*
 * NOTE! All mapped/uptodate combinations are valid:
 *
 *      Mapped  Uptodate        Meaning
 *
 *      No      No              "unknown" - must do get_block()
 *      No      Yes             "hole" - zero-filled
 *      Yes     No              "allocated" - allocated on disk, not read in
 *      Yes     Yes             "valid" - allocated and up-to-date in memory.
 *
 * "Dirty" is valid only with the last case (mapped+uptodate).
 */

/*
 * While block_write_full_page is writing back the dirty buffers under
 * the page lock, whoever dirtied the buffers may decide to clean them
 * again at any time.  We handle that by only looking at the buffer
 * state inside lock_buffer().
 *
 * If block_write_full_page() is called for regular writeback
 * (called_for_sync() is false) then it will redirty a page which has a locked
 * buffer.   This only can happen if someone has written the buffer directly,
 * with submit_bh().  At the address_space level PageWriteback prevents this
 * contention from occurring.
 */

// ====================================================================

/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        static int __block_write_full_page(struct inode *inode, struct page *page,
 #else
*/
int __block_write_full_page(struct inode *inode,
			sector_t iblock , // block number on Disk Partition
			unsigned int num_sectors ,
			unsigned char *buf,
/*
#endif
*/
			get_block_t *get_block
			/*
			#ifndef NO_VFS
				, struct writeback_control *wbc
			#endif // ndef NO_VFS
			*/
           )
{
 int err = 0;
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	unsigned long block;
	unsigned long last_block;
  #else
 */
	// unsigned int blocksize;
 /*
 #endif // USB_NO_VMM
 */
 struct buffer_head *bh, *head;
 int nr_underway = 0;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __block_write_full_page() Begin .. \n"));
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	BUG_ON(!PageLocked(page));
	if(!PageLocked(page)) {
        dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }
 #endif // ndef USB_NO_VMM
 */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
     #ifndef NO_INODE
        last_block = (i_size_read(inode) - 1) >> inode->i_blkbits;
     #endif // ndef NO_INODE
  #else
 */
 // #ifndef NO_INODE
 #ifndef INODE_REDUCED  
	blocksize = 1 << inode->i_blkbits;
 #endif // ndef INODE_REDUCED  
 // #endif // ndef NO_INODE
 /* USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] __block_write_full_page() : blocksize=%d \n",
	blocksize));
  */
 /*
 #endif
 */

 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        if(!page_has_buffers(page)) {
                if (!PageUptodate(page))
                        buffer_error();
                #ifndef NO_INODE
                        create_empty_buffers(page, 1 << inode->i_blkbits,
                                        (1 << BH_Dirty)|(1 << BH_Uptodate));
                #endif // ndef NO_INODE
           }
  #else
 */
 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] __block_write_full_page(1) : create_empty_buffers() .. \n"));
 head = create_empty_buffers(buf);
 /*
 head->b_size_sector = blocksize ;
 */
 // head->b_size_sector = 1 ;
 head->b_size_sector = num_sectors ;
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] __block_write_full_page() : Set Value : head->b_size_sector=%d \n",
    head->b_size_sector )) ;

 USBMS_DBG_MEM(iprintf("[USBMS MEM] __block_write_full_page() : head=%08XH \n",
        head));
 /*
 #endif // USB_NO_VMM
 */

 //      * Be very careful.  We have no exclusion from __set_page_dirty_buffers
 //      * here, and the (potentially unmapped) buffers may become dirty at
 //      * any time.  If a buffer becomes dirty here after we've inspected it
 //      * then we just miss that fact, and the page stays dirty.
 //      *
 //      * Buffers outside i_size may be dirtied by __set_page_dirty_buffers;
 //      * handle that here by just cleaning them.

 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        block = page->index << (PAGE_CACHE_SHIFT - inode->i_blkbits);
        head = page_buffers(page);
 #endif // ndef USB_NO_VMM
 */
 // #ifdef USB_NO_VMM // vitrtul memory management
	USBMS_DBG_MEM(iprintf("[USBMS MEM] __block_write_full_page() : iblock=%d \n",
        iblock));
	/*
	lblock = 1 ;
	USBMS_DBG_MEM(iprintf("[USBMS MEM] __block_write_full_page() : lblock=%d \n",
        lblock));
	*/
 // #endif // USB_NO_VMM

 bh = head;

 // #ifdef USB_NO_VMM // vitrtul memory management
        // nr = 1;
 // #endif // USB_NO_VMM

 // * Get all the dirty buffers mapped to disk addresses and
 // * handle any aliases from the underlying blockdev's mapping.
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
 do{
 #endif // ndef USB_NO_VMM
 */
	/*
	#ifndef USB_NO_VMM // vitrtul memory management
	if (block > last_block) {

                        // * mapped buffers outside i_size will occur, because
                        // * this page can be outside i_size when there is a
                        // * truncate in progress.
                        // *
                        // * if (buffer_mapped(bh))
                        // *    buffer_error();


                        // The buffer was zeroed by block_write_full_page()

                        clear_buffer_dirty(bh);
                        set_buffer_uptodate(bh);
	   }
	  else if (!buffer_mapped(bh) && buffer_dirty(bh))
	#endif // ndef USB_NO_VMM
	*/
       {
		/*
		#ifndef USB_NO_VMM // vitrtul memory management
                           if (buffer_new(bh))
                                   buffer_error();
        #endif // ndef USB_NO_VMM
		*/
		/*
        #ifndef USB_NO_VMM // vitrtul memory management
	                        err = get_block(inode, block, bh, 1);
		 #else
		*/
			err = get_block(inode, iblock, bh, 1);
		/*
		#endif // USB_NO_VMM
		*/

		if (!err) {
			USBMS_DBG_INFO(iprintf(
				"[USBMS INFO] __block_write_full_page() : Error !! \n"
				)) ;
			dprintf("[USBMS] Write Sector Internal Error !! \n") ;
		/*
        #ifndef USB_NO_VMM // vitrtul memory management
			goto recover;
		 #else
		*/
			goto done ;
		/*
		#endif // USB_NO_VMM
		*/
	   }

		/*
        #ifndef USB_NO_VMM // vitrtul memory management
			if (buffer_new(bh)) {
                  // blockdev mappings never come here
                  clear_buffer_new(bh);
	              unmap_underlying_metadata(bh->b_bdev,
                                            bh->b_blocknr);
               }
		#endif // ndef USB_NO_VMM
		*/
	   }
	/*
	#ifndef USB_NO_VMM // vitrtul memory management
			bh = bh->b_this_page;
	        block++;
	#endif // ndef USB_NO_VMM
	*/
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
   } while (bh != head);
 #endif // ndef USB_NO_VMM
 */

 /*
 #ifndef USB_NO_VMM // vitrtul memory management
 do{
 #endif // ndef USB_NO_VMM
 */
	get_bh(bh);
	/*
    #ifndef USB_NO_VMM // vitrtul memory management
	if (buffer_mapped(bh) && buffer_dirty(bh))
    #endif // ndef USB_NO_VMM
	*/
	   {
		/*
		#ifndef NO_VFS
                        if (wbc->sync_mode != WB_SYNC_NONE) {
                                lock_buffer(bh);
                           }
		#endif // ndef NO_VFS
		*/
		/*
        #ifndef USB_NO_VMM // vitrtul memory management
                          else {
                                        if (test_set_buffer_locked(bh)) {
                                                __set_page_dirty_nobuffers(page);
                                                continue;
                                           }
                           }
		#endif // ndef USB_NO_VMM
		*/
		/*
        #ifndef USB_NO_VMM // vitrtul memory management
        if (test_clear_buffer_dirty(bh))
        #endif // ndef USB_NO_VMM
		*/
           {
			/*
            #ifndef USB_NO_VMM // vitrtul memory management
                                if (!buffer_uptodate(bh))
                                        buffer_error();
            #endif // ndef USB_NO_VMM
			*/
			mark_buffer_async_write(bh);
           }
		/*
		#ifndef USB_NO_VMM // vitrtul memory management
                  else {
                        unlock_buffer(bh);
                   }
		#endif // ndef USB_NO_VMM
		*/
	   }
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
   } while ((bh = bh->b_this_page) != head);
 #endif // ndef USB_NO_VMM
 */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	BUG_ON(PageWriteback(page));
	SetPageWriteback(page);         // Keeps try_to_free_buffers() away
	unlock_page(page);
 #endif // ndef USB_NO_VMM
 */

 // The page may come unlocked any time after the *first* submit_bh()
 // call.  Be careful with its buffers.
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
 do{
 #endif // ndef USB_NO_VMM
 */
	/*
    #ifndef USB_NO_VMM // vitrtul memory management
	struct buffer_head *next = bh->b_this_page;

	if (buffer_async_write(bh))
    #endif // ndef USB_NO_VMM
	*/
	   {
        USBMS_DBG_INIT(iprintf(
			"[USBMS INIT] __block_write_full_page() : bh=%08XH \n",
   	        bh));
		/*
        USBMS_DBG_INIT(iprintf(
       	    "[USBMS INIT] __block_write_full_page() : bh->b_size=%d \n",
             bh->b_size )) ;
		*/
        USBMS_DBG_INIT(iprintf(
       	    "[USBMS INIT] __block_write_full_page() : bh->b_size_sector=%d \n",
             bh->b_size_sector )) ;
		/*
		#ifndef USBMS_DBG_READ_ONLY
		*/
		USBMS_DBG_FLOW(iprintf(
			"[USBMS FLOW] __block_write_full_page() : submit_bh() .. \n"));
		if (!submit_bh(WRITE, bh)) {
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __block_write_full_page() : End with Failed !! \n"));
			return -EIO ;
		   }
		/*
         #else
           #warning: USBMS_DBG_READ_ONLY : ON !!
           USBMS_DBG_INFO(iprintf("[USBMS INFO] __block_write_full_page(1) : Error : Read Only , Not Call submit_bh() !! \n"));
		#endif
		*/
		nr_underway++;
	   }

	put_bh(bh);
	/*
    #ifndef USB_NO_VMM // vitrtul memory management
                bh = next;
    #endif // ndef USB_NO_VMM
	*/
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
   } while (bh != head);
 #endif // ndef USB_NO_VMM
 */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	err = 0;
  #else
 */
	err = 1; // 1 : ok !
 /*
 #endif
 */
 // --------------
 done:
 // --------------

 if(nr_underway == 0) {

	// The page was marked dirty, but the buffers were
	// clean.  Someone wrote them back by hand with
	// ll_rw_block/submit_bh.  A rare case.
	/*                 
    #ifndef USB_NO_VMM // vitrtul memory management
                int uptodate = 1;

                do {
                        if (!buffer_uptodate(bh)) {
                                uptodate = 0;
                                break;
                           }
                        bh = bh->b_this_page;
                   } while (bh != head);

                if (uptodate)
                        SetPageUptodate(page);
                end_page_writeback(page);
    #endif // ndef USB_NO_VMM
	*/
   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __block_write_full_page() End. \n"));

 return err;

 /*                 
 #ifndef USB_NO_VMM // vitrtul memory management
 // --------------
 recover:
 // --------------
        
 // * ENOSPC, or some other error.  We may already have added some
 // * blocks to the file, so we need to write these out to avoid
 // * exposing stale data.
 // * The page is currently locked and not marked for writeback
         
 bh = head;
 // Recovery: lock and submit the mapped buffers
 
 // #ifndef USB_NO_VMM // vitrtul memory management
 //do{
 // #endif // ndef USB_NO_VMM
 
 	get_bh(bh);
	
    //#ifndef USB_NO_VMM // vitrtul memory management
    //            if (buffer_mapped(bh) && buffer_dirty(bh)) {
    //                    lock_buffer(bh);
    //#endif // ndef USB_NO_VMM
	
	mark_buffer_async_write(bh);
	
    //#ifndef USB_NO_VMM // vitrtul memory management
    //               } else {
    //                    
    //                    // * The buffer may have been set dirty during
    //                    // * attachment to a dirty page.
    //                     
    //                    clear_buffer_dirty(bh);
    //               }
	//#endif // ndef USB_NO_VMM
	
 
 //#ifndef USB_NO_VMM // vitrtul memory management
 //  } while ((bh = bh->b_this_page) != head);
 //#endif // ndef USB_NO_VMM
 

 
 //#ifndef USB_NO_VMM // vitrtul memory management
 //	SetPageError(page);
 //	BUG_ON(PageWriteback(page));
 //	SetPageWriteback(page);
 //	unlock_page(page);
 //#endif // ndef USB_NO_VMM
 
 
 //#ifndef USB_NO_VMM // vitrtul memory management
 //do{
 //#endif // ndef USB_NO_VMM
 
	
	//#ifndef USB_NO_VMM // vitrtul memory management
	//		struct buffer_head *next = bh->b_this_page;
    //#endif // ndef USB_NO_VMM
    
	
    //#ifndef USB_NO_VMM // vitrtul memory management
    //        if (buffer_async_write(bh)) {
	//			clear_buffer_dirty(bh);
	//            // #ifndef USBMS_DBG_READ_ONLY
	//				USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __block_write_full_page() : submit_bh() .. \n"));
    //                submit_bh(WRITE, bh);
    //            // #else
    //            //        USBMS_DBG_INFO(iprintf("[USBMS INFO] __block_write_full_page(2) : Error : Read Only , Not Call submit_bh() !! \n"));
    //            //#endif
	//			nr_underway++;
    //           }
    //    #endif // ndef USB_NO_VMM
	
	put_bh(bh);
	
    //#ifndef USB_NO_VMM // vitrtul memory management
    //            bh = next;
	//#endif // ndef USB_NO_VMM
	
 
 //#ifndef USB_NO_VMM // vitrtul memory management
 //	   } while (bh != head);
 //#endif // ndef USB_NO_VMM
 
 goto done;
 #endif // ndef USB_NO_VMM
 */
} // function __block_write_full_page()

// ====================================================================

/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        static int __block_prepare_write(struct inode *inode, struct page *page,
 #else
        static int __block_prepare_write(struct inode *inode, unsigned char *buf,
#endif
                unsigned from, unsigned to, get_block_t *get_block)
{
        unsigned block_start, block_end;
        sector_t block;
        int err = 0;
        unsigned blocksize, bbits;
        struct buffer_head *bh, *head, *wait[2], **wait_bh=wait;

        BUG_ON(!PageLocked(page));
        BUG_ON(from > PAGE_CACHE_SIZE);
        BUG_ON(to > PAGE_CACHE_SIZE);
        BUG_ON(from > to);

 #ifndef NO_INODE
        blocksize = 1 << inode->i_blkbits;
        #endif // ndef NO_INODE
        if (!page_has_buffers(page))
                create_empty_buffers(page, blocksize, 0);
        head = page_buffers(page);

 #ifndef NO_INODE
        bbits = inode->i_blkbits;
        #endif // ndef NO_INODE
 #ifndef USB_NO_VMM // vitrtul memory management
        block = (sector_t)page->index << (PAGE_CACHE_SHIFT - bbits);
 #endif // ndef USB_NO_VMM

        for(bh = head, block_start = 0; bh != head || !block_start;
 #ifndef USB_NO_VMM // vitrtul memory management
            block++, block_start=block_end, bh = bh->b_this_page) {
 #endif // ndef USB_NO_VMM
                block_end = block_start + blocksize;
                if (block_end <= from || block_start >= to) {
                        if (PageUptodate(page)) {
                                if (!buffer_uptodate(bh))
                                        set_buffer_uptodate(bh);
                        }
                        continue;
                }
                if (buffer_new(bh))
                        clear_buffer_new(bh);
                if (!buffer_mapped(bh)) {
                        err = get_block(inode, block, bh, 1);
                        if (err)
                                goto out;
                        if (buffer_new(bh)) {
                                clear_buffer_new(bh);
                                unmap_underlying_metadata(bh->b_bdev,
                                                        bh->b_blocknr);
                                if (PageUptodate(page)) {
                                        if (!buffer_mapped(bh))
                                                buffer_error();
                                        set_buffer_uptodate(bh);
                                        continue;
                                }
                                if (block_end > to || block_start < from) {
                                        void *kaddr;

                                        kaddr = kmap_atomic(page, KM_USER0);
                                        if (block_end > to)
                                                memset(kaddr+to, 0,
                                                        block_end-to);
                                        if (block_start < from)
                                                memset(kaddr+block_start,
                                                        0, from-block_start);
                                        flush_dcache_page(page);
                                        kunmap_atomic(kaddr, KM_USER0);
                                }
                                continue;
                        }
                }
                if (PageUptodate(page)) {
                        if (!buffer_uptodate(bh))
                                set_buffer_uptodate(bh);
                        continue; 
                }
                if (!buffer_uptodate(bh) && !buffer_delay(bh) &&
                     (block_start < from || block_end > to)) {
                        ll_rw_block(READ, 1, &bh);
                        *wait_bh++=bh;
                }
        }
        
        // If we issued read requests - let them complete.
         
        while(wait_bh > wait) {
                wait_on_buffer(*--wait_bh);
                if (!buffer_uptodate(*wait_bh))
                        return -EIO;
        }
        return 0;
out:
        
        // Zero out any newly allocated blocks to avoid exposing stale
        // data.  If BH_New is set, we know that the block was newly
        // allocated in the above loop.
         
        bh = head;
        block_start = 0;
        do {
                block_end = block_start+blocksize;
                if (block_end <= from)
                        goto next_bh;
                if (block_start >= to)
                        break;
                if (buffer_new(bh)) {
                        void *kaddr;

                        clear_buffer_new(bh);
                        if (buffer_uptodate(bh))
                                buffer_error();
                        kaddr = kmap_atomic(page, KM_USER0);
                        memset(kaddr+block_start, 0, bh->b_size);
                        kunmap_atomic(kaddr, KM_USER0);
                        set_buffer_uptodate(bh);
                        mark_buffer_dirty(bh);
                }
next_bh:
                block_start = block_end;
 #ifndef USB_NO_VMM // vitrtul memory management
                bh = bh->b_this_page;
 #endif // ndef USB_NO_VMM
        } while (bh != head);
        return err;
}
*/

/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        static int __block_commit_write(struct inode *inode, struct page *page,
 #else
        static int __block_commit_write(struct inode *inode, unsigned char *buf,
#endif
                unsigned from, unsigned to)
{
        unsigned block_start, block_end;
        int partial = 0;
        unsigned blocksize;
        struct buffer_head *bh, *head;

 #ifndef NO_INODE
        blocksize = 1 << inode->i_blkbits;
        #endif // ndef NO_INODE

        for(bh = head = page_buffers(page), block_start = 0;
            bh != head || !block_start;
 #ifndef USB_NO_VMM // vitrtul memory management
            block_start=block_end, bh = bh->b_this_page) {
 #endif // ndef USB_NO_VMM
                block_end = block_start + blocksize;
                if (block_end <= from || block_start >= to) {
                        if (!buffer_uptodate(bh))
                                partial = 1;
                } else {
                        set_buffer_uptodate(bh);
                        mark_buffer_dirty(bh);
                }
        }

        
        // If this is a partial write which happened to make all buffers
        // uptodate then we can optimize away a bogus readpage() for
        // the next read(). Here we 'discover' whether the page went
        // uptodate as a result of this (potentially partial) write.
        
 #ifndef USB_NO_VMM // vitrtul memory management
        if (!partial)
                SetPageUptodate(page);
 #endif // ndef USB_NO_VMM
        return 0;
}
*/
#if _USB2 > 0	// reduce function call layer
#else	// _USB2
/*
 * Generic "read page" function for block devices that have the normal
 * get_block functionality. This is most of the block device filesystems.
 * Reads the page asynchronously --- the unlock_buffer() and
 * set/clear_buffer_uptodate() functions propagate buffer state into the
 * page struct once IO has completed.
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        int block_read_full_page(struct page *page, get_block_t *get_block)
 #else
*/
int block_read_full_page(struct inode *inode,
                sector_t iblock , // block number on Disk Partition
				unsigned int num_sectors ,
                unsigned char *buf,
                get_block_t *get_block)
/*
#endif
*/
{
 #ifndef NO_INODE
	/*
	#ifndef USB_NO_VMM // vitrtul memory management
		struct inode *inode = page->mapping->host;
	#endif // ndef USB_NO_VMM
	*/
 #endif // ndef NO_INODE
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        sector_t iblock, lblock;
  #else
 */
 // sector_t lblock;
 /*
 #endif
 */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	struct buffer_head *bh, *head, *arr[MAX_BUF_PER_PAGE];
  #else
 */
	struct buffer_head *bh, *head , *arr ;
 /*
 #endif
 */
 unsigned int blocksize;
#ifndef _ETCPIP
 int nr, i;
 int fully_mapped = 1;
#endif	// ndef _ETCPIP


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] block_read_full_page() Begin .. \n"));
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        if (!PageLocked(page))
                PAGE_BUG(page);
        if (PageUptodate(page))
                buffer_error();
 #endif // ndef USB_NO_VMM
 */
 // #ifndef NO_INODE
 #ifndef INODE_REDUCED  
	blocksize = 1 << inode->i_blkbits;
 #endif // ndef INODE_REDUCED  
 // #endif // ndef NO_INODE
 USBMS_DBG_MEM(iprintf("[USBMS MEM] block_read_full_page() : blocksize=%d \n",
        blocksize));
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        if (!page_has_buffers(page))
                create_empty_buffers(page, blocksize, 0);
        head = page_buffers(page);
  #else
 */
 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] block_read_full_page(1) : create_empty_buffers() .. \n"
	));
 head = create_empty_buffers(buf);
 /*
 head->b_size = blocksize ;
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] block_read_full_page() : Set Value : head->b_size=%d \n",
    head->b_size )) ;
 */
 /*
 head->b_size_sector = blocksize ;
 */
 // head->b_size_sector = 1 ;
 head->b_size_sector = num_sectors ;
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] block_read_full_page() : Set Value : head->b_size_sector=%d \n",
    head->b_size_sector )) ;
 /*
 #endif // USB_NO_VMM
 */

 USBMS_DBG_MEM(iprintf("[USBMS MEM] block_read_full_page() : head=%08XH \n",
        head));
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        iblock = (sector_t)page->index << (PAGE_CACHE_SHIFT - inode->i_blkbits);
  #else
 */
    // iblock = 0 ; // block number on Disk Partition
 /*
 #endif
 */

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] block_read_full_page() : iblock=%d \n",
	iblock));

 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        #ifndef NO_INODE
                USBMS_DBG_FLOW(iprintf("[USBMS FLOW] block_read_full_page(2) : i_size_read() .. \n"));
                lblock = (i_size_read(inode)+blocksize-1) >> inode->i_blkbits;
         #endif // ndef NO_INODE
  #else
 */
	// lblock = 1 ;
 /*
 #endif
 */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	USBMS_DBG_MEM(iprintf("[USBMS MEM] block_read_full_page() : lblock=%d \n",
        lblock));
 #endif // ndef USB_NO_VMM
 */
 bh = head;
#ifndef _ETCPIP
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	nr = 0;
  #else
 */
	nr = 1;
 /*
 #endif
 */
 i = 0;
#endif	// ndef _ETCPIP

 /*
 #ifndef USB_NO_VMM // vitrtul memory management
 do{
 #endif // ndef USB_NO_VMM
 */
	/*
	#ifndef USB_NO_VMM // vitrtul memory management
	if (buffer_uptodate(bh))
		continue;
	#endif // ndef USB_NO_VMM
	*/
	/*
	#ifndef USB_NO_VMM // vitrtul memory management
	if (!buffer_mapped(bh))
	#endif // ndef USB_NO_VMM
	*/
	   {
#ifndef _ETCPIP
		fully_mapped = 0;
#endif	// ndef _ETCPIP
		/*
		#ifndef USB_NO_VMM // vitrtul memory management
		if (iblock < lblock)
		#endif // ndef USB_NO_VMM
		*/
		   {
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] block_read_full_page(3) : get_block() .. \n"));
			if (get_block(inode, iblock, bh, 0) > 0 ) {
				/*
				#ifndef USB_NO_VMM // vitrtul memory management
					SetPageError(page);
				#endif // ndef USB_NO_VMM
				*/
			   }
			  else {
				USBMS_DBG_FLOW(iprintf(
					"[USBMS FLOW] block_read_full_page(4) End with Failed !! \n"));
				return -EIO;
			   }
		   }
		/*
		#ifndef USB_NO_VMM // vitrtul memory management
			if (!buffer_mapped(bh)) {
				void *kaddr = kmap_atomic(page, KM_USER0);
                memset(kaddr + i * blocksize, 0, blocksize);
                flush_dcache_page(page);
                kunmap_atomic(kaddr, KM_USER0);
                set_buffer_uptodate(bh);
                continue;
               }
		#endif // ndef USB_NO_VMM
		*/
		// get_block() might have updated the buffer
		// synchronously
		/*
        #ifndef USB_NO_VMM // vitrtul memory management
        	if (buffer_uptodate(bh))
				continue;
        #endif // ndef USB_NO_VMM
		*/
	   }
	/*
	#ifndef USB_NO_VMM // vitrtul memory management
                arr[nr++] = bh;
         #else
	*/
                arr = bh;
	/*
        #endif
	*/
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
   } while (i++, iblock++, (bh = bh->b_this_page) != head);
 #endif // ndef USB_NO_VMM
 */

#ifndef _ETCPIP
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        if (fully_mapped)
                SetPageMappedToDisk(page);
 #endif // ndef USB_NO_VMM
 */
 if(!nr) {

	// All buffers are uptodate - we can set the page uptodate
    // as well. But not if get_block() returned an error.
	/*
    #ifndef USB_NO_VMM // vitrtul memory management
                if (!PageError(page))
                        SetPageUptodate(page);
                unlock_page(page);
    #endif // ndef USB_NO_VMM
	*/
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] block_read_full_page(5) End. \n"));
	return 0;
   }

 // Stage two: lock the buffers
 for(i = 0; i < nr; i++) {
	/*
	#ifndef USB_NO_VMM // vitrtul memory management
		bh = arr[i];
     #else
	*/
		bh = arr ;
	/*
    #endif
	*/
    lock_buffer(bh);
#endif	// ndef _ETCPIP

    USBMS_DBG_FLOW(iprintf("[USBMS FLOW] block_read_full_page(6) : mark_buffer_async_read() .. \n"));
    mark_buffer_async_read(bh);
#ifndef _ETCPIP
   }


 // Stage 3: start the IO.  Check for uptodateness
 // inside the buffer lock in case another process reading
 // the underlying blockdev brought it uptodate (the sct fix).

 for (i = 0; i < nr; i++) {
	/*
	#ifndef USB_NO_VMM // vitrtul memory management
        bh = arr[i];
      #else
	*/
        bh = arr ;
	/*
        #endif
	*/
	USBMS_DBG_INIT(iprintf("[USBMS INIT] block_read_full_page() : bh=%08XH \n",
            bh));
	#ifndef USB_NO_VMM // vitrtul memory management
	    USBMS_DBG_INIT(iprintf(
            "[USBMS INIT] block_read_full_page() : bh->b_size=%d \n",
                bh->b_size )) ;
	 #else
	    USBMS_DBG_INIT(iprintf(
            "[USBMS INIT] block_read_full_page() : bh->b_size_sector=%d \n",
                bh->b_size_sector )) ;
	#endif
	/*
    #ifndef USB_NO_VMM // vitrtul memory management
		if (buffer_uptodate(bh)) {
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] block_read_full_page(7) : end_buffer_async_read() .. \n"));
			end_buffer_async_read(bh, 1);
		   }
          else
	#endif // ndef USB_NO_VMM
	*/
		   {
#endif	// ndef _ETCPIP
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] block_read_full_page(8) : submit_bh() .. \n"));
			// submit_bh(READ, bh);
			if (!submit_bh(READ, bh)) {
				USBMS_DBG_FLOW(iprintf("[USBMS FLOW] block_read_full_page(9) : End with Failed !! \n"));
				return -EIO ;
			   }
#ifndef _ETCPIP
		   }
   }
#endif	// ndef _ETCPIP

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] block_read_full_page() End. \n"));
 // return 0;
 return 1 ;
}
#endif	// _USB2

/* utility function for filesystems that need to do work on expanding
 * truncates.  Uses prepare/commit_write to allow the filesystem to
 * deal with the hole.
 */
/*
int generic_cont_expand(struct inode *inode, loff_t size)
{
 #ifndef NO_INODE
        struct address_space *mapping = inode->i_mapping;
 #endif // ndef NO_INODE
 #ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        struct page *page;
 #endif
        unsigned long index, offset, limit;
        int err;

        err = -EFBIG;
 #ifndef _ETCPIP
        limit = current->rlim[RLIMIT_FSIZE].rlim_cur;
        if (limit != RLIM_INFINITY && size > (loff_t)limit) {
                send_sig(SIGXFSZ, current, 0);
                goto out;
        }
 #endif // ndef _ETCPIP
 // #ifndef INODE_REDUCED
        if (size > inode->i_sb->s_maxbytes)
                goto out;
 // #endif // ndef INODE_REDUCED

        offset = (size & (PAGE_CACHE_SIZE-1)); // Within page 

        // ugh.  in prepare/commit_write, if from==to==start of block, we 
        // skip the prepare.  make sure we never send an offset for the start
        // of a block
        
 // #ifndef INODE_REDUCED
        if ((offset & (inode->i_sb->s_blocksize - 1)) == 0) {
                offset++;
        }
 // #endif // ndef INODE_REDUCED
 #ifndef USB_NO_VMM // vitrtul memory management
        index = size >> PAGE_CACHE_SHIFT;
 #endif // ndef USB_NO_VMM
        err = -ENOMEM;
        page = grab_cache_page(mapping, index);
        if (!page)
                goto out;
        err = mapping->a_ops->prepare_write(NULL, page, offset, offset);
        if (!err) {
                err = mapping->a_ops->commit_write(NULL, page, offset, offset);
        }
 #ifndef USB_NO_VMM // vitrtul memory management
        unlock_page(page);
 #endif // ndef USB_NO_VMM
        page_cache_release(page);
        if (err > 0)
                err = 0;
out:
        return err;
}
*/

/*
 * For moronic filesystems that do not allow holes in file.
 * We may have to extend the file.
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        int cont_prepare_write(struct page *page, unsigned offset,
 #else
        int cont_prepare_write(unsigned char *buf, unsigned offset,
#endif
                unsigned to, get_block_t *get_block, loff_t *bytes)
{
        struct address_space *mapping = page->mapping;
        struct inode *inode = mapping->host;
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        struct page *new_page;
#endif
        unsigned long pgpos;
        long status;
        unsigned zerofrom;
 #ifndef NO_INODE
        unsigned blocksize = 1 << inode->i_blkbits;
        #endif // ndef NO_INODE
        void *kaddr;

 #ifndef USB_NO_VMM // vitrtul memory management
        while(page->index > (pgpos = *bytes>>PAGE_CACHE_SHIFT)) {
                status = -ENOMEM;
                new_page = grab_cache_page(mapping, pgpos);
                if (!new_page)
                        goto out;
                // we might sleep 
                if (*bytes>>PAGE_CACHE_SHIFT != pgpos) {
 #ifndef USB_NO_VMM // vitrtul memory management
                        unlock_page(new_page);
 #endif // ndef USB_NO_VMM
                        page_cache_release(new_page);
                        continue;
                }
                zerofrom = *bytes & ~PAGE_CACHE_MASK;
                if (zerofrom & (blocksize-1)) {
                        *bytes |= (blocksize-1);
                        (*bytes)++;
                }
                status = __block_prepare_write(inode, new_page, zerofrom,
                                                PAGE_CACHE_SIZE, get_block);
                if (status)
                        goto out_unmap;
                kaddr = kmap_atomic(new_page, KM_USER0);
                memset(kaddr+zerofrom, 0, PAGE_CACHE_SIZE-zerofrom);
                flush_dcache_page(new_page);
                kunmap_atomic(kaddr, KM_USER0);
                __block_commit_write(inode, new_page,
                                zerofrom, PAGE_CACHE_SIZE);
 #ifndef USB_NO_VMM // vitrtul memory management
                unlock_page(new_page);
 #endif // ndef USB_NO_VMM
                page_cache_release(new_page);
        }
 #endif // ndef USB_NO_VMM

        if (page->index < pgpos) {
                // completely inside the area 
                zerofrom = offset;
        } else {
                // page covers the boundary, find the boundary offset 
                zerofrom = *bytes & ~PAGE_CACHE_MASK;

                // if we will expand the thing last block will be filled 
                if (to > zerofrom && (zerofrom & (blocksize-1))) {
                        *bytes |= (blocksize-1);
                        (*bytes)++;
                }

                // starting below the boundary? Nothing to zero out 
                if (offset <= zerofrom)
                        zerofrom = offset;
        }
        status = __block_prepare_write(inode, page, zerofrom, to, get_block);
        if (status)
                goto out1;
        if (zerofrom < offset) {
                kaddr = kmap_atomic(page, KM_USER0);
                memset(kaddr+zerofrom, 0, offset-zerofrom);
                flush_dcache_page(page);
                kunmap_atomic(kaddr, KM_USER0);
                __block_commit_write(inode, page, zerofrom, offset);
        }
        return 0;
out1:
        ClearPageUptodate(page);
        return status;

out_unmap:
        ClearPageUptodate(new_page);
 #ifndef USB_NO_VMM // vitrtul memory management
        unlock_page(new_page);
 #endif // ndef USB_NO_VMM
        page_cache_release(new_page);
out:
        return status;
}
*/
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        int block_prepare_write(struct page *page, unsigned from, unsigned to,
 #else
        int block_prepare_write(unsigned char *buf, unsigned from, unsigned to,
#endif
                        get_block_t *get_block)
{
        struct inode *inode = page->mapping->host;
        int err = __block_prepare_write(inode, page, from, to, get_block);
        if (err)
                ClearPageUptodate(page);
        return err;
}
*/
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        int block_commit_write(struct page *page, unsigned from, unsigned to)
 #else
        int block_commit_write(unsigned char *buf, unsigned from, unsigned to)
#endif
{
        struct inode *inode = page->mapping->host;
        __block_commit_write(inode,page,from,to);
        return 0;
}
*/
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        int generic_commit_write(struct file *file, struct page *page,
 #else
        int generic_commit_write(struct file *file, unsigned char *buf,
#endif
                unsigned from, unsigned to)
{
        struct inode *inode = page->mapping->host;
 #endif // ndef USB_NO_VMM
 #ifndef USB_NO_VMM // vitrtul memory management
        loff_t pos = ((loff_t)page->index << PAGE_CACHE_SHIFT) + to;
 #endif // ndef USB_NO_VMM
        __block_commit_write(inode,page,from,to);
        
        // No need to use i_size_read() here, the i_size
        // cannot change under us because we hold i_sem.
         
        if (pos > inode->i_size) {
                i_size_write(inode, pos);
                mark_inode_dirty(inode);
        }
        return 0;
}
*/
/*
 * On entry, the page is fully not uptodate.
 * On exit the page is fully uptodate in the areas outside (from,to)
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        int nobh_prepare_write(struct page *page, unsigned from, unsigned to,
                        get_block_t *get_block)
 #else
        int nobh_prepare_write(unsigned char *buf, unsigned from, unsigned to,
                        get_block_t *get_block)
#endif
{
        struct inode *inode = page->mapping->host;
 #endif // ndef USB_NO_VMM
 #ifndef NO_INODE
        const unsigned blkbits = inode->i_blkbits;
        #endif // ndef NO_INODE
        const unsigned blocksize = 1 << blkbits;
        struct buffer_head map_bh;
 #ifndef USB_NO_VMM // vitrtul memory management
        struct buffer_head *read_bh[MAX_BUF_PER_PAGE];
 #endif // ndef USB_NO_VMM
        unsigned block_in_page;
        unsigned block_start;
        sector_t block_in_file;
        char *kaddr;
        int nr_reads = 0;
        int i;
        int ret = 0;
        int is_mapped_to_disk = 1;
        int dirtied_it = 0;

        if (PageMappedToDisk(page))
                return 0;

 #ifndef USB_NO_VMM // vitrtul memory management
        block_in_file = (sector_t)page->index << (PAGE_CACHE_SHIFT - blkbits);
 #endif // ndef USB_NO_VMM
 #ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        map_bh.b_page = page;
 #endif 

        
        // We loop across all blocks in the page, whether or not they are
        // part of the affected region.  This is so we can discover if the
        // page is fully mapped-to-disk.
         
        for (block_start = 0, block_in_page = 0;
                  block_start < PAGE_CACHE_SIZE;
                  block_in_page++, block_start += blocksize) {
                unsigned block_end = block_start + blocksize;
                int create;

                map_bh.b_state = 0;
                create = 1;
                if (block_start >= to)
                        create = 0;
                ret = get_block(inode, block_in_file + block_in_page,
                                        &map_bh, create);
                if (ret)
                        goto failed;
                if (!buffer_mapped(&map_bh))
                        is_mapped_to_disk = 0;
                if (buffer_new(&map_bh))
                        unmap_underlying_metadata(map_bh.b_bdev,
                                                        map_bh.b_blocknr);
                if (PageUptodate(page))
                        continue;
                if (buffer_new(&map_bh) || !buffer_mapped(&map_bh)) {
                        kaddr = kmap_atomic(page, KM_USER0);
                        if (block_start < from) {
                                memset(kaddr+block_start, 0, from-block_start);
                                dirtied_it = 1;
                        }
                        if (block_end > to) {
                                memset(kaddr + to, 0, block_end - to);
                                dirtied_it = 1;
                        }
                        flush_dcache_page(page);
                        kunmap_atomic(kaddr, KM_USER0);
                        continue;
                }
                if (buffer_uptodate(&map_bh))
                        continue;       // reiserfs does this 
                if (block_start < from || block_end > to) {
                        #ifndef _ETCPIP
                                struct buffer_head *bh = alloc_buffer_head(GFP_NOFS);
                         #else
                                struct buffer_head *bh = alloc_buffer_head();
                        #endif

                        if (!bh) {
                                ret = -ENOMEM;
                                goto failed;
                        }
                        bh->b_state = map_bh.b_state;
                        atomic_set(&bh->b_count, 0);
 #ifndef USB_NO_VMM // vitrtul memory management
                        bh->b_this_page = 0;
 #endif // ndef USB_NO_VMM
 #ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
                        bh->b_page = page;
  #else
                        bh->b_buf = buf ;
 #endif 
                        bh->b_blocknr = map_bh.b_blocknr;
                        bh->b_size = blocksize;
                        bh->b_data = (char *)(long)block_start;
                        bh->b_bdev = map_bh.b_bdev;
                        bh->b_private = NULL;
                        read_bh[nr_reads++] = bh;
                }
        }

        if (nr_reads) {
                ll_rw_block(READ, nr_reads, read_bh);
                for (i = 0; i < nr_reads; i++) {
                        wait_on_buffer(read_bh[i]);
                        if (!buffer_uptodate(read_bh[i]))
                                ret = -EIO;
                        free_buffer_head(read_bh[i]);
                        read_bh[i] = NULL;
                }
                if (ret)
                        goto failed;
        }

 #ifndef USB_NO_VMM // vitrtul memory management
        if (is_mapped_to_disk)
                SetPageMappedToDisk(page);
        SetPageUptodate(page);
 #endif // ndef USB_NO_VMM

        //
        // Setting the page dirty here isn't necessary for the prepare_write
        // function - commit_write will do that.  But if/when this function is
        // used within the pagefault handler to ensure that all mmapped pages
        // have backing space in the filesystem, we will need to dirty the page
        // if its contents were altered.
        
        if (dirtied_it)
                set_page_dirty(page);

        return 0;

failed:
        for (i = 0; i < nr_reads; i++) {
                if (read_bh[i])
                        free_buffer_head(read_bh[i]);
        }

        
        // Error recovery is pretty slack.  Clear the page and mark it dirty
        // so we'll later zero out any blocks which _were_ allocated.
         
        kaddr = kmap_atomic(page, KM_USER0);
        memset(kaddr, 0, PAGE_CACHE_SIZE);
        kunmap_atomic(kaddr, KM_USER0);
 #ifndef USB_NO_VMM // vitrtul memory management
        SetPageUptodate(page);
 #endif // ndef USB_NO_VMM
        set_page_dirty(page);
        return ret;
}
EXPORT_SYMBOL(nobh_prepare_write);
*/
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        int nobh_commit_write(struct file *file, struct page *page,
 #else
        int nobh_commit_write(struct file *file, unsigned char *buf,
                unsigned from, unsigned to)
#endif
{
        struct inode *inode = page->mapping->host;
 #endif // ndef USB_NO_VMM
#ifndef USB_NO_VMM // vitrtul memory management
        loff_t pos = ((loff_t)page->index << PAGE_CACHE_SHIFT) + to;
 #endif // ndef USB_NO_VMM

        set_page_dirty(page);
        if (pos > inode->i_size) {
                i_size_write(inode, pos);
                mark_inode_dirty(inode);
        }
        return 0;
}
*/
/*
EXPORT_SYMBOL(nobh_commit_write);
*/
/*
 * This function assumes that ->prepare_write() uses nobh_prepare_write().
 */
/*
int nobh_truncate_page(struct address_space *mapping, loff_t from)
{
        struct inode *inode = mapping->host;
 #ifndef NO_INODE
        unsigned blocksize = 1 << inode->i_blkbits;
        #endif // ndef NO_INODE
#ifndef USB_NO_VMM // vitrtul memory management
        pgoff_t index = from >> PAGE_CACHE_SHIFT;
 #endif // ndef USB_NO_VMM
        unsigned offset = from & (PAGE_CACHE_SIZE-1);
        unsigned to;
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        struct page *page;
#endif
        struct address_space_operations *a_ops = mapping->a_ops;
        char *kaddr;
        int ret = 0;

        if ((offset & (blocksize - 1)) == 0)
                goto out;

        ret = -ENOMEM;
        page = grab_cache_page(mapping, index);
        if (!page)
                goto out;

        to = (offset + blocksize) & ~(blocksize - 1);
        ret = a_ops->prepare_write(NULL, page, offset, to);
        if (ret == 0) {
                kaddr = kmap_atomic(page, KM_USER0);
                memset(kaddr + offset, 0, PAGE_CACHE_SIZE - offset);
                flush_dcache_page(page);
                kunmap_atomic(kaddr, KM_USER0);
                set_page_dirty(page);
        }
 #ifndef USB_NO_VMM // vitrtul memory management
        unlock_page(page);
 #endif // ndef USB_NO_VMM
        page_cache_release(page);
out:
        return ret;
}
*/
/*
EXPORT_SYMBOL(nobh_truncate_page);
*/
/*
int block_truncate_page(struct address_space *mapping,
                        loff_t from, get_block_t *get_block)
{
#ifndef USB_NO_VMM // vitrtul memory management
        unsigned long index = from >> PAGE_CACHE_SHIFT;
 #endif // ndef USB_NO_VMM
        unsigned offset = from & (PAGE_CACHE_SIZE-1);
        unsigned blocksize, iblock, length, pos;
        struct inode *inode = mapping->host;
 #endif // ndef USB_NO_VMM
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        struct page *page;
#endif
        struct buffer_head *bh;
        void *kaddr;
        int err;

 #ifndef NO_INODE
        blocksize = 1 << inode->i_blkbits;
        #endif // ndef NO_INODE
        length = offset & (blocksize - 1);

        // Block boundary? Nothing to do 
        if (!length)
                return 0;

        length = blocksize - length;
#ifndef USB_NO_VMM // vitrtul memory management
        iblock = index << (PAGE_CACHE_SHIFT - inode->i_blkbits);
 #endif // ndef USB_NO_VMM
        
        page = grab_cache_page(mapping, index);
 #endif // ndef USB_NO_VMM
        err = -ENOMEM;
        if (!page)
                goto out;

        if (!page_has_buffers(page))
                create_empty_buffers(page, blocksize, 0);

        //   Find the buffer that contains "offset" 
        bh = page_buffers(page);
        pos = blocksize;
        while (offset >= pos) {
 #ifndef USB_NO_VMM // vitrtul memory management
                bh = bh->b_this_page;
 #endif // ndef USB_NO_VMM
                iblock++;
                pos += blocksize;
        }

        err = 0;
        if (!buffer_mapped(bh)) {
                err = get_block(inode, iblock, bh, 0);
                if (err)
                        goto unlock;
                // unmapped? It's a hole - nothing to do 
                if (!buffer_mapped(bh))
                        goto unlock;
        }

        // Ok, it's mapped. Make sure it's up-to-date 
        if (PageUptodate(page))
                set_buffer_uptodate(bh);

        if (!buffer_uptodate(bh) && !buffer_delay(bh)) {
                err = -EIO;
                ll_rw_block(READ, 1, &bh);
                wait_on_buffer(bh);
                // Uhhuh. Read error. Complain and punt. 
                if (!buffer_uptodate(bh))
                        goto unlock;
        }

        kaddr = kmap_atomic(page, KM_USER0);
        memset(kaddr + offset, 0, length);
        flush_dcache_page(page);
        kunmap_atomic(kaddr, KM_USER0);

        mark_buffer_dirty(bh);
        err = 0;

 unlock:
 #ifndef USB_NO_VMM // vitrtul memory management
        unlock_page(page);
 #endif // ndef USB_NO_VMM
        page_cache_release(page);
 out:
        return err;
}
*/

#if _USB2 > 0	// reduce function call layer
#else	// _USB2
/*
 * The generic ->writepage function for buffer-backed address_spaces
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        int block_write_full_page(struct page *page, get_block_t *get_block,
 #else
*/
int block_write_full_page(struct inode *inode,
		sector_t iblock , // block number on Disk Partition
		unsigned int num_sectors ,
		unsigned char *buf, get_block_t *get_block
/*
#endif
*/
		/*
        #ifndef NO_VFS
                        , struct writeback_control *wbc
        #endif // ndef NO_VFS
		*/
)
{
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        struct inode * const inode = page->mapping->host;
 #endif // ndef USB_NO_VMM
 */
 loff_t i_size = i_size_read(inode);
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        const unsigned long end_index = i_size >> PAGE_CACHE_SHIFT;
 #endif // ndef USB_NO_VMM
 */
 unsigned offset;
 void *kaddr;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] block_write_full_page() Begin .. \n"));
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        // Is the page fully inside i_size? 
        if (page->index < end_index)
                return __block_write_full_page(inode, page, get_block
                                #ifndef NO_VFS
                                        , wbc
                                #endif // ndef NO_VFS
                                );
 #endif // ndef USB_NO_VMM
 */
 // Is the page fully outside i_size? (truncate in progress) 
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        offset = i_size & (PAGE_CACHE_SIZE-1);
        if (page->index >= end_index+1 || !offset) {
                
                //  The page may have dirty, unmapped buffers.  For example,
                //  they may have been added in ext3_writepage().  Make them
                //  freeable here, so the page does not leak.
                 
                block_invalidatepage(page, 0);
                unlock_page(page);
                return 0; // don't care 
           }
 #endif // ndef USB_NO_VMM
 */
        
 // * The page straddles i_size.  It must be zeroed out on each and every
 // * writepage invocation because it may be mmapped.  "A file is mapped
 // * in multiples of the page size.  For a file that is not a multiple of
 // * the  page size, the remaining memory is zeroed when mapped, and
 // * writes to that region are not written out to the file."
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        kaddr = kmap_atomic(page, KM_USER0);
        memset(kaddr + offset, 0, PAGE_CACHE_SIZE - offset);
        flush_dcache_page(page);
        kunmap_atomic(kaddr, KM_USER0);
 #endif // ndef USB_NO_VMM
 */
 USBMS_DBG_MEM(iprintf("[USBMS MEM] block_write_full_page() : iblock=%08XH \n",
	iblock));

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] block_write_full_page() End with Call __block_write_full_page() .. \n"));
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
        return __block_write_full_page(inode, page, get_block, wbc);
  #else
 */
	return __block_write_full_page(inode, iblock,
						num_sectors ,
						buf, get_block
                        #ifndef NO_VFS
                                , wbc
                        #endif // ndef NO_VFS
                        );
 /*
 #endif
 */
}
#endif	// _USB2

/*
sector_t generic_block_bmap(struct address_space *mapping, sector_t block,
                            get_block_t *get_block)
{
        struct buffer_head tmp;
 #ifndef NO_INODE
        struct inode *inode = mapping->host;
 #endif // ndef NO_INODE
        tmp.b_state = 0;
        tmp.b_blocknr = 0;
 #ifndef INODE_REDUCED  
        get_block(inode, block, &tmp, 0);
 #endif // ndef INODE_REDUCED  
        return tmp.b_blocknr;
}
*/

static int end_bio_bh_io_sync(struct bio *bio, unsigned int bytes_done, int err)
{
 struct buffer_head *bh = bio->bi_private;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] end_bio_bh_io_sync() Begin .. \n"));
 if (bio->bi_size) {
	return 1;
   }

 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] end_bio_bh_io_sync() : bh->b_end_io() .. \n"
	));
 bh->b_end_io(bh, test_bit(BIO_UPTODATE, &bio->bi_flags));

 // B. S. Liaw : It is impotatnt to call bio_put , otherwise memory leak
 bio_put(bio);

 // #ifdef NO_ELEVATOR
	// 2003/12/22 B. S. Liaw Add ..
	USBMS_DBG_FLOW(iprintf(
		"[USBMS FLOW] end_bio_bh_io_sync() : free_buffer_head() .. \n"
		));
	free_buffer_head(bh) ;
	bio->bi_private = NULL ;
 // #endif // NO_ELEVATOR 
 
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] end_bio_bh_io_sync() End. \n"));
 return 0;
}

int submit_bh(int rw, struct buffer_head * bh)
{
 struct bio *bio;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] submit_bh() Begin .. \n"));
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
 BUG_ON(!buffer_locked(bh));
 
 if (!buffer_locked(bh)) {
                dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
           }
 #endif // ndef USB_NO_VMM
 */

 /*
 #ifndef USB_NO_VMM // vitrtul memory management
 BUG_ON(!buffer_mapped(bh));
 
 if (!buffer_mapped(bh)) {
                dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
           }
 #endif // ndef USB_NO_VMM
 */

 USBMS_DBG_MEM(iprintf("[USBMS MEM] submit_bh() : bh=%08XH \n",
        bh));
 USBMS_DBG_INIT(iprintf("[USBMS INIT] submit_bh() : bh->b_end_io=%08XH \n",
        bh->b_end_io));
 /*
 BUG_ON(!bh->b_end_io);
 */
 if(!bh->b_end_io) {
	dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
   }
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
 if ((rw == READ || rw == READA) && buffer_uptodate(bh))
                buffer_error();

 if (rw == WRITE && !buffer_uptodate(bh))
                buffer_error();

 if (rw == READ && buffer_dirty(bh))
                buffer_error();
 #endif // ndef USB_NO_VMM
 */

 /* Only clear out a write error when rewriting */

 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	if (test_set_buffer_req(bh) && rw == WRITE)
                clear_buffer_write_io_error(bh);
 #endif // ndef USB_NO_VMM
 */

 /*
  * from here on down, it's all bio -- do the initial mapping,
  * submit_bio -> generic_make_request may further map this bio around
  */

 /*
 USBMS_DBG_INFO(iprintf("[USBMS INFO] submit_bh() : FixMe : Not call bio_alloc() !! \n")) ;
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] submit_bh() : bio_alloc() .. \n"));
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	// bio = bio_alloc(GFP_NOIO, 1);
	bio = bio_alloc(1);
  #else 
 */
	bio = global_bio ;
 /*
 #endif // NO_ELEVATOR
 */

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] submit_bh() : Set Value : bio=%08XH \n",
	bio
	));
 
 #ifndef USB_NO_VMM // vitrtul memory management
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] submit_bh() : bh->b_size=%d \n",
	    bh->b_size
	    ));
 #endif // ndef USB_NO_VMM

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] submit_bh() : bh->b_blocknr=%d \n",
	bh->b_blocknr
	));
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] submit_bh() : bh->b_bdev=%08XH \n",
	bh->b_bdev
	));
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] submit_bh() : bh_offset(bh)=%08XH \n",
	bh_offset(bh)
	));

 // #ifndef USB_NO_VMM // vitrtul memory management
	/*
	#ifndef USB_NO_VMM // vitrtul memory management
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] submit_bh() : bh->b_size>>9=%d \n",
			bh->b_size>>9
			));
        bio->bi_sector = bh->b_blocknr * (bh->b_size >> 9);
	#else
	*/
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] submit_bh() : bh->b_size_sector=%d \n",
		bh->b_size_sector
		));
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] submit_bh() : bh->b_blocknr=%d \n",
		bh->b_blocknr
		));
	// bio->bi_sector = bh->b_size_sector ;
	bio->bi_sector = bh->b_blocknr ;
	/*
	#endif
	*/

	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] submit_bh() : bio->bi_sector=%d \n",
		bio->bi_sector
		));
	bio->bi_bdev = bh->b_bdev;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] submit_bh() : bio->bi_bdev=%08XH \n",
		bio->bi_bdev
		));
	/*
    #ifndef USB_NO_VMM // vitrtul memory management
                // struct page is on include/linux/mm.h
                bio->bi_io_vec[0].bv_page = bh->b_page;
     #else
	*/
		bio->bi_io_vec[0].bv_buf = bh->b_buf;
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] submit_bh() : bio->bi_io_vec[0].bv_buf=%08XH \n",
			bio->bi_io_vec[0].bv_buf
			));
	/*
    #endif
	*/
	/*
	#ifndef USB_NO_VMM // vitrtul memory management
	        bio->bi_io_vec[0].bv_len = bh->b_size;
	 #else
	*/
#if 1  /* Charles 20040524 */
		bio->bi_hardsect_size = bdev_hardsect_size(bh->b_bdev);
		bio->bi_io_vec[0].bv_len = bh->b_size_sector * bio->bi_hardsect_size;
#else
	        bio->bi_io_vec[0].bv_len = bh->b_size_sector << 9 ;
#endif
	/*
	#endif
	*/
    bio->bi_io_vec[0].bv_offset = bh_offset(bh);
    USBMS_DBG_MEM(iprintf(
                "[USBMS MEM] submit_bh() : bio->bi_io_vec[0]->bv_offset=%08XH \n",
                bio->bi_io_vec[0].bv_offset
                ));

    bio->bi_vcnt = 1;
    bio->bi_idx = 0;
	/*
	#ifndef USB_NO_VMM // vitrtul memory management
        bio->bi_size = bh->b_size;
	 #else
	*/
#if 1  /* Charles 20040524 */
	bio->bi_size = bh->b_size_sector * bio->bi_hardsect_size;
#else	
        bio->bi_size = bh->b_size_sector << 9 ;
#endif
	/*
	#endif
	*/
    USBMS_DBG_INIT(iprintf(
         "[USBMS INIT] submit_bh() : Set Value : bio->bi_size=%d \n",
         bio->bi_size )) ;

    bio->bi_end_io = end_bio_bh_io_sync;
    bio->bi_private = bh;
  /*
  #else
        bio.bi_sector = bh->b_blocknr * (bh->b_size >> 9);
        USBMS_DBG_INIT(iprintf(
                "[USBMS INIT] submit_bh() : bio.bi_sector=%d \n",
                bio.bi_sector
                ));
        bio.bi_bdev = bh->b_bdev;
        USBMS_DBG_MEM(iprintf(
                "[USBMS MEM] submit_bh() : bio.bi_bdev=%08XH \n",
                bio.bi_bdev
                ));
 // #ifndef BLOCK_IO_REDUCED
        #ifndef USB_NO_VMM // vitrtul memory management
                // struct page is on include/linux/mm.h
                bio.bi_io_vec[0].bv_page = bh->b_page;
         #else
                bio.bi_io_vec[0].bv_buf = bh->b_buf;
        #endif 
        bio.bi_io_vec[0].bv_len = bh->b_size;
        bio.bi_io_vec[0].bv_offset = bh_offset(bh);
        USBMS_DBG_MEM(iprintf(
                "[USBMS MEM] submit_bh() : bio.bi_io_vec[0].bv_offset=%08XH \n",
                bio.bi_io_vec[0].bv_offset
                ));
 // #endif // ndef BLOCK_IO_REDUCED

        bio.bi_vcnt = 1;
        bio.bi_idx = 0;
        bio.bi_size = bh->b_size;
        USBMS_DBG_INIT(iprintf(
                "[USBMS INIT] submit_bh() : Set Value : bio.bi_size=%d \n",
                bio.bi_size )) ;

        bio.bi_end_io = end_bio_bh_io_sync;
        bio.bi_private = bh;
 #endif
 */

 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] submit_bh() end with Call submit_bio() .. \n"));
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
 */
        return submit_bio(rw, bio);
 /*
  #else
        return submit_bio(rw, &bio);
 
 #endif
 */
}

/**
 * ll_rw_block: low-level access to block devices (DEPRECATED)
 * @rw: whether to %READ or %WRITE or maybe %READA (readahead)
 * @nr: number of &struct buffer_heads in the array
 * @bhs: array of pointers to &struct buffer_head
 *
 * ll_rw_block() takes an array of pointers to &struct buffer_heads,
 * and requests an I/O operation on them, either a %READ or a %WRITE.
 * The third %READA option is described in the documentation for
 * generic_make_request() which ll_rw_block() calls.
 *
 * This function drops any buffer that it cannot get a lock on (with the
 * BH_Lock state bit), any buffer that appears to be clean when doing a
 * write request, and any buffer that appears to be up-to-date when doing
 * read request.  Further it marks as clean buffers that are processed for
 * writing (the buffer cache won't assume that they are actually clean until
 * the buffer gets unlocked).
 *
 * ll_rw_block sets b_end_io to simple completion handler that marks
 * the buffer up-to-date (if approriate), unlocks the buffer and wakes
 * any waiters. 
 *
 * All of the buffers must be for the same device, and must also be a
 * multiple of the current approved size for the device.
 */
/*
#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
void ll_rw_block(int rw, int nr, struct buffer_head *bhs[])
{
 int i;

 for (i = 0; i < nr; i++) {
	struct buffer_head *bh = bhs[i];

	#ifndef USB_NO_VMM // vitrtul memory management
        if (test_set_buffer_locked(bh))
                        continue;
	#endif // ndef USB_NO_VMM

	get_bh(bh);
	if (rw == WRITE) {
		bh->b_end_io = end_buffer_write_sync;
        
        // if (test_clear_buffer_dirty(bh))
        
           {
			#ifndef USBMS_DBG_READ_ONLY
				USBMS_DBG_FLOW(iprintf("[USBMS FLOW] ll_rw_block() : submit_bh() .. \n"));
                submit_bh(WRITE, bh);
             #else
                USBMS_DBG_INFO(iprintf("[USBMS INFO] ll_rw_block() : Error : Read Only , Not Call submit_bh() !! \n"));
            #endif
			continue;
           }
       } else {
		bh->b_end_io = end_buffer_read_sync;
        
        // if (!buffer_uptodate(bh))
	    
		   {
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] ll_rw_block() : submit_bh() .. \n"));
            submit_bh(rw, bh);
            continue;
		   }
	   }
	unlock_buffer(bh);
	put_bh(bh);
   }
}
#endif // ndef NO_ELEVATOR
*/
/*
 * For a data-integrity writeout, we need to wait upon any in-progress I/O
 * and then start new I/O and then wait upon it.
 */
/* B. S. Liaw : None Call this ??
void sync_dirty_buffer(struct buffer_head *bh)
{
 
 // WARN_ON(atomic_read(&bh->b_count) < 1);
 
 if (!((atomic_read(&bh->b_count) < 1)!=0)) {
        dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
   }
        
 // lock_buffer(bh);
        
        
 // if (test_clear_buffer_dirty(bh)) 
        
        {
                get_bh(bh);
                bh->b_end_io = end_buffer_write_sync;
            #ifndef USBMS_DBG_READ_ONLY
                        USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sync_dirty_buffer() : submit_bh() .. \n"));
                        submit_bh(WRITE, bh);
                 #else
                        USBMS_DBG_INFO(iprintf("[USBMS INFO] sync_dirty_buffer() : Error : Read Only , Not Call submit_bh() !! \n"));
                #endif
                wait_on_buffer(bh);
        }
 
  else {
         
        //      unlock_buffer(bh);
         
        }
 
}
*/
/*
 * Sanity checks for try_to_free_buffers.
 */
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        static void check_ttfb_buffer(struct page *page, struct buffer_head *bh)
 #else
        static void check_ttfb_buffer(unsigned char *buf, struct buffer_head *bh)
#endif
{
 #ifndef USB_NO_VMM // vitrtul memory management
        if (!buffer_uptodate(bh) && !buffer_req(bh)) {
                if (PageUptodate(page) && page->mapping
                        && buffer_mapped(bh)    // discard_buffer 
                        && S_ISBLK(page->mapping->host->i_mode))
                {
                        buffer_error();
                }
        }
 #endif // ndef USB_NO_VMM
}
*/
/*
 * try_to_free_buffers() checks if all the buffers on this particular page
 * are unused, and releases them if so.
 *
 * Exclusion against try_to_free_buffers may be obtained by either
 * locking the page or by holding its mapping's private_lock.
 *
 * If the page is dirty but all the buffers are clean then we need to
 * be sure to mark the page clean as well.  This is because the page
 * may be against a block device, and a later reattachment of buffers
 * to a dirty page will set *all* buffers dirty.  Which would corrupt
 * filesystem data on the same device.
 *
 * The same applies to regular filesystem pages: if all the buffers are
 * clean then we set the page clean and proceed.  To do that, we require
 * total exclusion from __set_page_dirty_buffers().  That is obtained with
 * private_lock.
 *
 * try_to_free_buffers() is non-blocking.
 */
/*
static inline int buffer_busy(struct buffer_head *bh)
{
        return atomic_read(&bh->b_count) |
                (bh->b_state & ((1 << BH_Dirty) | (1 << BH_Lock)));
}
*/
/*
static int
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        drop_buffers(struct page *page, struct buffer_head **buffers_to_free)
 #else
        drop_buffers(unsigned char *buf, struct buffer_head **buffers_to_free)
#endif
{
        struct buffer_head *head = page_buffers(page);
        struct buffer_head *bh;
        int was_uptodate = 1;

        bh = head;
        do {
                check_ttfb_buffer(page, bh);
 #ifndef USB_NO_VMM // vitrtul memory management
                if (buffer_write_io_error(bh))
                        set_bit(AS_EIO, &page->mapping->flags);
 #endif // ndef USB_NO_VMM
                if (buffer_busy(bh))
                        goto failed;
                if (!buffer_uptodate(bh) && !buffer_req(bh))
                        was_uptodate = 0;
 #ifndef USB_NO_VMM // vitrtul memory management
                bh = bh->b_this_page;
 #endif // ndef USB_NO_VMM
        } while (bh != head);

        if (!was_uptodate && PageUptodate(page))
                buffer_error();

        do {
 #ifndef USB_NO_VMM // vitrtul memory management
                struct buffer_head *next = bh->b_this_page;
 #endif // ndef USB_NO_VMM

                if (!list_empty(&bh->b_assoc_buffers))
                        __remove_assoc_queue(bh);
                bh = next;
        } while (bh != head);
        *buffers_to_free = head;
        __clear_page_buffers(page);
        return 1;
failed:
        return 0;
}
*/
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
        int try_to_free_buffers(struct page *page)
 #else
        int try_to_free_buffers(unsigned char *buf)
#endif
{
 #ifndef USB_NO_VMM // vitrtul memory management
        struct address_space * const mapping = page->mapping;
 #endif // ndef USB_NO_VMM
        struct buffer_head *buffers_to_free = NULL;
        int ret = 0;
 
 // BUG_ON(!PageLocked(page));
 
 if (!PageLocked(page)) {
                dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
           }
        if (PageWriteback(page))
                return 0;

 #ifndef USB_NO_VMM // vitrtul memory management
        if (mapping == NULL) {          // swapped-in anon page 
                ret = drop_buffers(page, &buffers_to_free);
                goto out;
        }
 #endif // ndef USB_NO_VMM
 
 #ifndef NO_SPINLOCK
        spin_lock(&mapping->private_lock);
 #endif // ndef NO_SPINLOCK
 
        ret = drop_buffers(page, &buffers_to_free);
        if (ret && !PageSwapCache(page)) {
                
                // If the filesystem writes its buffers by hand (eg ext3)
                // then we can have clean buffers against a dirty page.  We
                // clean the page here; otherwise later reattachment of buffers
                // could encounter a non-uptodate page, which is unresolvable.
                // This only applies in the rare case where try_to_free_buffers
                // succeeds but the page is not freed.
                 
                clear_page_dirty(page);
        }
 
 #ifndef NO_SPINLOCK
        spin_unlock(&mapping->private_lock);
 #endif // ndef NO_SPINLOCK
 
out:
        if (buffers_to_free) {
                struct buffer_head *bh = buffers_to_free;

                do {
 #ifndef USB_NO_VMM // vitrtul memory management
                        struct buffer_head *next = bh->b_this_page;
 #endif // ndef USB_NO_VMM
                        USBMS_DBG_FLOW(iprintf(
                                "[USBMS FLOW] try_to_free_buffers() : free_buffer_head() .. \n"
                                ));
                        free_buffer_head(bh);
                        bh = next;
                } while (bh != buffers_to_free);
        }
        return ret;
}
EXPORT_SYMBOL(try_to_free_buffers);
*/
/*
#ifndef USB_NO_VMM // vitrtul memory management
        // struct page is on include/linux/mm.h
int block_sync_page(struct page *page)
//#else
//int block_sync_page(unsigned char *buf)
/#endif
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] block_sync_page() : blk_run_queues() .. \n"));
 blk_run_queues();
 return 0;
}
#endif // ndef USB_NO_VMM
*/
/*
 * There are no bdflush tunables left.  But distributions are
 * still running obsolete flush daemons, so we terminate them here.
 *
 * Use of bdflush() is deprecated and will be removed in a future kernel.
 * The `pdflush' kernel threads fully replace bdflush daemons and this call.
 */
/*
// asmlinkage 
long sys_bdflush(int func, long data)
{
        static int msg_count;

        if (!capable(CAP_SYS_ADMIN))
                return -EPERM;

        if (msg_count < 5) {
                msg_count++;
                
                //printk(KERN_INFO
                //      "warning: process `%s' used the obsolete bdflush"
                //      " system call\n", current->comm);
                
 #ifndef _ETCPIP
                dprintf("[USBMS] warning: process `%s' used the obsolete bdflush"
                        " system call\n", current->comm);
 #endif // ndef _ETCPIP
                
                // printk(KERN_INFO "Fix your initscripts?\n");
                
                dprintf("[USBMS] Fix your initscripts?\n");
        }

        if (func == 1)
                do_exit(0);
        return 0;
}
*/

/*
 * Buffer-head allocation
 */
/* kmem_cache_t is on include/linux/slab.h
static kmem_cache_t *bh_cachep;
*/

/*
 * Once the number of bh's in the machine exceeds this level, we start
 * stripping them in writeback.
 */
/*
static int max_buffer_heads;

int buffer_heads_over_limit;
*/
/*
#ifndef _ETCPIP
struct bh_accounting {
        int nr;                 //  Number of live bh's 
        int ratelimit;          // Limit cacheline bouncing
};
#endif // ndef _ETCPIP
*/

/*
#ifndef _ETCPIP
static DEFINE_PER_CPU(struct bh_accounting, bh_accounting) = {0, 0};
#endif // ndef _ETCPIP
*/

/*
#ifndef _ETCPIP
static void recalc_bh_state(void)
{
        int i;
        int tot = 0;

        if (__get_cpu_var(bh_accounting).ratelimit++ < 4096)
                return;
        __get_cpu_var(bh_accounting).ratelimit = 0;
        for (i = 0; i < NR_CPUS; i++) {
                if (cpu_online(i))
                        tot += per_cpu(bh_accounting, i).nr;
        }
        buffer_heads_over_limit = (tot > max_buffer_heads);
}
#endif // ndef _ETCPIP
*/
/*
#ifndef _ETCPIP
        struct buffer_head *alloc_buffer_head(int gfp_flags)
 #else
*/
        struct buffer_head *alloc_buffer_head(void)
/*
#endif
*/
{
 /* kmem_cache_t is on include/linux/slab.h
 struct buffer_head *ret = kmem_cache_alloc(bh_cachep, gfp_flags);
 */
 struct buffer_head *ret ;


 USBMS_DBG_MEM(iprintf(
        "[USBMS MEM] alloc_buffer_head() : ret = UsbAllocMemC() , size=%d .. \n",
        sizeof(struct buffer_head)
        ));
 ret = (struct buffer_head *) UsbAllocMemC(USB_MEM_C_S
                                #ifdef USB_DBG_MEM_C_ON
                                        , sizeof(struct buffer_head) // size : 36
                        #endif
                        ) ;
 USBMS_DBG_MEM(iprintf(
        "[USBMS MEM] alloc_buffer_head() : ret=%08XH \n",
        ret
        ));

 if (ret) {
        /*
        #ifndef _ETCPIP
                preempt_disable();
                __get_cpu_var(bh_accounting).nr++;
        #endif // ndef _ETCPIP
        */
        /*
        #ifndef _ETCPIP
                recalc_bh_state();
                preempt_enable();
        #endif // ndef _ETCPIP
        */
        }
 return ret;
}
/*
EXPORT_SYMBOL(alloc_buffer_head);
*/
void free_buffer_head(struct buffer_head *bh)
{
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	// BUG_ON(!list_empty(&bh->b_assoc_buffers));
	if (!list_empty(&bh->b_assoc_buffers) {
                dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
           }
 #endif // ndef USB_NO_VMM
 */
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	/* kmem_cache_t is on include/linux/slab.h
	kmem_cache_free(bh_cachep, bh);
	*/
	USBMS_DBG_MEM(iprintf(
                "[USBMS MEM] free_buffer_head() : UsbFreeMemC(bh) .. \n"
                ));
	UsbFreeMemC(USB_MEM_C_S,(unsigned char *) bh
                #if defined(USB_DBG_MEM_C_ON)
                        , sizeof(struct buffer_head) // size : 36
                #endif
                );
 #endif // ndef NO_ELEVATOR
 /*
 #ifndef _ETCPIP
        preempt_disable();
        __get_cpu_var(bh_accounting).nr--;
 #endif // ndef _ETCPIP
 */
 /*
 #ifndef _ETCPIP
        recalc_bh_state();
        preempt_enable();
 #endif // ndef _ETCPIP
 */
}
/*
EXPORT_SYMBOL(free_buffer_head);
*/
/*
static void
init_buffer_head(void *data, kmem_cache_t *cachep, unsigned long flags)
{
 if ((flags & (SLAB_CTOR_VERIFY|SLAB_CTOR_CONSTRUCTOR)) ==
                            SLAB_CTOR_CONSTRUCTOR) {
                struct buffer_head * bh = (struct buffer_head *)data;

                memset(bh, 0, sizeof(*bh));
                INIT_LIST_HEAD(&bh->b_assoc_buffers);
        }
}
*/
/*      
#ifndef _ETCPIP
static void buffer_init_cpu(int cpu)
{
        struct bh_accounting *bha = &per_cpu(bh_accounting, cpu);
        struct bh_lru *bhl = &per_cpu(bh_lrus, cpu);

        bha->nr = 0;
        bha->ratelimit = 0;
        memset(bhl, 0, sizeof(*bhl));
}
#endif // ndef _ETCPIP
*/
/*      
#ifndef _ETCPIP
static int __devinit buffer_cpu_notify(struct notifier_block *self, 
                                unsigned long action, void *hcpu)
{
        long cpu = (long)hcpu;
        switch(action) {
        case CPU_UP_PREPARE:
                buffer_init_cpu(cpu);
                break;
        default:
                break;
        }
        return NOTIFY_OK;
}

static struct notifier_block __devinitdata buffer_nb = {
        .notifier_call  = buffer_cpu_notify,
};
#endif // ndef _ETCPIP
*/
/*
void __init buffer_init(void)
*/
void buffer_init(void)
{
 int i;
 /*
 int nrpages;
 */
 /* kmem_cache_t is on include/linux/slab.h
 bh_cachep = kmem_cache_create("buffer_head",
                        sizeof(struct buffer_head), 0,
                        0, init_buffer_head, NULL);
 */

 #ifndef NO_WAITQUEUE
        for (i = 0; i < ARRAY_SIZE(bh_wait_queue_heads); i++)
                init_waitqueue_head(&bh_wait_queue_heads[i].wqh);
 #endif // ndef NO_WAITQUEUE

        /*
         * Limit the bh occupancy to 10% of ZONE_NORMAL
         */
 /*
 #ifndef _ETCPIP
        nrpages = (nr_free_buffer_pages() * 10) / 100;
        max_buffer_heads = nrpages * (PAGE_SIZE / sizeof(struct buffer_head));
        buffer_cpu_notify(&buffer_nb, (unsigned long)CPU_UP_PREPARE,
                                (void *)(long)smp_processor_id());
        register_cpu_notifier(&buffer_nb);
 #endif // ndef _ETCPIP
 */
 
 // #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] buffer_init() : alloc_buffer_head() .. \n"));
	global_bh = alloc_buffer_head();
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] buffer_init() : bio_alloc() .. \n"));
	global_bio = bio_alloc(1);
 // #endif // NO_ELEVATOR
 
}

/*
EXPORT_SYMBOL(__bforget);
EXPORT_SYMBOL(__brelse);
EXPORT_SYMBOL(__wait_on_buffer);
EXPORT_SYMBOL(block_commit_write);
EXPORT_SYMBOL(block_prepare_write);
EXPORT_SYMBOL(block_read_full_page);
EXPORT_SYMBOL(block_sync_page);
EXPORT_SYMBOL(block_truncate_page);
EXPORT_SYMBOL(block_write_full_page);
EXPORT_SYMBOL(buffer_insert_list);
EXPORT_SYMBOL(cont_prepare_write);
EXPORT_SYMBOL(end_buffer_async_write);
EXPORT_SYMBOL(end_buffer_read_sync);
EXPORT_SYMBOL(end_buffer_write_sync);
EXPORT_SYMBOL(file_fsync);
EXPORT_SYMBOL(fsync_bdev);
EXPORT_SYMBOL(fsync_buffers_list);
EXPORT_SYMBOL(generic_block_bmap);
EXPORT_SYMBOL(generic_commit_write);
EXPORT_SYMBOL(generic_cont_expand);
EXPORT_SYMBOL(init_buffer);
EXPORT_SYMBOL(invalidate_bdev);
EXPORT_SYMBOL(ll_rw_block);
EXPORT_SYMBOL(mark_buffer_dirty);
EXPORT_SYMBOL(submit_bh);
EXPORT_SYMBOL(sync_dirty_buffer);
EXPORT_SYMBOL(unlock_buffer);
*/
