/*
 *  fs/partitions/check.c
 *
 *  Code extracted from drivers/block/genhd.c
 *  Copyright (C) 1991-1998  Linus Torvalds
 *  Re-organised Feb 1998 Russell King
 *
 *  We now have independent partition support from the
 *  block drivers, which allows all the partition code to
 *  be grouped in one location, and it to be mostly self
 *  contained.
 *
 *  Added needed MAJORS for new pairs, {hdi,hdj}, {hdk,hdl}
 */

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
	// for deciding whether to compile
	//  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP

/*
#include <linux/init.h>
#include <linux/module.h>
#include <linux/fs.h>
*/
#include "PoFile.h" // include/linux/file.h
#include "radix-tree.h" // include/linux/radix-tree.h
#include "PoFs.h" // <linux/fs.h>
/*
#include <linux/kmod.h>
#include <linux/ctype.h>
#include <linux/devfs_fs_kernel.h>
*/
#include "check.h"
/*
#include "devfs.h"

#include "acorn.h"
#include "amiga.h"
#include "atari.h"
#include "ldm.h"
#include "mac.h"
*/
#ifdef CONFIG_MSDOS_PARTITION
	#include "msdos.h"
#endif // CONFIG_MSDOS_PARTITION
/*
#include "nec98.h"
#include "osf.h"
#include "sgi.h"
#include "sun.h"
#include "ibm.h"
#include "ultrix.h"
*/

#ifdef CONFIG_EFI_PARTITION
	#include "efi.h"
#endif // CONFIG_EFI_PARTITION


#ifdef CONFIG_BLK_DEV_MD
	extern void md_autodetect_dev(dev_t dev);
#endif
/*
#ifdef USBMS_ONLY_ONE_DISK
	// 2003/12/17 B. S. Liaw add these ..
	// struct inode *disk_global_inode = NULL ;
	struct block_device *disk_global_bdev[USBMS_MAX_MINORS+1] ;
#endif // USBMS_ONLY_ONE_DISK
*/
#ifdef USBMS_ONLY_ONE_DISK
	unsigned char *disk_global_Sector_buf = NULL ;
#endif // USBMS_ONLY_ONE_DISK

int warn_no_part = 1; /*This is ugly: should make genhd removable media aware*/

static int (*check_part[])(struct inode *inode,struct parsed_partitions *, struct block_device *) = {
	/*
	 * Probe partition formats with tables at disk address 0
	 * that also have an ADFS boot block at 0xdc0.
	 */
 #ifdef CONFIG_ACORN_PARTITION_ICS
	adfspart_check_ICS,
 #endif
 #ifdef CONFIG_ACORN_PARTITION_POWERTEC
	adfspart_check_POWERTEC,
 #endif
 #ifdef CONFIG_ACORN_PARTITION_EESOX
	adfspart_check_EESOX,
 #endif

	/*
	 * Now move on to formats that only have partition info at
	 * disk address 0xdc0.  Since these may also have stale
	 * PC/BIOS partition tables, they need to come before
	 * the msdos entry.
	 */
 #ifdef CONFIG_ACORN_PARTITION_CUMANA
	adfspart_check_CUMANA,
 #endif
 #ifdef CONFIG_ACORN_PARTITION_ADFS
	adfspart_check_ADFS,
 #endif

 #ifdef CONFIG_EFI_PARTITION
	efi_partition,		/* this must come before msdos */
 #endif
 #ifdef CONFIG_LDM_PARTITION
	ldm_partition,		/* this must come before msdos */
 #endif
 #ifdef CONFIG_NEC98_PARTITION
	nec98_partition,	/* must be come before `msdos_partition' */
 #endif
 #ifdef CONFIG_MSDOS_PARTITION
	msdos_partition,
 #endif
 #ifdef CONFIG_OSF_PARTITION
	osf_partition,
 #endif
 #ifdef CONFIG_SUN_PARTITION
	sun_partition,
 #endif
 #ifdef CONFIG_AMIGA_PARTITION
	amiga_partition,
 #endif
 #ifdef CONFIG_ATARI_PARTITION
	atari_partition,
 #endif
 #ifdef CONFIG_MAC_PARTITION
	mac_partition,
 #endif
 #ifdef CONFIG_SGI_PARTITION
	sgi_partition,
 #endif
 #ifdef CONFIG_ULTRIX_PARTITION
	ultrix_partition,
 #endif
 #ifdef CONFIG_IBM_PARTITION
	ibm_partition,
 #endif
	NULL
};
 
/*
 * disk_name() is used by partition check code and the genhd driver.
 * It formats the devicename of the indicated disk into
 * the supplied buffer (of size at least 32), and returns
 * a pointer to that same buffer (for convenience).
 */

#ifdef USB_DBGSTRING_ON
char *disk_name(struct gendisk *hd, int part, char *buf)
{
	if (!part)
		snprintf(buf, BDEVNAME_SIZE, "%s", hd->disk_name);
	else if (isdigit(hd->disk_name[strlen(hd->disk_name)-1]))
		snprintf(buf, BDEVNAME_SIZE, "%sp%d", hd->disk_name, part);
	else
		snprintf(buf, BDEVNAME_SIZE, "%s%d", hd->disk_name, part);

	return buf;
}
#endif // USB_DBGSTRING_ON

#ifndef USB_DBGSTRING_ON
	// 2004/1/14 B. S. Liaw
	const char null_str[] = "" ;
#endif // ndef USB_DBGSTRING_ON

const char *bdevname(struct block_device *bdev, char *buf)
{
 #ifndef USBMS_ONLY_ONE_DISK
	int part = MINOR(bdev->bd_dev) - bdev->bd_disk->first_minor;
  #else
	int part = MINOR(bdev->bd_dev) - 0 ;
 #endif // USBMS_ONLY_ONE_DISK

 USBMS_DBG_INFO(iprintf(
	"[USBMS INFO] bdevname() : MINOR(bdev->bd_dev)=%d \n",
	MINOR(bdev->bd_dev) ));
 #ifndef USBMS_ONLY_ONE_DISK
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] bdevname() : bdev->bd_disk->first_minor=%d \n",
		bdev->bd_disk->first_minor));
 #endif // ndef USBMS_ONLY_ONE_DISK
 USBMS_DBG_INFO(iprintf("[USBMS INFO] bdevname() : part=%d \n", part));
 #ifdef USB_DBGSTRING_ON
	return disk_name(bdev->bd_disk, part, buf);
  #else
	// return NULL ;

	// 2004/1/14 B. S. Liaw
	return null_str ; ;
 #endif // ndef USB_DBGSTRING_ON
}
/*
EXPORT_SYMBOL(bdevname);
*/
/*
 * NOTE: this cannot be called from interrupt context.
 *
 * But in interrupt context you should really have a struct
 * block_device anyway and use bdevname() above.
 */
/* B. S. Liaw : None Call This ??
const char *__bdevname(dev_t dev, char *buffer)
{
	struct gendisk *disk;
	int part;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __bdevname() : get_gendisk() .. \n"));
	disk = get_gendisk(dev, &part);
	if (disk) {
		buffer = disk_name(disk, part, buffer);
		put_disk(disk);
	} else {
		snprintf(buffer, BDEVNAME_SIZE, "unknown-block(%u,%u)",
				MAJOR(dev), MINOR(dev));
	}

	return buffer;
}
*/
/*
EXPORT_SYMBOL(__bdevname);
*/
static struct parsed_partitions *
check_partition(struct inode *inode,struct gendisk *hd, struct block_device *bdev)
{
 struct parsed_partitions *state;
 int i, res;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] check_partition() Begin .. \n"));
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] check_partition() : state = UsbAllocMemC() size=%d .. \n",
	sizeof(struct parsed_partitions)
	));
 /*
 state = kmalloc(sizeof(struct parsed_partitions), GFP_KERNEL);
 */
 state = (struct parsed_partitions *) UsbAllocMemC(USB_MEM_C_M
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct parsed_partitions) // 256 : size : 3112
		        #endif								   // 16  : size : 232
           		) ;
 if (!state)
	return NULL;

 #ifdef CONFIG_DEVFS_FS
	if (hd->devfs_name[0] != '\0') {
		/*
		printk(KERN_INFO " /dev/%s:", hd->devfs_name);
		*/
		dprintf("[USBMS] /dev/%s:", hd->devfs_name);
		sprintf(state->name, "p");
	}
	else
 #endif
	 {
	#ifdef USB_DBGSTRING_ON
		disk_name(hd, 0, state->name);
		/*
		printk(KERN_INFO " %s:", state->name);
		*/
		dprintf("[USBMS] INFO %s:", state->name);
		if (isdigit(state->name[strlen(state->name)-1]))
			sprintf(state->name, "p");
	#endif // USB_DBGSTRING_ON
	}
	state->limit = hd->minors;
	i = res = 0;
	while (!res && check_part[i]) {
		USBMS_DBG_INIT(iprintf("[USBMS INIT] check_partition() : i=%d \n",i));
		memset(&state->parts, 0, sizeof(state->parts));
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] check_partition() : check_part() .. \n"));
		res = check_part[i++](inode,state, bdev);
	}
 if (res > 0)
		return state;

 if (!res) {
	/*
	printk(" unknown partition table\n");
	*/
	dprintf("[USBMS] unknown partition table !! \n");
   }
  else if (warn_no_part) {
	/*
	printk(" unable to read partition table\n");
	*/
	dprintf("[USBMS] Error : unable to read partition table !!\n");
   }

 USBMS_DBG_MEM(iprintf("[USBMS MEM] check_partition() : UsbFreeMemC(state) .. \n"));
 /*
 kfree(state);
 */
 UsbFreeMemC(USB_MEM_C_M,(unsigned char *) state
		#if defined(USB_DBG_MEM_C_ON)
			,sizeof(struct parsed_partitions)  // 256 : size : 3112
		#endif								   // 16  : size : 232
		);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] check_partition() End with Failed !! \n"));
 return NULL;
}

/*
 * sysfs bindings for partitions
 */

#ifndef NO_SYS_FS
   // struct attribute on sysfs.h
   struct part_attribute {
		struct attribute attr;
		ssize_t (*show)(struct hd_struct *,char *);
   };

   static ssize_t 
   part_attr_show(struct kobject * kobj, struct attribute * attr, char * page)
   {
	struct hd_struct * p = container_of(kobj,struct hd_struct,kobj);
	struct part_attribute * part_attr = container_of(attr,struct part_attribute,attr);
	ssize_t ret = 0;
	if (part_attr->show)
		ret = part_attr->show(p,page);
	return ret;
   }

   static struct sysfs_ops part_sysfs_ops = {
	.show	=	part_attr_show,
   };
#endif // ndef NO_SYS_FS
/*
#ifndef NO_SYS_FS
   static ssize_t part_dev_read(struct hd_struct * p, char *page)
   {
	struct gendisk *disk = container_of(p->kobj.parent,struct gendisk,kobj);
	dev_t dev = MKDEV(disk->major, disk->first_minor + p->partno); 
	return print_dev_t(page, dev);
   }
   static ssize_t part_start_read(struct hd_struct * p, char *page)
   {
	return sprintf(page, "%llu\n",(unsigned long long)p->start_sect);
   }
   static ssize_t part_size_read(struct hd_struct * p, char *page)
   {
	return sprintf(page, "%llu\n",(unsigned long long)p->nr_sects);
   }
   static ssize_t part_stat_read(struct hd_struct * p, char *page)
   {
	return sprintf(page, "%8u %8llu %8u %8llu\n",
		       p->reads, (unsigned long long)p->read_sectors,
		       p->writes, (unsigned long long)p->write_sectors);
   }

   static struct part_attribute part_attr_dev = {
	.attr = {.name = "dev", .mode = S_IRUGO },
	.show	= part_dev_read
   };
   static struct part_attribute part_attr_start = {
	.attr = {.name = "start", .mode = S_IRUGO },
	.show	= part_start_read
   };
   static struct part_attribute part_attr_size = {
	.attr = {.name = "size", .mode = S_IRUGO },
	.show	= part_size_read
   };
   static struct part_attribute part_attr_stat = {
	.attr = {.name = "stat", .mode = S_IRUGO },
	.show	= part_stat_read
   };

   // struct attribute on sysfs.h
   static struct attribute * default_attrs[] = {
		&part_attr_dev.attr,
		&part_attr_start.attr,
		&part_attr_size.attr,
		&part_attr_stat.attr,
		NULL,
	   };
#endif // ndef NO_SYS_FS
*/
extern struct subsystem block_subsys;
/*
#ifndef NO_SYS_FS
   static void part_release(struct kobject *kobj)
   {
	struct hd_struct * p = container_of(kobj,struct hd_struct,kobj);
	
	// kfree(p);
	
	UsbFreeMemC(USB_MEM_C_S,(unsigned char *) p
		#if defined(USB_DBG_MEM_C_ON)
			,sizeof(struct hd_struct)
		#endif
		);
   }

   struct kobj_type ktype_part = {
	.release	= part_release,
   // struct attribute on sysfs.h
	.default_attrs	= default_attrs,
	.sysfs_ops	= &part_sysfs_ops,
   };
#endif // ndef NO_SYS_FS
*/
void delete_partition(struct gendisk *disk, int part)
{
 struct hd_struct *p = disk->part[part-1];


 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] delete_partition() Begin .. \n" )) ;
 if (!p)
		return;
 if (!p->nr_sects)
		return;
 /*
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] delete_partition() : UsbFreeMemC(disk->part[part-1]) , size=%d .. \n" ,
	sizeof(struct hd_struct)
	));
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) disk->part[part-1]
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct hd_struct)   // size : 84
		        #endif
				);
 */
 disk->part[part-1] = NULL;
 p->start_sect = 0;
 p->nr_sects = 0;
 p->reads = p->writes = p->read_sectors = p->write_sectors = 0;
 #ifdef CONFIG_USB_DEVICEFS
	devfs_remove("%s/part%d", disk->devfs_name, part);
 #endif // CONFIG_USB_DEVICEFS
 #ifndef _ETCPIP
	kobject_unregister(&p->kobj);
 #endif // ndef _ETCPIP

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] delete_partition() : UsbFreeMemC(p]) , size=%d .. \n" ,
	sizeof(struct hd_struct)
	));
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) p
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct hd_struct)   // size : 84
		        #endif
				);
}

void add_partition(struct gendisk *disk, int part, sector_t start, sector_t len)
{
 struct hd_struct *p;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] add_partition() Begin .. \n"));
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] add_partition() : p = UsbAllocMemC() , size=%d .. \n" ,
	sizeof(struct hd_struct)
	));

 /*
 p = kmalloc(sizeof(*p), GFP_KERNEL);
 */
 p = (struct hd_struct *) UsbAllocMemC(USB_MEM_C_S
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct hd_struct)  // size : 84
		        #endif
           		) ;
 if (!p)
		return;
 /*	
 memset(p, 0, sizeof(*p));
 */
 p->start_sect = start;
 p->nr_sects = len;
 p->partno = part;
 /*
 #ifndef USBMS_ONLY_ONE_DISK
	USBMS_DBG_INFO(iprintf("[USBMS INFO] add_partition() : disk->major=%d \n",
		disk->major
		));
	USBMS_DBG_INFO(iprintf("[USBMS INFO] add_partition() : disk->first_minor=%d \n",
		disk->first_minor
		));
 #endif // ndef USBMS_ONLY_ONE_DISK
 */
 USBMS_DBG_PARTITION(iprintf("[USBMS PARTITION] add_partition() : part=%d \n",
	part
	));
 /*
 #ifndef USBMS_ONLY_ONE_DISK
	USBMS_DBG_INIT(iprintf("[USBMS INIT] add_partition() : MKDEV()=%04XH \n",
		MKDEV(disk->major, disk->first_minor + part)
		));
  #else
 */
	USBMS_DBG_PARTITION(iprintf("[USBMS PARTITION] add_partition() : MKDEV()=%04XH \n",
		MKDEV(0, 0 + part)
		));
 /*
 #endif // USBMS_ONLY_ONE_DISK
 */

 /*
 #ifdef CONFIG_USB_DEVICEFS
	devfs_mk_bdev(MKDEV(disk->major, disk->first_minor + part),
			S_IFBLK|S_IRUSR|S_IWUSR,
			"%s/part%d", disk->devfs_name, part);
 #endif // CONFIG_USB_DEVICEFS
 */
 #ifdef USB_DBGSTRING_ON
	snprintf(p->kobj.name,KOBJ_NAME_LEN,"%s%d",disk->kobj.name,part);
 #endif // USB_DBGSTRING_ON
 p->kobj.parent = &disk->kobj;
 /*
 #ifndef NO_SYS_FS
	p->kobj.ktype = &ktype_part;
 #endif // ndef NO_SYS_FS
 */
 kobject_register(&p->kobj);
 disk->part[part-1] = p;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] add_partition() End. \n"));
}

static void disk_sysfs_symlinks(struct gendisk *disk)
{
	struct device *target = get_device(disk->driverfs_dev);
 #ifndef NO_SYS_FS
	if (target) {
		sysfs_create_link(&disk->kobj,&target->kobj,"device");
		sysfs_create_link(&target->kobj,&disk->kobj,"block");
	}
 #endif // ndef NO_SYS_FS
}

extern int validDiskCnt;
extern int validPartitionCnt;
extern unsigned long validPartitionID[20];

/* Not exported, helper to add_disk(). */
int register_disk(struct gendisk *disk)
{
 struct parsed_partitions *state;
 struct block_device *bdev;
 char *s;
 int j;
 int err;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk() Begin .. \n"));

 #ifdef USB_DBGSTRING_ON
	strlcpy(disk->kobj.name,disk->disk_name,KOBJ_NAME_LEN);
	/* ewww... some of these buggers have / in name... */
	s = strchr(disk->kobj.name, '/');
	if (s)
		*s = '!';
 #endif // USB_DBGSTRING_ON

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk(1) : kobject_add() .. \n"));
 if ((err = kobject_add(&disk->kobj))) {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] register_disk() : kobject_add() Error !! \n"));
	dprintf("[USBMS] Error : Internal Failed !! \n") ;
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk(2) End with Failed !! \n"));
	return 0 ;
   }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk(3) : disk_sysfs_symlinks() .. \n"));
 disk_sysfs_symlinks(disk);

	/* No minors to use for partitions */
 #ifdef CONFIG_USB_DEVICEFS
	if (disk->minors == 1) {
		if (disk->devfs_name[0] != '\0')
			devfs_add_disk(disk);
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk(4) End with Failed !! \n"));
		return;
	   }

	/* always add handle for the whole disk */
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk(5) : devfs_add_partitioned() .. \n"));
	devfs_add_partitioned(disk);
 #endif // CONFIG_USB_DEVICEFS

	/* No such device (e.g., media were just removed) */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk(6) : get_capacity() .. \n"));
 if(!get_capacity(disk)) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk(7) : End with Failed !! \n"));
	return;
   }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk(8) : bdget_disk() .. \n"));
 bdev = bdget_disk(disk, 0);
 USBMS_DBG_MEM(iprintf("[USBMS MEM] register_disk() : bdev=%08XH \n",bdev));

 // #ifndef INODE_REDUCED  
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk(9) : blkdev_get() .. \n"));
 // blkdev_get() on fs/block_dev.c
 /*
 #ifndef _ETCPIP
 */
	if(blkdev_get(bdev, FMODE_READ, 0, BDEV_RAW) < 0)
 /*
  #else
	if(blkdev_get(bdev, FMODE_READ, 0, BDEV_RAW,disk) < 0)
 #endif
 */
	   {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk(10) : End with Failed !! \n"));
		return;
	   }
 // #endif // ndef INODE_REDUCED  

 USBMS_DBG_PARTITION(iprintf("[USBMS PARTITION] register_disk() : bdev->bd_inode->i_size=%d \n",
	bdev->bd_inode->i_size));
 USBMS_DBG_PARTITION(iprintf("[USBMS PARTITION] register_disk() : Check Partition .. \n"));
 dprintf("[USBMS] Check Partition .. \n");
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk(11) : check_partition() .. \n"));
 state = check_partition(bdev->bd_inode,disk, bdev);
 if (state) {
	USBMS_DBG_PARTITION(iprintf(
		"[USBMS PARTITION] register_disk() : Max (limit) Partitions = %d \n",
		state->limit));
	for(j = 1; j < state->limit; j++) {
		sector_t size = state->parts[j].size;
		sector_t from = state->parts[j].from;
		if (!size)
				continue;
		if (validPartitionCnt < 20) {
			validPartitionID[validPartitionCnt++] = j - 1;
		   }
		USBMS_DBG_PARTITION(iprintf(
			"[USBMS PARTITION] register_disk() : Add Partition (Logical Disk) .. \n"
			));
		dprintf("[USBMS] Add Logical Disk %d.. \n",j);
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk(12) : add_partition() .. \n"));
		add_partition(disk, j, from, size);
		#ifdef CONFIG_BLK_DEV_MD
			if (!state->parts[j].flags)
				continue;
			md_autodetect_dev(bdev->bd_dev+j);
		#endif
	   }
	//validDiskCnt = 1;

	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] register_disk() : UsbFreeMemC(state) .. \n"
		));
	/*
	kfree(state);
	*/
	UsbFreeMemC(USB_MEM_C_M,(unsigned char *) state
		#if defined(USB_DBG_MEM_C_ON)
			,sizeof(struct parsed_partitions)  // 256 : size : 3112
		#endif								   // 16  : size : 232
		);
   }
  else {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] register_disk() : check_partition() Error !! \n"));
	dprintf("[USBMS] Error : Check Partition Failed !! \n") ;

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk(13) End with Failed !! \n"));
	return -EIO ;
   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk(14) : blkdev_put() .. \n"));
 blkdev_put(bdev, BDEV_RAW);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] register_disk() End. \n"));
 return 1 ;
}

/*
#ifndef USBMS_ONLY_ONE_DISK

#ifndef USB_NO_VMM // vitrtul memory management
	int rescan_partitions(struct gendisk *disk, struct block_device *bdev)
 #else
int rescan_partitions(struct inode *inode,struct gendisk *disk, struct block_device *bdev)

#endif

{
 struct parsed_partitions *state;
 int p, res;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] rescan_partitions() Begin .. \n"));
 if (bdev->bd_part_count)
		return -EBUSY;

 #ifndef USBMS_ONLY_ONE_DISK
	res = invalidate_partition(disk, 0);
	if (res)
		return res;
 #endif // ndef USBMS_ONLY_ONE_DISK

 bdev->bd_invalidated = 0;
 for (p = 1; p < disk->minors; p++)
		delete_partition(disk, p);
 if (disk->fops->revalidate_disk)
   {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] rescan_partitions() : disk->fops->revalidate_disk() .. \n"));
	disk->fops->revalidate_disk(disk);
   }
	if (!get_capacity(disk) || !(state = check_partition(inode,disk, bdev)))
		return res;
	for (p = 1; p < state->limit; p++) {
		sector_t size = state->parts[p].size;
		sector_t from = state->parts[p].from;
		if (!size)
			continue;
		USBMS_DBG_INIT(iprintf("[USBMS INIT] rescan_partitions() : Add Partition .. \n"));
		add_partition(disk, p, from, size);
 #ifdef CONFIG_BLK_DEV_MD
		if (state->parts[p].flags)
			md_autodetect_dev(bdev->bd_dev+p);
 #endif
	}

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] rescan_partitions() : UsbFreeMemC(state) .. \n"
	));
 // kfree(state);
 
 UsbFreeMemC(USB_MEM_C_M,(unsigned char *) state
	#if defined(USB_DBG_MEM_C_ON)
		,sizeof(struct parsed_partitions)  // 256 : size : 3112
	#endif								   // 16  : size : 232
	);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] rescan_partitions() End. \n"));
 return res;
}
#endif // ndef USBMS_ONLY_ONE_DISK
*/
#if _USB2 > 0	// reduce function call layer
#include "buffer_head.h"
#endif	// _USB2

int read_dev_sector(struct inode *inode, struct block_device *bdev,
					sector_t n,
					unsigned int num_sectors ,
					unsigned char *Sector_buf) {
#if _USB2 > 0	// reduce function call layer
	struct buffer_head *head;
#else	// _USB2
#ifndef NO_INODE
	struct address_space *mapping = bdev->bd_inode->i_mapping;
#endif // ndef NO_INODE
#endif	// _USB2
	int i, retval;

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] read_dev_sector() Begin .. \n"));
	led_usb_flash(1);
	for (i = 0 ; i < 2 ; i++) {
#if _USB2 > 0	// reduce function call layer
//		retval = block_read_full_page(inode, n, num_sectors, Sector_buf, blkdev_get_block);
		{
			head = create_empty_buffers(Sector_buf);
			head->b_size_sector = num_sectors;
			if (blkdev_get_block(inode, n, head, 0) > 0 ) {
				mark_buffer_async_read(head);
				if (!submit_bh(READ, head))
					retval = -EIO;
				else
					retval = 1;
			} else
				retval = -EIO;
		}
#else	// _USB2
		retval = mapping->a_ops->readpage(inode, n, num_sectors, Sector_buf);
#endif	// _USB2
		if (retval > 0) {
			led_usb_flash(0);
			return 1;
		}
	}
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] read_dev_sector() Failed! \n"));

	led_usb_flash(0);
	return retval;
}

// ====================================================================
// 2003/12/23 B. S. Liaw Add this function ..
// unsigned char *write_dev_sector(struct inode *inode, struct block_device *bdev,
int write_dev_sector(struct inode *inode, struct block_device *bdev,
		 sector_t n,
		unsigned int num_sectors ,
		unsigned char *Sector_buf) {
#if _USB2 > 0	// reduce function call layer
	loff_t i_size = i_size_read(inode);
#else	// _USB2
#ifndef NO_INODE
	struct address_space *mapping = bdev->bd_inode->i_mapping;
#endif // ndef NO_INODE
#endif	// _USB2
	int i, retval;

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] write_dev_sector() Begin .. \n"));
	led_usb_flash(1);
	for (i = 0 ; i < 2 ; i++) {
#if _USB2 > 0	// reduce function call layer
		// retval = block_write_full_page(inode, n, num_sectors, Sector_buf);
		retval = __block_write_full_page(inode, n, num_sectors, Sector_buf, blkdev_get_block);
#else	// _USB2
		retval = mapping->a_ops->writepage(inode, n, num_sectors, Sector_buf);
#endif	// _USB2
		if (retval > 0) {
			led_usb_flash(0);
			return 1;
		}
	}
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] write_dev_sector(4) End. \n"));

	led_usb_flash(0);
	return retval;
}

// ====================================================================

void del_gendisk(struct gendisk *disk)
{
 int p;

 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] del_gendisk() Begin .. \n"));
 /* invalidate stuff */
 for(p = disk->minors - 1; p > 0; p--) {
	#ifndef USBMS_ONLY_ONE_DISK
		invalidate_partition(disk, p);
	#endif // ndef USBMS_ONLY_ONE_DISK
	USBMS_DBG_PLUG(iprintf("[USBMS PLUG] del_gendisk() : delete_partition() .. \n"));
	delete_partition(disk, p);
   }

 #ifndef USBMS_ONLY_ONE_DISK
	invalidate_partition(disk, 0);
 #endif // ndef USBMS_ONLY_ONE_DISK
 disk->capacity = 0;
 disk->flags &= ~GENHD_FL_UP;
 unlink_gendisk(disk);
 disk_stat_set_all(disk, 0);
 disk->stamp = disk->stamp_idle = 0;

 #ifndef NO_SYS_FS
	devfs_remove_disk(disk);
 #endif // ndef NO_SYS_FS

 if(disk->driverfs_dev) {
	#ifndef NO_SYS_FS
		sysfs_remove_link(&disk->kobj, "device");
		sysfs_remove_link(&disk->driverfs_dev->kobj, "block");
	#endif // ndef NO_SYS_FS
	USBMS_DBG_PLUG(iprintf("[USBMS PLUG] del_gendisk() : put_device() .. \n"));
	put_device(disk->driverfs_dev);
   }
 #ifndef _ETCPIP
	kobject_del(&disk->kobj);
 #endif // ndef _ETCPIP
}
