
// Name of File : scsi_mid.c
// ====================================================================

// drivers/scsi/scsi.c

/*
 *  scsi.c Copyright (C) 1992 Drew Eckhardt
 *         Copyright (C) 1993, 1994, 1995, 1999 Eric Youngdale
 *
 *  generic mid-level SCSI driver
 *      Initial versions: Drew Eckhardt
 *      Subsequent revisions: Eric Youngdale
 *
 *  <drew@colorado.edu>
 *
 *  Bug correction thanks go to :
 *      Rik Faith <faith@cs.unc.edu>
 *      Tommy Thorn <tthorn>
 *      Thomas Wuensche <tw@fgb1.fgb.mw.tu-muenchen.de>
 *
 *  Modified by Eric Youngdale eric@andante.org or ericy@gnu.ai.mit.edu to
 *  add scatter-gather, multiple outstanding request, and other
 *  enhancements.
 *
 *  Native multichannel, wide scsi, /proc/scsi and hot plugging
 *  support added by Michael Neuffer <mike@i-connect.net>
 *
 *  Added request_module("scsi_hostadapter") for kerneld:
 *  (Put an "alias scsi_hostadapter your_hostadapter" in /etc/modules.conf)
 *  Bjorn Ekwall  <bj0rn@blox.se>
 *  (changed to kmod)
 *
 *  Major improvements to the timeout, abort, and reset processing,
 *  as well as performance modifications for large queue depths by
 *  Leonard N. Zubkoff <lnz@dandelion.com>
 *
 *  Converted cli() code to spinlocks, Ingo Molnar
 *
 *  Jiffies wrap fixes (host->resetting), 3 Dec 1998 Andrea Arcangeli
 *
 *  out_of_space hacks, D. Gilbert (dpg) 990608
 */

// #ifdef _ETCPIP
#define __KERNEL__
        #ifdef _NEWLIB
          #include "string.h"
        #endif
   	#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsbMsInit.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP


/*
	#include <linux/config.h>
	#include <linux/module.h>
#include <linux/moduleparam.h>
*/
/*
#include <linux/kernel.h>
*/
#include "PoKernel.h"
/*
#include <linux/sched.h>
*/
#ifndef _ETCPIP
	#include <linux/timer.h>
 #else
	#ifndef NO_TIMER
		#include "PoTimer.h" // osport_timer.h
	#endif // ndef NO_TIMER
#endif
/*
#include <linux/string.h>
*/
#include "PoString.h" // linux/string.h
/*
#include <linux/slab.h>
*/
/*
#include <linux/stat.h>
*/
#include "stat.h"

/*
#include <linux/blkdev.h>
*/
#include "bitops.h" // B. S. Liaw add for test_bit()
#include "blkdev.h"
/*
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/completion.h>
#include <linux/devfs_fs_kernel.h>
*/

#define __KERNEL_SYSCALLS__

#ifndef _ETCPIP
	#include <linux/unistd.h>
	#include <linux/spinlock.h>
    #include <linux/kmod.h>
    #include <linux/interrupt.h>

#endif // ndef _ETCPIP

#include "scsi_ds.h"
#include "scsi_hosts.h"

#include "scsi_priv.h"
#include "scsi_logging.h"


/*
 * Definitions and constants.
 */

#define MIN_RESET_DELAY (2*HZ)

/* Do not call reset on error if we just did a reset within 15 sec. */
#define MIN_RESET_PERIOD (15*HZ)

/*
 * Macro to determine the size of SCSI command. This macro takes vendor
 * unique commands into account. SCSI commands in groups 6 and 7 are
 * vendor unique and we will depend upon the command length being
 * supplied correctly in cmd_len.
 */
#define CDB_SIZE(cmd)	(((((cmd)->cmnd[0] >> 5) & 7) < 6) ? \
				COMMAND_SIZE((cmd)->cmnd[0]) : (cmd)->cmd_len)

/*
 * Data declarations.
 */
unsigned long scsi_pid;
#ifdef USBMS_SCSI_SHORTCUT  /* Charles 20040526 */
unsigned long serial_number;
#else
static unsigned long serial_number;
#endif

/*
 * Note - the initial logging level can be set here to log events at boot time.
 * After the system is up, you may enable logging via the /proc interface.
 */
unsigned int scsi_logging_level;

const char *const scsi_device_types[MAX_SCSI_DEVICE_CODE] = {
	"Direct-Access    ",
	"Sequential-Access",
	"Printer          ",
	"Processor        ",
	"WORM             ",
	"CD-ROM           ",
	"Scanner          ",
	"Optical Device   ",
	"Medium Changer   ",
	"Communications   ",
	"Unknown          ",
	"Unknown          ",
	"Unknown          ",
	"Enclosure        ",
};
/*
static void scsi_softirq(void) ;
*/
static void scsi_softirq(struct scsi_cmnd *cmd) ;

// #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	// B. S. Liaw add this ..
	struct scsi_cmnd global_scsi_cmnd ;
	unsigned char * global_sense_buffer = (unsigned char *) NULL ;
// #endif // NO_ELEVATOR

/*
 * Function:    scsi_allocate_request
 *
 * Purpose:     Allocate a request descriptor.
 *
 * Arguments:   device    - device for which we want a request
 *		gfp_mask	- allocation flags passed to kmalloc
 *
 * Lock status: No locks assumed to be held.  This function is SMP-safe.
 *
 * Returns:     Pointer to request block.
 */
/*
struct scsi_request *scsi_allocate_request(struct scsi_device *sdev,
					   int gfp_mask)
*/
struct scsi_request *scsi_allocate_request(struct scsi_device *sdev)
{
 /*
 const int offset = ALIGN(sizeof(struct scsi_request), 4);
 const int size = offset + sizeof(struct request);
 */
 int offset ;
 int size ;
 struct scsi_request *sreq;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_allocate_request() Begin .. \n"));

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_allocate_request() : sizeof(struct scsi_request)=%08XH \n" ,
	sizeof(struct scsi_request) ));

 #ifndef _ETCPIP
	offset = ALIGN(sizeof(struct scsi_request), 4);
  #else
	offset = sizeof(struct scsi_request) ;
 #endif

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_allocate_request() : offset=%08XH \n" ,
	offset ));

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_allocate_request() : sizeof(struct request)=%d \n" ,
	sizeof(struct request) )) ;

 size = offset + sizeof(struct request);
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_allocate_request() : size=%d \n" ,
	size ));

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_allocate_request() : sreq = UsbAllocMemC() .. \n"
	));

 /*  
 sreq = kmalloc(size, gfp_mask);
 */
 sreq = (struct scsi_request *) UsbAllocMemC(USB_MEM_C_M
						#ifdef USB_DBG_MEM_C_ON
							, size  // size : 272
			    	    #endif
	    	       		) ;

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_allocate_request() : sreq=%08XH \n" ,
		sreq ));
 /*
 if (likely(sreq != NULL)) {
 */
 if ((sreq != NULL)) {
	/*
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] scsi_allocate_request() : clean sreq .. \n"
		));
	memset(sreq, 0, size);
	*/
	sreq->sr_request = (struct request *)(((char *)sreq) + offset);
	USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] scsi_allocate_request() : sreq->sr_request=%08XH \n" ,
			sreq->sr_request ));

	// #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		// B. S. Liaw Add these ..
		global_rq = sreq->sr_request ;
	// #endif // NO_ELEVATOR

	USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] scsi_allocate_request() : sdev=%08XH \n" ,
			sdev ));
	sreq->sr_device = sdev;
	USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] scsi_allocate_request() : sdev->host=%08XH \n" ,
			sdev->host ));
	sreq->sr_host = sdev->host;
	USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] scsi_allocate_request() : sreq->sr_host=%08XH \n" ,
			sreq->sr_host ));
	sreq->sr_magic = SCSI_REQ_MAGIC;
	USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_allocate_request() Set SCSI_REQ_MAGIC \n"));
	sreq->sr_data_direction = DMA_BIDIRECTIONAL;
	// #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_allocate_request() : Set Value sdev->request_queue->req = sreq->sr_request \n"));
		sdev->request_queue->req = sreq->sr_request ;
	// #endif // NO_ELEVATOR
   }
  else
   {
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] scsi_allocate_request() : Error : Null sreq=%08XH !! \n" ,
		sreq ));
   }
  
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_allocate_request() End. \n"));
 return sreq;
}

void __scsi_release_request(struct scsi_request *sreq)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __scsi_release_request() Begin .. \n"));
 /*
 if (likely(sreq->sr_command != NULL)) {
 */
 if((sreq->sr_command != NULL)) {
	struct scsi_cmnd *cmd = sreq->sr_command;

	sreq->sr_command = NULL;
	/*
	#ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		USBMS_DBG_INIT(iprintf("[USBMS INIT] __scsi_release_request() Not Call scsi_next_command() !! \n"));
	  #else
	*/
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __scsi_release_request() : scsi_next_command() .. \n"));
		scsi_next_command(cmd);
	/*
	#endif
	*/
   }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __scsi_release_request() End. \n"));
}

/*
 * Function:    scsi_release_request
 *
 * Purpose:     Release a request descriptor.
 *
 * Arguments:   sreq    - request to release
 *
 * Lock status: No locks assumed to be held.  This function is SMP-safe.
 */
void scsi_release_request(struct scsi_request *sreq)
{
 __scsi_release_request(sreq);


 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_release_request() : UsbFreeMemC(sreq) .. \n"
	));
 /*
 kfree(sreq);
 */
 UsbFreeMemC(USB_MEM_C_M,(unsigned char *) sreq
		#if defined(USB_DBG_MEM_C_ON)
			,sizeof(*sreq) // size : 272
		#endif
		);
}

struct scsi_host_cmd_pool {
 /* kmem_cache_t is on include/linux/slab.h
	kmem_cache_t	*slab;
 */
 unsigned int	users;
 #ifdef USB_DBGSTRING_ON
	char		*name;
 #endif // USB_DBGSTRING_ON
 /*
 unsigned int	slab_flags;
 */
 unsigned int	gfp_mask;
};

static struct scsi_host_cmd_pool scsi_cmd_pool = {
 #ifdef USB_DBGSTRING_ON
	.name		= "scsi_cmd_cache",
 #endif // USB_DBGSTRING_ON
 /*
 .slab_flags	= SLAB_HWCACHE_ALIGN,
 */
};

static struct scsi_host_cmd_pool scsi_cmd_dma_pool = {
 #ifdef USB_DBGSTRING_ON
	.name		= "scsi_cmd_cache(DMA)",
 #endif // USB_DBGSTRING_ON
 /*
 .slab_flags	= SLAB_HWCACHE_ALIGN|SLAB_CACHE_DMA,
 */
 .gfp_mask	= __GFP_DMA,
};


#ifndef NO_SEMPHORE
	static DECLARE_MUTEX(host_cmd_pool_mutex);
#endif // ndef NO_SEMPHORE

/*
static struct scsi_cmnd *__scsi_get_command(struct Scsi_Host *shost,
					    int gfp_mask)
*/
static struct scsi_cmnd *__scsi_get_command(struct Scsi_Host *shost)
{
 struct scsi_cmnd *cmd;

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	/* kmem_cache_t is on include/linux/slab.h
	cmd = kmem_cache_alloc(shost->cmd_pool->slab,
			gfp_mask | shost->cmd_pool->gfp_mask);
	*/
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] __scsi_get_command() : cmd = UsbAllocMemC() size=%d .. \n" ,
		sizeof(struct scsi_cmnd)
		));
	cmd = (struct scsi_cmnd *) UsbAllocMemC(USB_MEM_C_M
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct scsi_cmnd) // size : 212
		        #endif
           		) ;
 #else
	cmd = &global_scsi_cmnd ;
 #endif // NO_ELEVATOR
 if(!cmd) {
	dprintf("[USBMS] Error : out of memory !! \n") ;
	goto check_cmd;
   }

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	// B. S. Liaw : sense_buffer need uncached
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] __scsi_get_command() : cmd->sense_buffer = UsbAllocMemU() size=%d .. \n" ,
		SCSI_SENSE_BUFFERSIZE
		));
	cmd->sense_buffer = (unsigned char *) UsbAllocMemU(USB_MEM_U_S
				#ifdef USB_DBG_MEM_U_ON
					, SCSI_SENSE_BUFFERSIZE    // size : 64
		        #endif
           		) ;
  #else 
	cmd->sense_buffer = global_sense_buffer ;
 #endif // NO_ELEVATOR

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] __scsi_get_command() : cmd->sense_buffer=%08XH \n" ,
	cmd->sense_buffer
	));
 if(!cmd->sense_buffer) {
	dprintf("[USBMS] Error : out of memory !! \n") ;
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			USBMS_DBG_MEM(iprintf(
				"[USBMS MEM] __scsi_get_command() : UsbFreeMemC(cmd) .. \n"
				));
			UsbFreeMemC(USB_MEM_C_M,(unsigned char *) cmd
				#if defined(USB_DBG_MEM_C_ON)
				, sizeof(struct scsi_cmnd)  // size : 212
				#endif
				);
	#endif // ndef NO_ELEVATOR
	cmd = NULL ;
   }

 // -------------
 check_cmd:
 // -------------

 /*
 if (unlikely(!cmd)) {
 */
 if (!cmd) {
	unsigned long flags;
	/*
	#ifndef NO_SPINLOCK
			spin_lock_irqsave(&shost->free_list_lock, flags);
	#endif // ndef NO_SPINLOCK
	*/
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		/*
		if (likely(!list_empty(&shost->free_list))) {
		*/
		if (!list_empty(&shost->free_list)) {
			cmd = list_entry(shost->free_list.next,
					 struct scsi_cmnd, list);
			// #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] __scsi_get_command() : list_del_init(&cmd->list) .. \n"));
			list_del_init(&cmd->list);
			// #endif // ndef NO_ELEVATOR
		   }
	  #else
		if (shost->cmd) {
			USBMS_DBG_INFO(iprintf("[USBMS INFO] __scsi_get_command() : Get cmd from shost->cmd .. \n"));
			cmd = shost->cmd;
		   }
	#endif
	/*
	#ifndef NO_SPINLOCK
			spin_unlock_irqrestore(&shost->free_list_lock, flags);
	#endif // ndef NO_SPINLOCK
	*/
   }


 return cmd;
}

/*
 * Function:	scsi_get_command()
 *
 * Purpose:	Allocate and setup a scsi command block
 *
 * Arguments:	dev	- parent scsi device
 *		gfp_mask- allocator flags
 *
 * Returns:	The allocated scsi command structure.
 */
struct scsi_cmnd *scsi_get_command(struct scsi_device *dev, int gfp_mask)
{
 /*
 struct scsi_cmnd *cmd = __scsi_get_command(dev->host, gfp_mask);
 */
 struct scsi_cmnd *cmd ;
 unsigned char *tmp_sense_buffer_pr ; // B. S. Liaw add this !!


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_get_command() : Begin .. \n" )) ;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_get_command() : __scsi_get_command() .. \n" )) ;
 cmd = __scsi_get_command(dev->host);
 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_get_command() : cmd=%08XH \n",cmd)) ;

 #ifdef USBMS_BS_ASSERT
	if (!(   (unsigned long) cmd
		   > (unsigned long) 0x200000L
         )
	   )
	   {
		iprintf("[ASSERT][USBMS] scsi_get_command() : " ) ;
	    iprintf("Error : Invlid address cmd=%08XH \n",
			cmd ) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT

 /*
 if (likely(cmd != NULL)) {
 */
 if ((cmd != NULL)) {
	unsigned long flags;


	tmp_sense_buffer_pr = cmd->sense_buffer ; // B. S. Liaw : backup
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_get_command() : memset() .. \n" )) ;
	memset(cmd, 0, sizeof(*cmd));
	cmd->sense_buffer =	tmp_sense_buffer_pr ; // B. S. Liaw : restore

	cmd->device = dev;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_get_command() : Set Value : cmd->device=%08XH \n",cmd->device
		)) ;
	cmd->state = SCSI_STATE_UNUSED;
		cmd->owner = SCSI_OWNER_NOBODY;
	#ifndef NO_TIMER
		init_timer(&cmd->eh_timeout);
	#endif // ndef NO_TIMER
	/*
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_get_command() : INIT_LIST_HEAD(&cmd->list) .. \n"));
		INIT_LIST_HEAD(&cmd->list);
	#endif // ndef NO_ELEVATOR
	*/
	/*
	#ifndef NO_SPINLOCK
		spin_lock_irqsave(&dev->list_lock, flags);
	#endif // ndef NO_SPINLOCK
	*/
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_get_command() : dev->cmd_list=%08XH \n",
			dev->cmd_list ));
	#ifdef USBMS_BS_ASSERT
			if (!(   (unsigned long) &dev->cmd_list
				   > (unsigned long) 0x200000L
		         )
			   )
			   {
				iprintf("[ASSERT][USBMS] scsi_get_command() : " ) ;
    		    iprintf("Error : Invlid address &dev->cmd_list=%08XH \n",
						&dev->cmd_list) ;
			    iprintf("[ASSERT][USBMS] Continue ..\n") ;
    		   }
	#endif // USBMS_BS_ASSERT
 #endif // ndef NO_ELEVATOR
	/*
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_get_command() : list_add_tail(&cmd->list) .. \n"));
		list_add_tail(&cmd->list, &dev->cmd_list);
	 #else
	*/	
		USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_get_command() : Set dev->cmd .. \n"));
		dev->cmd = cmd ;
	/*	
	#endif
	*/
	/*
	#ifndef NO_SPINLOCK
		spin_unlock_irqrestore(&dev->list_lock, flags);
	#endif // ndef NO_SPINLOCK
	*/
   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_get_command() : End. \n" )) ;
 return cmd;
}				

/*
 * Function:	scsi_put_command()
 *
 * Purpose:	Free a scsi command block
 *
 * Arguments:	cmd	- command block to free
 *
 * Returns:	Nothing.
 *
 * Notes:	The command must not belong to any lists.
 */
void scsi_put_command(struct scsi_cmnd *cmd)
{
 struct Scsi_Host *shost = cmd->device->host;
 unsigned long flags;

	
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_put_command() Begin .. \n"));
 /* serious error if the command hasn't come from a device list */
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(&cmd->device->list_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	// BUG_ON(list_empty(&cmd->list));
	if(!((list_empty(&cmd->list))!=0)) {
		USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_put_command() : Error : cmd->list Empty !! \n"));
		dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }
  #else
 */
	if(!(shost->cmd)) {
		USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_put_command() : Error : Null shost->cmd !! \n"));
		// dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
	   }
 /*
 #endif
 */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_put_command() : list_del_init(&cmd->list) .. \n"));
	list_del_init(&cmd->list);
  #else
 */
	/*
	if (shost->cmd) {
		USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_put_command() : Clear from shost->cmd .. \n"));
		shost->cmd = NULL;
	   }
	*/
 /*
 #endif
 */

 /*
 #ifndef NO_SPINLOCK
	spin_unlock(&cmd->device->list_lock);
 #endif // ndef NO_SPINLOCK
 */
 /* changing locks here, don't need to restore the irq state */
 /*
 #ifndef NO_SPINLOCK
	spin_lock(&shost->free_list_lock);
 #endif // ndef NO_SPINLOCK
 */
 /*
 if (unlikely(list_empty(&shost->free_list))) {
 */
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	if (list_empty(&shost->free_list)) {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_put_command() : list_add(&cmd->list) .. \n"));
		list_add(&cmd->list, &shost->free_list);
		cmd = NULL;
	   }
  #else
	/*
	if (!(shost->cmd)) {
		USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_put_command() : Set shost->cmd .. \n"));
		shost->cmd = cmd ;
		cmd = NULL;
	   }
	*/
 #endif

 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(&shost->free_list_lock, flags);
 #endif // ndef NO_SPINLOCK
 */

 /*
 if (likely(cmd != NULL))
 */
 if((cmd != NULL))
   {
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		/* kmem_cache_t is on include/linux/slab.h
		kmem_cache_free(shost->cmd_pool->slab, cmd);
		*/
	
		// B. S. Liaw : sense_buffer need uncached
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] scsi_put_command() : UsbFreeMemU(cmd->sense_buffer) .. \n"
			));
		UsbFreeMemU(USB_MEM_U_S,(unsigned char *) cmd->sense_buffer
					#if defined(USB_DBG_MEM_U_ON)
						, SCSI_SENSE_BUFFERSIZE    // size : 64
					#endif
					);
		cmd->sense_buffer = NULL ;
	
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] scsi_put_command() : UsbFreeMemC(cmd) .. \n"
			));
		UsbFreeMemC(USB_MEM_C_M,(unsigned char *) cmd
				#if defined(USB_DBG_MEM_C_ON)
					, sizeof(struct scsi_cmnd) // size : 212
				#endif
				);
	#endif // ndef NO_ELEVATOR
	cmd = NULL ;
   }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_put_command() End. \n"));
}

/*
 * Function:	scsi_setup_command_freelist()
 *
 * Purpose:	Setup the command freelist for a scsi host.
 *
 * Arguments:	shost	- host to allocate the freelist for.
 *
 * Returns:	Nothing.
 */
int scsi_setup_command_freelist(struct Scsi_Host *shost)
{
 struct scsi_host_cmd_pool *pool;
 struct scsi_cmnd *cmd;


 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	/*
	#ifndef NO_SPINLOCK
		spin_lock_init(&shost->free_list_lock);
	#endif // ndef NO_SPINLOCK
	*/
	INIT_LIST_HEAD(&shost->free_list);
 #endif // ndef NO_ELEVATOR

 /*
  * Select a command slab for this host and create it if not
  * yet existant.
  */
 #ifndef NO_SEMPHORE
	down(&host_cmd_pool_mutex);
 #endif // ndef NO_SEMPHORE

 pool = (shost->unchecked_isa_dma ? &scsi_cmd_dma_pool : &scsi_cmd_pool);

 if (!pool->users) {
	/* kmem_cache_t is on include/linux/slab.h
		pool->slab = kmem_cache_create(pool->name,
				sizeof(struct scsi_cmnd), 0, // size : 212
				pool->slab_flags, NULL, NULL);
	*/
		
	/* kmem_cache_t is on include/linux/slab.h
		if (!pool->slab)
			goto fail;
	*/
	}

 pool->users++;
 shost->cmd_pool = pool;
 #ifndef NO_SEMPHORE
	up(&host_cmd_pool_mutex);
 #endif // ndef NO_SEMPHORE

 /*
  * Get one backup command for this host.
  */

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	/* kmem_cache_t is on include/linux/slab.h
	cmd = kmem_cache_alloc(shost->cmd_pool->slab,
			GFP_KERNEL | shost->cmd_pool->gfp_mask);
	*/
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_setup_command_freelist() : cmd = UsbAllocMemC() size=%d .. \n" ,
		sizeof(struct scsi_cmnd)
		));

	cmd = (struct scsi_cmnd *) UsbAllocMemC(USB_MEM_C_M
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct scsi_cmnd) // size : 212
		        #endif
           		) ;
 #else
	cmd = &global_scsi_cmnd ;
 #endif // NO_ELEVATOR
 if(!cmd) {
	dprintf("[USBMS] Error : out of memory !! \n") ;
	goto fail2;
   }

 // B. S. Liaw : sense_buffer need uncached
 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_setup_command_freelist() : cmd->sense_buffer = UsbAllocMemU() size=%d .. \n" ,
		SCSI_SENSE_BUFFERSIZE
		));
 cmd->sense_buffer = (unsigned char *) UsbAllocMemU(USB_MEM_U_S
					#ifdef USB_DBG_MEM_U_ON
						, SCSI_SENSE_BUFFERSIZE    // size : 64
			        #endif
	           		) ;
 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_setup_command_freelist() : cmd->sense_buffer=%08XH \n" ,
		cmd->sense_buffer
		));
 // #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	global_sense_buffer = cmd->sense_buffer ;
 // #endif // ndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 

  if(!cmd->sense_buffer) {
	dprintf("[USBMS] Error : out of memory !! \n") ;
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] scsi_setup_command_freelist() : UsbFreeMemC(cmd) .. \n"
			));
		UsbFreeMemC(USB_MEM_C_M,(unsigned char *) cmd
			#if defined(USB_DBG_MEM_C_ON)
				, sizeof(struct scsi_cmnd)  // size : 212
			#endif
			);
	#endif // ndef NO_ELEVATOR
	cmd = NULL ;
	goto fail2;
   }

 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_setup_command_freelist() : list_add(&cmd->list) .. \n"));
	list_add(&cmd->list, &shost->free_list);
  #else
	USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_setup_command_freelist() : Set shost->cmd .. \n"));
	shost->cmd = cmd ;
 #endif

 return 0;

 fail2:

 if (!--pool->users) 
   {
	/* kmem_cache_t is on include/linux/slab.h
		kmem_cache_destroy(pool->slab);
	*/
   }
 return -ENOMEM;

 fail:
 #ifndef NO_SEMPHORE
	up(&host_cmd_pool_mutex);
 #endif // ndef NO_SEMPHORE
 return -ENOMEM;

}

/*
 * Function:	scsi_destroy_command_freelist()
 *
 * Purpose:	Release the command freelist for a scsi host.
 *
 * Arguments:	shost	- host that's freelist is going to be destroyed
 */
void scsi_destroy_command_freelist(struct Scsi_Host *shost)
{
 // USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_destroy_command_freelist() Begin .. \n"));
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	while (!list_empty(&shost->free_list)) {
 #endif // ndef NO_ELEVATOR
		struct scsi_cmnd *cmd;

		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_destroy_command_freelist() Begin .. \n"));
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			cmd = list_entry(shost->free_list.next, struct scsi_cmnd, list);
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_destroy_command_freelist() : list_del_init(&cmd->list) .. \n"));
			list_del_init(&cmd->list);
		 #else
			cmd = shost->cmd ;
			USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_destroy_command_freelist() : Clear shost->cmd .. \n"));
			shost->cmd = NULL ;
		#endif
		/* kmem_cache_t is on include/linux/slab.h
		kmem_cache_free(shost->cmd_pool->slab, cmd);
		*/
		// B. S. Liaw : sense_buffer need uncached
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] scsi_destroy_command_freelist() : UsbFreeMemU(cmd->sense_buffer) .. \n"
			));
		UsbFreeMemU(USB_MEM_U_S,(unsigned char *) cmd->sense_buffer
					#if defined(USB_DBG_MEM_U_ON)
						, SCSI_SENSE_BUFFERSIZE    // size : 64
					#endif
					);
		cmd->sense_buffer = NULL ;

		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			USBMS_DBG_MEM(iprintf(
				"[USBMS MEM] scsi_destroy_command_freelist() : UsbFreeMemC(cmd) .. \n"
				));
			UsbFreeMemC(USB_MEM_C_M,(unsigned char *) cmd
				#if defined(USB_DBG_MEM_C_ON)
					, sizeof(struct scsi_cmnd)  // size : 212
				#endif
				);
		#endif // ndef NO_ELEVATOR

		cmd = NULL ;
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	   }
	#endif // ndef NO_ELEVATOR

 #ifndef NO_SEMPHORE
	down(&host_cmd_pool_mutex);
 #endif // ndef NO_SEMPHORE

 if(!--shost->cmd_pool->users) {
	/* kmem_cache_t is on include/linux/slab.h
	kmem_cache_destroy(shost->cmd_pool->slab);
	*/
	/*
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_destroy_command_freelist() : UsbFreeMemC(shost->cmd_pool) .. \n"
		));
	UsbFreeMemC(USB_MEM_C_S,(unsigned char *) shost->cmd_pool
		#if defined(USB_DBG_MEM_C_ON)
			, sizeof(shost->cmd_pool)
		#endif
		);
	*/
   }
 /*
 #ifndef NO_SEMPHORE
	up(&host_cmd_pool_mutex);
 #endif // ndef NO_SEMPHORE
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_destroy_command_freelist() End. \n"));
}

/*
 * Function:    scsi_dispatch_command
 *
 * Purpose:     Dispatch a command to the low-level driver.
 *
 * Arguments:   cmd - command block we are dispatching.
 *
 * Notes:
 */
int scsi_dispatch_cmd(struct scsi_cmnd *cmd)
{
 /*
 struct Scsi_Host *host = cmd->device->host;
 */
 struct Scsi_Host *host ;
 unsigned long flags = 0;
 unsigned long timeout;
 int rtn = 0;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_dispatch_cmd() Begin .. \n"));

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_dispatch_cmd() : cmd=%08XH \n",cmd
	)) ;
 #ifdef USBMS_BS_ASSERT
	if (!(   (unsigned long) cmd
		   > (unsigned long) 0x200000L
         )
	   )
	   {
		iprintf("[ASSERT][USBMS] scsi_dispatch_cmd() : " ) ;
	    iprintf("Error : Invlid address cmd=%08XH \n",
			cmd ) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_dispatch_cmd() : cmd->device=%08XH \n",cmd->device
	)) ;

 #ifdef USBMS_BS_ASSERT
	if (!(   (unsigned long) cmd->device
		   > (unsigned long) 0x200000L
         )
	   )
	   {
		iprintf("[ASSERT][USBMS] scsi_dispatch_cmd() : " ) ;
	    iprintf("Error : Invlid address cmd->device=%08XH \n",
			cmd->device ) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_dispatch_cmd() : cmd->device->host=%08XH \n",
	cmd->device->host
	)) ;
 #ifdef USBMS_BS_ASSERT
	if (!(   (unsigned long) cmd->device->host
		   > (unsigned long) 0x200000L
         )
	   )
	   {
		iprintf("[ASSERT][USBMS] scsi_dispatch_cmd() : " ) ;
	    iprintf("Error : Invlid address cmd->device->host=%08XH \n",
			cmd->device->host ) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT

 host = cmd->device->host;

 /* Assign a unique nonzero serial_number. */
 /* XXX(hch): this is racy */
 if (++serial_number == 0)
		serial_number = 1;
 cmd->serial_number = serial_number;
 cmd->pid = scsi_pid++;

 /* 
  * If SCSI-2 or lower, store the LUN value in cmnd.
  */
 if (cmd->device->scsi_level <= SCSI_2) {
		cmd->cmnd[1] = (cmd->cmnd[1] & 0x1f) |
			       (cmd->device->lun << 5 & 0xe0);
	}

 /*
  * We will wait MIN_RESET_DELAY clock ticks after the last reset so
  * we can avoid the drive not being ready.
  */
 timeout = host->last_reset + MIN_RESET_DELAY;
 /*
 #ifndef NO_TIMER
	if (host->resetting && time_before(jiffies, timeout)) {
		int ticks_remaining = timeout - jiffies;

		// * NOTE: This may be executed from within an interrupt
		// * handler!  This is bad, but for now, it'll do.  The irq
		// * level of the interrupt handler has been masked out by the
		// * platform dependent interrupt handling code already, so the
		// * sti() here will not cause another call to the SCSI host's
		// * interrupt handler (assuming there is one irq-level per
		// * host).
		
		while (--ticks_remaining >= 0)
           {
            
			// mdelay(1 + 999 / HZ);
			
			if (in_interrupt())
			   {
				mdelay(1 + 999 / HZ);
			   }
			  else
			   {
				wait_ms(1 + 999 / HZ);
			   } 
           }
		host->resetting = 0;
	   }

	scsi_add_timer(cmd, cmd->timeout_per_command, scsi_times_out);
 #endif // ndef NO_TIMER
 */

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_dispatch_cmd() : cmd->buffer=%08XH \n",
	cmd->buffer
	));

 /*
  * We will use a queued command if possible, otherwise we will
  * emulate the queuing and calling of completion function ourselves.
  */
 /*
 SCSI_LOG_MLQUEUE(3, printk("scsi_dispatch_cmnd (host = %d, "
				"channel = %d, target = %d, command = %p, "
				"buffer = %p, \nbufflen = %d, done = %p)\n",
				host->host_no, cmd->device->channel,
				cmd->device->id, cmd->cmnd, cmd->buffer,
				cmd->bufflen, cmd->done));
 */
 /*
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] scsi_dispatch_cmd() : host = %d, "
		"channel = %d, target = %d , command = %p \n "
	"             buffer = %08XH , bufflen = %d, done = %08XH)\n",
	host->host_no, cmd->device->channel,
	cmd->device->id, cmd->cmnd, cmd->buffer,
	cmd->bufflen, cmd->done));
 */
 USBMS_DBG_SCSI(iprintf(
	"[USBMS SCSI] scsi_dispatch_cmd() : host=%d, channel=%d, target=%d \n" ,
	host->host_no, cmd->device->channel,
	cmd->device->id
	));

 USBMS_DBG_SCSI(iprintf(
	"[USBMS SCSI]     command=%08XH  buffer=%08XH \n" ,
	cmd->cmnd, cmd->buffer
	));
 USBMS_DBG_SCSI(iprintf(
	"[USBMS SCSI]     bufflen=%d, done=%08XH \n" ,
	cmd->bufflen, cmd->done
	));

 #ifdef USBMS_BS_ASSERT
	if (cmd->bufflen) {
		if ( ! (IS_VALID_ADDR(cmd->buffer)) ) {
			iprintf("[ASSERT][USBMS] scsi_dispatch_cmd(1) : " ) ;
   		    iprintf("Error : Invalid Address !!\n" ) ;
	        iprintf("[ASSERT][USBMS] cmd->buffer=%08XH \n",
					cmd->buffer ) ;
	        iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
		if ( !(USB_IS_UN_CACHED(cmd->buffer)) ) {
			iprintf("[ASSERT][USBMS] scsi_dispatch_cmd(2) : " ) ;
		    iprintf("Error : Not Un-Cached !!\n" ) ;
		    iprintf("[ASSERT][USBMS] cmd->buffer=%08XH \n", cmd->buffer ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	   } // if has length
 #endif // USBMS_BS_ASSERT

 cmd->state = SCSI_STATE_QUEUED;
 cmd->owner = SCSI_OWNER_LOWLEVEL;

 /*
  * Before we queue this command, check if the command
  * length exceeds what the host adapter can handle.
  */
 if (CDB_SIZE(cmd) > cmd->device->host->max_cmd_len) {
	/*
	SCSI_LOG_MLQUEUE(3,
				printk("queuecommand : command too long.\n"));
	*/
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] scsi_dispatch_cmd(1) : Error : command too long !! \n"
		)) ;

	cmd->result = (DID_ABORT << 16);
	/*
	#ifndef NO_SPINLOCK
			spin_lock_irqsave(host->host_lock, flags);
	#endif // ndef NO_SPINLOCK
	*/
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_dispatch_cmd() : scsi_done() .. \n"));
	scsi_done(cmd);
	/*
	#ifndef NO_SPINLOCK
		spin_unlock_irqrestore(host->host_lock, flags);
	#endif // ndef NO_SPINLOCK
	*/
	goto out;
   }
 /*
 SCSI_LOG_MLQUEUE(3, printk("queuecommand : routine at %p\n",
				   host->hostt->queuecommand));
 */
 USBMS_DBG_SCSI(iprintf(
	"[USBMS SCSI] scsi_dispatch_cmd() : queuecommand() at %08XH \n",
	host->hostt->queuecommand));

 /*
 #ifndef NO_SPINLOCK
                spin_lock_irqsave(host->host_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 /*
 if (unlikely(test_bit(SHOST_CANCEL, &host->shost_state))) {
 */
 if(test_bit(SHOST_CANCEL, &host->shost_state)) {
	cmd->result = (DID_NO_CONNECT << 16);
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_dispatch_cmd() : scsi_done() .. \n"));
	scsi_done(cmd);
   }
  else {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_dispatch_cmd() : host->hostt->queuecommand() .. \n"));
#ifdef USBMS_NO_USB_TASK	// bnchen 09.May.2005
	rtn = scsi_issue_command(cmd, scsi_done);
#else
	rtn = host->hostt->queuecommand(cmd, scsi_done);
#endif
   }

 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(host->host_lock, flags);
 #endif // ndef NO_SPINLOCK
 */

 if(rtn) {
	/*
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_dispatch_cmd() : scsi_queue_insert() .. \n"));
		scsi_queue_insert(cmd,
				(rtn == SCSI_MLQUEUE_DEVICE_BUSY) ?
				 rtn : SCSI_MLQUEUE_HOST_BUSY);
	#endif // ndef NO_ELEVATOR
	*/
	/*
	SCSI_LOG_MLQUEUE(3,
		    printk("queuecommand : request rejected\n"));
	*/
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] scsi_dispatch_cmd(2) : Error : request rejected !! \n"
		));
	dprintf("[USBMS] Error : request rejected !! \n") ;
   }

 out:

 SCSI_LOG_MLQUEUE(3, printk("[USBMS SCSI] leaving scsi_dispatch_cmnd()\n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_dispatch_cmd() End. \n"));
 return rtn;
}

/*
 * Function:    scsi_init_cmd_from_req
 *
 * Purpose:     Queue a SCSI command
 * Purpose:     Initialize a struct scsi_cmnd from a struct scsi_request
 *
 * Arguments:   cmd       - command descriptor.
 *              sreq      - Request from the queue.
 *
 * Lock status: None needed.
 *
 * Returns:     Nothing.
 *
 * Notes:       Mainly transfer data from the request structure to the
 *              command structure.  The request structure is allocated
 *              using the normal memory allocator, and requests can pile
 *              up to more or less any depth.  The command structure represents
 *              a consumable resource, as these are allocated into a pool
 *              when the SCSI subsystem initializes.  The preallocation is
 *              required so that in low-memory situations a disk I/O request
 *              won't cause the memory manager to try and write out a page.
 *              The request structure is generally used by ioctls and character
 *              devices.
 */
void scsi_init_cmd_from_req(struct scsi_cmnd *cmd, struct scsi_request *sreq)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_init_cmd_from_req() Begin .. \n"));
 sreq->sr_command = cmd;

 cmd->owner = SCSI_OWNER_MIDLEVEL;
 cmd->cmd_len = sreq->sr_cmd_len;
 /*
 #ifndef NO_SCATTER_LIST
	cmd->use_sg = sreq->sr_use_sg;
 #endif // ndef NO_SCATTER_LIST
 */
 USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_init_cmd_from_req() : cmd->request = sreq->sr_request .. \n"));
 cmd->request = sreq->sr_request;
 memcpy(cmd->data_cmnd, sreq->sr_cmnd, sizeof(cmd->data_cmnd));
 cmd->serial_number = 0;
 cmd->serial_number_at_timeout = 0;
 cmd->bufflen = sreq->sr_bufflen;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_init_cmd_from_req() : Set Value : cmd->bufflen=%d \n",
	cmd->bufflen
	));
 cmd->buffer = sreq->sr_buffer;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_init_cmd_from_req() : Set Value : cmd->buffer=%08XH \n",
	cmd->buffer
	));
 cmd->retries = 0;
 cmd->allowed = sreq->sr_allowed;
 cmd->done = sreq->sr_done;
 cmd->timeout_per_command = sreq->sr_timeout_per_command;
 cmd->sc_data_direction = sreq->sr_data_direction;
 #ifndef NO_SCATTER_LIST
	cmd->sglist_len = sreq->sr_sglist_len;
 #endif // ndef NO_SCATTER_LIST
 cmd->underflow = sreq->sr_underflow;
 cmd->sc_request = sreq;
 memcpy(cmd->cmnd, sreq->sr_cmnd, sizeof(sreq->sr_cmnd));

 /*
  * Zero the sense buffer.  Some host adapters automatically request
  * sense on error.  0 is not a valid sense code.
  */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_init_cmd_from_req() : memset() .. \n"));
 /*
 memset(cmd->sense_buffer, 0, sizeof(sreq->sr_sense_buffer));
 */
 memset(cmd->sense_buffer, 0, SCSI_SENSE_BUFFERSIZE);
 // B. S. Liaw : request_buffer need uncached
 cmd->request_buffer = sreq->sr_buffer;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_init_cmd_from_req() : Set Value : cmd->request_buffer=%08XH .. \n",
	cmd->request_buffer
	));
 cmd->request_bufflen = sreq->sr_bufflen;
 
 #ifndef NO_SCATTER_LIST
	cmd->old_use_sg = cmd->use_sg;
 #endif // ndef NO_SCATTER_LIST
 
 if (cmd->cmd_len == 0)
		cmd->cmd_len = COMMAND_SIZE(cmd->cmnd[0]);
 cmd->old_cmd_len = cmd->cmd_len;
 cmd->sc_old_data_direction = cmd->sc_data_direction;
 cmd->old_underflow = cmd->underflow;

 /*
  * Start the timer ticking.
  */
 cmd->internal_timeout = NORMAL_TIMEOUT;
 cmd->abort_reason = 0;
 cmd->result = 0;

 SCSI_LOG_MLQUEUE(3, printk("Leaving scsi_init_cmd_from_req()\n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_init_cmd_from_req() End. \n"));
}

/*
 * Per-CPU I/O completion queue.
 */
/*
#ifndef _ETCPIP
static DEFINE_PER_CPU(struct list_head, scsi_done_q);
#endif // ndef _ETCPIP
*/
/**
 * scsi_done - Enqueue the finished SCSI command into the done queue.
 * @cmd: The SCSI Command for which a low-level device driver (LLDD) gives
 * ownership back to SCSI Core -- i.e. the LLDD has finished with it.
 *
 * This function is the mid-level's (SCSI Core) interrupt routine, which
 * regains ownership of the SCSI command (de facto) from a LLDD, and enqueues
 * the command to the done queue for further processing.
 *
 * This is the producer of the done queue who enqueues at the tail.
 *
 * This function is interrupt context safe.
 */
void scsi_done(struct scsi_cmnd *cmd)
{
 unsigned long flags;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_done() Begin .. \n"));

 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_done() : cmd=%08XH \n",
	cmd )) ;
 #ifdef USBMS_BS_ASSERT
	if ( ! (IS_VALID_ADDR(cmd)) ) {
		iprintf("[ASSERT][USBMS] scsi_done() : " ) ;
	    iprintf("Error : Invalid Address !!\n" ) ;
        iprintf("[ASSERT][USBMS] cmd=%08XH \n",
				cmd ) ; ;
        iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT
 /*
  * We don't have to worry about this one timing out any more.
  * If we are unable to remove the timer, then the command
  * has already timed out.  In which case, we have no choice but to
  * let the timeout function run, as we have no idea where in fact
  * that function could really be.  It might be on another processor,
  * etc, etc.
  */
 /*
 #ifndef NO_TIMER
	if (!scsi_delete_timer(cmd))
		return;
 #endif // ndef NO_TIMER
 */
 /*
  * Set the serial numbers back to zero
  */
 cmd->serial_number = 0;
 cmd->serial_number_at_timeout = 0;
 cmd->state = SCSI_STATE_BHQUEUE;
 cmd->owner = SCSI_OWNER_BH_HANDLER;

 /*
  * Next, enqueue the command into the done queue.
  * It is a per-CPU queue, so we just disable local interrupts
  * and need no spinlock.
  */
 /*
	local_irq_save(flags);
 #ifndef _ETCPIP
	list_add_tail(&cmd->eh_entry, &__get_cpu_var(scsi_done_q));
 #endif // ndef _ETCPIP
 */
 /*
 #ifndef _ETCPIP
	//  SCSI_SOFTIRQ is on include/linux/interrupt
	raise_softirq_irqoff(SCSI_SOFTIRQ);
	
  #else
 */
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_done() : scsi_softirq() .. \n"));
	// scsi_softirq() ;
	scsi_softirq(cmd) ;
 /*
 #endif
 */
 /*
 #ifndef _ETCPIP
	local_irq_restore(flags);
 #endif // ndef _ETCPIP
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_done() End. \n"));
}

/**
 * scsi_softirq - Perform post-interrupt processing of finished SCSI commands.
 *
 * This is the consumer of the done queue.
 *
 * This is called with all interrupts enabled.  This should reduce
 * interrupt latency, stack depth, and reentrancy of the low-level
 * drivers.
 */

/* struct softirq_action is on include/linux/interrupt.h
static void scsi_softirq(struct softirq_action *h)
*/
static void scsi_softirq(struct scsi_cmnd *cmd)
{
 /*
 #ifndef _ETCPIP
 LIST_HEAD(local_q);
 #endif // ndef _ETCPIP
 */
 // #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	// B. S. Liaw Add this !!
	// struct us_data *us ;
 // #endif // NO_ELEVATOR

#if (_USB2 == 1)
 // temporary solution - bonian chen 15.04.2005
 // BLOCK_PREEMPTION takes too long, it will cause other time critical task to crash
 // see also core\hub.c
#else	// defined (_USB2)
 BLOCK_PREEMPTION;
#endif	// defined (_USB2)
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_softirq() Begin .. \n"));
 /*
 #ifndef _ETCPIP
	local_irq_disable();
	list_splice_init(&__get_cpu_var(scsi_done_q), &local_q);
	local_irq_enable();
 #endif // ndef _ETCPIP
 */
 // #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	// B. S. Liaw Add this !!
 	// us = (struct us_data *)cmd->device->host->hostdata[0];
 // #endif // NO_ELEVATOR

 /*
 #ifndef _ETCPIP
 while (!list_empty(&local_q)) {
	struct scsi_cmnd *cmd = list_entry(local_q.next,
						   struct scsi_cmnd, eh_entry);
 
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_softirq(1) : list_del_init() .. \n"));
	list_del_init(&cmd->eh_entry);
 #endif // ndef _ETCPIP
 */
 switch (scsi_decide_disposition(cmd)) {
	case SCSI_SUCCESS:
		USBMS_DBG_SCSI(iprintf("[USBMS SCSI] scsi_softirq() : SCSI_SUCCESS \n"));
		/*
		 * Add to BH queue.
		 */
		/*
		SCSI_LOG_MLCOMPLETE(3,
					    printk("Command finished %d %d "
						   "0x%x\n",
		*/
		USBMS_DBG_SCSI(iprintf(
			"[USBMS SCSI] scsi_softirq() : Command finished %d %d result=%08XH \n",
					   cmd->device->host->host_busy,
					   cmd->device->host->host_failed,
						   cmd->result));

		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_softirq(2) : scsi_finish_command() .. \n"));
		scsi_finish_command(cmd);
		break;
	case NEEDS_RETRY:
		USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_softirq() : Error : NEEDS_RETRY !! \n"));
		/*
		 * We only come in here if we want to retry a
		 * command.  The test to see whether the
		 * command should be retried should be keeping
		 * track of the number of tries, so we don't
		 * end up looping, of course.
		 */
		SCSI_LOG_MLCOMPLETE(3, printk("[USBMS] Command needs retry "
						      "%d %d 0x%x\n",
					      cmd->device->host->host_busy,
					      cmd->device->host->host_failed,
						      cmd->result));

		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			// B. S. Liaw : This leader to recursive
			// scsi_retry_command()->scsi_dispatch_cmd()->queuecommand()
			scsi_retry_command(cmd);
		 #else
			// B. S. Liaw Add this !!
			// us->sm_state = US_STATE_IDLE ;
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_softirq(3) : scsi_finish_command() .. \n"));
			scsi_finish_command(cmd);
		#endif // NO_ELEVATOR
		break;
	case ADD_TO_MLQUEUE:
		USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_softirq() : Error : ADD_TO_MLQUEUE !! \n"));
		/* 
		 * This typically happens for a QUEUE_FULL
		 * message - typically only when the queue
		 * depth is only approximate for a given
		 * device.  Adding a command to the queue for
		 * the device will prevent further commands
		 * from being sent to the device, so we
		 * shouldn't end up with tons of things being
		 * sent down that shouldn't be.
		 */
		SCSI_LOG_MLCOMPLETE(3, printk("Command rejected as "
						      "device queue full, "
						      "put on ml queue %p\n",
						      cmd));
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_softirq(4) : scsi_queue_insert() .. \n"));
		scsi_queue_insert(cmd, SCSI_MLQUEUE_DEVICE_BUSY);
		break;
	default:
		USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_softirq() : Error : Failed !! \n"));
		/*
		 * Here we have a fatal error of some sort.
		 * Turn it over to the error handler.
		 */
		SCSI_LOG_MLCOMPLETE(3,
					    printk("Command failed %p %x "
						   "busy=%d failed=%d\n",
						   cmd, cmd->result,
					   cmd->device->host->host_busy,
					   cmd->device->host->host_failed));

		/*
		 * Dump the sense information too.
		 */
		if (status_byte(cmd->result) & CHECK_CONDITION) {
			SCSI_LOG_MLCOMPLETE(3, print_sense("bh", cmd));
		   }

		/*
		 * We only fail here if the error recovery thread
		 * has died.
		 */
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_softirq(5) : scsi_eh_scmd_add() .. \n"));
		if (!scsi_eh_scmd_add(cmd, 0)) {
				USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_softirq(6) : scsi_finish_command() .. \n"));
				scsi_finish_command(cmd);
			   }
   }

 /*
 #ifndef _ETCPIP
	}
 #endif // ndef _ETCPIP
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_softirq() End. \n"));
#if (_USB2 == 1)
 // temporary solution - bonian chen 15.04.2005
 // BLOCK_PREEMPTION takes too long, it will cause other time critical task to crash
 // see also core\hub.c
#else	// defined (_USB2)
 UNBLOCK_PREEMPTION;
#endif	// defined (_USB2)
}

/*
 * Function:    scsi_retry_command
 *
 * Purpose:     Send a command back to the low level to be retried.
 *
 * Notes:       This command is always executed in the context of the
 *              bottom half handler, or the error handler thread. Low
 *              level drivers should not become re-entrant as a result of
 *              this.
 */
int scsi_retry_command(struct scsi_cmnd *cmd)
{
 /*
  * Restore the SCSI command state.
  */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_retry_command() : scsi_setup_cmd_retry() .. \n"));
 scsi_setup_cmd_retry(cmd);

 /*
  * Zero the sense information from the last time we tried
  * this command.
  */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_retry_command() : memset() .. \n"));
 /*
 memset(cmd->sense_buffer, 0, sizeof(cmd->sense_buffer));
 */
 memset(cmd->sense_buffer, 0, SCSI_SENSE_BUFFERSIZE);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_retry_command() : scsi_dispatch_cmd() .. \n"));
 return scsi_dispatch_cmd(cmd);
}

/*
 * Function:    scsi_finish_command
 *
 * Purpose:     Pass command off to upper layer for finishing of I/O
 *              request, waking processes that are waiting on results,
 *              etc.
 */
void scsi_finish_command(struct scsi_cmnd *cmd)
{
 /*
 struct scsi_device *sdev = cmd->device;
 struct Scsi_Host *shost = sdev->host;
 */
 struct scsi_device *sdev ;
 struct Scsi_Host *shost ;
 struct scsi_request *sreq;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_finish_command() Begin .. \n"));

 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : &(cmd->device)=%08XH \n",
			&(cmd->device) )) ;
 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : cmd->device=%08XH \n",
			cmd->device )) ;
 sdev = cmd->device;

 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : &(sdev->host)=%08XH \n",
			&(sdev->host) )) ;
 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : sdev->host=%08XH \n",
			sdev->host )) ;
 shost = sdev->host;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_finish_command(1) : scsi_device_unbusy() .. \n")) ;
 scsi_device_unbusy(sdev);

 /*
  * Clear the flags which say that the device/host is no longer
  * capable of accepting new commands.  These are set in scsi_queue.c
  * for both the queue full condition on a device, and for a
  * host full condition on the host.
  *
  * XXX(hch): What about locking?
  */
 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : &(shost->host_blocked)=%08XH \n",
			&(shost->host_blocked) )) ;
 shost->host_blocked = 0;
 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : &(sdev->device_blocked)=%08XH \n",
			&(sdev->device_blocked) )) ;
 sdev->device_blocked = 0;

 /*
  * If we have valid sense information, then some kind of recovery
  * must have taken place.  Make a note of this.
  */
 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : &(cmd->sense_buffer)=%08XH \n",
			&(cmd->sense_buffer) )) ;
 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : cmd->sense_buffer=%08XH \n",
			cmd->sense_buffer )) ;

 #ifdef USBMS_BS_ASSERT
	if (!(   (unsigned long) cmd->sense_buffer
		   > (unsigned long) 0x200000L
	   ) )
	   {
		iprintf("[ASSERT][USBMS] scsi_finish_command() : " ) ;
	    iprintf("Error : Invlid address cmd->sense_buffer=%08XH \n",
				cmd->sense_buffer ) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT

 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : &(cmd->result)=%08XH \n",
			&(cmd->result) )) ;
 if (SCSI_SENSE_VALID(cmd))
		cmd->result |= (DRIVER_SENSE << 24);

 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : &(sdev->id)=%08XH \n",
			&(sdev->id) )) ;
 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : &(cmd->result)=%08XH \n",
			&(cmd->result) )) ;
 /*
 SCSI_LOG_MLCOMPLETE(3, dprintf("[USBMS] Notifying upper driver of completion "
 */
 USBMS_DBG_SCSI(iprintf(
	"[USBMS SCSI] scsi_finish_command() : &(cmd->result)=%08XH \n",
				"for device %d %x\n", sdev->id, cmd->result));

 cmd->owner = SCSI_OWNER_HIGHLEVEL;
 cmd->state = SCSI_STATE_FINISHED;

 /*
  * We can get here with use_sg=0, causing a panic in the upper level
  */
 /*
 #ifndef NO_SCATTER_LIST
	cmd->use_sg = cmd->old_use_sg;
 #endif // ndef NO_SCATTER_LIST
 */
 /*
  * If there is an associated request structure, copy the data over
  * before we call the completion function.
  */
 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : &(cmd->sc_request)=%08XH \n",
			&(cmd->sc_request) )) ;
 sreq = cmd->sc_request;
 if(sreq) {
	USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : &(sreq->sr_result)=%08XH \n",
			&(sreq->sr_result) )) ;
	USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : &(sreq->sr_command)=%08XH \n",
			&(sreq->sr_command) )) ;
	USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : &(sreq->sr_command->result)=%08XH \n",
			&(sreq->sr_command->result) )) ;
	if (sreq->sr_command) {
		sreq->sr_result = sreq->sr_command->result;
		if (sreq->sr_result) {
			USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : sreq->sr_sense_buffer=%08XH \n",
				sreq->sr_sense_buffer )) ;
			USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_finish_command() : sreq->sr_command->sense_buffer=%08XH \n",
				sreq->sr_command->sense_buffer )) ;
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_finish_command(2) : memcpy() .. \n"));
			memcpy(sreq->sr_sense_buffer,
					sreq->sr_command->sense_buffer,
				    sizeof(sreq->sr_sense_buffer));
	   	}
	}
  }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_finish_command(3) : cmd->done() , sd_rw_intr() .. \n"));
 cmd->done(cmd);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_finish_command() End. \n"));
}

/*
 * Function:	scsi_adjust_queue_depth()
 *
 * Purpose:	Allow low level drivers to tell us to change the queue depth
 * 		on a specific SCSI device
 *
 * Arguments:	sdev	- SCSI Device in question
 * 		tagged	- Do we use tagged queueing (non-0) or do we treat
 * 			  this device as an untagged device (0)
 * 		tags	- Number of tags allowed if tagged queueing enabled,
 * 			  or number of commands the low level driver can
 * 			  queue up in non-tagged mode (as per cmd_per_lun).
 *
 * Returns:	Nothing
 *
 * Lock Status:	None held on entry
 *
 * Notes:	Low level drivers may call this at any time and we will do
 * 		the right thing depending on whether or not the device is
 * 		currently active and whether or not it even has the
 * 		command blocks built yet.
 *
 * XXX(hch):	What exactly is device_request_lock trying to protect?
 */
void scsi_adjust_queue_depth(struct scsi_device *sdev, int tagged, int tags)
{
 /*
 #ifndef NO_SPINLOCK
	static spinlock_t device_request_lock = SPIN_LOCK_UNLOCKED;
 #endif // ndef NO_SPINLOCK
 */
 unsigned long flags;

 /*
  * refuse to set tagged depth to an unworkable size
  */
 if(tags <= 0)
		return;
 /*
  * Limit max queue depth on a single lun to 256 for now.  Remember,
  * we allocate a struct scsi_command for each of these and keep it
  * around forever.  Too deep of a depth just wastes memory.
  */
 if(tags > 256)
		return;
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(&device_request_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
	sdev->queue_depth = tags;
	switch(tagged) {
		case MSG_ORDERED_TAG:
			sdev->ordered_tags = 1;
			sdev->simple_tags = 1;
			break;
		case MSG_SIMPLE_TAG:
			sdev->ordered_tags = 0;
			sdev->simple_tags = 1;
			break;
		default:
			/*
			printk(KERN_WARNING "(scsi%d:%d:%d:%d) "
				"scsi_adjust_queue_depth, bad queue type, "
				"disabled\n", sdev->host->host_no,
				sdev->channel, sdev->id, sdev->lun); 
			*/
			dprintf("[USBMS] (scsi%d:%d:%d:%d) "
				"scsi_adjust_queue_depth, bad queue type, "
				"disabled\n", sdev->host->host_no,
				sdev->channel, sdev->id, sdev->lun); 
		case 0:
			sdev->ordered_tags = sdev->simple_tags = 0;
			sdev->queue_depth = tags;
			break;
	}
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(&device_request_lock, flags);
 #endif // ndef NO_SPINLOCK
}

/*
 * Function:	scsi_track_queue_full()
 *
 * Purpose:	This function will track successive QUEUE_FULL events on a
 * 		specific SCSI device to determine if and when there is a
 * 		need to adjust the queue depth on the device.
 *
 * Arguments:	sdev	- SCSI Device in question
 * 		depth	- Current number of outstanding SCSI commands on
 * 			  this device, not counting the one returned as
 * 			  QUEUE_FULL.
 *
 * Returns:	0 - No change needed
 * 		>0 - Adjust queue depth to this new depth
 * 		-1 - Drop back to untagged operation using host->cmd_per_lun
 * 			as the untagged command depth
 *
 * Lock Status:	None held on entry
 *
 * Notes:	Low level drivers may call this at any time and we will do
 * 		"The Right Thing."  We are interrupt context safe.
 */
int scsi_track_queue_full(struct scsi_device *sdev, int depth)
{
 #ifndef NO_TIMER
	if ((jiffies >> 4) == sdev->last_queue_full_time)
		return 0;

	sdev->last_queue_full_time = (jiffies >> 4);
	if (sdev->last_queue_full_depth != depth) {
		sdev->last_queue_full_count = 1;
		sdev->last_queue_full_depth = depth;
	} else {
		sdev->last_queue_full_count++;
	}

	if (sdev->last_queue_full_count <= 10)
		return 0;
	if (sdev->last_queue_full_depth < 8) {
				/* Drop back to untagged */
		scsi_adjust_queue_depth(sdev, 0, sdev->host->cmd_per_lun);
				return -1;
			}
	
	if (sdev->ordered_tags)
		scsi_adjust_queue_depth(sdev, MSG_ORDERED_TAG, depth);
			else
		scsi_adjust_queue_depth(sdev, MSG_SIMPLE_TAG, depth);
 #endif // ndef NO_TIMER
 return depth;
}

int scsi_device_get(struct scsi_device *sdev)
{
	struct class *class = class_get(&sdev_class);

	if (!class)
		goto out;
	if (test_bit(SDEV_DEL, &sdev->sdev_state))
		goto out;
	if (!try_module_get(sdev->host->hostt->module))
		goto out;
	if (!get_device(&sdev->sdev_gendev))
		goto out_put_module;
	atomic_inc(&sdev->access_count);
	class_put(&sdev_class);
	return 0;

 out_put_module:
	module_put(sdev->host->hostt->module);
 out:
	class_put(&sdev_class);
	return -ENXIO;
	}

void scsi_device_put(struct scsi_device *sdev)
{
 struct class *class = class_get(&sdev_class);


 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_put() Begin .. \n"));
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_put() : sdev=%08XH \n",sdev));
 if (!class)
		return;

 module_put(sdev->host->hostt->module);
 if (atomic_dec_and_test(&sdev->access_count))
		if (test_bit(SDEV_DEL, &sdev->sdev_state)) {
			USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_put() : &sdev->sdev_gendev=%08XH \n",
				&sdev->sdev_gendev));
			USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_put() : device_del() .. \n"));
			device_del(&sdev->sdev_gendev);
		   }
 // #ifndef _ETCPIP
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_put() : put_device() .. \n"));
 put_device(&sdev->sdev_gendev);
 // #endif // ndef _ETCPIP
 
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_put() : class_put() .. \n"));
 class_put(&sdev_class);
}

int scsi_device_cancel_cb(struct device *dev, void *data)
{
 struct scsi_device *sdev = to_scsi_device(dev);
 int recovery = *(int *)data;


 // USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_cancel_cb() Begin .. \n")) ;
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_cancel_cb() Begin .. \n")) ;
 USBMS_DBG_CALLBACK(iprintf("[USBMS CALLBACK] scsi_device_cancel_cb() : Begin .. \n"));

 USBMS_DBG_CALLBACK(iprintf("[USBMS CALLBACK] scsi_device_cancel_cb() scsi_device_cancel_cb() End with Call scsi_device_cancel() .. \n"));
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_cancel_cb() End with Call scsi_device_cancel() .. \n")) ;
 // USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_cancel_cb() End with Call scsi_device_cancel() .. \n")) ;
 return scsi_device_cancel(sdev, recovery);
}

/**
 * scsi_device_cancel - cancel outstanding IO to this device
 * @sdev:	pointer to struct scsi_device
 * @data:	pointer to cancel value.
 *
 **/
int scsi_device_cancel(struct scsi_device *sdev, int recovery)
{
 struct scsi_cmnd *scmd;
 LIST_HEAD(active_list);
 struct list_head *lh, *lh_sf;
 unsigned long flags;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_cancel() Begin .. \n"));
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_device_cancel() Begin .. \n"));
 set_bit(SDEV_CANCEL, &sdev->sdev_state);
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(&sdev->list_lock, flags);
 #endif // ndef NO_SPINLOCK
 */
 /*
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	list_for_each_entry(scmd, &sdev->cmd_list, list) {
  #else
 */
	USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_device_cancel() : Get sdev->cmd .. \n"));
	scmd=sdev->cmd ;
 /*
 #endif
 */

 if (scmd->request && scmd->request->rq_status != RQ_INACTIVE) {
		/*
		 * If we are unable to remove the timer, it means
		 * that the command has already timed out or
		 * finished.
		 */
		#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
			/*
			#ifndef NO_TIMER
			if (!scsi_delete_timer(scmd))
			   {
				continue;
			   }
			#endif // ndef NO_TIMER
			*/
		 #else
		/*
		#ifndef NO_TIMER
		if (scsi_delete_timer(scmd))
		#endif // ndef NO_TIMER
		*/
		#endif
		   {
			USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_device_cancel() : active_list=%08XH \n",
				active_list));
			#ifdef USBMS_BS_ASSERT
					if (!(   (unsigned long) &active_list
						   > (unsigned long) 0x200000L
				         )
					   )
					   {
						iprintf("[ASSERT][USBMS] scsi_device_cancel() : " ) ;
		    		    iprintf("Error : Invlid address &active_list=%08XH \n",
								&active_list) ;
					    iprintf("[ASSERT][USBMS] Continue ..\n") ;
    				   }
			#endif // USBMS_BS_ASSERT

			list_add_tail(&scmd->eh_entry, &active_list);
		   }
		   }
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	   }
 #endif // ndef NO_ELEVATOR
 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(&sdev->list_lock, flags);
 #endif // ndef NO_SPINLOCK
 */

 if(!list_empty(&active_list)) {
	list_for_each_safe(lh, lh_sf, &active_list) {
		scmd = list_entry(lh, struct scsi_cmnd, eh_entry);
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_cancel() : list_del_init() .. \n"));
		list_del_init(lh);
		if (recovery) {
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_cancel() : scsi_eh_scmd_add() .. \n")) ;
			scsi_eh_scmd_add(scmd, SCSI_EH_CANCEL_CMD);
		} else {
			scmd->result = (DID_ABORT << 16);
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_cancel() : scsi_finish_command() .. \n"));
			scsi_finish_command(scmd);
		}
	   }
   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_device_cancel() End. \n"));
 return 0;
}

/*
#ifndef _ETCPIP
MODULE_DESCRIPTION("SCSI core");
MODULE_LICENSE("GPL");

module_param(scsi_logging_level, int, S_IRUGO|S_IWUSR);
MODULE_PARM_DESC(scsi_logging_level, "a bit mask of logging levels");

#endif // ndef _ETCPIP
*/

/*
static int __init init_scsi(void)
*/
int init_scsi(void)
{
 int error, i;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] init_scsi() Begin .. \n"));

 #ifdef CONFIG_SCSI_LOGGING
	USBMS_DBG_INFO(iprintf("[USBMS INFO] init_scsi() : Set scsi_logging_level .. \n"));
	scsi_logging_level=0xFFFFFFFF;
 #endif // CONFIG_SCSI_LOGGING

 error = scsi_init_queue();
 if (error)
		return error;
 error = scsi_init_procfs();
	if (error)
		goto cleanup_queue;
 error = scsi_init_devinfo();
	if (error)
		goto cleanup_procfs;
 error = scsi_init_hosts();
	if (error)
		goto cleanup_devlist;
 error = scsi_init_sysctl();
	if (error)
		goto cleanup_hosts;

 // #ifndef NO_SYS_FS
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] init_scsi() : scsi_sysfs_register() .. \n"));
 error = scsi_sysfs_register();
 if (error)
		goto cleanup_sysctl;
 // #endif // ndef NO_SYS_FS
 /*
 #ifndef _ETCPIP
	for (i = 0; i < NR_CPUS; i++)
		INIT_LIST_HEAD(&per_cpu(scsi_done_q, i));
 #endif // ndef _ETCPIP
 */

 #ifdef CONFIG_USB_DEVICEFS
	devfs_mk_dir("scsi");
 #endif // CONFIG_USB_DEVICEFS
 /* SCSI_SOFTIRQ is on include/linux/interrupt.h
	open_softirq(SCSI_SOFTIRQ, scsi_softirq, NULL);
 */

 /*
 printk(KERN_NOTICE "SCSI subsystem initialized\n");
 */
 dprintf("[USBMS] SCSI subsystem initialized\n");
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] init_scsi() End. \n"));
 return 0;

 cleanup_sysctl:
	scsi_exit_sysctl();
 cleanup_hosts:
	scsi_exit_hosts();
 cleanup_devlist:
	scsi_exit_devinfo();
 cleanup_procfs:
	scsi_exit_procfs();
 cleanup_queue:
	scsi_exit_queue();
 /*
 printk(KERN_ERR "SCSI subsystem failed to initialize, error = %d\n",
	       -error);
 */
 dprintf("[USBMS] ERROR : SCSI subsystem failed to initialize, error = %d\n",
	       -error);
 
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] init_scsi() End with Fail !! \n"));
 return error;
}

static void __exit exit_scsi(void)
{
 // #ifndef NO_SYS_FS
	scsi_sysfs_unregister();
 // #endif // ndef NO_SYS_FS
	scsi_exit_sysctl();
	scsi_exit_hosts();
	scsi_exit_devinfo();
 #ifdef CONFIG_USB_DEVICEFS
	devfs_remove("scsi");
 #endif // CONFIG_USB_DEVICEFS
	scsi_exit_procfs();
	scsi_exit_queue();
}

/*
subsys_initcall(init_scsi);
module_exit(exit_scsi);
*/

// ====================================================================
// End of File : scsi_mid.c

