/* 
 *  linux/kernel/scsi_debug.c
 * vvvvvvvvvvvvvvvvvvvvvvv Original vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
 *  Copyright (C) 1992  Eric Youngdale
 *  Simulate a host adapter with 2 disks attached.  Do a lot of checking
 *  to make sure that we are not getting blocks mixed up, and PANIC if
 *  anything out of the ordinary is seen.
 * ^^^^^^^^^^^^^^^^^^^^^^^ Original ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 *
 *  This version is more generic, simulating a variable number of disk 
 *  (or disk like devices) sharing a common amount of RAM 
 *  
 *
 *  For documentation see http://www.torque.net/sg/sdebug25.html
 *
 *   D. Gilbert (dpg) work for Magneto-Optical device test [20010421]
 *   dpg: work for devfs large number of disks [20010809]
 *        forked for lk 2.5 series [20011216, 20020101]
 *        use vmalloc() more inquiry+mode_sense [20020302]
 *        add timers for delayed responses [20020721]
 *   Patrick Mansfield <patmans@us.ibm.com> max_luns+scsi_level [20021031]
 *   Mike Anderson <andmike@us.ibm.com> sysfs work [20021118]
 *   dpg: change style of boot options to "scsi_debug.num_tgts=2" and
 *        module options to "modprobe scsi_debug num_tgts=2" [20021221]
 */

// #ifdef _ETCPIP
#define __KERNEL__
        #ifdef _NEWLIB
          #include "string.h"
        #endif
   	#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsbMsInit.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP

/*
#include <linux/config.h>
#include <linux/module.h>
*/
/*
#include <linux/kernel.h>
*/
#include "PoKernel.h"
/*
#include <linux/sched.h>
*/
/*
#ifndef _ETCPIP
#include <linux/errno.h>
*/
#include "PoErrno.h" // osport_errno.h
/*
#include <linux/timer.h>
*/
#ifndef NO_TIMER
	#include "PoTimer.h" // osport_timer.h
#endif // ndef NO_TIMER

/*
#include <linux/types.h>
*/
/*
#include <linux/string.h>
*/
#include "PoString.h" // linux/string.h

/*
#include <linux/genhd.h>
*/
#include "block/genhd.h"

/*
#include <linux/fs.h>
*/
#include "PoFile.h" // include/linux/file.h
#include "radix-tree.h" // include/linux/radix-tree.h
#include "PoFs.h" // osport_fs.h

/*
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/moduleparam.h>
#include <linux/blkdev.h>
*/
#include "bitops.h" // B. S. Liaw add for test_bit()
#include "blkdev.h"

/*
#include "scsi.h"
*/
#include "scsi_ds.h"
/*
#include "hosts.h"
*/
#include "scsi_hosts.h"
#ifndef _ETCPIP
	#include <scsi/scsicam.h>
 #else
	#include "scsi_cam.h"
#endif
/*
#include <linux/stat.h>
*/
#include "stat.h"

/*
#ifndef _ETCPIP
	#ifndef LINUX_VERSION_CODE
	#include <linux/version.h>
	#endif
#endif // ndef _ETCPIP
*/

#include "scsi_logging.h"
#include "scsi_debug.h"



static const char * scsi_debug_version_str = "Version: 1.70 (20030507)";

/* Additional Sense Code (ASC) used */
#define NO_ADDED_SENSE 0x0
#define UNRECOVERED_READ_ERR 0x11
#define INVALID_OPCODE 0x20
#define ADDR_OUT_OF_RANGE 0x21
#define INVALID_FIELD_IN_CDB 0x24
#define POWERON_RESET 0x29
#define SAVING_PARAMS_UNSUP 0x39
#define THRESHHOLD_EXCEEDED 0x5d

#define SDEBUG_TAGGED_QUEUING 0 /* 0 | MSG_SIMPLE_TAG | MSG_ORDERED_TAG */

/* Default values for driver parameters */
#define DEF_NUM_HOST   1
#define DEF_NUM_TGTS   1
#define DEF_MAX_LUNS   1
/* With these defaults, this driver will make 1 host with 1 target
 * (id 0) containing 1 logical unit (lun 0). That is 1 device.
 */
#ifdef USBMS_REDUCE_RESOURCE
	#define DEF_DEV_SIZE_KB   4
 #else
	#define DEF_DEV_SIZE_MB   8
#endif

#define DEF_EVERY_NTH   0
#define DEF_DELAY   1
#define DEF_SCSI_LEVEL   3
/*
#define DEF_OPTS   0
*/
// B. S. Liaw
#define DEF_OPTS   0x0F

/* bit mask values for scsi_debug_opts */
#define SCSI_DEBUG_OPT_NOISE          1
#define SCSI_DEBUG_OPT_MEDIUM_ERR     2
#define SCSI_DEBUG_OPT_TIMEOUT        4
#define SCSI_DEBUG_OPT_RECOVERED_ERR  8

/* When "every_nth" > 0 then modulo "every_nth" commands:
 *   - a no response is simulated if SCSI_DEBUG_OPT_TIMEOUT is set
 *   - a RECOVERED_ERROR is simulated on successful read and write
 *     commands if SCSI_DEBUG_OPT_RECOVERED_ERR is set.
 */

/* when 1==SCSI_DEBUG_OPT_MEDIUM_ERR, a medium error is simulated at this
 * sector on read commands: */
#define OPT_MEDIUM_ERR_ADDR   0x1234 /* that's sector 4660 in decimal */

/* If REPORT LUNS has luns >= 256 it can choose "flat space" (value 1)
 * or "peripheral device" addressing (value 0) */
#define SAM2_LUN_ADDRESS_METHOD 0

#ifdef USBMS_REDUCE_RESOURCE
	static int scsi_debug_dev_size_kb = DEF_DEV_SIZE_KB;
 #else
	static int scsi_debug_dev_size_mb = DEF_DEV_SIZE_MB;
#endif
static int scsi_debug_num_tgts = DEF_NUM_TGTS; /* targets per host */
static int scsi_debug_opts = DEF_OPTS;
static int scsi_debug_every_nth = DEF_EVERY_NTH;
static int scsi_debug_cmnd_count = 0;
static int scsi_debug_delay = DEF_DELAY;
static int scsi_debug_max_luns = DEF_MAX_LUNS;
static int scsi_debug_scsi_level = DEF_SCSI_LEVEL;
static int scsi_debug_add_host = DEF_NUM_HOST;

#define DEV_READONLY(TGT)      (0)
#define DEV_REMOVEABLE(TGT)    (0)
#define PERIPH_DEVICE_TYPE(TGT) (TYPE_DISK);

static unsigned long sdebug_store_size;	/* in bytes */
static sector_t sdebug_capacity;	/* in sectors */

/* old BIOS stuff, kernel may get rid of them but some mode sense pages
   may still need them */
static int sdebug_heads;		/* heads per disk */
static int sdebug_cylinders_per;	/* cylinders per surface */
static int sdebug_sectors_per;		/* sectors per cylinder */

/* default sector size is 512 bytes, 2**9 bytes */
#define POW2_SECT_SIZE 9
#define SECT_SIZE (1 << POW2_SECT_SIZE)
#define SECT_SIZE_PER(TGT) SECT_SIZE

#define SDEBUG_SENSE_LEN 32

struct sdebug_dev_info {
	struct list_head dev_list;
	unsigned char sense_buff[SDEBUG_SENSE_LEN];	/* weak nexus */
	unsigned int channel;
	unsigned int target;
	unsigned int lun;
	struct sdebug_host_info *sdbg_host;
	char reset;
	char used;
};

struct sdebug_host_info {
	struct list_head host_list;
	struct Scsi_Host *shost;
	struct device dev;
	struct list_head dev_info_list;
};

#define to_sdebug_host(d)	\
	container_of(d, struct sdebug_host_info, dev)

static LIST_HEAD(sdebug_host_list);


// #ifdef _ETCPIP
// B. S. Liaw Add these ..
static LIST_HEAD(sdebug_bus_drvlist);
static LIST_HEAD(sdebug_bus_devlist);
static LIST_HEAD(scsidbg_shost_my_devices);
// #endif // _ETCPIP


/*
#ifndef NO_SPINLOCK
	static spinlock_t sdebug_host_list_lock = SPIN_LOCK_UNLOCKED;
#endif // ndef NO_SPINLOCK
*/
typedef void (* done_funct_t) (struct scsi_cmnd *);

struct sdebug_queued_cmd {
	int in_use;
 #ifndef NO_TIMER
	struct timer_list cmnd_timer;
 #endif // ndef NO_TIMER
	done_funct_t done_funct;
	struct scsi_cmnd * a_cmnd;
	int scsi_result;
};
static struct sdebug_queued_cmd queued_arr[SCSI_DEBUG_CANQUEUE];

static Scsi_Host_Template sdebug_driver_template = {
 .proc_info =		scsi_debug_proc_info,
 
 #ifdef USB_DBGSTRING_ON
	.name =			"SCSI DEBUG",
 #endif // USB_DBGSTRING_ON 
 
	.info =			scsi_debug_info,
	.slave_alloc =		scsi_debug_slave_alloc,
	.slave_configure =	scsi_debug_slave_configure,
	.slave_destroy =	scsi_debug_slave_destroy,
	.ioctl =		scsi_debug_ioctl,
	.queuecommand =		scsi_debug_queuecommand,
	.eh_abort_handler =	scsi_debug_abort,
	.eh_bus_reset_handler = scsi_debug_bus_reset,
	.eh_device_reset_handler = scsi_debug_device_reset,
	.eh_host_reset_handler = scsi_debug_host_reset,
	.bios_param =		scsi_debug_biosparam,
	.can_queue =		SCSI_DEBUG_CANQUEUE,
	.this_id =		7,
	.sg_tablesize =		64,
	.cmd_per_lun =		3,
	.max_sectors =		4096,
	.unchecked_isa_dma = 	0,
	.use_clustering = 	ENABLE_CLUSTERING,
 /*
 #ifndef _ETCPIP
	.module =		THIS_MODULE,
 #endif // ndef _ETCPIP
	*/
};

static unsigned char * fake_storep;	/* ramdisk storage */

static unsigned char spare_buff[SDEBUG_SENSE_LEN];

static int num_aborts = 0;
static int num_dev_resets = 0;
static int num_bus_resets = 0;
static int num_host_resets = 0;


 #ifndef NO_SPINLOCK
	static spinlock_t queued_arr_lock = SPIN_LOCK_UNLOCKED;

	// rwlock_t : on spinlock.h
	static rwlock_t atomic_rw = RW_LOCK_UNLOCKED;
#endif // ndef NO_SPINLOCK

static char sdebug_proc_name[] = "scsi_debug";

static int sdebug_driver_probe(struct device *);
static int sdebug_driver_remove(struct device *);
static struct bus_type pseudo_lld_bus;

static struct device_driver sdebug_driverfs_driver = {
 
 #ifdef USB_DBGSTRING_ON
	.name 		= sdebug_proc_name,
 #endif // USB_DBGSTRING_ON 
 
 .bus		= &pseudo_lld_bus,
 .probe          = sdebug_driver_probe,
 .remove         = sdebug_driver_remove,
};

static const int check_condition_result = 
		(DRIVER_SENSE << 24) | SAM_STAT_CHECK_CONDITION;

/* function declarations */
static int resp_inquiry(unsigned char * cmd, int target, unsigned char * buff,
			int bufflen, struct sdebug_dev_info * devip);
static int resp_mode_sense(unsigned char * cmd, int target,
			   unsigned char * buff, int bufflen,
			   struct sdebug_dev_info * devip);
static int resp_read(struct scsi_cmnd * SCpnt, int upper_blk, int block, 
		     int num, struct sdebug_dev_info * devip);
static int resp_write(struct scsi_cmnd * SCpnt, int upper_blk, int block,
		      int num, struct sdebug_dev_info * devip);
static int resp_report_luns(unsigned char * cmd, unsigned char * buff,
			    int bufflen, struct sdebug_dev_info * devip);
static void timer_intr_handler(unsigned long);
static struct sdebug_dev_info * devInfoReg(struct scsi_device * sdev);
static void mk_sense_buffer(struct sdebug_dev_info * devip, int key, 
			    int asc, int asq, int inbandLen);
static int check_reset(struct scsi_cmnd * SCpnt, 
		       struct sdebug_dev_info * devip);
static int schedule_resp(struct scsi_cmnd * cmnd, 
			 struct sdebug_dev_info * devip, 
			 done_funct_t done, int scsi_result, int delta_jiff);
static void init_all_queued(void);
static void stop_all_queued(void);
static int stop_queued_cmnd(struct scsi_cmnd * cmnd);
static int inquiry_evpd_83(unsigned char * arr, int dev_id_num,
                           const char * dev_id_str, int dev_id_str_len);
static void do_create_driverfs_files(void);
static void do_remove_driverfs_files(void);

static int sdebug_add_adapter(void);
static void sdebug_remove_adapter(void);
static struct device pseudo_primary;
static struct bus_type pseudo_lld_bus;

/*
#ifndef NO_SCATTER_LIST
   static unsigned char * scatg2virt(const struct scatterlist * sclp)
   {
	if (NULL == sclp)
		return NULL;
 #ifndef USB_NO_VMM // vitrtul memory management
	// struct page is on include/linux/mm.h
	else if (sclp->page)
		return (unsigned char *)page_address(sclp->page) + 
		       sclp->offset;
 #endif // ndef USB_NO_VMM
	else
		return NULL;
}
#endif // ndef NO_SCATTER_LIST
*/
static
int scsi_debug_queuecommand(struct scsi_cmnd * SCpnt, done_funct_t done)
{
 unsigned char *cmd = (unsigned char *) SCpnt->cmnd;
 int block, upper_blk, num;
 unsigned char *buff;
 int errsts = 0;
 int target = SCpnt->device->id;
 int bufflen = SCpnt->request_bufflen;
 unsigned long capac;
 struct sdebug_dev_info * devip = NULL;
 unsigned char * sbuff;
 int inj_recovered = 0;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_debug_queuecommand() Begin .. \n"));

 if (done == NULL)
		return 0;	/* assume mid level reprocessing command */

 #ifndef NO_SCATTER_LIST
	if (SCpnt->use_sg) { /* just use first element */
		struct scatterlist *sgpnt = (struct scatterlist *)
						SCpnt->request_buffer;

		buff = scatg2virt(&sgpnt[0]);
		bufflen = sgpnt[0].length;
		/* READ and WRITE process scatterlist themselves */
	}
	else 
 #endif // ndef NO_SCATTER_LIST
	  {
		buff = (unsigned char *) SCpnt->request_buffer;
	  }
 if (NULL == buff) {
		buff = spare_buff;	/* assume cmd moves no data */
		bufflen = SDEBUG_SENSE_LEN;
	}


 if(target == sdebug_driver_template.this_id) {
	/*
	printk(KERN_INFO "scsi_debug: initiator's id used as "
		       "target!\n");
	*/
    dprintf("[USBMS] scsi_debug: initiator's id used as target!\n");
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_debug_queuecommand(1) End with schedule_resp() .. \n"));
	return schedule_resp(SCpnt, NULL, done,
				     DID_NO_CONNECT << 16, 0);
   }

 if(SCpnt->device->lun >= scsi_debug_max_luns)
   { 
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_debug_queuecommand(2) End with schedule_resp() .. \n"));
	return schedule_resp(SCpnt, NULL, done, 
				     DID_NO_CONNECT << 16, 0);
   }
 devip = devInfoReg(SCpnt->device);
 if (NULL == devip)
    {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_debug_queuecommand(3) End with schedule_resp() .. \n"));
	return schedule_resp(SCpnt, NULL, done, 
				     DID_NO_CONNECT << 16, 0);
   }
if ((scsi_debug_every_nth > 0) &&
            (++scsi_debug_cmnd_count >= scsi_debug_every_nth)) {
                scsi_debug_cmnd_count =0;
		if (SCSI_DEBUG_OPT_TIMEOUT & scsi_debug_opts)
			return 0; /* ignore command causing timeout */
		else if (SCSI_DEBUG_OPT_RECOVERED_ERR & scsi_debug_opts)
			inj_recovered = 1; /* to reads and writes below */
        }

 switch (*cmd) {
	case INQUIRY:     /* mandatory */
		errsts = resp_inquiry(cmd, target, buff, bufflen, devip);
		break;
	case REQUEST_SENSE:	/* mandatory */
		/* Since this driver indicates autosense by placing the
		 * sense buffer in the scsi_cmnd structure in the response
		 * (when SAM_STAT_CHECK_CONDITION is set), the mid level
		 * shouldn't need to call REQUEST_SENSE */
		if (devip) {
			sbuff = devip->sense_buff;
			memcpy(buff, sbuff, (bufflen < SDEBUG_SENSE_LEN) ? 
					     bufflen : SDEBUG_SENSE_LEN);
			mk_sense_buffer(devip, 0, NO_ADDED_SENSE, 0, 7);
		} else {
			memset(buff, 0, bufflen);
			buff[0] = 0x70;
		}
		break;
	case START_STOP:
		errsts = check_reset(SCpnt, devip);
		break;
	case ALLOW_MEDIUM_REMOVAL:
		if ((errsts = check_reset(SCpnt, devip)))
			break;
		if (SCSI_DEBUG_OPT_NOISE & scsi_debug_opts)
			printk("\tMedium removal %s\n",
			       cmd[4] ? "inhibited" : "enabled");
		break;
	case SEND_DIAGNOSTIC:     /* mandatory */
		memset(buff, 0, bufflen);
		break;
	case TEST_UNIT_READY:     /* mandatory */
		memset(buff, 0, bufflen);
		break;
        case RESERVE:
		errsts = check_reset(SCpnt, devip);
		memset(buff, 0, bufflen);
                break;
        case RESERVE_10:
		errsts = check_reset(SCpnt, devip);
		memset(buff, 0, bufflen);
                break;
        case RELEASE:
		errsts = check_reset(SCpnt, devip);
		memset(buff, 0, bufflen);
                break;
        case RELEASE_10:
		errsts = check_reset(SCpnt, devip);
		memset(buff, 0, bufflen);
                break;
	case READ_CAPACITY:
		errsts = check_reset(SCpnt, devip);
		memset(buff, 0, bufflen);
		if (bufflen > 7) {
			capac = (unsigned long)sdebug_capacity - 1;
			buff[0] = (capac >> 24);
			buff[1] = (capac >> 16) & 0xff;
			buff[2] = (capac >> 8) & 0xff;
			buff[3] = capac & 0xff;
			buff[6] = (SECT_SIZE_PER(target) >> 8) & 0xff;
			buff[7] = SECT_SIZE_PER(target) & 0xff;
		}
		break;
	case READ_16:
	case READ_12:
	case READ_10:
	case READ_6:
		if ((errsts = check_reset(SCpnt, devip)))
			break;
		upper_blk = 0;
		if ((*cmd) == READ_16) {
			upper_blk = cmd[5] + (cmd[4] << 8) + 
				    (cmd[3] << 16) + (cmd[2] << 24);
			block = cmd[9] + (cmd[8] << 8) + 
				(cmd[7] << 16) + (cmd[6] << 24);
			num = cmd[13] + (cmd[12] << 8) + 
				(cmd[11] << 16) + (cmd[10] << 24);
		} else if ((*cmd) == READ_12) {
			block = cmd[5] + (cmd[4] << 8) + 
				(cmd[3] << 16) + (cmd[2] << 24);
			num = cmd[9] + (cmd[8] << 8) + 
				(cmd[7] << 16) + (cmd[6] << 24);
		} else if ((*cmd) == READ_10) {
			block = cmd[5] + (cmd[4] << 8) + 
				(cmd[3] << 16) + (cmd[2] << 24);
			num = cmd[8] + (cmd[7] << 8);
		} else {
			block = cmd[3] + (cmd[2] << 8) + 
				((cmd[1] & 0x1f) << 16);
			num = cmd[4];
		}
		errsts = resp_read(SCpnt, upper_blk, block, num, devip);
		if (inj_recovered && (0 == errsts)) {
			mk_sense_buffer(devip, RECOVERED_ERROR, 
					THRESHHOLD_EXCEEDED, 0, 18);
			errsts = check_condition_result;
		}
		break;
	case REPORT_LUNS:
		errsts = resp_report_luns(cmd, buff, bufflen, devip);
		break;
	case WRITE_16:
	case WRITE_12:
	case WRITE_10:
	case WRITE_6:
		if ((errsts = check_reset(SCpnt, devip)))
			break;
		upper_blk = 0;
		if ((*cmd) == WRITE_16) {
			upper_blk = cmd[5] + (cmd[4] << 8) + 
				    (cmd[3] << 16) + (cmd[2] << 24);
			block = cmd[9] + (cmd[8] << 8) + 
				(cmd[7] << 16) + (cmd[6] << 24);
			num = cmd[13] + (cmd[12] << 8) + 
				(cmd[11] << 16) + (cmd[10] << 24);
		} else if ((*cmd) == WRITE_12) {
			block = cmd[5] + (cmd[4] << 8) + 
				(cmd[3] << 16) + (cmd[2] << 24);
			num = cmd[9] + (cmd[8] << 8) + 
				(cmd[7] << 16) + (cmd[6] << 24);
		} else if ((*cmd) == WRITE_10) {
			block = cmd[5] + (cmd[4] << 8) + 
				(cmd[3] << 16) + (cmd[2] << 24);
			num = cmd[8] + (cmd[7] << 8);
		} else {
			block = cmd[3] + (cmd[2] << 8) + 
				((cmd[1] & 0x1f) << 16);
			num = cmd[4];
		}
		errsts = resp_write(SCpnt, upper_blk, block, num, devip);
		if (inj_recovered && (0 == errsts)) {
			mk_sense_buffer(devip, RECOVERED_ERROR, 
					THRESHHOLD_EXCEEDED, 0, 18);
			errsts = check_condition_result;
		}
		break;
	case MODE_SENSE:
	case MODE_SENSE_10:
		errsts = resp_mode_sense(cmd, target, buff, bufflen, devip);
		break;
	case SYNCHRONIZE_CACHE:
		memset(buff, 0, bufflen);
		break;
	default:
		if ((errsts = check_reset(SCpnt, devip)))
			break;
		mk_sense_buffer(devip, ILLEGAL_REQUEST, INVALID_OPCODE, 0, 18);
		errsts = check_condition_result;
		break;
	}

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_debug_queuecommand(4) End with schedule_resp() .. \n"));
 return schedule_resp(SCpnt, devip, done, errsts, scsi_debug_delay);
}

static int scsi_debug_ioctl(struct scsi_device *dev, int cmd, void *arg)
{
	if (SCSI_DEBUG_OPT_NOISE & scsi_debug_opts) {
		/*
		printk(KERN_INFO "scsi_debug: ioctl: cmd=0x%x\n", cmd);
		*/
		dprintf("[USBMS] scsi_debug: ioctl: cmd=0x%x\n", cmd);
	}
	return -EINVAL;
	/* return -ENOTTY; // correct return but upsets fdisk */
}

static int check_reset(struct scsi_cmnd * SCpnt, struct sdebug_dev_info * devip)
{
	if (devip->reset) {
		devip->reset = 0;
		mk_sense_buffer(devip, UNIT_ATTENTION, POWERON_RESET, 0, 18);
		return check_condition_result;
	}
	return 0;
}

#define SDEBUG_LONG_INQ_SZ 58
#define SDEBUG_MAX_INQ_ARR_SZ 128

static const char * vendor_id = "Linux   ";
static const char * product_id = "scsi_debug      ";
static const char * product_rev = "0004";

static int inquiry_evpd_83(unsigned char * arr, int dev_id_num, 
			   const char * dev_id_str, int dev_id_str_len)
{
	int num;

	/* Two identification descriptors: */
	/* T10 vendor identifier field format (faked) */
	arr[0] = 0x2;	/* ASCII */
	arr[1] = 0x1;
	arr[2] = 0x0;
	memcpy(&arr[4], vendor_id, 8);
	memcpy(&arr[12], product_id, 16);
	memcpy(&arr[28], dev_id_str, dev_id_str_len);
	num = 8 + 16 + dev_id_str_len;
	arr[3] = num;
	num += 4;
	/* NAA IEEE registered identifier (faked) */
	arr[num] = 0x1;	/* binary */
	arr[num + 1] = 0x3;
	arr[num + 2] = 0x0;
	arr[num + 3] = 0x8;
	arr[num + 4] = 0x51;	/* ieee company id=0x123456 (faked) */
	arr[num + 5] = 0x23;
	arr[num + 6] = 0x45;
	arr[num + 7] = 0x60;
	arr[num + 8] = (dev_id_num >> 24);
	arr[num + 9] = (dev_id_num >> 16) & 0xff;
	arr[num + 10] = (dev_id_num >> 8) & 0xff;
	arr[num + 11] = dev_id_num & 0xff;
	return num + 12;
}

static int resp_inquiry(unsigned char * cmd, int target, unsigned char * buff,
			int bufflen, struct sdebug_dev_info * devip)
{
	unsigned char pq_pdt;
	unsigned char arr[SDEBUG_MAX_INQ_ARR_SZ];
	int min_len = bufflen > SDEBUG_MAX_INQ_ARR_SZ ? 
			SDEBUG_MAX_INQ_ARR_SZ : bufflen;

	if (bufflen < cmd[4])
	   {
		/*
		printk(KERN_INFO "scsi_debug: inquiry: bufflen=%d "
		       "< alloc_length=%d\n", bufflen, (int)cmd[4]);
		*/
		dprintf("[USBMS] scsi_debug: inquiry: bufflen=%d "
		       "< alloc_length=%d\n", bufflen, (int)cmd[4]);
	   }
	memset(buff, 0, bufflen);
	memset(arr, 0, SDEBUG_MAX_INQ_ARR_SZ);
	pq_pdt = PERIPH_DEVICE_TYPE(target);
	arr[0] = pq_pdt;
	if (0x2 & cmd[1]) {  /* CMDDT bit set */
		mk_sense_buffer(devip, ILLEGAL_REQUEST, INVALID_FIELD_IN_CDB,
			       	0, 18);
		return check_condition_result;
	} else if (0x1 & cmd[1]) {  /* EVPD bit set */
		int dev_id_num, len;
		char dev_id_str[6];
		
		dev_id_num = ((devip->sdbg_host->shost->host_no + 1) * 2000) +
			     (devip->target * 1000) + devip->lun;
	//#ifndef USB_NO_LIB
		len = snprintf(dev_id_str, 6, "%d", dev_id_num);
	// #else
	//	dprintf("%d", dev_id_num);
	//#endif // ndef USB_NO_LIB
		len = (len > 6) ? 6 : len;
		if (0 == cmd[2]) { /* supported vital product data pages */
			arr[3] = 3;
			arr[4] = 0x0; /* this page */
			arr[5] = 0x80; /* unit serial number */
			arr[6] = 0x83; /* device identification */
		} else if (0x80 == cmd[2]) { /* unit serial number */
			arr[1] = 0x80;
			arr[3] = len;
			memcpy(&arr[4], dev_id_str, len);
		} else if (0x83 == cmd[2]) { /* device identification */
			arr[1] = 0x83;
			arr[3] = inquiry_evpd_83(&arr[4], dev_id_num,
						 dev_id_str, len);
		} else {
			/* Illegal request, invalid field in cdb */
			mk_sense_buffer(devip, ILLEGAL_REQUEST, 
					INVALID_FIELD_IN_CDB, 0, 18);
			return check_condition_result;
		}
		memcpy(buff, arr, min_len); 
		return 0;
	}
	/* drops through here for a standard inquiry */
	arr[1] = DEV_REMOVEABLE(target) ? 0x80 : 0;	/* Removable disk */
	arr[2] = scsi_debug_scsi_level;
	arr[4] = SDEBUG_LONG_INQ_SZ - 5;
	arr[7] = 0x3a; /* claim: WBUS16, SYNC, LINKED + CMDQUE */
	memcpy(&arr[8], vendor_id, 8);
	memcpy(&arr[16], product_id, 16);
	memcpy(&arr[32], product_rev, 4);
	memcpy(buff, arr, min_len);
	return 0;
}

/* <<Following mode page info copied from ST318451LW>> */ 

static int resp_err_recov_pg(unsigned char * p, int pcontrol, int target)
{	/* Read-Write Error Recovery page for mode_sense */
	unsigned char err_recov_pg[] = {0x1, 0xa, 0xc0, 11, 240, 0, 0, 0, 
					5, 0, 0xff, 0xff};

	memcpy(p, err_recov_pg, sizeof(err_recov_pg));
	if (1 == pcontrol)
		memset(p + 2, 0, sizeof(err_recov_pg) - 2);
	return sizeof(err_recov_pg);
}

static int resp_disconnect_pg(unsigned char * p, int pcontrol, int target)
{ 	/* Disconnect-Reconnect page for mode_sense */
	unsigned char disconnect_pg[] = {0x2, 0xe, 128, 128, 0, 10, 0, 0, 
					 0, 0, 0, 0, 0, 0, 0, 0};

	memcpy(p, disconnect_pg, sizeof(disconnect_pg));
	if (1 == pcontrol)
		memset(p + 2, 0, sizeof(disconnect_pg) - 2);
	return sizeof(disconnect_pg);
}

static int resp_format_pg(unsigned char * p, int pcontrol, int target)
{       /* Format device page for mode_sense */
        unsigned char format_pg[] = {0x3, 0x16, 0, 0, 0, 0, 0, 0,
                                     0, 0, 0, 0, 0, 0, 0, 0,
                                     0, 0, 0, 0, 0x40, 0, 0, 0};

        memcpy(p, format_pg, sizeof(format_pg));
        p[10] = (sdebug_sectors_per >> 8) & 0xff;
        p[11] = sdebug_sectors_per & 0xff;
        p[12] = (SECT_SIZE >> 8) & 0xff;
        p[13] = SECT_SIZE & 0xff;
        if (DEV_REMOVEABLE(target))
                p[20] |= 0x20; /* should agree with INQUIRY */
        if (1 == pcontrol)
                memset(p + 2, 0, sizeof(format_pg) - 2);
        return sizeof(format_pg);
}

static int resp_caching_pg(unsigned char * p, int pcontrol, int target)
{ 	/* Caching page for mode_sense */
	unsigned char caching_pg[] = {0x8, 18, 0x14, 0, 0xff, 0xff, 0, 0, 
		0xff, 0xff, 0xff, 0xff, 0x80, 0x14, 0, 0,     0, 0, 0, 0};

	memcpy(p, caching_pg, sizeof(caching_pg));
	if (1 == pcontrol)
		memset(p + 2, 0, sizeof(caching_pg) - 2);
	return sizeof(caching_pg);
}

static int resp_ctrl_m_pg(unsigned char * p, int pcontrol, int target)
{ 	/* Control mode page for mode_sense */
	unsigned char ctrl_m_pg[] = {0xa, 10, 2, 0, 0, 0, 0, 0,
				     0, 0, 0x2, 0x4b};

	memcpy(p, ctrl_m_pg, sizeof(ctrl_m_pg));
	if (1 == pcontrol)
		memset(p + 2, 0, sizeof(ctrl_m_pg) - 2);
	return sizeof(ctrl_m_pg);
}

static int resp_iec_m_pg(unsigned char * p, int pcontrol, int target)
{	/* Informational Exceptions control mode page for mode_sense */
	unsigned char iec_m_pg[] = {0x1c, 0xa, 0x08, 0, 0, 0, 0, 0,
				    0, 0, 0x0, 0x0};
	memcpy(p, iec_m_pg, sizeof(iec_m_pg));
	if (1 == pcontrol)
		memset(p + 2, 0, sizeof(iec_m_pg) - 2);
	return sizeof(iec_m_pg);
}

#define SDEBUG_MAX_MSENSE_SZ 256

static int resp_mode_sense(unsigned char * cmd, int target,
			   unsigned char * buff, int bufflen,
			   struct sdebug_dev_info * devip)
{
	unsigned char dbd;
	int pcontrol, pcode;
	unsigned char dev_spec;
	int alloc_len, msense_6, offset, len;
	unsigned char * ap;
	unsigned char arr[SDEBUG_MAX_MSENSE_SZ];
	int min_len = bufflen > SDEBUG_MAX_MSENSE_SZ ? 
			SDEBUG_MAX_MSENSE_SZ : bufflen;

	SCSI_LOG_LLQUEUE(3, printk("Mode sense ...(%p %d)\n", buff, bufflen));
	dbd = cmd[1] & 0x8;
	pcontrol = (cmd[2] & 0xc0) >> 6;
	pcode = cmd[2] & 0x3f;
	msense_6 = (MODE_SENSE == cmd[0]);
	alloc_len = msense_6 ? cmd[4] : ((cmd[7] << 8) | cmd[8]);
	/* printk(KERN_INFO "msense: dbd=%d pcontrol=%d pcode=%d "
		"msense_6=%d alloc_len=%d\n", dbd, pcontrol, pcode, "
		"msense_6, alloc_len); */
	if (bufflen < alloc_len)
	   {
		/*
		printk(KERN_INFO "scsi_debug: mode_sense: bufflen=%d "
		       "< alloc_length=%d\n", bufflen, alloc_len);
		*/
		dprintf("[USBMS] scsi_debug: mode_sense: bufflen=%d "
		       "< alloc_length=%d\n", bufflen, alloc_len);
	   }
	memset(buff, 0, bufflen);
	memset(arr, 0, SDEBUG_MAX_MSENSE_SZ);
	if (0x3 == pcontrol) {  /* Saving values not supported */
		mk_sense_buffer(devip, ILLEGAL_REQUEST, SAVING_PARAMS_UNSUP,
			       	0, 18);
		return check_condition_result;
	}
	dev_spec = DEV_READONLY(target) ? 0x80 : 0x0;
	if (msense_6) {
		arr[2] = dev_spec;
		offset = 4;
	} else {
		arr[3] = dev_spec;
		offset = 8;
	}
	ap = arr + offset;

	switch (pcode) {
	case 0x1:	/* Read-Write error recovery page, direct access */
		len = resp_err_recov_pg(ap, pcontrol, target);
		offset += len;
		break;
	case 0x2:	/* Disconnect-Reconnect page, all devices */
		len = resp_disconnect_pg(ap, pcontrol, target);
		offset += len;
		break;
        case 0x3:       /* Format device page, direct access */
                len = resp_format_pg(ap, pcontrol, target);
                offset += len;
                break;
	case 0x8:	/* Caching page, direct access */
		len = resp_caching_pg(ap, pcontrol, target);
		offset += len;
		break;
	case 0xa:	/* Control Mode page, all devices */
		len = resp_ctrl_m_pg(ap, pcontrol, target);
		offset += len;
		break;
	case 0x1c:	/* Informational Exceptions Mode page, all devices */
		len = resp_iec_m_pg(ap, pcontrol, target);
		offset += len;
		break;
	case 0x3f:	/* Read all Mode pages */
		len = resp_err_recov_pg(ap, pcontrol, target);
		len += resp_disconnect_pg(ap + len, pcontrol, target);
		len += resp_format_pg(ap + len, pcontrol, target);
		len += resp_caching_pg(ap + len, pcontrol, target);
		len += resp_ctrl_m_pg(ap + len, pcontrol, target);
		len += resp_iec_m_pg(ap + len, pcontrol, target);
		offset += len;
		break;
	default:
		mk_sense_buffer(devip, ILLEGAL_REQUEST, INVALID_FIELD_IN_CDB,
			       	0, 18);
		return check_condition_result;
	}
	if (msense_6)
		arr[0] = offset - 1;
	else {
		offset -= 2;
		arr[0] = (offset >> 8) & 0xff; 
		arr[1] = offset & 0xff; 
	}
	memcpy(buff, arr, min_len);
	return 0;
}

static int resp_read(struct scsi_cmnd * SCpnt, int upper_blk, int block, 
		     int num, struct sdebug_dev_info * devip)
{
        unsigned char *buff = (unsigned char *) SCpnt->request_buffer;
        int nbytes, sgcount;
 #ifndef NO_SCATTER_LIST
        struct scatterlist *sgpnt = NULL;
 #endif // ndef NO_SCATTER_LIST
        int bufflen = SCpnt->request_bufflen;
	unsigned long iflags;

	if (upper_blk || (block + num > sdebug_capacity)) {
		mk_sense_buffer(devip, ILLEGAL_REQUEST, ADDR_OUT_OF_RANGE, 
				0, 18);
		return check_condition_result;
	}
	if ((SCSI_DEBUG_OPT_MEDIUM_ERR & scsi_debug_opts) &&
	    (block <= OPT_MEDIUM_ERR_ADDR) && 
	    ((block + num) > OPT_MEDIUM_ERR_ADDR)) {
		mk_sense_buffer(devip, MEDIUM_ERROR, UNRECOVERED_READ_ERR, 
				0, 18);
		/* claim unrecoverable read error */
		return check_condition_result;
	}
 #ifndef NO_SPINLOCK
	// rwlock_t : on spinlock.h
	read_lock_irqsave(&atomic_rw, iflags);
 #endif // ndef NO_SPINLOCK
    sgcount = 0;
	nbytes = bufflen;
	/* printk(KERN_INFO "scsi_debug_read: block=%d, tot_bufflen=%d\n", 
	       block, bufflen); */
 #ifndef NO_SCATTER_LIST
	if (SCpnt->use_sg) {
		sgcount = 0;
		sgpnt = (struct scatterlist *) buff;
		buff = scatg2virt(&sgpnt[sgcount]);
		bufflen = sgpnt[sgcount].length;
	}
 #endif // ndef NO_SCATTER_LIST
	do {
		memcpy(buff, fake_storep + (block * SECT_SIZE), bufflen);
		nbytes -= bufflen;
 #ifndef NO_SCATTER_LIST
		if (SCpnt->use_sg) {
			block += bufflen >> POW2_SECT_SIZE;
			sgcount++;
			if (nbytes) {
				buff = scatg2virt(&sgpnt[sgcount]);
				bufflen = sgpnt[sgcount].length;
			}
		} else
 #endif // ndef NO_SCATTER_LIST
		 if (nbytes > 0) 
		{
			/*
			printk(KERN_WARNING "scsi_debug:resp_read: unexpected "
			       "nbytes=%d\n", nbytes);
			*/
			dprintf("[USBMS] scsi_debug:resp_read: unexpected "
			       "nbytes=%d\n", nbytes);
		}
	} while (nbytes);
 #ifndef NO_SPINLOCK
	// rwlock_t : on spinlock.h
	read_unlock_irqrestore(&atomic_rw, iflags);
 #endif // ndef NO_SPINLOCK
 return 0;
}

static int resp_write(struct scsi_cmnd * SCpnt, int upper_blk, int block,
		      int num, struct sdebug_dev_info * devip)
{
 unsigned char *buff = (unsigned char *) SCpnt->request_buffer;
 int nbytes, sgcount;
 #ifndef NO_SCATTER_LIST
        struct scatterlist *sgpnt = NULL;
 #endif // ndef NO_SCATTER_LIST
 int bufflen = SCpnt->request_bufflen;
	unsigned long iflags;

	if (upper_blk || (block + num > sdebug_capacity)) {
		mk_sense_buffer(devip, ILLEGAL_REQUEST, ADDR_OUT_OF_RANGE,
			       	0, 18);
		return check_condition_result;
	}

 #ifndef NO_SPINLOCK
	// rwlock_t : on spinlock.h
	write_lock_irqsave(&atomic_rw, iflags);
 #endif // ndef NO_SPINLOCK
    sgcount = 0;
	nbytes = bufflen;
 #ifndef NO_SCATTER_LIST
	if (SCpnt->use_sg) {
		sgcount = 0;
		sgpnt = (struct scatterlist *) buff;

		buff = scatg2virt(&sgpnt[sgcount]);
		bufflen = sgpnt[sgcount].length;
	}
 #endif // ndef NO_SCATTER_LIST
	do {
		memcpy(fake_storep + (block * SECT_SIZE), buff, bufflen);

		nbytes -= bufflen;
 #ifndef NO_SCATTER_LIST
		if (SCpnt->use_sg) {
			block += bufflen >> POW2_SECT_SIZE;
			sgcount++;
			if (nbytes) {
				buff = scatg2virt(&sgpnt[sgcount]);
				bufflen = sgpnt[sgcount].length;
			}
		} else 
 #endif // ndef NO_SCATTER_LIST
			if (nbytes > 0) 
		{
			/*
			printk(KERN_WARNING "scsi_debug:resp_write: "
			       "unexpected nbytes=%d\n", nbytes);
			*/
			dprintf("[USBMS] scsi_debug:resp_write: "
			       "unexpected nbytes=%d\n", nbytes);
		}
	} while (nbytes);
 #ifndef NO_SPINLOCK
	// rwlock_t : on spinlock.h
	write_unlock_irqrestore(&atomic_rw, iflags);
 #endif // ndef NO_SPINLOCK
	return 0;
}

static int resp_report_luns(unsigned char * cmd, unsigned char * buff,
			    int bufflen, struct sdebug_dev_info * devip)
{
	unsigned int alloc_len; 
	int lun_cnt, i, upper;
	int select_report = (int)cmd[2];
	struct scsi_lun *one_lun;

	alloc_len = cmd[9] + (cmd[8] << 8) + (cmd[7] << 16) + (cmd[6] << 24);
	if ((alloc_len < 16) || (select_report > 2)) {
		mk_sense_buffer(devip, ILLEGAL_REQUEST, INVALID_FIELD_IN_CDB,
			       	0, 18);
		return check_condition_result;
	}
	if (bufflen > 8) { /* can produce response with up to 16k luns 
			      (lun 0 to lun 16383) */
		memset(buff, 0, bufflen);
		lun_cnt = scsi_debug_max_luns;
		buff[2] = ((sizeof(struct scsi_lun) * lun_cnt) >> 8) & 0xff;
		buff[3] = (sizeof(struct scsi_lun) * lun_cnt) & 0xff;
		lun_cnt = min((int)((bufflen - 8) / sizeof(struct scsi_lun)), 
			      lun_cnt);
		one_lun = (struct scsi_lun *) &buff[8];
		for (i = 0; i < lun_cnt; i++) {
			upper = (i >> 8) & 0x3f;
			if (upper)
				one_lun[i].scsi_lun[0] = 
				    (upper | (SAM2_LUN_ADDRESS_METHOD << 6));
			one_lun[i].scsi_lun[1] = i & 0xff;
		}
	}
	return 0;
}

/* When timer goes off this function is called. */
static void timer_intr_handler(unsigned long indx)
{
	struct sdebug_queued_cmd * sqcp;
	unsigned long iflags;

	if (indx >= SCSI_DEBUG_CANQUEUE) {
		/*
		printk(KERN_ERR "scsi_debug:timer_intr_handler: indx too "
		       "large\n");
		*/
		dprintf("[USBMS] Error :scsi_debug:timer_intr_handler: indx too "
		       "large\n");
		return;
	}
	 #ifndef NO_SPINLOCK
			spin_lock_irqsave(&queued_arr_lock, iflags);
 #endif // ndef NO_SPINLOCK
	sqcp = &queued_arr[(int)indx];
	if (! sqcp->in_use) {
		/*
		printk(KERN_ERR "scsi_debug:timer_intr_handler: Unexpected "
		       "interrupt\n");
		*/
		dprintf("[USBMS] Error :scsi_debug:timer_intr_handler: Unexpected "
		       "interrupt\n");

		#ifndef NO_SPINLOCK
			spin_unlock_irqrestore(&queued_arr_lock, iflags);
		#endif // ndef NO_SPINLOCK
		return;
	}
	sqcp->in_use = 0;
	if (sqcp->done_funct) {
		sqcp->a_cmnd->result = sqcp->scsi_result;
		sqcp->done_funct(sqcp->a_cmnd); /* callback to mid level */
	}
	sqcp->done_funct = NULL;
	 #ifndef NO_SPINLOCK
			spin_unlock_irqrestore(&queued_arr_lock, iflags);
 #endif // ndef NO_SPINLOCK
}

static int scsi_debug_slave_alloc(struct scsi_device * sdp)
{
	if (SCSI_DEBUG_OPT_NOISE & scsi_debug_opts)
	   {
		/*
		printk(KERN_INFO "scsi_debug: slave_alloc <%u %u %u %u>\n",
		       sdp->host->host_no, sdp->channel, sdp->id, sdp->lun);
		*/
		dprintf("[USBMS] scsi_debug: slave_alloc <%u %u %u %u>\n",
		       sdp->host->host_no, sdp->channel, sdp->id, sdp->lun);
	   }
	return 0;
}

static int scsi_debug_slave_configure(struct scsi_device * sdp)
{
	struct sdebug_dev_info * devip; 

	if (SCSI_DEBUG_OPT_NOISE & scsi_debug_opts)
	   {
		/*
		printk(KERN_INFO "scsi_debug: slave_configure <%u %u %u %u>\n",
		       sdp->host->host_no, sdp->channel, sdp->id, sdp->lun);
		*/
		dprintf("[USBMS] scsi_debug: slave_configure <%u %u %u %u>\n",
		       sdp->host->host_no, sdp->channel, sdp->id, sdp->lun);
	   }
	if (sdp->host->max_cmd_len != SCSI_DEBUG_MAX_CMD_LEN)
		sdp->host->max_cmd_len = SCSI_DEBUG_MAX_CMD_LEN;
	devip = devInfoReg(sdp);
	sdp->hostdata = devip;
	if (sdp->host->cmd_per_lun)
		scsi_adjust_queue_depth(sdp, SDEBUG_TAGGED_QUEUING, 
					sdp->host->cmd_per_lun);
	return 0;
}

static void scsi_debug_slave_destroy(struct scsi_device * sdp)
{
	struct sdebug_dev_info * devip = 
				(struct sdebug_dev_info *)sdp->hostdata;

	if (SCSI_DEBUG_OPT_NOISE & scsi_debug_opts)
	   {
		/*
		printk(KERN_INFO "scsi_debug: slave_destroy <%u %u %u %u>\n",
		       sdp->host->host_no, sdp->channel, sdp->id, sdp->lun);
		*/
		dprintf("[USBMS] scsi_debug: slave_destroy <%u %u %u %u>\n",
		       sdp->host->host_no, sdp->channel, sdp->id, sdp->lun);
	   }
	if (devip) {
		/* make this slot avaliable for re-use */
		devip->used = 0;
		sdp->hostdata = NULL;
	}
}

static struct sdebug_dev_info * devInfoReg(struct scsi_device * sdev)
{
	struct sdebug_host_info * sdbg_host;
	struct sdebug_dev_info * open_devip = NULL;
	struct sdebug_dev_info * devip =
			(struct sdebug_dev_info *)sdev->hostdata;

	if (devip)
		return devip;
	sdbg_host = *(struct sdebug_host_info **) sdev->host->hostdata;
        if(! sdbg_host) {
			/*
                printk(KERN_ERR "Host info NULL\n");
			*/
            dprintf("[USBMS] Error :Unable to locate host info\n");
		return NULL;
        }
	list_for_each_entry(devip, &sdbg_host->dev_info_list, dev_list) {
		if ((devip->used) && (devip->channel == sdev->channel) &&
                    (devip->target == sdev->id) &&
                    (devip->lun == sdev->lun))
                        return devip;
		else {
			if ((!devip->used) && (!open_devip))
				open_devip = devip;
		}
	}
        if (open_devip) {
		open_devip->channel = sdev->channel;
		open_devip->target = sdev->id;
		open_devip->lun = sdev->lun;
		open_devip->sdbg_host = sdbg_host;
		open_devip->reset = 1;
		open_devip->used = 1;
		memset(open_devip->sense_buff, 0, SDEBUG_SENSE_LEN);
		open_devip->sense_buff[0] = 0x70;
		return open_devip;
        }
        return NULL;
}

static void mk_sense_buffer(struct sdebug_dev_info * devip, int key, 
			    int asc, int asq, int inbandLen)
{
	unsigned char * sbuff;

	sbuff = devip->sense_buff;
	memset(sbuff, 0, SDEBUG_SENSE_LEN);
	if (inbandLen > SDEBUG_SENSE_LEN)
		inbandLen = SDEBUG_SENSE_LEN;
	sbuff[0] = 0x70;
	sbuff[2] = key;
	sbuff[7] = (inbandLen > 7) ? (inbandLen - 8) : 0;
	sbuff[12] = asc;
	sbuff[13] = asq;
}

static int scsi_debug_abort(struct scsi_cmnd * SCpnt)
{
	if (SCSI_DEBUG_OPT_NOISE & scsi_debug_opts)
	   {
		/*
		printk(KERN_INFO "scsi_debug: abort\n");
		*/
		dprintf("[USBMS] scsi_debug: abort\n");
	   }
	++num_aborts;
	stop_queued_cmnd(SCpnt);
	return SCSI_SUCCESS;
}

static int scsi_debug_biosparam(struct scsi_device *sdev,
		struct block_device * bdev, sector_t capacity, int *info)
{
	int res;
	unsigned char *buf;

	if (SCSI_DEBUG_OPT_NOISE & scsi_debug_opts)
	   {
		/*
		printk(KERN_INFO "scsi_debug: biosparam\n");
		*/
		dprintf("[USBMS] scsi_debug: biosparam\n");
	   }
	buf = scsi_bios_ptable(bdev);
	if (buf) {
		res = scsi_partsize(buf, capacity, 
				    &info[2], &info[0], &info[1]);
		/*
		kfree(buf);
		*/
		UsbFreeMemC(USB_MEM_C_S,(unsigned char *) buf
				#if defined(USB_DBG_MEM_C_ON)
					,sizeof(*buf)
				#endif
				);

		if (! res)
			return res;
	}
	info[0] = sdebug_heads;
	info[1] = sdebug_sectors_per;
	info[2] = sdebug_cylinders_per;
	return 0;
}

static int scsi_debug_device_reset(struct scsi_cmnd * SCpnt)
{
	struct sdebug_dev_info * devip;

	if (SCSI_DEBUG_OPT_NOISE & scsi_debug_opts)
	   {
		/*
		printk(KERN_INFO "scsi_debug: device_reset\n");
		*/
		dprintf("[USBMS] scsi_debug: device_reset\n");
	   }
	++num_dev_resets;
	if (SCpnt) {
		devip = devInfoReg(SCpnt->device);
		if (devip)
			devip->reset = 1;
	}
	return SCSI_SUCCESS;
}

static int scsi_debug_bus_reset(struct scsi_cmnd * SCpnt)
{
	struct sdebug_host_info *sdbg_host;
        struct sdebug_dev_info * dev_info;
        struct scsi_device * sdp;
        struct Scsi_Host * hp;

	if (SCSI_DEBUG_OPT_NOISE & scsi_debug_opts) 
	   {
		/*
		printk(KERN_INFO "scsi_debug: bus_reset\n");
		*/
		dprintf("[USBMS] scsi_debug: bus_reset\n");
       }
	++num_bus_resets;
	if (SCpnt && ((sdp = SCpnt->device)) && ((hp = sdp->host))) {
		sdbg_host = *(struct sdebug_host_info **) hp->hostdata;
		if (sdbg_host) {
			list_for_each_entry(dev_info,
                                            &sdbg_host->dev_info_list,
                                            dev_list)
				dev_info->reset = 1;
		}
	}
	return SCSI_SUCCESS;
}

static int scsi_debug_host_reset(struct scsi_cmnd * SCpnt)
{
	struct sdebug_host_info * sdbg_host;
        struct sdebug_dev_info * dev_info;

	if (SCSI_DEBUG_OPT_NOISE & scsi_debug_opts)
		{
		/*
		printk(KERN_INFO "scsi_debug: host_reset\n");
		*/
		dprintf("[USBMS] scsi_debug: host_reset\n");
		}
	++num_host_resets;
 #ifndef NO_SPINLOCK
			spin_lock(&sdebug_host_list_lock);
 #endif // ndef NO_SPINLOCK
        list_for_each_entry(sdbg_host, &sdebug_host_list, host_list) {
                list_for_each_entry(dev_info, &sdbg_host->dev_info_list,
                                    dev_list)
                        dev_info->reset = 1;
        }
 #ifndef NO_SPINLOCK
			spin_unlock(&sdebug_host_list_lock);
 #endif // ndef NO_SPINLOCK
	stop_all_queued();
	return SCSI_SUCCESS;
}

/* Returns 1 if found 'cmnd' and deleted its timer. else returns 0 */
static int stop_queued_cmnd(struct scsi_cmnd * cmnd)
{
 unsigned long iflags;
 int k;
 struct sdebug_queued_cmd * sqcp;

 #ifndef NO_SPINLOCK
	spin_lock_irqsave(&queued_arr_lock, iflags);
 #endif // ndef NO_SPINLOCK

 for (k = 0; k < SCSI_DEBUG_CANQUEUE; ++k) {
	sqcp = &queued_arr[k];
	if (sqcp->in_use && (cmnd == sqcp->a_cmnd)) {
		#ifndef NO_TIMER
			del_timer_sync(&sqcp->cmnd_timer);
		#endif // ndef NO_TIMER
		sqcp->in_use = 0;
		sqcp->a_cmnd = NULL;
		break;
	   }
   }
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(&queued_arr_lock, iflags);
 #endif // ndef NO_SPINLOCK
 return (k < SCSI_DEBUG_CANQUEUE) ? 1 : 0;
}

/* Deletes (stops) timers of all queued commands */
static void stop_all_queued(void)
{
	unsigned long iflags;
	int k;
	struct sdebug_queued_cmd * sqcp;

	 #ifndef NO_SPINLOCK
			spin_lock_irqsave(&queued_arr_lock, iflags);
 #endif // ndef NO_SPINLOCK
	for (k = 0; k < SCSI_DEBUG_CANQUEUE; ++k) {
		sqcp = &queued_arr[k];
		if (sqcp->in_use && sqcp->a_cmnd) {
	#ifndef NO_TIMER
			del_timer_sync(&sqcp->cmnd_timer);
	#endif // ndef NO_TIMER
			sqcp->in_use = 0;
			sqcp->a_cmnd = NULL;
		}
	}
 #ifndef NO_SPINLOCK
			spin_unlock_irqrestore(&queued_arr_lock, iflags);
 #endif // ndef NO_SPINLOCK
}

/* Initializes timers in queued array */
static void init_all_queued(void)
{
	unsigned long iflags;
	int k;
	struct sdebug_queued_cmd * sqcp;

	 #ifndef NO_SPINLOCK
			spin_lock_irqsave(&queued_arr_lock, iflags);
 #endif // ndef NO_SPINLOCK
	for (k = 0; k < SCSI_DEBUG_CANQUEUE; ++k) {
		sqcp = &queued_arr[k];
	#ifndef NO_TIMER
		init_timer(&sqcp->cmnd_timer);
	#endif // ndef NO_TIMER
		sqcp->in_use = 0;
		sqcp->a_cmnd = NULL;
	}
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(&queued_arr_lock, iflags);
 #endif // ndef NO_SPINLOCK
}

static int schedule_resp(struct scsi_cmnd * cmnd, 
			 struct sdebug_dev_info * devip,
			 done_funct_t done, int scsi_result, int delta_jiff)
{
 int k, num; 


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] schedule_resp() Begin .. \n"));
 if ((SCSI_DEBUG_OPT_NOISE & scsi_debug_opts) && cmnd) {
	/*
	printk(KERN_INFO "scsi_debug: cmd ");
	*/
	dprintf("[USBMS] scsi debug : cmd ");
	for (k = 0, num = cmnd->cmd_len; k < num; ++k)
			printk("%02x ", (int)cmnd->cmnd[k]);
	printk("\n");
	if (scsi_result) {
		struct scsi_device * sdp = cmnd->device;
		/*
		printk(KERN_INFO "scsi_debug: ... <%u %u %u %u> "
			       "non-zero result=0x%x\n", sdp->host->host_no,
			       sdp->channel, sdp->id, sdp->lun, scsi_result);
		*/
		dprintf("[USBMS] scsi_debug: ... <%u %u %u %u> "
			       "non-zero result=0x%x\n", sdp->host->host_no,
			       sdp->channel, sdp->id, sdp->lun, scsi_result);
	   }
   }

 if (cmnd && devip) {
		/* simulate autosense by this driver */
		if (SAM_STAT_CHECK_CONDITION == (scsi_result & 0xff))
			memcpy(cmnd->sense_buffer, devip->sense_buff, 
			       (SCSI_SENSE_BUFFERSIZE > SDEBUG_SENSE_LEN) ?
			       SDEBUG_SENSE_LEN : SCSI_SENSE_BUFFERSIZE);
	}

 #ifndef NO_TIMER
 if(delta_jiff <= 0) {
 #endif // ndef NO_TIMER
	if (cmnd)
			cmnd->result = scsi_result;
	if (done)
	   {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] schedule_resp() : done() .. \n"));
		done(cmnd);
	   }
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] schedule_resp(1) End. \n"));
	return 0;
 #ifndef NO_TIMER
   } else {
	unsigned long iflags;
	int k;
	struct sdebug_queued_cmd * sqcp = NULL;

	/*
	#ifndef NO_SPINLOCK
			spin_lock_irqsave(&queued_arr_lock, iflags);
	#endif // ndef NO_SPINLOCK
	*/
	for (k = 0; k < SCSI_DEBUG_CANQUEUE; ++k) {
			sqcp = &queued_arr[k];
			if (! sqcp->in_use)
				break;
		}
	if (k >= SCSI_DEBUG_CANQUEUE) {
		/*
		#ifndef NO_SPINLOCK
			spin_unlock_irqrestore(&queued_arr_lock, iflags);
		#endif // ndef NO_SPINLOCK
		*/
		/*
		printk(KERN_WARNING "scsi_debug: can_queue exceeded\n");
		*/
		dprintf("[USBMS] scsi_debug: can_queue exceeded\n");
		return 1;	/* report busy to mid level */
	   }
	sqcp->in_use = 1;
	sqcp->a_cmnd = cmnd;
	sqcp->scsi_result = scsi_result;
	sqcp->done_funct = done;

	#ifndef NO_TIMER
		sqcp->cmnd_timer.function = timer_intr_handler;
		sqcp->cmnd_timer.data = k;
		sqcp->cmnd_timer.expires = jiffies + delta_jiff;
		add_timer(&sqcp->cmnd_timer);
	#endif // ndef NO_TIMER
	/*
	#ifndef NO_SPINLOCK
			spin_unlock_irqrestore(&queued_arr_lock, iflags);
	#endif // ndef NO_SPINLOCK
	*/
	if (cmnd)
			cmnd->result = 0;
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] schedule_resp(2) End. \n"));
	return 0;
   }
 #endif // ndef NO_TIMER
}

/* Set 'perm' (4th argument) to 0 to disable module_param's definition
 * of sysfs parameters (which module_param doesn't yet support).
 * Sysfs parameters defined explicitly below. 
 */
/*
module_param_named(num_tgts, scsi_debug_num_tgts, int, 0);
module_param_named(max_luns, scsi_debug_max_luns, int, 0);
module_param_named(scsi_level, scsi_debug_scsi_level, int, 0);
#ifdef USBMS_REDUCE_RESOURCE
	module_param_named(dev_size_mb, scsi_debug_dev_size_kb, int, 0);
 #else
	module_param_named(dev_size_mb, scsi_debug_dev_size_mb, int, 0);
#endif
module_param_named(opts, scsi_debug_opts, int, 0); // perm=0644 
module_param_named(every_nth, scsi_debug_every_nth, int, 0);
module_param_named(delay, scsi_debug_delay, int, 0); // perm=0644 
module_param_named(add_host, scsi_debug_add_host, int, 0); // perm=0644
*/
/*
MODULE_AUTHOR("Eric Youngdale + Douglas Gilbert");
MODULE_DESCRIPTION("SCSI debug adapter driver");
MODULE_LICENSE("GPL");

MODULE_PARM_DESC(num_tgts, "number of SCSI targets per host to simulate");
MODULE_PARM_DESC(max_luns, "number of SCSI LUNs per target to simulate");
MODULE_PARM_DESC(scsi_level, "SCSI level to simulate");
MODULE_PARM_DESC(dev_size_mb, "size in MB of ram shared by devs");
MODULE_PARM_DESC(opts, "1->noise, 2->medium_error, 4->...");
MODULE_PARM_DESC(every_nth, "timeout every nth command(def=100)");
MODULE_PARM_DESC(delay, "# of jiffies to delay response(def=1)");
MODULE_PARM_DESC(add_host, "0..127 hosts allowed(def=1)");
*/

static char sdebug_info[256];

static const char * scsi_debug_info(struct Scsi_Host * shp)
{
 
 sprintf(sdebug_info, "scsi_debug, %s, num_tgts=%d, "
		"dev_size_mb=%d, opts=0x%x", scsi_debug_version_str,
		#ifdef USBMS_REDUCE_RESOURCE
			scsi_debug_num_tgts, scsi_debug_dev_size_kb,
		 #else
			scsi_debug_num_tgts, scsi_debug_dev_size_mb,
		#endif
		scsi_debug_opts);
 return sdebug_info;
}

/* scsi_debug_proc_info
 * Used if the driver currently has no own support for /proc/scsi
 */
static int scsi_debug_proc_info(struct Scsi_Host *host, char *buffer, char **start, off_t offset,
				int length, int inout)
{
 int len, pos, begin;
 int orig_length;


 orig_length = length;

 if (inout == 1) {
	char arr[16];
	int minLen = length > 15 ? 15 : length;

	/*
	#ifndef _ETCPIP
	// CAP_SYS_ADMIN / CAP_SYS_RAWIO is on include/linux/capability.h
		if (!capable(CAP_SYS_ADMIN) || !capable(CAP_SYS_RAWIO))
			return -EACCES;
	#endif // ndef _ETCPIP
	*/
	memcpy(arr, buffer, minLen);
	arr[minLen] = '\0';

	USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_debug_proc_info() : sscanf() .. \n")) ;
		if (1 != sscanf(arr, "%d", &pos))
			return -EINVAL;
		scsi_debug_opts = pos;
		if (scsi_debug_every_nth > 0)
                        scsi_debug_cmnd_count = 0;
		return length;
	}
	begin = 0;
	pos = len = sprintf(buffer, "scsi_debug adapter driver, %s\n"
	    "num_tgts=%d, shared (ram) size=%d MB, opts=0x%x, "
	    "every_nth=%d(curr:%d)\n"
	    "delay=%d, max_luns=%d, scsi_level=%d\n"
	    "sector_size=%d bytes, cylinders=%d, heads=%d, sectors=%d\n"
	    "number of aborts=%d, device_reset=%d, bus_resets=%d, " 
	    "host_resets=%d\n",
	    scsi_debug_version_str, scsi_debug_num_tgts, 
		#ifdef USBMS_REDUCE_RESOURCE
		    scsi_debug_dev_size_kb,	scsi_debug_opts, scsi_debug_every_nth,
		#else
		    scsi_debug_dev_size_mb,	scsi_debug_opts, scsi_debug_every_nth,
		#endif
	    scsi_debug_cmnd_count, scsi_debug_delay,
	    scsi_debug_max_luns, scsi_debug_scsi_level,
	    SECT_SIZE, sdebug_cylinders_per, sdebug_heads, sdebug_sectors_per,
	    num_aborts, num_dev_resets, num_bus_resets, num_host_resets);
	if (pos < offset) {
		len = 0;
		begin = pos;
	}
	*start = buffer + (offset - begin);	/* Start of wanted data */
	len -= (offset - begin);
	if (len > length)
		len = length;
	return len;
}

static ssize_t sdebug_delay_show(struct device_driver * ddp, char * buf) 
{
 //#ifndef USB_NO_LIB
        return snprintf(buf, PAGE_SIZE, "%d\n", scsi_debug_delay);
 //#else
 //	dprintf("%d\n", scsi_debug_delay);
 //#endif
}

static ssize_t sdebug_delay_store(struct device_driver * ddp, 
				  const char * buf, size_t count)
{
 int delay;
 char work[20];


 USBMS_DBG_INFO(iprintf("[USBMS INFO] sdebug_delay_store() : sscanf() .. \n")) ;
        if (1 == sscanf((char *)buf, "%10s", work)) {
 USBMS_DBG_INFO(iprintf("[USBMS INFO] sdebug_delay_store() : sscanf() .. \n")) ;
		if ((1 == sscanf(work, "%d", &delay)) && (delay >= 0)) {
			scsi_debug_delay = delay;
			return count;
		}
	}
	return -EINVAL;
}
DRIVER_ATTR(delay, S_IRUGO | S_IWUSR, sdebug_delay_show, 
	    sdebug_delay_store)

static ssize_t sdebug_opts_show(struct device_driver * ddp, char * buf) 
{
 // #ifndef USB_NO_LIB
        return snprintf(buf, PAGE_SIZE, "0x%x\n", scsi_debug_opts);
 //#else
 //	dprintf("0x%x\n", scsi_debug_opts);
 //#endif
}

static ssize_t sdebug_opts_store(struct device_driver * ddp, 
				 const char * buf, size_t count)
{
        int opts;
	char work[20];

 USBMS_DBG_INFO(iprintf("[USBMS INFO] sdebug_opts_store() : sscanf() .. \n")) ;
        if (1 == sscanf((char *)buf, "%10s", work)) {
		if (0 == strnicmp(work,"0x", 2)) {
 USBMS_DBG_INFO(iprintf("[USBMS INFO] sdebug_opts_store() : sscanf() .. \n")) ;
			if (1 == sscanf(&work[2], "%x", &opts))
				goto opts_done;
		} else {
 USBMS_DBG_INFO(iprintf("[USBMS INFO] sdebug_opts_store() : sscanf() .. \n")) ;
			if (1 == sscanf(work, "%d", &opts))
				goto opts_done;
		}
	}
	return -EINVAL;
opts_done:
	scsi_debug_opts = opts;
	scsi_debug_cmnd_count = 0;
	return count;
}
DRIVER_ATTR(opts, S_IRUGO | S_IWUSR, sdebug_opts_show, 
	    sdebug_opts_store)

static ssize_t sdebug_num_tgts_show(struct device_driver * ddp, char * buf) 
{
 //#ifndef USB_NO_LIB
        return snprintf(buf, PAGE_SIZE, "%d\n", scsi_debug_num_tgts);
 //#else
 //	dprintf("%d\n", scsi_debug_num_tgts);
 //#endif
}

static ssize_t sdebug_num_tgts_store(struct device_driver * ddp, 
				     const char * buf, size_t count)
{
        int n;

 USBMS_DBG_INFO(iprintf("[USBMS INFO] sdebug_num_tgts_store() : sscanf() .. \n")) ;
	if ((count > 0) && (1 == sscanf((char *)buf, "%d", &n)) && (n >= 0)) {
		scsi_debug_num_tgts = n;
		return count;
	}
	return -EINVAL;
}
DRIVER_ATTR(num_tgts, S_IRUGO | S_IWUSR, sdebug_num_tgts_show, 
	    sdebug_num_tgts_store) 

static ssize_t sdebug_dev_size_mb_show(struct device_driver * ddp, char * buf) 
{
 #ifdef USBMS_REDUCE_RESOURCE
        return snprintf(buf, PAGE_SIZE, "%d\n", scsi_debug_dev_size_kb);
  #else
        return snprintf(buf, PAGE_SIZE, "%d\n", scsi_debug_dev_size_mb);
		//	dprintf("%d\n", scsi_debug_dev_size_mb);
 #endif
}

DRIVER_ATTR(dev_size_mb, S_IRUGO, sdebug_dev_size_mb_show, NULL) 

static ssize_t sdebug_every_nth_show(struct device_driver * ddp, char * buf) 
{
 // #ifndef USB_NO_LIB
        return snprintf(buf, PAGE_SIZE, "%d\n", scsi_debug_every_nth);
 // #else
 //	dprintf("%d\n", scsi_debug_every_nth);
 // #endif
}

static ssize_t sdebug_every_nth_store(struct device_driver * ddp, 
				      const char * buf, size_t count)
{
        int nth;

 USBMS_DBG_INFO(iprintf("[USBMS INFO] sdebug_every_nth_store() : sscanf() .. \n")) ;
	if ((count > 0) && (1 == sscanf((char *)buf, "%d", &nth)) && (nth >= 0)) {
		scsi_debug_every_nth = nth;
		scsi_debug_cmnd_count = 0;
		return count;
	}
	return -EINVAL;
}
DRIVER_ATTR(every_nth, S_IRUGO | S_IWUSR, sdebug_every_nth_show,
	    sdebug_every_nth_store) 

static ssize_t sdebug_max_luns_show(struct device_driver * ddp, char * buf) 
{
 // #ifndef USB_NO_LIB
        return snprintf(buf, PAGE_SIZE, "%d\n", scsi_debug_max_luns);
 //#else
 //	dprintf("%d\n", scsi_debug_max_luns);
 //#endif
}

static ssize_t sdebug_max_luns_store(struct device_driver * ddp, 
				     const char * buf, size_t count)
{
 int n;

 USBMS_DBG_INFO(iprintf("[USBMS INFO] sdebug_max_luns_store() : sscanf() .. \n")) ;
 if ((count > 0) && (1 == sscanf((char *)buf, "%d", &n)) && (n >= 0)) {
		scsi_debug_max_luns = n;
		return count;
	}
	return -EINVAL;
}
DRIVER_ATTR(max_luns, S_IRUGO | S_IWUSR, sdebug_max_luns_show, 
	    sdebug_max_luns_store) 

static ssize_t sdebug_scsi_level_show(struct device_driver * ddp, char * buf) 
{
 //#ifndef USB_NO_LIB
        return snprintf(buf, PAGE_SIZE, "%d\n", scsi_debug_scsi_level);
 //#else
 //	dprintf("%d\n", scsi_debug_scsi_level);
 //#endif
}

DRIVER_ATTR(scsi_level, S_IRUGO, sdebug_scsi_level_show, NULL) 

static ssize_t sdebug_add_host_show(struct device_driver * ddp, char * buf) 
{
 // #ifndef USB_NO_LIB
        return snprintf(buf, PAGE_SIZE, "%d\n", scsi_debug_add_host);
 //#else
 //	dprintf("%d\n", scsi_debug_add_host);
 //#endif
}

static ssize_t sdebug_add_host_store(struct device_driver * ddp, 
				     const char * buf, size_t count)
{
 int delta_hosts;
 char work[20];


 USBMS_DBG_INFO(iprintf("[USBMS INFO] sdebug_add_host_store() : sscanf() .. \n")) ;
        if (1 != sscanf((char *)buf, "%10s", work))
		return -EINVAL;
	{	/* temporary hack around sscanf() problem with -ve nums */
		int neg = 0;

		if ('-' == *work)
			neg = 1;
 USBMS_DBG_INFO(iprintf("[USBMS INFO] sdebug_add_host_store() : sscanf() .. \n")) ;
		if (1 != sscanf(work + neg, "%d", &delta_hosts))
			return -EINVAL;
		if (neg)
			delta_hosts = -delta_hosts;
	}
	if (delta_hosts > 0) {
		do {
			sdebug_add_adapter();
		} while (--delta_hosts);
	} else if (delta_hosts < 0) {
		do {
			sdebug_remove_adapter();
		} while (++delta_hosts);
	}
	return count;
}

DRIVER_ATTR(add_host, S_IRUGO | S_IWUSR, sdebug_add_host_show, 
	    sdebug_add_host_store)


static void do_create_driverfs_files()
{
	driver_create_file(&sdebug_driverfs_driver, &driver_attr_delay);
	driver_create_file(&sdebug_driverfs_driver, &driver_attr_opts);
	driver_create_file(&sdebug_driverfs_driver, &driver_attr_num_tgts);
	driver_create_file(&sdebug_driverfs_driver, &driver_attr_dev_size_mb);
	driver_create_file(&sdebug_driverfs_driver, &driver_attr_every_nth);
	driver_create_file(&sdebug_driverfs_driver, &driver_attr_max_luns);
	driver_create_file(&sdebug_driverfs_driver, &driver_attr_scsi_level);
	driver_create_file(&sdebug_driverfs_driver, &driver_attr_add_host);
}

static void do_remove_driverfs_files()
{
	driver_remove_file(&sdebug_driverfs_driver, &driver_attr_add_host);
	driver_remove_file(&sdebug_driverfs_driver, &driver_attr_scsi_level);
	driver_remove_file(&sdebug_driverfs_driver, &driver_attr_max_luns);
	driver_remove_file(&sdebug_driverfs_driver, &driver_attr_every_nth);
	driver_remove_file(&sdebug_driverfs_driver, &driver_attr_dev_size_mb);
	driver_remove_file(&sdebug_driverfs_driver, &driver_attr_num_tgts);
	driver_remove_file(&sdebug_driverfs_driver, &driver_attr_opts);
	driver_remove_file(&sdebug_driverfs_driver, &driver_attr_delay);
}
/*
static int __init scsi_debug_init(void)
*/
int scsi_debug_init(void)
{
 unsigned long sz;
 int host_to_add;
 int k;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_debug_init() Begin .. \n"));

 #ifdef USBMS_REDUCE_RESOURCE
	sdebug_store_size = (unsigned long) scsi_debug_dev_size_kb * 1024 ;
  #else
	sdebug_store_size = (unsigned long) scsi_debug_dev_size_mb * 1048576;
 #endif
 sdebug_capacity = sdebug_store_size / SECT_SIZE;

 /* play around with geometry, don't waste too much on track 0 */
 sdebug_heads = 8;
 sdebug_sectors_per = 32;

 #ifdef USBMS_REDUCE_RESOURCE
	if (scsi_debug_dev_size_kb >= 16) 
 #else
	if (scsi_debug_dev_size_mb >= 16) 
 #endif
			sdebug_heads = 32;

 #ifdef USBMS_REDUCE_RESOURCE
	 else if (scsi_debug_dev_size_kb >= 256)
 #else
	 else if (scsi_debug_dev_size_mb >= 256)
 #endif
			sdebug_heads = 64;

 sdebug_cylinders_per = (unsigned long)sdebug_capacity / 
			       (sdebug_sectors_per * sdebug_heads);
 if (sdebug_cylinders_per >= 1024) { 
		/* other LLDs do this; implies >= 1GB ram disk ... */
		sdebug_heads = 255;
		sdebug_sectors_per = 63;
		sdebug_cylinders_per = (unsigned long)sdebug_capacity / 
			       (sdebug_sectors_per * sdebug_heads);
	}

 sz = sdebug_store_size;

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_debug_init() : fake_storep = UsbAllocMemC() .. \n"
		));

 /*
 fake_storep = vmalloc(sz);
 */
 fake_storep = (unsigned char *)	UsbAllocMemC(USB_MEM_C_L
				#ifdef USB_DBG_MEM_C_ON
					, sdebug_store_size  // size : 8388608 or 4096
		        #endif
				) ;

 if (NULL == fake_storep) {
		/*
		printk(KERN_ERR "scsi_debug_init: out of memory, 1\n");
		*/
		USBMS_DBG_INFO(iprintf("[USBMS] ERROR : scsi_debug_init() : out of memory, 1 \n"));
		dprintf("[USBMS] ERROR : out of memory \n");
		return -ENOMEM;
	}
 memset(fake_storep, 0, sz);

 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_debug_init() : init_all_queued() .. \n"));
 init_all_queued();

 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_debug_init() : device_register() .. \n"));
 device_register(&pseudo_primary);

 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_debug_init() : bus_register() .. \n"));
 #ifndef NO_KSET
	bus_register(&pseudo_lld_bus);
  #else
	bus_register(&pseudo_lld_bus,&sdebug_bus_drvlist,&sdebug_bus_devlist) ;
 #endif
 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_debug_init() : driver_register() .. \n"));
 driver_register(&sdebug_driverfs_driver);
 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_debug_init() : do_create_driverfs_files() .. \n"));
 do_create_driverfs_files();

 sdebug_driver_template.proc_name = (char *)sdebug_proc_name;

 host_to_add = scsi_debug_add_host;
 scsi_debug_add_host = 0;

 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_debug_init() : add Adapters .. \n"));
 for (k = 0; k < host_to_add; k++) {
	USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_debug_init() : k=%d .. \n",k));
	if (sdebug_add_adapter()) {
		/*
        printk(KERN_ERR "scsi_debug_init: "
                        "sdebug_add_adapter failed k=%d\n", k);
		*/
		USBMS_DBG_INIT(iprintf(
			"[USBMS INIT] scsi_debug_init() : Error : sdebug_add_adapter() Failed k=%d !! \n",
			k));
        dprintf("[USBMS] Error : scsi debug add adapter failed k=%d\n", k);
        break;
	   }
   }

 if(SCSI_DEBUG_OPT_NOISE & scsi_debug_opts) {
	/*
	printk(KERN_INFO "scsi_debug: ... built %d host(s)\n",
	       scsi_debug_add_host);
	*/
	dprintf("[USBMS] scsi_debug() : ... built %d host(s)\n",
	       scsi_debug_add_host);
   }

 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] scsi_debug_init() End. \n"
	));
 return 0;
}

static void __exit scsi_debug_exit(void)
{
	int k = scsi_debug_add_host;


	/* free up adapters here ?? */
	stop_all_queued();
	for (; k; k--)
		sdebug_remove_adapter();
	do_remove_driverfs_files();
	driver_unregister(&sdebug_driverfs_driver);
	bus_unregister(&pseudo_lld_bus);
	device_unregister(&pseudo_primary);
 /*
 vfree(fake_storep);
 */
 UsbFreeMemU(USB_MEM_U_M, (unsigned char *) fake_storep
		#if defined(USB_DBG_MEM_U_ON)
			, DEF_DEV_SIZE_KB * 1024 // size : 8388608 or 4096
		#endif
		) ;
}

/*
device_initcall(scsi_debug_init);
module_exit(scsi_debug_exit);
*/

static struct device pseudo_primary = {
	.bus_id		= "pseudo_0",
};

static int pseudo_lld_bus_match(struct device *dev, 
                          struct device_driver *dev_driver)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] pseudo_lld_bus_match() Begin .. \n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] pseudo_lld_bus_match() End. \n"));
 return 1;
}

static struct bus_type pseudo_lld_bus = {
 #ifdef USB_DBGSTRING_ON 
        .name = "pseudo",
 #endif // USB_DBGSTRING_ON 
        .match = pseudo_lld_bus_match,
};

static void sdebug_release_adapter(struct device * dev)
{
 struct sdebug_host_info *sdbg_host;

 sdbg_host = to_sdebug_host(dev);

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] sdebug_release_adapter() : sdbg_host = UsbFreeMemC(sdbg_host) .. \n"
		));
 /*
 kfree(sdbg_host);
 */
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) sdbg_host
		#if defined(USB_DBG_MEM_C_ON)
			,sizeof(struct sdebug_host_info)
		#endif
		);
}

static int sdebug_add_adapter(void)
{
 int k, devs_per_host;
 int error = 0;
 struct sdebug_host_info *sdbg_host;
 struct sdebug_dev_info *sdbg_devinfo;
 struct list_head *lh, *lh_sf;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sdebug_add_adapter() Begin .. \n"));
 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] sdebug_add_adapter() : sdbg_host = UsbAllocMemC() .. \n"
		));
 /*
 sdbg_host = kmalloc(sizeof(*sdbg_host),GFP_KERNEL);
 */
 sdbg_host = (struct sdebug_host_info *) UsbAllocMemC(USB_MEM_C_M
					#ifdef USB_DBG_MEM_C_ON
						, sizeof(struct sdebug_host_info) // size : 228
		    	    #endif
	           		) ;

 if (NULL == sdbg_host) {
	/*           
    printk(KERN_ERR "%s: out of memory at line %d\n",
                __FUNCTION__, __LINE__);
	*/
    USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] sdebug_add_adapter() : Error : out of memory at line %d\n",
        __LINE__));
	dprintf("[USBMS] Error : out of memory !!\n") ;
	
    return -ENOMEM;
   }

 /*
 memset(sdbg_host, 0, sizeof(*sdbg_host));
 */
 INIT_LIST_HEAD(&sdbg_host->dev_info_list);

 devs_per_host = scsi_debug_num_tgts * scsi_debug_max_luns;

 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] sdebug_add_adapter() : devs_per_host=%d\n",
	devs_per_host
	));

 for (k = 0; k < devs_per_host; k++) {

	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] sdebug_add_adapter() : sdbg_devinfo = UsbAllocMemC() .. \n"
		));

	/*
    sdbg_devinfo = kmalloc(sizeof(*sdbg_devinfo),GFP_KERNEL);
	*/
	sdbg_devinfo = (struct sdebug_dev_info *) UsbAllocMemC(USB_MEM_C_S
						#ifdef USB_DBG_MEM_C_ON
								, sizeof(struct sdebug_dev_info)
			    	    #endif
		           		) ;

	if (NULL == sdbg_devinfo) {
		/*
        printk(KERN_ERR "%s: out of memory at line %d\n",
                        __FUNCTION__, __LINE__);
		*/
        dprintf("[USBMS] Error :%s: out of memory at line %d\n",
                               __FUNCTION__, __LINE__);
        error = -ENOMEM;
		goto clean;
       }
	/*
	memset(sdbg_devinfo, 0, sizeof(*sdbg_devinfo));
	*/
    sdbg_devinfo->sdbg_host = sdbg_host;

 USBMS_DBG_MEM(iprintf("[USBMS MEM] sdebug_add_adapter() : sdbg_host->dev_info_list=%08XH \n",
			sdbg_host->dev_info_list ));
		#ifdef USBMS_BS_ASSERT
			if (!(   (unsigned long) &sdbg_host->dev_info_list
				   > (unsigned long) 0x200000L
		         )
			   )
			   {
				iprintf("[ASSERT][USBMS] sdebug_add_adapter() : " ) ;
    		    iprintf("Error : Invlid address &sdbg_host->dev_info_list=%08XH \n",
						&sdbg_host->dev_info_list) ;
			    iprintf("[ASSERT][USBMS] Continue ..\n") ;
    		   }
		#endif // USBMS_BS_ASSERT
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] sdebug_add_adapter() : list_add_tail(sdbg_devinfo) .. \n"
		));
    list_add_tail(&sdbg_devinfo->dev_list,
                  &sdbg_host->dev_info_list);
   }

 /*
 #ifndef NO_SPINLOCK
        spin_lock(&sdebug_host_list_lock);
 #endif // ndef NO_SPINLOCK
 */

 USBMS_DBG_MEM(iprintf("[USBMS MEM] sdebug_add_adapter() : sdebug_host_list=%08XH \n",
			sdebug_host_list ));
		#ifdef USBMS_BS_ASSERT
			if (!(   (unsigned long) &sdebug_host_list
				   > (unsigned long) 0x200000L
		         )
			   )
			   {
				iprintf("[ASSERT][USBMS] sdebug_add_adapter() : " ) ;
    		    iprintf("Error : Invlid address &sdebug_host_list=%08XH \n",
						&sdebug_host_list) ;
			    iprintf("[ASSERT][USBMS] Continue ..\n") ;
    		   }
		#endif // USBMS_BS_ASSERT
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] sdebug_add_adapter() : list_add_tail(sdbg_host) .. \n"
	));
 list_add_tail(&sdbg_host->host_list, &sdebug_host_list);

 /*
 #ifndef NO_SPINLOCK
        spin_unlock(&sdebug_host_list_lock);
 #endif // ndef NO_SPINLOCK
 */

 sdbg_host->dev.bus = &pseudo_lld_bus;
 sdbg_host->dev.parent = &pseudo_primary;
 sdbg_host->dev.release = &sdebug_release_adapter;

 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] sdebug_add_adapter() : sprintf() .. \n"
	));
 sprintf(sdbg_host->dev.bus_id, "adapter%d", scsi_debug_add_host);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sdebug_add_adapter() : device_register() .. \n"));
 error = device_register(&sdbg_host->dev);

 if(error)
   {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] sdebug_add_adapter() : device_register() Failed !! \n"));
	goto clean;
   }

 ++scsi_debug_add_host;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sdebug_add_adapter() End. \n"));
 return error;

 clean:

 list_for_each_safe(lh, lh_sf, &sdbg_host->dev_info_list) {
	sdbg_devinfo = list_entry(lh, struct sdebug_dev_info,
					  dev_list);

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sdebug_add_adapter() : list_del() .. \n"));
	list_del(&sdbg_devinfo->dev_list);

	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] sdebug_add_adapter() : UsbFreeMemC(sdbg_devinfo) .. \n"
		));
	/*
	kfree(sdbg_devinfo);
	*/
	UsbFreeMemC(USB_MEM_C_S,(unsigned char *) sdbg_devinfo
			#if defined(USB_DBG_MEM_C_ON)
				,sizeof(struct sdebug_dev_info)
			#endif
		   );
   }

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] sdebug_add_adapter() : UsbFreeMemC(sdbg_host) .. \n"
		));
 /*
 kfree(sdbg_host);
 */
 UsbFreeMemC(USB_MEM_C_M,(unsigned char *) sdbg_host
		#if defined(USB_DBG_MEM_C_ON)
			,sizeof(struct sdebug_host_info) // size : 228
		#endif
		);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sdebug_add_adapter() End with Fail !! \n"));
 return error;
}

static void sdebug_remove_adapter()
{
 struct sdebug_host_info * sdbg_host = NULL;


 /*
 #ifndef NO_SPINLOCK
			spin_lock(&sdebug_host_list_lock);
 #endif // ndef NO_SPINLOCK
 */
 if (!list_empty(&sdebug_host_list)) {
	sdbg_host = list_entry(sdebug_host_list.prev,
							struct sdebug_host_info, host_list);
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sdebug_remove_adapter() : list_del() .. \n"));
	list_del(&sdbg_host->host_list);
   }
 /*
 #ifndef NO_SPINLOCK
			spin_unlock(&sdebug_host_list_lock);
 #endif // ndef NO_SPINLOCK
 */
 if (!sdbg_host)
		return;

        device_unregister(&sdbg_host->dev);
        --scsi_debug_add_host;
}

static int sdebug_driver_probe(struct device * dev)
{
 int error = 0;
 struct sdebug_host_info *sdbg_host;
 struct Scsi_Host *hpnt;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sdebug_driver_probe() Begin .. \n"));
 sdbg_host = to_sdebug_host(dev);

 #ifndef _ETCPIP
	hpnt = scsi_host_alloc(&sdebug_driver_template, sizeof(sdbg_host));
  #else
	hpnt = scsi_host_alloc(&sdebug_driver_template, sizeof(sdbg_host) ,
			&scsidbg_shost_my_devices ) ;
 #endif

 if(NULL == hpnt) {
	/*
    printk(KERN_ERR "%s: scsi_register failed\n", __FUNCTION__);
	*/
    USBMS_DBG_INFO(iprintf("[USBMS INFO] sdebug_driver_probe() : Error : scsi_register failed !! \n"));
    dprintf("[USBMS] Error : scsi register failed !! \n");
    error = -ENODEV;
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sdebug_driver_probe() End with Fail !! \n"));
	return error;
   }

 sdbg_host->shost = hpnt;
 *((struct sdebug_host_info **)hpnt->hostdata) = sdbg_host;
 if ((hpnt->this_id >= 0) && (scsi_debug_num_tgts > hpnt->this_id))
		hpnt->max_id = scsi_debug_num_tgts + 1;
	else
		hpnt->max_id = scsi_debug_num_tgts;
 hpnt->max_lun = scsi_debug_max_luns;

 error = scsi_add_host(hpnt, &sdbg_host->dev);
 if (error) {
	/*
    printk(KERN_ERR "%s: scsi_add_host failed\n", __FUNCTION__);
	*/
    USBMS_DBG_INFO(iprintf("[USBMS INFO] sdebug_driver_probe() : Error : scsi_add_host !! \n"));
    dprintf("[USBMS] Error : scsi add host failed !! \n");

    error = -ENODEV;
	scsi_host_put(hpnt);
   } else
	scsi_scan_host(hpnt);


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sdebug_driver_probe() End. \n"));
 return error;
}

static int sdebug_driver_remove(struct device * dev)
{
 struct list_head *lh, *lh_sf;
 struct sdebug_host_info *sdbg_host;
 struct sdebug_dev_info *sdbg_devinfo;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sdebug_driver_remove() Begin .. \n"));
 sdbg_host = to_sdebug_host(dev);

 if (!sdbg_host) {
	/*
	printk(KERN_ERR "%s: Unable to locate host info\n",
                      __FUNCTION__);
	*/
	USBMS_DBG_INFO(iprintf("[USBMS INFO] sdebug_driver_remove() : Error : Unable to locate host info !!\n")) ;
	dprintf("[USBMS] Error : Unable to locate host info !!\n") ;
    return -ENODEV;
   }

 scsi_remove_host(sdbg_host->shost);

 list_for_each_safe(lh, lh_sf, &sdbg_host->dev_info_list) {
	sdbg_devinfo = list_entry(lh, struct sdebug_dev_info,
                                          dev_list);
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sdebug_driver_remove() : list_del() .. \n"));
    list_del(&sdbg_devinfo->dev_list);

	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] sdebug_driver_remove() : UsbFreeMemC(sdbg_devinfo) .. \n"
		));
	/*
	kfree(sdbg_devinfo);
    */
	UsbFreeMemC(USB_MEM_C_S,(unsigned char *) sdbg_devinfo
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(*sdbg_devinfo)
					#endif
					);
   }

 scsi_host_put(sdbg_host->shost);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sdebug_driver_remove() End. \n"));
 return 0;
}

