
// Name of File : sd.c
// ====================================================================

/*
 *      sd.c Copyright (C) 1992 Drew Eckhardt
 *           Copyright (C) 1993, 1994, 1995, 1999 Eric Youngdale
 *
 *      Linux scsi disk driver
 *              Initial versions: Drew Eckhardt
 *              Subsequent revisions: Eric Youngdale
 *	Modification history:
 *       - Drew Eckhardt <drew@colorado.edu> original
 *       - Eric Youngdale <eric@andante.org> add scatter-gather, multiple 
 *         outstanding request, and other enhancements.
 *         Support loadable low-level scsi drivers.
 *       - Jirka Hanika <geo@ff.cuni.cz> support more scsi disks using 
 *         eight major numbers.
 *       - Richard Gooch <rgooch@atnf.csiro.au> support devfs.
 *	 - Torben Mathiasen <tmm@image.dk> Resource allocation fixes in 
 *	   sd_init and cleanups.
 *	 - Alex Davis <letmein@erols.com> Fix problem where partition info
 *	   not being read in sd_open. Fix problem where removable media 
 *	   could be ejected after sd_open.
 *	 - Douglas Gilbert <dgilbert@interlog.com> cleanup for lk 2.5.x
 *
 *	Logging policy (needs CONFIG_SCSI_LOGGING defined):
 *	 - setting up transfer: SCSI_LOG_HLQUEUE levels 1 and 2
 *	 - end of transfer (bh + scsi_lib): SCSI_LOG_HLCOMPLETE level 1
 *	 - entering sd_ioctl: SCSI_LOG_IOCTL level 1
 *	 - entering other commands: SCSI_LOG_HLQUEUE level 3
 *	Note: when the logging level is set by the user, it must be greater
 *	than the level indicated above to trigger output.	
 */

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
	// for deciding whether to compile
	//  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsbMsInit.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP

/*
#include <linux/config.h>
#include <linux/module.h>
*/
/*
#include <linux/fs.h>
*/
#include "PoFile.h" // include/linux/file.h
#include "radix-tree.h" // include/linux/radix-tree.h
#include "PoFs.h" // osport_fs.h

/*
#include <linux/kernel.h>
*/
#include "PoKernel.h"
/*
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/bio.h>
*/
#include "bio.h"
/*
#include <linux/genhd.h>
*/
#include "block/genhd.h"
/*
#include <linux/hdreg.h>
*/
#include "hdreg.h"
/*
#include <linux/errno.h>
*/
#include "PoErrno.h" // osport_errno.h
/*
#ifndef _ETCPIP
	#include <linux/interrupt.h>
	#include <linux/init.h>
	#include <linux/reboot.h>
	#include <linux/vmalloc.h>
*/
/*
#include <linux/blkdev.h>
*/
#include "bitops.h" // B. S. Liaw add for test_bit()
#include "blkdev.h"
/*
#include <linux/blkpg.h>
#include <asm/uaccess.h>
*/

#include "scsi_ds.h"
#include "scsi_hosts.h"

/*
#include <scsi/scsi_driver.h>
*/
// B. S. Liaw Add this
#include "base/PoDeviceSc.h" // #include/linux/device.h , 2.6.0-test7
#include "scsi_driver.h"

/*
#include <scsi/scsi_ioctl.h>
*/
#include "scsi_ioctl.h"
/*
#include <scsi/scsicam.h>
*/
#include "scsi_cam.h"

#include "scsi_logging.h"

#include "config.h"
//#include "blockio.h"      // on /kernel/fatfs
#include "part.h" // on /kernel/fs
#include "nbios.h"

struct DiskParamS USBDiskParam = {
	2880,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

int validDiskCnt = 1;
int validPartitionCnt = 0;
unsigned char PartitionType[20];
unsigned long validPartitionID[20];
unsigned long partitionMask[32] = {
	0x00000001, 0x00000002, 0x00000004, 0x00000008, 
	0x00000010, 0x00000020, 0x00000040, 0x00000080, 
	0x00000100, 0x00000200, 0x00000400, 0x00000800, 
	0x00001000, 0x00002000, 0x00004000, 0x00008000, 
	0x00010000, 0x00020000, 0x00040000, 0x00080000, 
	0x00100000, 0x00200000, 0x00400000, 0x00800000, 
	0x01000000, 0x02000000, 0x04000000, 0x08000000, 
	0x10000000, 0x20000000, 0x40000000, 0x80000000
};

// ====================================================================

#ifdef USBMS_ONLY_ONE_DISK
	struct gendisk *gd;
#endif // USBMS_ONLY_ONE_DISK

// ====================================================================

/*
 * Remaining dev_t-handling stuff
 */
#ifdef USBMS_REDUCE_RESOURCE
	#define SD_MAJORS	(1)
	#define SD_DISKS	(SD_MAJORS << USBMS_PART_BIT_SHIFT)
 #else
	#define SD_MAJORS	16
	#define SD_DISKS	(SD_MAJORS << 4)
#endif

/*
 * Time out in seconds for disks and Magneto-opticals (which are slower).
 */
/*
#define SD_TIMEOUT		(30 * HZ)
#define SD_MOD_TIMEOUT		(75 * HZ)
*/
#define SD_TIMEOUT		(30)
#define SD_MOD_TIMEOUT	(75)


#define ALLOC_FAILURE_MSG "[USBMS] SCSI Disk : Error : Out of memory !! \n" 

/*
 * Number of allowed retries
 */
#define SD_MAX_RETRIES		5

#ifdef USBMS_ONLY_ONE_DISK
	// 2003/12/17 B. S. Liaw add these ..
	// struct inode *disk_global_inode = NULL ;
	struct block_device *disk_global_bdev[USBMS_MAX_MINORS+1] ;
	struct block_device *whole = NULL ;
	int UsbDiskOpened[USBMS_MAX_MINORS+1] ;
#endif // USBMS_ONLY_ONE_DISK

struct scsi_disk {
	struct scsi_driver *driver;	/* always &sd_template */
	struct scsi_device *device;
	struct gendisk	*disk;
	unsigned int	openers;	/* protected by BKL for now, yuck */
	sector_t	capacity;	/* size in 512-byte sectors */
	u32		index;
	u8		media_present;
	u8		write_prot;
	unsigned	WCE : 1;	/* state of disk WCE bit */
	unsigned	RCD : 1;	/* state of disk RCD bit, unused */
};
/*
static unsigned long sd_index_bits[SD_DISKS / BITS_PER_LONG];
*/
#ifdef USBMS_REDUCE_RESOURCE
	#define SD_INDEX_SZ (SD_DISKS+1)
 #else
	#define SD_INDEX_SZ (SD_DISKS / BITS_PER_LONG) 
#endif
static unsigned long sd_index_bits[SD_INDEX_SZ];
/*
#ifndef NO_SPINLOCK
	static spinlock_t sd_index_lock = SPIN_LOCK_UNLOCKED;
#endif // ndef NO_SPINLOCK
*/
static int sd_revalidate_disk(struct gendisk *disk);
static void sd_rw_intr(struct scsi_cmnd * SCpnt);
static int sd_probe(struct device *);
static int sd_remove(struct device *);
static void sd_shutdown(struct device *dev);
static void sd_rescan(struct device *);
static int sd_init_command(struct scsi_cmnd *);
#ifdef USB_DBGSTRING_ON
	static void sd_read_capacity(struct scsi_disk *sdkp, char *diskname,
		 struct scsi_request *SRpnt, unsigned char *buffer);
 #else
	static void sd_read_capacity(struct scsi_disk *sdkp,
		 struct scsi_request *SRpnt, unsigned char *buffer);
#endif

static struct scsi_driver sd_template = {
 /*
 #ifndef _ETCPIP
	.owner			= THIS_MODULE,
 #endif // ndef _ETCPIP
 */
 .gendrv = { // struct device_driver
 
	#ifdef USB_DBGSTRING_ON
		.name		= "sd",
	#endif // USB_DBGSTRING_ON
 
	.probe		= sd_probe,
	.remove		= sd_remove,
	.shutdown	= sd_shutdown,
   },
 .rescan		= sd_rescan,
 .init_command	= sd_init_command,
};

#ifndef USBMS_ONLY_ONE_DISK
static int sd_major(int major_idx)
{
 #ifdef USBMS_REDUCE_RESOURCE
	return SCSI_DISK_MAJOR;
  #else
	switch (major_idx) {
		case 0:
			return SCSI_DISK0_MAJOR;
		case 1 ... 7:
			return SCSI_DISK1_MAJOR + major_idx - 1;
		case 8 ... 15:
			return SCSI_DISK8_MAJOR + major_idx - 8;
		default:
			BUG();
			return 0;	/* shut up gcc */
		}
 #endif
}
#endif // ndef USBMS_ONLY_ONE_DISK

static inline struct scsi_disk *scsi_disk(struct gendisk *disk)
{
	return container_of(disk->private_data, struct scsi_disk, driver);
}

/**
 *	sd_init_command - build a scsi (read or write) command from
 *	information in the request structure.
 *	@SCpnt: pointer to mid-level's per scsi command structure that
 *	contains request and into which the scsi command is written
 *
 *	Returns 1 if successful and 0 if error (or cannot be done now).
 **/
static int sd_init_command(struct scsi_cmnd * SCpnt)
{
 unsigned int this_count, timeout;
 struct gendisk *disk;
 sector_t block;
 struct scsi_device *sdp = SCpnt->device;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_init_command() Begin .. \n"));
 timeout = SD_TIMEOUT;
 if (SCpnt->device->type != TYPE_DISK)
		timeout = SD_MOD_TIMEOUT;

 /*
  * these are already setup, just copy cdb basically
  */
 if(SCpnt->request->flags & REQ_BLOCK_PC) {
	struct request *rq = SCpnt->request;

	if (sizeof(rq->cmd) > sizeof(SCpnt->cmnd))
			return 0;

	memcpy(SCpnt->cmnd, rq->cmd, sizeof(SCpnt->cmnd));
	if (rq_data_dir(rq) == WRITE)
			SCpnt->sc_data_direction = DMA_TO_DEVICE;
	else if (rq->data_len)
			SCpnt->sc_data_direction = DMA_FROM_DEVICE;
	else
			SCpnt->sc_data_direction = DMA_NONE;

	this_count = rq->data_len;
	if (rq->timeout)
			timeout = rq->timeout;

	SCpnt->transfersize = rq->data_len;
	goto queue;
   }

 /*
  * we only do REQ_CMD and REQ_BLOCK_PC
  */
 // REQ_CMD : regular fs rw request
 if(!(SCpnt->request->flags & REQ_CMD)) {
	// REQ_CMD : regular fs rw request
	USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_init_command() : Error : Not regular fs rw request !! \n"));
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_init_command() End with Failed !! \n"));
	return 0;
   }

 disk = SCpnt->request->rq_disk;
 block = SCpnt->request->sector;
#if 1  /* Charles 20040531 */
	this_count = SCpnt->request_bufflen / 512;
#else
	this_count = SCpnt->request_bufflen >> 9;
#endif

 #ifdef USB_DBGSTRING_ON
	/*
	SCSI_LOG_HLQUEUE(1, printk("sd_init_command: disk=%s, block=%llu, "
			    "count=%d\n", disk->disk_name, (unsigned long long)block, this_count));
	*/
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] sd_init_command: disk=%s, block=%u, "
	    "count=%d blocks\n", disk->disk_name, block, this_count));
  #else
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] sd_init_command: block=%u, "
	    "count=%d blocks\n", block, this_count));
 #endif

 if(!sdp || !sdp->online
	/*
	#ifndef NO_SCATTER_LIST
		||	block + SCpnt->request->nr_sectors > get_capacity(disk)
	 #else
	*/
		||	block + 1 > get_capacity(disk)
	/*
	#endif
	*/
	) {
	#ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
		SCSI_LOG_HLQUEUE(2, printk("Finishing %ld sectors\n", 
				 SCpnt->request->nr_sectors));
	#endif // ndef NO_ELEVATOR
	SCSI_LOG_HLQUEUE(2, printk("Retry with 0x%p\n", SCpnt));
	return 0;
   }

 if (sdp->changed) {
	/*
	 * quietly refuse to do anything to a changed disc until 
	 * the changed bit has been reset
	 */
	/* printk("SCSI disk has been changed. Prohibiting further I/O.\n"); */
	return 0;
   }
 #ifdef USB_DBGSTRING_ON
	/*
	SCSI_LOG_HLQUEUE(2, printk("%s : block=%llu\n",
				   disk->disk_name, (unsigned long long)block));
	*/
	// USBMS_DBG_INIT(iprintf("[USBMS INIT] sd_init_command() : [%s] : block=%u\n",
	//			   disk->disk_name, block));
	USBMS_DBG_SECTOR(iprintf("[USBMS SECTOR] sd_init_command() : [%s] : block=%u\n",
				   disk->disk_name, block));
 #else
	// USBMS_DBG_INIT(iprintf("[USBMS INIT] sd_init_command() : block=%u\n",
	//			   block));
	USBMS_DBG_SECTOR(iprintf("[USBMS SECTOR] sd_init_command() :  block=%u\n",
				   block));
 #endif
 #if 0  /* Charles 20040528 */
 /*
  * If we have a 1K hardware sectorsize, prevent access to single
  * 512 byte sectors.  In theory we could handle this - in fact
  * the scsi cdrom driver must be able to handle this because
  * we typically use 1K blocksizes, and cdroms typically have
  * 2K hardware sectorsizes.  Of course, things are simpler
  * with the cdrom, since it is read-only.  For performance
  * reasons, the filesystems should be able to handle this
  * and not force the scsi disk driver to use bounce buffers
  * for this.
  */
 if (sdp->sector_size == 1024) {
		if ((block & 1)
		/*
		#ifndef NO_SCATTER_LIST
			 || (SCpnt->request->nr_sectors & 1)
		#endif // ndef
		*/
			) {
			/*
			printk(KERN_ERR "sd: Bad block number requested");
			*/
			dprintf("[USBMS] Error :sd: Bad block number requested");
			return 0;
		} else {
			block = block >> 1;
			this_count = this_count >> 1;
		}
	}
 if (sdp->sector_size == 2048) {
		if ((block & 3)
		/*
		#ifndef NO_SCATTER_LIST
			 || (SCpnt->request->nr_sectors & 3)
		#endif // ndef
		*/
			) {
			/*
			printk(KERN_ERR "sd: Bad block number requested");
			*/
			dprintf("[USBMS] Error :sd: Bad block number requested");
			return 0;
		} else {
			block = block >> 2;
			this_count = this_count >> 2;
		}
	}
 if (sdp->sector_size == 4096) {
		if ((block & 7)
		/*
		#ifndef NO_SCATTER_LIST
		 || (SCpnt->request->nr_sectors & 7)
		#endif // ndef
		*/
		) {
			/*
			printk(KERN_ERR "sd: Bad block number requested");
			*/
			dprintf("[USBMS] Error :sd: Bad block number requested");
			return 0;
		} else {
			block = block >> 3;
			this_count = this_count >> 3;
		}
	}
 #endif
 
 if (rq_data_dir(SCpnt->request) == WRITE) {
		if (!sdp->writeable) {
			return 0;
		}
		SCpnt->cmnd[0] = WRITE_6;
		SCpnt->sc_data_direction = DMA_TO_DEVICE;
	} else if (rq_data_dir(SCpnt->request) == READ) {
		SCpnt->cmnd[0] = READ_6;
		SCpnt->sc_data_direction = DMA_FROM_DEVICE;
	} else {
		/*
		printk(KERN_ERR "sd: Unknown command %lx\n", 
		       SCpnt->request->flags);
		*/
		dprintf("[USBMS] Error :sd: Unknown command %lx\n", 
		       SCpnt->request->flags);
	/* overkill 	panic("Unknown sd command %lx\n", SCpnt->request->flags); */
	return 0;
   }

 /*
 SCSI_LOG_HLQUEUE(2, printk("%s : %s %d/%ld 512 byte blocks.\n", 
 */
 #ifdef USB_DBGSTRING_ON
	USBMS_DBG_SCSI(iprintf("[USBMS SCSI] sd_init_command: [%s] %s %d/%ld 512 byte blocks.\n",
		disk->disk_name, (rq_data_dir(SCpnt->request) == WRITE) ? 
		"writing" : "reading", this_count, SCpnt->request->nr_sectors));
  #else
	/*
	#ifndef NO_SCATTER_LIST
		USBMS_DBG_SCSI(iprintf("[USBMS SCSI] sd_init_command() : %s %d/%ld 512 byte blocks.\n",
			(rq_data_dir(SCpnt->request) == WRITE) ? 
			"writing" : "reading", this_count, SCpnt->request->nr_sectors));
	 #else
	*/
		USBMS_DBG_SCSI(iprintf("[USBMS SCSI] sd_init_command() : %s %d 512 byte blocks.\n",
			(rq_data_dir(SCpnt->request) == WRITE) ? 
			"writing" : "reading", this_count));
	/*
	#endif
	*/
 #endif

 SCpnt->cmnd[1] = 0;
	
 if (block > 0xffffffff) {
		SCpnt->cmnd[0] += READ_16 - READ_6;
		SCpnt->cmnd[2] = sizeof(block) > 4 ? (unsigned char) (block >> 56) & 0xff : 0;
		SCpnt->cmnd[3] = sizeof(block) > 4 ? (unsigned char) (block >> 48) & 0xff : 0;
		SCpnt->cmnd[4] = sizeof(block) > 4 ? (unsigned char) (block >> 40) & 0xff : 0;
		SCpnt->cmnd[5] = sizeof(block) > 4 ? (unsigned char) (block >> 32) & 0xff : 0;
		SCpnt->cmnd[6] = (unsigned char) (block >> 24) & 0xff;
		SCpnt->cmnd[7] = (unsigned char) (block >> 16) & 0xff;
		SCpnt->cmnd[8] = (unsigned char) (block >> 8) & 0xff;
		SCpnt->cmnd[9] = (unsigned char) block & 0xff;
		SCpnt->cmnd[10] = (unsigned char) (this_count >> 24) & 0xff;
		SCpnt->cmnd[11] = (unsigned char) (this_count >> 16) & 0xff;
		SCpnt->cmnd[12] = (unsigned char) (this_count >> 8) & 0xff;
		SCpnt->cmnd[13] = (unsigned char) this_count & 0xff;
		SCpnt->cmnd[14] = SCpnt->cmnd[15] = 0;
	} else if ((this_count > 0xff) || (block > 0x1fffff) ||
		   SCpnt->device->use_10_for_rw) {
		USBMS_DBG_SCSI(iprintf("[USBMS SCSI] sd_init_command() : use_10_for_rw .. \n")) ;
		if (this_count > 0xffff)
			this_count = 0xffff;

		SCpnt->cmnd[0] += READ_10 - READ_6;
		SCpnt->cmnd[2] = (unsigned char) (block >> 24) & 0xff;
		SCpnt->cmnd[3] = (unsigned char) (block >> 16) & 0xff;
		SCpnt->cmnd[4] = (unsigned char) (block >> 8) & 0xff;
		SCpnt->cmnd[5] = (unsigned char) block & 0xff;
		SCpnt->cmnd[6] = SCpnt->cmnd[9] = 0;
		SCpnt->cmnd[7] = (unsigned char) (this_count >> 8) & 0xff;
		SCpnt->cmnd[8] = (unsigned char) this_count & 0xff;
	} else {
		if (this_count > 0xff)
			this_count = 0xff;

		SCpnt->cmnd[1] |= (unsigned char) ((block >> 16) & 0x1f);
		SCpnt->cmnd[2] = (unsigned char) ((block >> 8) & 0xff);
		SCpnt->cmnd[3] = (unsigned char) block & 0xff;
		SCpnt->cmnd[4] = (unsigned char) this_count;
		SCpnt->cmnd[5] = 0;
	}
	SCpnt->request_bufflen = SCpnt->bufflen =
			this_count * sdp->sector_size;

	/*
	 * We shouldn't disconnect in the middle of a sector, so with a dumb
	 * host adapter, it's safe to assume that we can at least transfer
	 * this many bytes between each connect / disconnect.
	 */
	SCpnt->transfersize = sdp->sector_size;
	SCpnt->underflow = this_count << 9;

 // -------------
 queue:
 // -------------

 SCpnt->allowed = SD_MAX_RETRIES;
 SCpnt->timeout_per_command = timeout;

 /*
  * This is the completion routine we use.  This is matched in terms
  * of capability to this function.
  */
 SCpnt->done = sd_rw_intr;

 /*
 #ifdef USBMS_DBG_READ_ONLY
	#warning: USBMS_DBG_READ_ONLY : ON !!
	if (rq_data_dir(SCpnt->request) == WRITE) {
		USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_init_command() : Error : Read Only , Return Failed !! \n"));
		return 0 ;
	   }
 #endif // USBMS_DBG_READ_ONLY
 */

 /*
  * This indicates that the command is ready from our end to be
  * queued.
  */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_init_command() End. \n"));
 return 1;
}

/**
 *	sd_open - open a scsi disk device
 *	@inode: only i_rdev member may be used
 *	@filp: only f_mode and f_flags may be used
 *
 *	Returns 0 if successful. Returns a negated errno value in case 
 *	of error.
 *
 *	Note: This can be called from a user context (e.g. fsck(1) )
 *	or from within the kernel (e.g. as a result of a mount(1) ).
 *	In the latter case @inode and @filp carry an abridged amount
 *	of information as noted above.
 **/
static int sd_open(
	#ifndef NO_INODE
		struct inode *inode,
	 #else
		struct gendisk *disk ,
	#endif // ndef NO_INODE  
	struct file *filp)
{
 
 #ifndef NO_INODE
	/*
	struct gendisk *disk = inode->i_bdev->bd_disk;
	*/
	struct gendisk *disk ;
 #endif // ndef NO_INODE
 /* 
 struct scsi_disk *sdkp = scsi_disk(disk);
 */
 struct scsi_disk *sdkp ;
 /*
 struct scsi_device *sdev = sdkp->device;
 */
 struct scsi_device *sdev ;
 int retval;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_open() Begin .. \n"));
 USBMS_DBG_MEM(iprintf("[USBMS MEM] sd_open() : inode=%08XH \n",
	inode));
 USBMS_DBG_MEM(iprintf("[USBMS MEM] sd_open() : inode->i_bdev=%08XH \n",
	inode->i_bdev));
 USBMS_DBG_MEM(iprintf("[USBMS MEM] sd_open() : inode->i_bdev->bd_disk=%08XH \n",
	inode->i_bdev->bd_disk));
 #ifndef NO_INODE
	disk = inode->i_bdev->bd_disk;
 #endif // ndef NO_INODE
 USBMS_DBG_MEM(iprintf("[USBMS MEM] sd_open() : disk=%08XH \n",disk));
 sdkp = scsi_disk(disk);
 USBMS_DBG_MEM(iprintf("[USBMS MEM] sd_open() : sdkp=%08XH \n",sdkp));
 sdev = sdkp->device;
 USBMS_DBG_MEM(iprintf("[USBMS MEM] sd_open() : sdev=%08XH \n",sdev));

 #ifdef USB_DBGSTRING_ON
	SCSI_LOG_HLQUEUE(3, printk("sd_open: disk=%s\n", disk->disk_name));
  #else
	SCSI_LOG_HLQUEUE(3, printk("sd_open: .. \n"));
 #endif

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_open(1) : scsi_device_get() .. \n"));
 retval = scsi_device_get(sdev);
 if(retval)
   {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_open(2) End with Failed !! \n"));
	return retval;
   }
 /*
  * If the device is in error recovery, wait until it is done.
  * If the device is offline, then disallow any access to it.
  */
 retval = -ENXIO;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_open(3) : scsi_block_when_processing_errors() .. \n"));
 if (!scsi_block_when_processing_errors(sdev))
		goto error_out;

 #ifdef USBMS_NO_REMOVAL
	goto normal_out; // B. S. Liaw add this !!
 #endif // USBMS_NO_REMOVAL
 

 if(sdev->removable || sdkp->write_prot)
   {
	#ifndef USBMS_NO_REMOVAL
	#ifndef NO_INODE
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_open(4) : check_disk_change() .. \n"));
		check_disk_change(inode->i_bdev);
	#endif // ndef NO_INODE
	#endif // ndef USBMS_NO_REMOVAL
   }

 /*
  * If the drive is empty, just let the open fail.
 */
 retval = -ENOMEDIUM;
 if (sdev->removable && !sdkp->media_present &&
	    !(filp->f_flags & O_NDELAY))
			goto error_out;

 /*
  * If the device has the write protect tab set, have the open fail
  * if the user expects to be able to write to the thing.
  */
 retval = -EROFS;
 if (sdkp->write_prot && (filp->f_mode & FMODE_WRITE))
			goto error_out;

 /*
  * It is possible that the disk changing stuff resulted in
  * the device being taken offline.  If this is the case,
  * report this to the user, and don't pretend that the
  * open actually succeeded.
  */
 retval = -ENXIO;
 if (!sdev->online)
		goto error_out;

 if(!sdkp->openers++ && sdev->removable) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_open(5) : scsi_block_when_processing_errors() .. \n"));
	if (scsi_block_when_processing_errors(sdev))
	   {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_open(6) : scsi_set_medium_removal() .. \n"));
		scsi_set_medium_removal(sdev, SCSI_REMOVAL_PREVENT);
	   }
   }

 // -----------
 normal_out: // B. S. Liaw add this !!
 // -----------

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_open() End. \n"));
 return 0;

 // -----------
 error_out:
 // -----------
 scsi_device_put(sdev);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_open(7) End with Failed !! \n"));
 return retval;	
}

/**
 *	sd_release - invoked when the (last) close(2) is called on this
 *	scsi disk.
 *	@inode: only i_rdev member may be used
 *	@filp: only f_mode and f_flags may be used
 *
 *	Returns 0. 
 *
 *	Note: may block (uninterruptible) if error recovery is underway
 *	on this disk.
 **/
static int sd_release(
	#ifndef NO_INODE  
		struct inode *inode,
	 #else
		struct gendisk *disk,
	#endif // ndef NO_INODE  
	struct file *filp)
{
 #ifndef NO_INODE
	struct gendisk *disk = inode->i_bdev->bd_disk;
 #endif // ndef NO_INODE
 struct scsi_disk *sdkp = scsi_disk(disk);
 struct scsi_device *sdev = sdkp->device;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_release() Begin .. \n "));

 #ifdef USB_DBGSTRING_ON
 /*
 SCSI_LOG_HLQUEUE(3, printk("sd_release: disk=%s\n", disk->disk_name));
 */
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] sd_release() : disk=%s\n", disk->disk_name));
  #else
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] sd_release() : .. \n"));
 #endif
 if(!--sdkp->openers && sdev->removable) {
	if (scsi_block_when_processing_errors(sdev))
		scsi_set_medium_removal(sdev, SCSI_REMOVAL_ALLOW);
   }

 /*
  * XXX and what if there are packets in flight and this close()
  * XXX is followed by a "rmmod sd_mod"?
  */
 scsi_device_put(sdev);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_release() End. \n "));
 return 0;
}
/*
static int sd_hdio_getgeo(struct block_device *bdev, struct hd_geometry *loc)
{
	struct scsi_disk *sdkp = scsi_disk(bdev->bd_disk);
	struct scsi_device *sdp = sdkp->device;
	struct Scsi_Host *host = sdp->host;
	int diskinfo[4];

	// default to most commonly used values 
        diskinfo[0] = 0x40;	// 1 << 6 
       	diskinfo[1] = 0x20;	// 1 << 5 
       	diskinfo[2] = sdkp->capacity >> 11;
	
	// override with calculated, extended default, or driver values 
	if (host->hostt->bios_param)
		host->hostt->bios_param(sdp, bdev, sdkp->capacity, diskinfo);
	else
		scsicam_bios_param(bdev, sdkp->capacity, diskinfo);
	
	//if (put_user(diskinfo[0], &loc->heads))
	//	return -EFAULT;
	//if (put_user(diskinfo[1], &loc->sectors))
	//	return -EFAULT;
	//if (put_user(diskinfo[2], &loc->cylinders))
	//	return -EFAULT;
	
	//if (put_user((unsigned)get_start_sect(bdev),
	//             (unsigned long *)&loc->start))
	//	return -EFAULT;
	
	return 0;
}
*/
/**
 *	sd_ioctl - process an ioctl
 *	@inode: only i_rdev/i_bdev members may be used
 *	@filp: only f_mode and f_flags may be used
 *	@cmd: ioctl command number
 *	@arg: this is third argument given to ioctl(2) system call.
 *	Often contains a pointer.
 *
 *	Returns 0 if successful (some ioctls return postive numbers on
 *	success as well). Returns a negated errno value in case of error.
 *
 *	Note: most ioctls are forward onto the block subsystem or further
 *	down in the scsi subsytem.
 **/

/*
#ifndef USBMS_NO_IOCTL

static int sd_ioctl(
	#ifndef NO_INODE  
		struct inode * inode,
	 #else
		struct block_device *bdev ,
	#endif // ndef NO_INODE  
	struct file * filp, 
    unsigned int cmd, unsigned long arg)
{
 #ifndef NO_INODE  
	struct block_device *bdev = inode->i_bdev;
 #endif // ndef NO_INODE  
 struct gendisk *disk = bdev->bd_disk;
 struct scsi_device *sdp = scsi_disk(disk)->device;
 int error;
 
   
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_ioctl() Begin .. \n"));
 SCSI_LOG_IOCTL(1, printk("sd_ioctl: disk=%s, cmd=0x%x\n",
						disk->disk_name, cmd));

 
 // * If we are in the middle of error recovery, don't let anyone
 // * else try and use this device.  Also, if error recovery fails, it
 // * may try and take the device offline, in which case all further
 // * access to the device is prohibited.
 
 if (!scsi_block_when_processing_errors(sdp))
		return -ENODEV;

	if (cmd == HDIO_GETGEO) {
		if (!arg)
			return -EINVAL;
		return sd_hdio_getgeo(bdev, (struct hd_geometry *)arg);
	}

	
	// Send SCSI addressing ioctls directly to mid level, send other
	// ioctls to block level and then onto mid level if they can't be
	// resolved.
	
	switch (cmd) {
		case SCSI_IOCTL_GET_IDLUN:
		case SCSI_IOCTL_GET_BUS_NUMBER:
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_ioctl() End with Call scsi_ioctl() .. \n"));
			return scsi_ioctl(sdp, cmd, (void *)arg);
		default:
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_ioctl() : scsi_cmd_ioctl() .. \n"));
			error = scsi_cmd_ioctl(bdev, cmd, arg);
			if (error != -ENOTTY)
				return error;
	}
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_ioctl() End with Call scsi_ioctl() .. \n"));
 return scsi_ioctl(sdp, cmd, (void *)arg);
}
#endif // ndef USBMS_NO_IOCTL
*/
static void set_media_not_present(struct scsi_disk *sdkp)
{
	sdkp->media_present = 0;
	sdkp->capacity = 0;
	sdkp->device->changed = 1;
}

/**
 *	sd_media_changed - check if our medium changed
 *	@disk: kernel device descriptor 
 *
 *	Returns 0 if not applicable or no change; 1 if change
 *
 *	Note: this function is invoked from the block subsystem.
 **/
static int sd_media_changed(struct gendisk *disk)
{
 struct scsi_disk *sdkp = scsi_disk(disk);
 struct scsi_device *sdp = sdkp->device;
 int retval;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_media_changed() Begin .. \n"));
 #ifdef USB_DBGSTRING_ON
	 SCSI_LOG_HLQUEUE(3, printk("sd_media_changed: disk=%s\n",
						disk->disk_name));
  #else
	 SCSI_LOG_HLQUEUE(3, printk("sd_media_changed ..\n" ));
 #endif

 if (!sdp->removable)
   {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_media_changed(1) End with Not Changed. \n"));
	return 0;
   }

 /*
  * If the device is offline, don't send any commands - just pretend as
  * if the command failed.  If the device ever comes back online, we
  * can deal with it then.  It is only because of unrecoverable errors
  * that we would ever take a device offline in the first place.
  */
 if (!sdp->online)
		goto not_present;

	/*
	 * Using TEST_UNIT_READY enables differentiation between drive with
	 * no cartridge loaded - NOT READY, drive with changed cartridge -
	 * UNIT ATTENTION, or with same cartridge - GOOD STATUS.
	 *
	 * Drives that auto spin down. eg iomega jaz 1G, will be started
	 * by sd_spinup_disk() from sd_revalidate_disk(), which happens whenever
	 * sd_revalidate() is called.
	 */
 retval = -ENODEV;
 if(scsi_block_when_processing_errors(sdp)) {
	#ifdef USBMS_NO_IOCTL
		USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_media_changed(2) Not Call scsi_ioctl() !! \n"));
	 #else
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_media_changed(2) : scsi_ioctl() .. \n"));
		retval = scsi_ioctl(sdp, SCSI_IOCTL_TEST_UNIT_READY, NULL);
	#endif // ndef USBMS_NO_IOCTL
   }

	/*
	 * Unable to test, unit probably not ready.   This usually
	 * means there is no disc in the drive.  Mark as changed,
	 * and we will figure it out later once the drive is
	 * available again.
	 */
 if (retval)
		 goto not_present;

	/*
	 * For removable scsi disk we have to recognise the presence
	 * of a disk in the drive. This is kept in the struct scsi_disk
	 * struct and tested at open !  Daniel Roche (dan@lectra.fr)
	 */
 sdkp->media_present = 1;

 retval = sdp->changed;
 sdp->changed = 0;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_media_changed() End. \n"));
 return retval;

 not_present:
 set_media_not_present(sdkp);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_media_changed(3) End with Changed. \n"));
 return 1;
}

static void sd_rescan(struct device *dev)
{
 struct scsi_disk *sdkp = dev_get_drvdata(dev);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_rescan() Begin .. \n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_rescan() : sd_revalidate_disk() .. \n"));
 sd_revalidate_disk(sdkp->disk);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_rescan() End. \n"));
}

static struct block_device_operations sd_fops = {
 /*
 #ifndef _ETCPIP
	.owner			= THIS_MODULE,
 #endif // ndef _ETCPIP
 */
 .open			= sd_open,
 .release		= sd_release,
 #ifndef USBMS_NO_IOCTL
	.ioctl			= sd_ioctl,
 #endif // ndef USBMS_NO_IOCTL
 .media_changed		= sd_media_changed,
 .revalidate_disk	= sd_revalidate_disk,
};

/**
 *	sd_rw_intr - bottom half handler: called when the lower level
 *	driver has completed (successfully or otherwise) a scsi command.
 *	@SCpnt: mid-level's per command structure.
 *
 *	Note: potentially run from within an ISR. Must not block.
 **/
static void sd_rw_intr(struct scsi_cmnd * SCpnt)
{
 int result = SCpnt->result;
 int this_count = SCpnt->bufflen / SCpnt->device->sector_size;
 int good_sectors = (result == 0 ? this_count : 0);
 sector_t block_sectors = 1;
 sector_t error_sector;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_rw_intr() Begin .. \n"));

 #ifdef CONFIG_SCSI_LOGGING
	#ifdef USB_DBGSTRING_ON
		/*
		SCSI_LOG_HLCOMPLETE(1, printk("sd_rw_intr: %s: res=0x%x\n", 
		*/
		SCSI_LOG_HLCOMPLETE(1, iprintf("[USBMS SCSI] sd_rw_intr() : %s: res=0x%x\n", 
				SCpnt->request->rq_disk->disk_name, result));
	 #else
		SCSI_LOG_HLCOMPLETE(1, iprintf("[USBMS SCSI] sd_rw_intr() : res=0x%x\n", 
				result));
	#endif
	if (0 != result) {
		/*
		SCSI_LOG_HLCOMPLETE(1, printk("sd_rw_intr: sb[0,2,asc,ascq]"
		*/
		SCSI_LOG_HLCOMPLETE(1, iprintf("[USBMS SCSI] sd_rw_intr() : sb[0,2,asc,ascq]"
				"=%x,%x,%x,%x\n", SCpnt->sense_buffer[0],
			SCpnt->sense_buffer[2], SCpnt->sense_buffer[12],
			SCpnt->sense_buffer[13]));
	}
 #endif
 /*
  Handle MEDIUM ERRORs that indicate partial success.  Since this is a
  relatively rare error condition, no care is taken to avoid
  unnecessary additional work such as memcpy's that could be avoided.
 */

 /* An error occurred */
 if (driver_byte(result) != 0 && 	/* An error occurred */
	    (SCpnt->sense_buffer[0] & 0x7f) == 0x70) { /* Sense current */
		switch (SCpnt->sense_buffer[2]) {
		case MEDIUM_ERROR:
			if (!(SCpnt->sense_buffer[0] & 0x80))
				break;
			error_sector = (SCpnt->sense_buffer[3] << 24) |
			(SCpnt->sense_buffer[4] << 16) |
			(SCpnt->sense_buffer[5] << 8) |
			SCpnt->sense_buffer[6];
		// #ifndef USB_NO_VMM // vitrtul memory management
			if (SCpnt->request->bio != NULL)
				block_sectors = bio_sectors(SCpnt->request->bio);
		// #endif // ndef USB_NO_VMM
			switch (SCpnt->device->sector_size) {
			case 1024:
				error_sector <<= 1;
				if (block_sectors < 2)
					block_sectors = 2;
				break;
			case 2048:
				error_sector <<= 2;
				if (block_sectors < 4)
					block_sectors = 4;
				break;
			case 4096:
				error_sector <<=3;
				if (block_sectors < 8)
					block_sectors = 8;
				break;
			case 256:
				error_sector >>= 1;
				break;
			default:
				break;
			}

			error_sector &= ~(block_sectors - 1);
			good_sectors = error_sector - SCpnt->request->sector;
			if (good_sectors < 0 || good_sectors >= this_count)
				good_sectors = 0;
			break;

		case RECOVERED_ERROR:
			/*
			 * An error occurred, but it recovered.  Inform the
			 * user, but make sure that it's not treated as a
			 * hard error.
			 */
			print_sense("sd", SCpnt);
			SCpnt->result = 0;
			SCpnt->sense_buffer[0] = 0x0;
			good_sectors = this_count;
			break;

		case ILLEGAL_REQUEST:
			if (SCpnt->device->use_10_for_rw &&
			    (SCpnt->cmnd[0] == READ_10 ||
			     SCpnt->cmnd[0] == WRITE_10))
			   {
				USBMS_DBG_INIT(iprintf("[USBMS INIT] sd_rw_intr() : Set ->use_10_for_rw=0 .. \n")) ;
				SCpnt->device->use_10_for_rw = 0;
			   }

			if (SCpnt->device->use_10_for_ms &&
			    (SCpnt->cmnd[0] == MODE_SENSE_10 ||
			     SCpnt->cmnd[0] == MODE_SELECT_10))
			   {
				USBMS_DBG_INIT(iprintf("[USBMS INIT] sd_rw_intr() : Set ->use_10_for_ms=0 .. \n")) ;
				SCpnt->device->use_10_for_ms = 0;
			   }
			break;

		default:
			break;
		}
	}
 /*
  * This calls the generic completion function, now that we know
  * how many actual sectors finished, and how many sectors we need
  * to say have failed.
  */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_rw_intr() : scsi_io_completion() .. \n"));
 scsi_io_completion(SCpnt, good_sectors, block_sectors);
}

static int media_not_present(struct scsi_disk *sdkp, struct scsi_request *srp)
{
	if (!srp->sr_result)
		return 0;
	if (!(driver_byte(srp->sr_result) & DRIVER_SENSE))
		return 0;
	if (srp->sr_sense_buffer[2] != NOT_READY &&
	    srp->sr_sense_buffer[2] != UNIT_ATTENTION)
	    	return 0;
	if (srp->sr_sense_buffer[12] != 0x3A) /* medium not present */
		return 0;

	set_media_not_present(sdkp);
	return 1;
}

/*
 * spinup disk - called only in sd_revalidate_disk()
 */
#ifdef USB_DBGSTRING_ON
	static void
	sd_spinup_disk(struct scsi_disk *sdkp, char *diskname,
 #else
	static void
	sd_spinup_disk(struct scsi_disk *sdkp,
#endif
	       struct scsi_request *SRpnt, unsigned char *buffer) {
 unsigned char cmd[10];
 unsigned long spintime_value = 0;
 int retries, spintime;
 unsigned int the_result;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_spinup_disk() Begin .. \n"));
 spintime = 0;

 /* Spin up drives, as required.  Only do this at boot time */
 /* Spinup needs to be done for module loads too. */
 do {
	retries = 0;

	spintime_value = get_sys_time() + (100 * HZ);
	do {
		cmd[0] = TEST_UNIT_READY;
		memset((void *) &cmd[1], 0, 9);

		SRpnt->sr_cmd_len = 0;
		SRpnt->sr_sense_buffer[0] = 0;
		SRpnt->sr_sense_buffer[2] = 0;
		SRpnt->sr_data_direction = DMA_NONE;

		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_spinup_disk() : scsi_wait_req() .. \n"));
		scsi_wait_req(SRpnt, (void *) cmd, (void *) buffer,
				       0/*512*/, SD_TIMEOUT, SD_MAX_RETRIES);

		the_result = SRpnt->sr_result;
		retries++;
	   } while (get_sys_time() < spintime_value && 
			 (!scsi_status_is_good(the_result) ||
			  ((driver_byte(the_result) & DRIVER_SENSE) &&
			   SRpnt->sr_sense_buffer[2] == UNIT_ATTENTION)));

	/*
	 * If the drive has indicated to us that it doesn't have
	 * any media in it, don't bother with any of the rest of
	 * this crap.
	 */
	if (media_not_present(sdkp, SRpnt))
			return;

	if ((driver_byte(the_result) & DRIVER_SENSE) == 0) {
			/* no sense, TUR either succeeded or failed
			 * with a status error */
			if(!spintime && !scsi_status_is_good(the_result))
			  {
			#ifdef USB_DBGSTRING_ON
				/*
				printk(KERN_NOTICE "%s: Unit Not Ready, error = 0x%x\n", diskname, the_result);
				*/
				dprintf("[USBMS] Error : [%s] : Unit Not Ready, error = 0x%x\n", diskname, the_result);
			#else
				dprintf("[USBMS] Error : Unit Not Ready, error = 0x%x\n", the_result);
			#endif
			  }
			break;
		}
					
		/*
		 * The device does not want the automatic start to be issued.
		 */
		if (sdkp->device->no_start_on_add) {
			break;
		}

		/*
		 * If manual intervention is required, or this is an
		 * absent USB storage device, a spinup is meaningless.
		 */
		if (SRpnt->sr_sense_buffer[2] == NOT_READY &&
		    SRpnt->sr_sense_buffer[12] == 4 /* not ready */ &&
		    SRpnt->sr_sense_buffer[13] == 3) {
			break;		/* manual intervention required */

		/*
		 * Issue command to spin up drive when not ready
		 */
		} else if (SRpnt->sr_sense_buffer[2] == NOT_READY) {
			unsigned long time1;
			if (!spintime) {
			#ifdef USB_DBGSTRING_ON
				/*
				printk(KERN_NOTICE "%s: Spinning up disk...",
				       diskname);
				*/
				dprintf("[USBMS] [%s]: Spinning up disk...",
				       diskname);
			#else
				dprintf("[USBMS] Spinning up disk...");
			#endif

				cmd[0] = START_STOP;
				cmd[1] = 1;	/* Return immediately */
				memset((void *) &cmd[2], 0, 8);
				cmd[4] = 1;	/* Start spin cycle */
				SRpnt->sr_cmd_len = 0;
				SRpnt->sr_sense_buffer[0] = 0;
				SRpnt->sr_sense_buffer[2] = 0;

				SRpnt->sr_data_direction = DMA_NONE;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_spinup_disk() : scsi_wait_req() .. \n"));
				scsi_wait_req(SRpnt, (void *)cmd, 
					      (void *) buffer, 0/*512*/, 
					      SD_TIMEOUT, SD_MAX_RETRIES);
		#ifndef NO_TIMER
				spintime_value = jiffies;
		#endif // ndef NO_TIMER
			}
			spintime = 1;
			time1 = HZ;
			/* Wait 1 second for next try */
			do {
		    #ifndef _ETCPIP
				current->state = TASK_UNINTERRUPTIBLE;
				time1 = schedule_timeout(time1);
		    #endif // ndef _ETCPIP
			} while(time1);
			printk(".");
		} else {
			/* we don't understand the sense code, so it's
			 * probably pointless to loop */
			if(!spintime) {
			#ifdef USB_DBGSTRING_ON
				/*
				printk(KERN_NOTICE "%s: Unit Not Ready, sense:\n", diskname);
				*/
				dprintf("[USBMS] NOTICE [%s] : Unit Not Ready, sense:\n", diskname);
			#else
				dprintf("[USBMS] NOTICE : Unit Not Ready, sense:\n");
			#endif
				print_req_sense("", SRpnt);
		}
			break;
		}
				
	#ifndef NO_TIMER
	} while (spintime &&
		 time_after(spintime_value + 100 * HZ, jiffies));
	#else
	} while (spintime) ;
	#endif // ndef NO_TIMER

	if (spintime) {
		if (scsi_status_is_good(the_result))
			printk("ready\n");
		else {
			printk("not responding...\n");
			sdkp->media_present = 0;	// bonian chen - 02 Jun 2005
										// for disk that could not spin up
		}
	}
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_spinup_disk() End. \n"));
}

/*
 * read disk capacity
 */
#ifdef USB_DBGSTRING_ON
	static void
	sd_read_capacity(struct scsi_disk *sdkp, char *diskname,
 #else
	static void
	sd_read_capacity(struct scsi_disk *sdkp,
#endif
		 struct scsi_request *SRpnt, unsigned char *buffer) {
 unsigned char cmd[16];
 struct scsi_device *sdp = sdkp->device;
 int the_result, retries;
 int sector_size = 0;
 int longrc = 0;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_read_capacity() Begin .. \n"));

 repeat:

 retries = 3;
 do {
	if (longrc) {
			memset((void *) cmd, 0, 16);
			cmd[0] = SERVICE_ACTION_IN;
			cmd[1] = SAI_READ_CAPACITY_16;
			cmd[13] = 12;
			memset((void *) buffer, 0, 12);
		} else {
			cmd[0] = READ_CAPACITY;
			memset((void *) &cmd[1], 0, 9);
			memset((void *) buffer, 0, 8);
		}
		
	SRpnt->sr_cmd_len = 0;
	SRpnt->sr_sense_buffer[0] = 0;
	SRpnt->sr_sense_buffer[2] = 0;
	SRpnt->sr_data_direction = DMA_FROM_DEVICE;

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_read_capacity() : scsi_wait_req() .. \n"));
	scsi_wait_req(SRpnt, (void *) cmd, (void *) buffer,
			      longrc ? 12 : 8, SD_TIMEOUT, SD_MAX_RETRIES);

	if (media_not_present(sdkp, SRpnt))
			return;

	the_result = SRpnt->sr_result;
	retries--;

   } while (the_result && retries);

 if(the_result && !longrc) {
	#ifdef USB_DBGSTRING_ON
	/*
	printk(KERN_NOTICE "%s : READ CAPACITY failed.\n"
		       "%s : status=%x, message=%02x, host=%d, driver=%02x \n",
		       diskname, diskname,
		       status_byte(the_result),
		       msg_byte(the_result),
		       host_byte(the_result),
		       driver_byte(the_result));
	*/
	dprintf("[USBMS] Error : [%s] : READ CAPACITY failed.\n"
		       "[%s] : status=%x, message=%02x, host=%d, driver=%02x \n",
		       diskname, diskname,
		       status_byte(the_result),
		       msg_byte(the_result),
		       host_byte(the_result),
		       driver_byte(the_result));
	#else
	dprintf("[USBMS] Error : READ CAPACITY failed.\n"
		       " status=%x, message=%02x, host=%d, driver=%02x \n",
		       status_byte(the_result),
		       msg_byte(the_result),
		       host_byte(the_result),
		       driver_byte(the_result));
	#endif
	if (driver_byte(the_result) & DRIVER_SENSE)
			print_req_sense("sd", SRpnt);
		else
		 {
	#ifdef USB_DBGSTRING_ON
			printk("%s : sense not available. \n", diskname);
	#else
			dprintf("[USBMS] sense not available. \n");
	#endif
		 }

	/* Set dirty bit for removable devices if not ready -
	 * sometimes drives will not report this properly. */
	if (sdp->removable &&
		    SRpnt->sr_sense_buffer[2] == NOT_READY)
			sdp->changed = 1;

	/* Either no media are present but the drive didn't tell us,
	   or they are present but the read capacity command fails */
	/* sdkp->media_present = 0; -- not always correct */
	sdkp->capacity = 0x200000; /* 1 GB - random */
	USBMS_DBG_INIT(iprintf("[USBMS INIT] sd_read_capacity(1) : Set Value : sdkp->capacity=%d \n",
		sdkp->capacity));

	return;
   } else if (the_result && longrc) {
	/* READ CAPACITY(16) has been failed */
	#ifdef USB_DBGSTRING_ON
	/*
	printk(KERN_NOTICE "%s : READ CAPACITY(16) failed.\n"
		       "%s : status=%x, message=%02x, host=%d, driver=%02x \n",
		       diskname, diskname,
		       status_byte(the_result),
		       msg_byte(the_result),
		       host_byte(the_result),
		       driver_byte(the_result));
	*/
	dprintf("[USBMS] Error : [%s] : READ CAPACITY(16) failed.\n"
		       "[%s] : status=%x, message=%02x, host=%d, driver=%02x \n",
		       diskname, diskname,
		       status_byte(the_result),
		       msg_byte(the_result),
		       host_byte(the_result),
		       driver_byte(the_result));
	/*
	printk(KERN_NOTICE "%s : use 0xffffffff as device size\n",
		       diskname);
	*/
	dprintf("[USBMS] NOTICE %s : use 0xffffffff as device size\n",
		       diskname);
	#else
	dprintf("[USBMS] Error : READ CAPACITY(16) failed.\n"
		       "status=%x, message=%02x, host=%d, driver=%02x \n",
		       status_byte(the_result),
		       msg_byte(the_result),
		       host_byte(the_result),
		       driver_byte(the_result));
	dprintf("[USBMS] NOTICE : use 0xffffffff as device size \n");
	#endif
	sdkp->capacity = 1 + (sector_t) 0xffffffff;		
	USBMS_DBG_INIT(iprintf("[USBMS INIT] sd_read_capacity(2) : Set Value : sdkp->capacity=%d \n",
			sdkp->capacity));
	goto got_data;
   }	
	
 if(!longrc) {
	sector_size = (buffer[4] << 24) |
			(buffer[5] << 16) | (buffer[6] << 8) | buffer[7];
	if (buffer[0] == 0xff && buffer[1] == 0xff &&
		    buffer[2] == 0xff && buffer[3] == 0xff) {
		if(sizeof(sdkp->capacity) > 4) {
			#ifdef USB_DBGSTRING_ON
				/*
				printk(KERN_NOTICE "%s : very big device. try to use"
				       " READ CAPACITY(16).\n", diskname);
				*/
				dprintf("[USBMS] NOTICE %s : very big device. try to use"
				       " READ CAPACITY(16).\n", diskname);
			#else
				dprintf("[USBMS] NOTICE very big device. try to use"
				       " READ CAPACITY(16).\n");
			#endif
			longrc = 1;
			goto repeat;
		   }
		  else {
			#ifdef USB_DBGSTRING_ON
				/*
				printk(KERN_ERR "%s: too big for kernel.  Assuming maximum 2Tb\n", diskname);
				*/
				dprintf("[USBMS] Error :%s: too big for kernel.  Assuming maximum 2Tb\n", diskname);
			 #else
				dprintf("[USBMS] Error : too big for kernel.  Assuming maximum 2Tb\n");
			#endif
		   }
	   }
	sdkp->capacity = 1 + (((sector_t)buffer[0] << 24) |
			(buffer[1] << 16) |
			(buffer[2] << 8) |
			buffer[3]);			
	USBMS_DBG_INIT(iprintf(
			"[USBMS INIT] sd_read_capacity(3) : Set Value : sdkp->capacity=%d, sector_size=%d\n",
			sdkp->capacity, sector_size));
   }
  else {
	sdkp->capacity = 1 + (((u64)buffer[0] << 56) |
			((u64)buffer[1] << 48) |
			((u64)buffer[2] << 40) |
			((u64)buffer[3] << 32) |
			((sector_t)buffer[4] << 24) |
			((sector_t)buffer[5] << 16) |
			((sector_t)buffer[6] << 8)  |
			(sector_t)buffer[7]);
	sector_size = (buffer[8] << 24) |
			(buffer[9] << 16) | (buffer[10] << 8) | buffer[11];
	USBMS_DBG_INIT(iprintf("[USBMS INIT] sd_read_capacity(4) : Set Value : sdkp->capacity=%d, sector_size=%d\n",
			sdkp->capacity, sector_size));
   }

 // -----------------
 got_data:
 // -----------------

 if(sector_size == 0) {
	sector_size = 512;
	#ifdef USB_DBGSTRING_ON
	/*
	printk(KERN_NOTICE "%s : sector size 0 reported, "
		       "assuming 512.\n", diskname);
	*/
	dprintf("[USBMS] Error : [%s] : sector size 0 reported, "
		       "assuming 512.\n", diskname);
	 #else
		dprintf("[USBMS] Error : sector size 0 reported, "
		       "assuming 512.\n");
	#endif
   }

 if(sector_size != 512 &&
    sector_size != 1024 &&
    sector_size != 2048 &&
    sector_size != 4096 &&
    sector_size != 256) {
	#ifdef USB_DBGSTRING_ON
	/*
	printk(KERN_NOTICE "%s : unsupported sector size "
		       "%d.\n", diskname, sector_size);
	*/
	dprintf("[USBMS] Error : [%s] : unsupported sector size "
		       "%d.\n", diskname, sector_size);
	 #else
		dprintf("[USBMS] Error :  unsupported sector size "
		       "%d.\n", sector_size);
	#endif
	/*
		 * The user might want to re-format the drive with
		 * a supported sectorsize.  Once this happens, it
		 * would be relatively trivial to set the thing up.
		 * For this reason, we leave the thing in the table.
		 */
	sdkp->capacity = 0;
   }

   {
	/*
	 * The msdos fs needs to know the hardware sector size
	 * So I have created this table. See ll_rw_blk.c
	 * Jacques Gelinas (Jacques@solucorp.qc.ca)
	 */
	int hard_sector = sector_size;
	sector_t sz = sdkp->capacity * (hard_sector/256);
	request_queue_t *queue = sdp->request_queue;
	sector_t mb;

	blk_queue_hardsect_size(queue, hard_sector);
	/* avoid 64-bit division on 32-bit platforms */
	mb = sz >> 1;
	sector_div(sz, 1250);
	mb -= sz - 974;
	sector_div(mb, 1950);
	#ifdef USB_DBGSTRING_ON
		/*
		printk(KERN_NOTICE "SCSI device %s: "
		       "%llu %d-byte hdwr sectors (%llu MB)\n",
		       diskname, (unsigned long long)sdkp->capacity,
		       hard_sector, (unsigned long long)mb);
		*/
		dprintf("[USBMS] SCSI device [%s]: %d %d-byte Hardware Sectors (%d MB)\n",
			diskname, sdkp->capacity,
		    hard_sector, mb);
	 #else
		dprintf("[USBMS] SCSI : %d (%08XH) %d-byte Hardware Sectors (%d MB)\n",
			sdkp->capacity,	sdkp->capacity,
		    hard_sector, mb);
	#endif

	// 2003/12/19 B. S. Liaw Add ..
	USBDiskParam.TotalBlocks = sdkp->capacity ;

   }

#if 0  /* Charles 20040524 */
 /* Rescale capacity to 512-byte units */
 if (sector_size == 4096) {
	sdkp->capacity <<= 3;
   }
 else if (sector_size == 2048) {
	sdkp->capacity <<= 2;
   }
 else if (sector_size == 1024) {
	sdkp->capacity <<= 1;
   }
 else if (sector_size == 256) {
	sdkp->capacity >>= 1;
   }
#endif

 sdkp->device->sector_size = sector_size;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_read_capacity() End. \n"));
}

/* called with buffer of length 512 */
static inline int
sd_do_mode_sense(struct scsi_request *SRpnt, int dbd, int modepage,
		 unsigned char *buffer, int len, struct scsi_mode_data *data)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_do_mode_sense() : __scsi_mode_sense() .. \n"));
 return __scsi_mode_sense(SRpnt, dbd, modepage, buffer, len,
				 SD_TIMEOUT, SD_MAX_RETRIES, data);
}

/*
 * read write protect setting, if possible - called only in sd_revalidate_disk()
 * called with buffer of length 512
 */
#ifdef USB_DBGSTRING_ON
	static void
	sd_read_write_protect_flag(struct scsi_disk *sdkp, char *diskname,
 #else
	static void
	sd_read_write_protect_flag(struct scsi_disk *sdkp,
#endif
		   struct scsi_request *SRpnt, unsigned char *buffer) {
	int res;
 // B. S. Liaw : request_buffer need uncached
 /*
 struct scsi_mode_data data;
 */
 struct scsi_mode_data *data;


 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] sd_read_write_protect_flag() : data = UsbAllocMemU() .. \n"
		));
 data = (struct scsi_mode_data *) UsbAllocMemU(USB_MEM_U_S
				#ifdef USB_DBG_MEM_U_ON
					, sizeof(struct scsi_mode_data)
		        #endif
           		) ;
 if(!data) {
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] sd_read_write_protect_flag() : Error : Out of Memory for data !! \n"
		));
	dprintf("[USBMS] Error : Out of Memory !! \n");
	return ;
   }

	/*
	 * First attempt: ask for all pages (0x3F), but only 4 bytes.
	 * We have to start carefully: some devices hang if we ask
	 * for more than is available.
	 */
 /*
 res = sd_do_mode_sense(SRpnt, 0, 0x3F, buffer, 4, &data);
 */
 res = sd_do_mode_sense(SRpnt, 0, 0x3F, buffer, 4, data);

	/*
	 * Second attempt: ask for page 0
	 * When only page 0 is implemented, a request for page 3F may return
	 * Sense Key 5: Illegal Request, Sense Code 24: Invalid field in CDB.
	 */
	if (!scsi_status_is_good(res))
	   {
		/*
		res = sd_do_mode_sense(SRpnt, 0, 0, buffer, 4, &data);
		*/
		res = sd_do_mode_sense(SRpnt, 0, 0, buffer, 4, data);
	   }

	/*
	 * Third attempt: ask 255 bytes, as we did earlier.
	 */
	if (!scsi_status_is_good(res))
	   {
		/*
		res = sd_do_mode_sense(SRpnt, 0, 0x3F, buffer, 255, &data);
		*/
		res = sd_do_mode_sense(SRpnt, 0, 0x3F, buffer, 255, data);
	   }

	if (!scsi_status_is_good(res)) {
	#ifdef USB_DBGSTRING_ON
	   /*
		printk(KERN_WARNING
		       "%s: test WP failed, assume Write Enabled\n", diskname);
		*/
		dprintf(
		       "%s: test WP failed, assume Write Enabled\n", diskname);
	  #else
		dprintf(
		       "[USBMS]  test WP failed, assume Write Enabled\n");
	#endif
	} else {
		/*
		sdkp->write_prot = ((data.device_specific & 0x80) != 0);
		*/
		sdkp->write_prot = ((data->device_specific & 0x80) != 0);
	#ifdef USB_DBGSTRING_ON
		/*
		printk(KERN_NOTICE "%s: Write Protect is %s\n", diskname,
		       sdkp->write_prot ? "on" : "off");
		*/
		dprintf("[USBMS] [%s]: Write Protect is %s\n", diskname,
		       sdkp->write_prot ? "on" : "off");
		/*
		printk(KERN_DEBUG "%s: Mode Sense: %02x %02x %02x %02x\n",
		       diskname, buffer[0], buffer[1], buffer[2], buffer[3]);
		*/
		dprintf("[USBMS] [%s]: Mode Sense: %02x %02x %02x %02x\n",
		       diskname, buffer[0], buffer[1], buffer[2], buffer[3]);
	 #else
		dprintf("[USBMS] Write Protect is %s\n",
		       sdkp->write_prot ? "on" : "off");
		dprintf("[USBMS] Mode Sense: %02x %02x %02x %02x\n",
		       buffer[0], buffer[1], buffer[2], buffer[3]);
	#endif
	}

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] sd_read_write_protect_flag() : UsbFreeMemU(data) .. \n"
		));
 UsbFreeMemU(USB_MEM_U_S,(unsigned char *) data
			#if defined(USB_DBG_MEM_U_ON)
				,sizeof(struct scsi_mode_data)
			#endif
			) ;
}

/*
 * sd_read_cache_type - called only from sd_revalidate_disk()
 * called with buffer of length 512
 */
#ifdef USB_DBGSTRING_ON
static void
sd_read_cache_type(struct scsi_disk *sdkp, char *diskname,
 #else
	static void
	sd_read_cache_type(struct scsi_disk *sdkp,
#endif
		   struct scsi_request *SRpnt, unsigned char *buffer) {
 #ifndef USBMS_SD_NO_CACHE
	int len = 0, res;

	const int dbd = 0;	   /* DBD */
	const int modepage = 0x08; /* current values, cache page */
	// B. S. Liaw : request_buffer need uncached
	/*
	struct scsi_mode_data data;
	*/
	struct scsi_mode_data *data;

 
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_read_cache_type() Begin .. \n"));

	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] sd_read_cache_type() : data = UsbAllocMemU() .. \n"
		));
	data = (struct scsi_mode_data *) UsbAllocMemU(USB_MEM_U_S
				#ifdef USB_DBG_MEM_U_ON
					, sizeof(struct scsi_mode_data)
		        #endif
           		) ;
 if(!data) {
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] sd_read_cache_type() : Error : Out of Memory for data !! \n"
		));
	dprintf("[USBMS] Error : Out of Memory !! \n");
	return ;
   }

 /* cautiously ask */
 /*
 res = sd_do_mode_sense(SRpnt, dbd, modepage, buffer, 4, &data);
 */
 res = sd_do_mode_sense(SRpnt, dbd, modepage, buffer, 4, data);

 if (!scsi_status_is_good(res))
		goto bad_sense;

 /* that went OK, now ask for the proper length */
 /*
 len = data.length;
 */
 len = data->length;
 /*
  * We're only interested in the first three bytes, actually.
  * But the data cache page is defined for the first 20.
  */
 if (len < 3)
		goto bad_sense;
 if (len > 20)
		len = 20;

 /* Take headers and block descriptors into account */
 /*
 len += data.header_length + data.block_descriptor_length;
 */
 len += data->header_length + data->block_descriptor_length;

 /* Get the data */
 /*
 res = sd_do_mode_sense(SRpnt, dbd, modepage, buffer, len, &data);
 */
 res = sd_do_mode_sense(SRpnt, dbd, modepage, buffer, len, data);

 if (scsi_status_is_good(res)) {
	const char *types[] = {
			"write through", "none", "write back",
			"write back, no read (daft)"
		};
	int ct = 0;
	/*
	int offset = data.header_length +
			data.block_descriptor_length + 2;
	*/
	int offset = data->header_length +
			data->block_descriptor_length + 2;


	sdkp->WCE = ((buffer[offset] & 0x04) != 0);
	sdkp->RCD = ((buffer[offset] & 0x01) != 0);

	ct =  sdkp->RCD + 2*sdkp->WCE;
		/*
		printk(KERN_NOTICE "SCSI device %s: drive cache: %s\n",
		       diskname, types[ct]);
		*/
	dprintf("[USBMS] NOTICE SCSI device %s: drive cache: %s\n",
		       diskname, types[ct]);

	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] sd_read_cache_type() : UsbFreeMemU(data) .. \n"
		));
	UsbFreeMemU(USB_MEM_U_S,(unsigned char *) data
			#if defined(USB_DBG_MEM_U_ON)
				,sizeof(struct scsi_mode_data)
			#endif
			) ;
	return;
   }

 bad_sense:

 if ((SRpnt->sr_sense_buffer[0] & 0x70) == 0x70
     && (SRpnt->sr_sense_buffer[2] & 0x0f) == ILLEGAL_REQUEST
     /* ASC 0x24 ASCQ 0x00: Invalid field in CDB */
     && SRpnt->sr_sense_buffer[12] == 0x24
     && SRpnt->sr_sense_buffer[13] == 0x00) {
    /*
	printk(KERN_NOTICE "%s: cache data unavailable\n",
			       diskname);
	*/
	dprintf("[USBMS] NOTICE %s: cache data unavailable\n",
			       diskname);
	} else {
	/*
	printk(KERN_ERR "%s: asking for cache data failed\n",
		       diskname);
	*/
	dprintf("[USBMS] Error :%s: asking for cache data failed\n",
		       diskname);
   }
  #else
	USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_read_cache_type() : USBMS_SD_NO_CACHE !! \n"));
 #endif // USBMS_SD_NO_CACHE

 #ifdef USB_DBGSTRING_ON
	/*
	printk(KERN_ERR "%s: assuming drive cache: write through\n",
		       diskname);
	*/
	dprintf("[USBMS] %s: assuming drive cache: write through\n",
		       diskname);
 #else
	dprintf("[USBMS] assuming drive cache: write through\n");
 #endif

 sdkp->WCE = 0;
 sdkp->RCD = 0;

 #ifndef USBMS_SD_NO_CACHE
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] sd_read_cache_type() : UsbFreeMemU(data) .. \n"
		));
	UsbFreeMemU(USB_MEM_U_S,(unsigned char *) data
			#if defined(USB_DBG_MEM_U_ON)
				,sizeof(struct scsi_mode_data)
			#endif
			) ;
 #endif // ndef USBMS_SD_NO_CACHE
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_read_cache_type() End. \n"));
}

/**
 *	sd_revalidate_disk - called the first time a new disk is seen,
 *	performs disk spin up, read_capacity, etc.
 *	@disk: struct gendisk we care about
 **/
static int sd_revalidate_disk(struct gendisk *disk)
{
 struct scsi_disk *sdkp = scsi_disk(disk);
 struct scsi_device *sdp = sdkp->device;
 struct scsi_request *sreq;
 unsigned char *buffer;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_revalidate_disk() Begin .. \n"));
 #ifdef USB_DBGSTRING_ON
	SCSI_LOG_HLQUEUE(3, printk("[USBMS] sd_revalidate_disk: disk=%s\n", disk->disk_name));
  #else
	SCSI_LOG_HLQUEUE(3, printk("[USBMS] sd_revalidate_disk .. \n"));
 #endif

 /*
  * If the device is offline, don't try and read capacity or any
  * of the other niceties.
  */
 if (!sdp->online)
		goto out;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_revalidate_disk() : scsi_allocate_request() .. \n"));
 /*
 sreq = scsi_allocate_request(sdp, GFP_KERNEL);
 */
 sreq = scsi_allocate_request(sdp);
 if (!sreq) {
		/*
		printk(KERN_WARNING "(sd_init_onedisk:) Request allocation "
		       "failure.\n");
		*/
		dprintf("sd_revalidate_disk() : (sd_init_onedisk:) Request allocation "
		       "failure.\n");
		goto out;
	}

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] sd_revalidate_disk() : buffer = UsbAllocMemU() .. \n"
		));

 // B. S. Liaw : request_buffer need uncached
 /*
 buffer = kmalloc(512, GFP_KERNEL | __GFP_DMA);
 */
 buffer = (unsigned char *)	UsbAllocMemU(USB_MEM_U_M
				#ifdef USB_DBG_MEM_U_ON
					, 512
		        #endif
           		) ;
 if(!buffer) {
	/*
	printk(KERN_WARNING "(sd_revalidate_disk:) Memory allocation "
	*/
	/*
	dprintf("(sd_init_onedisk:) Memory allocation "
		       "failure.\n");
	*/
	dprintf(ALLOC_FAILURE_MSG);
	goto out_release_request;
   }

 /* defaults, until the device tells us otherwise */
 sdp->sector_size = 512;
 sdkp->capacity = 0;
 sdkp->media_present = 1;
 sdkp->write_prot = 0;
 sdkp->WCE = 0;
 sdkp->RCD = 0;

 USBMS_DBG_INIT(iprintf("[USBMS INIT] sd_revalidate_disk() === 1 === Spin Up Disk .. \n"));
 
 #ifdef USB_DBGSTRING_ON
	sd_spinup_disk(sdkp, disk->disk_name, sreq, buffer);
  #else
	sd_spinup_disk(sdkp, sreq, buffer);
 #endif

 /*
  * Without media there is no reason to ask; moreover, some devices
  * react badly if we do.
  */
 if (sdkp->media_present) {
	USBMS_DBG_INIT(iprintf("[USBMS INIT] sd_revalidate_disk() === 2 === Read Capacity .. \n"));
	#ifdef USB_DBGSTRING_ON
		sd_read_capacity(sdkp, disk->disk_name, sreq, buffer);
	 #else
		sd_read_capacity(sdkp, sreq, buffer);
	#endif
	if (sdp->removable)
	   {
		USBMS_DBG_INIT(iprintf("[USBMS INIT] sd_revalidate_disk() === 3 === Check Write Protect .. \n"));
		#ifdef USB_DBGSTRING_ON
			sd_read_write_protect_flag(sdkp, disk->disk_name,
					sreq, buffer);
		 #else
			sd_read_write_protect_flag(sdkp,
					sreq, buffer);
		#endif
	   }
	USBMS_DBG_INIT(iprintf("[USBMS INIT] sd_revalidate_disk() === 4 === Read Cache Type .. \n"));
	#ifdef USB_DBGSTRING_ON
		sd_read_cache_type(sdkp, disk->disk_name, sreq, buffer);
	 #else
		sd_read_cache_type(sdkp, sreq, buffer);
	#endif
   }
	
 USBMS_DBG_INIT(iprintf("[USBMS INIT] sd_revalidate_disk() === 5 === Set Capacity .. \n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_revalidate_disk() : set_capacity() .. \n"));
 set_capacity(disk, sdkp->capacity);

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] sd_revalidate_disk() : UsbFreeMemU(buffer) , size=%d .. \n" ,
	512
	));
 /*
 kfree(buffer);
 */
 UsbFreeMemU(USB_MEM_U_M,(unsigned char *) buffer
			#if defined(USB_DBG_MEM_U_ON)
				,512
			#endif
			) ;
		
 out_release_request: 
	scsi_release_request(sreq);
 out:
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_revalidate_disk() End. \n"));
 return 0;
}

/**
 *	sd_probe - called during driver initialization and whenever a
 *	new scsi device is attached to the system. It is called once
 *	for each scsi device (not just disks) present.
 *	@dev: pointer to device object
 *
 *	Returns 0 if successful (or not interested in this scsi device 
 *	(e.g. scanner)); 1 when there is an error.
 *
 *	Note: this function is invoked from the scsi mid-level.
 *	This function sets up the mapping between a given 
 *	<host,channel,id,lun> (found in sdp) and new device name 
 *	(e.g. /dev/sda). More precisely it is the block device major 
 *	and minor number that is chosen here.
 *
 *	Assume sd_attach is not re-entrant (for time being)
 *	Also think about sd_attach() and sd_remove() running coincidentally.
 **/
extern int usb_block_read(unsigned long VolumeID, unsigned long SectorNo, void *Buffer);
extern int usb_block_readex(unsigned long VolumeID, unsigned long SectorNo, void *Buffer, unsigned short max_prefetch_sectors);
extern int usb_block_write(unsigned long VolumeID, unsigned long SectorNo, void *Buffer);

static int sd_probe(struct device *dev)
{
 struct scsi_device *sdp = to_scsi_device(dev);
 struct scsi_disk *sdkp;
 #ifndef USBMS_ONLY_ONE_DISK
	struct gendisk *gd;
 #endif // ndef USBMS_ONLY_ONE_DISK
 u32 index;
 int error;
 #ifdef USBMS_ONLY_ONE_DISK
	int i ;
 #endif // USBMS_ONLY_ONE_DISK
 int j ;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_probe() Begin .. \n"));
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] sd_probe() Begin .. \n"));

 // #if defined(USB_DBG_MEM_C_ON) || defined(USB_DBG_MEM_U_ON)
 // #ifdef USBMS_DBG_INIT_ON
 #ifdef USB_DBG_MEM_COUNT_ON
	iprintf("[USB MEM_COUNT] sd_probe(1) ====== show USB Mem Count .. \n"
    	) ;
	showUsbMemC() ;
	showUsbMemU() ;
 #endif // USB_DBG_MEM_COUNT_ON


 #ifdef USBMS_ONLY_ONE_DISK
	// disk_global_Sector_buf = NULL ;
	for(i=0;i<USBMS_MAX_MINORS+1;i++) {
		disk_global_bdev[i] = NULL ;
	   }
	whole = NULL ;

	for(i=0;i<USBMS_MAX_MINORS+1;i++) {
		UsbDiskOpened[i] = NULL_VALUE ;
		USBMS_DBG_PLUG(iprintf("[USBMS PLUG] sd_probe() : Set Value : UsbDiskOpened[%d]=%d\n",
			i,UsbDiskOpened[i]));
		//USBMS_DBG_INIT(iprintf("[USBMS INIT] sd_probe() : Set Value : UsbDiskOpened[%d]=%d\n",
		//	i,UsbDiskOpened[i]));
	   }
 #endif // USBMS_ONLY_ONE_DISK

 error = -ENODEV;
 if ((sdp->type != TYPE_DISK) && (sdp->type != TYPE_MOD))
		goto out;

 SCSI_LOG_HLQUEUE(3, printk("sd_attach: scsi device: <%d,%d,%d,%d>\n", 
			 sdp->host->host_no, sdp->channel, sdp->id, sdp->lun));

 error = -ENOMEM;

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] sd_probe() : sdkp = UsbAllocMemC() .. \n"
		));

 /*
 sdkp = kmalloc(sizeof(*sdkp), GFP_KERNEL);
 */
 sdkp = (struct scsi_disk *) UsbAllocMemC(USB_MEM_C_S
				#ifdef USB_DBG_MEM_C_ON
					, sizeof(struct scsi_disk) 
		        #endif
           		) ;

 if (!sdkp)
	goto out;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_probe() : alloc_disk() .. \n")) ;
 /*
 gd = alloc_disk(16);
 */
 gd = alloc_disk(USBMS_MAX_MINORS);
 if (!gd) {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_probe() : alloc_disk() Failed !! \n")) ;
	goto out_free;
   }
 /*
 #ifndef NO_SPINLOCK
	spin_lock(&sd_index_lock);
 #endif // ndef NO_SPINLOCK
 */
 index = find_first_zero_bit(sd_index_bits, SD_DISKS);
 USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_probe() : index=%d \n",
	index )) ;

 index -= 24 ;
 USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_probe(1) : FixMe : index-=24 !! \n"
	)) ;

 USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_probe() : index=%d \n",
	index )) ;

 if(index == SD_DISKS) {
	/*
	#ifndef NO_SPINLOCK
		spin_unlock(&sd_index_lock);
	#endif // ndef NO_SPINLOCK
	*/
	error = -EBUSY;
	goto out_put;
   }
 USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_probe() : set_bit() : index=%d \n",
	index )) ;
 /*
 __set_bit(index, sd_index_bits);
 set_bit(index, sd_index_bits);
 */
 USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_probe(2) : FixMe : index + 24 !! \n"
	)) ;
 set_bit(index+24, sd_index_bits);
 /*
 #ifndef NO_SPINLOCK
	spin_unlock(&sd_index_lock);
 #endif // ndef NO_SPINLOCK
 */
 sdkp->device = sdp;
 sdkp->driver = &sd_template;
 sdkp->disk = gd;
 sdkp->index = index;
 sdkp->openers = 0;
 /*
 gd->major = sd_major(index >> 4);
 gd->first_minor = (index & 15) << 4;
 gd->minors = 16;
 */
 #ifndef USBMS_ONLY_ONE_DISK
	gd->major = sd_major(index >> USBMS_PART_BIT_SHIFT);
 #endif // ndef USBMS_ONLY_ONE_DISK
 #ifdef USBMS_REDUCE_RESOURCE
	#ifndef USBMS_ONLY_ONE_DISK
		gd->first_minor = 0 ;
	#endif // ndef USBMS_ONLY_ONE_DISK
   #else
	gd->major = sd_major(index >> 4);
	gd->first_minor = (index & 15) << 4;
	gd->minors = 16;
	// gd->first_minor = (index & (USBMS_MAX_MINORS-1)) << USBMS_PART_BIT_SHIFT ;
 #endif
 gd->minors = USBMS_MAX_MINORS ;

 #ifndef USBMS_ONLY_ONE_DISK
	USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_probe() : gd->major=%d \n",
		gd->major )) ;
	USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_probe() : gd->first_minor=%d \n",
		gd->first_minor )) ;
 #endif // ndef USBMS_ONLY_ONE_DISK
 USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_probe() : gd->minors=%d \n",
	gd->minors )) ;

 gd->fops = &sd_fops;

 #ifdef USB_DBGSTRING_ON
	 if (index >= 26) {
		sprintf(gd->disk_name, "sd%c%c",
				'a' + index/26-1,'a' + index % 26);
	   }
	  else {
		sprintf(gd->disk_name, "sd%c", 'a' + index % 26);
	   }
 #endif // USB_DBGSTRING_ON

 /*
 #ifdef CONFIG_USB_DEVICEFS
	strcpy(gd->devfs_name, sdp->devfs_name);
 #endif // CONFIG_USB_DEVICEFS
 */
 gd->private_data = &sdkp->driver;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] sd_probe() : Set Value : gd->private_data=%08XH \n",
	gd->private_data));

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_probe() : sd_revalidate_disk() .. \n"));
 sd_revalidate_disk(gd);

 gd->driverfs_dev = &sdp->sdev_gendev;
 gd->flags = GENHD_FL_DRIVERFS;
 if (sdp->removable)
		gd->flags |= GENHD_FL_REMOVABLE;
 gd->queue = sdkp->device->request_queue;

 // #ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_probe(3) : FixMe : gd->queue->req !! \n"));
	/*
	USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_probe() : gd->queue->req = sdkp->sr_request .. \n"));
	gd->queue->req = sdkp->sr_request ;
	*/
 // #endif // NO_ELEVATOR

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_probe() : dev_set_drvdata() .. \n"));
 dev_set_drvdata(dev, sdkp);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_probe() : add_disk() .. \n"));
 if ( add_disk(gd) <=0 ) {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_probe() : Error : add_disk() Failed !! \n"));
	dprintf("[USBMS] Error : Check Disk Failed !! \n") ;
	error = -EIO ;
	goto out_put;
   }

 #ifdef USB_DBGSTRING_ON
	/*
	printk(KERN_NOTICE "Attached scsi %sdisk %s at scsi%d, channel %d, "
	       "id %d, lun %d\n", sdp->removable ? "removable " : "",
	       gd->disk_name, sdp->host->host_no, sdp->channel,
	       sdp->id, sdp->lun);
	*/
	USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_probe() : Attached scsi %sdisk %s at scsi%d, channel %d, "
	       "\n", sdp->removable ? "removable " : "",
	       gd->disk_name, sdp->host->host_no, sdp->channel
	       ));

  #else
	USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_probe() : Attached scsi %sdisk at scsi%d, channel %d, "
	       "\n", sdp->removable ? "removable " : "",
	       sdp->host->host_no, sdp->channel
	       ));

 #endif // ndef USB_DBGSTRING_ON

 USBMS_DBG_INFO(iprintf("[USBMS INFO]              id %d, lun %d\n",
	       sdp->id, sdp->lun));

 // #if defined(USB_DBG_MEM_C_ON) || defined(USB_DBG_MEM_U_ON)
 // #ifdef USBMS_DBG_INIT_ON
 #ifdef USB_DBG_MEM_COUNT_ON
	iprintf("[USB MEM_COUNT] sd_probe(2) ====== show USB Mem Count .. \n"
    	) ;
	showUsbMemC() ;
	showUsbMemU() ;
 #endif // USBMS_DBG_INIT_ON

 #ifndef USBMS_DBG_DISABLE_FATIF
	USBMS_DBG_PLUG(iprintf(
		"[USBMS PLUG] sd_probe() : mount FAT system .. \n"
		)) ;
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] sd_probe() : mount FAT system .. \n"
		));
	dprintf("[USBMS] mount FAT system .. \n");
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_probe() : partition_mount() .. \n"));

#if 0
	USBDiskParam.readIO = (void *)usb_block_read;
	USBDiskParam.writeIO = (void *)usb_block_write;

	USBDiskParam.readBlkIO = (void *)usb_block_readex;

	//USBDiskParam.rdAllIO = (void *);
	USBDiskParam.wrAllIO = (void *)blk_sync_cache;
#endif

	//USBDiskParam.unPlugIO = (void *);
	
	dprintf("[USBMS] total Logical Disk : %d \n",
			validPartitionCnt);
	// registe every partition.  partition_mount() will judge which partition is valid.
	for (j = 0 ; j < validDiskCnt ; j++) {
		for (i = 0 ; i < validPartitionCnt ; i++) {
			USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_probe() : mount %d .. \n", i));
			if (diskMount_status(j) & partitionMask[i]) {
				dprintf("[USBMS] mount partition %d\n", i);
				if (partition_mount(validPartitionID[i], PartitionType[i], &USBDiskParam) < 0) {
					// when mounting fail, disable this config
					diskMount_set_status(j, diskMount_status(j) & (~partitionMask[i]));
					dprintf("[USBMS] partition %d mount fail\n", i);
				} else
					dprintf("[USBMS] partition %d mount success\n", i);
			}
		}
	}

	// update netbios share resource validation
	#if defined(_NB_SHARE) && defined(_NB_FILE)
	updateShareConfPlug(STYPE_DISKTREE, 0);
	#endif // defined(_NB_SHARE) && defined(_NB_FILE)
 #else
	iprintf("[USBMS DISABLE_FATIF] sd_probe() : Fat Interace Disabled , Not Call partition_mount() !! \n") ;
 #endif // ndef USBMS_DBG_DISABLE_FATIF

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_probe() End. \n"));

 return 0;

 out_put:
	put_disk(gd);
 out_free:
 /*
 kfree(sdkp);
 */
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) sdkp
	#if defined(USB_DBG_MEM_C_ON)
		,sizeof(struct scsi_disk)
	#endif
	);

 out:

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_probe() End with Failed !! \n"));
 return error;
}

/**
 *	sd_remove - called whenever a scsi disk (previously recognized by
 *	sd_probe) is detached from the system. It is called (potentially
 *	multiple times) during sd module unload.
 *	@sdp: pointer to mid level scsi device object
 *
 *	Note: this function is invoked from the scsi mid-level.
 *	This function potentially frees up a device name (e.g. /dev/sdc)
 *	that could be re-used by a subsequent sd_probe().
 *	This function is not called when the built-in sd driver is "exit-ed".
 **/
static int sd_remove(struct device *dev)
{
 struct scsi_disk *sdkp = dev_get_drvdata(dev);
 #ifdef USBMS_ONLY_ONE_DISK
	int i ;
 #endif // USBMS_ONLY_ONE_DISK
 struct block_device *bdev;  /* Charles 20040421 */

 // USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_remove() Begin .. \n" )) ;
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] sd_remove() Begin .. \n" )) ;


 #ifdef USBMS_ONLY_ONE_DISK
	for(i=0;i<USBMS_MAX_MINORS+1;i++) {
		UsbDiskOpened[i] = NULL_VALUE ;
		USBMS_DBG_PLUG(iprintf("[USBMS PLUG] sd_remove(1) : Set Value : UsbDiskOpened[%d]=%d\n",
			i,UsbDiskOpened[i]));
		// USBMS_DBG_INIT(iprintf("[USBMS INIT] sd_remove() : Set Value : UsbDiskOpened[%d]=%d\n",
		// 	i,UsbDiskOpened[i]));
	   }
	whole = NULL ;

	for(i=0;i<USBMS_MAX_MINORS+1;i++) {
		if(disk_global_bdev[i] && disk_global_bdev[i]->bd_inode) {
#if 1  /* Charles 20040421 */
			bdev = disk_global_bdev[i]->bd_contains;
			if (bdev && bdev->bd_inode)
			{
				USBMS_DBG_PLUG(iprintf("[USBMS PLUG] sd_remove(2) : destroy_inode() .. \n")) ;
				destroy_inode(bdev->bd_inode) ;
			}
#endif
			USBMS_DBG_PLUG(iprintf("[USBMS PLUG] sd_remove(3) : destroy_inode() .. \n")) ;
			destroy_inode(disk_global_bdev[i]->bd_inode) ;
		   }
		disk_global_bdev[i] = NULL ;
	   }

 #endif // USBMS_ONLY_ONE_DISK



 /*
 // in partitions/check.c
 del_gendisk(sdkp->disk);
 */
 /*
 #ifndef NO_SPINLOCK
	spin_lock(&sd_index_lock);
 #endif // ndef NO_SPINLOCK
 */
 USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_remove() : clear_bit() : sdkp->index=%d \n",
	sdkp->index )) ;
 /*
 clear_bit(sdkp->index, sd_index_bits);
 */
 USBMS_DBG_INFO(iprintf("[USBMS INFO] sd_remove() : FixMe : sdkp->index + 24 !! \n"
	)) ;
 clear_bit(sdkp->index+24, sd_index_bits);

 /*
 #ifndef NO_SPINLOCK
	spin_unlock(&sd_index_lock);
 #endif // ndef NO_SPINLOCK
 */
 sd_shutdown(dev);
 put_disk(sdkp->disk);

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] sd_remove() : UsbFreeMemC(sdkp) , size=%d .. \n" ,
	sizeof(struct scsi_disk)
	));
 /*
 kfree(sdkp);
 */
 UsbFreeMemC(USB_MEM_C_S,(unsigned char *) sdkp
		#if defined(USB_DBG_MEM_C_ON)
			,sizeof(struct scsi_disk)
		#endif
		);

 #ifdef USBMS_ONLY_ONE_DISK
	USBMS_DBG_PLUG(iprintf("[USBMS PLUG] sd_remove(4) : del_gendisk() .. \n" )) ;
	del_gendisk(gd) ;
	USBMS_DBG_PLUG(iprintf("[USBMS PLUG] sd_remove(5) : disk_release() .. \n" )) ;
	disk_release(gd) ;
 #endif // USBMS_ONLY_ONE_DISK

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_remove() End. \n" )) ;
 return 0;
}

/*
 * Send a SYNCHRONIZE CACHE instruction down to the device through
 * the normal SCSI command structure.  Wait for the command to
 * complete.
 */
static void sd_shutdown(struct device *dev)
{
	struct scsi_device *sdp = to_scsi_device(dev);
	struct scsi_disk *sdkp;
	struct scsi_request *sreq;
	int retries, res;

	sdkp = dev_get_drvdata(dev);
	if (!sdkp)
               return;         /* this can happen */

	if (!sdp->online || !sdkp->WCE)
		return;
 #ifdef USB_DBGSTRING_ON
	/*
	printk(KERN_NOTICE "Synchronizing SCSI cache for disk %s: ",
			sdkp->disk->disk_name);
	*/
	dprintf("[USBMS] NOTICE Synchronizing SCSI cache for disk %s: ",
			sdkp->disk->disk_name);
	  #else
	dprintf("[USBMS] NOTICE Synchronizing SCSI cache for disk : "
			);
	#endif // ndef USB_DBGSTRING_ON

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_shutdown() : scsi_allocate_request() .. \n"));
 /*
 sreq = scsi_allocate_request(sdp, GFP_KERNEL);
 */
 sreq = scsi_allocate_request(sdp);
	if (!sreq) {
		printk("FAILED\n  No memory for request\n");
		return;
	}

	sreq->sr_data_direction = DMA_NONE;
	for (retries = 3; retries > 0; --retries) {
		unsigned char cmd[10] = { 0 };

		cmd[0] = SYNCHRONIZE_CACHE;
		/*
		 * Leave the rest of the command zero to indicate
		 * flush everything.
		*/
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] sd_shutdown() : scsi_wait_req() .. \n"));
		scsi_wait_req(sreq, cmd, NULL, 0, SD_TIMEOUT, SD_MAX_RETRIES);
		if (sreq->sr_result == 0)
			break;
	}

	res = sreq->sr_result;
	if (res) {
		/*
		printk(KERN_WARNING "FAILED\n  status = %x, message = %02x, "
				    "host = %d, driver = %02x\n  ",
				    status_byte(res), msg_byte(res),
				    host_byte(res), driver_byte(res));
		*/
		dprintf("[USBMS] FAILED  status = %x, message = %02x, "
				    "host = %d, driver = %02x\n  ",
				    status_byte(res), msg_byte(res),
				    host_byte(res), driver_byte(res));
			if (driver_byte(res) & DRIVER_SENSE)
				print_req_sense("sd", sreq);
	}

	scsi_release_request(sreq);
	printk("\n");
}

/**
 *	init_sd - entry point for this driver (both when built in or when
 *	a module).
 *
 *	Note: this function registers this driver with the scsi mid-level.
 **/
/*
static int __init init_sd(void)
*/
int init_sd(void)
{
 int majors = 0, i;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] init_sd() Begin .. \n"));
 SCSI_LOG_HLQUEUE(3, printk("init_sd: sd driver entry point\n"));

/*
 #ifdef USBMS_ONLY_ONE_DISK
	disk_global_Sector_buf = NULL ;
 #endif // USBMS_ONLY_ONE_DISK
 */ 
 
 USBMS_DBG_MEM(iprintf("[USBMS MEM] init_sd() : disk_global_Sector_buf = UsbAllocMemU() size=%d .. \n",USBMS_SECTOR_SIZE));
 disk_global_Sector_buf = (unsigned char *) UsbAllocMemU(USB_MEM_U_L
			#ifdef USB_DBG_MEM_U_ON
				, USBMS_SECTOR_SIZE * USBMS_BLK_RD_CACHE_LINE_LEN  // size : 8K
		        #endif
           		) ;
 
 #ifdef USBMS_ONLY_ONE_DISK
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] init_sd() : disk_global_Sector_buf=%08XH \n",
		disk_global_Sector_buf
		));
	USBMS_DBG_BLOCK(iprintf(
		"[USBMS BLOCK] init_sd() : disk_global_Sector_buf=%08XH \n",
		disk_global_Sector_buf
		));
	if (!disk_global_Sector_buf) {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] init_sd() End with Failed !! \n"));
		return  -ENODEV;
	   }
 #endif // USBMS_ONLY_ONE_DISK
 USBMS_DBG_INIT(iprintf("[USBMS INIT] sd_init() : disk_global_Sector_buf=%08XH\n", disk_global_Sector_buf));


 #ifdef USBMS_ONLY_ONE_DISK
	for(i=0;i<USBMS_MAX_MINORS+1;i++) {
		disk_global_bdev[i] = NULL ;
	   }

	for(i=0;i<USBMS_MAX_MINORS+1;i++) {
		UsbDiskOpened[i] = NULL_VALUE ;
		USBMS_DBG_INIT(iprintf(
			"[USBMS INIT] init_sd() : Set Value : UsbDiskOpened[%d]=%d\n",
			i,UsbDiskOpened[i]));
	   }
 #endif // USBMS_ONLY_ONE_DISK


 // #ifdef _ETCPIP
 USBMS_DBG_INIT(iprintf("[USBMS INIT] init_sd() : Clear sd_index_bits[] .. \n"));
 for (i = 0; i < (SD_INDEX_SZ); i++) {
	sd_index_bits[i] = 0x00UL;
   }
 // #endif // _ETCPIP

 for (i = 0; i < SD_MAJORS; i++) {
	#ifndef USBMS_ONLY_ONE_DISK
		if (register_blkdev(sd_major(i), "sd") == 0) {
	 #else
		if (register_blkdev("sd") == 0) {
	#endif // USBMS_ONLY_ONE_DISK
			majors++;
		   }
   }

 if (!majors)
		return -ENODEV;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] init_sd() End with Call scsi_register_driver() .. \n"));
 return scsi_register_driver(&sd_template.gendrv);
}

/**
 *	exit_sd - exit point for this driver (when it is a module).
 *
 *	Note: this function unregisters this driver from the scsi mid-level.
 **/
static void __exit exit_sd(void)
{
 int i;

 SCSI_LOG_HLQUEUE(3, printk("exit_sd: exiting sd driver\n"));

 scsi_unregister_driver(&sd_template.gendrv);
 for (i = 0; i < SD_MAJORS; i++) {
	#ifndef USBMS_ONLY_ONE_DISK
		unregister_blkdev(sd_major(i), "sd");
	 #else
		unregister_blkdev("sd");
	#endif // USBMS_ONLY_ONE_DISK
   }
}

/*
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Eric Youngdale");
MODULE_DESCRIPTION("SCSI disk (sd) driver");
*/
/*
module_init(init_sd);
module_exit(exit_sd);
*/

// ====================================================================
// End of File : sd.c

